import sys
sys.path.append('N:\\dev\\pyle\\')
import labrad
from labrad.units import Unit
from labrad.types import Value
from pyle.util import sweeptools as st
from pyle.plotting import dstools as ds
V, GHz, MHz, Hz, dBm, dB, K = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K')]
import numpy as np
import time
from time import sleep

def pna_scan(sample, name = '', freqsweep = st.r[8.3:8.6:0.1, GHz],
             num_points = 800, averages = 1, power = -20*dBm, bandwidth = 100 * Hz):
    
    """Do a PNA Scan for a range of frequencies. The PNA is limited to 16000 points per sweep, so we can break
    long scans into chunks of high resolution. The variable freqsweep tells the PNA how to break up a long scan into chunks.
    For example: if we were to do freqsweep = st.r[8.0:9.0:0.5, GHz] that tells the PNA to sweep from 8.0 to 9.0 GHz in chunks of 500MHz.
    This means that num_points will be the number of points per chunk, with a maximum of 16000.
    
    If averages > 1, a subdirectory will be created where all of the raw data is stored. Once all of the raw data has come in, an averaged
    version of the data will be created and saved in the current directory. 
    """
    
    cxn = sample._cxn
    
    # num_points is the number of points per pna sweep
        
    if sample['port'] == 'Port1':
        independents = ['frequency [GHz]']
        dependents = ['Magnitude (S11) [dB]', 'Phase (S11) [rad]',
                                                   'Magnitude (S21) [dB]', 'Phase (S21) [rad]']
    elif sample['port'] == 'Port2':
        independents = ['frequency [GHz]']
        dependents = ['Magnitude (S22) [dB]', 'Phase (S22) [rad]',
                                                   'Magnitude (S12) [dB]', 'Phase (S12) [rad]']
    
    # get the parameters from the data vault
    parameters = {}
    parameters.update(sample)
    parameters['bandwidth'] = bandwidth
    parameters['freq_start'] = freqsweep[0]
    parameters['freq_stop'] = freqsweep[-1]
    parameters['power'] = power
    parameters['num_points'] = num_points
    startTime = time.time()
    
    def get_temp():
        adr = cxn.adr_server
        adr.select_device(sample['deviceName'])
        temp = float(adr.ruox_status()[0])
        return temp
    
    parameters['temp_start'] = get_temp()
    
    def get_pna_data(start_freq, stop_freq):
        pna.frequency_range(start_freq, stop_freq)
        while True:
            try:
                raw_data = pna.freq_sweep()
            except Exception, e:
                continue
            break	
        f = raw_data[0].asarray
        s = raw_data[1].asarray.T
        f.shape = (len(f), 1)
        pna_data = np.hstack((f, s))
        return pna_data
    
    def getMagPhase(c): #returns magnitude and phase of pna data
        return 2 * 10 * np.log10(abs(c)), np.angle(c)
    
    #pna = cxn.agilent_n5230a_pna
    pna = cxn.pna_x
    pna.select_device(sample['pna_gpib'])
    pna.power(power)
    pna.bandwidth(bandwidth)
    pna.num_points(int(num_points))
    
    if sample['port'] == 'Port1':
        pna.s_parameters(['S11', 'S21'])
    elif sample['port'] == 'Port2':
        pna.s_parameters(['S12', 'S22'])
        
    # set up data vault parameters
    dv = sample._cxn.data_vault
    dv.cd('')
    dv.cd(sample._dir, True)
    if averages > 1:
        nameString = 'raw ' + sample['_datasetname'] + ' ' + name + str(time.time())
        dv.mkdir(nameString)
        dv.cd(nameString)
    else:
        nameString = 'PNA Scan ' + sample['_datasetname'] + ' ' + name
        
    for dummy in range(averages): # save all of the raw data (if averaging)
        dv.new(nameString, independents, dependents)
        p = dv.packet()
        for parm in parameters:
            p.add_parameter(parm, parameters[parm])
        p.send()
        
        for k in range(len(freqsweep)-1):
            
            chunk_start_freq = freqsweep[k]
            chunk_stop_freq = freqsweep[k+1] - freqsweep.range.step/num_points # this ugly subtraction value 
                                                                               # stops the data from being double valued at
                                                                               # the points where the data is stitched together.
                                                                               # it calculates the actual frequency step from the PNA
            
            raw_data = get_pna_data(chunk_start_freq, chunk_stop_freq)
            
            # add live view later
            freqs = raw_data[:,0]*1e-9 # (downconvert to GHz)
            S_through = getMagPhase(raw_data[:,1]) # S12 or S21, depending on PORT
            S_reflected = getMagPhase(raw_data[:,2]) # S11 or S22 depending on PORT
            
            data = np.array(np.vstack((freqs.T, S_reflected[0], S_reflected[1], S_through[0], S_through[1]))).T.astype(float)
            
            dv.add(data)
            try:
                dv.add_parameter('time', startTime)
            except Exception, e:
                continue
            
        
        print 'finished set number ' + str(dummy)
        
    if averages > 1:
    # get all of the averaged data    
        raw_data = [ds.getDataset(dv, m+1) for m in range(averages)] # +1 since the data vault starts at set #1
        
        for i in range(len(raw_data)):
            var = raw_data[i]
            var0 = [x[0] for x in var]	#Frequency
            var1 = [10**(x[1]/20) for x in var] #Magnitude
            var2 = [x[2] for x in var]	#Phase
            var3 = [x[3] for x in var]	#S11
            var4 = [x[4] for x in var] 	#S11
            reData = [var1[k]*np.cos(var2[k]) for k in range(len(var))]
            imData = [var1[k]*np.sin(var2[k]) for k in range(len(var))]
            for j in range(len(var)):
                raw_data[i][j] = [var0[j], reData[j], imData[j], var3[j], var4[j]]
            
        averaged_data = sum(raw_data)/averages
        
        for i in range(len(averaged_data)):
            var = averaged_data[i]
            mData = 20*np.log10((var[1]**2 + var[2]**2)**.5)
            pData = np.arctan2(var[2],var[1])
            averaged_data[i] = [var[0], mData, pData, var[3], var[4]]
        
        # set up data vault parameters
        dv = sample._cxn.data_vault
        dv.cd('')
        dv.cd(sample._dir, True)
        nameString = 'PNA Scan(avg)' + sample['_datasetname'] + ' ' + name
        dv.new(nameString, independents, dependents)
        p = dv.packet()
        for parm in parameters:
            p.add_parameter(parm, parameters[parm])
        p.send()
        dv.add(averaged_data)
        
        currentTime = time.time()
                
        dv.add_parameter('temp_stop', get_temp())
        dv.add_parameter('averages', averages)
        dv.add_parameter('time_elapsed_min', (currentTime - startTime)/60)
    
    