# Utilities for Saving a file to the data Vault continuously

# author:       David Low
# date          120827



import labrad
import time
import davidutil as util

LOGFILE = "datavaultSaver"

class dvsaver(object):
    ''' Parent Class of all data saver objects.
        Changing this will change all functionality
        Never instantiate this class.
        
        All Data Vault saver classes are meant to be used ONLY in a with-as
        structure.  Any other use WILL result in an error.  Note that the 
        class' cxn object is ONLY created inside an __enter__.   PLEASE do not
        try and be clever and instantiate a dvsaver object, then call the 
        __enter__ method.  If you circumevent safety features, you're going to 
        have a bad time.
        
        Please use the following syntax (found in powerSweep)
        
            with dvsaver_temp(resonator folder path) as dvs_temp:
                pass
            pass
            
        This will allow dvsaver to always close and open nicely.  Note that if
        one uses the above code for opening and closing dvsaver objects, you
        will never end a script with open connection objects from dvsaver.  
        Also, the number of connection objects equals the number of dvsaver
        objects open.  
    '''
    dependents = [("Qc", "", ""),
                 ("Qi", "", ""),
                 ("F0", "", "GHz"),
                 ("phi", "", "rad")
                ]
    subpath = []
    
    def __enter__(self):
        ''' Used directly after the constructor in a with-as statement.  
            The only way to use the datavault Saver.
            Creates a cxn connection object and prints to stdout and the log 
            files that we have indeed created data vault objects.  
        '''
        self.__class__.cxn = labrad.connect()
        self.__class__.dv  = self.__class__.cxn.data_vault()
        util.printLog(''.join(["Cxn for data Saver ",
                               str(self.subpath[1]),
                               " connected"]),
                      [LOGFILE], "datavaultSaver -> dvsaver -> __enter__")
        return self

    def __exit__(self, type, value, traceback):
        ''' Used at the conclusion of a with-as statement
            Disconnects the cxn object and prints to stdout and log
        '''
        self.__class__.cxn.disconnect()
        util.printLog(''.join(["Cxn for data Saver ",
                               str(self.subpath[1]),
                               " disconnected"]),
                      [LOGFILE], "datavaultSaver -> dvsaver -> __exit__")
        
    def __init__(self, resfolder):
        ''' Constructor takes the input resfolder path and inserts at
            the beginning of the subpath
        '''
        self.subpath.insert(0, resfolder)

    def startDataSet(self, sample):
        ''' makes a data set.  sample._cxn is NOT used '''
        
        localdv  = self.__class__.dv
        fullpath = sample._dir[:-1] + self.subpath
        print fullpath
        self.gotodir(fullpath)
        
        name = ''.join(["Qc,Qi,F0,phi vs ", 
                        self.__class__.independents[0][0]
                       ])

        dvparam = self.makeParams(sample)
        localdv.new(name, self.__class__.independents, 
                          self.dependents)
        p = localdv.packet()
        
        for param in dvparam:
            p.add_parameter(param, dvparam[param])
        p.send()

    def restartDataSet(self, sample, data):
        ''' restarts a data set if there is an error '''
        isSuccess = -1
        localdv = self.__class__.dv
        fullpath = sample._dir[:-1] + self.subpath
        print fullpath
        self.gotodir(fullpath)
        
        prevData = []
        prevParm = []
        for i in range(10):
            try:
                localdv.open(len(localdv.dir()[1]))
                prevData = localdv.get()
                prevParm = localdv.get_parameters()
                break
            except Exception, e:
                util.printErr(str(e), [LOGFILE], "Restarting Data vault saver")
        # } end of for loop
        
        for i in range(10):
            try:
                self.startDataSet(sample)
                self.addParams(prevParm)
                self.addData(  prevData)
                
                self.addData(data)
                isSuccess = 0
                break
            except Exception, e:
                util.printErr(str(e), [LOGFILE], "Restarting Data vault saver")
        # } end of for loop
        
        return isSuccess
        
        
    def addData(self, data):
        for i in range(10):
            try:
                self.__class__.dv.add(data)
                break
            except Exception, e:
                util.printErr(str(e), [LOGFILE], str(__file__) + ", addData")
    
    def gotodir(self, path):
        ''' navigates the class' dataVault to the correct folder, creating 
            them as it goes.
            This seems to work.  If you suspect something's amiss, 
            return and catch ldv (return ldv; self.__class__.dv = gotodir(path) 
            sort of thing).
        '''
        ldv = self.__class__.dv
        if path[0] == '':
            ldv.cd('')
            path = path[1:]
        for folder in path:
            if folder in ldv.dir()[0]:
                ldv.cd(folder)
            else:
                ldv.mkdir(folder)
                ldv.cd(folder)
        
    def makeParams(self, sample):
        ''' Just the parameters stored in the long term data saving files '''
        dvparams                            = {}
        dvparams['Physical Attenuation']    = sample['physicalAttenuation']
        dvparams['Resonator Number']        = sample['resnum']
        return dvparams
        
    def addParam(self, name, val): 
        listparam = [list(key) for key in self.__class__.dv.get_parameters()]
        listparamkeys = [a[0] for a in listparam]
        if listparamkeys.__contains__(name):
           return
        # } end if
        
        try:
            self.__class__.dv.add_parameter(name, val)
        except Exception, e:
            util.printErr(str(e),[LOGFILE], str(__file__) + ", addParam() ")
    
    def addParams(self, listOfTwoples):
        listOfTwoples = list(listOfTwoples)
        totallist = []
        goodTwoples = []
        totalparam = self.__class__.dv.get_parameters()
        
        for tuple in totalparam:
            totallist.append(tuple[0])
        
        for tuple in listOfTwoples:
            if tuple[0] not in totallist:
                goodTwoples.append(tuple)
                
        try:
            self.__class__.dv.add_parameters(goodTwoples)
        except Exception, e:
            util.printErr(str(e),[LOGFILE], str(__file__) + ", addParam() ")
    

# The following classes are the ones meant to be instantiated.
# If you wish to track a different independent variable, please add another
# child class of dvsaver.  
        
class dvsaver_temp(dvsaver):
    independents = [("Fridge Temperature", "K")]
    
    def __init__(self, resfolder):
        self.subpath = ['temp']
        dvsaver.__init__(self, resfolder)
    
class dvsaver_npho(dvsaver):
    independents = [("Number of Photons", "")]
    
    def __init__(self, resfolder):
        self.subpath = ['numpho']
        dvsaver.__init__(self, resfolder)

class dvsaver_lognpho(dvsaver):
    independents = [("Log10 Number of Photons", "")]
    dependents = [("Log10 Qc", "", ""),
                 ("Log10 Qi", "", ""),
                 ("Log F0", "", "Log10 GHz"),
                 ("Log phi", "", "Log10 rad")]
    
    def __init__(self, resfolder):
        self.subpath = ['lognumpho']
        dvsaver.__init__(self, resfolder)
    
    def _logNestedList(self, list):
        import numpy as np
        logedlist = []
        for i in list:
            tmplist = [np.log10(np.abs(j)) for j in i]
            logedlist.append(tmplist)
        return logedlist
        
    def _logList(self, list):
        import numpy as np
        return [np.log10(np.abs(val)) for val in list]
    

class dvsaver_time(dvsaver):
    independents = [("Time", "s")]
  
    def __init__(self, resfolder):
        self.subpath = ['time']
        dvsaver.__init__(self, resfolder)
        
class dvsaver_curr(dvsaver):
    independents = [("Current", "mA")]
    
    def __init__(self, resfolder):
        self.subpath = ['current']
        dvsaver.__init__(self, resfolder)

# These may or may not work anymore.  Run at your own peril.
def _runtest():
    # sufficient to open two of these classes and write data simultaniously
    THEUSR  = 'David'
    USRROOT = ['',THEUSR]
    TESTPATH = ['tests','tests_powersweep','dataVaultSaver','plots003', 'currentscan']
    from pyle.workflow  import switchSession
    from labrad.units   import dBm
    with labrad.connect() as cxn:
        reg = cxn.registry()
        reg.cd(USRROOT)
        prevsample = reg.get('sample')
        reg.set('sample', TESTPATH)
        sample = switchSession(cxn, user=THEUSR)
        sample['physicalAttenuation'] = -70*dBm
        sample['resnum']              = 1
    
        #print "sample at: " + str(sample._dir)
        reg.cd(sample._dir)
        
        sampledata=[ [8532969.2975348886,
                      132998.03104659007,
                      8205949.1692330446,
                      4982852391.4912453,
                      -0.14526027139063077],
                     [2728030.9528733715,
                      134845.34363265408,
                      7635707.0903478106,
                      4982853081.8583832,
                      -0.17794248476860972],
                     [828199.14802218438,
                      129757.15159382115,
                      6885760.2426854633,
                      4982853416.7055359,
                      -0.14037380748512363],
                     [270574.87452213687,
                      135056.66266433161,
                      5965242.3124016896,
                      4982853531.3600445,
                      -0.13421023968236248],
                     [80147.793016581083,
                      127950.90205766742,
                      4497007.6058881134,
                      4982853536.0524883,
                      -0.14252873928560345],
                     [25866.646990621266,
                      133193.40603193664,
                      3443753.9216545615,
                      4982853661.5289268,
                      -0.1488558680298355],
                     [7859.3690515050121,
                      131714.31010564274,
                      2451127.7975810925,
                      4982853854.7305918,
                      -0.14030495278636421],
                     [2308.8950141773907,
                      128812.08795943637,
                      1587337.011621281,
                      4982853474.4448509,
                      -0.15602979292804653],
                     [715.9377154557701,
                      135963.05164718698,
                      1117060.3558582922,
                      4982853402.0081377,
                      -0.14194059000422823],
                     [196.65112320914875,
                      130677.33831556293,
                      726211.74039046303,
                      4982852820.6951237,
                      -0.14950524032480644],
                     [52.574688460548252,
                      130713.98573187843,
                      461141.33934434404,
                      4982853265.2441521,
                      -0.14924216917334993],
                     [14.265540263428617,
                      131063.8271870139,
                      338303.35266228026,
                      4982853091.0913582,
                      -0.18803630573812483],
                     [4.1199479080741046,
                      130935.00636782584,
                      290269.80287006299,
                      4982853241.1134443,
                      -0.13640484645184189],
                     [1.1969942503298441,
                      134562.37675393841,
                      251657.55432132748,
                      4982855934.2099867,
                      -0.13980327152862368]]
        sampledata2= [0.35205026991137656,
                      132729.84178149025,
                      228652.87703180898,
                      4982853777.5611296,
                      -0.13906501877552144]
        
        
        with dvsaver_npho('Resxx') as dvsnpho:
            with dvsaver_lognpho('Resxx') as dvslognpho:
                dvsnpho.startDataSet(sample)
                dvslognpho.startDataSet(sample)
                
        
                dvsnpho.addData(sampledata)
                dvslognpho.addData(dvslognpho._logNestedList(sampledata))
                
                dvslognpho.addParam("q(-10)",100)
                dvsnpho.addParam("q(-10)", 100)

                
                dvsnpho.restartDataSet(sample, sampledata2)
                dvslognpho.restartDataSet(sample, 
                                        dvslognpho._logList(sampledata2))
                                        
                dvslognpho.addParam("q(-20)",100)
                dvsnpho.addParam("q(-20)", 100)
            
        reg.set('sample', prevsample)
