# Utilities for powerSweep

# author        : david low
# date          : 120827
# unit tested   : 120816, not all tests are still valid

import re
import davidutil as util

from labrad.units   import Hz, dBm

util.safePyle()
from pyle.workflow  import switchSession



def skippower(sample, LOGFILE, source = 'powerSweep -> powerSweep'):
    ''' prints message when one skips power '''
    stout = ''.join(["Skipping power ", str(sample['power']), 
                     defaultMsg(sample)])
    util.printLog(stout, LOGFILE, source)

def skipres(sample, LOGFILE, source = 'powerSweep -> powerSweep'):
    ''' prints message when one skips resonator '''
    stout = ''.join(["Skipping resonator ", str(sample['resnum']),
                      defaultMsg(sample)])
    util.printLog(stout, LOGFILE, source)

def getpnaName(sample):
    ''' returns pna name (just the numbers and the last letter) '''
    m = re.search(r"(\d\d\d\d[A-Za-z])", regget(sample, 'pna'))
    pna = regget(sample,'pna')
    print 'sample',sample,'pna',pna
    return m.group(0)

def writeToVault(dv, dvparams, filename, independents, dependents, alldata):
    ''' writes a data set to the data vault.  Nothing fancy '''
    for i in range(10):
        try:
            dv.new(filename, independents, dependents)
            p = dv.packet()
            for param in dvparams:
                p.add_parameter(param, dvparams[param])
            p.send()
            dv.add(alldata)
            return
        except Exception, e:
            util.printErr(str(e),[LOGFILE],"writetovault")
    
        
def resetSig(sample):
    ''' resets the sample SIGINT variables.  they must be defined first '''
    allkeys = sample.keys()
    for key in allkeys:
        m = re.search(r"^z_SIGINT_\w+$", key)
        if (m != None):
            sample[key] = 0
            
def regget(sample, key):
    ''' gets a key from the registry (NOT the sample, which is where in 
         the registry the sample variable points to) '''
    reg = sample._cxn.registry()
    reg.cd(sample._dir[:-1])
    return reg.get(key)

def regset(sample, key, value):
    ''' sets a key in the registry (NOT the sample) '''
    reg = sample._cxn.registry()
    reg.cd(sample._dir[:-1])
    reg.set(key, value)

def getpowerlist(sample):
    ''' gets the list of powers from the registry '''
    reg = sample._cxn.registry()
    reg.cd(sample._dir[:-1])
    for key in reg.dir()[1]:
        m = re.search(r"[pP]ower[-_][rR]es(\d\d)",key)
        if (m != None and int(m.group(1)) == sample['resnum']):
                return regget(sample,key)
    return regget(sample,'power_default')

def resetSample(sample):
    ''' resets the sample variable. nothing interesting '''
    sample['y_powerindex']   = 0
    sample['power']          = regget(sample,'power_default')[0]
    sample['resnum']         = regget(sample,'reslist')[0]

def defaultMsg(sample):
    ''' fancy print statement for current status of powersweeper '''
    space = '    '
    stout = ''.join(
            ["\n",
             "Status:\n",
             space, "Resonator = ", str(sample['resnum']),              "\n",
             space, "f0        = ", str(sample['f0']),                  "\n",
             space, "VNA Power = ", str(sample['power']),               "\n",
             space, "Q         = ", str(sample['q0']),               "\n",
             space, "qtol      = ", str(sample['x_qtolerance']),          "\n",
             space, "cal scans = ", str(sample['calibrationScanList']), "\n",
             space, "ave scans = ", str(sample['averageScanList']),     "\n"
            ])
    return stout
     
def paramsToString(params):
    return ''.join(["(Qi, f0, Qc, phi) = ",
                    "({0:1.3e}, {1:1.9e}, {2:1.3e}, {3:1.3e})".format(
                                params[0], params[1], params[2], params[3])
                    ])
     
def clearSample_pwr(sample): 
    ''' clear sample in scope of a power '''
    sample['averageScanList']     = []
    sample['calibrationScanList'] = []

     
# unit tests: passed as of 120816
def _runtests():
    import labrad
    with labrad.connect() as cxn:
        reg = cxn.registry()
        reg.cd(['','David'])
        powerSweepDir = reg.get('sample')
        reg.set('sample', ['tests','test_powersweep','powerSweepUtil',
                            'currentscan'])
        sample = switchSession(cxn, user='David')
        
        errors  = 0
        errors += _test_writeToVault(sample)
        errors += _test_resetSig(sample)
        errors += _test_regget(sample)
        errors += _test_getpnaName(sample)
        errors += _test_getpowerlist(sample)
        errors += _test_resetSample(sample)
        errors += _test_defaultMsg(sample)
        errors += _test_skippower(sample)
        errors += _test_skipres(sample)
        
        print "Total Errors: " + str(errors)
        
        reg.cd(['','David'])
        reg.set('sample', powerSweepDir)
     
def _test_skippower(sample):
    skippower(sample, ["testlog"])
    skippower(sample, ["testlog"], "testsource")
    return 0
    
def _test_skipres(sample):
    skipres(sample, ["testlog"])
    skipres(sample, ["testlog"], "testsource")
    return 0

def _test_getpnaName(sample):
    errors = 0
    tests = ["aAkd543kds1234A", "12345 1234A"]
    for test in tests:
        regset(sample,'pna',test)
        errors += int( not getpnaName(sample) == '1234A')
    print "Errors: getpnaName = " + str(errors)
    return errors

def _test_writeToVault(sample):
    errors = 0
    name = "mpthg"
    dv = sample._cxn.data_vault()
    dv.cd(sample._dir)
    testval = [[1,2,3],[4,5,6],[7,8,9]]
    testparam = {"What is your name?": "My name is Sir Lancelot of Camelot",
                 "What is your quest?": "To seek the Holy Grail",
                 "What is your favorite colour?": "Blue"
                 }
    independents = [("independent", "dBm")]
    dependents   = [("day", "legend1", "K"), ("aliens", "virus", "mA")]
    writeToVault(dv, testparam, name,
                 independents, dependents, testval)
    dv.open(len(dv.dir()[1]))
    dataarr = dv.get()
    #alldata = dv.get()
    #dataarr = alldata.asarray()
    errors += int (not dataarr[0] == testval[0])
    errors += int (not dataarr[2] == testval[2])
    #errors += int (dv.get_parameter(testparam.keys()[0]) == 
    #               testparam[testparam.keys()[0]] )
    print "Errors: writeToVault = " + str(errors)
    return errors

def _test_resetSig(sample):
    errors = 0
    sample['z_SIGINT_cleanexit']    = 1
    sample['z_SIGINT_gotores']      = 1
    sample['z_SIGINT_gotopower']    = 1
    sample['z_SIGINT_skippower']    = 1
    sample['z_SIGINT_skipres']      = 1
   
    resetSig(sample)
    errors += int(not sample['z_SIGINT_cleanexit']  == 0)
    errors += int(not sample['z_SIGINT_gotores']    == 0)
    errors += int(not sample['z_SIGINT_gotopower']  == 0)
    errors += int(not sample['z_SIGINT_skippower']  == 0)
    errors += int(not sample['z_SIGINT_skipres']    == 0)
    print "Errors: resetSig: " + str(errors)
    return errors

def _test_regget(sample):
    tests  = {"me": 1, "so": 2, "doe":3}
    errors = 0
    for t in tests:
        regset(sample,t,tests[t])
        errors += int(not regget(sample,t) == tests[t])
    print "Errors: regget and regset = " + str(errors)
    return errors

def _test_getpowerlist(sample):
    errors = 0
    dict = {'power_res02':[2,3,5,7,11,13,17,19,23,29,31,37],
            'Power_Res10':[10,1,1,2,3,5,8,13,21,34,55,89]}
    for k in dict:
        regset(sample, k, dict[k])
        sample['resnum'] = dict[k][0]
        errors += int(not getpowerlist(sample) == dict[k])
    print "Errors: getpowerlist = " + str(errors)
    return errors

def _test_resetSample(sample):
    regset(sample, 'power_default', [1,2,3,4,5])
    regset(sample, 'reslist', [11,12,13,14,15])
    resetSample(sample)
    errors  = 0
    errors += int(not sample['y_resindex'] == 0)
    errors += int(not sample['resnum']     == 11)
    errors += int(not sample['power']      == 1)
    print "Errors: resetSample = " + str(errors)
    return errors

def _test_defaultMsg(sample):
    sample['resnum']              = 1.0
    sample['f0']                  = 2.0*Hz
    sample['power']               = 3.0*dBm
    sample['q0']               = 4.0
    sample['qtolerance']          = 5.0
    sample['calibrationScanList'] = [6,7,8,9,10]
    sample['averageScanList']     = [11,12,13,14,15]
    return 0