'''FluxNoiseAnalysis/echo.py

For analyzing spin echo data

'''

import numpy as np, matplotlib.pyplot as plt
from scipy.optimize import leastsq, curve_fit
from pyle.plotting import dstools
import ramsey, meta, ds, spectroscopy
from units import *

def analyzeEcho(ds, T1=None, plot=True, title=''):
    ''' analyze echo for a dataset '''
    time = ds.data[:,0]
    probs = ds.data[:, 1:]
    freq = ds.parameters['fringeFrequency']['GHz']
    if T1:
        env = ramsey.scaledEnvelope(ds, T1)
    else:
        env = ramsey.makeEnvelope(ds)
    
    fitFunc = lambda v,x: v[0] + v[1]*np.exp(-x/v[2])*np.exp(-(x**2/v[3]))
    vGuess = [0.01, 0.85, 1000.0, 1000.0**2]
    errorFunc = lambda v,x,y: fitFunc(v,x)-y
    result = leastsq(errorFunc, vGuess, args=(time,env), maxfev=10000, full_output=True)
    v = result[0]
    cov=result[1]
    success = result[4]
    #Jim's formula for error bars. Got this from the Resonator fit server.
    rsd = (result[2]['fvec']**2).sum()/(len(time)-len(v))
    bars = np.sqrt(rsd*np.diag(result[1]))
    
    if plot:
        title += ' $T_{\\phi 1} = %.0f\\,ns$, $T_{\\phi 2} = %.0f\\,ns$' % (v[2], np.sqrt(v[3]))
        fig = dstools.plotDataset1D(np.vstack((time,env)).T,[('Time','ns')],
                            [('Envelope Probability', 'Data','')],title=title)
        #fig.show()
        ax = fig.get_axes()[0]
        if T1:
            ax.plot(time, ramsey.makeEnvelope(ds), color='r', label='Before T1 Scale')
        ax.plot(time, fitFunc(v,time),'g',linewidth=4, label='Fit')
        ax.legend(numpoints=1)
        fig.show()
    return {'Tphi1':v[2]*ns, 'dTphi1':bars[2]*ns,
            'Tphi2':np.sqrt(v[3])*ns, 'dTphi2':np.sqrt(bars[3])*ns}
            
def analyzeZset(dv, path, convertZ = True, convert_dfdphi=True, plot=False):
    sets = meta.findDatasets(dv, path, "Spin Echo")
    zs = []
    tphi1s = []
    tphi2s = []
    for s in sets:
        try:
            d = ds.getDataset(dv, s[1], s[0])
            zs.append(float(d.parameters['stack.frame0.arg']['z']))
            r = analyzeEcho(d, plot=plot, title="z = %.3f" % zs[-1])
            tphi1s.append(r['Tphi1']['ns'])
            tphi2s.append(r['Tphi2']['ns'])
            print "did %s" % s[1]
        except ValueError as e:
            # for z = None, lets ignore those datasets
            #print e
            print "skipped %s" % s[1]
            
    if convertZ or convert_dfdphi:
        # now get a Z spectroscopy to convert to freq
        zds = meta.findStarred(dv._cxn, 'Spectroscopy Z func', path)
        d = ds.getDataset(dv, int(zds), path)
        f_p, A, B = spectroscopy.fitSpectroscopy(d, plot=False, quiet=True)
        q = ds.getQubit(d)
        fc_GHz = (q.f10 - q.f21)['GHz']
        def freqFunc(x): return spectroscopy.freqFunc(x, f_p, A, B, f_c)
        zs = [spectroscopy.freqFunc(x, f_p, A, B, fc_GHz) for x in zs]
        xlabel = 'Frequency [GHz]'
        if convert_dfdphi:
            zs = [spectroscopy.dfdphi_f(x, f_p, fc_GHz)**2 for x in zs]
            xlabel = '$ (df / d\Phi_0)^2 $'
    else:
        xlabel = 'z bias [A.U.]'
    
    f = plt.figure()
    ax = f.add_subplot(111)
    ax.plot(zs, tphi1s, 'bo', label='$T_{\phi1}$')
    ax.plot(zs, tphi2s, 'ro', label='$T_{\phi2}$')
    ax.set_xlabel(xlabel, fontsize=19)
    ax.set_ylabel('$T_\phi$ [ns]', fontsize=19)
    ax.legend(numpoints=1)
    ax.set_ylim(0, 20000)
    

