'''

FluxNoiseAnalysis/ramsey.py

For analysis of Ramsey datasets.

'''
import itertools
import numpy as np
from pyle.plotting import dstools
from scipy.optimize import leastsq, curve_fit
import matplotlib.pyplot as plt

from labrad.units import Value, Unit
ns = Unit('ns')
pi = np.pi

import pyle.analysis.ramseyIntegrals as ramseyIntegrals
    
def makeEnvelope(dataset):
    xp = dataset.data[:,1]
    xm = dataset.data[:,3]
    yp = dataset.data[:,2]
    ym = dataset.data[:,4]
    env = np.sqrt((xp-xm)**2+(yp-ym)**2)
    return env
    
def scaledEnvelope(dataset,T1):
    if type(T1) is Value:
        T1 = T1['ns']
    T1 = float(T1)
    t = dataset.data[:,0]
    env = makeEnvelope(dataset)
    envScaled = env/np.exp(-t/(2.0*T1))
    return envScaled
    
def ramseyTomo_noLog(dataset, T1=None, plot=True, title=''):
    """
    Basic Ramsey plotting/fitting. Taken and modified from Dan's original.
    
    T1[ns] -- if given, scale out the 2*T1 exponential decay
    
    Returns dict with keys:
    Tphi1, dTphi1: value and error of white/linear/exponential coefficient
    Tphi2, dTpih2: value and error of 1/f/quadratic/gaussian coefficient
    
    Note that if T1 is not given, then Tphi1 is actually a combination of Tphi1
    and T1, i.e. 1/Tphi_returned = 1/(2*T1) + 1/Tphi
    """
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    freq = dataset.parameters['fringeFrequency']['GHz']
    
    if T1 is None:
        env = makeEnvelope(dataset)
    else:
        env = scaledEnvelope(dataset, T1)


    # parameters v = [offset, amplitude, T_phi1, T_phi2]
    fitFunc = lambda v,x: v[0] + v[1]*np.exp(-x/v[2])*np.exp(-(x/v[3])**2)
    vGuess = [0.02, 0.85, 1000.0, 1000.0]
    errorFunc = lambda v,x,y: fitFunc(v,x)-y
    result = leastsq(errorFunc, vGuess, args=(time,env), maxfev=10000, full_output=True)
    v = result[0]
    cov=result[1]
    success = result[4]
    #Jim's formula for error bars. Got this from the Resonator fit server.
    rsd = (result[2]['fvec']**2).sum()/(len(time)-len(v))
    bars = np.sqrt(rsd*np.diag(result[1]))
    #Plot the fitted curve
    if plot:
        title += ' $T_{\\phi 1} = %.0f\\,ns$, $T_{\\phi 2} = %.0f\\,ns$' % (v[2], v[3])
        fig = dstools.plotDataset1D(np.vstack((time,env)).T,[('Time','ns')],
                            [('Envelope Probability', 'Data','')],title=title)
        #fig.show()
        ax = fig.get_axes()[0]
        if T1:
            ax.plot(time, makeEnvelope(dataset), color='r', label='Before T1 Scale')
        ax.plot(time, fitFunc(v,time),'g',linewidth=4, label='Fit')
        ax.legend(numpoints=1)
        fig.show()
    return {'Tphi1':v[2]*ns, 'dTphi1':bars[2]*ns,
            'Tphi2':v[3]*ns, 'dTphi2':bars[3]*ns}


def fitForS(dataset, T1, alpha=1.0, dfdphi=None, plot=True, fMinTime=3600, title=None):
    ''' Fit a Ramsey dataset to extract S_phi for a given alpha.
    dataset -- must be Ramsey dataset
    T1 -- either a value, or float in ns
    alpha -- float, the alpha to use (default 1.0)
    dfdphi -- frequency [GHz] per flux [Phi_0 (flux quantum)]
              (for unit conversion, generally extracted from Z spectroscopy)
    '''
    # prep arguments
    if dfdphi is None:
        dfdphi = dataset.parameters.dfdPhi0['GHz']
    if isinstance(T1, Value):
        T1 = T1['ns']
    
    env = makeEnvelope(dataset)
    times = dataset.data[:,0]
    # strip out t=0
    env = env[times>0]
    times = times[times>0]
    fMin = 1.0 / (fMinTime * 10**9)     
    fMax = 1e7 / 1e9 # 10 MHz
    integralFunc = ramseyIntegrals.interpForAlpha(alpha, times, fMin, fMax)
    def fitFunc(x, amp, b, offset):
        return amp*np.exp(-x/(2*T1)) * np.exp(-(x**(1.0+alpha))*b*integralFunc(x))+offset
    popt,cov = curve_fit(fitFunc,times,env,[1.0,1e-12,0.02])
    
    a,b,c = popt
    print b
    S = 2.0*(1.0/(2*pi)**2)*b*dfdphi**-2
    #noiseAmplitudes = np.hstack((noiseAmplitudes,np.sqrt(S)))
    #T2s = np.hstack((T2s, 1.0/np.sqrt(b)))
    
    if plot:
        ax = plt.figure().add_subplot(111)
        ax.plot(times, env, '.', label="Envelope Data")
        ax.plot(times, fitFunc(times, *popt), '-', label="Envelope Fit")
        ax.set_xlabel("Time [ns]")
        ax.set_ylabel("Envelope Probability")
        ax.legend(numpoints=1, loc='upper right')
        if title:
            ax.set_title(title)
    return S
    
def fitForAlpha(dataset, T1, S, dfdphi=None, plot=True, fMinTime=3600):
    ''' Fit a Ramsey dataset to extract alpha for a given S_phi
    dataset, T1, dfdphi -- as fitForS
    S -- value, or float in microPhi_0 per root Hz
    '''
    if dfdphi is None:
        dfdphi = dataset.parameters.dfdpPhi0['GHz']
    if isinstance(T1, Value):
        T1 = T1['ns']
    if isinstance(S, Value):
        S = S['uPhi^2/Hz']
    S /= 10**12
    
    env = makeEnvelope(dataset)
    times = dataset.data[:,0]
    # strip out t=0
    env = env[times>0]
    times = times[times>0]
    fMin = 1.0 / (fMinTime * 10**9)     
    fMax = 1e7 / 1e9 # 10 MHz
    
    b = S * (2*pi)**2 / 2.0 * dfdphi**2
    def fitFunc(x, amp, alpha, offset):
        integralFunc = ramseyIntegrals.interpForAlpha(alpha, times, fMin, fMax)
        return amp*np.exp(-x/(2*T1)) * np.exp(-(x**(1.0+alpha))*b*integralFunc(x)) + offset
    popt,cov = curve_fit(fitFunc, times, env, [1.0, 1.0, 0.02])
    
    if plot:
        ax = plt.figure().add_subplot(111)
        ax.plot(times, env, '.', label="Envelope Data")
        ax.plot(times, fitFunc(times, *popt), '-', label="Envelope Fit")
        ax.set_xlabel("Time [ns]")
        ax.set_ylabel("Envelope Probability")
        ax.legend(numpoints=1, loc='upper right')
    return popt[1]
        
