'''

FluxNoiseAnalysis/spectroscopy.py

Code for spectroscopy analysis: Z, flux, and readout resonator spectroscopy.

'''

import numpy as np
import scipy
import matplotlib.pyplot as plt

def freqFunc(zb, f_p, A, B, f_c):
    ''' Xmon qubit freq vs bias (symmetric junctions)
    
    f(z_bias) = f_p * Sqrt[Cos(A*z_bias + B)] - f_c
    '''
    return f_p*np.sqrt(np.abs(np.cos(A*zb + B))) - f_c

def dfdphi(zb, f_p, A, B):
    ''' df/dphi as a function of bias
    the (analytic) derivative of freqFunc--note that f_c has dropped out
    works out to:
    
    df/dphi(zb) = -pi/2 * f_p * Sin[ A*zb + b ] / Sqrt[ Cos( A*zb + B ) ]
    '''
    return -np.pi/2 * f_p * np.sin(A*zb+B)/np.sqrt(np.cos(A*zb+B))
    
def dfdphi_f(f, f_p, f_c):
    ''' df/dphi as a function of f (!!)
    (analytically) solve freqFunc for z_bias(frequency) and insert that into
    df/dphi. note that A, B have dropped out (think about it). comes out to:
    
    df/dphi(f) = -pi/2 * f_p * Sqrt[ (f_p / (f+f_c))^4 - 1]
    '''
    #TODO: check the math, heh
    return -np.pi/2 * f_p * np.sqrt(np.power(f_p/(f+f_c), 4) - 1)

def _getExtrema(data, axisCol, indepCol, valueCol, max=True):
    ''' Gets the value of axisCol where valueCol is max/min, keeping indepCol as independent.
    Useful for reducing a 2D dataset to a 1D one (e.g. flux, Z spectroscopy). '''
    ivals, avals = [], []
    if max:
        f = np.argmax
    else:
        f = np.argmin
    for indepval in np.unique(data[:,indepCol]):
        ivals.append(indepval)
        thisRow = data[data[:,indepCol]==indepval]
        avals.append(thisRow[f(thisRow[:,valueCol]), axisCol])
    return np.array([ivals, avals]).T
    
def fluxSpectroscopy(dataset, plot=True, V_op=None, guessA = 1.0, guessB = 0.0,
                    whichMeasure=0, quiet=False):
    ''' Fits and extracts operating slope from a flux spectroscopy dataset.
    f_10 [GHz] = f_p * Sqrt( Cos(A * V_bias - B) ) - f_c'''
    q = dataset['parameters'][dataset['parameters']['config'][dataset['parameters']['measure'][whichMeasure]]]
    d = _getExtrema(dataset.data, 0,1,2, True)
    fc_GHz = (q.f10 - q.f21)['GHz']
    if V_op is None:
        V_op = q.biasOperate['V']
    
    def fitFunc(v, f_p, A, B):
        return freqFunc(v, f_p, A, B, fc_GHz)
    #def freqFunc(v, f_p, A, B):
    #    return f_p * np.sqrt(np.abs(np.cos(A*v - B))) - fc_GHz
    popt = (d[:,1].max(), guessA, guessB)
    popt, pcov = scipy.optimize.curve_fit(fitFunc, d[:,0], d[:,1], popt)
    f_p, A, B = popt
    dfdv = -f_p*np.sin(A*V_op-B)*A/(2*np.sqrt(np.cos(A*V_op-B)))
    #print popt
    if not quiet:
        print "Period: %.2f V" % (np.pi / A)
        print "df/dv: %.2f GHz/V" % dfdv
        print "df/dphi: %.2f GHz/Phi0" % (dfdv * np.pi/A)
    if plot:
        plt.figure()
        plt.plot(d[:,0], d[:,1], 'bo')
        plt.plot(d[:,0], fitFunc(d[:,0], *popt), 'b-', linewidth=2.0)
        plt.plot(np.linspace(0,1.0,1000), fitFunc(np.linspace(0, 1.0,1000), *popt), 'b-', linewidth=2.0)
        plt.plot(V_op, fitFunc(V_op, *popt), 'ro', markersize=10)
        dpts = np.linspace(V_op-0.1, V_op+0.1)
        plt.plot(dpts, dfdv*(dpts-V_op)+fitFunc(V_op, *popt), 'r--', linewidth=2.0)
        plt.xlabel('Flux Bias [V]', fontsize=19)
        plt.ylabel("Frequency [GHz]", fontsize=19)
    
def fitSpectroscopy(dataset, whichMeasure=0, plot=False, V_op = None, asym=False, 
                    zbMin=None, zbMax=None, returnDacAmp=False,
                    guessA = 1.0, guessB = 0.5, guessC = 0.0, quiet=False):
    ''' Fits a spectroscopy dataset to the qubit model.
    the data should have been taken with the Spectroscopy Z func.
    Qubit model:    f(phi) = f_p * Cos( pi * phi - offset )^(1/2) - f_c
    or:             f(phi) = f_p * Cos( A * V_bias - B )^(1/2) - f_c
    ARGUMENTS:
        dataset
        whichMeasure=0
            which qubit in 'measure' we're working with
            i.e. we use config[measure[whichMeasure]]
            if you only measured one (as you should have), leave whichMeasure=0.
        plot=False
            if True, plot the data + fit
        V_op=None
            if not None, calculate df/dPhi0 at V_op and return it
            
    RETURNS:
        if V_op is None:
            (f_p, A, B) -- from model above
        else:
            dfdPhi0 -- evaluated at V_op
            if returnDacAmp:
                (dfdPhi0, dfdDacAmp) -- df/dDacAmp = df/dphi * A / pi
            
    for an asymmetric squid we have
    I_m^2 = I_1^2 + I_2^2 + 2I_1 I_2 Cos(2*pi*phi / Phi0)
        '''
    if type(dataset) is list:
        d = np.concatenate([x.data for x in dataset])
        dataset = dataset[0]
    else:
        d = dataset.data
    if zbMin is not None:
        d = d[d[:,1] >= zbMin]
    if zbMax is not None:
        d = d[d[:,1] <= zbMax]
    q = dataset['parameters'][dataset['parameters']['config'][dataset['parameters']['measure'][whichMeasure]]]
    # TODO: check for dataset name / scan type?
    # get the maximum frequencies
    zbs = []
    freqs = []
    for zbVal in np.unique(d[:,1]):
        inds = d[:,1] == zbVal
        sd = d[ inds ]
        zbs.append(zbVal)
        freqs.append(sd[ sd[:,2] == sd[:,2].max(), 0][0])
        
    fc_GHz = (q.f10 - q.f21)['GHz']
    if asym:
        def fitFunc(zb, amp, A, B, C):
            return amp * np.power((C + np.cos(A*zb + B)**2), 1.0/4.0) - fc_GHz
        popt = [max(freqs), guessA, guessB, guessC]
    else:
        def fitFunc(zb, amp, A, B):
            return freqFunc(zb, amp, A, B, fc_GHz)
        popt = [max(freqs), guessA, guessB]
    try:
        popt, pcov = scipy.optimize.curve_fit(fitFunc, np.array(zbs), np.array(freqs), popt)
        if asym:
            f_p, A, B, C = popt
            if not quiet:
                print "A = %s, B = %s, f_p = %s, C = %s" % (A, B, f_p, C)
        else:
            f_p, A, B = popt
            if not quiet:
                print "A = %s, B = %s, f_p = %s" % (A, B, f_p)
        fit = True
    except RuntimeError:
        print "Unable to fit"
        fit = False
        f_p, A, B, C, D= None, None, None, None, None
    
    retval = (f_p, A, B)
    if V_op is not None:
        if fit:
            if asym:
                dfdphi = -np.pi/2 * f_p * D * np.cos(A*V_op+B) * np.sin(A*V_op+B) / np.power(C+D*np.cos(A*V_op+B)**2, 3.0/4.0)
            else:
                dfdphi = -np.pi/2 * f_p * np.sin(A*V_op + B) / np.sqrt(np.cos(A*V_op + B))
            dfdDacAmp = dfdphi * A / np.pi
        else:
            idx = (np.abs(np.array(zbs) - V_op)).argmin()
            dfdDacAmp = np.polyfit(zbs[idx-2:idx+3], freqs[idx-2:idx+3], 1)[0]
            dfdphi = 0
        if returnDacAmp:
            retval = dfdphi, dfdDacAmp
        else:
            retval = dfdphi
        if not quiet:
            print "df/dphi = %.2f" % dfdphi
            print "df/dDacAmp = %.2f" % dfdDacAmp
        
    if plot:
        f = plt.figure()
        ax = f.add_subplot(111)
        ax.plot(zbs, freqs, 'bo')
        leg = ['Data']
        if fit:
            maxzb = -B / A
            zbps = np.linspace(min(zbs + [maxzb*1.2]), max(zbs + [maxzb*1.2]))
            ax.plot(zbps, fitFunc(np.array(zbps), *popt), 'g-', linewidth=2)
            #pts = np.linspace(-2,2,1000)
            #ax.plot(fitFunc(pts, *popt), pts, 'b--', linewidth=2)
            leg.append('Fit ($\sqrt{\cos}$)')
        if V_op is not None:
            zbps2 = np.linspace(V_op - 0.1, V_op + 0.1)
            if fit:
                ax.plot(V_op, fitFunc(V_op, *popt), 'ro', markersize=10.0)
                ax.plot(zbps2, dfdphi/np.pi*A * (zbps2 - V_op) + fitFunc(V_op, *popt), 'r-', linewidth = 2)
            else:
                ax.plot(V_op, freqs[idx], 'ro', markersize=10.0)
                ax.plot(zbps2, dfdphi * (zbps2 - V_op) + freqs[idx], 'r-', linewidth = 2)
            
            leg += ['Operating Point', 'Calculated Slope']
        ax.set_ylabel("Frequency [GHz]");
        ax.set_xlabel("Dac Amplitude");
        plt.legend(leg, numpoints=1)

    return retval
    
def _fitAsym(zbs, freqs, fc_GHz):
    def freqFunc(zb, amp, A, B, C, D):
        return amp * np.power((C + D * np.cos(A*zb + B)), 1.0/4.0) - fc_GHz
