''' rbDephasing.py

Analyze randomized benchmarking dataset together with T1, echo, ramsey, etc, to
compare dephasing/decoherence numbers.

Key functions:

Extracting r_c vs gate time for RB data:
[optional kw args white=T/F, telegraph=T/F, correlated=T/F, will turn on fitting
with the given noise model]
orbit_analysis(dvwRef, dvwOrbit)
full_analysis(dvwRef, dvwIdleList)

Comparisons:
orbit_vs_ramsey(dvwRef, dvwOrbit, dvwRamsey)
orbit_vs_freq(dvwRefs, dvwOrbits)

Fitting RB and ramsey [lower level code used by analysis funcs above]:
fitRb(times, r_c, white, telegraph, correlated)
fitRamsey(times, envelope probabilities, white, telegraph, correlated)

'''

import numpy as np, matplotlib.pyplot as plt
from scipy.optimize import curve_fit

import labrad.units as U

import pyle.dataking.benchmarking.randomizedBenchmarking as rb
from pyle.datavault import DataVaultWrapper
import pyle.plotting.dstools as dstools#, pyle.util.dvgrep as dvgrep

def print_fit(popt, pcov, name = ''):
    print name + ' fit params:'
    for x,y in zip(popt, np.sqrt(np.diag(pcov))):
        if y > 1:
            print '%.0f +- %.0f    ' % (x,y), 
        else:
            print '%.5f +- %.5f    ' % (x,y),
    print

def interleaved_rc_single_m(p_ref, f, m, A, B, d=2):
    ''' get r_c from a single m, given a reference value.
    p_ref = p from the reference curve
    f = fidelity for this sequence
    m = num gates in this sequence
    A0,B0 = parameters to use for this sequence
    d = 2**n where n = num qubits '''
    pc = ((f - B)/A)**(1/float(m))
    return (d-1)/float(d) * (1 - pc/p_ref)
    
def orbit_vs_ramsey(dvwRef, dvwOrbit, dvwRamsey):
    r_orbit = orbit_analysis(dvwRef, dvwOrbit, plot=False)
    ramsey = dstools.doAverages(dvwRamsey)
    plot_rb_on_ramsey(r_orbit, ramsey[:,0], ramsey[:,5])
    
def orbit_vs_freq(dvwRefs, dvwOrbits, plot=True, noisy=True):
    rs = []
    f10s = []
    popts = []
    if plot:
        ax = plt.figure().add_subplot(111)
        ax.set_xlabel("%s Length" % dvwOrbits[0].parameters['interleaved'])
        ax.set_ylabel("r_c")
    for i, (ref, orb) in enumerate(zip(dvwRefs, dvwOrbits)):
        f10s.append(ref.parameters[ref.parameters['measure'][0]]['f10'])
        r = orbit_analysis(ref, orb, noisy=False, plot=False)
        if noisy:
            print
            print "F10 = %.2f" % f10s[-1]
        popt = fit_telegraph(r, plot=False, noisy=noisy)
        rs.append(r), popts.append(popt)        
        if plot:
            l = len(plt.rcParams['axes.color_cycle'])
            color = plt.rcParams['axes.color_cycle'][i % l]
            ax.plot(r[:,0], r[:,1], 'o', label="F10 = %.2f data" % f10s[-1], color=color)
            ax.plot(r[:,0], white_and_telegraph_func(r[:,0], *popt), '-', label="F10 = %.2f fit" % f10s[-1], color=color)
    if plot:
        ax.legend(numpoints=1, loc='best')
    return rs, f10s, popts
    
def orbit_analysis(dvwRef, dvwOrbit, A=None, B=None, d=2, noisy=True, plot=True,
                    white=False, telegraph=False, correlated=False, veryNoisy=False):
    ''' analyze an orbit dataset vs idle length. '''
    r, params = rb.reference(dvwRef, A=A, B=B, noisy=False, plot=False)
    if noisy:
        print "From reference: A=%s, B=%s, p=%s" % (params['A'], params['B'], params['p'])
    k = dvwOrbit.parameters['k']
    lengths = np.unique(np.array(dvwOrbit[:,1]))
    rcs = np.zeros_like(lengths)
    for i, l in enumerate(lengths):
        data = dvwOrbit[dvwOrbit[:,1]==l]
        m = data[0,3]
        if not np.all(data[:,3]==m):
            print "ERROR: not all sequences of length=%s have the same m=%s, skipping" % (l, m)
            continue
        if veryNoisy:
            print "for l=%s, m=%s" % (l, m)
        f = np.mean(data[:,2])
        rcs[i] = interleaved_rc_single_m(params['p'], f, m, params['A'], params['B'])
    r = np.vstack([lengths, rcs]).T
    if white or telegraph or correlated:
        popt, pcov = fitRb(r[:,0], r[:,1], white, telegraph, correlated)
    else:
        popt = None
    if plot:
        ax = plt.figure().add_subplot(111)
        ax.set_xlabel("%s Length" % dvwOrbit.parameters['interleaved'])
        ax.set_ylabel("r_c")
        ax.plot(r[:,0], r[:,1], 'o')
        if popt is not None:
            func = getRbFunc(white, telegraph, correlated)
            ax.plot(r[:,0], func(r[:,0], *popt), '-')
    return r[r[:,1] > 0]
    
def rcs_all_ms(dvwRef, dvwInt, noisy=True, d=2, A=None, B=None, plot=True):
    ''' get the r_c for this interleaved dataset, doing each m individually '''
    # do reference
    r, params = rb.reference(dvwRef, A=A, B=B, plot=plot, noisy=False)
    if noisy:
        print "From reference: A=%s, B=%s, p=%s" % (params['A'], params['B'], params['p'])
    # loop over m's
    k = dvwInt.parameters['k']
    
    ms = np.array(np.unique(dvwInt[:,1]))
    rcs = np.zeros_like(ms)
    for i, m in enumerate(ms):
        data = dvwInt[dvwInt[:,1]==m]
        # check that we have all our k's
        if data[:,0].size % k != 0:
            if noisy:
                print "Error for m=%s: expected k=%s, but found %s different k's. Skipping." % (m, k, data[:,0].size)
            continue
        # average the fidelities
        f = np.mean(data[:,2])
        rcs[i] = interleaved_rc_single_m(params['p'], f, m, params['A'], params['B'], d=d)

    r = np.vstack([ms, rcs]).T
    return r[r[:,1] > 0]

def compare_ms_with_full(dvwRef, dvwInt, A=None, B=None, noisy=True, plot=False):
    rcs = rcs_all_ms(dvwRef, dvwInt, A=A, B=B, noisy=noisy, plot=False)
    r, rc, params_ref, params_inter = rb.interleaved(dvwRef, dvwInt, A, B, plot=False, noisy=False)
    if plot:
        ax = plt.figure().add_subplot(111)
        ax.plot(rcs[:,0], rcs[:,1], 'o')
        ax.axhline(rc, linestyle='--')
        ax.set_xlabel('single m used')
        ax.set_ylabel('r_c')
        plt.draw()
        
def analyze_full(dvwRef, dvwInts, A=None, B=None, noisy=True, plot=True, tg_key='identityWaitLen',
                white=False, telegraph=False, correlated=False):
    r, params = rb.reference(dvwRef, A=A, B=B, plot=False, noisy=False)
    A, B = params['A'], params['B']
    if noisy:
        print "From reference: A=%s, B=%s, p=%s" % (A, B, params['p'])
    meas = dvwRef.parameters['measure'][0]
    times, rcs, rc_errs = [], [], []
    for d in dvwInts:
        if d.parameters['measure'][0] != meas:
            raise ValueError("Measure qubits not matched for reference and %s" % str(d))
        t = d.parameters[meas][tg_key]
        r, rc, params_ref, params_inter = rb.interleaved(dvwRef, d, A=A, B=B, plot=False, noisy=False)
        times.append(t['ns'])
        rcs.append(rc)
        rc_errs.append(params_inter['rcerr'])
    times, rcs, rc_errs = np.array(times), np.array(rcs), np.array(rc_errs)
    if white or telegraph or correlated:
        popt, pcov = fitRb(times, rcs, white, telegraph, correlated)
        if noisy:
            print "popt from fit: %s" % popt
    else:
        popt = None
    if plot:
        ax = plt.figure().add_subplot(111)
        ax.set_xlabel('Idle Length [ns]')
        ax.set_ylabel('r_c')
        ax.plot(times, rcs, 'o')
        if popt is not None:
            func = getRbFunc(white, telegraph, correlated)
            ax.plot(times, func(times, *popt), '-')

        plt.draw()
    ret = [times, rcs, rc_errs]
    if white or telegraph or correlated:
        ret += [popt, pcov]
    return ret
    
def plot_all_fits(ref, ints, A=None, B=0.49):
    t, rc, rc_errs, wtp, wtpc = analyze_full(ref, ints, A=A,B=B, noisy=False, plot=False, white=True, telegraph=True)
    t, rc, rc_errs, wp, wpc = analyze_full(ref, ints, A=A,B=B, noisy=False, plot=False, white=True)
    t, rc, rc_errs, wcp, wcpc = analyze_full(ref, ints, A=A,B=B, noisy=False, plot=False, white=True, correlated=True)
    t, rc, rc_errs, tp, tpc = analyze_full(ref, ints, A=A,B=B, noisy=False, plot=False, telegraph=True)
    for name, popt, pcov in zip(['w+t', 'w', 'w+c', 't'], [wtp, wp, wcp, tp], [wtpc, wpc, wcpc, tpc]):
        print_fit(popt, pcov, name)
    wtf, wf, wcf, tf = getRbFunc(True, True, False), getRbFunc(True, False, False), getRbFunc(True, False, True), getRbFunc(False, True, False)
    ax = plt.figure().add_subplot(111)
    ax.set_xlabel('Idle Length [ns]')
    ax.set_ylabel('r_c')
    ax.plot(t, rc, 'o')
    ax.plot(t, wtf(t, *wtp), '-', label="w+t")
    ax.plot(t, wf(t, *wp), '-', label="w")
    ax.plot(t, wcf(t, *wcp), '-', label="w+c")
    ax.plot(t, tf(t, *tp), '-', label="t")
    ax.legend(numpoints=1, loc='best')
    
#################
#### Fitting ####
#################
    
def white_func(tg, tw):
    ''' <phi^2(t)> = 2*t/t_white for white noise. '''
    return 2*tg/tw

def telegraph_func(tg, tt, extra):
    ''' <phi^2(t)> for telegraph noise, see Korotkov "Notes on RB dephasing error" '''
    return (tg-tt*(1-np.exp(-tg/tt)))/extra
    
def correlated_func(tg, tc):
    ''' <phi^2(t)> = (t/t_corr)^2 for correlated (e.g. 1/f) noise '''
    return (tg/tc)**2


def telegraph_quad(tg, tt, extra):
    return tg**2 / (tt*extra)
def telegraph_lin(tg, tt, extra):
    return (tg-tt)/extra
    
def plot_rb_on_ramsey(pts, t_ramsey, p_ramsey, ax=None, color=None, label=None):
    if ax is None:
        ax = plt.figure().add_subplot(111)
        ax.set_xlabel("Ramsey Delay/ORBIT Idle Time [ns]")
        ax.set_ylabel("Ramsey Envelope/Converted orbit error")
        ax.set_title("* = ORBIT, o = Ramsey")
    l1 = ax.plot(t_ramsey, p_ramsey, 'o-', label=label)[0]
    amp = p_ramsey[0]
    l2 = ax.plot(pts[:,0], amp*np.exp(-3*pts[:,1]), '*--')[0]
    if color is not None:
        l1.set_color(color)
        l2.set_color(color)

def getRbFunc(white, telegraph, correlated, squareArgs=False):
    def func(t, *args):
        if squareArgs:
            args = [x**2 for x in args]
        phi = 0
        i = 0
        if white:
            phi += white_func(t, args[i])
            i += 1
        if telegraph:
            phi += telegraph_func(t, args[i], args[i+1])
            i += 2
        if correlated:
            phi += correlated_func(t, args[i])
            i += 1
        return 1./6. * phi
    return func

def getRamseyFunc(white, telegraph, correlated):
    def func(t, A, B, *args):
        inner = 0
        i = 0
        if white:
            inner += white_func(t, args[i])
            i += 1
        if telegraph:
            inner += telegraph_func(t, args[i], args[i+1])
            i += 2
        if correlated:
            inner += correlated_func(t, args[i])
            i += 1
        return A*np.exp(-0.5*inner)+B
    return func

def fitRb(t, err, white=True, telegraph=True, correlated=False):
    ''' returns popt, pcov for a least-sq fit for RB error vs idle length data.
    functional form is err = 1/6 * phi(t)
    where phi is some combo of white, telegraph, and correlated noise.
    '''
    if not(white or telegraph or correlated):
        raise ValueError("Must have at least one noise type!")
    func = getRbFunc(white, telegraph, correlated, squareArgs=True)
    guess = []
    if white:
        guess.append(3000)
    if telegraph:
        guess += [200, 4.1e3]
    if correlated:
        guess.append(2000)
    guess = np.sqrt(np.array(guess))
    popt, pcov = curve_fit(func, t, err, guess, maxfev=int(1e5))
    popt = popt**2
    pcov = pcov**2
    if type(pcov) is float and pcov==float('inf'):
        pcov = np.ones((len(popt), len(popt)))/0.0
    return popt, pcov

def fitRamsey(times, probs, white=True, telegraph=True, correlated=False):
    ''' returns popt, pcov for a least-sq fit for ramsey data.
    functional form is A*np.exp(-0.5 * noise_func(...)) + B
    popt = (A, B, ...) where ... are arguments to noise_func, and will vary
    based on what you're trying to fit. each switch in the arguments (white,
    telegraph, correlated) adds another term to noise func, and the corresponding
    arguments to the fit as well.
    '''
    if not(white or telegraph or correlated):
        raise ValueError("Must have at least one noise type!")
    func = getRamseyFunc(white, telegraph, correlated)
    guess = [0.9, 0.02]
    if white:
        guess.append(3000)
    if telegraph:
        guess += [200, 4.1e3]
    if correlated:
        guess.append(2000)
    popt, pcov = curve_fit(func, times, probs, guess, maxfev=int(1e5))
    return popt, pcov
    
def ramseys(dvw, ramseyIndices, measure = 'q2', plot=True, noisy=True, white=True, telegraph=True, correlated=False,
            ramsey_vs_orbit=False, orbitRefs=None, orbitIdles=None):
    ''' Plot ramseys at different points.
    '''
    datas = {}
    fits = {}
    for x in ramseyIndices:
        d = dvw[x]
        f = round(d.parameters[measure]['f10']['GHz'], 2)
        if f not in datas:
            datas[f] = []
        datas[f].append(d)
    freqs = datas.keys()
    freqs.sort()
    for f in freqs:
        d = datas[f][0]
        for d2 in datas[f][1:]:
            d = np.append(d, d2, axis=0)
        # now average the data
        times = np.unique(d[:,0])
        probs = np.zeros_like(times)
        stds = np.zeros_like(probs)
        for i, t in enumerate(times):
            probs[i] = d[d[:,0]==t,5].mean()
            stds[i] = d[d[:,0]==t,5].std()
        # adjust for T1
        #t1 = findT1(dvw, f*U.GHz)
        #if t1:
        #    probs /= np.exp(-times/(2.0*t1))
        #    stds /= np.exp(-times/(2.0*t1))        
        #else:
        #    print "No T1 found for f=%s" % f
        #    del datas[f]
        #    freqs.remove(f)
        datas[f] = (times, probs, stds)
        popt, pcov = fitRamsey(times, probs, white=white, telegraph=telegraph, correlated=correlated)
        fits[f] = (popt, pcov)
    if noisy:
        for f in freqs:
            popt = fits[f][0]
            print
            print
            print "F10 = %.2f fit params" % f
            print "A=%.2f, B=%.2f" % tuple(popt[:2])
            i = 2
            print "Timescales:"
            if white:
                print "White: %.0f   " % popt[i],
                i += 1
            if telegraph:
                print "Telegraph: %.0f, %.0f   " % (popt[i], popt[i+1]),
                i += 2
            if correlated:
                print "Correlated: %.0f   " % popt[i],
                i += 1
    if plot:
        func = getRamseyFunc(white, telegraph, correlated)
        ax = plt.figure().add_subplot(111)
        for i, f in enumerate(freqs):
            l = len(plt.rcParams['axes.color_cycle'])
            color = plt.rcParams['axes.color_cycle'][i % l]
            ax.plot(datas[f][0], datas[f][1], 'o', label="F10 = %.2f" % f, color=color)
            ax.plot(datas[f][0], func(datas[f][0], *fits[f][0]), '-', color=color)
        ax.legend(numpoints=1, loc='best')
        ax.set_xlabel("Ramsey delay [ns]")
        ax.set_ylabel("Ramsey envelope")
        
    if ramsey_vs_orbit:
        ax = plt.figure().add_subplot(111)
        for i, f in enumerate(freqs):
            l = len(plt.rcParams['axes.color_cycle'])
            color = plt.rcParams['axes.color_cycle'][i % l]
            for ref, idle in zip(orbitRefs, orbitIdles):
                fref = ref.parameters[measure]['f10']
                if np.abs(f-fref) < 0.01:
                    break
            else:
                print "No ORBIT found for f=%s" % f
                continue
            r_orbit = orbit_analysis(ref, idle, plot=False)
            plot_rb_on_ramsey(r_orbit, datas[f][0], datas[f][1], ax=ax, color=color, label="F10=%.2f" % f)
        ax.set_xlabel("Idle Time/Ramsey Delay [ns]")
        ax.set_ylabel("Ramsey envelope / Converted ORBIT data")
        ax.legend(numpoints=1, loc='best')
        
    return datas, fits

#################
##### MISC ######
#################

def analyzeT1(dvwT1, plot=True):
    ''' Fit a T1 dataset, return T1 in ns. '''
    def t1func(t, T1, amp, offset):
        return amp*np.exp(-t/T1) + offset
    popt, pcov = curve_fit(t1func, dvwT1[:,0], dvwT1[:,1], (5000, 1.0, 0.04))
    if plot:
        t = np.linspace(0, dvwT1[:,0].max(), 200)
        ax = plt.figure().add_subplot(111)
        ax.plot(dvwT1[:,0], dvwT1[:,1], '.', label="Data")
        ax.plot(t, t1func(t, *popt), '-', label="Fit, T1 = %.1f us" % (popt[0]/1000,))
        ax.legend(numpoints=1)
        ax.set_xlabel("Time", fontsize=19)
        ax.set_ylabel("Probability", fontsize=19)
    return popt[0]
    
def findT1(dvw, freq, resolution=0.01*U.GHz, plot=False, noisy=False):
    ''' Find a T1 dataset close to this frequency, in this directory,
    and return the T1 (or None if no dataset found) '''
    grep = dvgrep.dvgrep(dvw._cxn.data_vault, 'T1', '/'.join(dvw._dir), [], noisy=False)
    for path, name in grep:
        d = dvw[name]
        q = d.parameters['measure'][0]
        f = d.parameters[q]['f10']
        if np.abs(f['GHz'] - freq['GHz']) < resolution['GHz']:
            if noisy:
                print "Found dataset %s" % name
            return analyzeT1(d, plot=plot)
    return None

def _envelope(d):
    #return np.sqrt((d[:,1] - d[:,3])**2 + (d[:,2] - d[:,4])**2)
    return d[:,5]
def _t1envelope(d, T1):
    return _envelope(d)/np.exp(-d[:,0]/(2.0*T1))

def verifySimilarity(dvw1, dvw2, freqResolution=1*U.MHz):
    ''' check that these datasets have the same config, measure, and qubit
    frequencies (to within freqResolution). '''
    p1,p2 = dvw1.parameters, dvw2.parameters
    if list(p1['config']) != list(p2['config']):
        raise ValueError("%s and %s do not have same config!" % (dvw1.name, dvw2.name))
    if list(p1['measure']) != list(p2['measure']):
        raise ValueError("%s and %s do not have same measure!" % (dvw1.name, dvw2.name))
    for m in list(p1['measure']):
        diff = abs(p1[m]['f10']['MHz']-p2[m]['f10']['MHz'])
        if diff > freqResolution['MHz']:
            raise ValueError("%s f10 differs by %.0f MHz on %s and %s" % (m, diff, dvw1.name, dvw2.name))
    return True
