import numpy as np

def rawRepCodeDataToMeasureOutcome(experimentBits, herald=False):
    """
    Convert raw experimental data into measure outcomes
    experimentBits.shape = (number of total qubits data+measure, stats, cycles+1 (herald taken out))
    For a single stat, experimentBits looks like (herald taken out, see heraldRawRepCodeData)):
    Cycles
            Final meas
            | X 
    X X X X |
            | X
    X X X X |
            | X
            
    returns for single stat, measureOutcomes looks like:
    Cycles
    X X X X X
    X X X X X
    """
    if herald:
        experimentBits = heraldRawRepCodeData(experimentBits)
    
    dataBits = experimentBits[...,-1]
    finalMeasureBits = dataMeasureToStabilizer(dataBits)
    finalMeasureBits = np.expand_dims(finalMeasureBits, -1)
    
    numQubits = experimentBits.shape[0]
    measureQubitIdx = np.arange(1, numQubits, 2)
    
    measureOutcomes = np.concatenate((experimentBits[measureQubitIdx, :, :-1], finalMeasureBits), axis = 2)
    
    return measureOutcomes
    
    
def heraldRawRepCodeData(experimentBits, noisy=False):
    """
    experimentBits.shape = (number of total qubits data+measure, stats, cycles+1+2)
    For a single stat, data looks like:
    
    Herald = 2 retriggers
        Cycles
                  Final meas
    X   |         | X 
      X | X X X X |
    X   |         | X
      X | X X X X |
    X   |         | X
    
    RETURNS cond. on all being zero
            Final meas
            | X 
    X X X X |
            | X
    X X X X |
            | X
    """
    
    oneChannelIndices = []
    for idx, statesOneChannel in enumerate(experimentBits):
        if idx % 2: # idx is odd, its a data qubit
            indices = np.argwhere(statesOneChannel[...,0] == 0) # data qubits are first demod
        else: # idx is even, its a measure qubit
            indices = np.argwhere(statesOneChannel[...,1] == 0) # measure qubits are second demod
        oneChannelIndices.append(np.squeeze(indices))
    allChannelIndices = reduce(np.intersect1d, oneChannelIndices)
    
    stats = experimentBits.shape[1]
    successRate = len(allChannelIndices)/float(stats)
    warningRate = 0.75
    if noisy:
        print "(readout) Herald success rate: %s" % successRate
    elif successRate < warningRate:
        print "(readout) WARNING: Herald success rate: %s" % successRate
    
    heraldedExperimentBits = experimentBits[:, allChannelIndices, 2:]
    return heraldedExperimentBits
    
def dataMeasureToStabilizer(dataBits):
    """
    rawBits.shape = (d+1 measurements of data qubits, stats)
    returns measureBits.shape =  (d with extracted stabilizer outcome, stats)
    """
    numData, stats = dataBits.shape
    measureBits = np.zeros((numData/2, stats), dtype='int')
    
    for sidx in range(stats):
        for idx in range(numData/2):
            outcome0, outcome1 = dataBits[idx*2:idx*2+2, sidx]
            measureBits[idx,sidx] = outcome0 ^ outcome1
    
    return measureBits
    
def measureOutcomeToDetection(measureBits, QND=True):
    """
    This assumes input state of |000> or |111>!
    This is because the emptyMeasure bits have to be generated with an input state in mind.
    measureBits.shape = (measure qubit #, stats, cycle #)
    returns detectionBits
    detectionBits.shape = (measure qubit #, stats, cycle #)
    """
    if QND:
        numQubits, stats, numCycles = measureBits.shape
        detectionBits = np.zeros((numQubits, stats, numCycles), dtype='int')
        # add two empty rounds of detection so first round of detection can resolve
        emptyMeasure = np.zeros((numQubits, stats, 2), dtype='int')
        paddedMeasureBits = np.concatenate((emptyMeasure, measureBits), axis = 2)
        
        for sidx in range(stats):
            for midx in range(numQubits):
                for idx in range(numCycles):
                    bit = paddedMeasureBits[midx, sidx, idx+2]
                    priorBit = paddedMeasureBits[midx, sidx, idx]
                    # if QND, detection event if measurement differs from value 2 rounds prior
                    detectionBits[midx, sidx, idx] = bit ^ priorBit
    else:
        raise NotImplementedError
    return detectionBits
    
def averageDetections(detectionBits):
    return np.mean(detectionBits)