import numpy as np
import scipy.optimize as opt
import matplotlib.pyplot as plt

import labrad
from labrad.units import Unit, Value
MHz,GHz = (Unit(ss) for ss in ['MHz','GHz'])

from pyle.util.structures import ValueArray
from pyle.dataking import util
from pyle.plotting import dstools
import pyle.datavault as datavault

from pyle.dataking import adjust

### UTILITY
def complex_fit(freqs, S, varsGuess, func):
    e = lambda v, f, y: abs(func(f,v)-y)
    popt, pcov, infodict, mesg, ier = opt.leastsq(e, varsGuess, args = (freqs,S), maxfev=1000, full_output=True)
    return popt
    
def optimize(dep, indep, model, guess, maxfev=0):
    """Find parameters to fit a model function to data"""
    def errorFunc(v):
        return abs(model(indep, v) - dep)
    result = opt.leastsq(errorFunc, guess, maxfev=maxfev)
    return result
    
def getLastDataset(sample, use_new_wrapper=False):
    if not use_new_wrapper:
        with labrad.connect() as cxn:
            dv = cxn.data_vault
            dataset = dstools.getDeviceDataset(dv,- 1, session=sample._dir)
        return dataset
    else:
        #Is this a good idea?
        return datavault.DataVaultWrapper(s)[-1]
        
def nonCompactPhase_rad(xs, ys):
    """Get phase angle of a set of x/y points
    
    The phases returned by this function will _not_ lie on a compact set. In
    other words, if the points used as input represent three cycles around a
    circle, the output phases will run from 0 to 3*(2pi), rather than going
    periodically from 0 to 2pi and around again.
    
    I didn't know about numpy.unwrap() when I wrote this.
    """    
    phases_rad = np.array([])
    rotator = 1.0
    lastPhase_rad = 0.0
    for x,y in zip(xs,ys):
        z = x + (1.0j*y)
        dPhase_rad = np.angle(rotator*z)
        phases_rad = np.hstack((phases_rad, lastPhase_rad + dPhase_rad))
        lastPhase_rad = lastPhase_rad + dPhase_rad
        rotator = np.exp(-1.0j*lastPhase_rad)
    return phases_rad
    
def discriminate(S):
    return list(S.imag<0)
    
def fit_circle(x, y):
    """Get the center and radius of a circle
    
    RETURNS
    (xc_2b, yc_2b, R_2b)
     xc_2b: x coordinate of the circle's center
     yc_2b: y coordinate of the circle's center
     R_2b: fitted radius of the circle
     
    I have no idea how this function works - DTS
    But it does indeed work - DTS
    """
    x_m = np.mean(x)
    y_m = np.mean(y)
    def calc_R(xc, yc):
        return np.sqrt((x-xc)**2 + (y-yc)**2)
    def f_2b(c):
        Ri = calc_R(*c)
        return Ri - Ri.mean()
    def Df_2b(c):
        xc, yc     = c
        df2b_dc    = np.empty((len(c), x.size))
        Ri = calc_R(xc, yc)
        df2b_dc[ 0] = (xc - x)/Ri
        df2b_dc[ 1] = (yc - y)/Ri
        df2b_dc       = df2b_dc - df2b_dc.mean(axis=1)[:, np.newaxis]
        return df2b_dc
    center_estimate = x_m, y_m
    center_2b, ier = opt.leastsq(f_2b, center_estimate, Dfun=Df_2b, col_deriv=True)
    xc_2b, yc_2b = center_2b
    Ri_2b        = calc_R(xc_2b, yc_2b)
    R_2b         = Ri_2b.mean()
    residu_2b    = sum((Ri_2b - R_2b)**2)
    residu2_2b   = sum((Ri_2b**2-R_2b**2)**2)
    return xc_2b, yc_2b, R_2b
    
def dipResonatorGuess(freqs_GHz, Sinv):
    """Get rough fit parameters for calibrated dip resonator
    
    Here it is assumed that S inv has the background transmission
    and cable phase delays removed.
    
    I've tested this function and it works pretty damn well. DTS
    
    RETURNS
    f0, Q, Qc, phi_rad
    
    f0 - resonance frequency
    Qi - internal Q factor
    Qc - coupling Q factor
    phi_rad - circle angle offset in radians
    """
    x = Sinv.real
    y = Sinv.imag
    mags = np.sqrt(x**2 + y**2)
    #Find the center and radius of the resonance circle
    xc, yc, r = fit_circle(x,y)
    #Estimate the phi and resonance frequency
    phi_rad = np.arctan2(yc, xc-1)
    f0_GHz = freqs_GHz[np.argmax(abs(mags))]
    #Now find the 3dB points so we can get Q.
    #Translate the circle so that it's center lies on the origin,
    Sinv = Sinv - (xc + 1.0j*yc)
    #rotate it so that the resonance point is on the real axis,
    Sinv = Sinv / np.exp(1.0j*phi_rad)
    #and then find the frequencies at which the y coordinate is maximum
    #and miniumum.
    phases = np.angle(Sinv)
    pos3dBFreq_GHz = freqs_GHz[np.argmax(Sinv.imag)]
    neg3dBFreq_GHz = freqs_GHz[np.argmin(Sinv.imag)]
    #deltaF is the difference between these two frequencies.
    deltaFreq_GHz = abs(pos3dBFreq_GHz - neg3dBFreq_GHz)
    Qi = f0_GHz/deltaFreq_GHz
    Qc = Qi/(2*r)
    return (f0_GHz, Qi, Qc, phi_rad)

### RESPONSE MODELS
    
def dipResonatorModel(freq, f0, Qi, Qc, phi):
    """S21^-1 as a function of frequency and other parameters of a dip resonator
    
    f0 - resonance frequency
    Qi - internal quality factor
    Qc - coupling quality factor
    phi - angular offset of the resonance circle
    
    This model assumes that scattered voltage has been corrected for
    background amplitude and phase. Therefore, the off resonant point
    should be at (1,0) in the complex plane, and the circle should extend
    toward the positive real axis, NOT toward the origin.
    
    NOTE: The scattering convention here is OPPOSITE that used in network
    analyzers. In a network analyzer, the "forward" direction is always
    pointing into the device under test. Therefore, a wave originating
    from port 1, travelling through the device, and then arriving at port
    2 appears to port 2 to be going "backwards". We take the RF signal
    coming out of the system to have the same sign convention as the
    signal going into the system, so we differ from a network analyzer
    by complex conjugation.
    """
    return (1+(Qi/Qc)*np.exp(1.0j*phi)/(1+2.0j*Qi*((freq/f0)-1))).conj()
    
def dipResonatorModel_leastsq(freq, v):
    """ Fit S^-1 with the cable delay and background transmission removed.
    
    This function is designed with the fit parameters as a single array
    for compatibility with scipy.optimize.leastsq.
    
    v[0] - f0
    v[1] - Qi
    v[2] - Qc
    v[3] - phi
    
    The purpose of this function is to wrap up the fit parameters into a
    single array so that the function is compatible with
    scipy.optimize.leastsq. It is just a passthrough to
    dipResonatorModel_calibrated with all nonleading arguments wrapped
    into an array.
    """
    return dipResonatorModel(freq, *v)
    
def parampGainModel(freq, f0, A, Gamma, B, C):
    """Lorentzian function with additional linear background"""
    return (A/np.pi)*(0.5*Gamma)/(((freq-f0)**2)+((0.5*Gamma)**2)) + B*freq + C
    
def parampGainModel_leastsq(freq, v):
    return parampGainModel(freq, *v)
    
def parampPhaseModel(freq, A, f0, width, offset_rad):
    return A*np.arctan((freq-f0)/width) + offset_rad

def parampPhaseModel_leastsq(freq, v):
    return parampPhaseModel(freq, *v)

def purcellResonatorModel(freq, f0, Qi, Ql, phi):
    """Capacitively coupled input dip resonator"""
    return (((Ql/Qi)+2j*Ql*((freq/f0)-1))/(np.exp(1.0j*phi)+2j*Ql*((freq/f0)-1))).conj()

def purcellResonatorModel_leastsq(freq, v):
    return purcellResonatorModel(freq, *v)
    
### DATA CORRECTIONS

def removeBackgroundPhase(q, freqs_GHz, V):
    """Remove RF cable delay phase from a voltage transmission.
    
    PARAMETERS:
    freqs_GHz - array[float]: array of frequencies in GHz
    V - array[complex]: array of transmission amplitudes in complex form,
        eg I + iQ
    """
    p, _ = q['calReadoutRfParams']
    p = (p[0]['rad/GHz'], p[1]['rad'])
    expectedPhase_rad = np.polyval(p, freqs_GHz)
    expectedPhasor = np.exp(1.0j*expectedPhase_rad)
    vCorrected = V/expectedPhasor
    return vCorrected
    
def removeBackgroundAmplitude(q, freqs_GHz, V):
    """Remove background transmission amplitude from transmission data
    
    PARAMETERS
    V - array[complex]: array of transmission amplitudes in complex form
    
    """
    powerFactor = 10**((q['readoutPower']['dBm']-q['calReadoutMagnitudeParams'][1]['dBm'])/10) #Account for different readout power
    timeFactor = (q['adcFilterWindow'][1]-q['adcFilterWindow'][0])/(q['calReadoutMagnitudeParams'][2])
    p = q['calReadoutMagnitudeParams'][0]
    expectedAmp = np.polyval(p, freqs_GHz)*np.sqrt(powerFactor)*timeFactor
    vCorrected = V/expectedAmp
    return vCorrected

def simpleCorrectReadout(q, V):
    '''
    Simple correction for readout.  All this does is rotate by a phase,
    and subtract an imaginary component.  These should be the phase
    and amplitude of (z_pulse + z_nopulse)/2.  The resulting signal
    should be real and centered on zero, with imaginary component being
    only noise.  The problem with this is that it has no correction for 
    readout frequency.  Also, as-is this doesn't scale with readout power.
    '''
    calAngle = q['calReadoutAngle']
    calMag = q['calReadoutOffset']
    Z = V * np.exp(1j*(np.pi/2 - calAngle)) - 1j*calMag
    return Z

def correctParampResponse(q, freq_GHz, V):
    gainParams = np.array(q['calReadoutParampGainFunc'])
    phaseParams = np.array(q['calReadoutParampPhaseFunc'])
    #expectedGain = parampGainModel(freq_GHz, *gainParams)
    #expectedPhaseParamp = parampPhaseModel(freq_GHz, *phaseParams)
    expectedGain = np.polyval(q['calReadoutParampGainFunc'], freq_GHz)
    expectedPhaseParamp = np.polyval(q['calReadoutParampPhaseFunc'], freq_GHz)    
    expectedV = expectedGain * np.exp(1.0j*expectedPhaseParamp)
    return V/expectedV
    
def correctEverything(q, freq_GHz, V, correctParamp=True, debug=False):
    if debug:
        plt.figure()
        plt.plot(V.real,V.imag)
    S = removeBackgroundPhase(q, freq_GHz, V)
    if debug:
        plt.figure()
        plt.plot(S.real,S.imag)
    S = removeBackgroundAmplitude(q, freq_GHz, S)
    if debug:
        plt.figure()
        plt.plot(S.real,S.imag)
    if correctParamp:
        S = correctParampResponse(q, freq_GHz, S)
    if debug:
        plt.figure()
        plt.plot(S.real,S.imag)
    return S
    
### CALIBRATION MEASUREMENTS
    
def getParampParameters(dataset):
    """Gets paramp paramters from data, using previously measured cable corrections"""
    #Get the measured qubit's parameters
    config = dataset['parameters']['config']
    measure = dataset['parameters']['measure'][0]
    measuredQubit = config[measure]
    q = dataset['parameters'][measuredQubit]
    #Parse data
    I = dataset.data[:,1]
    Q = dataset.data[:,2]
    V = I + 1.0j*Q
    freq = ValueArray(dataset.data[:,0],dataset.variables[0][0][1])
    #Correct for cable phase and non-paramp magnitudes
    S = removeBackgroundPhase(q, freq['GHz'], V)
    S = removeBackgroundAmplitude(q, freq['GHz'], S)       #Removed this line Feb2013, makes paramp correction account for _all_ magnitude correction
    mag = np.sqrt(S.real**2 + S.imag**2)
    phase_rad = nonCompactPhase_rad(S.real, S.imag)
    #GAIN PARAMETER FIT
    #Plot curves and ask user for fit ranges
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(freq['GHz'], mag, '.')
    plt.show()
    low = float(raw_input('Low frequency cutoff [GHz]: '))
    high = float(raw_input('High frequency cutoff [GHz]: '))
    #Make array mask and mask data
    mask = np.logical_or(freq['GHz']<low, freq['GHz']>high)
    freqMasked_GHz = freq['GHz'][mask]
    magMasked = mag[mask]
    #Show points being fit
    ax.plot(freqMasked_GHz, magMasked, 'r.')
    plt.show()
    #Guess parameters and do the fit
    idx = np.argmax(magMasked)
    f0Guess_GHz = freqMasked_GHz[idx]
    gammaGuess = 0.01 #TODO: guess this more intelligently
    AGuess = magMasked[idx]*np.pi*gammaGuess/2.0
    guess = [f0Guess_GHz, AGuess, gammaGuess, 0, 0]
    result = optimize(magMasked, freqMasked_GHz, parampGainModel_leastsq, guess)
    gainParams = result[0]
    #Show the fit line
    ax.plot(freq['GHz'], parampGainModel_leastsq(freq['GHz'], gainParams), 'r-')
    plt.show()
    #PHASE PARAMETER FIT
    residualPhase_rad = phase_rad
    residualPhaseMasked_rad = phase_rad[mask]
    #Plot the curve
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(freq['GHz'], residualPhase_rad, 'b.')
    ax.plot(freqMasked_GHz, residualPhaseMasked_rad, 'r.')
    plt.show()
    #Fit
    guess = [3, gainParams[0], 0.005, np.pi] #XXX More intelligent guess for width needed, should be able to get this from gamma in the gain fit
    result = optimize(residualPhaseMasked_rad, freqMasked_GHz, parampPhaseModel_leastsq, guess)
    phaseParams = result[0]
    #Show the fit line
    ax.plot(freq['GHz'], parampPhaseModel_leastsq(freq['GHz'], phaseParams), 'r-')
    return gainParams, phaseParams

def getParampParameters_polynomial(dataset):
    """Gets paramp paramters from data, using previously measured cable corrections"""
    #Get the measured qubit's parameters
    config = dataset['parameters']['config']
    measure = dataset['parameters']['measure'][0]
    measuredQubit = config[measure]
    q = dataset['parameters'][measuredQubit]
    #Parse data
    I = dataset.data[:,1]
    Q = dataset.data[:,2]
    V = I + 1.0j*Q
    freq = ValueArray(dataset.data[:,0],dataset.variables[0][0][1])
    #Correct for cable phase and non-paramp magnitudes
    S = removeBackgroundPhase(q, freq['GHz'], V)
    S = removeBackgroundAmplitude(q, freq['GHz'], S)
    mag = np.sqrt(S.real**2 + S.imag**2)
    phase_rad = nonCompactPhase_rad(S.real, S.imag)
    phase_rad = np.unwrap(phase_rad)
    #GAIN PARAMETER FIT
    #Plot curves and ask user for fit ranges
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(freq['GHz'], mag, '.')
    plt.show()
    def adjustMask(freqs_GHz, mags):
        traces = [{'x': freqs_GHz, 'y': mags, 'args': ('b.',)}]
        params = [{'name': 'Left side start', 'val': freqs_GHz[0], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'b'},
                  {'name': 'Left side end', 'val': freqs_GHz[0], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'r'},
                  {'name': 'Right side start', 'val': freqs_GHz[-1], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'g'},
                  {'name': 'Right side end', 'val': freqs_GHz[-1], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'k'}]
        result = adjust.adjust(params, traces)
        return result
    results = adjustMask(freq['GHz'], mag)
    calStart = Value(results['Left side start'], 'GHz')
    calStop = Value(results['Right side end'], 'GHz')
    dipStart = Value(results['Left side end'])
    dipStop = Value(results['Right side start'])
    #Make array mask and mask data
    mask = np.logical_or(
                         np.logical_and(freq['GHz']>calStart['GHz'],freq['GHz']<dipStart['GHz']),
                         np.logical_and(freq['GHz']>dipStop['GHz'],freq['GHz']<calStop['GHz'])
                        )
    freqMasked_GHz = freq['GHz'][mask]
    magMasked = mag[mask]
    #Show points being fit
    ax.plot(freqMasked_GHz, magMasked, 'r.')
    plt.show()
    #Do the fit
    result = np.polyfit(freqMasked_GHz, magMasked, 1)
    gainParams = result
    #Show the fit line
    ax.plot(freq['GHz'], np.polyval(result, freq['GHz']), 'r-')
    plt.show()
    #PHASE PARAMETER FIT
    residualPhase_rad = phase_rad
    residualPhaseMasked_rad = phase_rad[mask]
    #Plot data
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(freq['GHz'], residualPhase_rad, 'b.')
    ax.plot(freqMasked_GHz, residualPhaseMasked_rad, 'r.')
    plt.show()
    #Fit
    result = np.polyfit(freqMasked_GHz, residualPhaseMasked_rad, 1)
    phaseParams = result
    #Show the fit line
    ax.plot(freq['GHz'], np.polyval(phaseParams, freq['GHz']), 'r-')
    return gainParams, phaseParams
