#Standard packages
import numpy as np
from math import atan2
import scipy.optimize as opt
from scipy.optimize import leastsq
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt

#LabRAD
import labrad
from labrad.units import Unit,Value
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
import pyle.envelopes as env
from pyle.pipeline import returnValue, FutureList
from pyle.dataking import envelopehelpers as eh
from pyle.dataking import measurement
from pyle.dataking import squid
from pyle.dataking import fpgaseqTransmon
from pyle.dataking.fpgaseqTransmon import runQubits as runQubits
from pyle.util import sweeptools as st
from pyle.dataking import utilMultilevels as ml
from pyle.plotting import dstools
from pyle.fitting import fitting
from pyle.util import structures
ValueArray = structures.ValueArray

from pyle import gateCompiler as gc
from pyle import gates
from pyle.gates import PiPulse, Spectroscopy, Wait, Measure, MeasureWait

from pyle.plotting import plothelper as ph

from pyle.dataking import sweeps
from pyle.dataking import util

import readout

"""
def template(sample, measure, name=<default name>,
             <scan ranges with defaults>,
             <sequence options>,
             <other data related options like stats>,
             save=False, collect=False, noisy=True,
             <scripting options like plots>):
    pass
"""

### UTILITY ###

def readDataSafe(s, num):
    dv = s._cxn.data_vault
    dv.cd(s._dir)
    dv.open(num)
    data = []
    while 1:
        dat = dv.get(100)
        if dat != [[]]:
            data.append(dat)
        else:
            break
    data = np.vstack(data)
    return data

def raw2Phase(q, freq_GHz, result):
    I,Q = result
    V = I+1.0j*Q
    S = readout.correctEverything(q, freq_GHz, V)
    return np.angle(S-0.5)
    
def minimaColumns(data):
    """From a 2D dataset, find the minimum of each column of data"""
    xVals = set()
    
    for x in data[:,0]:
        xVals.add(x)
    
    xVals = np.array([x for x in xVals])
    
    yMins = []
    for x in xVals:
        mask = data[:,0]==x
        maskedData = data[mask,1:]
        maskedData = np.column_stack((maskedData[:,0], np.unwrap(maskedData[:,1])))
        idx = np.argmin(maskedData[:,1])
        idxMin = max(0,idx-10)
        idxMax = min(len(maskedData[:,0]), idx+10)
        p = np.polyfit(maskedData[idxMin:idxMax,0],maskedData[idxMin:idxMax,1],2)
        yMin = -p[1]/(2*p[0])
        plt.figure()
        plt.plot(maskedData[:,0],maskedData[:,1])
        plt.plot(maskedData[:,0],np.polyval(p,maskedData[:,0]),'r-')
        yMins.append(yMin)
    yMins = np.hstack(yMins)
    return np.vstack((xVals,yMins)).T

### RESONATOR SCANS

def T1Coherent(sample, measure, time_delay, name="T1 Phase Coherent", stats=100, save=True, collect=False, regParams = None):
    if regParams is None:
        sampleReg, qubits = util.loadQubits(sample, write_access=False)
        q = qubits[measure]
    else:
        sampleReg,q = regParams
    
    axes = [(time_delay, 'Time')]
    deps = [("Amplitude", "I", ""),("Amplitude","Q", "")]
    kw = {'stats': stats}
    q['readout'] = True
    dataset = sweeps.prepDataset(sampleReg, name, axes, deps, measure=measure, kw=kw)
    devs = gc.registryWrapper2GateCompiler(sampleReg, qubits)
    q = devs[measure]

    def func(server, t):
        alg = gc.Algorithm()
        alg[PiPulse([q])]
        alg[Wait([q], t)]
        alg[Measure(devs)]
        alg.compile()
        print t
        data = yield runQubits(server, devs, stats, dataFormat='iq')
        Is,Qs = data[0]
        returnValue(np.vstack((Is,Qs)).T)
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    if collect:
        return data
        
def readoutIqPhaseCoherent(sample, measure, name='Readout IQ',
                           freq=None, power=None, readoutLen=None,
                           piPulse=False,
                           stats=150, dataFormat='iq',
                           save=True, collect=False, noisy=True,
                           plot=True, regParams = None):
    """Clone of readoutIq adapted for coherent paramp pumping
    
    The biggest difference here is that we can't sweep the carrier
    anymore. Therefore, in order to sweep the readout frequency we have
    to change the sideband, keeping the carrier constant. This requires
    different phase calibration.
    
    The other difference is that we make the readout system its own
    device. This shows up only in as the call to
    registryWrapper2GateCompiler(sampleReg, qubits)
    which builds all of the necessary gateCompiler style Agents and
    gives the qubits references to their readout devices, and vice
    versa.
    """
    raise Exception('Depricated. Use functions in singleQubit instead')
    if regParams is None:
        sampleReg, qubits = util.loadQubits(sample, write_access=False)
        q = qubits[measure]
    else:
        sampleReg,qubits = regParams
        q = qubits[measure]
    
    if freq is None:
        freq = q['readoutFrequency']
    if power is None:
        power = q['readoutPower']
    if readoutLen is None:
        readoutLen = q['readoutLen']
        
    axes = [(freq, 'Frequency'),(power,'Power'),(readoutLen,'Readout Length')]
    deps = [('Clicks', 'I for %s' %q.__name__, ''), ('Clicks', 'Q for %s'%q.__name__, '')]
    kw = {'stats': stats, 'power':power}        
    
    dataset = sweeps.prepDataset(sampleReg, name, axes, deps, measure=measure, kw=kw)    
    
    devs = gc.registryWrapper2GateCompiler(sampleReg, qubits)
    q = devs[measure]
    q['readout'] = True
    
    def func(server, f, power, readoutLen):
        q['readoutFrequency'] = f
        q['readoutPower'] = power
        q['readoutLen'] = readoutLen
        
        alg = gc.Algorithm()
        if piPulse:
            alg[PiPulse([q])]
        alg[Measure(devs)]
        alg.compile()
        
        if noisy: print f
        data = yield runQubits(server, devs, stats, dataFormat=dataFormat)
        Is,Qs = data[0]
        returnValue(np.vstack((Is,Qs)).T)
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)     
    if plot:
        plt.figure()
        plt.plot(data[:,1],data[:,2],'.')
        plt.grid()
        for i,freq in enumerate(data[:,0]):
            plt.text(data[i,1],data[i,2],str(data[i,0]))
        plt.xlabel('I')
        plt.ylabel('Q')
    if collect:
        return data
        
def ramseyCoherent(Sample, measure, name='Ramsey Coherent',
           delay=st.r[0:200:1,ns], phase=0.0, fringeFreq = 50*MHz,
           tomo=True,
           stats=600L, randomize=False, averages = 1,
           save = True, collect=False, noisy=True):
    """Ramsey sequence on one qubit. Can be single phase or 4-axis tomo, and
    can have randomized time axis and/or averaging over the time axis multiple
    times
    
    PARAMETERS
    sample: object defining qubits to measure, loaded from registry
    measure - scalar: number of qubit to measure. Only one qubit allowed.
    delay - iterable: time axis
    fringeFreq - value [Mhz]: Desired frequency of Ramsey fringes
    stats - scalar: number of times a point will be measured per iteration over
            the time axis. That the actual number of times a point will be
            measured is stats*averages
    name - string: Name of dataset.
    save - bool: Whether or not to save data to the datavault
    noisy - bool: Whether or not to print out probabilities while the scan runs
    collect - bool: Whether or not to return data to the local scope.
    randomize - bool: Whether or not to randomize the time axis.
    averages - scalar: Number of times to iterate over the time axis.
    tomo - bool: Set True if you want to measure all four tomo axes, False if
                 you only want the X axis (normal Ramsey fringes).
    """
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q = qubits[measure]
    Q = Qubits[measure]
    q['readout'] = True
    #Randomize time axis if you want
    if randomize:
        delay = st.shuffle(delay)
    
    #Generator that produces time delays. Iterates over the list of delays as many times as specified by averages.
    delay_gen = st.averagedScan(delay,averages,noisy=noisy)
    
    axes = [(delay_gen(), 'Delay'),(phase, 'Phase')]
    #If you want XY state tomography then we use all four pi/2 pulse phases
    if tomo:
        deps = [('Phase', '+X', 'rad'),('Phase', '+Y', 'rad'),
                ('Phase', '-X', 'rad'),('Phase', '-Y', 'rad')]
        tomoPhases = {'+X': 0.0, '+Y': 0.25, '-X': -0.5, '-Y': -0.25} #[+X, +Y, -X, -Y] in CYCLES
    #Otherwise we only do a final pi/2 pulse about the +X axis.
    else:
        deps = [('Phase', '', '')]
        tomoPhases = {'+X': 0.0}
    
    kw = {'averages': averages, 'stats': stats, 'fringeFrequency': fringeFreq}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    dt = q['piFWHM']
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    def func(server, delay, phase):
        reqs = []
        for tomoPhase in tomoPhases:
            alg = gc.Algorithm()
            alg[gates.PiHalfPulse([q], phase = 0.0)]
            alg[gates.Wait([q],delay)]
            alg[gates.PiHalfPulse([q], phase = 2*np.pi*(fringeFreq['GHz']*delay['ns'] + tomoPhases[tomoPhase]))]
            alg[gates.Measure(devs)]
            alg.compile()
            reqs.append(runQubits(server, devs, stats, dataFormat='iq'))
        results = yield FutureList(reqs)
        data = [r[0] for r in results]
        phases = [raw2Phase(q, q['readoutFrequency']['GHz'], dat) for dat in data]
        returnValue(phases)
    #Run the scan and save data
    data = sweeps.grid(func, axes, dataset = save and dataset, collect=collect, noisy=noisy)
    if collect:
        return data
    
### QUIBIT PULSE TUNEUP

def lollipopPhaseCoherent(Sample, measure, name='Lollipop',
                          stats=300, freq=None, power=None,
                          pumpAmplitude=None,
                          dataFormat='iq',
                          plot=True, save=True):
    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    if freq is None:
        freq = q['readoutFrequency']
    if power is None:
        power = q['readoutPower']
    if pumpAmplitude is None:
        pumpAmplitude = q['readoutDevice']['pumpAmplitude']
        
    if not any( [np.iterable(x) for x in [freq, power, pumpAmplitude] ]):
        freq = [freq]
        
    axes = [(freq, 'Frequency'), (power, 'Power'), (pumpAmplitude, 'pumpAmplitude')]
    deps = [('I', 'I no pulse', 'clicks'), ('Q', 'Q no pulse', 'clicks'),
            ('I', 'I with pulse', 'clicks'), ('Q', 'Q with pulse', 'clicks')]
    kw = {'stats':stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    

    #q = gc.AttrDict2Qubit(q, gc.Transmon)
    q['readout'] = True
    
    def func(server, freq, power, pumpAmp):
        q['readoutPower'] = power
        q['readoutFrequency'] = freq
        q['readoutDevice']['pumpAmplitude'] = pumpAmp
        reqs = []
        print "readout with power %s freq %s pumpamp %s" % (power, freq, pumpAmp)
        #Measure with no pi pulse
        alg = gc.Algorithm()
        alg[Measure(devs)]
        alg.compile()
        reqs.append(runQubits(server, devs, stats, dataFormat=dataFormat))
        #Measure with pi pulse
        alg = gc.Algorithm()
        alg[PiPulse([q])]
        alg[Measure(devs)]
        alg.compile()
        reqs.append(runQubits(server, devs, stats, dataFormat=dataFormat))
        #Run the experiment
        data = yield FutureList(reqs)
        noPulse, pulse = data
        InoPulse, QnoPulse = noPulse[0]
        IPulse, QPulse = pulse[0]
        returnValue(np.vstack((InoPulse,QnoPulse,IPulse,QPulse)).T)
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)        
    
    vNoPulse = data[:,1]+1.0j*data[:,2]
    SNoPulse = readout.correctEverything(q, data[:,0], vNoPulse)
    vPulse = data[:,3]+1.0j*data[:,4]
    SPulse = readout.correctEverything(q, data[:,0], vPulse)
    
    #Plots
    if plot:
        plt.figure()
        plt.title('Corrected IQ data')
        plt.plot(SNoPulse.real, SNoPulse.imag, label='|0>')
        plt.plot(SPulse.real, SPulse.imag, label='|1>')
        plt.legend(loc='upper right')
        plt.figure()
        plt.title('abs[S] (corrected)')
        plt.plot(data[:,0],abs(SNoPulse))
        plt.plot(data[:,0],abs(SPulse))
        plt.figure()
        plt.title('Absolute separation (corrected)')
        plt.plot(data[:,0],abs(SPulse-SNoPulse))
    return data
    
def readoutTimeTracePhaseCoherent(Sample, measure,
                                  freq=None, power=None,
                                  piPulse=False,
                                  averages=150, stats=150,
                                  collect=False,
                                  plot=True, regParams=None):
    server = Sample._cxn.qubit_sequencer
    if regParams is None:
        sample, qubits = util.loadQubits(Sample)
    else:
        sample, qubits = regParams
    q = qubits[measure]
    if freq is None:
        freq = q['readoutFrequency']
    if power is None:
        power = q['readoutPower']
    
    if averages%stats != 0:
        raise Exception("Cannot do %d averages with %d stats"%(averages, stats))
        
    repetitions = averages//stats

    q['readout'] = True    
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]

    
    q['readoutFrequency'] = freq
    q['readoutPower'] = power
    q['adc mode'] = 'average'
    ISum = np.zeros(8192) #Number of samples in an ADC time trace, should not be hardcoded!
    QSum = np.zeros(8192)
    
    for repetition in range(repetitions):
        alg = gc.Algorithm()
        if piPulse:
            alg[PiPulse([q])]
        alg[Measure(devs)]
        alg.compile()
         
        data = runQubits(server, devs, stats, dataFormat='iqRaw', debug=True).wait()
        Iraw = data[0][0].asarray
        Qraw = data[0][1].asarray
        ISum += Iraw
        QSum += Qraw
        
    if plot:
        z = ISum+1.0j*QSum
        ft = np.fft.fft(z)
        fftFreqs = np.fft.fftfreq(8192, 2)
        plt.figure()
        plt.semilogy(fftFreqs, np.abs(ft)**2)
        def rms(x):
            return np.sqrt(np.sum(x**2)/len(x))
        numClippedI = sum(ISum>252) + sum(ISum<-254)
        numClippedQ = sum(QSum>252) + sum(QSum<-254)
        fig  =  plt.figure()
        descriptor = "I: rms=%d, clipped=%d, Q: rms=%d clipped=%d"%(rms(ISum), numClippedI, rms(QSum), numClippedQ)
        print descriptor
        plt.title('Normalized counts - power=%s'%q['readoutPower'])
        ax = fig.add_subplot(111)
        ax.plot(ISum/float(averages), 'b', linewidth=1)
        ax.plot(QSum/float(averages), 'r', linewidth=1)
    if collect:
        return ISum,QSum
    
### MULTIPLEXED READOUT ###
    
def readoutTimeTrace_multiQubit(Sample, name='Readout time trace',
                                averages=150, stats=150,
                                piPulseQ0 = False, piPulseQ1 = False,
                                collect=False, noisy=True,
                                regParams=None):
    server = Sample._cxn.qubit_sequencer
    if regParams is None:
        sampleReg, qubits = util.loadQubits(Sample)
    else:
        sampleReg, qubits = regParams
    
    if averages%stats != 0:
        raise Exception("Cannot do %d averages with %d stats"%(averages, stats))
    
    repetitions = averages//stats
    
    #Only one qubit need be read out as they share the same time trace.
    #The readout DAC signal is added so long as qubit['rr'] is set to
    #something. Setting qubit['readout'] only adds the qubit to the
    #timing order. 2013 Feb 21 DTS/EJ
    qubits[0]['readout']=True
    for q in qubits:
        q['adc mode'] = 'average'
    qubits = [gc.AttrDict2Qubit(q, gc.Transmon) for q in qubits]
    
    dataSummed = np.zeros((8192,2))
    
    for repetition in range(repetitions):
        alg = gc.Algorithm()
        if piPulseQ0:
            alg[PiPulse([qubits[0]])]
        if piPulseQ1:
            alg[PiPulse([qubits[1]])]
        alg[Measure(qubits)]
        alg.compile()
        
        data = runQubits(server, qubits, stats, dataFormat='iqRaw').wait()
        data = np.vstack((data[0][i] for i in range(2))).T
        dataSummed += data
    return dataSummed
    
def iqVsTime_multiQubit(Sample, name='Two qubit IQ vs Time',
                        averages=1, stats=1, repetitions=20,
                        filterBw=None, downsampleFactor=10, timeMax_ns=17000):
    filterPts=512
    sampleReg, qubits = util.loadQubits(Sample)
    
    indeps = [('Repetition', ''), ('State',''), ('Time', 'ns')]
    deps = []
    for qName in ["First Q", "Second Q"]:
        deps.extend([
                     ('Amplitude', 'I, %s'%qName, ''),
                     ('Amplitude', 'Q, %s'%qName, '')
                     ])
    kw = {'stats':stats, 'averages':averages, 'filterBw':filterBw, 'downsampleFactor':downsampleFactor, 'filterPts':filterPts}
    dataset = sweeps.prepDataset(sampleReg, name, axes=[],
              explicitIndependents=indeps, dependents=deps,
              measure=None, kw=kw, cxn=Sample._cxn)
    
    STATES = {'00':(False,False),
              '01':(False,True),
              '10':(True,False),
              '11':(True,True)}
    
    with dataset:
        for repetition in range(repetitions):
            for state in STATES:
                data = readoutTimeTrace_multiQubit(Sample, averages=averages, stats=stats,
                                                   piPulseQ0=STATES[state][0], piPulseQ1=STATES[state][1],
                                                   collect=True, regParams = (sampleReg, qubits))
                zs = []
                t = np.linspace(0, (8192*2)-2, 8192)
                repetitionVector = np.zeros(8192)+repetition
                stateVector = np.zeros(8192)+int(state)
                downsampler = slice(0, -1, downsampleFactor)
                for q in qubits:
                    sbFreq = q['readoutFrequency'] - q['readout fc']
                    phasor = np.exp(1j*2*np.pi*sbFreq['GHz']*2*np.arange(len(t)))
                    z = phasor * (data[:,0]+1.0j*data[:,1])
                    if filterBw:
                        lpfilter = spsignal.firwin(filterPts, (filterBw/(250*MHz))[''])
                        z = np.convolve(z, lpfilter, 'same')
                    zs.append(z)
                dataset.add(np.column_stack((
                                       repetitionVector[downsampler],
                                       t[downsampler],
                                       stateVector[downsampler],
                                       zs[0].real[downsampler],
                                       zs[0].imag[downsampler],
                                       zs[1].real[downsampler],
                                       zs[1].imag[downsampler])
                                       ))
    
def acStarkShiftPhaseCoherent(Sample, measure, name='AC Stark Shift Phase Coherent',
                 freq=None, amplitudeSquared=None, excitationLen=800*ns,
                 stats=150,
                 save=True, collect=False, noisy=True):
    print("THIS FUNCTION USES A HACK TO GET THE RESONATOR EXCITATION PULSE")
    print("Make sure the excitationLen is smaller than PREPAD")
    print("PREPAD: %d ns"%fpgaseqTransmon.PREPAD)
    print("excitationLen: %s"%excitationLen)
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    
    if freq is None:
        freq = st.r[q['f10']['GHz']-0.1:q['f10']['GHz']+0.02:0.002,GHz]
    #if power is None:
    #    power = st.r[-40:-30:1,dBm]
    if amplitudeSquared is None:
        amplitudeSquared = np.arange(0.0001,1,0.05)
    
    axes = [(amplitudeSquared, 'dac amplitude squared'),(freq, 'Frequency')]
    deps = [('Phase','','rad')]
    
    refPower = q['readoutPower']
    refOffsets = q['readoutIqOffset']

    kw = {'stats': stats, 'refPower':refPower}        
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    q['readout'] = True
    
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    def func(server, ampSquared, f):
        df = q['readoutFrequency']-q['readoutDevice']['carrierFrequency']
        q['f10'] = f
        #amp = eh.power2amp(pwr)
        amp = np.sqrt(ampSquared)
        alg = gc.Algorithm()
        alg[PiPulse([q], alpha=0)]
        alg[Measure(devs)]
        alg.compile()
        #Add a resonator ringup pulse
        q['rr'] += env.mix(env.flattop(-excitationLen, excitationLen, 20, amp=amp), df)
        data  = yield runQubits(server, devs, stats, dataFormat='iq')
        phase = raw2Phase(q, q['readoutFrequency']['GHz'], data[0])
        if phase < -2: # hack to avoid grapher screwing up the phase plot: put the angle() branch cut where we want it
            phase += 2*np.pi
        print phase
        returnValue([phase])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
        
### Real time readout scans ###
# Moving stuff into singleQubit
def iqVsTimePhaseCoherent(Sample, measure, name='IQ vs time',
                          freq=None, power=None,
                          averages=1, stats=1,
                          timeMax_ns=17000,
                          save=True, collect=False, noisy=True, regParams=None):
    """A time trace of the readout signal IQ with and without an excitation pulse
    
    There is no demodulation, downsampling, for filtering.
    """
    if regParams is None:
        sampleReg, qubits = util.loadQubits(Sample, write_access=False)
    else:
        sampleReg, qubits = regParams
    q = qubits[measure]
    if freq is not None:
        q['readoutFrequency'] = freq
    if power is not None:
        q['readoutPower'] = power
    
    indeps = [('Time','ns')]
    deps=[('Amplitude','I no pulse',''),
          ('Amplitude','Q no pulse',''),
          ('Amplitude','I with pulse',''),
          ('Amplitude','Q with pulse','')]
    kw = {'stats':stats, 'averages':averages}
    dataset = sweeps.prepDataset(sampleReg, name, axes=[], explicitIndependents = indeps, dependents=deps, measure=measure, kw=kw, cxn=Sample._cxn)  
    
    INoPulse,QNoPulse = readoutTimeTracePhaseCoherent(Sample, measure,
                                         freq=freq, power=power,
                                         piPulse=False,
                                         averages=averages, stats=stats,
                                         collect=True,
                                         plot=False, regParams=(sampleReg,qubits))
    IPulse,QPulse = readoutTimeTracePhaseCoherent(Sample, measure,
                                         freq=freq, power=power,
                                         piPulse=True,
                                         averages=averages, stats=stats,
                                         collect=True,
                                         plot=False, regParams=(sampleReg,qubits))
    t = np.linspace(0,(8192*2)-2,8192)
    tMask = t<timeMax_ns
    data = np.vstack((
                      t[tMask],
                      INoPulse[tMask],
                      QNoPulse[tMask],
                      IPulse[tMask],
                      QPulse[tMask])).T
    with dataset:
        dataset.add(data)
    if collect:
        return data
    
def iqVsTime2D(Sample, measure, name='IQ vs time',
               freq=None, power=None, sbFreq=None,
               averages=1, stats=1, repetitions=20,
               filterBw=None, downsampleFactor=10, timeMax_ns=17000,
               collect=False, noisy=True):
    """IQ vs time traces"""
    filterPts = 512
    sampleReg, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    if freq is not None:
        q['readoutFrequency'] = freq
    if power is not None:
        q['readoutPower'] = power
    if sbFreq is None:
        sbFreq = q['readoutFrequency'] - q['readout fc']
    
    indeps = [('Repetition', ''), ('Time','ns')]
    deps=[('Amplitude','I no pulse',''),
          ('Amplitude','Q no pulse',''),
          ('Amplitude','I with pulse',''),
          ('Amplitude','Q with pulse','')]
    kw = {'stats':stats, 'averages':averages, 'filterBw':filterBw, 'downsampleFactor':downsampleFactor,'filterPts':filterPts}
    dataset = sweeps.prepDataset(sampleReg, name, axes=[], explicitIndependents = indeps, dependents=deps, measure=measure, kw=kw, cxn=Sample._cxn)  
    #Loop over repetitions
    with dataset:
        for repetition in range(repetitions):
            INoPulse,QNoPulse = readoutTimeTrace(Sample, measure,
                                                 freq=freq, power=power, sbFreq=sbFreq,
                                                 piPulse=False,
                                                 averages=averages, stats=stats,
                                                 collect=True,
                                                 plot=False, regParams=(sampleReg,qubits))
            IPulse,QPulse = readoutTimeTrace(Sample, measure,
                                                 freq=freq, power=power, sbFreq=sbFreq,
                                                 piPulse=True,
                                                 averages=averages, stats=stats,
                                                 collect=True,
                                                 plot=False, regParams=(sampleReg,qubits))
            #Go to DC, filtering out idler
            phasor = np.exp(1j*2*np.pi*sbFreq['GHz']*2*np.arange(len(IPulse))) #Factor of 2 for 2ns sampler interval on A/D board.
            t = np.linspace(0, (8192*2)-1, 8192)
            zNoPulse = phasor * (INoPulse + 1j*QNoPulse)
            zPulse = phasor * (IPulse + 1j*QPulse)
            if filterBw:
                lpfilter = spsignal.firwin(filterPts, (filterBw/(250*MHz))[''])
                zNoPulse = np.convolve(zNoPulse, lpfilter, 'same')
                zPulse = np.convolve(zPulse, lpfilter, 'same')
            INoPulse,QNoPulse = zNoPulse.real,zNoPulse.imag
            IPulse,QPulse = zPulse.real, zPulse.imag
            repetitionVector = np.zeros(8192)+repetition
            downsampler = slice(0,-1,downsampleFactor)
            tMask = t<timeMax_ns
            dataset.add(np.vstack((
                                   repetitionVector[tMask][downsampler],
                                   t[tMask][downsampler],
                                   INoPulse[tMask][downsampler],
                                   QNoPulse[tMask][downsampler],
                                   IPulse[tMask][downsampler],
                                   QPulse[tMask][downsampler])).T)
    
def iqVsTime2DPhaseCoherent(Sample, measure, name='IQ vs time Phase Coherent',
                            freq=None, power=None, sbFreq=None,
                            averages=1, stats=1, repetitions=20,
                            filterBw=None, downmix=False, downsampleFactor=1, timeMax_ns=2000,
                            regParams=None):
    """IQ vs time traces"""
    filterPts = 512
    if regParams is None:
        sampleReg, qubits = util.loadQubits(Sample, write_access=False)
    else:
        sampleReg, qubits = regParams
    q = qubits[measure]
    if freq is not None:
        q['readoutFrequency'] = freq
    if power is not None:
        q['readoutPower'] = power
    
    indeps = [('Repetition', ''), ('Time','ns')]
    deps=[('Amplitude','I no pulse',''),
          ('Amplitude','Q no pulse',''),
          ('Amplitude','I with pulse',''),
          ('Amplitude','Q with pulse','')]
    kw = {'stats':stats, 'averages':averages, 'repeitions': repetitions, 'filterBw':filterBw, 'downsampleFactor':downsampleFactor,'filterPts':filterPts}
    dataset = sweeps.prepDataset(sampleReg, name, axes=[], explicitIndependents = indeps, dependents=deps, measure=measure, kw=kw, cxn=Sample._cxn)  
    #Loop over repetitions
    with dataset:
        for repetition in range(repetitions):
            INoPulse,QNoPulse = readoutTimeTracePhaseCoherent(Sample, measure,
                                                              freq=freq, power=power,
                                                              piPulse=False,
                                                              averages=averages, stats=stats,
                                                              collect=True,
                                                              plot=False, regParams=(sampleReg,qubits))
            IPulse,QPulse = readoutTimeTracePhaseCoherent(Sample, measure,
                                                          freq=freq, power=power,
                                                          piPulse=True,
                                                          averages=averages, stats=stats,
                                                          collect=True,
                                                          plot=False, regParams=(sampleReg,qubits))
            t = np.linspace(0, (8192*2)-2, 8192)
            if downmix or filterBw:
                raise Exception('This is ported from the old, phase incoherent version. Needs checking')
                if sbFreq is None:
                    sbFreq = q['readoutFrequency'] - q['readout fc']
                #Go to DC, filtering out idler
                phasor = np.exp(1j*2*np.pi*sbFreq['GHz']*2*np.arange(len(IPulse))) #Factor of 2 for 2ns sampler interval on A/D board.
                zNoPulse = phasor * (INoPulse + 1j*QNoPulse)
                zPulse = phasor * (IPulse + 1j*QPulse)
                if filterBw:
                    lpfilter = spsignal.firwin(filterPts, (filterBw/(250*MHz))[''])
                    zNoPulse = np.convolve(zNoPulse, lpfilter, 'same')
                    zPulse = np.convolve(zPulse, lpfilter, 'same')
                INoPulse,QNoPulse = zNoPulse.real,zNoPulse.imag
                IPulse,QPulse = zPulse.real, zPulse.imag
            repetitionVector = np.zeros(8192)+repetition
            downsampler = slice(0,None,downsampleFactor)
            tMask = t<timeMax_ns
            dataset.add(np.vstack((
                                   repetitionVector[tMask][downsampler],
                                   t[tMask][downsampler],
                                   INoPulse[tMask][downsampler],
                                   QNoPulse[tMask][downsampler],
                                   IPulse[tMask][downsampler],
                                   QPulse[tMask][downsampler])).T)
    
def iqVsPowerTime(Sample, measure, name='IQ traces vs power',
                  freq=None, powers=None, sbFreq=-50.0*MHz,
                  averages=1, stats=1, iterations=100, downsampleFactor=10,
                  save=True, collect=False, noisy=True):
    sampleReg, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    if freq is not None:
        q['readoutFrequency'] = freq
    if powers is None:
        raise Exception('Choose power scan')
    
    indeps = [('Power', 'dBm'), ('Iteration',''), ('Time','ns')]
    deps=[('Amplitude','I, no pulse','clicks'),
          ('Amplitude','Q, no pulse','clicks'),
          ('Amplitude','I, pulse','clicks'),
          ('Amplitude','Q, pulse','clicks')]
    kw={'stats':stats,'averages':averages}
    idlerFreq = q['parAmpPumpFreq'] - q['readoutFrequency']
    idlerSbFreq = idlerFreq - q['readout fc']    
    dataset = sweeps.prepDataset(sampleReg, name, axes=[], explicitIndependents = indeps, dependents=deps, measure=measure, kw=kw, cxn=Sample._cxn)  
    #Loop over repetitions
    with dataset:
        for power in powers:
            print(power)
            for iteration in range(iterations):
                INoPulse,QNoPulse = readoutTimeTrace(Sample, measure,
                                                     freq=freq, power=power, sbFreq=sbFreq,
                                                     piPulse=False,
                                                     averages=averages, stats=stats,
                                                     save=False, collect=True,
                                                     plot=False, regParams=(sampleReg,qubits))
                IPulse,QPulse = readoutTimeTrace(Sample, measure,
                                                     freq=freq, power=power, sbFreq=sbFreq,
                                                     piPulse=True,
                                                     averages=averages, stats=stats,
                                                     save=False, collect=True,
                                                     plot=False, regParams=(sampleReg,qubits))
                phasor = np.exp(1j*2*np.pi*sbFreq['GHz']*2*np.arange(len(IPulse)))

                idler_phasor = np.exp(1j*2*np.pi*idlerSbFreq['GHz']*2*np.arange(len(IPulse)))
                filter_bw = 2*MHz
                lpfilter = spsignal.firwin(512, (filter_bw/(250*MHz))[''])
                t = np.linspace(0, (8192*2)-1, 8192)
                zNoPulse = np.convolve(phasor * (INoPulse + 1j*QNoPulse), lpfilter, 'same')
                zPulse = np.convolve(phasor * (IPulse + 1j*QPulse), lpfilter, 'same')
                INoPulse,QNoPulse = zNoPulse.real,zNoPulse.imag
                IPulse,QPulse = zPulse.real, zPulse.imag
                powerVector = np.zeros(8192)+power['dBm']
                iterationVector = np.zeros(8192)+iteration
                downsampler = slice(0,-1,downsampleFactor)
                data = np.vstack((powerVector[downsampler],
                                  iterationVector[downsampler],
                                  t[downsampler],
                                  INoPulse[downsampler],
                                  QNoPulse[downsampler],
                                  IPulse[downsampler],
                                  QPulse[downsampler])).T
                dataset.add(data)

                
def signalIdlerResponse(Sample, measure,  name='Signal / Idler Response',
                          signal_freq=None, power=None,
                          averages=1, stats=1,
                          timeMax_ns=17000,
                          save=True, collect=False, noisy=True, regParams=None):
    """A time trace of the readout signal IQ with and without an excitation pulse
    
    There is no demodulation, downsampling, for filtering.
    """
    if regParams is None:
        sampleReg, qubits = util.loadQubits(Sample, write_access=False)
    else:
        sampleReg, qubits = regParams
    q = qubits[measure]
    if signal_freq is not None:
        q['readoutFrequency'] = signal_freq 
    else:
        signal_freq = q['readoutFrequency']
    idler_freq = sampleReg[q['readoutConfig']]['pumpFrequency'] - signal_freq

    if power is not None:
        q['readoutPower'] = power
    
    indeps = [('Time','ns')]
    deps=[('Amplitude','I Cos[w_signal t]',''),
          ('Amplitude','Q Cos[w_signal t]',''),
          ('Amplitude','I Sin[w_signal t]',''),
          ('Amplitude','I Sin[w_signal t]',''),
          ('Amplitude','I Cos[w_idler t]',''),
          ('Amplitude','Q Cos[w_idler t]',''),
          ('Amplitude','I Sin[w_idler t]',''),
          ('Amplitude','I Sin[w_idler t]','')]
          
    kw = {'stats':stats, 'averages':averages}
    dataset = sweeps.prepDataset(sampleReg, name, axes=[], explicitIndependents = indeps, dependents=deps, measure=measure, kw=kw, cxn=Sample._cxn)   
    q['readoutPhaseShift'] = 0
    I1,Q1 = readoutTimeTracePhaseCoherent(Sample, measure,
                                         freq=signal_freq, power=power,
                                         piPulse=False,
                                         averages=averages, stats=stats,
                                         collect=True,
                                         plot=False, regParams=(sampleReg,qubits))
    q['readoutPhaseShift'] = np.pi/2
    I2,Q2 = readoutTimeTracePhaseCoherent(Sample, measure,
                                         freq=signal_freq, power=power,
                                         piPulse=False,
                                         averages=averages, stats=stats,
                                         collect=True,
                                         plot=False, regParams=(sampleReg,qubits))
    q['readoutPhaseShift'] = 0    
    I3,Q3 = readoutTimeTracePhaseCoherent(Sample, measure,
                                         freq=idler_freq, power=power,
                                         piPulse=False,
                                         averages=averages, stats=stats,
                                         collect=True,
                                         plot=False, regParams=(sampleReg,qubits))
    q['readoutPhaseShift'] = np.pi/2 
    I4,Q4 = readoutTimeTracePhaseCoherent(Sample, measure,
                                         freq=idler_freq, power=power,
                                         piPulse=False,
                                         averages=averages, stats=stats,
                                         collect=True,
                                         plot=False, regParams=(sampleReg,qubits))                                         
    t = np.linspace(0,(8192*2)-2,8192)
    tMask = t<timeMax_ns
    data = np.vstack((
                      t[tMask],
                      I1[tMask],
                      Q1[tMask],
                      I2[tMask],
                      Q2[tMask],
                      I3[tMask],
                      Q3[tMask],
                      I4[tMask],
                      Q4[tMask])).T
    with dataset:
        dataset.add(data)
    if collect:
        return data
                
def phaseTraceVsPower(data, start, stop):
    timeMask = np.logical_and(data[:,2]<stop,data[:,2]>start)
    data = data[timeMask,:]
    powers=set()
    repetitions=set()
    
    for p in data[:,0]:
        powers.add(p)
    for repetition in data[:,1]:
        repetitions.add(repetition)
    for power in powers:
        plt.figure()
        plt.title('%s dBm'%power)
        mask = data[:,0]==power
        dataAtPower = data[mask,1:]
        for repetition in repetitions:
            mask = dataAtPower[:,0]==repetition
            dataAtIteration = dataAtPower[mask,1:]
            
            t = dataAtIteration[:,0]
            phaseNoPulse = np.unwrap(dataAtIteration[:,1])
            phasePulse = np.unwrap(dataAtIteration[:,2])
            plt.plot(t*np.cos(phaseNoPulse),t*np.sin(phaseNoPulse),'b-')
            plt.plot(t*np.cos(phasePulse),t*np.sin(phasePulse),'r-')

def iqVsTime2DVsPhase(sample, measure, base_name='IQ vs time',
                averages=1, stats=1, repetitions=20, timeMax_ns=2000):
    """
    Take time trace datasets while varying the readout phase.
    """
    sampleReg, qubits = util.loadQubits(sample, write_access=False)
    q = qubits[measure]
    phases = np.arange(8)*np.pi/8
    for phi in phases:
        q['readoutPhaseShift'] = phi
        ds_name = base_name + " phi=%.2f pi" % (phi/np.pi)
        iqVsTime2DPhaseCoherent(sample, measure, ds_name, averages=averages,
            stats=stats, repetitions=repetitions, timeMax_ns=timeMax_ns,
            regParams = (sampleReg, qubits))

def paramp_gain_saturation(sample, measure, power=None):
    '''
    Measure paramp gain saturation
    '''
    if power is None:
        power = ValueArray(np.arange(-50,-25,1), dBm)
    lollipopPhaseCoherent(sample, measure, stats=600, name="Paramp Gain Saturation", 
        dataFormat='iq', save=True, power=power)

def paramp_gain_vs_pump(sample, measure, pumpAmplitude=None):
    '''
    Measure paramp gain vs. pump amplitude.  This should look roughly parabolic,
    with two solutions for any given gain.  The lower pump amplitude choice has
    less noise.
    '''
    if pumpAmplitude is None:
        pumpAmplitude = np.arange(.5,1.0, 0.01)
    lollipopPhaseCoherent(sample, measure, stats=600, name="Paramp gain vs. pump", 
        dataFormat='iq', save=True, pumpAmplitude=pumpAmplitude)
        
def dealWithIt(data, start, stop):
    repetitions = set()
    
    figNoPulse = plt.figure()
    plt.title('Readout signal phase vs time for $|0 \\rangle $')
    plt.xlabel('Time [ns]')
    plt.ylabel('Phase [rad]')
    
    figPulse = plt.figure()
    plt.title('Readout signal phase vs time for $|1 \\rangle $')
    plt.xlabel('Time [ns]')
    plt.ylabel('Phase [rad]')
    
    figBoth = plt.figure()
    plt.title('Readout signal phase vs time')
    plt.xlabel('Time [ns]')
    plt.ylabel('Phase [rad]')
    
#    figAll = plt.figure()
#    plt.grid()
#    plt.title('Readout signal phase vs time')
#    ax0 = figAll.add_subplot(3,1,0)
#    ax1 = figAll.add_subplot(3,1,1)
#    ax10 = figAll.add_subplot(3,1,2)
#    for ax in [ax0, ax1, ax10]:
#        ax.set_xlabel('Time [ns]')
#        ax.set_ylabel('Phase [rad]')
    
    for repetition in data[:,0]:
        repetitions.add(repetition)
    for repetition in repetitions:
        mask = data[:,0]==repetition
        dataMasked = data[mask,1:]
        dataMasked = dataMasked[start:stop,:]
        
        plt.figure(figNoPulse.number)
        plt.plot(dataMasked[:,0],np.unwrap(dataMasked[:,1]),'b-')
        
        plt.figure(figPulse.number)
        plt.plot(dataMasked[:,0],np.unwrap(dataMasked[:,2]),'r-')

        plt.figure(figBoth.number)
        plt.plot(dataMasked[:,0],np.unwrap(dataMasked[:,1]),'b-', label='$|0 \\rangle $')
        plt.plot(dataMasked[:,0],np.unwrap(dataMasked[:,2]),'r-', label='$|1 \\rangle $')
        
#        plt.figure(figAll.number)
#        ax0.plot(dataMasked[:,0],np.unwrap(dataMasked[:,1]),'b-', label='$|0 \\rangle $')
#        ax1.plot(dataMasked[:,0],np.unwrap(dataMasked[:,2]),'r-', label='$|1 \\rangle $')
#        ax10.plot(dataMasked[:,0],np.unwrap(dataMasked[:,1]),'b-', label='$|0 \\rangle $')
#        ax10.plot(dataMasked[:,0],np.unwrap(dataMasked[:,2]),'r-', label='$|1 \\rangle $')

### PHASE COHERENT READOUT ###

