import numpy as np
import scipy.optimize

import labrad
from labrad.units import Unit,Value
ns,MHz = (Unit(s) for s in ['ns','MHz'])

from pyle.util.structures import ValueArray
from pyle.dataking import util
from pyle.dataking import transmon as tr
from pyle.plotting import dstools

import matplotlib.pyplot as plt


#CHANGELOG

#Nov 10 2012
#At this point I will assume all IF frequencies are fixed for a given qubit.
#This allows me to deal only with the RF cable delay. I believe this is a
#useful first step, and possibly necessary given the variability in
#propagation time through the IF differential amplifiers as a function of
#IF frequency.

#How to process the data

#Hardware demod
#We get an IQ value of the transmitted voltage wave, ie V. Our goal is to put this on
#a resonance circle. We have to deal with the fact that we have phase from the cables,
#and we have to deal with the fact that we have nonunity background transmission.
#Let's allow the resonator model to ONLY describe the resonance itself, ie. assumes
#unity transmission and phase=0 off resonance. We'll deal with the phase and
#magnitude separately.
#
# Things that affect the transmission data:
# a. cable RF delay
# b. Resonance circle itself
# c. Paramp if it's there
#
#CALIBRATION
# 1. Measure RF phase from cabling. This gives a slope and a y-intercept
#    point. This is independent of everything else in the experiment. These
#    values allow you to completely remove RF cable phase for a fixed IF.
# 2. Measure background cable transmission. The cabling is now done in the
#    sense that at your current power, any measurement with no interesting
#    features will have no frequency dependance. If you keep track of the
#    power at which you measured this background, you can also change
#    readout power.
# 3. Measure the resonance dip without the paramp in town. This tells
#    you things like Qi, Qc, f0, and that phi parameter.
# 4. Put the paramp where you want it and measure transmission. Remove
#    the cable phase and amplitude to get the paramp phase and amplitude.
#    Make sure you only fit data not inside the resonator.
#    
#CORRECTION
# 1. Remove cable phase and amplitude.
# 2. Remove paramp.
#


def resonanceDip(sample, measure, freq, power=None, name='Readout resonator fit'):
    sampleReg, qubits, Qubits = util.loadQubits(sample, write_access=True)
    qubit = qubits[measure]
    Qubit = Qubits[measure]
    if power is not None:
        qubit['readout power'] = power
    data = tr.readoutIq(sample, freq=freq, sb_freq=-50*MHz,
                        measure=measure, stats=1200, dataFormat='iq',
                        save=True, name=name, collect=True, plot=False,
                        regParams=(sampleReg, qubit))
    dataset = tr.getLastDataset(sample)
    freqUnit = dataset.variables[0][0][1]
    frequencies = ValueArray(dataset.data[:,0],freqUnit)
    V = dataset.data[:,1] + 1.0j*dataset.data[:,2]
    S = removeRfPhase(qubit, frequencies['GHz'], V)
    S = removeBackgroundAmplitude(qubit, frequencies['GHz'], S)
    fig = plt.figure()
    plt.plot(S.real,S.imag,'.')
    #Go from S to S^-1
    Sinv = S**(-1.0)
    #Find guesses for fit parameters
    guess = dipResonatorGuess_calibrated(frequencies['GHz'], Sinv)
    #Fit the data
    fitResult = optimize(freq_GHz, Sinv, dipResonatorModel_calibrated_leastsq, guess)
    f0, qi, qc, phi = fitResult[0]
    Q['calReadoutResonatorParams'] = (f0, qi, qc, phi)
    