#!/usr/bin/env python

import numpy as np
import time
import scipy
from scipy.optimize import leastsq,curve_fit
from scipy.special import erf, erfc
import matplotlib
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt

from labrad.units import Unit,Value
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]

import pyle.envelopes as env
from pyle.pipeline import returnValue, FutureList
from pyle.dataking import envelopehelpers as eh
from pyle.dataking import measurement
from pyle.dataking import squid
from pyle.dataking.fpgaseqTransmon import runQubits as runQubits
from pyle.util import sweeptools as st
from math import atan2
from pyle.dataking import utilMultilevels as ml
from pyle.plotting import dstools
from pyle.sim import pyleSim as ps
from pyle.fitting import fitting
from pyle.util import structures
from pyle.plotting import dstools as ds
import pyle.dataking.util as util
from pyle.sim import pyleSim as ps

from pyle import gateCompiler as gc
from pyle import gates
from pyle import tomo
import pyle.quantum as quantum

def dualBlockTest(Sample, measure, delay=500*ns, dual_block=True):
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    q['readout'] = True
    q['adc mode'] = 'demodulate'    
    alg = gc.Algorithm()
    alg[gates.PiPulse([q])]
    if dual_block:
        alg[gates.DualBlockWaitDetune(devs, q, len=delay, w=15*ns, amp=0.5)]
    else:
        alg[gates.Detune([q],len=delay,w=15*ns,amp=0.5)]     
    alg[Measure(devs)]
    alg.compile()
    rv = runQubits(Sample._cxn.qubit_sequencer, devs, stats=100, dataFormat='iqRaw', debug=True)
    return rv.wait()

def qubitDecay(sample, tDelay=structures.ValueArray(np.linspace(0, 4000, 40), ns),
               name='Tomography decay', save=False): 

    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    qubits = gc.AttrDict2Qubit(qubits, gc.Transmon)
    q, Q = qubits[measure], Qubits[measure]
    
    opList = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, 1)) # 1 qubit only
    
    axes = [(tDelay, 'Delay')]
    deps = [('Bloch X', '', ''),('Bloch Y', '', ''),('Bloch Z', '', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    rrCharacteristics,rrFreq,rrS21,rrS21fit = getrrCharacteristics(q)
   
    sb_freq_ro=rrCharacteristics[8]    
    
    fread = optimalReadoutFrequency(q,includeXi=True)
    q['readout'] = True    
    q['readout frequency']=fread
    if not val.sbSweep:    
        q['readout fc'] = fread - sb_freq_ro     
    
    def func(server, tDelay):
        
        probs = []
        
        for op in opList:
            alg = gc.Algorithm()
            alg[gates.PiPulse([q])]
            alg[gates.TestDelay(tDelay)]
            alg[gates.Tomography([q], op)]
            alg[gates.Measure([q])]
            alg.compile()    
            
            preprocess([q])
            ans = yield FutureList([runQubits(server, qubits, stats, dataFormat='iq')])
            Probs,Probsstd,Smean = getProbs([q],stats,ans,[[fread]])
            
            probs.append(Probs[0])
         
            # get rho
            
            # map rho onto bloch sphere
        [rho, residual] = tomo.qst_mle_evan(np.array(probs), tomo.octomo_ops, stats, 1)
        S = quantum.rho2bloch(rho)
        returnValue([x for x in S[1:4]])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
def T1StartDelay(Sample, measure, delay = st.r[50:150e3:1e3, ns], name="T1", stats=1200, 
       save=True, collect=False, state=1):
    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    axes = [(delay, 'Time')]
    deps = [("Probabiliy |%s>" % state,'','')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)

    def func(server, t):
        alg = gc.Algorithm()
        alg[gates.MoveToState([q], 0, state)]
        if t<1e3*ns:
            alg[gates.Wait([q], t)]
        else:
            start_delay = (t['ns'] //4)*4*ns
            wait_delay = t - start_delay
            alg[gates.Wait([q], wait_delay)]
            q['readoutDevice']['readoutStartDelay'] = start_delay
        alg[gates.Measure(devs)]
        alg.compile()
        print t
        data = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = np.mean(sq.readoutProbabilities(q,np.array(Is)+1.0j*np.array(Qs)))
        
        returnValue([prob])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    if collect:
        return data

def ramsey_LF(Sample, measure, name='Ramsey LF Noise', delay=500*ns,
    phase = np.pi/2, interval=1000*us, stats=30000, averages=1, save=False, collect=True):

    """Low frequency phase noise estimator based on Ramsey interference.
    
    PARAMETERS
    sample: object defining qubits to measure, loaded from registry
    measure - scalar: number of qubit to measure. Only one qubit allowed.
    delay - scalar: Ramsey delay.  This should be large, but less than T2*
    phase - scalar: The ramsey phase. pi/2 should give the best sensitivity.
    interval - scalar: The measurement interval.  THis sets the frequency 
                resolution for the FFT
    stats - scalar: number of times a point will be measured per iteration over
            the time axis.
    name - string: Name of dataset.
    save - bool: Whether or not to save data to the datavault
    collect - bool: Whether or not to return data to the local scope.
    averages - scalar: Number of sequence blocks to run
    """
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    q['biasOperateSettle'] = interval    

    axes = [(np.arange(repetitions), 'Repetition')]
    deps = [('I', 'I', ''), ('Q', 'Q', '')]
    kw = {'averages': averages, 'stats': stats, 'delay':delay, 'interval': interval, 'phase':phase}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw,explicitIndependents=[('Time','ns')])
    def func(server, rep):
        reqs = []

        alg = gc.Algorithm()
        alg[gates.PiHalfPulse([q], phase = 0.0)]
        alg[gates.Wait([q], delay)]
        alg[gates.PiHalfPulse([q], phase = phase)]
        alg[gates.Measure(devs)]
        alg.compile()
        reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        results = yield FutureList(reqs)
        Is,Qs = results[0][0]
        ts = np.arange(len(Is))*interval['ns']
        returnValue(np.column_stack(ts, Is, Qs))
    #Run the scan and save data
    data = sweeps.grid(func, axes, dataset = save and dataset, collect=collect, noisy=noisy)
    if collect:
        return data

def blochTrajectory(sample, theta=np.linspace(0, np.pi, 10),  alpha = 0.5, measure=0, stats=1500,
             name='Bloch Sphere Trajectory', save=True, collect=False, noisy=True):
             

    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    qubits = gc.AttrDict2Qubit(qubits, gc.Transmon)
    q, Q = qubits[measure], Qubits[measure]
    
    opList = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, 1)) # 1 qubit only
    
    axes = [(theta, 'Rotation Angle')]
    deps = [('Bloch X', '', ''),('Bloch Y', '', ''),('Bloch Z', '', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    rrCharacteristics,rrFreq,rrS21,rrS21fit = getrrCharacteristics(q)
   
    sb_freq_ro=rrCharacteristics[8]    
    
    fread = optimalReadoutFrequency(q,includeXi=True)
    q['readout'] = True    
    q['readout frequency']=fread
    if not val.sbSweep:    
        q['readout fc'] = fread - sb_freq_ro     
    
    def func(server, rotTheta):          
        
        probs = []
        
        if noisy: print 'angle is :',rotTheta
        
        for op in opList:
            
            if noisy: print op
        
            alg = gc.Algorithm()
            alg[gates.RotPulse([q], rotTheta)]
            alg[gates.Tomography([q], op)]
            alg[gates.Measure([q])]
            alg.compile()    
            
            preprocess([q])
            ans = yield FutureList([runQubits(server, qubits, stats, dataFormat='iq')])
            Probs,Probsstd,Smean = getProbs([q],stats,ans,[[fread]])
            
            probs.append(Probs[0])
         
            # get rho
            
            # map rho onto bloch sphere
        [rho, residual] = tomo.qst_evan(np.array(probs), tomo.octomo_ops, stats, 1)
        S = quantum.rho2bloch(rho)
        returnValue([x for x in S[1:4]])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data 
