import numpy as np
from scipy.optimize import leastsq
from scipy.special import erf, erfc
import pylab as plt
import labrad
from labrad.units import ns, GHz, MHz, mV, V, dBm, us,rad
from labrad.units import Unit,Value
from math import atan2


from pyle.pipeline import returnValue, FutureList
from pyle.dataking import util
from pyle import envelopes as env
from pyle.dataking import envelopehelpers as eh
from pyle.dataking.DevJim.CatchRelease import multiqubit as mq
from pyle.dataking import sweeps
from pyle.dataking.DevJim.CatchRelease.fpgaseq import runQubits
from pyle.util import sweeptools as st
from pyle.dataking import utilMultilevels as ml
from pyle.plotting import dstools
from pyle.fitting import fitting
from pyle.dataking import qubitpulsecal as qpc
import pyle.dataking.DevJim.CatchRelease.gateCompiler as gc
import pyle.gates as gates


"""
TODO: Static Coupling Calibrations
"""    

def fluxCouplerCrosstalk(sample, couplerScan=st.r[-500:500:10,mV], measure=0, coupler=1, spectroscopyAmp = None, startFluxShift = 0*mV,
                  fluxBelow=2*mV, fluxAbove=2*mV, fluxStep=0.1*mV, sb_freq=0*GHz, stats=300L, plot=True,
                  name='Flux - Static Coupler Crosstalk', save=True, collect=False, noisy=True, update=True):
    """Find the static coupler bias vs. flux bias voltage. A fit to the data is
    stored in the registry for the measured qubit as calFluxCouplerBias.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    couplerScan - iterable (V): Voltage to apply to static coupler bias.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    spectroscopyAmp - float: Spectroscopy pulse amplitude
    fluxBelow - float: How far below center flux to scan
    fluxAbove - float: How far above center flux to scan
    fluxStep - float: Step between scanned flux points
    sb_freq - float (Hz): Sideband frequency for spectroscopy pulse
    stats - scalar: Number of times a point will be measured.
    plot - bool: Whether to plot data and fit curve
    name - string: Name of dataset.
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    update - bool: Whether to update timingLabRRUwave registry key
    """
    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    
    operateBias = q['biasOperate']
    readoutBias = q['biasReadout']
    measAmplFunc = mq.get_mpa_func(q)
    if spectroscopyAmp is not None:
        q['spectroscopyAmp'] = spectroscopyAmp
    couplerScan = np.array([bias[mV] for bias in couplerScan])
    couplerScan = couplerScan[np.argsort(np.abs(couplerScan))]
    fluxScan = np.arange(-fluxBelow[mV], fluxAbove[mV], fluxStep[mV])
    fluxScan = fluxScan[np.argsort(abs(fluxScan))]
    fluxPoints = len(fluxScan)
    fluxStep = fluxStep[mV]
    
    sweepData = {
        'fluxFunc': np.array([startFluxShift[mV]]),
        'fluxIndex': 0,
        'couplerIndex': 0, 
        'flux': 0*fluxScan,
        'prob': 0*fluxScan,
        'maxima': 0*couplerScan,
    }
    
    axes = [('Flux Bias Shift', 'mV'), ('Coupler Bias', 'mV')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)
    
    def sweep():
        for couplerBias in couplerScan:
            center = np.polyval(sweepData['fluxFunc'], couplerBias)
            center = st.nearest(center, fluxStep)
            for dFlux in center + fluxScan:
                yield dFlux*mV, couplerBias*mV
    
    def func(server, args):
        dFlux, couplerBias = args
        c['biasOperate'] = couplerBias
        q['biasOperate'] = operateBias + dFlux
        q['biasReadout'] = readoutBias + dFlux
        q['measureAmp'] = measAmplFunc(operateBias + dFlux)      
        q.xy = eh.spectroscopyPulse(q, 0, sb_freq)
        q.z = eh.measurePulse(q, q['spectroscopyLen'] + q['piLen'])
        prob = yield runQubits(server, devs, stats, probs=[1])
        
        flux_idx = sweepData['fluxIndex']
        sweepData['flux'][flux_idx] = dFlux[mV]
        sweepData['prob'][flux_idx] = prob[0]
        if flux_idx + 1 == fluxPoints:
            # one row is done.  find the maximum and update the spectroscopy fit
            coupler_idx = sweepData['couplerIndex']
            paramsGuess = [0.5, sweepData['flux'][np.argmax(sweepData['prob'])], 0.15, 0.02]
            fits, cov, fitFunc = fitting.fitCurve('gaussian', sweepData['flux'], sweepData['prob'], paramsGuess)
            sweepData['maxima'][coupler_idx] = fits[1]
            sweepData['fluxFunc'] = np.polyfit(couplerScan[:coupler_idx+1],
                                               sweepData['maxima'][:coupler_idx+1],
                                               1)
            sweepData['fluxIndex'] = 0
            sweepData['couplerIndex'] += 1
        else:
            # just go to the next point
            sweepData['fluxIndex'] = flux_idx + 1
        returnValue([dFlux, couplerBias, prob[0]])
    sweeps.run(func, sweep(), dataset=save and dataset, collect=collect, noisy=noisy)
    
    # create a flux function and return it
    poly = sweepData['fluxFunc']
    coupler_idx = sweepData['couplerIndex']
    if plot:
        indices = np.argsort(sweepData['maxima'][:coupler_idx])
        data = np.vstack((sweepData['maxima'][:coupler_idx][indices],couplerScan[0:coupler_idx][indices])).T
        fig=dstools.plotDataset1D(data,[('Flux Bias Shift','mV')],[('Coupler Bias','','mV')],style='.',legendLocation=None,show=False,
                                  markersize=15,title='')
        fig.get_axes()[0].plot(np.polyval(poly,data[:,1]),data[:,1],'r',linewidth=3)
        fig.show()
    if update:
        Q['calFluxCouplerBias'] = poly
    return poly


def resSwapSpectroscopyStatic(sample, zpaScan=None, couplings=st.r[-0.9:0.9:0.0025,V],
                      stats=600L, measure=0, coupler=1, paramName='Bus', tBuf=10*ns,
                      name='Resonator Swap Spectroscopy Static', save=True, collect=True, noisy=True):
    """2-D spectroscopy of resonator vs static coupler bias and qubit zpa. Measures
    resonator frequency by exciting qubit and then swapping to resonator with
    varying qubit zpa (akin to Swap Spectroscopy). 
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    zpaScan - iterable: Qubit zpas over which to swap from qubit to resonator.
    couplings - iterable: Static coupler bias
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    tBuf - time: Buffer time
    name - string: Name of dataset.
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs"""
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    
    couplerBiasOld = c['biasOperate']
    if zpaScan is None:
        minZpa = mq.freq2zpa(sample,measure,q['readoutFrequency']-10*MHz)
        maxZpa = mq.freq2zpa(sample,measure,q['readoutFrequency']+10*MHz)
        zpaStep = (maxZpa-minZpa)/100.
        zpaScan = st.r[minZpa:maxZpa:zpaStep]
    axes = [(couplings,'Coupler Bias'),(zpaScan, 'Qubit Swap ZPA')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(Sample, name, axes, measure=measure, kw=kw)
    piLen = q['piLen']    
    swapTime = q['swapTime'+paramName]

    def func(server, coupling, zpa):
        fluxCouplerXtalk = q['calFluxCouplerBias']
        dFlux = (np.polyval(fluxCouplerXtalk,coupling[mV]) - np.polyval(fluxCouplerXtalk,c['biasOperate'][mV]))*mV
        c['biasOperate'] = coupling
        q['biasOperate'] += dFlux
        q['biasReadout'] += dFlux
        t = 0
        q.xy = eh.mix(q, eh.piPulseHD(q, t+piLen/2))
        t += (piLen + tBuf)
        q.z = env.rect(t, swapTime, zpa)
        t += swapTime + tBuf
        q.z += eh.measurePulse(q, t)
        return runQubits(server, devs, stats, probs=[1])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data


def fitResSpectroscopyStatic(dir, datasetId, sample=None, minBias=None, maxBias=None, minZPA=None, maxZPA=None,
                             minFitBias=None, maxFitBias=None, name='Resonator Freq Fit vs Static Tuning',
                             savePath=None, save=True, collect=True, update=True, plot=True):
    """Fit resonator Freq with static z pulse tuning on the adjustable coupler. Requires
    data from resSwapSpectroscopyStatic. A fit to the data is stored in the registry for the
    measured qubit as calSwapCouplerBias.
    
    INPUT PARAMETERS
    dir: Location of data from resT1DynamicCoupler (eg, s._dir)
    datasetId - int: Dataset number for data from resSwapSpectroscopyStatic
    sample: Object defining qubits to measure, loaded from registry. Only needed if update is True.
    minBias - float: Minimum bias for which to determine swap Amp
    maxBias - float: Maximum bias for which to determine swap Amp
    minZPA - float: Minimum frequency to use to determine swap Amp
    maxZPA - float: Maximum frequency to use to determine swap Amp
    minBias - float: Minimum bias to use in fitting
    maxBias - float: Maximum bias to use in fitting
    name - string: Name of dataset.
    savePath: Location in datavault to which to save data
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    update - bool: Whether to update calSwapCouplerBias registry key
    plot - bool: Whether to plot data corrected for FastBias resolution and fit.
    """
    if sample==None and update:
        raise Exception('Must give sample to update registry')
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datasetId, session=dir)
    data = dataset.data
    couplingUnit = Unit(dataset.variables[0][0][1])
    params = dataset.parameters
        
    #Determine the squid biases & separation between biases
    biases = np.unique(data[:,0])
    #Underestimate of separation between biases (in case multiple copies of same bias kept)
    dbias = np.abs(biases[-1]-biases[0])/len(biases)
    # If min,max not defined, select min and max over data.
    if minBias==None:
        minBias = np.min(biases)
    if maxBias==None:
        maxBias = np.max(biases)
    if minZPA==None:
        minZPA = np.min(data[:,1])
    if maxZPA==None:
        maxZPA = np.max(data[:,1])
    # Determine biases in [minBias,maxBias]
    biasesCut = [bias for bias in biases if (bias>=minBias and bias<=maxBias)]
    
    # For each bias in [minBias,maxBias], fit T1.
    zpas = []
    for bias in biasesCut:
        # Select only data with the desired bias.
        dataBias = data[np.abs(data[:,0]-bias)<dbias/2.0,1:]
        dataBiasCut = dataBias[np.logical_and(dataBias[:,0]>=minZPA, dataBias[:,0]<=maxZPA)]
        zpaMinProb = dataBiasCut[np.argmin(dataBiasCut[:,1]),0]
        vGuess = [-0.5, zpaMinProb, 0.01, 0.5]
        fits, cov, fitFunc = fitting.fitCurve('gaussian', dataBiasCut[:,0], dataBiasCut[:,1], vGuess)
        zpas.append(fits[1])
    newdata = np.array([biasesCut,zpas]).T
        
    # Now save T1s to the data vault
    if save:
        if savePath==None:
            savePath = dir
        with labrad.connect() as cxn:
            dv = cxn.data_vault
            dv.cd(savePath)
            dv.new(name,[dataset.variables[0][0]],[('Qubit Swap ZPA','','')])
            dv.add(newdata)
            for key in sorted(params):
                if key in params.config:
                    for qubitkey in sorted(params[key]):
                        dv.add_parameter(key+'.'+qubitkey,params[key][qubitkey])
                else:
                    dv.add_parameter(key,params[key])
            dv.add_parameter('datasetId',datasetId)
            dv.add_parameter('dataPath',dir)
            dv.add_parameter('minBias',minBias)
            dv.add_parameter('maxBias',maxBias)
            dv.add_parameter('minZPA',minZPA)
            dv.add_parameter('maxZPA',maxZPA)
    if update:
        if minFitBias==None:
            minFitBias = minBias
        if maxFitBias==None:
            maxFitBias = maxBias
        
        # Correcting experimental data for resolution of FastBiases
        measure = params['measure'][0]
        q = params[params['config'][0]]
        fluxStep = 2.5*V/2**15 # FastBias resolution
        couplingFluxStep = (fluxStep/q['calFluxCouplerBias'][0]) # Convert resolution into coupling units
        dataCut = newdata[np.logical_and(newdata[:,0]>minFitBias,newdata[:,0]<maxFitBias)]
        dFluxStep = dataCut[:,0]/couplingFluxStep[couplingUnit]-np.round(dataCut[:,0]/couplingFluxStep[couplingUnit]) #Determine fraction of FastBias step for given point
        dzpas = np.array([mq.freq2zpa(sample,0,mq.bias2freq(sample,0,q['biasOperate']-d*fluxStep)*GHz) for d in dFluxStep]) #Offset in swapAmps due to FastBias resolution
        couplings = dataCut[:,0]
        swapAmps = dataCut[:,1]+dzpas
        
        width = maxFitBias-minFitBias
        mid = (maxFitBias+minFitBias)/2.0
        minVal = swapAmps[np.argmin(np.abs(couplings-mid))]
        amp = (swapAmps[np.argmin(np.abs(couplings-(mid+maxFitBias)/2.0))]-minVal)/(1-np.sqrt(2))
        fits,cov,fitFunc = fitting.fitCurve('secant',couplings,swapAmps,[amp,width,mid,minVal])
        sample,qubits,Qubits = util.loadQubits(sample,write_access=True)
        Q = Qubits[measure]
        Q['calSwapCouplerBias'] = fits
        
        if plot:
            fig = dstools.plotDataset1D(np.array([couplings,swapAmps]).T, [dataset.variables[0][0]],
                                        [(dataset.variables[0][1][0],'',dataset.variables[0][1][1])],
                                        marker='.', markersize=15)
            ax = fig.get_axes()[0]
            ax.plot(couplings,fitFunc(couplings,*fits),'k')
            ax.grid()
            ax.set_title(dataset['path'])
            fig.show()
    if collect:
        return newdata
    
    
def getResSwapStatic(qubit, couplerBias, correctFastBias=True):
    """Return resonator swap amplitude for a given static coupler bias"""
    amp,width,mid,minVal = qubit['calSwapCouplerBias']
    print amp, width,mid,minVal
    trueSwap = fitting.secant(couplerBias,amp,width,mid,minVal) #Swap without correcting for FastBias resolution
    if not correctFastBias:
        return trueSwap
    
    fluxStep = 2.5*V/2**14 # FastBias resolution
    couplingFluxStep = (fluxStep/qubit['calFluxCouplerBias'][0]) # Convert resolution into coupling units
    dFluxStep = couplerBias['V']/couplingFluxStep['V']-np.round(couplerBias['V']/couplingFluxStep['V']) #Determine fraction of FastBias step for given point
    
    #mq.bias2freq
    pF = qubit['calFluxFunc']
    bias = qubit['biasOperate']-dFluxStep*fluxStep
    freq = (((bias-qubit['biasStepEdge'])[V]-pF[1])/pF[0])**(0.25)*GHz
    
    #mq.freq2zpa
    pZ = qubit['calZpaFunc']
    dzpa = pZ[1] + pZ[0]*(freq['GHz']**4)
    # dzpa is offset in swapAmps due to FastBias resolution
    
    correctedSwap = trueSwap-dzpa
    return correctedSwap


def resT1StaticCoupler(sample, delay=np.logspace(-2,.8,101)*us,couplings=st.r[0:0.2:0.02,mV],
                       paramName='Bus',stats=600L, measure=0, coupler=1,
                       name='Resonator T1 Static Coupling', save=True, collect=True, noisy=True):
    """Measure resonator T1 with static z pulse tuning on the adjustable coupler
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    delay - iterable (us): Resonator delay time (between excite/readout)
    couplings - iterable: Static coupler bias during resonator delay time
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    name - string: Name of dataset.
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    
    axes = [(couplings,'Coupler Bias'),(delay, 'Resonator T1 Delay')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)
            
    sl = q['swapTime'+paramName]
    sa = q['swapAmp'+paramName]
    
    def func(server, coupling, delay):
        fluxCouplerXtalk = q['calFluxCouplerBias']
        dFlux = (np.polyval(fluxCouplerXtalk,coupling[mV]) - np.polyval(fluxCouplerXtalk,c['biasOperate'][mV]))*mV
        c['biasOperate'] = coupling
        q['biasOperate'] += dFlux
        q['biasReadout'] += dFlux
        #q['swapAmp'] = np.polyval(q['calSwapCouplerBias'],coupling[V])     
        q['swapAmp'] = getResSwapStatic(q, coupling)
        print 'Vals: ',coupling,q['biasOperate'],q['swapAmp']
        q.xy = eh.mix(q, eh.piPulseHD(q, 0))
        q.z = env.rect(q.piLen/2, sl, q['swapAmp'])
        q.z += env.rect(q.piLen/2+sl+delay, sl, q['swapAmp'])
        q.z += eh.measurePulse(q, q.piLen/2+sl+delay+sl)
        return runQubits(server, devs, stats, probs=[1])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data


def operateQubitCouplerStatic(qubit, coupler, couplerBias, paramName='Bus'):
    """Update qubit parameters to operate at the specified static coupler bias."""
    fluxCouplerXtalk = qubit['calFluxCouplerBias']
    dFlux = (np.polyval(fluxCouplerXtalk,couplerBias[mV]) - np.polyval(fluxCouplerXtalk,coupler['biasOperate'][mV]))*mV
    print 'couplerBias: %s, dFlux: %s' % (couplerBias, dFlux)
    coupler['biasOperate'] = couplerBias
    qubit['biasOperate'] += dFlux
    qubit['biasReadout'] += dFlux
    qubit['biasStepEdge'] += dFlux
    qubit['swapAmp'+paramName] = getResSwapStatic(qubit, couplerBias)


def operateCouplerStatic(sample, couplerBias, paramName='Bus', measure=0, coupler=1):
    """Update qubit parameters to operate at the specified static coupler bias."""
    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    Q = Qubits[measure]
    C = Qubits[coupler]
    operateQubitCoupler(Q,C,couplerBias,paramName)


"""
TODO: Test Delays and Utilities
"""
    
    
def testCouplerZDelayWrtQubit(sample, t0=st.r[-30:30:0.25,ns],measure=0, coupler=1,
                              tBuf=5*ns, paramName='Bus', czpas=st.r[0.8:-0.8:-0.05], stats=1200, update=True,
                              save=True, name='Coupler Z-Qubit Test Delay', plot=True, noisy=True, collect=False):   
    """ This is to calibrate the test delay between qubit z pulse and adjustable
    coupler dynamic z pulse.
    
    Note: If SZ is not a float or int, will take 2D scan. Once scan is complete or
    cancelled, will prompt for which SZ to use to calculate testdelay time.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    t0 - iterable (us): Delay time between qubit and coupler z-pulses.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    measureC - scalar: ?????
    measureR - scalar: ??????
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    czpas - iterable: Amplitude of coupler z-pulse.
    stats - scalar: Number of times a point will be measured.
    update - bool: Whether to update timingLabRRUwave registry key
    save - bool: Whether to save data to the datavault.
    name - string: Name of dataset.
    plot - bool: Whether to plot data and fit curve
    noisy - bool: Whether to print out probabilities while the scan runs
    collect - bool: Whether or not to return data.
    """
    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    C = Qubits[coupler]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    
    axes = [(czpas,'Coupler ZPA'),(t0, 'Delay of Coupler wrt Qubit')]
    kw = {'stats': stats,}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)
    
    swapTime = q['swapTime'+paramName]
    
    def func(server, cz,t0):
        if t0>0:
            t=0
        else:
            t = -t0
        q.xy = eh.mix(q, eh.piPulse(q, t-q['piLen']/2))
        c.z = env.rect(t+t0, swapTime, cz)
        q.z = env.rect(t, swapTime, q['swapAmp'+paramName])
        measBuf = (t0>0)*t0 + tBuf
        q.z += eh.measurePulse(q, t+swapTime+measBuf)
        return runQubits(server, devs, stats, probs=[1])
    result = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    
    if not (isinstance(cz,int) or isinstance(cz,float)):
        selectBias = float(raw_input('Select Coupler ZPA: '))
        biasSep = cz[1]-cz[0]
        result = result[(np.abs(result[:,0]-selectBias))<biasSep/2.0]
        result = result[:,1:]
    
    zpl = swapTime[ns]*0.8
    translength = 0.8*q['piFWHM'][ns]
    def fitfunc(x, p):
        return (p[1] -
                p[2] * 0.5*erfc((x - (p[0] - zpl/2.0)) / translength) -
                p[3] * 0.5*erf((x - (p[0] + zpl/2.0)) / translength))
    x, y = result.T
    fit, _ok = leastsq(lambda p: fitfunc(x, p) - y, [0.0, 0.05, 0.85, 0.85])
    if plot:
        plt.figure()
        plt.plot(x, y, '.')
        plt.plot(x, fitfunc(x, fit))
    print 'Coupler Dynamic Z lag:', fit[0]
    if update:
        print 'Coupler Dynamic Z delay corrected by %g ns' % fit[0]
        C['timingLagWrtMaster'] += fit[0]*ns
    if collect:
        return result
    
    
def testResDelay(sample, delay=st.r[-80:80:0.5,ns], pulseLength=8*ns, pulseSep=20*ns, amp=0.2,
                 stats=600L, measure=0, coupler=1, resonator=2, paramName='Bus',
                 cz=0.0, name='Resonator-qubit test delay',
                 tBuf=100*ns, save=True, collect=False, noisy=True, plot=True, update=True):
    """Test delay between resonator uwave drive and qubit z pulse line
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    delay - iterable (us): Delay time from qubit z-pulse center to center between
        two resonator pulses
    pulseLength - time: FWHM of Gaussian pulses on resonator
    pulseSep - time: Time between resonator pulses
    amp - float: Amplitude of resonator pulses
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    SZ - iterable: Dynamic (Fast) coupler zpa during resonator pulses
    name - string: Name of dataset.
    tBuf - time: Time after qubit z-pulse to wait before measure pulse
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    plot - bool: Whether to plot data and fit curve
    update - bool: Whether to update timingLabRRUwave registry key
    """
    raise Exception('Fix pulse sequence.')
    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    R = Qubits[resonator]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    r = devs[resonator]
        
    axes = [(delay, 'Delay of Res wrt Qubit')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)
    
    def func(server, start):
        sb_freq = q['readoutFrequency'] - q['readout fc']
        checkTime = q['swapTime'+paramName]/2 + start - pulseSep - pulseLength
        if checkTime>0:
            t = q['swapTime'+paramName]/2 + start
        else:
            t = pulseSep + pulseLength
        # t is time in middle of two rr/cz pulses
        r.xy = env.mix(env.gaussian(t-pulseSep, pulseLength, amp=amp), sb_freq)
        c.z = env.rect(t-pulseSep-pulseLength,2*pulseLength, cz)
        r.xy += env.mix(env.gaussian(t+pulseSep, pulseLength, amp=-amp), sb_freq)
        c.z += env.rect(t+pulseSep-pulseLength,2*pulseLength, cz)
        q.z = env.rect(t-start-q['swapTime'+paramName]/2, q['swapTime'+paramName], q['swapAmp'+paramName])
        q.z += eh.measurePulse(q, t-start+q['swapTime'+paramName]/2+tBuf)
        return runQubits(server, devs, stats, probs=[1])
    result = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    
    topLen = q['swapTime'+paramName]['ns']*2*0.8
    translength = pulseLength[ns]
    def fitfunc(x, p):
        return (p[1] -
                p[2] * 0.5*erfc((x - (p[0] - topLen/2.0)) / translength) -
                p[3] * 0.5*erf((x - (p[0] + topLen/2.0)) / translength))
    x, y = result.T
    fit, _ok = leastsq(lambda p: fitfunc(x, p) - y, [0.0, 0.05, 0.85, 0.85])
    if plot:
        plt.figure()
        plt.plot(x, y, '.')
        plt.plot(x, fitfunc(x, fit))
    print 'uwave lag:', -fit[0]
    if update:
        print 'uwave lag corrected by %g ns' % fit[0]
        R['timingLagWrtMaster'] += fit[0]*ns
    if collect:
        return result
    

def resonatorSpectroscopy(sample, freqScan=None, swapTime=300*ns, stats=600L,
                          measure=0, coupler=1, resonator=1, cz=0.0,paramName='Bus',amplitude=None,
       name='Resonator spectroscopy', save=True, collect=True, noisy=True,tBuf=20*ns):
    """A single pi-pulse on one qubit, with other qubits also operated.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    freqScan - iterable: Frequencies at which to drive resonator.
    swapTime - time: Time to swap between resonator and qubit.
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    cz - float: Coupler zpa during resonator drive
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    amplitude - float or None: Amplitude of resonator drive
    name - string: Name of dataset.
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    tBuf - time: Buffer time
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    r = devs[resonator]
        
    if amplitude is not None:
        r['spectroscopyAmp'] = amplitude
    if freqScan is None:
        freqScan = st.r[r['f10']['GHz']-0.002:r['f10']['GHz']+0.002:0.00002,GHz]
    axes = [(freqScan, 'Resonator frequency')]
    kw = {'stats': stats, 'Coupler ZPA': cz}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)
    
    def func(server, freq):
        sb_freq = freq-r['fc']
        r['f10'] = freq
        r.xy = eh.spectroscopyPulse(r, 0, sb_freq)
        c.z = env.rect(0,r['spectroscopyLen']+0.0*ns, cz)  
        q.z = env.rect(r['spectroscopyLen']+tBuf, swapTime, q['swapAmp'+paramName])
        q.z += eh.measurePulse(q, r['spectroscopyLen']+tBuf+swapTime+tBuf)
        return runQubits(server, devs, stats, probs=[1])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data

    
def resonatorT1(sample, delay=np.logspace(-2,.85,101)*us, paramName='Bus', stats=1200L, cz=0.0,
                measure=0, coupler=1, name='resonator T1 MQ', save=True, collect=True, noisy=True):
    """Measures T1 of resonator using qubit to excite/readout.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    delay - iterable (us): Resonator delay time (between excite/readout)
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    stats - scalar: Number of times a point will be measured.
    zpa - iterable or float: Qubit zpa during resonator delay time
    cz - float: Coupler zpa during resonator delay time
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    name - string: Name of dataset.
    excited - bool: Whether to excite resonator into |2>
    adi - bool: Whether to use trapezoidal qubit z-pulses instead of rectangular
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    
    axes = [(delay, 'T1 delay')]
    kw = {'stats':stats, 'Coupler ZPA':cz}
    dataset = sweeps.prepDataset(sample, paramName+' '+name, axes, measure=measure, kw=kw)    
        
    swapLen = q['swapTime'+paramName]
    swapAmp = q['swapAmp'+paramName]
    
    def func(server, delay):
        q.xy = eh.mix(q, eh.piPulseHD(q, 0))
        q.z = env.rect(q['piLen']/2, swapLen, swapAmp)
        c.z = env.rect(q['piLen']/2+swapLen, delay, cz)
        q.z += env.rect(q['piLen']/2+swapLen+delay, swapLen, swapAmp)
        q.z += eh.measurePulse(q, q.piLen/2+swapLen+delay+swapLen)
        return runQubits(server, devs, stats, probs=[1])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data


def fitResT1Coupler(dir, datasetId, minBias=None, maxBias=None, minTime=None, maxTime=None,
       name='Resonator T1 Fit vs Dynamic Tuning', savePath=None, save=True, collect=True):
    """Fit resonator T1 with z pulse tuning on the adjustable coupler. Requires
    data from resT1DynamicCoupler or resT1StaticCoupler.
    
    INPUT PARAMETERS
    dir: Location of data from resT1DynamicCoupler (eg, s._dir)
    datasetId - int: Dataset number for data from resT1DynamicCoupler
    minBias - float: Minimum bias for which to fit T1
    maxBias - float: Maximum bias for which to fit T1
    minTime - float: Minimum time to use in fitting T1
    maxTime - float: Maximum time to use in fitting T1
    name - string: Name of dataset.
    savePath: Location in datavault to which to save data
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    """
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datasetId, session=dir)
    data = dataset.data
    timeUnit = Unit(dataset.variables[0][1][1])
        
    #Determine the squid biases & separation between biases
    biases = np.unique(data[:,0])
    #Underestimate of separation between biases (in case multiple copies of same bias kept)
    dbias = np.abs(biases[-1]-biases[0])/len(biases)
    # If min,max not defined, select min and max over data.
    if minBias==None:
        minBias = np.min(biases)
    if maxBias==None:
        maxBias = np.max(biases)
    if minTime==None:
        minTime = np.min(data[:,1])
    if maxTime==None:
        maxTime = np.max(data[:,1])
    # Determine biases in [minBias,maxBias]
    biasesCut = [bias for bias in biases if (bias>=minBias and bias<=maxBias)]
    
    # For each bias in [minBias,maxBias], fit T1.
    t1s = []
    for bias in biasesCut:
        # Select only data with the desired bias.
        dataBias = data[np.abs(data[:,0]-bias)<dbias/2.0,1:]
        dataBiasCut = dataBias[np.logical_and(dataBias[:,0]>minTime, dataBias[:,0]<maxTime)]
        vGuess = [0.8, 1.0, 0.05]
        fits, cov, fitFunc = fitting.fitCurve('exponential', dataBiasCut[:,0], dataBiasCut[:,1], vGuess)
        t1s.append(fits[1])
    newdata = np.array([biasesCut,t1s]).T
        
    # Now save T1s to the data vault
    if save:
        if savePath==None:
            savePath = dir
        params = dataset.parameters
        with labrad.connect() as cxn:
            dv = cxn.data_vault
            dv.cd(savePath)
            dv.new(name,[dataset.variables[0][0]],[('Resonator T1','','us')])
            dv.add(newdata)
            for key in params.keys():
                if key in params.config:
                    for qubitkey in params[key].keys():
                        dv.add_parameter(key+'.'+qubitkey,params[key][qubitkey])
                else:
                    dv.add_parameter(key,params[key])
            dv.add_parameter('datasetId',datasetId)
            dv.add_parameter('dataPath',dir)
            dv.add_parameter('minBias',minBias)
            dv.add_parameter('maxBias',maxBias)
            dv.add_parameter('minTime',minTime)
            dv.add_parameter('maxTime',maxTime)
    if collect:
        return newdata


"""
TODO: Dynamic Coupling Calibrations
"""

def resT1CouplerDynamic(sample, delay=np.logspace(-2,.8,101)*us, czpas=st.r[0:0.2:0.02,None],
                        paramName='Bus', stats=600L, measure=0, coupler=1,
                        name='resonator T1 with dynamic tuning', save=True, collect=True, noisy=True):
    """Measure resonator T1 with dynamic z pulse tuning on the adjustable coupler
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    delay - iterable (us): Resonator delay time (between excite/readout)
    czpas - iterable: Dynamic (Fast) coupler zpa during resonator delay time
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    name - string: Name of dataset.
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    
    axes = [(czpas,'Coupler ZPA'),(delay, 'Resonator T1 Delay')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)    
            
    swapLen = q['swapTime'+paramName]
    swapAmp = q['swapAmp'+paramName]
    
    def func(server, cz, delay):
        q.xy = eh.mix(q, eh.piPulseHD(q, 0))
        q.z = env.rect(q['piLen']/2, swapLen, swapAmp)
        c.z = env.rect(q['piLen']/2+swapLen, delay, cz)
        q.z += env.rect(q['piLen']/2+swapLen+delay, swapLen, swapAmp)
        q.z += eh.measurePulse(q, q['piLen']/2+swapLen+delay+swapLen)
        return runQubits(server, devs, stats, probs=[1])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data
    
        
def resSpectroscopyDynamic(sample, freqScan=st.r[6.52:6.586:0.0005,GHz], swapTime=300*ns,
                      stats=600L, measure=0, coupler=1, resonator=2, paramName='Bus',
                      amplitude=None, tBuf=10*ns, czpas=st.r[-0.9:0.9:0.0025],
                      name='Resonator Spectroscopy SZ', save=True, collect=True, noisy=True):
    """2-D spectroscopy of resonator vs coupler SZ and drive frequency.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    freqScan - iterable: Frequencies at which to drive resonator.
    swapTime - time: Time to swap between resonator and qubit.
    czpas - iterable: Coupler zpa during resonator drive
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    amplitude - float or None: Amplitude of resonator drive
    tBuf - time: Buffer time
    name - string: Name of dataset.
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    r = devs[resonator]
        
    if amplitude is not None:
        r['spectroscopyAmp'] = amplitude
    if freqScan is None:
        freqScan = st.r[q['readoutFrequency']['GHz']-0.002:q['readoutFrequency']['GHz']+0.002:0.00002,GHz]
    axes = [(czpas,'Coupler ZPA'),(freqScan, 'Resonator frequency')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)    

    def func(server, cz, freq):
        sb_freq = freq-r['fc']
        r.xy = eh.spectroscopyPulse(r, 0, sb_freq)
        c.z = env.rect(0,q['spectroscopyLen'], cz)
        q.z = env.rect(q['spectroscopyLen']+tBuf, swapTime, q['swapAmp'+paramName])
        q.z += eh.measurePulse(q, q['spectroscopyLen']+tBuf+swapTime+tBuf)
        return runQubits(server, devs, stats, probs=[1])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data
               
        
def resSwapSpectroscopyDynamic(sample, zpaScan=None, czpas=st.r[-0.9:0.9:0.0025],
                      stats=600L, measure=0, coupler=1, resonator=2, paramName='Bus', tBuf=10*ns,
                      name='Resonator Swap Spectroscopy SZ', save=True, collect=True, noisy=True):
    """2-D spectroscopy of resonator vs coupler SZ and qubit zpa. Measures
    resonator frequency by exciting qubit and then swapping to resonator with
    varying qubit zpa (akin to Swap Spectroscopy).
    
    Not yet tested.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    zpaScan - iterable: Qubit zpas over which to swap from qubit to resonator.
    czpas - iterable: Coupler zpa during resonator drive
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    paramName - string: Name of swap registry keys (swapAmpName and swapTimeName).
    tBuf - time: Buffer time
    name - string: Name of dataset.
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs"""
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    r = devs[resonator]
        
    if zpaScan is None:
        minZpa = mq.freq2zpa(sample,measure,r['f10']-2*MHz)
        maxZpa = mq.freq2zpa(sample,measure,r['f10']+2*MHz)
        zpaStep = mq.freq2zpa(sample,measure,r['f10'])-mq.freq2zpa(sample,measure,r['f10']+0.02*MHz)
        zpaScan = st.r[minZpa:maxZpa:zpaStep]
    axes = [(czpas,'squid z pulse amp'),(zpaScan, 'Qubit Swap ZPA')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(Sample, name, axes, measure=measure, kw=kw)
    piLen = q['piLen']    
    swapTime = q['swapTime'+paramName]

    def func(server, cz, zpa):
        t = 0
        q.xy = eh.mix(q, eh.piPulseHD(q, t+piLen/2))
        t += (piLen + tBuf)
        # Start cz pulse (located at end of pulse sequence)
        t += tBuf
        q.z = env.rect(t, swapTime, zpa)
        t += swapTime + tBuf
        c.z = env.rect(piLen+tBuf, t-(piLen+tBuf), cz)
        t+= tBuf        
        q.z += eh.measurePulse(q, t)
        return runQubits(server, devs, stats, probs=[1])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data


def fitResSpectroscopyDynamic(dir, datasetId, minBias=None, maxBias=None, minFreq=None, maxFreq=None,
       name='Resonator Freq Fit vs Dynamic Tuning', savePath=None, save=True, collect=True):
    """Fit resonator Freq with dynamic z pulse tuning on the adjustable coupler. Requires
    data from resSpectroscopyDynamic.
    
    INPUT PARAMETERS
    dir: Location of data from resSpectroscopyDynamic (eg, s._dir)
    datasetId - int: Dataset number for data from resSpectroscopyDynamic
    minBias - float: Minimum bias for which to fit T1
    maxBias - float: Maximum bias for which to fit T1
    minFreq - float: Minimum frequency to use in fitting T1
    maxFreq - float: Maximum frequency to use in fitting T1
    name - string: Name of dataset.
    savePath: Location in datavault to which to save data
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    """
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datasetId, session=dir)
    data = dataset.data
    print 'data loaded'
    freqUnit = Unit(dataset.variables[0][1][1])
    params = dataset.parameters
        
    #Determine the squid biases & separation between biases
    biases = np.unique(data[:,0])
    #Underestimate of separation between biases (in case multiple copies of same bias kept)
    dbias = np.abs(biases[-1]-biases[0])/len(biases)
    # If min,max not defined, select min and max over data.
    if minBias==None:
        minBias = np.min(biases)
    if maxBias==None:
        maxBias = np.max(biases)
    if minFreq==None:
        minFreq = np.min(data[:,1])
    if maxFreq==None:
        maxFreq = np.max(data[:,1])
    # Determine biases in [minBias,maxBias]
    biasesCut = [bias for bias in biases if (bias>=minBias and bias<=maxBias)]
    
    # For each bias in [minBias,maxBias], fit T1.
    freqs = []
    for bias in biasesCut:
        print bias
        # Select only data with the desired bias.
        dataBias = data[np.abs(data[:,0]-bias)<dbias/2.0,1:]
        dataBiasCut = dataBias[np.logical_and(dataBias[:,0]>=minFreq, dataBias[:,0]<=maxFreq)]
        freqMaxProb = dataBiasCut[np.argmax(dataBiasCut[:,1]),0]
        vGuess = [0.5, freqMaxProb, 0.0001, 0.05]
        fits, cov, fitFunc = fitting.fitCurve('gaussian', dataBiasCut[:,0], dataBiasCut[:,1], vGuess)
        freqs.append(fits[1])
    newdata = np.array([biasesCut,freqs]).T
        
    # Now save T1s to the data vault
    if save:
        if savePath==None:
            savePath = dir
        with labrad.connect() as cxn:
            dv = cxn.data_vault
            dv.cd(savePath)
            dv.new(name,[dataset.variables[0][0]],[('Resonator Frequency','','GHz')])
            dv.add(newdata)
            for key in params.keys():
                if key in params.config:
                    for qubitkey in params[key].keys():
                        dv.add_parameter(key+'.'+qubitkey,params[key][qubitkey])
                else:
                    dv.add_parameter(key,params[key])
            dv.add_parameter('datasetId',datasetId)
            dv.add_parameter('dataPath',dir)
            dv.add_parameter('minBias',minBias)
            dv.add_parameter('maxBias',maxBias)
            dv.add_parameter('minFreq',minFreq)
            dv.add_parameter('maxFreq',maxFreq)
    if collect:
        return newdata


def fitResFreqDynamic(sample, datasetId, dir=None, minBias=None, maxBias=None, excludeBias=False, 
                 name='Resonator Freq Fit vs Dynamic Tuning', update=True, plot=True):
    """Fit resonator Freq with dynamic z pulse tuning on the adjustable coupler. Requires
    data from fitResSpectroscopyDynamic. A fit to the data is stored in the registry for the
    measured qubit as calCouplerDynamicFreq.
    
    INPUT PARAMETERS
    dir: Location of data from fitResSpectroscopyDynamic (eg, s._dir)
    datasetId - int: Dataset number for data from fitResSpectroscopyDynamic
    sample: Object defining qubits to measure, loaded from registry. Only needed if update is True.
    minBias - float: Minimum bias for which to determine swap Amp
    maxBias - float: Maximum bias for which to determine swap Amp
    minZPA - float: Minimum frequency to use to determine swap Amp
    maxZPA - float: Maximum frequency to use to determine swap Amp
    minBias - float: Minimum bias to use in fitting
    maxBias - float: Maximum bias to use in fitting
    name - string: Name of dataset.
    savePath: Location in datavault to which to save data
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    update - bool: Whether to update calCouplerDynamicFreq registry key
    plot - bool: Whether to plot data corrected for FastBias resolution and fit.
    """
    if dir==None:
        dir = sample._dir
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datasetId, session=dir)
    data = dataset.data
    params = dataset.parameters
        
    #Determine the squid biases & separation between biases
    biases = data[:,0]
    #Underestimate of separation between biases (in case multiple copies of same bias kept)
    dbias = np.min(biases[1:]-biases[:-1])
    if dbias <1e-8:
        raise Exception('Biases do not all have only one frequency.')
    # If min,max not defined, select min and max over data.
    if minBias==None:
        minBias = np.min(biases)
    if maxBias==None:
        maxBias = np.max(biases)
    if not excludeBias:
        excludeBias = [1e5]
    excludeBias = np.array(excludeBias)
    # Determine biases in [minBias,maxBias] or in excludeBias
    whichCut = np.array([(bias>=minBias and bias<=maxBias and np.min(np.abs(excludeBias-bias))>(dbias/4.0)) for bias in biases])
    biasesCut = biases[whichCut]
    freqsCut = data[whichCut,1]
    
    width = maxBias-minBias
    mid = (maxBias+minBias)/2.0
    minVal = freqsCut[np.argmin(np.abs(biasesCut-mid))]
    amp = (freqsCut[np.argmin(np.abs(biasesCut-(mid+maxBias)/2.0))]-minVal)/(1-np.sqrt(2))
    fits,cov,fitFunc = fitting.fitCurve('secant',biasesCut,freqsCut,[amp,width,mid,minVal])
    
    if update:
        sample,qubits,Qubits = util.loadQubits(sample,write_access=True)
        Q = Qubits[params['measure'][0]]
        Q['calCouplerDynamicFreq'] = fits
        
    if plot:
        fig = dstools.plotDataset1D(np.array([biasesCut,freqsCut]).T, [dataset.variables[0][0]],
                                    [dataset.variables[1][0]],
                                    marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.plot(biasesCut,fitFunc(biasesCut,*fits),'k')
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    return fits
    
    
def getResFreqDynamic(qubit, couplerBias, correctFastBias=True):
    """Return resonator frequency for a given dynamic coupler bias"""
    amp,width,mid,minVal = qubit['calCouplerDynamicFreq']
    trueSwap = fitting.secant(couplerBias,amp,width,mid,minVal)
    return trueSwap*GHz
    
    
def getResFreq(sample, qubitNum, couplerBias, correctFastBias=True):
    """Return resonator frequency for a given dynamic coupler bias"""
    sample, qubits = util.loadQubits(sample)
    qubit = qubits[qubitNum]
    return getResFreqDynamic(qubit,couplerBias)


def pulseshapeDynamic(sample, measure=0, swapName='Bus', tBuf=5*ns,
                      biasStep=0.005, minPeak=0.4, stopAt=0.3, minCount=2, time=None,
                      height=-1.0, initSettling=1000*ns, useOldRatesAmps = False,
                      pulseLen=10*ns, pulseAmp=1.0, freq=None, sb_freq=50*MHz,
                      stats=300, save=True, name='Coupler Pulse Shape', plot=True,
                      noisy=True, update=True):
    """
    Determine coupler bias pulse shape. A fit to the data is stored in the registry for the
    measured qubit as settling(Rates/Amplitudes)Coupler.
    
    INPUT PARAMETERS
    Qubit parameters:
        sample: Object defining qubits to measure, loaded from registry.
        measure - scalar: Number of qubit to measure. Only one qubit allowed.
        swapName - string: Name of swap registry keys (swapAmpName and swapTimeName).
        tBuf - time: Buffer time.
    Definition of scan iterables:
        biasStep - float: Step size for coupler bias.
        minPeak - float: Minimum qubit probability for probability peak.
        stopAt - float: Max qubit probability at which to stop scan once peak has been found.
        minCount - int: Minimum number of peak probabilities measured to declare peak found.
        time - iterable time: How long after defining pulse to have resonator drive pulse.
    Coupler bias pulse:
        height - float: Amplitude of coupler bias pulse.
        initSettling - time: Initial settling time at coupler bias of height prior to applying
            pulse to be measured. Should be long enough that coupler bias certain to be
            constant.
        useOldRatesAmps - bool: Whether to use original registry keys or to blank them out.
            Use original keys to determine if correction code working. Can not update
            registry keys if set to True.
    Resonator microwave pulse:
        pulseLen - time: Length of resonator microwave drive pulse.
        pulseAmp - float: Amplitude of resonator microwave drive pulse.
        freq - frequency: Frequency of resonator microwave drive pulse.
        sb_freq - frequency: Sideband mixing frequency.
    Data collection:
        stats - int: Number of times a point will be measured.
        save - bool: Whether to save data to the datavault.
        name - string: Name of dataset.
        plot - bool: Whether to plot data.
        noisy - bool: Whether to print out probabilities while the scan runs.
        update - bool: Whether to update registry keys. Can not update registry keys if
            useOldRatesAmps set to True.
    """
    raise Exception('Fix pulse sequence.')
    if useOldRatesAmps and update:
        raise Exception('Using settling correction for the experiment and then updating will lead to incorrect settling parameters!') 
    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    if time is None:
        time = range(3, 25) + [round(10**(0.02*i)) for i in range(70, 110)]
        time = [t*ns for t in time]
    if freq is None:
        freq = q['readout frequency']
    q['readout fc'] =  freq - sb_freq  

    axes = [('time after step', time[0].units), ('z offset', '')]
    kw = {'step height': height, 'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)
    
    scanInfo = {'peakFound': False, 'lowCount': 0, 'highCount': 0}
    def sweep():
        center = 0
        for t in time:
            scanInfo['time'] = t
            scanInfo['center'] = center
            yield t, center
            low = center
            high = center
            scanInfo['lowCount'] = 0
            scanInfo['highCount'] = 0
            scanInfo['peakFound'] = False
            while ((scanInfo['lowCount'] < minCount) or
                   (scanInfo['highCount'] < minCount) or
                   (not scanInfo['peakFound'])):
                if (scanInfo['lowCount'] < minCount) or (not scanInfo['peakFound']):
                    low -= biasStep
                    yield t, low
                if (scanInfo['highCount'] < minCount) or (not scanInfo['peakFound']):
                    high += biasStep
                    yield t, high
            center = round(0.5*(low+high)/biasStep)*biasStep
    
    def func(server, args):
        t, ofs = args
        if not useOldRatesAmps:
            q['settlingAmplitudesCoupler'] = []
            q['settlingRatesCoupler'] = []
        q.rr = env.mix(env.rect(initSettling+t-pulseLen/2., pulseLen, pulseAmp), sb_freq)
        q.cz = env.rect(0, initSettling+t+pulseLen+tBuf, height + ofs)
        q.cz += env.rect(initSettling, t+pulseLen+tBuf, -height)
        q.z = env.rect(initSettling+t+pulseLen+tBuf, q['swapTime'+swapName], q['swapAmp'+swapName])
        q.z += eh.measurePulse(q, initSettling+t+pulseLen+2*tBuf+q['swapTime'+swapName])
        probData = yield runQubits(server, devs, stats, probs=[1])
        prob = probData[0]
        if t == scanInfo['time']:
            if prob >= minPeak:
                scanInfo['peakFound'] = True
            side = 'highCount' if ofs > scanInfo['center'] else 'lowCount'
            if prob < stopAt:
                scanInfo[side] += 1
            else:
                scanInfo[side] = 0
        returnValue([t, ofs, prob])
    pulsedata = sweeps.run(func, sweep(), dataset=save and dataset, noisy=noisy)
    
    p, func = qpc._getstepfunc(pulsedata, height, plot=plot,
                               ind=dataset.independents, dep=dataset.dependents)
    
    if update:
        Q['settlingRatesCoupler'] = p[2::2]
        Q['settlingAmplitudesCoupler'] = p[1::2]/float(height)
        
    
"""
TODO: S-Scanning Circles
"""

def circlePlot(session,dataNum=1,calData=False,xGuess=0,yGuess=0,pGuess=[0.0,0.0],ampRescale=False,calUpdate=False):
    """ Plot circle for s_scanning. Apply calibration if desired.
    
    INPUT PARAMETERS
    session: Directory where data saved.
    dataNum - int: Number of dataset.
    calData - bool: Whether to apply calibration to data.
    xGuess - float: Guess for x-axis center of circle.
    yGuess - float: Guess for y-axis center of circle.
    pGuess - [float,float]: Guess for line describing phase of transmission
        line in format [intercept~0,slope~300]
    ampRescale - bool: Whether to rescale data such that fit circle radius is 1.
    calUpdate - bool: Whether to find fit parameters for circle.
    """
    with labrad.connect() as cxn:
        dv = cxn.data_vault
        dataset = dstools.getDeviceDataset(dv,datasetId=dataNum,session=session)
    data = dataset.data
    params = dataset.parameters
    q = params[params['config'][params['measure'][0]]]
    freqs = data[:,0]
    phases = data[:,1]
    amps = data[:,2]
    
    freqCutMin = float(raw_input('Resonance Peak Min (in GHz): '))
    freqCutMax = float(raw_input('Resonance Peak Max (in GHz): '))
    
    if calUpdate:
        params = circleCal(data,freqCutMin,freqCutMax,xGuess=xGuess,yGuess=yGuess,pGuess=pGuess)
    else:
        params = [q.get('calCircleX',xGuess), q.get('calCircleY',yGuess), q.get('calCircleR',np.mean(amps)),[q.get('calCircleP0',pGuess[0]),q.get('calCircleP1',pGuess[1])]]
        
    if calData:
        x = amps*np.cos(phases)
        y = amps*np.sin(phases)
        xCentered = x-params[0]
        yCentered = y-params[1]
        sCentered = xCentered + 1j * yCentered
        amps = np.abs(sCentered)
        phases = np.angle(sCentered)
    if ampRescale:
        amps /= params[2]

    x = amps*np.cos(phases)
    y = amps*np.sin(phases)
    plt.figure(2)
    plt.plot(x,y,'bo-')
    
    if calData:
        phases = np.mod(phases-freqs*params[3][1]-params[3][0],2*pi)
        x = amps*np.cos(phases)
        y = amps*np.sin(phases)
        plt.figure(2)
        plt.plot(x,y,'go-')
    return freqs,x,y,params
    
    
def circleCal(data,freqCutMin,freqCutMax,xGuess=0,yGuess=0,pGuess=[0.0,300.0],ampRescale=False):
    """ Determine calibration for s_scanning circle.
    
    INPUT PARAMETERS
    data: Circle data.
    xGuess - float: Guess for x-axis center of circle.
    yGuess - float: Guess for y-axis center of circle.
    pGuess - [float,float]: Guess for line describing phase of transmission
        line in format [intercept,slope]
    ampRescale - bool: Whether to rescale data such that fit circle radius is 1.
    """
    #freqs = data[:,0]
    if freqCutMax<freqCutMin:
        freqCutMin,freqCutMax = freqCutMax,FreqCutMin
    dFreq = data[1,0]-data[0,0]
    whereResonator = np.logical_or(data[:,0]<=freqCutMin,data[:,0]>=freqCutMax)
    dataCut = data[whereResonator,:]
    
    x=dataCut[:,2]*np.cos(dataCut[:,1])
    y=dataCut[:,2]*np.sin(dataCut[:,1])
    
    def calc_R(xc, yc):
        """ Calculate the distance of each 2D points from the center (xc, yc) """
        return sqrt((x-xc)**2 + (y-yc)**2)

    def f_2(c):
        """ calculate the algebraic distance between the 2D points and the mean circle centered at c=(xc, yc) """
        Ri = calc_R(*c)
        return Ri - Ri.mean()
    
    center_2, ier = leastsq(f_2, [xGuess, yGuess])

    xc_2, yc_2 = center_2
    Ri_2 = calc_R(xc_2, yc_2)
    R_2 = Ri_2.mean()
    
    x_new = x-xc_2
    y_new = y-yc_2
    phase_new = np.unwrap(np.angle(x_new+1j*y_new))
    plt.figure(3)
    plt.plot(dataCut[:,0],phase_new,'-')
    num2Pis = int(raw_input('Number of 2*pi to add for freq>resonator: '))
    plt.close(3)
    phaseCutMin = np.min(np.where(whereResonator==False)[0])
    addPhaseCut = num2Pis*2*np.pi*(np.arange(len(phase_new))>=phaseCutMin)
    phase_new = phase_new+addPhaseCut
    plt.figure(3)
    plt.plot(dataCut[:,0],phase_new,'-')
    
    def fitfunc(freq,p):
        return p[0]+p[1]*freq
    def errfunc(p):
        return phase_new-fitfunc(dataCut[:,0],p)
    p,ok = leastsq(errfunc, pGuess)
    print p
    plt.figure(3)
    plt.plot(dataCut[:,0],np.unwrap(p[0]+p[1]*dataCut[:,0])+addPhaseCut)
    
    return xc_2, yc_2, R_2, p
    
        
        
def s_scanning(Sample, freqs=None, amps=.01, zpa=0.0, couplerBias=0, pulseLen=None, 
               sb_freq = -50*MHz, measure=0, coupler=1, resonator=2, stats=150,
               tail=2800*ns, update = False,
               save=True, name='S parameter scanning', collect=True, noisy=True):
    """ This is the original s_scaning with sz"""
    sample, qubits, Qubits = util.loadQubits(sample, write_access=True)
    r, R = qubits[resonator], Qubits[resonator]
    r['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    r = devs[resonator]
    
    if freqs is None:
        f = st.nearest(q['readoutFrequency'][GHz], 1e-5)
        freqs = st.r[f-0.0015:f+0.0015:1e-5, GHz]
    if pulseLen is None:
        pulseLen = q['readout length']
    if update:
        name = 'Cali ' + name
        if r.get('calCircleX',None) is not None:
            r.pop('calCircleX')
            r.pop('calCircleY')
            r.pop('calCircleP0')
            r.pop('calCircleP1')
            r.pop('calCircleR')
    q['biasReadout'] = q['biasOperate']
    r['adc mode'] = 'demodulate'
    
    axes = [(amps,'amplitude'), (freqs, 'Frequency')]
    deps = [('Phase', 'S11 for %s'%q.__name__, rad) for q in [q]]+ [('Amplitude','S11 for %s'%q.__name__,'') for q in [q]]    
    kw = {'stats': stats,
          'zpa': zpa,
          'Coupler Bias': couplerBias,
          'Pulse Length':pulseLen}        
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
                     
    def func(server, ampCurr, freqCurr):
        amp =[]
        phase =[]
        r['fc'] = freqCurr - sb_freq
        r['adc demod frequency'] = freqCurr-r['fc']
        r.xy = env.mix(env.rect(0, pulseLen, ampCurr), sb_freq)
        q.z = env.rect(0, pulseLen+tail, zpa)
        c.z = env.rect(0,pulseLen+tail, couplerBias)

        if noisy: print freqCurr, ampCurr
        
        data = yield FutureList([runQubits(server, devs, stats, dataFormat='iq')])        
        I_mean = np.mean(data[0][0][0])
        Q_mean = np.mean(data[0][0][1])
        ref = I_mean + 1j*Q_mean
        amp.append(abs(I_mean+1j*Q_mean))
        phase.append(atan2(Q_mean,I_mean))
        returnValue(phase+amp)
    
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        freq,x,y,params = circlePlot(dataNum=-1,session=sample._dir,calData=True,xGuess=0,yGuess=0,
                                     pGuess=[0.0,350.0],ampRescale=True,calUpdate=True)
        x0,y0,r0,p = params
        R['calCircleX']=x0
        R['calCircleY']=y0
        R['calCircleR']=r0
        R['calCircleP0']=p[0]
        R['calCircleP1']=p[1]
    if collect:
        return data
    
    
def sScanningDynamicCoupler(sample, SZs, dirT1, numT1, dirF, numF):
    with labrad.connect() as cxn:
        dv=cxn.data_vault
        datasetT1 = dstools.getDeviceDataset(dv,numT1,dirT1)
        datasetF = dstools.getDeviceDataset(dv,numF,dirF)
    freqBiases = datasetF.data[:,0]
    freqs = datasetF.data[:,1]
    t1Biases = datasetT1.data[:,0]
    t1s = datasetT1.data[:,1]
    if not all(freqBiases==t1Biases):
        raise Exception('Biases must be equal in freq, T1')
    dfreqs = 1e-3/(2*np.pi*t1s)
    biasSelects = [np.where(np.abs(t1Biases-SZ)<1e-8)[0][0] for SZ in SZs]
    szSelects = t1Biases[biasSelects]
    freqSelects = freqs[biasSelects]
    dfreqSelects = dfreqs[biasSelects]
    for freq,dfreq,SZ in zip(freqSelects,dfreqSelects,szSelects):
        s_scanning(sample, SZ=SZ, name='S Scanning '+str(SZ), freq=np.linspace((freq-20*dfreq),(freq+20*dfreq),4000)*GHz)


"""
TODO: Coherent
"""

def coherent(sample, probeLen=st.r[0:200:2,ns], amps=st.r[0:1:0.05], freq=None, pulseLen=100*ns,
             mode='rectangle', pulseParam=None, couplingOn=None, paramName='Bus',
             stats=600L, measure=0, coupler=1, resonator=2, sb_freq=50*MHz,
             name='Coherent state', save=True, collect=True, noisy=True):
    """
    Calibration of photon numbers by varying swapTime to Fourier decompose into Rabi times.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    probeLen - list of times: Length of time to swap between qubit and resonator.
    amps - list of float: Amplitude of drive pulse
    freq - frequency: Resonator drive frequency.
    pulseLen - time: Length of drive pulse.
    mode - string: Type of envelope for drive pulse.
    pulseParam: Additional parameters needed for drive pulse envelope (eg. time constant)
    couplingOn - float: Squid dynamic bias during capture
    couplingOff - float: Squid dynamic bias for reading out (releasing) photon
    paramName - string: Name of swapAmp and swapTime registry keys
    stats - scalar: Number of times a point will be measured.
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    sb_freq - frequency: Sideband frequency of resonator drive source
    name - string: File name for data vault
    save - bool: Whether to save data to the datavault.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print out probabilities while the scan runs
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    c = devs[coupler]
    r = devs[resonator]
    
    name += ' '+mode.capitalize()
    if couplingOn is None:
        couplingOn = 0.0
    if freq is None:
        freq = r['f10']
    q['readout fc'] =  freq - sb_freq  
    axes = [(probeLen, 'Measure pulse length'),(amps, 'Resonator uwave drive Amp')]
    kw = {'stats': stats, 'biasCouplerDynamic':couplingOn, 'pulseLen':pulseLen,
          'pulseParams':pulseParam, 'Pulse Frequency':freq}
    dataset = sweeps.prepDataset(sample, name, axes, measure=measure, kw=kw)
    
    def func(server, currLen, currDrive):
        if mode== 'gaussian':
            r.xy = env.gaussian(pulseLen/2, pulseLen, amp=currDrive)
            len = pulseLen
        elif mode=='rectangle':
            r.xy = env.rect(0, pulseLen, currDrive)
            len = pulseLen
        elif mode=='exponential':
            timeConstant = pulseParam
            r.xy = env.exponential(0, pulseLen, timeConstant, currDrive)
            len = pulseLen
        elif mode=='exponentialBuildup':
            timeConstant, lenRectangle = pulseParam
            r.xy = env.rect(0,lenRectangle,currDrive)
            r.xy += env.exponential(lenRectangle, pulseLen, timeConstant, currDrive)
            len = pulseLen + lenRectangle
        else:
            raise Exception('Specified mode does not exist.')        
        r.xy = env.mix(r.xy, sb_freq)
        
        c.z = env.rect(0, len, couplingOn)
        q.z = env.rect(len, currLen, q['swapAmp'+paramName])
        q.z +=eh.measurePulse(q, len+currLen+5*ns)
        return runQubits(server, devs, stats, probs=[1])
    
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=noisy)
    if collect:
        return data