
import numpy as np

from pyle import envelopes as env
from pyle.dataking import envelopehelpers as eh
from pyle.registry import AttrDict

from labrad.units import Unit, Value

ns, GHz, us = (Unit(un) for un in ['ns', 'GHz', 'us'])

def getQubit(qubitType,name='q'):
    q = AttrDict()
    q['__name__'] = name
    q['piAmp'] = 0.6
    q['piLen'] = 20*ns
    q['piFWHM'] = 8*ns
    q['uwavePhase'] = 0.0
    q['readoutLen'] = 200*ns
    q['readoutAmp'] = 0.1
    q['readoutWidth'] = 10*ns
    q['readoutFrequency'] = 4.0*GHz
    q['readout fc'] = 4.1*GHz
    q['f10'] = 6.1*GHz
    q['f21'] = 6.05*GHz
    q['fc'] = 6.0*GHz
    q['measureAmp'] = 0.1
    q['measureLen'] = 200*ns
    return AttrDict2Qubit(q, qubitType)
    
def doStuff():
    q0 = getQubit(Transmon)
    q1 = getQubit(Transmon)
    alg = Algorithm()
    alg[PiPulse([qubit0])][Wait([qubit0], 20*ns)][PiPulse([qubit0])]
    alg[PiPulse([qubit1])][Wait([qubit1], 100*ns)][PiPulse([qubit1])]
    alg.compile()
    return qubit0,qubit1
    
###############
### DEVICES ###
###############

class Agent(AttrDict): #I hate this but I did it to not break code using the . as dict item access. Sorry.
    def __init__(self):
        AttrDict.__init__(self)
        self['_t'] = 0.0*ns
        self['gates'] = []
    
class Qubit(Agent):
    """
    Each Qubit object has a current time, self._t which is the current
    start time of this qubit's sequence.
    """
    def __init__(self):
        Agent.__init__(self)

    def excite(self):
        t = self['_t']
        l = self['piLen']
        self['xy'] += eh.mix(self, eh.piPulse(self, t-(l/2)))
        self['_t'] -= l
        
    def measure(self):
        raise Exception('Not implemented')
        
class PhaseQubit(Qubit):
    def measure(self, state=1):
        # Add measure pulse
        measureTime = self['_t']-(self['measureLenFall']+self['measureLenTop'])
        self['z'] += eh.measurePulse(self, measureTime, state=state)
        self['_t'] -= (self['measureLenFall']+self['measureLenTop'])
        # Add resonator readout pulse if:
        #    a) Reading out qubit
        #    b) 'readoutConfig' device exists and is a Readout object,
        if self.get('readout',False):
            readoutDevice = self.get('readoutDevice',None)
            if isinstance(readoutDevice,Readout):
                carrierFrequency = self['readoutDevice']['carrierFrequency']
                # carrierFrequency = q['readout fc']
                self['adc demod frequency'] = self['readoutFrequency'] - carrierFrequency
                df = self['readoutFrequency'] - carrierFrequency
                self['readoutAmp'] = eh.power2amp(self['readoutPower'])
                phase_shift = self.get('readoutPhaseShift', None)
                #Mix with the envelope starting at zero. No shifting going on here.
                #We simply build the readout signal referenced to "zero" and then
                #put in a start delay in fpgaseqTransmon.
                e1 = env.flattop(0, self['readoutLen'], w=self['readoutWidth'], amp=self['readoutAmp'])
                self['rr'] = env.mix(e1, df)
                if  phase_shift is not None and phase_shift != 0:
                    #print "Phase shifting readout envelope by %s pi rad" % (phase_shift/np.pi)
                    self['rr'] = np.exp(1j*phase_shift) * self['rr']
    
class Readout(Agent):
    pass
    
class CoherentPumpReadout(Readout):
    def __init__(self):
        Agent.__init__(self)
        self['qubits'] = []
    
    def measure(self):
        pumpSideband = self['pumpFrequency'] - (self['carrierFrequency']*self['pumpHarmonic'])
        self['readoutLen'] = max([q['readoutLen'] for q in self['qubits']])
        pumpDelay = self.get('pumpDelay', 0*ns)
        if pumpDelay['ns'] > 0:
            #print "Using pump delay of %s" % pumpDelay  
            pass
        # There is a 500 nanosecond extra pad on the end of the pump to capture the ring-down.
        # There is certainly a better way to do this.
        self['pp'] = env.mix(env.flattop(pumpDelay['ns'], self['readoutLen']-2*pumpDelay['ns']+1500*ns, w=self['readoutWidth'], amp=self['pumpAmplitude']), pumpSideband)
        #print "pump amplitude %s" % self['pumpAmplitude']
    
class SimpleReadout(Readout):
    def __init__(self):
        Agent.__init__(self)
        self['qubits'] = []
    
    def measure(self):
        pass
        
class Resonator(Agent):
    def __init__(self):
        Agent.__init__(self)
    
    def measure(self):
        pass
        
class TunableCoupler(Agent):
    def __init__(self):
        Agent.__init__(self)
    
    def measure(self):
        pass

class Transmon(Qubit):
    def measure(self):
        carrierFrequency = self['readoutDevice']['carrierFrequency']
        df = self['readoutFrequency'] - carrierFrequency
        #self['_readoutStartTime'] = - self['readoutLen']
        self['readoutAmp'] = eh.power2amp(self['readoutPower'])
        phase_shift = self.get('readoutPhaseShift', None)
        #Mix with the envelope starting at zero. No shifting going on here.
        #We simply build the readout signal referenced to "zero" and then
        #put in a start delay in fpgaseqTransmon.
        e1 = env.flattop(0, self['readoutLen'], w=self['readoutWidth'], amp=self['readoutAmp'])
        self['rr'] = env.mix(e1, df)
        if  phase_shift is not None and phase_shift != 0:
            #print "Phase shifting readout envelope by %s pi rad" % (phase_shift/np.pi)
            self['rr'] = np.exp(1j*phase_shift) * self['rr']
                
class TransmonPreEmphasis(Transmon):
    def measure(self):
        carrierFrequency = self['readoutDevice']['carrierFrequency']
        df = self['readoutFrequency'] - carrierFrequency
        #self['_readoutStartTime'] = - self['readoutLen']
        self['readoutAmp'] = eh.power2amp(self['readoutPower'])
        phase_shift = self.get('readoutPhaseShift', None)
        #Mix with the envelope starting at zero. No shifting going on here.
        #We simply build the readout signal referenced to "zero" and then
        #put in a start delay in fpgaseqTransmon.
        e1 = env.flattop(0, self['readoutLen'], w=self['readoutWidth'], amp=self['readoutAmp'])
        e2 = env.flattop(0, self['preEmphLen'], w=self['readoutWidth'], amp=self['readoutAmp']*(1-self['preEmphFactor']))
        self['rr'] = env.mix(e1+e2, df)
        if  phase_shift is not None and phase_shift != 0:
            #print "Phase shifting readout envelope by %s pi rad" % (phase_shift/np.pi)
            self['rr'] = np.exp(1j*phase_shift) * self['rr']

CONSTRUCTORS = {
                'phaseQubit': PhaseQubit,
                'transmon': Transmon,
                'resonator' : Resonator,
                'simpleReadout': SimpleReadout,
                'coherentPumpReadout': CoherentPumpReadout,
                'transmonPreEmph': TransmonPreEmphasis,
                'tunableCoupler': TunableCoupler
            }

def AttrDict2Qubit(q, qubitType):
    """Convert from a pure registry.AttrDict to a Qubit"""
    raise Exception('Depricated: use registryWrapper2GateCompiler instead')
    qubit = qubitType()
    #Copy all values from q into qubit
    for key,value in q.items():
        qubit[key] = value
    object.__setattr__(qubit, '__name__', q.__name__)
    return qubit
    
def AttrDict2QubitForSim(q, qubitType):
    """Convert from a pure registry.AttrDict to a Qubit"""
    qubit = qubitType()
    #Copy all values from q into qubit
    for key,value in q.items():
        qubit[key] = value
    object.__setattr__(qubit, '__name__', q.__name__)
    return qubit
    
def attrDict2GateCompilerObject(deviceAsAttrDict):
    devType = deviceAsAttrDict['_type']
    dev = CONSTRUCTORS[devType]()
    for key,value in deviceAsAttrDict.items():
        dev[key]=value
    object.__setattr__(dev, '__name__', deviceAsAttrDict.__name__)
    return dev
    
def registryWrapper2GateCompiler(sample, qubits):
    #Make a copy of all qubits
    devs = [q for q in qubits]
    if not any([q.get('readout', False) for q in qubits]):
        raise RuntimeError("No readout devices selected.  q['readout']=True must be set before calling registryWrapper2GateCompiler")
    #Add readout attr dicts to devs
    #for q in qubits:
    for q in [qubit for qubit in qubits if qubit['_type']=='transmon']: # resonators are not read out, only transmons
        readoutDev = q['readoutConfig']
        if readoutDev not in [d.__name__ for d in devs]:
            devs.append(sample[readoutDev])
    for q in [qubit for qubit in qubits if qubit['_type']=='phaseQubit']: # resonators are not read out, only transmons
        readoutDev = q.get('readoutConfig',None)
        if readoutDev:
            if readoutDev not in [d.__name__ for d in devs]:
                devs.append(sample[readoutDev])
    #Convert to gate compiler classes
    devs = [attrDict2GateCompilerObject(dev) for dev in devs]
    readoutDevices = [d for d in devs if isinstance(d,Readout)]
    #Map of device name -> device
    devPointers = {}
    for dev in devs:
        devPointers[dev.__name__] = dev
    #Give each qubit device a pointer to its readout device
    #and give each readout device a pointer to its qubits
    for dev in devs:
        readoutConfig = dev.get('readoutConfig', None)
        if readoutConfig:
            dev['readoutDevice'] = devPointers[readoutConfig]
            if isinstance(devPointers[readoutConfig], Readout):
                devPointers[readoutConfig]['qubits'].append(dev)
    #Add readout channels to the readout devices
    for readoutDevice in readoutDevices:
        for q in readoutDevice['qubits']:
            if q.get('readout', False):
                readoutDevice['channels'].append(('readout%s'%q.__name__, ('ADC', [readoutDevice['readoutADC']])))
    return devs
    

##################
### ALGORITHMS ###
##################

class Algorithm(object):

    def __init__(self, gates=None, agents=None):
        self.gates = [] if gates is None else gates
        self.agents = [] if agents is None else agents
                
    def __getitem__(self, gate):
        #For each agent involved in this gate, if we haven't seen this agent
        #before, add it to our list of agents.
        for agent in gate.agents:
            if agent not in self.agents: self.agents.append(agent)
        #Add this gate to our list of gates.
        self.gates.append(gate)
        return self
        
    def compile(self,correctXtalkZ=False,config=None,beSmart=False):
        """Compile the algorithm, updating all participating agents with relevant sequence data"""
        #Initialize all participating agents
        #correctCrosstalkZ = self.correctCrosstalkZ
        for agent in self.agents:
            agent['_t'] = 0.0*ns
            #set all envelopes to zero
            agent['xy'] = env.ZERO
            agent['xy_s'] = []
            agent['z'] = env.ZERO
            agent['z_s'] = []
            agent['rr'] = env.ZERO
            agent['xy_phase'] = 0.0 # this is for compensating for z phase shifts. e.g. from a swap
        #Compile the sequence
        if beSmart:
            #print 'being smart pass 1'
            for gate in self.gates[::-1]:
                gate.compile(smartrun=1)
            #print 'being smart pass 2'
            for gate in self.gates[::-1]:
                gate.compile(smartrun=2)
        else:
            for gate in self.gates[::-1]:
                gate.compile()            
        for agent in self.agents:
            agent['xy_s'] = [agent['xy']] + agent['xy_s']
            #print "agent %s has %d xy blocks" % (agent.__name__, len(agent['xy_s']))
            agent['z_s'] = [agent['z']] + agent['z_s']
            #print "agent %s has %d z blocks" % (agent.__name__, len(agent['z_s']))
        #Now we correct for the fact that we built the sequence backwards.
        #Find the longest sequence and shift all sequences to the right by that
        #amount.
        tShift = -1*min([ag.get('_t', 0.0*ns)['ns'] for ag in self.agents])
        for agent in self.agents:
            if isinstance(agent, PhaseQubit):
                #For transmons, don't shift readout pulse, because it wasn't built backwards
                #     as delay accounted for in fpgaseqTransmon.
                #For phase qubits, readout pulse is built backwards
                agent['rr'] = env.shift(agent['rr'], dt=tShift)
            agent['xy'] = env.shift(agent['xy'], dt=tShift)
            agent['z']  = env.shift(agent['z'],  dt=tShift)
            agent['xy_s'] = [env.shift(x, dt=tShift) for x in agent['xy_s']]
            agent['z_s']  = [env.shift(x,  dt=tShift) for x in agent['z_s']]
            #Store the time of the end of coherence manipulations, ie. the time
            #at which the redaout pulse starts. Note that this is shifted to
            #the right by tShift.
            agent['_t'] = tShift*ns
        
        #if correctXtalkZ:
            #correct for Z crosstalk
            #eh.correctCrosstalkZ(self.agents)

        if correctXtalkZ:
            #qubits need to be put in order for the Xtalk matrix

            if config is None:
                #alphabetical order: q0,q1,q2... etc
                names=[]
                for agent in self.agents:
                    names.append( agent.__name__ )                   
                n,qubits = zip( *sorted(zip(names,self.agents)))
            else:
                #in order of the config
                #master = ['q3','q9','q8','q5']
                #target = ['q5','q8','q3']
                #indices = [master.index(element) for element in target]
                #i,targetnew = zip( *sorted(zip(indices,target)))
                #out: ['q3','q8','q5']
                
                qubits=[]
                names=[]
                for qname in config:
                    for agent in self.agents:
                        if agent.__name__==qname:
                            qubits.append(agent)                            
                            names.append( agent.__name__ )
            print 'gateCompiler: order for Z Xtalk: ', names
                
            #correct for Z crosstalk
            qub=[]
            for q in qubits:
                if q['_type'] == 'transmon':
                    qub.append(q)
                
            eh.correctCrosstalkZ( qub )
            
    def compileNoShift(self):
        """Compile the algorithm, updating all participating agents with relevant sequence data"""
        #Initialize all participating agents
        for agent in self.agents:
            agent['_t'] = 0.0*ns
            #set all envelopes to zero
            agent['xy'] = env.ZERO
            agent['z'] = env.ZERO
            agent['rr'] = env.ZERO
        for gate in self.gates[::-1]:
            gate.compile()

def PrintGateEnvelopes(agents):
    idx=0
    for q in agents:
        print "agent %s has t_: %f" % (q['__name__'], q['_t'])
        for channel in ['xy_s', 'z_s']:
            for idx, e in enumerate(q[channel]):
                print "Envelope %d for agent %s channel %s has span (%f,%f)" % (idx, q['__name__'],channel, e.start, e.end)
##############
### EXAMPLE ##
##############
"""
from pyle.dataking import util
from pyle import gateCompiler as gc
sample, qubits = util.loadQubits(s)
q = qubits[0]
q['readout']=True
q = gc.registryWrapper2GateCompiler(sample, [q])[0]
#q = gc.AttrDict2Qubit(q, gc.Transmon)
alg = gc.Algorithm()
alg[gates.PiPulse([q])][gates.Wait([q], 10*ns)][gates.PiPulse([q])]
alg.compile()

t=np.linspace(-500,500,1001)
s=q['xy'](t)

plt.figure()
plt.plot(t,s)
plt.show()
"""



"""
import time
from pyle.dataking import util
from pyle import gateCompiler as gc

sample, qubits = util.loadQubits(s)
q = qubits[0]
q['readout']=True
q = gc.registryWrapper2GateCompiler(sample, [q])[0]
#q = gc.AttrDict2Qubit(q, gc.Transmon)

alg = gc.Algorithm()
t0a=np.linspace(0,1000,201)*ns
for t0 in t0a:
    alg[gates.PiPulse([q])]
alg.compile()

tic=time.time()
t=np.linspace(-500,500,1001)
s=q['xy'](t)
toc=time.time()
print toc-tic

alg.compile(beSmart=True)

tic=time.time()
t=np.linspace(-500,500,1001)
s=q['xy'](t)
toc=time.time()
print toc-tic

plt.figure()
plt.plot(t,s)
plt.show()
"""
