import numpy as np
import inspect
import labrad
from labrad.types import Value

import pyle
from pyle.pipeline import pmap, returnValue

def analyze_caller(start_depth=1, max_frames=float("inf")):
    '''
    This function inspects the call stack and creates a dictionary from the call stack
    which can be placed in a data set.  start_depth is the number of frames to skip
    including analyze_caller itself.  At most max_frames stack frames will be inspected.
    '''
    frames = inspect.stack()
    nframes = min(max_frames, len(frames)-start_depth)
    parameters={}
    for k in range(nframes):
        frame = frames[k+start_depth]
        parameters['stack.frame%d.file' % k] = frame[1]
        parameters['stack.frame%d.func' % k] = frame[3]
        parameters['stack.frame%d.line' % k] = frame[2]
        args, varargs, keywords, locals = inspect.getargvalues(frame[0])
        parameters['stack.frame%d.args' % k] = inspect.formatargvalues(args, varargs, keywords, locals)
        for a in args:
            parameters['stack.frame%d.arg.%s' % (k, a)] = repr(locals[a])
        if frame[1] == '<ipython console>':
            break
    return parameters


def prepDataset(sample, name, axes=None, dependents=None, measure=None, cxn=None, kw=None, explicitIndependents=None):
    """Prepare dataset for a sweep.
    
    This function builds a dictionary of keyword arguments to be used to create
    a Dataset object for a sweep.  Sample should be a dict-like object
    (usually a copy of the sample as returned by loadQubits) that contains current
    parameter settings.  Name is the name of the dataset, which will get prepended
    with a list indicating which qubits are in the sample config, and which of them
    are to be measured.  kw is a dictionary of additional parameters that should
    be added to the dataset.
    
    Axes can be specified explicitly as a tuple of (<name>, <unit>), or else
    by value for use with grid sweeps.  In the latter case (grid sweep), you should
    specify axes as (<value>, <name>).  If value is iterable, the axis will be
    added to the list of dependent variables so the value can be swept (we look
    at element [0] to get the units); if value is not iterable it will be added
    to the dictionary of static parameters.
    
    Dependents is either a list of (<name>, <label>, <unit>) designations for the
    dependent variables, or None.  If no list is provided, then the dependents are
    assumed to be probabilities.  In this case, the measure variable is used to
    determine the appropriate set of probabilities: for one qubit, we assume only
    P1 will be measured, while for N qubits all 2^N probabilities are assumed to
    be measured, in the order P00...00, P00...01, P00...10,..., P11...11.  If this
    is not what you want, you must specify the independent variables explicitly.
    
    Note that measure can be None (all qubits assumed to be measured), an integer
    (just one qubit measured, identified by index in sample['config']) or a list
    of integers (multiple qubits measured).
    
    explicitIndependents are independent variables which are explicitly included
    in the argument of returnValue in func. These variables are added after any
    independent variables given by axes. It must be either None (default) or a
    list of tuples. If explicitIndependents is not None, the tuples must be of the
    form ('label','unit'), where 'label' is the label for the variable, which has
    unit 'unit'.
    """
    conf = list(sample['config'])
    
    # copy parameters
    kw = {} if kw is None else dict(kw)
    kw.update(sample) # copy all sample data
    
    if measure is None:
        measure = range(len(conf))
    elif isinstance(measure, (int, long)):
        measure = [measure]
    
    if hasattr(measure, 'params'):
        # this is a Measurer
        kw.update(measure.params())
    else:
        kw['measure'] = measure
    
    # Store information about the call stack.  This is useful if you copy
    # a scan or file but don't update the dataset name.  start_depth=2
    # skips prepDataset's stack frame.
    kw.update(analyze_caller(start_depth=2, max_frames=2))
    # update dataset name to reflect which qubits are measured
    for i, q in enumerate(conf):
        if i in kw['measure']:
            conf[i] = '|%s>' % q
    name = '%s: %s' % (', '.join(conf), name)
    
    # create list of independent vars
    independents = []
    for num, (param, label) in enumerate(axes):
        if isinstance(param, str):
            # param specified as string name
            independents.append((param, label))
            print 'Parameter %d: %s' %(num,param)
        elif np.iterable(param):
            # param value will be swept
            try:
                units = param[0].units
            except Exception:
                units = ''
            independents.append((label, units))
            print 'Parameter %d: %s' %(num,label)
        else:
            # param value is static
            kw[label] = param
            print 'Single Parameter %d: %s' %(num,label)
    if explicitIndependents is not None:
        for indep in explicitIndependents:
            independents.append(indep)
    
    # create list of dependent vars
    if dependents is None:
        if hasattr(measure, 'dependents'):
            # this is a Measurer
            dependents = measure.dependents()
        else:
            n = len(measure)
            if n == 1:
                labels = ['|1>']
            else:
                labels = ['|%s>' % bin(i)[2:].rjust(n,'0') for i in xrange(2**n)]
            dependents = [('Probability', s, '') for s in labels]
    
    return pyle.Dataset(
        path=list(sample._dir),
        name=name,
        independents=independents,
        dependents=dependents,
        params=kw,
        cxn=cxn
    )


def run(func, sweep, save=True, dataset=None,
        abortable=True, abortPrefix=[],
        collect=True, noisy=True, pipesize=10):
    """Run a function pipelined over an iterable sweep.
    
    func: function that will be called once for each value in the sweep.
          should be written to work with pipelining, and should
          return sequence objects ready to be sent to the qubit sequencer.
    sweep: an iterable which returns successive values to be
           passed to func
    
    abortable: if True, check for keypresses to allow the sweep to be aborted cleanly
    save: if True, create a new dataset (using ds_info) and save all data to it
    collect: if True, collect the data into an array and return it
    noisy: if True, print each row of data as it comes in
    
    dataset: a dataset that will be called with the iterable of data to be saved
    pipesize: the number of pipelined calls to func that should be run in parallel

    
    The following additional parameters usually only need to be modified for defining
    new types of sweep, such as grid:
    
    abortPrefix: passed along to checkAbort for abortable sweeps
    """    
    # wrap the function to pass the qubit sequencer as the first param
    def wrapped(val):
        ans = yield func(None, val)
        ans = np.asarray(ans)
        returnValue(ans)
        
    def singleParam(params,num):
        if not len(params):
            yield ()
        else:
            param, rest = params[0], params[1:]
            print 'Parameter %d' %num
            try:
                units = param.units
                print '     Original Value: %f %s' %(param.value,param.units)
                val = Value(float(raw_input('     Desired Value [%s]: ' %units)),units)
            except Exception:
                print '     Original Value: %f' %param
                val = float(raw_input('     Desired Value: '))
            for all in singleParam(rest,num+1):
                yield (val,) + all
        
    def getParams(sweep):
        firstSweep = sweep.next()
        yield singleParam(firstSweep,0).next()
    
    # Build the generator that returns results of func. Note that this generator
    # doesn't execute any code yet, and won't until .next() is called. 
    iter = pmap(wrapped, getParams(sweep), size=pipesize)
    # run the iterable and discard
    for _ in iter: pass
    
    
def grid(func, axes, **kw):
    """Run a pipelined sweep on a grid over the given list of axes.
    
    The axes should be specified as a list of (value, label) tuples.
    We iterate over each axis that is iterable, leaving others constant.
    Func should be written to return only the dependent variable data
    (e.g. probabilities), and the independent variables that are being
    swept will be prepended automatically before the data is passed along.
    
    All other keyword arguments to this function are passed directly to run.
    """
    def gridSweep(axes):
        if not len(axes):
            yield ()
        else:
            (param, _label), rest = axes[0], axes[1:]
            if np.iterable(param): # TODO: different way to detect if something should be swept
                for val in param:
                    for all in gridSweep(rest):
                        yield (val,) + all
            else:
                for all in gridSweep(rest):
                    yield (param,) + all
    
    # pass in all params to the function, but only prepend swept params to data
    def wrapped(server, args):
        all = args
        ans = yield func(None, *all)
        ans = np.asarray(ans)
        returnValue(ans)
    
    return run(wrapped, gridSweep(axes), abortPrefix=[1], **kw)