import numpy as np
import scipy.optimize

from pyle.util.structures import ValueArray
from pyle.dataking.DevJim.QND import couplerResonator as cr

import matplotlib.pyplot as plt

### UTILITY

#def nonCompactPhase(xs,ys):
#    """Get phase angle of a set of x/y points
#    
#    The phases returned by this function will _not_ lie on a compact set. In
#    other words, if the points used as input represent three cycles around a
#    circle, the output phases will run from 0 to 3*(2pi), rather than going
#    periodically from 0 to 2pi and around again.
#    """    
#    phases_rad = np.array([])
#    rotator = 1.0
#    lastPhase_rad = 0.0
#    for x,y in zip(xs,ys):
#        z = x + (1.0j*y)
#        dPhase_rad = np.angle(rotator*z)
#        phases_rad = np.hstack((phases_rad, lastPhase_rad + dPhase_rad))
#        lastPhase_rad = lastPhase_rad + dPhase_rad
#        rotator = np.exp(-1.0j*lastPhase_rad)
#    return phases_rad
    
def optimize(dep, indep, model, guess):
    """Find parameters to fit a model function to data"""
    def errorFunc(v, indep, dep):
        return abs(model(indep, v) - dep)
    result = scipy.optimize.leastsq(errorFunc, guess, args=(indep, dep))
    return result
    
#### MODEL FUNCTIONS FOR DIP RESONATORS WITHOUT ANY CALIBRATION OF THE SCATTERING DATA
#
#def dipResonatorModel(freq, f0, Qi, Qc, phi, phasePerFreq, phaseOffs, magOffRes_dB):
#    """S12^-1 as a function of frequency and other parameters for a dip resonator
#    
#    f0 - resonance frequency
#    Qi - internal quality factor
#    Qc - coupling quality factor
#    phi - a phase angle arising from mismatched impedances near the resonator
#    phasePerFreq - phase accumulation/frequency of cabling
#    phaseOffs - constant offset phase
#    magOffRes_dB - off resonant transmission magnitude in dB
#    """
#    return (10**(-1.0*magOffRes_dB/20.0))*\
#           np.exp(-1.0j*2*np.pi*(phasePerFreq*freq))*\
#           (1+(Qi/Qc)*np.exp(1.0j*phi)/(1+2.0j*Qi*(freq/f0-1)))
#    
#def dipResonatorModel_leastsq(freq, v):
#    """S21^-1 as a function of frequency and other parameters of a dip resonator
#    
#    v[0] - f0
#    v[1] - Qi
#    v[2] - Qc
#    v[3] - phi
#    v[4] - slope of phase vs frequency in rad/GHz
#    v[5] - offset phase
#    v[6] - off resonant magnitude in dB (voltage dB is 20*log10(number))
#    
#    The purpose of this function is to wrap up the fit parameters into a
#    single array so that the function is compatible with
#    scipy.optimize.leastsq. It is just a passthrough to
#    dipResonatorModel with all nonleading arguments wrapped into an array.
#    """
#    return dipResonatorModel(freq, *v)
#    
#### MODEL FUNCTIONS FOR DIP RESONATORS ASSUMING BACKBROUND PHASE AND AMPLITUDE HAS BEEN REMOVED
#    
#def dipResonatorModel_calibrated(freq, f0, Qi, Qc, phi):
#    """S21^-1 as a function of frequency and other parameters of a dip resonator
#    
#    f0 - resonance frequency
#    Qi - internal quality factor
#    Qc - coupling quality factor
#    phi - angular offset of the resonance circle
#    
#    This model assumes that scattered voltage has been corrected for
#    background amplitude and phase. Therefore, the off resonant point
#    should be at (1,0) in the complex plane, and the circle should extend
#    toward the positive real axis, NOT toward the origin.
#    """
#    return 1+(Qi/Qc)*np.exp(1.0j*phi)/(1+2.0j*Qi*(freq/f0-1))
#    
#def dipResonatorModel_calibrated_leastsq(freq, v):
#    """ Fit S^-1 with the cable delay and background transmission removed.
#    
#    This function is designed with the fit parameters as a single array
#    for compatibility with scipy.optimize.leastsq.
#    
#    v[0] - f0
#    v[1] - Qi
#    v[2] - Qc
#    v[3] - phi
#    
#    The purpose of this function is to wrap up the fit parameters into a
#    single array so that the function is compatible with
#    scipy.optimize.leastsq. It is just a passthrough to
#    dipResonatorModel_calibrated with all nonleading arguments wrapped
#    into an array.
#    """
#    return dipResonatorModel_calibrated(freq, *v)
#    
#### MODEL FUNCTIONS FOR PARAMP GAIN AND PHASE DELAY
#    
#def parampGainModel(freq, f0, A, Gamma, B, C):
#    """Lorentzian function with additional linear background"""
#    return (A/np.pi)*(0.5*Gamma)/(((freq-f0)**2)+((0.5*Gamma)**2)) + B*freq + C
#    
#def parampGainModel_leastsq(freq, v):
#    return parampGainModel(freq, *v)
#    
#def parampPhaseModel(freq, f0, width, offset_rad):
#    return np.arctan((freq-f0)/width) + offset_rad
#    
#def parampPhaseModel_leastsq(freq, v):
#    return parampPhaseModel(freq, *v)
    
### FITTING
    
def fit_circle(x, y):
    """Get the center and radius of a circle
    
    RETURNS
    (xc_2b, yc_2b, R_2b)
     xc_2b: x coordinate of the circle's center
     yc_2b: y coordinate of the circle's center
     R_2b: fitted radius of the circle
     
    I have no idea how this function works - DTS
    """
    x_m = np.mean(x)
    y_m = np.mean(y)
    def calc_R(xc, yc):
        return np.sqrt((x-xc)**2 + (y-yc)**2)
    def f_2b(c):
        Ri = calc_R(*c)
        return Ri - Ri.mean()
    def Df_2b(c):
        xc, yc     = c
        df2b_dc    = np.empty((len(c), x.size))
        Ri = calc_R(xc, yc)
        df2b_dc[ 0] = (xc - x)/Ri
        df2b_dc[ 1] = (yc - y)/Ri
        df2b_dc       = df2b_dc - df2b_dc.mean(axis=1)[:, np.newaxis]
        return df2b_dc
    center_estimate = x_m, y_m
    center_2b, ier = scipy.optimize.leastsq(f_2b, center_estimate, Dfun=Df_2b, col_deriv=True)
    xc_2b, yc_2b = center_2b
    Ri_2b        = calc_R(xc_2b, yc_2b)
    R_2b         = Ri_2b.mean()
    residu_2b    = sum((Ri_2b - R_2b)**2)
    residu2_2b   = sum((Ri_2b**2-R_2b**2)**2)
    return xc_2b, yc_2b, R_2b
    
def dipResonatorGuess_calibrated(freqs, Sinv):
    """Get rough fit parameters for calibrated dip resonator
    
    Here it is assumed that S inv has the background transmission
    and cable phase delays removed
    
    RETURNS
    f0, Q, Qc, phi_rad
    
    f0 - resonance frequency
    Qi - internal Q factor
    Qc - coupling Q factor
    phi_rad - circle angle offset in radians
    """
    x = Sinv.real
    y = Sinv.imag
    mags = np.sqrt(x**2 + y**2)
    #Find the center and radius of the resonance circle
    xc, yc, r = fit_circle(x,y)
    #Estimate the phi and resonance frequency
    phi_rad = np.arctan2(yc, xc-1)
    f0 = freqs[mags.index(np.max(mags))]
    #Now find the 3dB points so we can get Q.
    #Translate the circle so that it's center lies on the origin,
    Sinv = Sinv - (xc + 1.0j*yc)
    #rotate it so that the resonance point is on the real axis,
    Sinv = Sinv / np.exp(1.0j*phi_rad)
    #and then find the frequencies at which the y coordinate is maximum
    #and miniumum.
    phases = np.angle(Sinv)
    find = lambda a,l:min(l,key=lambda x:abs(x-a))
    Pos3DbF = freqs[p.index(find(np.pi/2,p))]
    Neg3DbF = freqs[p.index(find(-1*np.pi/2,p))]
    #deltaF is the difference between these two frequencies.
    deltaF = abs(Pos3DbF - Neg3DbF)
    Qi = f0/deltaF

    #TODO: check that fits make sense!
    #TODO: check that this definition of Qc is correct
    return [f0, Qi, Q/(2*r), phi_rad]
    
def resonanceFit(dir, dataNum, calFunction, calData=None):
    #Calibrate and normalize the data
    SCalibrated = calFunction(q, sData, calData)
    
    #Go from S to S^-1
    Sinv = SCalibrated**(-1.0)
    #Find guesses for fit parameters
    guess = dipResonatorGuess_calibrated(freqs, Sinv)
    #Fit the data
    fitResult = optimize(freq_GHz, Sinv, dipResonatorModel_calibrated_leastsq, guess)
    f0, qi, qc, phi, phasePerFreq, phaseOffs, magOffRes_dB = fitResult[0]
    #return <stuff>
    
### DATA ACQUISITION SCRIPTS

#For a fixed sideband frequency f, the only phase contributions are from the
#device under test (resonator) and the phase delay through the cables.
    
def linearCalibrate(dir, dataNum, calData=None):
    calibratedData = cr.circlePlot(dir,dataNum,calData=True,xGuess=0,yGuess=0,pGuess=[0.0,0.0],ampRescale=True,calUpdate=True)
    return np.array([calibratedData[0],calibratedData[1]+1j*calibratedData[2]]).T

#def parampResponse(sample, measure, freq, stats=150):
#    tr.readoutIq(sample, freq=freq, measure=measure, stats=1200, dataFormat='iq',
#              save=True, plot=False, name='Readout IQ')
#    with labrad.connect() as cxn:
#        dv = cxn.data_vault
#        dataset = dstools.getDeviceDataset(dv,-1,session=sample._dir)
#    gainParams, phaseParams = getParampParameters(dataset)
#    
#    
#    
#def getParampParameters(dataset):
#    """This is a long function"""
#    #Get the measured qubit's parameters
#    config = dataset['parameters']['config']
#    measure = dataset['parameters']['measure'][0]
#    measuredQubit = config[measure]
#    q = dataset['parameters'][measuredQubit]
#    #Parse data
#    I = dataset.data[:,1]
#    Q = dataset.data[:,2]
#    mag = np.sqrt(I**2 + Q**2)
#    phase_rad = nonCompactPhase(I, Q)
#    freq = ValueArray(dataset.data[:,0],dataset.variables[0][0][1])
#    #GAIN PARAMETER FIT
#    #Plot curves and ask user for fit ranges
#    fig = plt.figure()
#    ax = fig.add_subplot(111)
#    ax.plot(freq['GHz'], mag, '.')
#    plt.show()
#    low = float(raw_input('Low frequency cutoff [GHz]: '))
#    high = float(raw_input('High frequency cutoff [GHz]: '))
#    #Make array mask and mask data
#    mask = np.logical_or(freq['GHz']<low, freq['GHz']>high)
#    freqMasked_GHz = freq['GHz'][mask]
#    magMasked = mag[mask]
#    #Show points being fit
#    ax.plot(freqMasked_GHz, magMasked, 'r.')
#    plt.show()
#    #Guess parameters and do the fit
#    idx = np.argmax(magMasked)
#    f0Guess_GHz = freqMasked_GHz[idx]
#    gammaGuess = 0.01 #TODO: guess this more intelligently
#    AGuess = magMasked[idx]*np.pi*gammaGuess/2.0
#    guess = [f0Guess_GHz, AGuess, gammaGuess, 0, 0]
#    result = optimize(magMasked, freqMasked_GHz, parampGainModel_leastsq, guess)
#    gainParams = result[0]
#    #Show the fit line
#    ax.plot(freq['GHz'], parampGainModel_leastsq(freq['GHz'], gainParams), 'r')
#    plt.show()
#    #PHASE PARAMETER FIT
#    expectedCablePhase_rad = phase_rad[0] + 2*np.pi*q['calReadoutCableDelay']['ns']*(freq['GHz']-freq['GHz'][0])
#    residualPhase_rad = phase_rad - expectedCablePhase_rad
#    residualPhaseMasked_rad = residualPhase_rad[mask]
#    #Plot the curve
#    fig = plt.figure()
#    ax = fig.add_subplot(111)
#    ax.plot(freq['GHz'], residualPhase_rad, 'b.')
#    ax.plot(freqMasked_GHz, residualPhaseMasked_rad, 'r.')
#    plt.show()
#    #Fit
#    guess = [gainParams[0], 0.005, np.pi] #XXX More intelligent guess for width needed, should be able to get this from gamma in the gain fit
#    result = optimize(residualPhaseMasked_rad, freqMasked_GHz, parampPhaseModel_leastsq, guess)
#    phaseParams = result[0]
#    #Show the fit line
#    ax.plot(freq['GHz'], parampPhaseModel_leastsq(freq['GHz'], phaseParams), 'r')
#    return gainParams, phaseParams
#    
#def correctParampResponse(q, freq, V):
#    gainParams = q['calReadoutParampGainFunc'].asarray
#    delay = q['calReadoutParampDelay']+q['calReadoutCableDelay']
#    #Compute gain curve
#    expectedGain = parampGainModel(freq, *gainParams)
#    #Compute phase background
#    theta0 = np.angle(V[0])
#    theta = theta0 + 2*np.pi*delay['ns']*(freq-freq[0])
#    expectedV = expectedGain * np.exp(1.0j*theta)
#    return V/expectedV
    