import numpy as np
import itertools
import labrad
from labrad.units import ns, GHz, MHz, mV, V, dBm, us,rad
from labrad.units import Unit,Value
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import pyplot as plt
from scipy.optimize import leastsq

from pyle.pipeline import returnValue, FutureList
from pyle.dataking import util
from pyle import envelopes as env
from pyle.dataking import envelopehelpers as eh
from pyle.dataking import sweeps
from pyle.dataking.DevJim.QND.fpgaseq_QND import runQubits
from pyle.util import sweeptools as st
from pyle.plotting import dstools
from pyle.dataking.DevJim.QND import fpgaseq_QND
from pyle.fitting import fitting
from pyle.dataking.DevJim.QND import couplerResonator as cr
from pyle.analysis import numberAnalysis as na

fpgaseq_QND.PREPAD = 1000


"""
TODO: Pulse
"""

def pulseIQ(sample, amps=None, freqs=None, captureBias=0.0, pulseLens=100*ns, readBias=None, couplerDelays=0*ns,
            mode='rectangle', pulseParam=None, delays=0, releaseLens=2000*ns, sb_freq=50*MHz, resolution=2,
            save=False, stats=150, statsCut=3000, measure=0, name='', collect=False, noisy=False):
    """ Apply a pulse on resonator with coupler open, close coupler, then open up to
    release photon.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    amps - float: Amplitude of drive pulse
    freqs - frequency: Resonator drive frequency.
    captureBias - float: Squid dynamic bias during capture
    pulseLens - time: Length of input pulse
    readBias - float: Squid dynamic bias for reading out (releasing) photon
    mode - str: Type of input pulse
    pulseParam - ?: Additional parameters for input pulse
    delays - time: Time during which coupling turned off
    releaseLens - time: Duration of release pulse.
    sb_freq - frequency: Sideband frequency of resonator drive source
    resolution - float (in ns): Time resolution for average mode. Max is 2ns.
    save - bool: Whether to save data to the datavault.
    stats - scalar: Number of times a point will be measured. Total signal is sum
        of each signal (so average is trace/states).
    statsCut - scalar: Number of times a point will be measured per rep. Total reps
        per point is stats/statsCut (rounded up).
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    name - str: Directory to which to save data.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print status updates
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    
    q['readoutType'] = 'resonator'
    q['biasReadout'] = q['biasOperate']
    q['adc_start_delay'] = 0*ns
    q['readout DAC start delay'] = 0*ns
    q['adc mode'] = 'average'
    q['adc demod frequency'] = sb_freq
    if amps is None:
        amps = q['readout amplitude']
    if freqs is None:
        freqs = q['readout frequency']
    if readBias is None:
        readBias = captureBias
    
    name = 'IQ '+mode.capitalize()+' '+name
    axes = [(amps,'Pulse Amplitude'),(freqs,'Pulse Frequency'),(pulseLens,'Pulse Length'),
            (delays,'Release Delay'),(releaseLens,'Release Length'),(couplerDelays,'Coupler Delays')]
    deps = [('Clicks', 'I for %s' %q.__name__, ''), ('Clicks', 'Q for %s' %q.__name__, '')]
    kw = {'stats':stats,
          'biasCouplerCapture':captureBias,
          'biasCouplerRead':readBias,
          'pulseMode':mode,
          'pulseParams':pulseParam,
          'resolution':resolution}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw,
                                 explicitIndependents=[('Time','ns')])
    
    def func(server, amp, freq, pulseLen, delay, releaseLen,couplerDelay):
        data_output = np.empty([0,3])
        for subpixel in list(st.r[0:(2-.5*resolution):resolution,ns]):
            statsLeft=stats
            count = 0
            Is = np.zeros(8192)
            Qs = np.zeros(8192)
            while statsLeft>0:
                q['readout fc'] = freq - sb_freq
                q['readout'] = True
                t = 0
                if mode=='gaussian':
                    t += pulseLen/2
                    q.rr = env.gaussian(t, pulseLen, amp=amp)
                    t += pulseLen/2
                elif mode=='rectangle':
                    q.rr = env.rect(t, pulseLen, amp=amp)
                    t += pulseLen
                elif mode=='exponential':
                    timeConstant = pulseParam
                    q.rr = env.exponential(t, pulseLen, timeConstant, amp)
                    t += pulseLen
                elif mode=='exponentialBuildup':
                    timeConstant, lenRectangle = pulseParam
                    q.rr = env.rect(t,lenRectangle,amp)
                    t += lenRectangle
                    q.rr += env.exponential(t, pulseLen, timeConstant, amp)
                    t += pulseLen
                else:
                    raise Exception('Specified mode does not exist.')
                q.rr = env.mix(q.rr, sb_freq)
                q.cz = env.rect(0,t+couplerDelay,captureBias)
                q.cz += env.rect(t+delay+couplerDelay,releaseLen,readBias)
                
                q.rr = env.shift(q.rr, subpixel)
                q.cz = env.shift(q.cz, subpixel)
                
                if statsLeft>statsCut:
                    statsNow = statsCut
                else:
                    statsNow = statsLeft
                statsLeft -= statsNow
                count += 1
                if noisy:
                    print delay,count
                ans = yield runQubits(server, qubits, statsNow, dataFormat='iq')
                Is += np.asarray(ans[0][0])
                Qs += np.asarray(ans[0][1])
            data_output = np.vstack((data_output,1.0*np.vstack((np.array(range(0,2*len(Is),2))-subpixel[ns],Is,Qs)).T))
        returnValue(data_output)
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    if collect:
        return data
    

def freqScan(sample, pulseLen, couplerDelay, pulseParam, captureBias=0.60, mode='exponential', stats=1500000, scale=None):
    if mode == 'exponential' and pulseParam>0:
        maxAmp = .03 * (1+90./pulseParam) / (np.exp(float(pulseLen)/pulseParam)-np.exp(-pulseLen/90.))
    else:
        raise Exception('maxAmp not yet defined.')
    ampRange = st.r[0:maxAmp:(maxAmp/100.)]
    if scale == None:
        cr.coherent(sample, mode=mode, pulseLen=pulseLen*ns, pulseParam=pulseParam*ns, amps=ampRange,
                    freq=6.5525*GHz, couplingOn=captureBias, noisy=False, collect=False)
        with labrad.connect() as cxn:
            dv = cxn.data_vault
            scale = na.coherentanalysis(dv,-1,sample._dir,minContrast=1.2,plot=False,qT1=250.,nfit=15)[2]
    freqRange = [freq for freq in st.r[6.552:6.555:.0005]]
    for freq in freqRange:
        pulseIQ(sample, mode=mode, pulseLens=pulseLen*ns, pulseParam=pulseParam*ns, freqs=freq*GHz,
                couplerDelays=couplerDelay*ns, amps=np.sqrt(30)/scale, resolution=1, delays=30*ns,
                name='Len%d Tau%d Freq%.5f Delay%.1f' %(pulseLen,pulseParam,freq,couplerDelay),
                captureBias=captureBias, stats=stats, statsCut=6000, save=True)


"""
TODO: Plots
"""
    
def plotPulseIQ(dir,datanum,plot=True, timeInit=400, ratioQtoI=1.0, delaySplit=None, maxRelease=None,
                collect=False, noisy=True, plotInit=0, plotRawIQ=False, remove10MHz=False, resolutionSplit=True):
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datanum, session=dir)
        dv.cd(dir)
        dv.open(datanum)
        title = dv.get_name().split('- ')[1]
    data = dataset.data
    stats = dataset.parameters['stats']
    resolution = float(dataset.parameters.get('resolution',2))
    
    # Determine all time scales. Pulse between timeInitScaled and timePulse. Release between
    # timeDelay and timeEnd. Sequence ended at timeComplete
    timeInitScaled = timeInit/resolution
    timePulse = timeInitScaled + dataset.parameters['Pulse Length'][ns]/resolution
    if ' Exponentialbuildup ' in title:
        timePulse += dataset.parameters['pulseParams'][1][ns]/resolution
    delay = dataset.parameters['Release Delay']
    if delay>0:
        timeDelay = timePulse + delay[ns]/resolution
    else:
        timeDelay = timePulse
    timeComplete = timeDelay + dataset.parameters['Release Length'][ns]/resolution
    if maxRelease is None:
        timeEnd = timeComplete
    else:
        timeEnd = timeDelay + maxRelease/resolution

    # Remove DC offset from each subpixel trace and sort data.
    if resolutionSplit:
        vals = (data[:,1]+1j*data[:,2]/ratioQtoI)/stats
        acVals = np.array([vals[i*8192:(i+1)*8192]-np.mean(vals[i*8192:(i*8192+.75*timeInit/2)]) for i in range(len(vals)/8192)]).flatten()
        zeroedData = np.array([data[:,0],np.real(acVals),np.imag(acVals)]).T
        dataSort = zeroedData[np.argsort(data[:,0])]
    else:
        vals = data[:,1]+1j*data[:,2]/ratioQtoI
        acVals = vals-np.mean(vals[:(.75*timeInitScaled)])
        dataSort = np.array([data[:,0],np.real(acVals),np.imag(acVals)]).T
    times = dataSort[:,0]
    reals = dataSort[:,1]
    imags = dataSort[:,2]
    
    # Remove 10MHz background sine wave (from clock)
    if remove10MHz:
        timeSine = np.append(times[(.1*timeInitScaled):(.75*timeInitScaled)],times[(1.1*timeComplete):])
        realSine = np.append(reals[(.1*timeInitScaled):(.75*timeInitScaled)],reals[(1.1*timeComplete):])
        imagSine = np.append(imags[(.1*timeInitScaled):(.75*timeInitScaled)],imags[(1.1*timeComplete):])
        guess = np.array([.06*stats,0.01,0.5,0])
        #sine(x, amp, freq, phase, offset)
        realFit, cov, realFitFunc = fitting.fitCurve('sine', timeSine, realSine, guess)
        imagFit, cov, imagFitFunc = fitting.fitCurve('sine', timeSine, imagSine, guess)
        realZeroed = reals-(realFit[3]+realFit[0]*np.sin(2*np.pi*(times*realFit[1]+realFit[2])))
        imagZeroed = imags-(imagFit[3]+imagFit[0]*np.sin(2*np.pi*(times*imagFit[1]+imagFit[2])))
        zeroedVals = realZeroed + 1j*imagZeroed
    else:
        zeroedVals = reals+1j*imags
    
    if delaySplit is not None:
        timeSplit = timePulse+delaySplit*(timeDelay-timePulse)
        pulseNoise = resolution*np.sum(np.abs(zeroedVals[(0.15*timeInitScaled):(0.85*timeInitScaled)])**2)/(0.7*timeInitScaled)
        pulseReflect = resolution*np.sum(np.abs(zeroedVals[timeInitScaled:timePulse])**2)
        pulseClosed = resolution*np.sum(np.abs(zeroedVals[timePulse:timeSplit])**2)
        pulseCatch = resolution*np.sum(np.abs(zeroedVals[timeSplit:timeEnd])**2)
    else:
        pulseNoise = resolution*np.sum(np.abs(zeroedVals[(0.15*timeInitScaled):(0.85*timeInitScaled)])**2)/(0.7*timeInitScaled)
        pulseReflect = resolution*np.sum(np.abs(zeroedVals[timeInitScaled:timePulse])**2)
        pulseClosed = resolution*np.sum(np.abs(zeroedVals[timePulse:timeDelay])**2)
        pulseCatch = resolution*np.sum(np.abs(zeroedVals[timeDelay:timeEnd])**2)
        
    if noisy:
        print 'Average Noise Power (LSBs^2ns): ',pulseNoise
        print 'Integral of (I^2+Q^2)/ns (Pulse): ',pulseReflect
        print 'Integral of (I^2+Q^2)/ns (Closed): ',pulseClosed
        print 'Integral of (I^2+Q^2)/ns (Release): ',pulseCatch
        
    plotInitTime = plotInit/resolution
    cutTimes = times[plotInitTime:timeEnd]
    cutZeroedVals = zeroedVals[plotInitTime:timeEnd]
    if plotRawIQ:
        cutVals = (vals[np.argsort(data[:,0])])[plotInitTime:timeEnd]
    else:
        cutVals = cutZeroedVals
    if plot:
        fig = dstools.plotDataset1D(np.array([cutTimes,np.real(cutVals),np.imag(cutVals),np.abs(cutZeroedVals)]).T,
                                    [dataset.variables[0][0]],
                                    [dataset.variables[1][0],dataset.variables[1][1],('Clicks','Magnitude','')],
                                    marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    if collect:
        return pulseNoise,pulseReflect,pulseClosed,pulseCatch
    
    
def plotNoiseCancellation(dir,datanum, ratioQtoI=1.0, delaySplit=None, maxRelease=None, collect=False, plot2D=True,
                          noisy=True, plotInit=0, plotRawIQ=False, plot=True, plotPhase=False, timeInit=400, plot3D=False,
                          remove10MHz=False, phase10MHz=0.1, sb_freq=50*MHz, nfft=4001, filterFunc=None):
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datanum, session=dir)
        dv.cd(dir)
        dv.open(datanum)
        title = dv.get_name().split(': ')[1]
    data = dataset.data
    stats = dataset.parameters['stats']
    resolution = float(dataset.parameters.get('resolution',2))
    
    # Determine all time scales. Pulse between timeInitScaled and timePulse. Release between
    # timeDelay and timeEnd. Sequence ended at timeComplete
    timeInitScaled = timeInit/resolution
    timePulse = timeInitScaled + dataset.parameters['Pulse Length'][ns]/resolution
    if ' Exponentialbuildup ' in title:
        timePulse += dataset.parameters['pulseParams'][1][ns]/resolution
    delay = dataset.parameters['Release Delay']
    if delay>0:
        timeDelay = timePulse + delay[ns]/resolution
    else:
        timeDelay = timePulse
    timeComplete = timeDelay + dataset.parameters['Release Length'][ns]/resolution
    if maxRelease is None:
        timeEnd = timeComplete
    else:
        timeEnd = timeDelay + maxRelease/resolution

    # Remove DC offset from each subpixel trace and sort data.
    vals = (data[:,1]+1j*data[:,2]/ratioQtoI)/stats
    acVals = np.array([vals[i*8192:(i+1)*8192]-np.mean(vals[i*8192:(i*8192+.75*timeInit/2)]) for i in range(len(vals)/8192)]).flatten()
    zeroedData = np.array([data[:,0],np.real(acVals),np.imag(acVals)]).T
    dataSort = zeroedData[np.argsort(data[:,0])]
    times = dataSort[:,0]
    reals = dataSort[:,1]
    imags = dataSort[:,2]
    if remove10MHz:
        timeSine = np.append(times[(.1*timeInitScaled):(.75*timeInitScaled)],times[(1.1*timeComplete):])
        realSine = np.append(reals[(.1*timeInitScaled):(.75*timeInitScaled)],reals[(1.1*timeComplete):])
        imagSine = np.append(imags[(.1*timeInitScaled):(.75*timeInitScaled)],imags[(1.1*timeComplete):])
        guess = np.array([.06*stats,0.01,0.5,0])
        #sine(x, amp, freq, phase, offset)
        realFit, cov, realFitFunc = fitting.fitCurve('sine', timeSine, realSine, guess)
        imagFit, cov, imagFitFunc = fitting.fitCurve('sine', timeSine, imagSine, guess)
        realZeroed = reals-(realFit[3]+realFit[0]*np.sin(2*np.pi*(times*realFit[1]+realFit[2])))
        imagZeroed = imags-(imagFit[3]+imagFit[0]*np.sin(2*np.pi*(times*imagFit[1]+imagFit[2])))
        zeroedVals = realZeroed + 1j*imagZeroed
    else:
        zeroedVals = reals+1j*imags
        
    unmixed = zeroedVals * np.exp(1j*(2*np.pi*sb_freq[GHz])*times)
    
    if filterFunc is not None:
        fftVals = np.fft.fft(unmixed, nfft)
        fftFreqs = np.fft.fftfreq(nfft,resolution*1e-9)
        filter = filterFunc(fftFreqs)
        filteredVals = np.fft.ifft(fftVals*filter, nfft)
    else:
        filteredVals = unmixed
        
    plotInitTime = plotInit/resolution
    cutTimes = times[plotInitTime:timeEnd]
    cutZeroedVals = filteredVals[plotInitTime:timeEnd]
    if plotRawIQ:
        cutVals = (vals[np.argsort(data[:,0])])[plotInitTime:timeEnd]
    else:
        cutVals = cutZeroedVals
    phases = np.angle(cutZeroedVals)
    if plot:
        fig = dstools.plotDataset1D(np.array([cutTimes,np.real(cutVals),np.imag(cutVals),np.abs(cutZeroedVals)]).T,
                                    [dataset.variables[0][0]],
                                    [dataset.variables[1][0],dataset.variables[1][1],('Clicks','Magnitude','')],
                                    marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    if plotPhase:
        fig = dstools.plotDataset1D(np.array([cutTimes,phases]).T, [dataset.variables[0][0]],
                                    [('Phase','Reflected','Rads')], marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    if plot3D:
        def findPhaseError(phase):
            expPhase = np.exp(1.j*phase)
            return 1.*(np.real(cutVals)*np.real(expPhase)+np.imag(cutVals)*np.imag(expPhase))#*np.abs(cutVals)
        phaseBest = leastsq(findPhaseError,0)[0]+np.pi/2.
        maxRadius = np.max(np.abs(cutVals))
        expBest = np.exp(1.j*phaseBest)
        x = np.array([[1,1],[-1,-1]])*maxRadius*np.real(expBest)
        y = np.array([[1,1],[-1,-1]])*maxRadius*np.imag(expBest)
        z = np.array([[np.min(cutTimes),np.max(cutTimes)],[np.min(cutTimes),np.max(cutTimes)]])
        
        fig = plt.figure()
        ax = Axes3D(fig)
        ax.plot_surface(x,y,z,color='b')
        ax.scatter(np.real(cutVals[:timeDelay]),np.imag(cutVals[:timeDelay]),cutTimes[:timeDelay],c='k',marker='o')
        ax.scatter(np.real(cutVals[timeDelay:]),np.imag(cutVals[timeDelay:]),cutTimes[timeDelay:],c='r',marker='o')
        fig.show()
        
        magBest = np.sqrt((np.real(cutVals)*np.real(expBest))**2+(np.imag(cutVals)*np.imag(expBest))**2)
        fig = dstools.plotDataset1D(np.array([cutTimes,np.abs(cutZeroedVals),magBest]).T,
                                    [dataset.variables[0][0]],
                                    [('Magnitude','Magnitude','LSBs'),('Magnitude','Planar Projection','LSBs')],
                                    marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    if plot2D:
        fig = plt.figure()
        dstools.plotDataset1D(np.array([np.real(cutVals[:(timeDelay-plotInitTime)]),np.imag(cutVals[:(timeDelay-plotInitTime)])]).T,
                                    [('Real(ADC)','LSBs')],[('Imag(ADC)','','LSBs')],
                                    marker='.', markersize=15, fig=fig)
        dstools.plotDataset1D(np.array([np.real(cutVals[(timeDelay-plotInitTime):]),np.imag(cutVals[(timeDelay-plotInitTime):])]).T,
                                    [('Real(ADC)','LSBs')],[('Imag(ADC)','','LSBs')],
                                    marker='.', markersize=15, fig=fig)
    #if collect:
        #return pulseNoise,pulseReflect,pulseClosed,pulseCatch


"""
TODO: Depreciated
"""

def fockIQ(sample, paramName='', readBias=0, delays=100*ns, releaseLen=2000*ns, sb_freq=50*MHz, resolution=2,
            save=True, stats=150, statsCut=3000, measure=0, name='', collect=False, noisy=False):
    """ Excite the qubit, swap to resonator, then open coupler to release photon.
    
    INPUT PARAMETERS
    sample: Object defining qubits to measure, loaded from registry.
    paramName - str: Name of swapAmp and swapTime registry keys.
    readBias - float: Squid dynamic bias for reading out (releasing) photon
    delays - time: Time during which coupling turned off
    releaseLens - time: Duration of release pulse.
    sb_freq - frequency: Sideband frequency of resonator drive source
    resolution - float (in ns): Time resolution for average mode. Max is 2ns.
    save - bool: Whether to save data to the datavault.
    stats - scalar: Number of times a point will be measured. Total signal is sum
        of each signal (so average is trace/states).
    statsCut - scalar: Number of times a point will be measured per rep. Total reps
        per point is stats/statsCut (rounded up).
    measure - scalar: Number of qubit to measure. Only one qubit allowed.
    name - str: Directory to which to save data.
    collect - bool: Whether or not to return data.
    noisy - bool: Whether to print status updates
    """
    sample, qubits = util.loadQubits(sample)
    q = qubits[measure]
    
    q['readoutType'] = 'resonator'
    q['biasReadout'] = q['biasOperate']
    q['adc_start_delay'] = 0*ns
    q['readout DAC start delay'] = 0*ns
    q['adc mode'] = 'average'
    q['adc demod frequency'] = sb_freq
    q['readout fc'] = q['readout frequency'] - sb_freq
    
    name = 'Fock IQ '+name
    axes = [(delays,'Release Delay')]
    deps = [('Clicks', 'I for %s' %q.__name__, ''), ('Clicks', 'Q for %s' %q.__name__, '')]
    kw = {'stats':stats,
          'Coupler Bias':readBias,
          'resolution':resolution}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw,
                                 explicitIndependents=[('Time','ns')])
    
    def func(server, delay):
        data_output = np.empty([0,3])
        for subpixel in list(st.r[0:(2-.5*resolution):resolution,ns]):
            statsLeft=stats
            count = 0
            Is = np.zeros(8192)
            Qs = np.zeros(8192)
            while statsLeft>0:
                q['readout'] = True
                t = 0
                q.xy = eh.mix(q, eh.piPulse(q, t+q['piLen']/2))
                t += q['piLen']
                q.z = env.rect(t, q['swapTime'+paramName], q['swapAmp'+paramName])
                t += q['swapTime'+paramName]
                q.cz = env.rect(t+delay,releaseLen,readBias)
                
                q.xy = env.shift(q.xy, subpixel)
                q.z = env.shift(q.z, subpixel)
                q.cz = env.shift(q.cz, subpixel)
                
                if statsLeft>statsCut:
                    statsNow = statsCut
                else:
                    statsNow = statsLeft
                statsLeft -= statsNow
                count += 1
                if noisy:
                    print subpixel,count
                ans = yield runQubits(server, qubits, statsNow, dataFormat='iq')
                Is += np.asarray(ans[0][0])
                Qs += np.asarray(ans[0][1])
            data_output = np.vstack((data_output,1.0*np.vstack((np.array(range(0,2*len(Is),2))-subpixel[ns],Is,Qs)).T))
        returnValue(data_output)
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    if collect:
        return data
    
def plotPulseIQNoMag(dir,datanum,plot=True, timeInit=400, ratioQtoI=1.0, delaySplit=None, maxRelease=None,
                collect=False, noisy=True, plotInit=0, plotRawIQ=False, remove10MHz=False):
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datanum, session=dir)
        dv.cd(dir)
        dv.open(datanum)
        title = dv.get_name().split(': ')[1]
    data = dataset.data
    stats = dataset.parameters['stats']
    resolution = float(dataset.parameters.get('resolution',2))
    
    # Determine all time scales. Pulse between timeInitScaled and timePulse. Release between
    # timeDelay and timeEnd. Sequence ended at timeComplete
    timeInitScaled = timeInit/resolution
    timePulse = timeInitScaled + dataset.parameters['Pulse Length'][ns]/resolution
    if ' Exponentialbuildup ' in title:
        timePulse += dataset.parameters['pulseParams'][1][ns]/resolution
    delay = dataset.parameters['Release Delay']
    if delay>0:
        timeDelay = timePulse + delay[ns]/resolution
    else:
        timeDelay = timePulse
    timeComplete = timeDelay + dataset.parameters['Release Length'][ns]/resolution
    if maxRelease is None:
        timeEnd = timeComplete
    else:
        timeEnd = timeDelay + maxRelease/resolution

    # Remove DC offset from each subpixel trace and sort data.
    vals = (data[:,1]+1j*data[:,2]/ratioQtoI)/stats
    acVals = np.array([vals[i*8192:(i+1)*8192]-np.mean(vals[i*8192:(i*8192+.75*timeInit/2)]) for i in range(len(vals)/8192)]).flatten()
    zeroedData = np.array([data[:,0],np.real(acVals),np.imag(acVals)]).T
    dataSort = zeroedData[np.argsort(data[:,0])]
    times = dataSort[:,0]
    reals = dataSort[:,1]
    imags = dataSort[:,2]
    
    # Remove 10MHz background sine wave (from clock)
    if remove10MHz:
        timeSine = np.append(times[(.1*timeInitScaled):(.75*timeInitScaled)],times[(1.1*timeComplete):])
        realSine = np.append(reals[(.1*timeInitScaled):(.75*timeInitScaled)],reals[(1.1*timeComplete):])
        imagSine = np.append(imags[(.1*timeInitScaled):(.75*timeInitScaled)],imags[(1.1*timeComplete):])
        guess = np.array([.06*stats,0.01,0.5,0])
        #sine(x, amp, freq, phase, offset)
        realFit, cov, realFitFunc = fitting.fitCurve('sine', timeSine, realSine, guess)
        imagFit, cov, imagFitFunc = fitting.fitCurve('sine', timeSine, imagSine, guess)
        realZeroed = reals-(realFit[3]+realFit[0]*np.sin(2*np.pi*(times*realFit[1]+realFit[2])))
        imagZeroed = imags-(imagFit[3]+imagFit[0]*np.sin(2*np.pi*(times*imagFit[1]+imagFit[2])))
        zeroedVals = realZeroed + 1j*imagZeroed
    else:
        zeroedVals = reals+1j*imags
    
    if delaySplit is not None:
        timeSplit = timePulse+delaySplit*(timeDelay-timePulse)
        pulseNoise = resolution*np.sum(np.abs(zeroedVals[(0.15*timeInitScaled):(0.85*timeInitScaled)])**2)/(0.7*timeInitScaled)
        pulseReflect = resolution*np.sum(np.abs(zeroedVals[timeInitScaled:timePulse])**2)
        pulseClosed = resolution*np.sum(np.abs(zeroedVals[timePulse:timeSplit])**2)
        pulseCatch = resolution*np.sum(np.abs(zeroedVals[timeSplit:timeEnd])**2)
    else:
        pulseNoise = resolution*np.sum(np.abs(zeroedVals[(0.15*timeInitScaled):(0.85*timeInitScaled)])**2)/(0.7*timeInitScaled)
        pulseReflect = resolution*np.sum(np.abs(zeroedVals[timeInitScaled:timePulse])**2)
        pulseClosed = resolution*np.sum(np.abs(zeroedVals[timePulse:timeDelay])**2)
        pulseCatch = resolution*np.sum(np.abs(zeroedVals[timeDelay:timeEnd])**2)
        
    if noisy:
        print 'Integral of (I^2+Q^2)/ns (Pulse): ',pulseReflect
        print 'Integral of (I^2+Q^2)/ns (Closed): ',pulseClosed
        print 'Integral of (I^2+Q^2)/ns (Release): ',pulseCatch
        
    plotInitTime = plotInit/resolution
    cutTimes = times[plotInitTime:timeEnd]
    cutZeroedVals = zeroedVals[plotInitTime:timeEnd]
    if plotRawIQ:
        cutVals = (vals[np.argsort(data[:,0])])[plotInitTime:timeEnd]
    else:
        cutVals = cutZeroedVals
    if plot:
        fig = dstools.plotDataset1D(np.array([cutTimes,np.real(cutVals),np.imag(cutVals)]).T,
                                    [dataset.variables[0][0]],
                                    [dataset.variables[1][0],dataset.variables[1][1]],
                                    marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    if collect:
        return pulseNoise,pulseReflect,pulseClosed,pulseCatch
    
def plotFockIQ(dir, datanum, plot=True, timeInit=400, ratioQtoI=1.0, fit=False, guess=None, collect=False, noisy=True, maxRelease=None):
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datanum, session=dir)
        dv.cd(dir)
        dv.open(datanum)
        title = dv.get_name().split(': ')[1]
    data = dataset.data
    dataSort = data[np.argsort(data[:,0])]
    stats = dataset.parameters['stats']
    times = dataSort[:,0]
    resolution = float(dataset.parameters.get('resolution',2))
    timeInit = timeInit/resolution
    if maxRelease is None:
        timeEnd = timeInit + dataset.parameters['Release Length'][ns]/resolution
    else:
        timeEnd = timeInit + maxRelease/resolution

    vals = (dataSort[:,1]+1j*dataSort[:,2]/ratioQtoI)/stats
    zeroedVals = vals-np.mean(vals[:(timeInit*3/4)])
    
    if fit:   
        cutTimes = times[timeInit:timeEnd]-times[timeInit]
        cutVals = zeroedVals[timeInit:timeEnd]
        if guess is None:
            guess = [0.1, 50.0, 0.0]
        fits, cov, fitFunc = fitting.fitCurve('exponential', cutTimes, np.abs(cutVals), guess)
        if noisy:
            print 'T1: %f ns' %fits[1]
    
    if plot:
        if fit:
            timeShift = times-times[timeInit]
            timePlot = timeShift * (timeShift>0)
            fitVals = fitting.exponential(timePlot,fits[0],fits[1],fits[2])
            plotData = np.array([times,np.real(zeroedVals),np.imag(zeroedVals),np.abs(zeroedVals),fitVals]).T
            plotDeps = [dataset.variables[1][0],dataset.variables[1][1],('Clicks','Magnitude',''),('','Fit','')]
        else:
            plotData = np.array([times,np.real(zeroedVals),np.imag(zeroedVals),np.abs(zeroedVals)]).T
            plotDeps = [dataset.variables[1][0],dataset.variables[1][1],('Clicks','Magnitude','')]
        fig = dstools.plotDataset1D(plotData, [dataset.variables[0][0]], plotDeps, marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
#    if collect:
#        return pulseReflect,pulseCatch
        

def columns2Matrix(pts, independentColumns=[0,1], dependentColumn=-1, default=np.NaN, steps=None, tol=1e-11):
    """Converts data in columns format into a 2D array.
    
    No special order of the datapoints is required.
    The spacing between pixels is the median of nonzero changes
    of independent variables in neigboring lines.
    """

    #n = len(pts[:,0])
    dims = np.size(independentColumns)
    indepVals = []
    indices = []
    sizes = []
    for col in pts[:,independentColumns].T:
        newIndep = np.array([col[0]])
        locs = [0]
        for elem in col[1:]:
            minArg = np.argmin(np.abs(newIndep-elem))
            if abs(newIndep[minArg]-elem)>tol:
                locs.append(len(newIndep))
                newIndep=np.append(newIndep,elem)
            else:
                locs.append(minArg)
        indepVals.append(newIndep)
        indices.append(locs)
        sizes.append(len(newIndep))
    indices = np.array(indices).T
    pts = pts[:,dependentColumn]
    if np.iterable(dependentColumn):
        data = np.resize(default, sizes + [len(dependentColumn)])
    else:
        data = np.resize(default, sizes)
    #Create the 2D image array
    for index,pt in zip(indices,pts):
        data[tuple(index)] = pt
        
    return data, indepVals
        
    
def plotSweep2D(dir, datanum, timeInit = 400/2, ratioQtoI=1.0, collect=True,
                plotColumns=None, plotMagnitudeTimes=False, plotIntegrals=False,):
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datanum, session=dir)
    data = dataset.data
    matrix,(sweepVals,times) = columns2Matrix(data, dependentColumn=[2,3])
    iReshape = matrix[:,:,0]
    qReshape = matrix[:,:,1]
    stats = dataset.parameters['stats']
    convertToNs = Value(1,dataset.variables[0][0][1])[ns]
    sweepVars = convertToNs * sweepVals
    timePulse = timeInit + dataset.parameters['Pulse Length'][ns]/2
    delay = dataset.parameters['Release Delay']
    if delay>0:
        timeDelay = timePulse + delay[ns]/2
    else:
        timeDelay = timePulse
    timeEnd = timeDelay + dataset.parameters['Release Length'][ns]/2

    vals = (iReshape+1j*qReshape/ratioQtoI)/stats
    acVals = vals.T-np.mean(vals[:,:(timeInit*3/4)],axis=1)
    
    pulseReflect = 2*np.sum(np.abs(acVals[timeInit:timePulse])**2,axis=0)
    pulseCatch = 2*np.sum(np.abs(acVals[timeDelay:timeEnd])**2,axis=0)
    
    cutTimes = times[:timeEnd]
    cutVals = acVals[:timeEnd]
    
    if plotColumns is not None:
        for column in plotColumns:
            fig = dstools.plotDataset1D(np.array([cutTimes,np.real(cutVals[:,column]),np.imag(cutVals[:,column]),np.abs(cutVals[:,column])]).T,
                                        [dataset.variables[0][1]],
                                        [dataset.variables[1][0],dataset.variables[1][1],('Clicks','Magnitude','')],
                                        marker='.', markersize=15)
            ax = fig.get_axes()[0]
            ax.set_title(dataset['path'])
            fig.show()
            
    if plotMagnitudeTimes:
        dataColumns = np.array([elem for elem in itertools.product(sweepVars,cutTimes)]).T
        dataColumns = np.vstack((dataColumns,np.reshape(np.abs(cutVals.T),-1))).T
        fig = dstools.plotDataset2D(dataColumns, dataset.variables[0], [('Amplitude','Magnitude','Clicks')])
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    
    if plotIntegrals:
        fig = dstools.plotDataset1D(np.array([sweepVars,pulseReflect]).T, [dataset.variables[0][0]],
                                    [('Energy','Reflected','Clicks^2*ns')], marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
        
        fig = dstools.plotDataset1D(np.array([sweepVars,pulseCatch]).T, [dataset.variables[0][0]],
                                    [('Energy','Caught','Clicks^2*ns')], marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    
    if collect:
        return np.array([sweepVars,pulseReflect,pulseCatch]).T,dataset
        
    
def plotPulseLength2D(dir, datanum, timeInit = 400/2, ratioQtoI=1.0, collect=True, save=False,
                plotColumns=None, plotMagnitudeTimes=False, plotIntegrals=False, delaySplit=None):
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datanum, session=dir)
    data = dataset.data
    matrix,(sweepVals,times) = columns2Matrix(data, dependentColumn=[2,3])
    iReshape = matrix[:,:,0]
    qReshape = matrix[:,:,1]
    stats = dataset.parameters['stats']
    convertToNs = Value(1,dataset.variables[0][0][1])[ns]
    sweepVars = convertToNs * sweepVals
    timePulse = timeInit + sweepVars/2
    delay = dataset.parameters['Release Delay']
    if delay>0:
        timeDelay = timePulse + delay[ns]/2
    else:
        timeDelay = timePulse
    timeEnd = timeDelay + dataset.parameters['Release Length'][ns]/2

    vals = (iReshape+1j*qReshape/ratioQtoI)/stats
    acVals = vals.T-np.mean(vals[:,:(timeInit*3/4)],axis=1)
    
    if delaySplit is not None:
        timeSplit = timePulse+delaySplit*(timeDelay-timePulse)
        
        pulseReflect = 2*np.array([np.sum(np.abs(acValCol[timeInit:splitTime])**2) for (splitTime,acValCol) in zip(timeSplit,acVals.T)])
        pulseCatch = 2*np.array([np.sum(np.abs(acValCol[splitTime:endTime])**2) for (splitTime,endTime,acValCol) in zip(timeSplit,timeEnd,acVals.T)])
    else:
        pulseReflect = 2*np.array([np.sum(np.abs(acValCol[timeInit:pulseTime])**2) for (pulseTime,acValCol) in zip(timePulse,acVals.T)])
        pulseCatch = 2*np.array([np.sum(np.abs(acValCol[delayTime:endTime])**2) for (delayTime,endTime,acValCol) in zip(timeDelay,timeEnd,acVals.T)])
    
    cutTimes = times[:np.max(timeEnd)]
    cutVals = acVals[:np.max(timeEnd)]
    
    if plotColumns is not None:
        for column in plotColumns:
            fig = dstools.plotDataset1D(np.array([cutTimes,np.real(cutVals[:,column]),np.imag(cutVals[:,column]),np.abs(cutVals[:,column])]).T,
                                        [dataset.variables[0][1]],
                                        [dataset.variables[1][0],dataset.variables[1][1],('Clicks','Magnitude','')],
                                        marker='.', markersize=15)
            ax = fig.get_axes()[0]
            ax.set_title(dataset['path'])
            fig.show()
            
    if plotMagnitudeTimes:
        dataColumns = np.array([elem for elem in itertools.product(sweepVars,cutTimes)]).T
        dataColumns = np.vstack((dataColumns,np.reshape(np.abs(cutVals.T),-1))).T
        fig = dstools.plotDataset2D(dataColumns, dataset.variables[0], [('Amplitude','Magnitude','Clicks')])
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    
    if plotIntegrals:
        fig = dstools.plotDataset1D(np.array([sweepVars,pulseReflect]).T, [dataset.variables[0][0]],
                                    [('Energy','Reflected','Clicks^2*ns')], marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
        
        fig = dstools.plotDataset1D(np.array([sweepVars,pulseCatch]).T, [dataset.variables[0][0]],
                                    [('Energy','Caught','Clicks^2*ns')], marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    
    dataOutput = np.array([sweepVars,pulseReflect,pulseCatch]).T
    if save:
        with labrad.connect() as cxn:
            dv = cxn.data_vault
            dv.cd(dir)
            dv.open(datanum)
            title = dv.get_name().split(': ')[1]
            dv.new('Integrals '+title, dataset['variables'][0][:-1],[('Integral','Pulse','Clicks^2*ns'),('Integral','Caught','Clicks^2*ns')])
            dv.add(1.0*dataOutput)
            params = dataset['parameters']
            params['Integral TimeInit'] = 2*timeInit*ns
            params['Integral OmitDelay'] = delaySplit
            params['Integral QtoI'] = ratioQtoI
            for elem in sorted(params):
                if isinstance(params[elem],dict):
                    qubit = sorted(params[elem])
                    for key in qubit:
                        dv.add_parameter(elem+'.'+key, params[elem][key])
                else:
                    dv.add_parameter(elem,params[elem])
    
    if collect:
        return dataOutput,dataset
        
    
def plotReleaseDelay2D(dir, datanum, timeInit = 400/2, ratioQtoI=1.0, delaySplit=None, collect=True,
                plotColumns=None, plotMagnitudeTimes=False, plotIntegrals=False, save=False):
    with labrad.connect() as cxn:
        #Get the data from the data vault
        dv=cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, datasetId=datanum, session=dir)
    data = dataset.data
    matrix,(sweepVals,times) = columns2Matrix(data, dependentColumn=[2,3])
    iReshape = matrix[:,:,0]
    qReshape = matrix[:,:,1]
    stats = dataset.parameters['stats']
    timePulse = timeInit + dataset.parameters['Pulse Length'][ns]/2
    convertToNs = Value(1,dataset.variables[0][0][1])[ns]
    sweepVars = convertToNs * sweepVals
    timeDelay = timePulse + sweepVars/2
    timeEnd = timeDelay + dataset.parameters['Release Length'][ns]/2

    vals = (iReshape+1j*qReshape/ratioQtoI)/stats
    acVals = vals.T-np.mean(vals[:,:(timeInit*3/4)],axis=1)
    
    if delaySplit is not None:
        timeSplit = timePulse+delaySplit*(timeDelay-timePulse)
        pulseReflect = resolution*np.sum(np.abs(zeroedVals[timeInit:timePulse])**2)
        pulseClosed = resolution*np.array([np.sum(np.abs(acValCol[timePulse:splitTime])**2) for (splitTime,acValCol) in zip(timeSplit,acVals.T)])
        pulseCatch = resolution*np.sum(np.abs(zeroedVals[timeSplit:timeEnd])**2)
        pulseReflect = 2*np.array([np.sum(np.abs(acValCol[timeInit:splitTime])**2) for (splitTime,acValCol) in zip(timeSplit,acVals.T)])
        pulseCatch = 2*np.array([np.sum(np.abs(acValCol[splitTime:endTime])**2) for (splitTime,endTime,acValCol) in zip(timeSplit,timeEnd,acVals.T)])
    else:
        pulseReflect = 2*np.sum(np.abs(acVals[timeInit:timePulse])**2,axis=0)
        pulseCatch = 2*np.array([np.sum(np.abs(acValCol[delayTime:endTime])**2) for (delayTime,endTime,acValCol) in zip(timeDelay,timeEnd,acVals.T)])
    
    cutTimes = times[:np.max(timeEnd)]
    cutVals = acVals[:np.max(timeEnd)]
    
    if plotColumns is not None:
        for column in plotColumns:
            fig = dstools.plotDataset1D(np.array([cutTimes,np.real(cutVals[:,column]),np.imag(cutVals[:,column]),np.abs(cutVals[:,column])]).T,
                                        [dataset.variables[0][1]],
                                        [dataset.variables[1][0],dataset.variables[1][1],('Clicks','Magnitude','')],
                                        marker='.', markersize=15)
            ax = fig.get_axes()[0]
            ax.set_title(dataset['path'])
            fig.show()
            
    if plotMagnitudeTimes:
        dataColumns = np.array([elem for elem in itertools.product(sweepVars,cutTimes)]).T
        dataColumns = np.vstack((dataColumns,np.reshape(np.abs(cutVals.T),-1))).T
        fig = dstools.plotDataset2D(dataColumns, dataset.variables[0], [('Amplitude','Magnitude','Clicks')])
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    
    if plotIntegrals:
        fig = dstools.plotDataset1D(np.array([sweepVars,pulseReflect]).T, [dataset.variables[0][0]],
                                    [('Energy','Reflected','Clicks^2*ns')], marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
        
        fig = dstools.plotDataset1D(np.array([sweepVars,pulseCatch]).T, [dataset.variables[0][0]],
                                    [('Energy','Caught','Clicks^2*ns')], marker='.', markersize=15)
        ax = fig.get_axes()[0]
        ax.grid()
        ax.set_title(dataset['path'])
        fig.show()
    
    dataOutput = np.array([sweepVars,pulseReflect,pulseCatch]).T
    if save:
        with labrad.connect() as cxn:
            dv = cxn.data_vault
            dv.cd(dir)
            dv.open(datanum)
            title = dv.get_name().split(': ')[1]
            dv.new('Integrals '+title, dataset['variables'][0][:-1],[('Integral','Pulse','Clicks^2*ns'),('Integral','Caught','Clicks^2*ns')])
            dv.add(1.0*dataOutput)
            params = dataset['parameters']
            params['Integral TimeInit'] = 2*timeInit*ns
            params['Integral OmitDelay'] = delaySplit
            params['Integral QtoI'] = ratioQtoI
            for elem in sorted(params):
                if isinstance(params[elem],dict):
                    qubit = sorted(params[elem])
                    for key in qubit:
                        dv.add_parameter(elem+'.'+key, params[elem][key])
                else:
                    dv.add_parameter(elem,params[elem])
            
    
    if collect:
        return dataOutput,dataset