import numpy as np
import random
import itertools
import copy
from pyle.math import tensor, dagger, dot3
from pyle import quantum as quantum
from pyle import tomo as tomo
from pyle import envelopes as env
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from pyle.datavault import DataVaultWrapper
from pyle.registry import AttrDict
from matplotlib.mlab import find
from labrad.units import Unit,Value
V, mV, us, ns, GHz, MHz = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz')]

class RandomizedBenchMarking(object):
    def __init__(self, numQubits,gateSet = None):
        if gateSet is None:
            gateSet=['I','X','Y','X/2','Y/2','-X','-Y','-X/2','-Y/2']
        self.gateSet = gateSet
        self.numQubits = numQubits
        if numQubits is 1:
            self.fullSingleGateSet = self.gateSet
            self.fullGateSet = self.fullSingleGateSet
        if numQubits is 2:
            self.fullSingleGateSet = list(itertools.product(self.gateSet,self.gateSet))
            self.fullGateSet = self.fullSingleGateSet + [('CZ',)*numQubits]
        elif numQubits >2:
            raise Exception("Haven't implemented >2 qubits")
        '''
        self.singleGateDict = {'I': np.eye(2),
            'X': np.array([[0,1],[1,0]]),
            'Y': np.array([[0,-1j],[1j,0]]),
            'X/2': (np.eye(2) + 1j * np.array([[0,1],[1,0]]))/np.sqrt(2),
            'Y/2': (np.eye(2) - 1j * np.array([[0,-1j],[1j,0]]))/np.sqrt(2),
            '-X': -np.array([[0,1],[1,0]]),
            '-Y': -np.array([[0,-1j],[1j,0]]),
            '-X/2': (np.eye(2) - 1j * np.array([[0,1],[1,0]]))/np.sqrt(2),
            '-Y/2': (np.eye(2) + 1j * np.array([[0,-1j],[1j,0]]))/np.sqrt(2)}
        '''
        self.singleGateDict = {'I': tomo.Rmat(tomo.sigmaI, 0),
                               'X': tomo.Rmat(tomo.sigmaX, np.pi),
                               'Y': tomo.Rmat(tomo.sigmaY, np.pi),
                               'X/2': tomo.Rmat(tomo.sigmaX, np.pi/2.),
                               'Y/2': tomo.Rmat(tomo.sigmaY, np.pi/2.),
                               '-X': tomo.Rmat(tomo.sigmaX, -np.pi),
                               '-Y': tomo.Rmat(tomo.sigmaY, -np.pi),
                               '-X/2': tomo.Rmat(tomo.sigmaX, -np.pi/2.),
                               '-Y/2': tomo.Rmat(tomo.sigmaY, -np.pi/2.),
                               'H': np.array([[1,1],[1,-1]])/np.sqrt(2),
                               'Z': np.eye(2),
                               'ZW': np.eye(2),                               
                               'CZR': np.eye(2), # Control Z trajectory with unloaded bus
                               'CZRL': tomo.Rmat(tomo.sigmaZ, np.pi), # Control Z trajectory with loaded bus
                               'SWAP': np.eye(2)}
        self.fullGateDict = {}
        if numQubits is 1:
            self.fullGateDict = self.singleGateDict
        if numQubits is 2:
            for gate in self.fullSingleGateSet:
                self.fullGateDict[gate] = tensor([self.singleGateDict[subGate] for subGate in gate[::-1]]) # we reverse the order of gate
                                                                                                           # we do this because while we pass in gates
                                                                                                           # like ('X/2','Y/2') for X/2 on q0 and Y/2 on q1
                                                                                                           # X/2 tensor Y/2 means 
                                                                                                           # X/2 on q1 with Y/2 on q0, so we must swap
            self.fullGateDict[('CZ','CZ')] = np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]])
            #self.fullGateDict[('CNOT','CNOT')] = np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]]) #control=1,target=0
            self.fullGateDict[('CNOT','CNOT')] = np.array([[1,0,0,0],[0,0,0,1],[0,0,1,0],[0,1,0,0]])  #control=0,target=1
    def randGen(self, iterations, entangling=False, interleaved=False): 
        '''
        This generates a random list of strings that correspond to gates
        e.g. ['X','I','-Y/2','X',....]
        '''
        if interleaved:
            gateList = []
            for k in range(iterations[-1]):
                gateList.append(random.sample(self.fullSingleGateSet,1)[0])
                gateList.append(interleaved)
        elif entangling:
            gateList = [random.sample(self.fullGateSet,1)[0] for k in range(iterations[-1])]
        else:
            gateList = [random.sample(self.fullSingleGateSet,1)[0] for k in range(iterations[-1])]
        
        return gateList
                
    def computeUnitary(self, gateList):

        '''
        This takes an entire sequence of gates and finds the equivalent unitary
        '''
        
        U = np.eye(2**self.numQubits)

        for op in gateList:
            U = np.dot(self.fullGateDict[op],U)
            
        return U

    def finishGateList(self, gateList, entangling = False):

        '''
        This takes the random list of strings that correspond to gates, from randGen
        and generates a final gate that will make the entire sequence map the ground 
        state to the ground state
        '''
        
        U = self.computeUnitary(gateList)
        psi0 = np.zeros((2**self.numQubits,1)) #column vector
        psi0[0] = 1
        finalGate=None
        if entangling:
            for gatei in self.fullSingleGateSet:
                for gatef in self.fullSingleGateSet:
                    Uend = dot3(self.fullGateDict[gatef], self.fullGateDict[('CZ','CZ')], self.fullGateDict[gatei])
                    Uf = np.dot(Uend, U)
                    psif = np.dot(Uf, psi0)
                    if abs(np.dot(dagger(psif), psi0))>0.999:
                        finalGate = [gatei]+[('CZ','CZ')]+[gatef]
                        continue
        else:
            for gate in self.fullSingleGateSet:
                Uf = np.dot(self.fullGateDict[gate],U)
                psif = np.dot(Uf, psi0)
                if abs(np.dot(dagger(psif), psi0))>0.999:
                    finalGate = [gate]
                    continue
            if finalGate is None:
                #try 2 gates
                for gatei in self.fullSingleGateSet:
                    for gatef in self.fullSingleGateSet:
                        Uend = np.dot(self.fullGateDict[gatef], self.fullGateDict[gatei])
                        Uf = np.dot(Uend, U)
                        psif = np.dot(Uf, psi0)
                        if abs(np.dot(dagger(psif), psi0))>0.999:
                            finalGate = [gatei]+[gatef]
                            continue                
                
        fullGateList = gateList + finalGate
        
        return fullGateList

    def finishGateListNew(self, gateList, entangling = False):
        '''
        Picks a final set of gates that will make the unitary of the entire sequence 
        be the identity, neglecting a global phase.
        '''        
        U = self.computeUnitary(gateList)
        psi0 = np.zeros((2**self.numQubits,1)) #column vector
        psi0[0] = 1
        finalGate=None
        if entangling:
            for gatei in self.fullSingleGateSet:
                for gatef in self.fullSingleGateSet:
                    Uend = dot3(self.fullGateDict[gatef], self.fullGateDict[('CZ','CZ')], self.fullGateDict[gatei])
                    Uf = np.dot(Uend, U)
                    howCloseToI = abs(np.trace(Uf))**2/(1.*2**(2*self.numQubits))
                    if howCloseToI>0.9999:
                        finalGate = [gatei]+[('CZ','CZ')]+[gatef]
                        break
        else:
            for gate in self.fullSingleGateSet:
                Uf = np.dot(self.fullGateDict[gate],U)
                howCloseToI = abs(np.trace(Uf))**2/(1.*2**(2*self.numQubits))
                if howCloseToI>0.9999:
                    finalGate = [gate]
                    break
            if finalGate is None:
                #try 2 gates
                for gatei in self.fullSingleGateSet:
                    for gatef in self.fullSingleGateSet:
                        Uend = np.dot(self.fullGateDict[gatef], self.fullGateDict[gatei])
                        Uf = np.dot(Uend, U)
                        howCloseToI = abs(np.trace(Uf))**2/(1.*2**(2*self.numQubits))
                        if howCloseToI>0.9999:
                            finalGate = [gatei]+[gatef]
                            break
        fullGateList = gateList + finalGate        
        return fullGateList
 
        
    def testSequence(self, length = 10, entangling=False, interleaved=None):
        
        gateList = self.randGen([length], entangling=entangling, interleaved=interleaved)
        print 'gateList: %s' % gateList
        
        Ui = self.computeUnitary(gateList)
        
        fullGateList = self.finishGateList(gateList, entangling=entangling)
        print 'fullGateList: %s' % fullGateList
        
        U = self.computeUnitary(fullGateList)
        psi0 = np.zeros((2**self.numQubits,1)) #column vector
        psi0[0] = 1
        psif = np.dot(U, psi0)
        fidelity = np.dot(dagger(psif),psi0)
        
        return abs(fidelity)
        
        
class RandomizedBenchMarkingClifford(object):
    '''
    Generates Clifford gate list for Randomized Benchmarking, 1 & 2 qubits. CZ-based.
    For 1 qubit : generates cliffords from the 1 qubit clifford set C1, with |C1|=24
    For 2 qubits: generates cliffords from the 2 qubit clifford set C2, with |C2|=11520 from 4 classes:
    (CZ-based clifford set C2)
        single qubit class (no CZs): 576
        CNOT-like class (1 CZ): 5184
        ISWAP-like class (2 CZs): 5184
        SWAP-like class (3 CZs): 576    
    The gate and gateCompiler automatically compress the sequence such that the single qubit gates may overlap between cliffords.
        ie:  
         q0: CZ  X/2  I   -X/2 CZ 
         q1: CZ -Y/2  X    I   CZ
        ->
         q0: CZ  X/2 -X/2  CZ 
         q1: CZ -Y/2  X    CZ
         
    It also gets rid of the first single qubit gate from the elementary 2 qubit gates as it is absorbed in C1.
    
    gateList format:
        format: [ 1st Clifford: [ [qubit0 gates] , [qubit1 gates] ] , 
                  2nd Clifford: [ [qubit0 gates] , [qubit1 gates] ] ,
                                                                       ]
        qubit0 gates:  'I','X' in order of execution
        so:
        [ [ ['I','X'],['Y','Y'] ] , ...]
        does first an identity and Y on qubit 0 & 1, then X & Y on qubit 0 and 1.
    Rami Barends July 2013		
    '''

    def __init__(self, numQubits,randomizeOverIswapAndSwapParts=False):
        self.numQubits = numQubits
        self.gateLengths = {}        
        #Clifford generator set
        self.cliffordGeneratorSet=['I','X','Y','X/2','Y/2','-X','-Y','-X/2','-Y/2']
        #single qubit cliffords C1, and S1, from Andrzej's mathematica script
        #average single qubit gates per clifford: 1.875
        #DO NOT CHANGE!
        self.gateSetC1=[['I'],['X/2'],['-X/2'],['X'],
                    ['Y/2'],['-Y/2'],['Y'],
                    ['-X/2','Y/2','X/2'],['-X/2','-Y/2','X/2'],['Y','X'],
                    ['Y','X/2'],['Y','-X/2'],['X','Y/2'],['X','-Y/2'],
                    ['X/2','Y/2','X/2'],['-X/2','Y/2','-X/2'],
                    ['Y/2','X/2'],['Y/2','-X/2'],['-Y/2','X/2'],['-Y/2','-X/2'],
                    ['X/2','Y/2'],['X/2','-Y/2'],['-X/2','Y/2'],['-X/2','-Y/2']]
                    
        #these S1 gates are elements fo the single qubit Clifford group:
        #S1 with absorption of Y2:
        #[['Y/2','I'],['Y/2','Y/2','X/2'],['Y/2','-X/2','-Y/2']] = 
        #[['Y/2'],['Y','X/2'],['-X/2','-Y/2','X/2']] (these are single qubit Cliffords)        
        self.gateSetS1=[['I'],['Y/2','X/2'],['-X/2','-Y/2']] 
        self.gateSetS1_with_absorption_of_Y2 = [['Y/2'],['Y','X/2'],['-X/2','-Y/2','X/2']] 
        #note that ['-X/2','-Y/2','X/2'] = ['Y/2','-X/2','-Y/2'], the advantage of keeping it this way is that it is a physical clifford from the above list
        self.gateSetS1_with_absorption_of_X2 = [['X/2'],['X/2','Y/2','X/2'],['-Y/2']]
        
        #8 ways of writing iswaps
        all_iswaps=[
        [['X/2','CZ','X/2','CZ','-X/2'],['Y/2','CZ','-Y/2','CZ','-Y/2']],
        [['X/2','CZ','-X/2','CZ','-X/2'],['-Y/2','CZ','-Y/2','CZ','Y/2']],
        [['Y/2','CZ','-Y/2','CZ','-Y/2'],['X/2','CZ','X/2','CZ','-X/2']],
        [['Y/2','CZ','Y/2','CZ','-Y/2'],['-X/2','CZ','X/2','CZ','X/2']],
        [['-X/2','CZ','X/2','CZ','X/2'],['Y/2','CZ','Y/2','CZ','-Y/2']],
        [['-X/2','CZ','-X/2','CZ','X/2'],['-Y/2','CZ','Y/2','CZ','Y/2']],
        [['-Y/2','CZ','-Y/2','CZ','Y/2'],['X/2','CZ','-X/2','CZ','-X/2']],
        [['-Y/2','CZ','Y/2','CZ','Y/2'],['-X/2','CZ','-X/2','CZ','X/2']]
        ]
        
        #24 ways of writing swaps
        all_swaps=[
        [['X/2','CZ','-X/2','CZ','X/2','CZ'],['-X/2','CZ','X/2','CZ','-X/2','CZ']],
        [['Y/2','CZ','-Y/2','CZ','Y/2','CZ'],['-Y/2','CZ','Y/2','CZ','-Y/2','CZ']],
        [['-X/2','CZ','X/2','CZ','-X/2','CZ'],['X/2','CZ','-X/2','CZ','X/2','CZ']],
        [['-Y/2','CZ','Y/2','CZ','-Y/2','CZ'],['Y/2','CZ','-Y/2','CZ','Y/2','CZ']],
        [['I','CZ','X/2','CZ','-X/2','CZ','X/2'],['I','CZ','-X/2','CZ','X/2','CZ','-X/2']],
        [['I','CZ','Y/2','CZ','-Y/2','CZ','Y/2'],['I','CZ','-Y/2','CZ','Y/2','CZ','-Y/2']],
        [['I','CZ','-X/2','CZ','X/2','CZ','-X/2'],['I','CZ','X/2','CZ','-X/2','CZ','X/2']],
        [['I','CZ','-Y/2','CZ','Y/2','CZ','-Y/2'],['I','CZ','Y/2','CZ','-Y/2','CZ','Y/2']],
        [['I','CZ','X/2','CZ','-X/2','CZ','I'],['X/2','CZ','-X/2','CZ','X/2','CZ','-X/2']],
        [['I','CZ','Y/2','CZ','-Y/2','CZ','I'],['Y/2','CZ','-Y/2','CZ','Y/2','CZ','-Y/2']],
        [['I','CZ','-X/2','CZ','X/2','CZ','I'],['-X/2','CZ','X/2','CZ','-X/2','CZ','X/2']],
        [['I','CZ','-Y/2','CZ','Y/2','CZ','I'],['-Y/2','CZ','Y/2','CZ','-Y/2','CZ','Y/2']],
        [['X/2','CZ','-X/2','CZ','X/2','CZ','-X/2'],['I','CZ','X/2','CZ','-X/2','CZ','I']],
        [['X/2','CZ','Y/2','CZ','-Y/2','CZ','-X/2'],['Y/2','CZ','X/2','CZ','-X/2','CZ','-Y/2']],
        [['X/2','CZ','-X/2','CZ','X/2','CZ','I'],['-X/2','CZ','X/2','CZ','-X/2','CZ','I']],
        [['X/2','CZ','-Y/2','CZ','Y/2','CZ','-X/2'],['-Y/2','CZ','X/2','CZ','-X/2','CZ','Y/2']],
        [['Y/2','CZ','-Y/2','CZ','Y/2','CZ','-Y/2'],['I','CZ','Y/2','CZ','-Y/2','CZ','I']],
        [['Y/2','CZ','X/2','CZ','-X/2','CZ','-Y/2'],['X/2','CZ','Y/2','CZ','-Y/2','CZ','-X/2']],
        [['Y/2','CZ','-X/2','CZ','X/2','CZ','-Y/2'],['-X/2','CZ','Y/2','CZ','-Y/2','CZ','X/2']],
        [['Y/2','CZ','-Y/2','CZ','Y/2','CZ','I'],['-Y/2','CZ','Y/2','CZ','-Y/2','CZ','I']],
        [['-X/2','CZ','X/2','CZ','-X/2','CZ','X/2'],['I','CZ','-X/2','CZ','X/2','CZ','I']],
        [['-X/2','CZ','X/2','CZ','-X/2','CZ','I'],['X/2','CZ','-X/2','CZ','X/2','CZ','I']],
        [['-X/2','CZ','Y/2','CZ','-Y/2','CZ','X/2'],['Y/2','CZ','-X/2','CZ','X/2','CZ','-Y/2']],
        [['-X/2','CZ','-Y/2','CZ','Y/2','CZ','X/2'],['-Y/2','CZ','-X/2','CZ','X/2','CZ','Y/2']],
        [['-Y/2','CZ','Y/2','CZ','-Y/2','CZ','Y/2'],['I','CZ','-Y/2','CZ','Y/2','CZ','I']],
        [['-Y/2','CZ','X/2','CZ','-X/2','CZ','Y/2'],['X/2','CZ','-Y/2','CZ','Y/2','CZ','-X/2']],
        [['-Y/2','CZ','Y/2','CZ','-Y/2','CZ','I'],['Y/2','CZ','-Y/2','CZ','Y/2','CZ','I']],
        [['-Y/2','CZ','-X/2','CZ','X/2','CZ','Y/2'],['-X/2','CZ','-Y/2','CZ','Y/2','CZ','X/2']]        
        ]
        
        '''
        generate 2 qubit cliffords:
        We're going to construct the qubit gates, single qubit-oriented. That's why there are no 'I's in the two qubit gate parts
        We'll align the CZ's to each other later. This way we can move overlapping Cliffords into each other:
        ie:  
         q0: CZ  X/2  I   -X/2 CZ 
         q1: CZ -Y/2  X    I   CZ
        ->
         q0: CZ  X/2 -X/2  CZ 
         q1: CZ -Y/2  X    CZ
        '''
        
        if numQubits==2:
            twoqubitS1 = list(itertools.product(self.gateSetS1,self.gateSetS1)) #list of S1 tuples
            twoqubitS1 = [ list(t) for t in twoqubitS1 ] #now as a list of lists
            
            twoqubitS1_with_absorption_of_I_Y2 = list(itertools.product(self.gateSetS1,self.gateSetS1_with_absorption_of_Y2)) #list of S1,S1* tuples
            twoqubitS1_with_absorption_of_I_Y2 = [ list(t) for t in twoqubitS1_with_absorption_of_I_Y2 ]
            
            twoqubitS1_with_absorption_of_Y2_X2 = list(itertools.product(self.gateSetS1_with_absorption_of_Y2,self.gateSetS1_with_absorption_of_X2)) #list of S1,S1* tuples
            twoqubitS1_with_absorption_of_Y2_X2 = [ list(t) for t in twoqubitS1_with_absorption_of_Y2_X2 ]
            
            #self.twoqubitS1_with_absorption_of_I_Y2=twoqubitS1_with_absorption_of_I_Y2
            #all 576 single qubit operations      
            gateSetC2_sq =  list(itertools.product(self.gateSetC1,self.gateSetC1)) #list of tuples        
            gateSetC2_sq =[ list(t) for t in gateSetC2_sq ] #now as a list of lists

            
            #CNOT-like class: 5184 elements
            twoqubitgatepart=[ ['CZ'] , ['CZ'] ] #order: control, target. Only for twoqubitS1_with_absorption_of_I_Y2
            #twoqubitgatepart=[ ['CZ'] , ['CZ','Y/2'] ] #order: control, target
            #twoqubitgatepart=[ ['CZ'] , ['-Y/2','CZ','Y/2'] ] #order: control, target            
            #add twoqubitgatepart to C1's
            gateSetC2_cnot_woS1 = [ [C2[0]+twoqubitgatepart[0],C2[1]+twoqubitgatepart[1]] for C2 in gateSetC2_sq ]
            #add S1's
            gateSetC2_cnot=[]
            for S1 in twoqubitS1_with_absorption_of_I_Y2:
                gateSetC2_cnot.extend( [ [gate[0]+S1[0], gate[1]+S1[1]] for gate in gateSetC2_cnot_woS1] )

                
            #ISWAP-like class: 5184 elements
            if randomizeOverIswapAndSwapParts:
                gateSetC2_iswap_woS1=[]
                for C2 in gateSetC2_sq:
                    twoqubitgatepart = random.sample(all_iswaps,1)[0]
                    gateSetC2_iswap_woS1.append( [C2[0]+twoqubitgatepart[0],C2[1]+twoqubitgatepart[1]] )                  
                #add S1's
                gateSetC2_iswap=[]
                for S1 in twoqubitS1:
                    gateSetC2_iswap.extend( [ [gate[0]+S1[0], gate[1]+S1[1]] for gate in gateSetC2_iswap_woS1] )                    
            else:
                #full ISWAP:
                #[['-Y/2','CZ','Y/2','CZ','Y/2'],['-X/2','CZ','-X/2','CZ','X/2']]
                #with absorption of first and last single qubit gates by C1 and S1:
                twoqubitgatepart=[['CZ','Y/2','CZ'],['CZ','-X/2','CZ']] #order: control, target            
                #twoqubitgatepart=[['CZ','-Y/2','CZ','-Y/2'],['CZ','X/2','CZ','-X/2']] #order: control, target          
                #add twoqubitgatepart to C1's
                gateSetC2_iswap_woS1 = [ [C2[0]+twoqubitgatepart[0],C2[1]+twoqubitgatepart[1]] for C2 in gateSetC2_sq ]          
                #add S1's
                gateSetC2_iswap=[]
                for S1 in twoqubitS1_with_absorption_of_Y2_X2:
                    gateSetC2_iswap.extend( [ [gate[0]+S1[0], gate[1]+S1[1]] for gate in gateSetC2_iswap_woS1] )

                
            #SWAP-like class: 576 elements
            if randomizeOverIswapAndSwapParts:
                gateSetC2_swap=[]
                for C2 in gateSetC2_sq:
                    twoqubitgatepart = random.sample(all_swaps,1)[0]
                    gateSetC2_swap.append( [C2[0]+twoqubitgatepart[0],C2[1]+twoqubitgatepart[1]] )                  
            else:             
                twoqubitgatepart=[['CZ','-Y/2','CZ','Y/2','CZ'],['CZ','Y/2','CZ','-Y/2','CZ','Y/2']] #order: control, target          
                #add twoqubitgatepart to C1's
                gateSetC2_swap = [ [C2[0]+twoqubitgatepart[0],C2[1]+twoqubitgatepart[1]] for C2 in gateSetC2_sq ]                
            gateSetC2=[]
            gateSetC2.extend(gateSetC2_sq)
            gateSetC2.extend(gateSetC2_cnot)
            gateSetC2.extend(gateSetC2_iswap)
            gateSetC2.extend(gateSetC2_swap)
            
        else:
            gateSetC2=[]
        self.gateSetC2=gateSetC2
        
        if self.numQubits==1:
            self.fullGateSet=[]
            for g in self.gateSetC1:
                self.fullGateSet.append([g])
        elif self.numQubits==2:
            self.fullGateSet=copy.deepcopy(self.gateSetC2)
        else:
            raise Exception('Not implemented for >2 qubits')        
        
        self.singleGateDictNames=['I','X','Y','X/2','Y/2','-X','-Y','-X/2','-Y/2','IGN','IW','IWSE','Zpi'] #this only needs to be the list of single qubit gates of the two qubit clifford generators
        self.singleGateDict = {'I': tomo.Rmat(tomo.sigmaI, 0),
                               'IW': tomo.Rmat(tomo.sigmaI, 0),
                               'IWSE': tomo.Rmat(tomo.sigmaI, 0),
                               'SE' : tomo.Rmat(tomo.sigmaX, np.pi),
                               'IGN': tomo.Rmat(tomo.sigmaI, 0),
                               'X': tomo.Rmat(tomo.sigmaX, np.pi),
                               'Y': tomo.Rmat(tomo.sigmaY, np.pi),
                               'X/2': tomo.Rmat(tomo.sigmaX, np.pi/2.),
                               'Y/2': tomo.Rmat(tomo.sigmaY, np.pi/2.),
                               '-X': tomo.Rmat(tomo.sigmaX, -np.pi),
                               '-Y': tomo.Rmat(tomo.sigmaY, -np.pi),
                               '-X/2': tomo.Rmat(tomo.sigmaX, -np.pi/2.),
                               '-Y/2': tomo.Rmat(tomo.sigmaY, -np.pi/2.),
                               'H': np.array([[1,1],[1,-1]])/np.sqrt(2),
                               'Zpi': np.array([[-1j,0],[0,1j]]),                               
                               'Zpi/2': tomo.Rmat(tomo.sigmaZ, np.pi/2.),
                               'Zpi/4': tomo.Rmat(tomo.sigmaZ, np.pi/4.),
                               '2T': tomo.Rmat(tomo.sigmaZ, np.pi/2.),
                               'Z': np.eye(2),
                               'ZW': np.eye(2),                               
                               'SWAP': np.eye(2),
                               'ACZShape': np.eye(2),
                               'ACZShapeW': np.eye(2)}
        self.fullGateDict = {}
        if numQubits is 1:
            self.fullGateDict = self.singleGateDict
        if numQubits is 2:
            #twoQubitCliffordGeneratorSet = list(itertools.product(self.cliffordGeneratorSet,self.cliffordGeneratorSet))
            twoQubitIter = list(itertools.product(self.singleGateDictNames,self.singleGateDictNames))
            for gate in twoQubitIter:
                self.fullGateDict[gate] = tensor([self.singleGateDict[subGate] for subGate in gate[::-1]]) # we reverse the order of gate
                                                                                                           # we do this because while we pass in gates
                                                                                                           # like ('X/2','Y/2') for X/2 on q0 and Y/2 on q1
                                                                                                           # X/2 tensor Y/2 means 
                                                                                                           # X/2 on q1 with Y/2 on q0, so we must swap
            self.fullGateDict[('CZ','CZ')] = np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]])         
            #self.fullGateDict[('CNOT','CNOT')] = np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]]) #control=1,target=0
            self.fullGateDict[('CNOT','CNOT')] = np.array([[1,0,0,0],[0,0,0,1],[0,0,1,0],[0,1,0,0]])  #control=0,target=1
            
        #precompute unitaries   
        cliffordUnitaries=[]
        for gate in self.fullGateSet:
            cliffordUnitaries.append( self.computeUnitary([gate]) )
        self.cliffordUnitaries=cliffordUnitaries
        
    def randGen(self, iterations, interleaved=False,finish=False): 
        '''
        iterations = m !!!
        This generates a random list of strings that correspond to gates
        e.g. ['X','I','-Y/2','X',....]
        ex:  interleaved = [['CZ'],['CZ']]. MAKE SURE IT IS A CLIFFORD GENERATOR, otherwise there may not be a reversal gate
        '''
        U = np.eye(2**self.numQubits) 
        if interleaved:
            Uinterleaved =  self.computeUnitary([interleaved]) #get unitary of interleaved
            gateList = []
            for k in range(iterations[-1]):
                idx=random.randint(0, len(self.fullGateSet)-1 )
                gateList.append( self.fullGateSet[idx] )
                U=np.dot( self.cliffordUnitaries[idx] , U )
                gateList.append(interleaved)
                U=np.dot( Uinterleaved , U )
        else:
            gateList = []
            for k in range(iterations[-1]):
                idx=random.randint(0, len(self.fullGateSet)-1 )
                gateList.append( self.fullGateSet[idx] )
                U=np.dot( self.cliffordUnitaries[idx] , U )
                
        if finish:
            finalGate=None
            for idx,gate in enumerate(self.fullGateSet):
                Uclifford=self.cliffordUnitaries[idx] #use precomputed unitaries, faster
                Uf = np.dot(Uclifford, U)        
                howCloseToI = abs(np.trace(Uf))**2/(1.*2**(2*self.numQubits))
                if howCloseToI>0.9999:
                    finalGate = gate
                    break
            if finalGate is None:
                raise Exception('Final gate could not be found (when interleaving: use a Clifford generator).')
            gateList = gateList + [finalGate]  
                
        return gateList         
        
    def gateListToOpList(self,gateListIn):
        '''
        This is per clifford!
        Turns  [ ['A','B','C' ..] , ['D','E','F'...]] into  [ [A,D],[B,E]...] etc
        Padds with IGNs to align CZ, and make the sequences the same length
        '''        
        gateList=copy.deepcopy(gateListIn)
        if self.numQubits==1:
            out=[]
            for el0 in gateList[0]:
                #print el0
                out.append([el0])
        elif self.numQubits==2:
            numCZs=gateList[0].count('CZ')
            if numCZs!=gateList[1].count('CZ'):
                raise Exception('dissimilar number of two qubit gates')
            if numCZs>0:
                #use the first CZ to align, and pad IGN's around to get the same length
                idx0=gateList[0].index('CZ');idx1=gateList[1].index('CZ')             
                if idx0<idx1:
                    gateList[0]=['IGN',]*(idx1-idx0)+gateList[0]
                elif idx0>idx1:
                    gateList[1]=['IGN',]*(idx0-idx1)+gateList[1]    
            len0=len(gateList[0]);len1=len(gateList[1])
            if len0<len1:
                gateList[0]+=['IGN',]*(len1-len0)                         
            elif len0>len1:
                gateList[1]+=['IGN',]*(len0-len1)
            out=[]
            for el0,el1 in zip(gateList[0],gateList[1]):
                out.append([el0,el1])
        return out
        
    def computeUnitary(self, gateList):
        '''
        This takes an entire sequence of gates and finds the equivalent unitary
        '''        
        U = np.eye(2**self.numQubits)
        for singleGate in gateList:
            #print singleGate
            operations = self.gateListToOpList(singleGate)
            #print operations
            for op in operations:
                if type(op) is list and len(op)>1:
                    # ['X','Y'] -> ('X','Y')
                    #print len(op)
                    U = np.dot(self.fullGateDict[ tuple(op) ],U)
                elif type(op) is list and len(op)==1:
                    # ['X'] -> 'X', because tuple of this is ('X',), which wont work
                    U = np.dot(self.fullGateDict[ op[0] ],U) ###
                else:
                    #'X'
                    U = np.dot(self.fullGateDict[op],U)            
        return U

    def finishGateList(self, gateList):
        '''
        Picks a final gate from the Clifford group that will make the unitary of the entire sequence 
        be the identity, neglecting a global phase.
        '''        
        U = self.computeUnitary(gateList)
        psi0 = np.zeros((2**self.numQubits,1)) #column vector
        psi0[0] = 1
        finalGate=None
        for idx,gate in enumerate(self.fullGateSet):
            #Uclifford = self.computeUnitary([gate])
            Uclifford=self.cliffordUnitaries[idx] #use precomputed unitaries, faster
            Uf = np.dot(Uclifford, U)        
            howCloseToI = abs(np.trace(Uf))**2/(1.*2**(2*self.numQubits))
            #psif = np.dot(Uf, psi0)
            if howCloseToI>0.9999:
                finalGate = gate
                #print 'found! ',idx
                #print Uf
                break
        if finalGate is None:
            raise Exception('Final gate could not be found (when interleaving: use a Clifford generator).')
        fullGateList = gateList + [finalGate]        
        return fullGateList        
        
    def computeGateLength(self,gateList):
        '''
        Computes the total time needed for the whole gate list. Overestimates, as it does not take into account Cliffords being moved into each other
        '''
        totalTime=0.*ns
        totalTimeArray=[]
        for singleGate in gateList:
            #print singleGate
            operations = self.gateListToOpList(singleGate)
            #print operations
            for op in operations:
                #print op
                sopTime=0.*ns
                for idx,sop in enumerate(op):
                    sopTime = max(sopTime, self.gateLengths[ sop ][idx] )
                    #print sop,sopTime,totalTime
                totalTime+=sopTime
            totalTimeArray+=[totalTime] #this way, we can find the max index for which we can stash in the gates into the sram
        return totalTimeArray    
    
    def computeAvgCliffordLength(self):
        '''
        Computes the average length of the Cliffords
        '''
        totalTimeArray=[]
        for clifford in self.fullGateSet:
            totalTime=0.*ns
            #print singleGate
            operations = self.gateListToOpList(clifford)
            #print operations
            for op in operations:
                #print op
                sopTime=0.*ns
                for idx,sop in enumerate(op):
                    sopTime = max(sopTime, self.gateLengths[ sop ][idx] )
                    #print sop,sopTime
                totalTime+=sopTime
            totalTimeArray+=[totalTime] #this way, we can find the max index for which we can stash in the gates into the sram
            avgTime = np.mean(np.array(totalTimeArray))
        return avgTime  

    def gatesPerClifford(self):
        ''' compute average number of physical gates per clifford '''
        if self.numQubits==1:
            s=0
            for gate in self.fullGateSet:
               s+=np.alen(gate[0])
            return (1.*s)/np.alen(self.fullGateSet)
        elif self.numQubits==2:
            s=0
            for gate in self.fullGateSet:
               s+=np.alen(gate[0])
               s+=np.alen(gate[1])
            return (1.*s)/np.alen(self.fullGateSet)            
        
    def setGateLengths(self,devs):
        '''
        Makes a dictionary with the length of I,X,X/2,CZ from reg:
        gateLengths={ 'I':[0*ns,0*ns] , 'X':[piLenq0,piLenq1] etc...}
        '''
        gateLengths={}
        if self.numQubits==1: 
            set=self.singleGateDictNames
            set+=['H','Z','ACZShape','ACZShapeW','ZW','Zpi','IW','Zpi/2','Zpi/4','2T','IWSE','SE']            
        elif self.numQubits==2:
            set=self.singleGateDictNames+['CZ','IW','Zpi','Zpi/2','Zpi/4']
        for gate in set:
            for devidx,dev in enumerate(devs):
                padding = dev.get('xyPadding',0.0*ns)
                if np.alen(padding) == 1:
                    padding = [padding, padding] # padding before and after the pulse, default symmetric padding if not specified
                padding = sum(padding)
                addtoDic=False                
                if gate is 'I':
                    gateLength=[ dev.get( 'identityLen' , np.min( [dev['piLen'], dev['piHalfLen']] )*ns ) ]
                    addtoDic=True
                if gate in ['IW','IWSE','SE']:
                    gateLength=[ dev.get('identityWaitLen',0.*ns) ]
                    addtoDic=True
                if gate is 'IGN':
                    gateLength=[0.*ns]
                    addtoDic=True                    
                if gate in ['ACZShape', 'ACZShapeW']:
                    if dev.get('aczTarget',False):
                        cztime = [ dev[dev['aczTarget']+'aczTime'] ]
                        gateLength = cztime
                        addtoDic=True                
                if gate in ['Z','ZW']:
                    gateLength=[ dev['detuneLen']+2.*dev['detuneW']+2.*dev.get('detunePadding',0.0) ]
                    addtoDic=True                        
                elif gate in ['X/2','Y/2','-X/2','-Y/2']:
                    gateLength=[ dev['piHalfLen']+padding ]
                    addtoDic=True
                elif gate in ['X','Y','-X','-Y']:
                    gateLength=[ dev['piLen']+padding ]
                    addtoDic=True
                elif gate in ['CZ'] and devidx==0:
                    cztime=dev[ devs[1].__name__+'aczTime' ] + 2.*dev[ devs[1].__name__+'aczW' ]
                    gateLength = [ cztime , cztime ]
                    addtoDic=True
                elif gate in ['H']:
                    gateLength=[ dev['piHalfLen']+dev['piLen']+2.*padding ]
                    addtoDic=True
                elif gate in ['Zpi']:
                    gateLength=[ dev.get('piLenZ',0.0*ns) ]
                    addtoDic=True  
                elif gate in ['Zpi/2']:
                    gateLength=[ dev.get('piHalfLenZ',0.0*ns) ]
                    addtoDic=True 
                elif gate in ['Zpi/4']:
                    gateLength=[ dev.get('piQuarterLenZ',0.0*ns) ]
                    addtoDic=True    
                elif gate in ['2T']:
                    gateLength=[ 2.*dev.get('piQuarterLenZ',0.0*ns) ]
                    addtoDic=True                     
                if addtoDic:    
                    if gate in gateLengths:   
                        gateLengths[gate]+=gateLength
                    else:
                        gateLengths[gate]=gateLength
        self.gateLengths = gateLengths        
                    
    def testSequence(self, length = 10, interleaved=None):        
        gateList = self.randGen([length], interleaved=interleaved)
        print 'gateList: %s' % gateList        
        Ui = self.computeUnitary(gateList)        
        fullGateList = self.finishGateList(gateList)
        print 'fullGateList: %s' % fullGateList        
        U = self.computeUnitary(fullGateList)
        #psi0 = np.zeros((2**self.numQubits,1)) #column vector
        #psi0[0] = 1.
        #psif = np.dot(U, psi0)
        #fidelity = np.dot(dagger(psif),psi0)
        fidelity = abs(np.trace(U))**2/(1.*2**(2*self.numQubits))        
        return abs(fidelity)
   
    def test2design(self):
        listToTest=self.cliffordUnitaries
        N=len(listToTest)
        runningSum=0.
        for k in range(N):
            print 'percentage completed: %f' %(100*(1.*k)/(N))
            a=dagger(listToTest[k])
            for l in range(N):
                runningSum+=abs( np.trace( np.dot(a,listToTest[l]) ) )**4
        runningSum*=1./(2*N**2)
        return runningSum

# TODO
# Make fitting easier x
# remove legacy code x
# make it handle 012 state discrimination
        
p0 = [0.999, np.pi/4., np.pi/4] # free A and B
p0AorB = [0.999, np.pi/4.] # A set
p0AB = [0.999] # A and B set


def fitFunc(A=None, B=None):
    if A and B:
        return lambda m, p: A*p**m + B, p0AB
    elif A:
        return lambda m, p, B: A*p**m + B, p0AorB
    elif B:
        return lambda m, p, A: A*p**m + B, p0AorB
    else:
        return lambda m, p, A, B: A*p**m + B, p0
        
def fitData(ms, sequence_fidelities, A=None, B=None, p0=None):
    func, p0default = fitFunc(A, B)
    if p0 is None:
        p0 = p0default
    p1, cov = curve_fit(func, ms, sequence_fidelities, p0 = p0, maxfev = int(1e5))
    p1err = np.sqrt( abs( np.diag(cov) ) ) #var[element x] = covariance matrix [x,x]    
    if A and B:
        params = {'A':A,'B':B,'p':p1[0],'Aerr':0.,'Berr':0.,'perr':p1err[0]}
    elif A:
        params = {'A':A,'B':p1[1],'p':p1[0],'Aerr':0.,'Berr':p1err[1],'perr':p1err[0]}
    elif B:
        params = {'A':p1[1],'B':B,'p':p1[0],'Aerr':p1err[1],'Berr':0.,'perr':p1err[0]}
    else:
        params = {'A':p1[1],'B':p1[2],'p':p1[0],'Aerr':p1err[1],'Berr':p1err[2],'perr':p1err[0]}
        
        
    return params
    

def fitdata(x,y,yerr,order,A0=None,B0=None):
    #p1,p1err,func,A0,B0,p,A0err,B0err,perr=fitdata(dvwData,A0,B0)
    #x = m from Magesan et al., PRL 109, 080505 (2012)

    if order==0:
        C=None
        Cerr=None
        """Zeroth order model, Eq.2 in PRL 106, 180504 (2011)"""
        if A0 is None and B0 is None:
            #func = lambda m, A0, p, B0: np.sin(A0)**2*p**m+np.sin(B0)**2 # see PRL 106, 180504 (2011)
            #p0 = [np.pi/4., 0.99, np.pi/4.]
            func = lambda m, A0, p, B0: A0*p**m+B0 # see PRL 106, 180504 (2011)
            p0 = [0.5, 0.99, 0.4]
        elif A0 is None and B0:
            func = lambda m, A0, p: A0*p**m+B0 # see PRL 106, 180504 (2011)
            p0 = [0.6, 0.99]  
        else:
            #we have A0 and B0
            func = lambda m, p: A0*p**m+B0 # see PRL 106, 180504 (2011)
            p0 = [0.99]

        p1, cov = curve_fit(func, x, y, sigma=yerr,p0 = p0, maxfev = 1e5)
        p1err = np.sqrt( abs( np.diag(cov) ) ) #var[element x] = covariance matrix [x,x]
        
        if A0 is None and B0 is None:    
            A0, p, B0 = p1
            A0err, perr, B0err = p1err
        elif A0 is None and B0:
            A0, p = p1
            A0err, perr = p1err        
            B0err =0.
        else:
            p=p1[0]
            perr=p1err[0]
            A0err=0.
            B0err=0.
            
    elif order==1:
        """First order model, Eq.2 in PRL 106, 180504 (2011)"""
        if A0 is None and B0 is None:
            #func = lambda m, A0, p, B0,C: np.sin(A0)**2*p**m+C*(m-1)*p**(m-2)+np.sin(B0)**2 # see PRL 106, 180504 (2011)
            #p0 = [np.pi/4., 0.99, np.pi/4.,0.]
            func = lambda m, A0, p, B0,C: A0*p**m+C*(m-1)*p**(m-2)+B0 # see PRL 106, 180504 (2011)
            p0 = [0.8, 0.99, 0.2,0.]
        elif A0 is None and B0:
            func = lambda m, A0, p, C: A0*p**m+C*(m-1)*p**(m-2)+B0 # see PRL 106, 180504 (2011)
            p0 = [0.8, 0.99, 0.]            
        else:
            #we have A0 and B0
            func = lambda m, p,C: A0*p**m+C*(m-1)*p**(m-2) + B0 # see PRL 106, 180504 (2011)
            p0 = [0.99,0.]

        p1, cov = curve_fit(func, x, y, sigma=yerr,p0 = p0, maxfev = 1e5)
        p1err = np.sqrt( abs( np.diag(cov) ) ) #var[element x] = covariance matrix [x,x]
        
        if A0 is None and B0 is None:    
            A0, p, B0,C = p1
            A0err, perr, B0err,Cerr = p1err
        elif A0 is None and B0:
            A0, p, C = p1
            A0err, perr, Cerr = p1err
            B0err=0.
            print 'bla'
        else:
            p=p1[0]
            perr= p1err[0]
            A0err=0.
            B0err=0.
            C=p1[1]
            Cerr=p1err[1]
    else:
        raise Exception('not implemented')
    print p1,cov
    print 'A0: %s, B0: %s, C: %s' % ( A0, B0, C)
    print 'A0err: %s, B0err: %s, Cerr: %s' % ( A0err, B0err, Cerr)
    return AttrDict({'A0':A0, 'B0':B0, 'C':C, 'p':p, 'fitFunc':func,'p1':p1,'A0err':A0err,'B0err':B0err, 'Cerr':Cerr, 'perr':perr, 'p1err':p1err})
     
  
    
def pToErr(p,d):
    r = (1.- 1./d) * (1.-p)
    return r

def pToRErr(perr,d):
    rerr = perr * (1.- 1./d)
    return rerr
    
def ppcToErr(p,pc,d):
    rc = (d-1.)/(1.0*d) * (1.-pc/p)
    return rc
 
def ppcToRErr(perr,pierr,p,pc,d):
    rcerr = (d-1.)/(1.0*d) *  ( pierr/pc + perr/p) * pc/p  
    return rcerr
 
def msFidelitiesToMeanStd(raw_ms, sequence_fidelities):
    
    setMs = np.sort(list(set(raw_ms)))
    
    sequence_fidelities_mean = []
    sequence_fidelities_std = []
    ms = []
    for m in setMs:
        idx = np.argwhere(raw_ms==m)
        sequence_fidelities_mean.append(np.mean(sequence_fidelities[idx]))
        sequence_fidelities_std.append(np.std(sequence_fidelities[idx])/np.sqrt(len(idx))) # divide by sqrt(k) for 1/sqrt(n) statistics
        ms.append(m)
    
    return ms, np.array(sequence_fidelities_mean), np.array(sequence_fidelities_std)
    
def plotRef(ms, sequence_fidelities, params):
    
    ms, fid, std = msFidelitiesToMeanStd(ms, sequence_fidelities)
    
    func, _ = fitFunc(params['A'], params['B'])
    
    plt.figure()
    plt.plot(ms, func(ms, params['p']),'-k')
    plt.errorbar(ms, fid, std, fmt='ob')
    plt.xlabel('M')
    plt.ylabel('Sequence Fidelity')
    plt.title('Reference p: %0.4f' % params['p'])
    
def plotInterleaved(ms_ref, sequence_fidelities_ref, params_ref, 
                    ms_inter, sequence_fidelities_inter, params_inter):
    
    ms_ref, fid_ref, std_ref = msFidelitiesToMeanStd(ms_ref, sequence_fidelities_ref)
    func_ref, _ = fitFunc(params_ref['A'], params_ref['B'])
    
    ms_inter, fid_inter, std_inter = msFidelitiesToMeanStd(ms_inter, sequence_fidelities_inter)
    func_inter, _ = fitFunc(params_inter['A'], params_inter['B'])
    
    plt.figure()
    plt.plot(ms_ref, func_ref(ms_ref, params_ref['p']),'-b')
    plt.errorbar(ms_ref, fid_ref, std_ref, fmt='ob')    
    plt.plot(ms_inter, func_inter(ms_inter, params_inter['p']),'-r')
    plt.errorbar(ms_inter, fid_inter, std_inter, fmt='or')
    plt.xlabel('M')
    plt.ylabel('Sequence Fidelity')
    plt.title('Reference p: %0.4f, Interleaved p: %0.4f' % (params_ref['p'],params_inter['p']))
    
def plotSimultaneous(ms_raw, all_seq_fids, all_params):
    
    all_fids = []
    all_stds = []
    all_funcs = []
    for seq, params in zip(all_seq_fids, all_params):
        ms, fid, std = msFidelitiesToMeanStd(ms_raw, seq)
        all_fids.append(fid)
        all_stds.append(std)
        func, _ = fitFunc(params['A'], params['B'])
        all_funcs.append(func)
    
    plt.figure()
    for fid,std,func,params in zip(all_fids, all_stds, all_funcs, all_params):
        plt.plot(ms, func(ms, params['p']),'-')
        plt.errorbar(ms, fid, std, fmt='o')
    plt.title(''.join(['Reference p: %0.4f, ' % params['p'] for params in all_params]))       
    
def plotLeakage(ms_raw, leakage, params):
    
    pass
    
def reference(dvwData, A=None, B=None, plot=True, noisy=True):

    qubits = [dvwData.parameters[meas] for meas in dvwData.parameters['measure']]
    d = 2**len(qubits)
    
    ms = dvwData[:,1]
    sequence_fidelities = dvwData[:,2]
    
    params = fitData(ms, sequence_fidelities, A, B)
    if plot:
        plotRef(ms, sequence_fidelities, params)
        
    r = pToErr(params['p'], d)
    rerr = pToRErr(params['perr'], d)    
    params['r']=r
    params['rerr']=rerr
    if noisy:
        print 'Average Error: %0.4f' % r
        print 'A: %0.3f, B: %0.3f' % (params['A'], params['B'])        
    return r, params
    
def interleaved(dvwDataRef, dvwDataInterleaved, A=None, B=None, plot=True, noisy=True): 
    
    qubits = [dvwDataRef.parameters[meas] for meas in dvwDataRef.parameters['measure']]
    d = 2**len(qubits)
    
    ms_ref = dvwDataRef[:,1]
    sequence_fidelities_ref = dvwDataRef[:,2]
    params_ref = fitData(ms_ref, sequence_fidelities_ref, A, B)
    
    ms_inter = dvwDataInterleaved[:,1]
    sequence_fidelities_inter = dvwDataInterleaved[:,2]
    params_inter = fitData(ms_inter, sequence_fidelities_inter, params_ref['A'], params_ref['B'])
    
    r = pToErr(params_ref['p'],d)
    rc = ppcToErr(params_ref['p'],params_inter['p'],d)
    rcerr = ppcToRErr(params_ref['perr'],params_inter['perr'],params_ref['p'],params_inter['p'],d)
    params_inter['rcerr']=rcerr
    
    if plot:
        plotInterleaved(ms_ref, sequence_fidelities_ref, params_ref,
                        ms_inter, sequence_fidelities_inter, params_inter)
    
    if noisy:
        print 'Average Reference Error: %0.4f' % r
        print 'Average Interleaved Error: %0.4f' % rc
        print 'A: %0.3f, B: %0.3f' % (params_ref['A'], params_ref['B'])
    
    return r, rc, params_ref, params_inter
    
def simultaneous(dvwData, A=None, B=None, plot=True, noisy=True):
    measure = dvwData.parameters['measure']
    correlatedProbs = dvwData.parameters['correlatedProbs']
    qubits = [dvwData.parameters[meas] for meas in dvwData.parameters['measure']]
    qubitsRB = [dvwData.parameters[meas] for meas in dvwData.parameters['measureRB']]
    if correlatedProbs:
        # please implement:
        d = 2**len(qubits)
        raise Exception('Use plotSimultaneousRB')
    else:
        d=2
        all_seq_fids = []
        all_params = []
        all_rs = []
        all_rs_errs = []
        ms = dvwData[:,1]
        for q in dvwData.parameters['measureRB']:
            idx = measure.index(q)
            sequence_fidelities = dvwData[:,2+idx]
            all_seq_fids.append(sequence_fidelities)
        for seq_fid in all_seq_fids:
            params = fitData(ms, seq_fid, A, B)
            r = pToErr(params['p'], d)
            rerr = pToRErr(params['perr'], d)    
            params['r']=r
            params['rerr']=rerr
            all_rs.append(r)
            all_params.append(params)
            if noisy:
                print 'Average Reference Error: %0.4f' % r
                print 'A: %0.3f, B: %0.3f' % (params['A'], params['B'])
        
        if plot:
            plotSimultaneous(ms, all_seq_fids, all_params)
            
        return all_rs, all_rs_errs, all_params
        
def leakage(dvwData, noisy=True, plot=True):
    
    ms = dvwData[:,1]
    leakage = dvwData[:,3]
    
    params = fitData(ms, 1-leakage, None, None)
    if plot:
        plotRef(ms, 1-leakage, params)
        
    return params

# def plotSingleRB(dvwData,order=0,A0=None,B0=None,M=None):
    # # raise Exception('Deprecated - use plotMultiRB instead')
    # qname = dvwData.parameters['config'][dvwData.parameters['measure'][0]]

    # gateNumbers = []
    # fidelities = []
    # for column in dvwData[:,1:].transpose():
        # gateNumbers = np.hstack((gateNumbers, dvwData[:,0].transpose()))
        # fidelities = np.hstack((fidelities, column))

    # # p1, cov = curve_fit(fitFunc, gateNumbers, fidelities, p0 = p0, maxfev = 1e5)
    # # p1err = np.sqrt( abs( np.diag(cov) ) ) #var[element x] = covariance matrix [x,x]
    
    # rawFidelityData = dvwData[:,1:]
    # fidelitySTD = np.std(rawFidelityData,axis=1,ddof=1)
    # fidelityMean = np.mean(rawFidelityData,axis=1)
    # plt.errorbar(dvwData[:,0], fidelityMean, yerr=fidelitySTD,fmt='ob',label = 'q %s data' % qname)

    # result=fitdata(dvwData[:,0],fidelityMean,1./(fidelitySTD**2.),order,A0,B0) #averaged data, using wi=1/sigma_i**2, with sigma_i**2 the 
    # p1=result['p1']
    # A0=result['A0']
    # B0=result['B0']
    # A0err=result['A0err']
    # B0err=result['B0err']    
    # p=result['p']
    # perr=result['perr']    
    
    # plotGateNumbers = np.arange(min(gateNumbers),max(gateNumbers))

    # plt.plot(plotGateNumbers, result['fitFunc'](plotGateNumbers, *p1),'-r',label = 'q %s fit' % qname)

    # # A0, p, B0 = p1
    # # A0err, perr, B0err = p1err
    
    
    # d=2**1 # 1 qubit
    
    # r = 1-p-(1-p)/d # d=2**n the dimension of the system
    # rerr = perr* (1.- 1./d)
    # fidelity = 1-r
    # fidelityerr = perr* (1.- 1./d)
    # print 'p: %s  +- %s' % (p,perr)
    # print 'r: %s  +- %s' % (r,rerr)      
    # print 'Fidelity: %s  +- %s' % (fidelity,fidelityerr)
    # # print 'A0: %s, B0: %s' % (np.sin(A0)**2, np.sin(B0)**2)
    # print 'A0: %s, B0: %s' % ( A0, B0 )    

    # plt.xlabel('Number of Gates')
    # plt.ylabel('Sequence Fidelity')
    # plt.title('Randomized Benchmarking Multi %s fidelity: %s, p: %s' % (qname, 1-r, p))
    # plt.legend()
    # return r, p, A0, B0, d

def plotSimultaneousRB(dvwData,cols,order=0,A0=None,B0=None,collectAll=False):
    '''
    syntax = plotSimultaneousRB(dvw[..],[0,2])  #adds P00+P10
    '''

    # qnames = [dvwData.parameters['config'][meas] for meas in dvwData.parameters['measure']]
    qnames = dvwData.parameters['measure']
    # seq = int(dvwData.parameters['stack.frame0.arg.seq'])
    
    
    k = dvwData.parameters.get('k',None)
    if k is None:
        k = int(dvwData.parameters['stack']['frame0']['arg']['k'])
    gateNumbers = []
    fidelities = []    
    
    dataFormat=dvwData.parameters.get('dataFormat',None)
    if dataFormat in ['stacked2D','2D','stacked2Dswapped','2Dswapped']:
        if dataFormat in ['stacked2D','2D']:        
            m_axisidx=0
        elif dataFormat in ['stacked2Dswapped','2Dswapped']:
            m_axisidx=1
        else:
            raise Exception('not implemented')
        ms = set(dvwData[:,m_axisidx])
            
        # # k = dvwData.parameters['k']
        xAxis = []
        fidelityMean=np.matrix(np.zeros(shape=(len(ms),4))) #oredefine
        fidelitySTD=np.matrix(np.zeros(shape=(len(ms),4)))
        for idx,m in enumerate(ms):
            currInds = np.argwhere(dvwData[:,m_axisidx]==m)
            if m_axisidx==0:
                if len(currInds)<k: # if incomplete column, cut it off
                    continue
            currFidelities = [np.array(dvwData[ind,2:][0]) for ind in currInds] #this has all P00,P01,P10 and P11 for a certain m
            fidelityMean[ idx,:]= np.mean( np.matrix(currFidelities),axis=0 )  #this a single array of P00,P01,P10 and P11 for a single m, averaged over all sequences
            fidelitySTD[ idx,:] = np.std( np.matrix(currFidelities),axis=0,ddof=1 )
            xAxis.append(m)
        xAxis = np.array(xAxis)
    else:
        raise Exception('not implemented, use simultaneousold')

    fidelityMeanAdded=np.matrix(np.zeros(shape=(len(xAxis),1))) #predefine a column vector
    fidelityVarAdded=np.matrix(np.zeros(shape=(len(xAxis),1)) )

    for col in cols:
        fidelityMeanAdded += fidelityMean[:,col]
        # fidelityVarAdded += (fidelitySTD[:,col])**2 #yes, we need to add the variance, assuming it's uncorrelated
        fidelityVarAdded += np.multiply( fidelitySTD[:,col] , fidelitySTD[:,col] ) #yes, we need to add the variance, assuming it's uncorrelated
    fidelitySTDAdded = np.sqrt(fidelityVarAdded)
    fidelityMeanAdded, =np.array( fidelityMeanAdded[:,0].T ) #turns a column vector into an array
    fidelitySTDAdded, =np.array( fidelitySTDAdded[:,0].T )
   
    result=fitdata(xAxis,fidelityMeanAdded,1./(fidelitySTDAdded**2.),order,A0,B0)     
    
    p1=result['p1']
    A0=result['A0']
    B0=result['B0']
    p=result['p']
    perr=result['perr']  
    fidelitySTDAdded=fidelitySTDAdded/np.sqrt(k) #standard deviation from the mean
    col='r'
    plt.errorbar(xAxis, fidelityMeanAdded, yerr=fidelitySTDAdded,fmt='o'+col,markerfacecolor='none',markeredgecolor=col,label = 'MQ data')    
    xAxisForFit=np.linspace(-3,max(xAxis),1001)
    plt.plot(xAxisForFit, result['fitFunc'](xAxisForFit, *p1),'-'+col,label = 'MQ fit')
   
    d=2**1 # n qubits, d=2**n the dimension of the system  
    r = (1.- 1./d) * (1.-p)
    rerr = perr * (1.- 1./d)
    fidelity = 1-r    
    fidelityerr = perr * (1.- 1./d)
    
    print 'p: %s  +- %s' % (p,perr)
    print 'r: %s  +- %s' % (r,rerr)    
    print 'Fidelity: %s  +- %s' % (fidelity,fidelityerr)
    print 'A0: %s, B0: %s' % ( A0, B0 )
    
    axismode=dvwData.parameters.get('axismode',None)
    if axismode=='m':
        plt.xlabel('m - number of sets of Cliffords/gates+interleaves')
    else:
        plt.xlabel('number of gates')
    plt.ylabel('Added Sequence Fidelity of '+str(cols) )
    # plt.title('Randomized Benchmarking Multi %s fidelity: %s, p: %s' % (qnames, 1-r, p))
    # plt.legend()
    plt.grid(color='gray', linestyle='dashed')
    if collectAll:
        return r, rerr, p, A0, B0, d, perr,axismode, xAxis,fidelityMeanAdded,fidelitySTDAdded,xAxisForFit,result['fitFunc'](xAxisForFit, *p1)
    else:
        return r, rerr, p, A0, B0, d, perr,axismode
    
    
# def fitdata(x,y,yerr,order,A0=None,B0=None):
    # # p1,p1err,func,A0,B0,p,A0err,B0err,perr=fitdata(dvwData,A0,B0)
    # # x = m from Magesan et al., PRL 109, 080505 (2012)

    # if order==0:
        # C=None
        # Cerr=None
        # """Zeroth order model, Eq.2 in PRL 106, 180504 (2011)"""
        # if A0 is None and B0 is None:
            # # func = lambda m, A0, p, B0: np.sin(A0)**2*p**m+np.sin(B0)**2 # see PRL 106, 180504 (2011)
            # # p0 = [np.pi/4., 0.99, np.pi/4.]
            # func = lambda m, A0, p, B0: A0*p**m+B0 # see PRL 106, 180504 (2011)
            # p0 = [0.5, 0.99, 0.4]
        # elif A0 is None and B0:
            # func = lambda m, A0, p: A0*p**m+B0 # see PRL 106, 180504 (2011)
            # p0 = [0.6, 0.99]  
        # else:
            # # we have A0 and B0
            # func = lambda m, p: A0*p**m+B0 # see PRL 106, 180504 (2011)
            # p0 = [0.99]

        # p1, cov = curve_fit(func, x, y, sigma=yerr,p0 = p0, maxfev = 1e5)
        # p1err = np.sqrt( abs( np.diag(cov) ) ) #var[element x] = covariance matrix [x,x]
        
        # if A0 is None and B0 is None:    
            # A0, p, B0 = p1
            # A0err, perr, B0err = p1err
        # elif A0 is None and B0:
            # A0, p = p1
            # A0err, perr = p1err        
            # B0err =0.
        # else:
            # p=p1[0]
            # perr=p1err[0]
            # A0err=0.
            # B0err=0.
            
    # elif order==1:
        # """First order model, Eq.2 in PRL 106, 180504 (2011)"""
        # if A0 is None and B0 is None:
            # # func = lambda m, A0, p, B0,C: np.sin(A0)**2*p**m+C*(m-1)*p**(m-2)+np.sin(B0)**2 # see PRL 106, 180504 (2011)
            # # p0 = [np.pi/4., 0.99, np.pi/4.,0.]
            # func = lambda m, A0, p, B0,C: A0*p**m+C*(m-1)*p**(m-2)+B0 # see PRL 106, 180504 (2011)
            # p0 = [0.8, 0.99, 0.2,0.]
        # elif A0 is None and B0:
            # func = lambda m, A0, p, C: A0*p**m+C*(m-1)*p**(m-2)+B0 # see PRL 106, 180504 (2011)
            # p0 = [0.8, 0.99, 0.]            
        # else:
            # # we have A0 and B0
            # func = lambda m, p,C: A0*p**m+C*(m-1)*p**(m-2) + B0 # see PRL 106, 180504 (2011)
            # p0 = [0.99,0.]

        # p1, cov = curve_fit(func, x, y, sigma=yerr,p0 = p0, maxfev = 1e5)
        # p1err = np.sqrt( abs( np.diag(cov) ) ) #var[element x] = covariance matrix [x,x]
        
        # if A0 is None and B0 is None:    
            # A0, p, B0,C = p1
            # A0err, perr, B0err,Cerr = p1err
        # elif A0 is None and B0:
            # A0, p, C = p1
            # A0err, perr, Cerr = p1err
            # B0err=0.
            # print 'bla'
        # else:
            # p=p1[0]
            # perr= p1err[0]
            # A0err=0.
            # B0err=0.
            # C=p1[1]
            # Cerr=p1err[1]
    # else:
        # raise Exception('not implemented')
    # print p1,cov
    # print 'A0: %s, B0: %s, C: %s' % ( A0, B0, C)
    # print 'A0err: %s, B0err: %s, Cerr: %s' % ( A0err, B0err, Cerr)
    # return AttrDict({'A0':A0, 'B0':B0, 'C':C, 'p':p, 'fitFunc':func,'p1':p1,'A0err':A0err,'B0err':B0err, 'Cerr':Cerr, 'perr':perr, 'p1err':p1err})
     
   
# def plotMultiRB(dvwData,order=0,A0=None,B0=None,M=None,col='b',collectAll=False,plot=True):    
    # qnames = dvwData.parameters['measure']
    # gateNumbers = []
    # fidelities = []
    # k=dvwData.parameters.get('k')
    # dataFormat=dvwData.parameters.get('dataFormat',None)
    # if dataFormat == 'stacked' or dataFormat == 'stackedInterleaved':
        # # it's in P00, P01, P10, P11, let's just pull the P00 out
        # gateNumbers=dvwData[:,0]
        # if M is not None:
            # # correct for measurement fidelity
            # Probs=dvwData[:,1::]
            # correctedProbs=(np.dot(np.linalg.inv(M),Probs.T)).T
            # fidelities=np.array(correctedProbs[:,0])
            # print type(fidelities)
        # else:
            # fidelities=dvwData[:,1]
        # rows=np.fix((len(fidelities)/k))
        # fidelities=fidelities[0:rows*k] #this cuts off incomplete data
        # rawFidelityData = np.reshape(fidelities,(len(fidelities)/k,k))
        # xAxis = np.reshape(gateNumbers,(len(fidelities)/k,k))[:,0]    
        # fidelitySTD = np.std(rawFidelityData,axis=1,ddof=1)
        # fidelityMean = np.mean(rawFidelityData,axis=1)
    # elif dataFormat in ['stacked2D','2D','stacked2Dswapped','2Dswapped']:
        # if dataFormat in ['stacked2D','2D']:        
            # m_axisidx=0
        # elif dataFormat in ['stacked2Dswapped','2Dswapped']:
            # m_axisidx=1
        # else:
            # raise Exception('not implemented')
        # ms = set(dvwData[:,m_axisidx])
            
        # fidelityMean=[]
        # fidelitySTD=[]
        # k = dvwData.parameters['k']
        # xAxis = []
        # for m in ms:
            # currInds = np.argwhere(dvwData[:,m_axisidx]==m)
            # if m_axisidx==0:
                # if len(currInds)<k: # if incomplete column, cut it off
                    # continue
            # currFidelities = [dvwData[ind,2] for ind in currInds]
            # fidelityMean.append(np.mean(currFidelities))
            # fidelitySTD.append(np.std(currFidelities,ddof=1))
            # xAxis.append(m)
            
            
        # fidelityMean = np.array(fidelityMean)
        # fidelitySTD = np.array(fidelitySTD)
        # xAxis = np.array(xAxis)
    # else:
        # for column in dvwData[:,1:].transpose():
            # gateNumbers = np.hstack((gateNumbers, dvwData[:,0].transpose()))
            # fidelities = np.hstack((fidelities, column))
        # rawFidelityData = dvwData[:,1:]
        # fidelitySTD = np.std(rawFidelityData,axis=1,ddof=1)
        # fidelityMean = np.mean(rawFidelityData,axis=1)
        # xAxis=dvwData[:,0]
    # result=fitdata(xAxis,fidelityMean,1./(fidelitySTD**2.),order,A0,B0) #averaged data, using wi=1/sigma_i**2, with sigma_i**2 the variance    
    
    # p1=result['p1']
    # A0=result['A0']
    # B0=result['B0']
    # p=result['p']
    # perr=result['perr'] #perr is a standard deviation
    # fidelitySTD=fidelitySTD/np.sqrt(k) #standard deviation from the mean

    # xAxisForFit=np.linspace(0.,max(xAxis),1001)    
    # if plot:
        # plt.errorbar(xAxis, fidelityMean, yerr=fidelitySTD,fmt='o'+col,markerfacecolor='none',markeredgecolor=col,label = 'MQ data')    
        # plt.plot(xAxisForFit, result['fitFunc'](xAxisForFit, *p1),'-'+col,label = 'MQ fit')
   
    # d=2**(len(qnames)) # n qubits, d=2**n the dimension of the system  
    # r = (1.- 1./d) * (1.-p)
    # rerr = perr * (1.- 1./d)
    # fidelity = 1-r    
    # fidelityerr = perr * (1.- 1./d)
    
    # print 'p: %s  +- %s' % (p,perr)
    # print 'r: %s  +- %s' % (r,rerr)    
    # print 'Fidelity: %s  +- %s' % (fidelity,fidelityerr)
    # print 'A0: %s, B0: %s' % ( A0, B0 )
    
    # axismode=dvwData.parameters.get('axismode',None)
    
    # if plot:
        # if axismode=='m':
            # plt.xlabel('m - number of sets of Cliffords/gates+interleaves')
        # else:
            # plt.xlabel('number of gates')
        # plt.ylabel('Sequence Fidelity')
        # # plt.title('Randomized Benchmarking Multi %s fidelity: %s, p: %s' % (qnames, 1-r, p))
        # # plt.legend()
        # plt.grid(color='gray', linestyle='dashed')
        
    # if collectAll:
        # return r, p, A0, B0, d, perr,axismode, xAxis, fidelityMean, fidelitySTD, xAxisForFit, result['fitFunc'](xAxisForFit, *p1), result
    # else:
        # return r, p, A0, B0, d, perr,axismode


    
# def analyzeInterleavedRB(dvwDataControl, dvwDataInterleaved,order=0,A0m=None,B0m=None,M=None,plot=True):
    # r, p, A0, B0, d,perr,axismode = plotMultiRB(dvwDataControl,order,A0m,B0m,M,col='b',plot=plot)
    # ri, pi, A0i, B0i, d,pierr,axismodei = plotMultiRB(dvwDataInterleaved,order,A0m,B0m,M,col='g',plot=plot)
    
    # if axismode is axismodei is 'm':
        # # plotted vs m
        # pc = pi 
        # rcerr = (d-1.)/(1.0*d) *  ( pierr/pi + perr/p) * pc/p        
    # elif axismode is axismodei:
        # # plotted vs number of gates.
        # # because of this, for 'm=1', we have n_gates =2 (excluding recovery), and pc = pi^2, as the x axis is number of gates, not m
        # print 'OLD DATA!'
        # pc = pi **2  
        # rcerr = (d-1.)/(1.0*d) *  ( 2.* pierr/pi + perr/p) * pc/p        
    # else:
        # raise Exception('dissimilar axis mode')
    # # print 'd=',d
    # rc = (d-1.)/(1.0*d) * (1.-pc/p)
    # E1=(d-1.)/d * ( abs(p-pc/p) + (1.-p) )
    # E2=2.*(d**2-1.)*(1.-p)/p/d**2    +     4.*np.sqrt(1-p)*np.sqrt(d**2-1.)/p
    # E=min(E1,E2)
    # print 'rc is: %s +- %s' % (rc,rcerr)
    # print 'rc must lie between: %s and %s' % (rc-E,rc+E)    
    # print 'Fidelity: %s +- %s' % (1-rc,rcerr)    
    # return rc,r,ri,rcerr

def getM(data):
    """Get measfid from tomography. RB"""
    # dvw = DataVaultWrapper(Sample)
    # data=dvw[ds]
    measure = data.parameters['measure']
    Probs=np.matrix(data[:,1:])
    nqubits=len(measure)
    # analyse data
    if nqubits==1:
        # measurement fidelity matrix definition: Pactual= inv(M) * Pmeas, Pmeas = M*Pactual -> M = Pmeas*inv(Pactual)
        Pideal=np.array([[1,0,],[0,1]])
        Pmeas=np.hstack( [ Probs[0,:].T , Probs[5,:].T ])
        M=np.dot(Pmeas,np.linalg.inv(Pideal))
     
    if nqubits==2:
        # 0-II
        # 5-IX
        # 30-XI
        # 35-XX
        Pideal=np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0,],[0,0,0,1]])
        Pmeas=np.hstack( [ Probs[0,:].T , Probs[5,:].T, Probs[30,:].T, Probs[35,:].T ])
        M=np.dot(Pmeas,np.linalg.inv(Pideal))
    return M   
    
def getA0B0(data):
    """Get A0, B0 from tomography. RB"""
    # dvw = DataVaultWrapper(Sample)
    # data=dvw[ds]
    measure = data.parameters['measure']
    Probs=np.matrix(data[:,1:])
    nqubits=len(measure)
    # analyse data
    if nqubits==1:
        # measurement fidelity matrix definition: Pactual= inv(M) * Pmeas, Pmeas = M*Pactual -> M = Pmeas*inv(Pactual)
        Pideal=np.array([[1,0,],[0,1]])
        Pmeas=np.hstack( [ Probs[0,:].T , Probs[5,:].T ])
        M=np.dot(Pmeas,np.linalg.inv(Pideal))
     
    if nqubits==2:
        # 0-II
        # 5-IX
        # 30-XI
        # 35-XX
        Pideal=np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0,],[0,0,0,1]])
        Pmeas=np.hstack( [ Probs[0,:].T , Probs[5,:].T, Probs[30,:].T, Probs[35,:].T ])
        M=np.dot(Pmeas,np.linalg.inv(Pideal))

    groundstate=np.zeros((2**nqubits,1));groundstate[0,0]=1.
    rbend=np.ones((2**nqubits,1))/(2**nqubits)

    Pmeas_rbstart=np.dot(M,groundstate)
    Pmeas_rbend=np.dot(M,rbend)
    print Pmeas_rbstart
    print Pmeas_rbend

    B0=Pmeas_rbend[0,0]
    A0=Pmeas_rbstart[0,0] -B0    
    # print np.linalg.inv(M)*Pmeas
    return A0,B0    

def getlength(RBClass,maxtime,interleaved=False,mstart=1,mstep=5):
    #gets the maximum m which fits in maxtime. make sure to set the gate times first
    averageover=5
    stayinloop=True
    m=mstart
    while stayinloop:
        seqLengths=[]
        for av in range(averageover):
            seqf=RBClass.randGen([m],interleaved=interleaved,finish=True)  
            seqLength= RBClass.computeGateLength(seqf)
            seqLengths.append( seqLength[-1] )
        if np.max(seqLengths)>maxtime:
            np.max(seqLengths), m
            #sequence too long, decrease m and get out
            m-=mstep            
            stayinloop=False
        else:
            m+=mstep        
    return m

'''
        if finish:
            finalGate=None
            howCloseToI=np.zeros(len(self.fullGateSet))
            for idx,gate in enumerate(self.fullGateSet):
                Uclifford=self.cliffordUnitaries[idx] #use precomputed unitaries, faster
                Uf = np.dot(Uclifford, U)        
                howCloseToI[idx] = abs(np.trace(Uf))**2
            idx=find(howCloseToI> (0.9999* 2**(2*self.numQubits)) )
            if np.alen(idx)==1:
                finalGate = self.fullGateSet[idx]
            if finalGate is None:
                raise Exception('Final gate could not be found (when interleaving: use a Clifford generator).')
            gateList = gateList + [finalGate]   
'''

'''
def fitFunc(A=None, B=None):
    if A and B:
        return lambda m, p: A*p**m + B, p0AB
    elif A:
        return lambda m, p, B: A*p**m + np.sin(B)**2, p0AorB
    elif B:
        return lambda m, p, A: np.sin(A)**2*p**m + B, p0AorB
    else:
        return lambda m, p, A, B: np.sin(A)**2*p**m + np.sin(B)**2, p0
        
def fitData(ms, sequence_fidelities, A=None, B=None, p0=None):
    func, p0default = fitFunc(A, B)
    if p0 is None:
        p0 = p0default
    p1, cov = curve_fit(func, ms, sequence_fidelities, p0 = p0, maxfev = int(1e5))
    if A and B:
        params = {'A':A,'B':B,'p':p1[0]}
    elif A:
        params = {'A':A,'B':np.sin(p1[1])**2,'p':p1[0]}
    elif B:
        params = {'A':np.sin(p1[1])**2,'B':B,'p':p1[0]}
    else:
        params = {'A':np.sin(p1[1])**2,'B':np.sin(p1[2])**2,'p':p1[0]}
    return params
'''