import numpy as np
from pyle import gateCompiler as gc  
from pyle.dataking import sweeps as sweeps
import labrad
from labrad.units import Unit, Value, Complex
import copy
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
  
class Parameters(object):
    def __init__(self, qubitAndParamNames, devs, measure, differential=True):
        if not isinstance(measure, list):
            measure = [measure]
        qNames = set([q for q,name in qubitAndParamNames])
        paramDict = {}
        for qName in qNames:
            parameters = [p for q,p in qubitAndParamNames if q == qName]
            tempDict = {}
            for parameter in parameters:
                tempDict[parameter] = devs[qName][parameter]
            paramDict[qName] = tempDict
        self.paramDict = paramDict
        self.devs = devs
        indexDict = {}
        for idx,meas in enumerate(measure):
            indexDict[meas] = idx
        self.indexDict = indexDict
        self.differential = differential
    
    def makeInputsAxesDeps(self, nelderMead = True, sweepRange=None):
        if nelderMead:
            axes = [('Function Evaluation','')]
            deps = [('Sequence Error', '', ''),('Sequence Error STD', '', '')]
        else:
            axes = []
            deps = []
        inputs = []
        for qName in np.sort(self.paramDict.keys()): # this would be way better with an ordered dict in python 2.7
            for keyName in np.sort(self.paramDict[qName].keys()):
                values = self.paramDict[qName][keyName]
                if not isinstance(values, list):
                    if self.differential:
                        inputs.append(0)
                    else:
                        inputs.append(value)
                    if nelderMead:
                        deps.append(( qName + '_'+str(keyName) , '' , '' ))
                    else:
                        axes.append((sweepRange, qName + '_'+str(keyName)))
                else:
                    for idx,value in enumerate(values):
                        if self.differential:
                            inputs.append(0)
                        else:
                            inputs.append(value)
                        if nelderMead:
                            deps.append(( qName + '_'+str(keyName)+'_%d' % idx , '' , '' ))
                        else:
                            axes.append((sweepRange, qName + '_'+str(keyName)+'_%d' % idx))
        return axes, deps, inputs
        
    def argsToCurrParamDict(self, args):
        args = list(args)
        currParamDict = copy.deepcopy(self.paramDict)
        for qName in np.sort(currParamDict.keys()): # this would be way better with an ordered dict in python 2.7
            for keyName in np.sort(currParamDict[qName].keys()):
                values = currParamDict[qName][keyName]
                if isinstance(values, list) and isinstance(values[0],Value):
                    currUnit = values[0].unit #I'm so sorry
                elif isinstance(values,Value):
                    currUnit = values.unit
                else:
                    currUnit = 1
                if self.differential:
                    delta_values = [args.pop(0) for k in range(np.alen(values))]
                    if np.alen(values)>1:
                        shifted_values = list(np.array(delta_values) + np.array(values))
                        currParamDict[qName][keyName] = [sval*currUnit for sval in shifted_values]
                    else:
                        currParamDict[qName][keyName] = delta_values[0]*currUnit + values
                else:
                    raise NotImplementedError
        return currParamDict
        
    def updateQubits(self, Qubits, currParamDict, noisy=False):
        for qName in currParamDict.keys():
            idx = self.indexDict[qName]
            Qubit = Qubits[idx]
            for param in currParamDict[qName].keys():
                if noisy:
                    print "%s['%s']: %s -> %s" % (qName, param, self.devs[qName][param], currParamDict[qName][param])
                Qubit[param] = currParamDict[qName][param]

def addQubitNamesToParamNames(paramNames, qubit):
    qubitAndParamNames = [[qubit.__name__, param] for param in paramNames]
    return qubitAndParamNames

def updateNM(Sample, dvwData, iteration = 'min'):

    if iteration == None:
        optimalParams = list(dvwData[-1,3:])
    elif iteration == 'min':
        bestIdx = np.argmin(dvwData[:,1])
        optimalParams = list(dvwData[bestIdx,3:])
        print 'best IDX is: %s' % (bestIdx+1)
    elif iteration == 'max':
        bestIdx = np.argmax(dvwData[:,1])
        optimalParams = list(dvwData[bestIdx,3:])
        print 'best IDX is: %s' % (bestIdx+1)
    else:
        optimalParams = list(dvwData[iteration-1,3:])
        
    measure = dvwData.parameters['measure']
    qubitAndParamNames=dvwData.parameters['qubitAndParamNames']
    _, devs, _, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    param = Parameters(qubitAndParamNames, devs, measure)
    currParamDict = param.argsToCurrParamDict(optimalParams)
    param.updateQubits(Qubits, currParamDict, noisy=True)
        
def makeFuncWrapper(averages, func, axes, deps, measure, sample, noisy=True):
        
    def funcWrapper(args):
        args=list(args)
        def seqGen():
            for currAvg in range(averages):
                yield args + [currAvg]
        mySeqGen = seqGen()
        sub_dataset = sweeps.prepDataset(sample, 'dummy dataset', axes, deps, measure=measure)
        data = sweeps.run(func, mySeqGen, dataset = sub_dataset, save=False, noisy=False)
        avgError = np.mean(data)
        std = np.std(data)/np.sqrt(averages)
        if noisy:
            print 'avg error: %s, std: %s' % (avgError, std)
        return avgError, std
    
    return funcWrapper
                