import numpy as np
import time
import itertools

from labrad.units import Value, Unit

dBm, ns, us, GHz = (Unit(ss) for ss in ['dBm', 'ns', 'us', 'GHz'])

import scipy.optimize
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.envelopes as env
from pyle.dataking.fpgaseqTransmon import runQubits
import pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.util as util
import pyle.dataking.sweeps as sweeps
import pyle.dataking.singleQubit as sq
import pyle.dataking.multipleQubit as mq
import pyle.pipeline as pipeline
from pyle.pipeline import FutureList, returnValue
import pyle.util.sweeptools as st
import pyle.analysis.iq as iq
import pyle.fitting.fitting as fitting
import pyle.dataking.adjust as adjust

import pyle.quantum as quantum
import pyle.tomo as tomo

#Crap that shouldn't be here
import pyle.analysis.iq as iq
import pyle.analysis.readout as roa
import matplotlib.pyplot as plt

def parampTuneFlux(Sample, measure, flux, freq,
                   save=True, measureLen=None):
    """Scan paramp flux bias to find best snr improvement
    
    freq - Value[Hz]: The frequency of the tone you want to use for probing
           the paramp response. Probably should not be a qubit readout
           frequency because that would be in or near a resonance dip.
    measureLen - Value[ns]: Length of probe tone. Defaults to qubit readout
                 length, however you probably want to use a longer tone.
    """
    sampleReg, devs = util.loadGateCompilerDevices(Sample, measure,
                                                       write_access=False)
    q = devs[measure]
    if measureLen is not None:
        q['readoutLen'] = measureLen
        q['adcFilterWindow'] = [q['adcFilterWindow'][0],
                                q['adcFilterWindow'][0]+measureLen]
    data = sq.readoutIq(Sample, measure, freq=freq, dataFormat='iqRaw',
                        name='Paramp bias flux',
                        parampFlux=flux, readoutLen=measureLen,
                        save=save, collect=True, plot=False,
                        regParams=(sampleReg, devs))
    signal = []
    noise = []
    fluxOut = []
    fluxUnit = data.indep[0][1]
    for f in flux:
        idx = data[:,0] == f[fluxUnit]
        z = data[idx,1] + 1.0j*data[idx,2]
        signal.append(np.mean(z))
        noise.append(np.std(z))
        fluxOut.append(f)
    snrOff = (signal[10]/noise[10])**2
    snr = (np.abs(signal)/noise)**2 / snrOff
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(fluxOut, 10*np.log10(snr))
    
def calculateReadoutCenters(Sample, measure, 
                            stats=900, states=None,
                            power=None, adcFilterWindow=None,
                            save=False, update=True, plot=False, noisy=True):
    """
    Find centers of IQ clusters for various qubit states
    """
    if states is None:
        states=[0,1]
    if update:
        sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
        Q = Qubits[measure]
    else:
        sample, qubits = util.loadQubits(Sample, write_access=False)
    
    name = 'iqCenters - '+' '.join(['|%d>'%state for state in states])
    dataset = sq.lollipop(Sample, measure, name=name,
                          power=power,
                          adcFilterWindow=adcFilterWindow,
                          states=states, stats=stats, dataFormat='iqRaw', 
                          plot=False, save=save)
    S = [dataset[:,2*i+1]+1.0j*dataset[:,2*i+2] for i in range(len(states))]
    centers = [iq.iqCenter(z) for z in S]
        
    colors = ['b','r','g']
    fig = plt.figure()
    ax = fig.add_subplot(111)
    for z,c,color in zip(S, centers, colors):
        ax.plot(z.real, z.imag, color+'.', alpha=0.8)
        ax.plot(c.real, c.imag, 'k.', markersize=20)
    Q['readoutCenters'] = centers
        
def iqVsTime(Sample, measure, name='IQ Trace',
             repetitions=10, states=None, readoutLengths=None,
             herald=True, heraldLengths=None, heraldPowers=None,
             ringdownDelay=None,
             tMax=17000*ns, tMin=0*ns, averages=1, stats=1,
             save=True, collect=False, noisy=True, regParams=None):
    """IQ time traces vs time on multiple qubits
    
    State heralding enabled with herald=True
    If This is set a measure pulse will be done on each qubit prior to state
    preparation and measurement. Because this heralding pulse is done before
    the state preparation you should only process data in which the result of
    the heralding pulse yields a ground state for ALL qubits.
    """
    print("You must invoke an early start time for the ADC for this to work")
    print("You must invoke an early start time for the ADC for this to work")
    print("You must invoke an early start time for the ADC for this to work")
    time.sleep(0.5)
    #Load data from registry or from passed in argument

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

    #Default settings
    if ringdownDelay is None:
        ringdownDelay = 500*ns
    if readoutLengths is not None:
        for q, rl in zip(qubits, readoutLengths):
            q['readoutLen'] = rl
    if heraldLengths is None:
        heraldLengths = [q['readoutLen'] for q in qubits]
    if heraldPowers is None:
        heraldPowers = [q['readoutPower'] for q in qubits]
    if states is None:
        states = ['{0:0{1}b}'.format(n, len(qubits)) for n in range(2**len(qubits))]
    if repetitions > 1:
        repetitions = np.arange(repetitions)
    #Assume we measure |0> and |1> for each qubit
    assert len(states) == 2**len(qubits)        
    #Pulse sequence constants
    piLenMax = max([q['piLen'] for q in qubits])
    heraldLenMax = max(heraldLengths)
    #Set up dataset
    axes = [(repetitions,'Repetition')]
    deps = []
    for state in states:
        #qubitName = Sample['config'][i]
        deps.append(('Amplitude', 'I |%s>'%(state,), ''))
        deps.append(('Amplitude', 'Q |%s>'%(state,), ''))
        
    kw = {'stats':stats, 'averages':averages}
    if herald:
        kw['heraldLengths'] = heraldLengths
        kw['ringdownDelay'] = ringdownDelay
        kw['heraldPowers'] = heraldPowers
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,
                                 explicitIndependents = [('Time','ns')],
                                 kw=kw)
    timeAxis = np.arange(8192)*2
    tMask = np.logical_and(timeAxis <= tMax['ns'], timeAxis >= tMin['ns'])
    #Define scan function
    def func(server, repetition):
        for q in qubits:
            q['adc mode'] = 'average'
        reqs=[]
        for state in states:
            alg = gc.Algorithm(agents=devs)
            #State herald
            #Excite qubits appropriate for this multi-qubit state.
            #For two qubits A and B, if you Kronecker product their basis
            #states you get |0A 0B>, |0A 1B>, |1A 0B>, |1A 1B> in that
            #order. Therefore, an index into the state string, eg '10',
            #is also the correct index into the ordered list of qubits.
            for idx, q in enumerate(alg.qubits):
                thisQubitState = int(state[idx])
                if thisQubitState == 1:
                    alg[gates.PiPulse([q])]
                elif thisQubitState == 0:
                    pass #Don't need to do anything, not even wait.
                else:
                    raise RuntimeError('|%d> not supported yet'%thisQubitState)
            #Measure all qubit states
            alg[gates.Measure(alg.qubits)]
            #If we want state heralding put a measure pulse in the prepad
            if herald:
                for q, hl, hp in zip(alg.qubits, heraldLengths, heraldPowers):
                    alg[gates.Measure([q],
                                delay=-piLenMax-ringdownDelay-heraldLenMax)]
            alg.compile()
            # This is a horrible hack to keep from returning redundant data
            for q in alg.qubits[1:]:
                print "Disabling readout for qubit %s" % (q.__name__)
                q['readout'] = False
            if len(alg.qubits) > 1:
                print "alg.q0 channels: %s" % (alg.q0['readoutDevice']['channels'])
                alg.q0['readoutDevice']['channels'] = alg.q0['readoutDevice']['channels'][0:-(len(alg.qubits)-1)]
                print "alg.q0 channels: %s" % (alg.q0['readoutDevice']['channels'])
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        data = yield FutureList(reqs)
        #Data is a list, one element per state
        #Each element has one entry per qubit
        data = np.array(data)
        #+0.0 is to cast to float. Not sure why but this is needed to
        #make data vault happy
        data = np.reshape(data, (2*len(states), -1)) + 0.0
        data = np.vstack((timeAxis[tMask], data[:,tMask])).T
        returnValue(data)
    data = sweeps.grid(func, axes, dataset=dataset, save=save, collect=collect,
                       noisy=False, pipesize=2)
    if collect:
        return data
    
def readoutDephasing(Sample, measure, name='Readout Dephasing',
                     powers=st.r[-60:-30:2, dBm], excitationLen=None,
                     excitationFreq=None, df=None, decayLen = 400*ns, 
                     phases=np.linspace(0, 4*np.pi, 40, endpoint=False),
                     stats=900, save=True, mode = 'both'):
    """
    Measure the qubit dephasing due to readout resonator photons
    
    Sequence is pi/2 pulse, excite readout resonator, allow resonator to ring
    down, pi/2 pulse, measure. This is essentially a Ramsey fringe with a
    readout resonator pulse inserted between the two pi/2 pulses.
    
    excitationLen - Value[ns] : Length of pulse that puts photons in readout
                    resonator, thus dephasing the qubit. Defaults to
                    q['readoutLen'].
    excitationFreq - Value[ns]: Frequency of resontaor excitation pulse.
                     Default to q['readoutFrequency'].
    df - Value[MHz]: Range of frequencies over which to scan excitationFreq.
         If None excitationFreq is not scanned.
    decayLen - Value[ns]: Extra time waited after the resonator excitation
               pulse before final pi/2 pulse
    """
    sample, devs, Devs = util.loadGateCompilerDevices(Sample, measure, write_access=True)
    q = devs[measure]
        
    if excitationLen is None:
        excitationLen = q['readoutLen']
    if excitationFreq is None:
        excitationFreq = q['readoutFrequency']
    if df is not None:
        excitationFreq = st.r[excitationFreq['GHz']-df['GHz']:excitationFreq['GHz']+df['GHz']:df['GHz']/20, GHz]
    
    axes = [(powers, 'Power'), (excitationFreq, 'Excitation Frequency'),
            (phases, 'Ramsey Phase')]
    deps = [('Probability', 'P', '')]
    kw = {'stats': stats, 'decayLen': decayLen, 'excitationLen': excitationLen,
          'mode': mode}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,
                                 kw={'stats':stats})
    if mode not in ('first', 'second', 'both'):
        raise RuntimeError('Invalid Mode')
        
    def func(server, power, ex_freq, phase):
        alg = gc.Algorithm(agents=devs)
        if mode in ('first', 'both'):
            alg[gates.PiHalfPulse([q])]
        else:
            alg[gates.Wait([q], q['piLen'])]
        alg[gates.Wait([q], excitationLen+decayLen)]
        if mode in ('second', 'both'):
            alg[gates.PiHalfPulse([q], phase=phase)]
        else:
            alg[gates.PiHalfPulse([q], phase=phase, amp=0)]
        
        alg[gates.Measure([q])]
        alg.compile()
        amp = eh.power2amp(power)
        sb_freq = ex_freq - q['readoutDevice']['carrierFrequency'] #sideband freq for readout
        q['rr'] += env.mix(env.flattop(-excitationLen-q['piHalfLen']-decayLen,
                           excitationLen, 20, amp=amp), sb_freq)
        data = yield runQubits(server, alg.agents, stats=stats, dataFormat='iqRaw')        
        Is, Qs = data[0]
        prob = sq.readoutProbabilities(q, Is.asarray+1.0j*Qs.asarray)
        pipeline.returnValue([np.mean(prob)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)

def readout_histogram_herald(Sample, measure, name='Readout Heralded', stats=100, ringdown_time=1000*ns, save=True, plot=True):
    sample, devs, q = gc.loadQubits(Sample, measure, write_access=False)
    
    state = [0,1]
    axes = [(state, 'State')]
    deps = [('Amplitude', 'I', 'clicks'), ('Amplitude', 'Q', 'clicks')]
    kw = {'stats':stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, state):
        alg = gc.Algorithm(agents=devs)
        alg[gates.MoveToState([alg.q], 0, state)]
        alg[gates.Measure([alg.q])]
        alg.compile()
        add_herald(alg, ringdown_time=ringdown_time)
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', localDemod=True)
        data = do_post_select(data, alg)
        pipeline.returnValue(np.column_stack(data[0]))
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    zero_mask = data['State'] == 0
    one_mask = data['State'] == 1
    SNoPulse = data['I'][zero_mask] + 1j*data['Q'][zero_mask]
    SPulse = data['I'][one_mask] + 1j*data['Q'][one_mask]
    center, axis, (mask0,mask1), (ax1,ax2) = iq.discriminationAxis(SNoPulse, SPulse, plot=plot)
    err_0 = np.sum(np.real((SNoPulse - center) / axis) > 0.0)*1.0 / len(SNoPulse)
    err_1 =np.sum(np.real((SPulse - center) / axis) < 0.0)*1.0 / len(SPulse)
    print "|0> State error: %f, |1> state error %f"  % (err_0, err_1)
    return data
    
def readoutCrossTalkSignal(Sample, measure, name='Readout Crosstalk Phase', stats=900, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    phase = np.linspace(0, 2*np.pi, 20, endpoint=False)
    axes=[(phase, 'Phase')]
    kw = {'stats': stats}
    deps = [('Amplitude', 'I', 'clicks'), ('Amplitude', 'Q', 'clicks')]
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, phase):
        alg = gc.Algorithm(agents=devs)
        alg.q[1]['readoutPhaseShift'] = phase
        alg[gates.Measure([alg.q[1]])]
        alg.compile()
        result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is, Qs = result[0]
        pipeline.returnValue([np.mean(Is), np.mean(Qs)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)

def readoutCrossTalk(Sample, measure, name='Readout Crosstalk', stats=900, herald=False, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    nqubits = len(q)
    states = range(2**nqubits)
    axes = [(states, 'State')]
    deps = []
    kw = {'stats':stats}
    for idx in range(2**nqubits):
        deps.append(('Counts', 'State %d' % idx, ''))
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, state):
        alg = gc.Algorithm(agents=devs)
        print "state %d" % state
        if isinstance(alg.q, dict):
            alg.q = [alg.q]
        for idx, qubit in enumerate(reversed(alg.q)):
            if (state>>idx)&1:
                print "Pi pulse on qubit %s" % qubit.__name__
                alg[gates.PiPulse([qubit])]
        alg[gates.Measure(alg.q)]
        alg.compile()
        if herald:
            add_herald(alg, 1000*ns)
            localDemod=True
        else:
            localDemod=False
        result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', localDemod=localDemod)
        if herald:
            result = do_post_select(result, alg)
        states=0
        for idx, qubit in enumerate(alg.q):
            Is,Qs = result[idx]
            prob = sq.readoutProbabilities(qubit, np.asarray(Is)+1.0j*np.asarray(Qs))
            states += prob*2**(nqubits-idx-1)
        rv = np.array([np.sum(states==x) for x in range(2**nqubits)]) * 1.0
        pipeline.returnValue(rv)
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    return data
    
def add_herald(alg, ringdown_time=500*ns):
    if not alg.compiled:
        raise RuntimeError('Cannot add post selection pulse to sequence until it is compiled')
    start=np.inf
    end=-np.inf
    for dev in alg.agents:
        if 'xy' in dev:
            start = min(dev['xy'].start, start)
            end = max(dev['xy'].end, end)
        if 'z' in dev:
            start = min(dev['xy'].start, start)
            end = max(dev['xy'].end, end)
    main_sequence_length = end-start
    min_prepad = 0
    for dev in alg.agents:
        if dev.get('readout',False):
            start_time = (dev['readoutLen'] + main_sequence_length + ringdown_time)
            dev.measure(delay=-start_time)
            min_prepad = min(min_prepad, start_time)
    if (200*ns + start_time) > fpgaseqTransmon.PREPAD:
        raise RuntimeError("Early start for readout doesn't leave enough prepad")

def do_post_select(data, alg):
    data_out = []
    mask = 1
    qubits = [q for q in alg.agents if q.get('readout', False)]
    for idx, qubit in enumerate(qubits):
        I1, Q1, I2, Q2 = data[idx]
        z1 = np.asarray(I2) + 1j*np.asarray(Q2)
        x = np.real((z1 - qubit['readoutCenterZ']) / qubit['readoutAxis'])
        mask = np.logical_and(mask, x<0)
    print "Post selection success rate: %f" % (np.mean(mask))
    for idx, qubit in enumerate(qubits):
        I1, Q1, I2, Q2 = data[idx]
        x = np.vstack((I1[mask], Q1[mask]))
        data_out.append(x)
    data_out = np.asarray(data_out)
    return data_out

global best_so_far
best_so_far = [0.0, [], []]
def optimize_bell_state(Sample, measure):
    global best_so_far
    best_so_far = [0.0, [], []]

    initial = ( Sample['r12']['q1sqrtSwapTime']['ns'],
                Sample['r12']['q1sqrtSwapAmp'],
                Sample['r12']['q2swapTime']['ns'],
                Sample['r12']['q2swapAmp'],
                Sample['q1']['settlingRates'][0],
                Sample['q1']['settlingAmplitudes'][0],
                Sample['q2']['settlingRates'][0],
                Sample['q2']['settlingAmplitudes'][0] )
    
    idx = [0]
    def func(parameters):
        global best_so_far
        Sample['r12']['q1sqrtSwapTime'] = parameters[0]* ns
        Sample['r12']['q1sqrtSwapAmp'] = parameters[1]
        Sample['r12']['q2swapTime'] = parameters[2] * ns
        Sample['r12']['q2swapAmp'] = parameters[3]
        Sample['q1']['settlingRates'] = [parameters[4]]
        Sample['q1']['settlingAmplitudes'] = [parameters[5]]
        Sample['q2']['settlingRates'] = [parameters[6]]
        Sample['q2']['settlingAmplitudes'] = [parameters[7]]
        rho, fval = swaping_bell_state(Sample, measure, stats=600, save=False)
        fef = quantum.fully_entangled_fraction(rho)
        print "iteration %d, result %f, current record: %f" % (idx[0], fef, best_so_far[0])
        if fef > best_so_far[0]:
            print "New best result"
            best_so_far[0] = fef
            best_so_far[1] = parameters
            best_so_far[2] = rho
        idx[0] = idx[0]+1 
        return 1-fef
    try:    
        result = scipy.optimize.fmin(func, initial)
    finally:
        Sample['r12']['q1sqrtSwapTime'] = initial[0]* ns
        Sample['r12']['q1sqrtSwapAmp'] = initial[1]
        Sample['r12']['q2swapTime'] = initial[2] * ns
        Sample['r12']['q2swapAmp'] = initial[3]
        Sample['q1']['settlingRates'] = [initial[4]]
        Sample['q1']['settlingAmplitudes'] = [initial[5]]
        Sample['q2']['settlingRates'] = [initial[6]]
        Sample['q2']['settlingAmplitudes'] = [initial[7]]        
    return result

def get_nested(dictionary, key):
    tmp = dictionary
    keyname, _, keyidx = key.partition(':')
    for k in keyname.split('.'):
        tmp = tmp[k]
    if keyidx:
        return tmp[int(keyidx)]
    else:
        return tmp

def set_nested(dictionary, key, value):
    tmp = dictionary
    keypath, _, keyname = key.rpartition('.')
    if keypath:
        for k in keypath.split('.'):
            tmp = tmp[k]
    keyname, _, keyidx = keyname.partition(':')
    if keyidx:
        tmp[keyname][int(keyidx)] = value
    else:
        tmp[keyname] = value
    
def optimize_bell_parameter(Sample, measure, parameter_name, param_range_mult=0.5, stats=300, npts=11, update=False, plot=False, do_tomo=True, herald=False):

    x0 = get_nested(Sample, parameter_name)
    print "x0: %s" % (x0,)
    if isinstance(x0, Value) and x0.unit is not None:
        xunit = x0.unit
        xval = x0.value
    else:
        xunit = 1.0
        xval = x0
    if param_range_mult > 0:
        values = np.linspace(xval*(1-param_range_mult), xval*(1.0+param_range_mult), npts)
    else:
        values = np.linspace(xval*(+param_range_mult), xval*(-param_range_mult), npts)
    result = 0.0*values
    data = []
    try:
        for idx, value in enumerate(values):
            set_nested(Sample, parameter_name, value*xunit)
            print "Next value: %s" % (get_nested(Sample, parameter_name),)
            if do_tomo:
                rho, fval = swaping_bell_state(Sample, measure, save=False, stats=stats, correctXtalkZ=True)
                fef = quantum.fully_entangled_fraction(rho)
                data.append(rho)
                print "bell fidelity: %f" % (fef,)
                result[idx] = fef
            else:
                probs = swaping_bell_state(Sample, measure, save=False,correctXtalkZ=True, stats=stats, do_tomo=False, herald = herald)
                data.append(probs)
                result[idx] = probs[3]
    finally:
        set_nested(Sample, parameter_name, x0)

    p = np.polyfit(values, result, 2)
    extremum =  -p[1]/(2*p[0])

    if update:
        adjuster=adjust.Adjust()
        adjuster.plot(values, result)
        adjuster.x_param('parameter_value', extremum)
        new_value = adjuster.run()
        if new_value:
            set_nested(Sample, parameter_name, new_value['parameter_value'])
    elif plot:
        plt.figure()
        plt.plot(values, result, '.')
        plot_range = np.linspace(np.min(values),np.max(values), 50)
        plt.plot(plot_range, np.polyval(p, plot_range), '-')
        print "extremum: %f" % (extremum,)
    return values, result, data

def time_delay_swap(Sample, measure, name='Time delay swap', delays=None, stats=300, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    if delays is None:
        delays = st.r[-30:30:2, ns]
        
    axes = [(delays, 'Time Lag %s'%(measure[1],))]
    deps = [('P', 'Transfer Probability -> %s' % measure[1], ''), ('P', 'Transfer Probability -> %s' % measure[0], '')]
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,
                                 kw={'stats':stats})
                                 
    def func(server, delay):
        alg1 = gc.Algorithm(agents=devs)
        alg1.q1['timingLagWrtMaster'] = delay
        alg1[gates.PiPulse([alg1.q0])]
        alg1[gates.ResonatorSwap([alg1.q0,alg1.agents[4]])]
        alg1[gates.ResonatorSwap([alg1.q1,alg1.agents[4]])]
        alg1[gates.Wait(alg1.qubits, 25*ns)]
        alg1[gates.Measure(alg1.qubits)]
        alg1.compile()
        result = yield runQubits(server, alg1.agents, stats, dataFormat='iqRaw')
        Is, Qs = result[1]
        prob1 = sq.readoutProbabilities(alg1.q1, Is.asarray+1.0j*Qs.asarray)

        alg2 = gc.Algorithm(agents=devs)
        alg2.q1['timingLagWrtMaster'] = delay
        alg2[gates.PiPulse([alg2.q1])]
        alg2[gates.ResonatorSwap([alg2.q1,alg2.agents[4]])]
        alg2[gates.ResonatorSwap([alg2.q0,alg2.agents[4]])]
        alg2[gates.Wait(alg2.qubits, 25*ns)]
        alg2[gates.Measure(alg2.qubits)]
        alg2.compile()        
        result = yield runQubits(server, alg2.agents, stats, dataFormat='iqRaw')
        Is, Qs = result[0]
        prob2 = sq.readoutProbabilities(alg2.q0, Is.asarray+1.0j*Qs.asarray)
        
        pipeline.returnValue([np.mean(prob1), np.mean(prob2)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=10)

def stray_population(Sample, measure, name='Test bus population', reps=10, stats=300, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)

    axes = [(range(reps), 'Rep')]
    deps = [('P', 'Qubit population', ''), ('P', 'Bus population', '')]
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,
                                 kw={'stats':stats})
    def func(server, delay):
        alg = gc.Algorithm(agents=devs)
        alg[gates.Measure(alg.qubits)]
        alg.compile() 
        result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is, Qs = result[0]
        prob_q = sq.readoutProbabilities(alg.q0, Is.asarray+1.0j*Qs.asarray)
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.ResonatorSwap([alg.q0, alg.agents[4]])]
        alg[gates.Measure(alg.qubits)]
        alg.compile() 
        result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is, Qs = result[0]
        prob_b = sq.readoutProbabilities(alg.q0, Is.asarray+1.0j*Qs.asarray)
       
        pipeline.returnValue([np.mean(prob_q), np.mean(prob_b)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=10)
    return data

def half_swap_tuner(Sample, measure, name='Half-swap tuner', swaplen=st.r[5:10:.2, ns], delays = st.r[0:3:.1, ns], stats=300, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    axes = [(swaplen, 'Swap Length'), (delays, 'Delay')]
    deps = [('P', 'P one', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,
                                 kw={'stats':stats})
    def func(server, swaplen, delay):
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        alg[gates.ResonatorSqrtSwap([alg.q0, alg.agents[4]])]
        alg[gates.Wait([alg.q0], 10*ns+ delay)]
        alg[gates.ResonatorSqrtSwap([alg.q0, alg.agents[4]])]
        alg[gates.Wait([alg.q0], 10*ns+ delay)]
        alg[gates.ResonatorSqrtSwap([alg.q0, alg.agents[4]])]
        alg[gates.Wait([alg.q0], 10*ns+ delay)]
        alg[gates.ResonatorSqrtSwap([alg.q0, alg.agents[4]])]
        
        alg[gates.Measure(alg.qubits)]
        alg.compile()
        result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is, Qs = result[0]
        prob = sq.readoutProbabilities(alg.q0, Is.asarray+1.0j*Qs.asarray)
        pipeline.returnValue([np.mean(prob)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=10)
    return data

def swaptunerN(Sample, measure, name="Swap Tuner N", swaps=2, swaplen=None, swapamp=None, herald=False, stats=300, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)

    if swaplen is None:
        swaplen0 = devs[4]['%sswapTime'%q.__name__]
        swaplen = st.r[0.9*swaplen0['ns']:1.1*swaplen0['ns']:.01*swaplen0['ns'], ns]
    if swapamp is None:
        swapamp0 = devs[4]['%sswapAmp'%q.__name__]
        swapamp = np.linspace(0.97*swapamp0, 1.02*swapamp0, 21)
        
    axes = [(swaplen, 'Swap Length'), (swapamp, 'Swap Amplitude')]
    deps = [('Counts', 'P1', '')]
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,
                                 kw={'stats':stats})
                                 
    def func(server, l, a):
        min_prob = 1.0
        for delay in st.r[0:3:.3, ns]:
            alg = gc.Algorithm(agents=devs)
            alg.agents[4]['%sswapTime'%alg.q0.__name__] = l
            alg.agents[4]['%sswapAmp'%alg.q0.__name__] = a
            alg[gates.PiPulse([alg.q0])]
            for idx in range(swaps):
                alg[gates.ResonatorSwap([alg.q0, alg.agents[4]])]
                alg[gates.Wait(alg.agents, 10*ns + delay)]
            alg[gates.Measure(alg.qubits)]
            alg.compile()
    #        if herald:
    #            add_herald(alg, ringdown_time=500*ns)
            result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', localDemod=False)
    #        if herald:
    #            result = do_post_select(result, alg)
            Is, Qs = result[0]
            prob = np.mean(sq.readoutProbabilities( alg.q0,Is.asarray+1j*Qs.asarray))
            print "probability at delay %s: %s" % (delay, prob)
            min_prob = min(prob, min_prob)
        pipeline.returnValue([min_prob])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=10)
    return data

    
def swaping_bell_state(Sample, measure, name='Swap based bell state', delay=50*ns, xtalk_m = None, stats=300, herald=False, save=True, correctXtalkZ=False, do_tomo=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    if do_tomo:
        ops = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, 2))
        axes = [(range(len(ops)), 'Tomo op')]
    else:
        ops = [('I', 'I')]
        axes = [(1, 'Tomo op')]
    op_Us = [ np.kron(a[0], a[1]) for a in tomo.gen_qst_tomo_ops(tomo.octomo_ops, 2)]
    #op_names = ['.'.join(op) for op in ops]
    axes = [(range(len(ops)), 'Tomo op')]
    deps = [('Counts', '{0:0{1}b}'.format(n, len(q)), '') for n in range(2**len(q))]
    kw = {'stats': stats}
    states = ['{0:0{1}b}'.format(n, len(q)) for n in range(2**len(q))]
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,
                                 kw={'stats':stats})
    def func(server, operation):
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        alg[gates.ResonatorSwap([alg.q0, alg.agents[4]])]
        alg[gates.Wait(alg.agents, 10*ns)]
        alg[gates.ResonatorSqrtSwap([alg.q1,alg.agents[4]])]
        alg[gates.Wait(alg.agents, 10*ns)]
        alg[gates.ResonatorSwap([alg.q0, alg.agents[4]])]
        alg[gates.Wait(alg.agents, 10*ns)]
        #alg[gates.PiHalfPulse([alg.q0])]
        #alg[gates.PiPulse([alg.q1])]
        alg[gates.Tomography(alg.qubits, ops[operation])]
        alg[gates.Wait(alg.agents, 10*ns)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        if herald:
            print "Operation: %s / %s" % (operation, ops[operation])
            add_herald(alg, ringdown_time=500*ns)
        result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', localDemod=bool(herald))
        if herald:
            result = do_post_select(result, alg)
            
        prob = mq.mqReadoutProbabilities(result, alg.qubits, measure, correlated=True)
        pipeline.returnValue(prob)
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=1 if herald else 10)
    if do_tomo:
        try:
            rho, rho_norm, fval, resid = tomo.qst_mle_evan(data[:,1:], op_Us, counts=stats, F=xtalk_m)
            print "tomo.qst_mle_evan success, fval: %s" % (fval,)
        except Exception:
            rho = tomo.qst_mle(data[:,1:], op_Us, F=xtalk_m)
            print "tomo.qst_mle_evan failed, using original"
            fval = np.inf
        gbf = quantum.fully_entangled_fraction(rho)
        t = quantum.tangle(rho)
        #print "rho entropy %f" % (quantum.entropy(rho),)
        #print "rho tangle: %f" % (t,)
        #print "gen. bell f: %f" % (gbf,)
        return rho, fval
    else:
        error_value = np.sqrt(np.sum((data[0,1:] - np.array([0.0,0.5,0.5,0.0]))**2)/4.0)
        print "data: ", data[0,1:]
        print "error_value: ", error_value
        return data[0,1:]

def print_tomo_results(ds):
    ops = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, 2))
    op_Us = [ np.kron(a[0], a[1]) for a in tomo.gen_qst_tomo_ops(tomo.octomo_ops, 2)]
    for idx, op in enumerate(ops):
        print op, ds[idx,1:]*ds.parameters['stats']
    