# Control envelopes in time domain and frequency domain
#
# For a pair of functions g(t) <-> h(f) we use the following
# convention for the Fourier Transform:
#
#          / +inf
#         |
# h(f) =  | g(t) * exp(-j*2pi*f*t) dt
#         |
#        / -inf
#
#          / +inf
#         |
# g(t) =  | h(f) * exp(2j*pi*f*t) df
#         |
#        / -inf
#
# Note that we are working with frequency in GHz, rather than
# angular frequency.  Also note that the sign convention is opposite
# to what is normally taken in physics.  But this is the convention
# used here and in the DAC deconvolution code, so you should use it.

# Also, this convention is better :)
#
# To get Mathematica to use this convention, add the option FourierParameters -> {0, -2*Pi}
#
# The physics convention comes from a very unfortunate choice of sign in
# Heisenberg/Schrodinger's equation, or Schrodinger picked the bad sign
# because of unfortunate pre-existing physics conventions, not sure. - DTS

from __future__ import absolute_import # have to do this so we get math std library

import math
import numpy as np
import scipy
import types
import traceback

import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad, au = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad', 'au')]

from scipy.special import erf
from scipy.interpolate import interp1d
import scipy.interpolate as interpolate
from scipy.integrate import quad

import matplotlib.pyplot as plt

from pyle.util import convertUnits
import pyle.util.cache
from pyle.signalprocessing import interp1d_cubic
import pyle.dataking.zfuncs as zfuncs
enable_caching = True
cache_hits = 0
cache_misses = 0


#exp = pyle.util.cache.LRU_Cache(np.exp, pyle.util.cache.keyfunc_ndarray, 1024)
exp = np.exp # To disable caching for exponentials

class Envelope(object):
    """Represents a control envelope as a function of time or frequency.
    
    Envelopes can be added to each other or multiplied by constant values.
    Multiplication of two envelopes and addition of a constant value (other
    than zero) are not equivalent in time and fourier domains, so these
    operations are not supported.
    
    Envelopes keep track of their start and end time, and when added
    together the new envelope will use the earliest start and latest end,
    to cover the entire range of its constituent parts.
    
    Envelopes can be evaluated as functions of time or frequency using the
    fourier flag.  By default, they are evaluated as a function of time.
    """
    @convertUnits(start='ns', end='ns')
    def __init__(self, start=None, end=None):
        self.start = start
        self.end = end

    def timeFunc(self, t):
        raise NotImplementedError('Envelope timeFunc must be overridden')

    def freqFunc(self, f):
        raise NotImplementedError('Envelope freqFunc must be overridden')
    
    def duration(self):
        return self.end - self.start
    def __call__(self, x, fourier=False):
        if fourier:
            return self.freqFunc(x)
        else:
            return self.timeFunc(x)

    def __add__(self, other):
        if isinstance(other, Envelope):
            return EnvSum(self, other)
        else:
            # if we try to add envelopes with the built in sum() function,
            # the first envelope is added to 0 before adding the rest.  To support
            # this, we add a special case here since adding 0 in time or fourier
            # is equivalent
            if other == 0:
                return self
            return NotImplemented
    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, Envelope):
            return EnvSum(self, -other)
        else:
            # if we try to add envelopes with the built in sum() function,
            # the first envelope is added to 0 before adding the rest.  To support
            # this, we add a special case here since adding 0 in time or fourier
            # is equivalent
            if other == 0:
                return self
            return NotImplemented
        
    def __rsub__(self, other):
        if isinstance(other, Envelope):
            return EnvSum(-self, other)
        else:
            # if we try to add envelopes with the built in sum() function,
            # the first envelope is added to 0 before adding the rest.  To support
            # this, we add a special case here since adding 0 in time or fourier
            # is equivalent
            if other == 0:
                return -self
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, Envelope):
            return NotImplemented
        else:
            return EnvProd(self, other)
    __rmul__ = __mul__

    def __div__(self, other):
        if isinstance(other, Envelope):
            return NotImplemented
        else:
            return EnvProd(self, 1.0/other)
        
    def __rdiv__(self, other):
        if isinstance(other, Envelope):
            return NotImplemented
        else:
            return EnvProd(EnvInv(self), other)

    def __neg__(self):
        return -1 * self
    
    def __pos__(self):
        return self

class DynamicEnvelope(Envelope):
    '''
    This replaces the old Envelope class which took the time/freq functions
    as parameters, usually referring to a nested function/closure.

    This makes debuging, profiling, and interpeting stack traces hard, so it is
    deprecated.
    '''
    def __init__(self, timeFunc, freqFunc, start=None, end=None):
        self.__freqFunc = freqFunc
        self.__timeFunc = timeFunc
        Envelope.__init__(start, end)
    def freqFunc(self, f):
        return self.__freqFunc(f)
    def timeFunc(self, t):
        return self.__timeFunc(t)

def memoize_envelope(cls):
    '''
    Decorator for memoizing envelope classes.
    If all of the parameters to __init__ are the same, and the function
    is evaulated at the same frequency points, return the same data.
    
    The array is set to be read-only.  This prevents accidental
    corruption of the cache.  We could also just return a copy
    if that turns out to be a problem.
    '''
    if not enable_caching:
        return cls
    old_init = cls.__init__
    old_freqFunc = cls.freqFunc
    
    def __init__(self, *args, **kwargs):
        # NOTE: Caching now defaults to True
        cache = kwargs.pop('cache', True)
        if cache:
            key = (args, tuple(kwargs.items()))
            self._instance_key = key
        old_init(self, *args, **kwargs)
    def freqFunc(self, f):
        if not enable_caching or not hasattr(self, '_instance_key'):  # Allow global disabling of caching
            return old_freqFunc(self, f)
        data_key = (f[0], f[-1], len(f))
        key = (self._instance_key, data_key)
        try:
            x = self._cache[key]
            global cache_hits
            cache_hits += 1
            return x
        except KeyError:
            global cache_misses
            cache_misses += 1
            x = old_freqFunc(self, f)
            x.setflags(write=False)
            self._cache[key] = x
            return x
        except TypeError:
            print("Warning: unable to hash parameters for envelope of type %s" % (cls.__name__,))
            return old_freqFunc(self, f)
    __init__.__doc__ = old_init.__doc__
    freqFunc.__doc__ = (old_freqFunc.__doc__ or "Envelope Frequency Function") + '\n\nCaching added by decorator memoize_envelope'
    cls._cache={}
    cls.freqFunc = freqFunc
    if not hasattr(cls, '_memoize_manual_key'):
        cls.__init__ = __init__
    return cls
    
class EnvSum(Envelope):
    '''
    Helper class to support __add__ and __sub__
    '''
    def __init__(self, a, b):
        self.a = a
        self.b = b
        start, end = timeRange((self.a, self.b))
        Envelope.__init__(self, start, end)
    def timeFunc(self, t):
        return self.a.timeFunc(t) + self.b.timeFunc(t)
    def freqFunc(self, f):
        return self.a.freqFunc(f) + self.b.freqFunc(f)

class EnvProd(Envelope):
    '''
    Helper class to support __mul__, __div__, and __neg__.  Represents multiplication by a scalar
    '''
    def __init__(self, envIn, const):
        self.envIn = envIn
        self.const = const
        Envelope.__init__(self, envIn.start, envIn.end)
    def timeFunc(self, t):
        return self.envIn.timeFunc(t) * self.const
    def freqFunc(self, f):
        return self.envIn.freqFunc(f) * self.const

class EnvInv(Envelope):
    '''
    Helper class to support division of a scalar by an
    envelope. (__rdiv__).  I don't know why this is helpful -- ERJ
    '''
    def __init__(self, envIn):
        self.envIn = envIn
        Envelope.__init__(self, envIn.start, envIn.end)
    def timeFunc(self, t):
        return 1.0/self.envIn.timeFunc(t)
    def freqFunc(self, f):
        return 1.0/self.envIn.freqFunc(f)

class EnvConvFOnly(Envelope):
    '''
    Helper class that convolves an envelope with a smoothing function
    **in the frequency domain only**.  Convolution is expensive in the
    time domain, and hard to get right.  so we just pass that data
    through directly -- it is only used for plotting anyway.
    '''
    def __init__(self, env_in, env_filter):
        self.env_in = env_in
        self.env_filter = env_filter
    def timeFunc(self, t):
        return self.env_in.timeFunc(t)
    def freqFunc(self, f):
        return self.env_in.freqFunc(f) * self.env_filter.freqFunc(f)

class EnvZero(Envelope):
    '''
    Zero envelope.  Duh.
    '''
    def timeFunc(self, t):
        return 0*t
    def freqFunc(self, f):
        return 0*f

def envelope_factory_t0(cls):
    '''
    Use only on envelopes that take a t0 as their first argument.
    This will construct envelopes at time zero and shift them, allowing
    for better caching.
    '''
    # I would like to rewrite this in a way that it modifies the class rather than
    # wrapping it in a factory, but I didn't see a clean way to do that -- ERJ
    @convertUnits(t0='ns')
    def factory(t0, *args, **kwargs):
        if not enable_caching or t0==0.0:   # If caching is disabled, this doesn't help, so keep it simple
            return cls(t0, *args, **kwargs) # Also, avoid extra shift if t0 is already zero.
        x = cls(0.0, *args, **kwargs)
        y = shift(x, t0)
        return y
    factory.__doc__ = 'Envelope factory for type %s.  Access original class via __klass__ attribute' % (cls.__name__,)
    factory.__klass__ = cls
    return factory

class NumericalPulse(Envelope):
    '''
    Base class for numerical pulses.  In a subclass, provide a
    timeFunc method.  freqFunc will interpolate and FFT to get the
    frequency space representation, and convolve with a Gaussian to
    reduce ringing.

    IMPORTANT NOTE: Convolution (width w) with a Gaussian is done only
    for the fourier domain.  Therefore, timeFunc provided should NOT
    contain the Gaussian, and it will not show up when e.g., plotted
    on the computer screen, but it will show up on the scope.
    
    The interpolation parameters can be overridden by changing the
    class variables NP_dt, NP_padTime, and NP_w.

    This is a more object oriented approach than the numerical_pulse
    decorator below.
    '''
    NP_dt=0.5*ns
    NP_padTime=1000*ns
    NP_w=0.0*ns

    def freqFunc(self, f):
        """
        A numerically defined frequency representation from NumericalPulse
        """
        ### FREQUENCY DOMAIN INTERPOLATING FUNCTION ###
        # Python DFT convention:
        #       N - 1
        #       ----
        #       \            -2*pi*i*n*k/N
        # a_k =  >    a_n * e
        #       /
        #       ----
        #      n = 0
        # But the actual FT relates to the DFT through a(v_k) = DFT(k)*T/N
        # Need to have the Fourier transform of [times,samples]
        # Best way to do this would probably be to pad the GRAPE pulse with starting values
        # (maybe remove the DC componant) and take the DFT

        dt = self.NP_dt['ns']
        padTime = self.NP_padTime['ns']
        w = self.NP_w['ns']

        numpts = 2**int(np.ceil(np.log2((self.end-self.start+2*padTime)/dt)))
        t = np.linspace(self.start-padTime, self.end+padTime, numpts, endpoint=False)
        assert len(t)==numpts
        dt = t[1]-t[0]  # Actual dt may be smaller to ensure power-of-two points

        samples = self.timeFunc(t)
        freqs = np.fft.fftshift(np.fft.fftfreq(numpts,dt))
        sampleSpectrum = np.fft.fft(samples)
        sampleSpectrum = dt*np.fft.fftshift(sampleSpectrum)
        # shift the function to its origin for a slowly evolving phase
        # envelope which is easy to interpolate If the signal is
        # symmetric, the imaginary part should be zero.  We'll subtract
        # this phase later with fast_phase

        sampleSpectrum = sampleSpectrum * np.exp(2.0j * np.pi * (self.end-self.start+2*padTime)/2.0 * freqs)
        interpSpecFuncR = interp1d_cubic(freqs, np.real(sampleSpectrum),fill_value=0)
        interpSpecFuncI = interp1d_cubic(freqs, np.imag(sampleSpectrum),fill_value=0)

        if w>0.0:
            # kernel = gaussian(0, w, 2*np.sqrt(np.log(2)/np.pi) / w)(f, fourier=True)
            kernel = cosine(0, self.w, amp = 2/self.w) # area=1
        else:
            kernel = 1.0

        # At this point, the middle of the pulse will be at time zero.  We need to shift it by end-start/2
        #fast_phase = 1 
        fast_phase = -2.0 * np.pi * (self.end+self.start)/2  * f

        return (interpSpecFuncR(f) + 1j*interpSpecFuncI(f)) * np.exp(1j*fast_phase) * kernel
        
@convertUnits(dt='ns', padTime='ns', w='ns')
def numerical_pulse(dt, padTime, w=0.0):
    '''
    Decorator to implement numerical pulses that do not have an analytic expression
    for the Fourier transform.  Just implement an Envelope subclass that has a 
    timeFunc, then decorate it with numerical_pulse() which will assign a freqFunc
    '''
    # This should be a subclass of Envelope rather than a decorator,
    # and particular numerical pulse classes should derive from it.  I
    # made it this way so that you could specify the interpolation
    # parameters (dt, padTime, w) in the decorator, but those should
    # be class variables that can be overridden. See NumericalPulse
    # above for a version that works like that.

    def decorator(cls):
        def freqFunc(self, f):
            """
            A numerically defined sequence
            
            INPUTS
            env: envelope with a timeFunc
            dt: time stepping for evaluating the time func
            padding: this needs to be large to get resolution in the frequency domain for the linear interpolation. For cubic it doesn't need to be big.
            
            IMPORTANT NOTE: Convolution (width w) with a Gaussian is done HERE, but only for the fourier domain.
            Therefore, timeFunc passed in through EnvIn should NOT contain the Gaussian.
            Hence, it will NOT show up in the timeFunc (plotted on the computer screen), but it will show up on the scope.
            """
            ### FREQUENCY DOMAIN INTERPOLATING FUNCTION ###
            # Python DFT convention:
            #       N - 1
            #       ----
            #       \            -2*pi*i*n*k/N
            # a_k =  >    a_n * e
            #       /
            #       ----
            #      n = 0
            # But the actual FT relates to the DFT through a(v_k) = DFT(k)*T/N
            # Need to have the Fourier transform of [times,samples]
            # Best way to do this would probably be to pad the GRAPE pulse with starting values
            # (maybe remove the DC componant) and take the DFT
        
            numpts = 2**int(np.ceil(np.log2((self.end-self.start+2*padTime)/self.dt)))
            t = np.linspace(self.start-padTime, self.end+padTime, numpts, endpoint=False)
            assert len(t)==numpts
            dt = t[1]-t[0]
            samples = self.timeFunc(t)
            freqs = np.fft.fftshift(np.fft.fftfreq(numpts,dt))
            sampleSpectrum = np.fft.fft(samples)
            sampleSpectrum = dt*np.fft.fftshift(sampleSpectrum)
            # shift the function to its origin for a slowly evolving phase envelope, for easy interpolation. 
            # If the signal is symmetric, the imaginary part should be zero.
            # We'll subtract this phase later with fast_phase
            sampleSpectrum = sampleSpectrum * np.exp(2.0j * np.pi * (self.end-self.start+2*padTime)/2.0 * freqs)
            interpSpecFuncR = interp1d_cubic(freqs, np.real(sampleSpectrum),fill_value=0)
            interpSpecFuncI = interp1d_cubic(freqs, np.imag(sampleSpectrum),fill_value=0)
            if w>0.0:
                kernel = gaussian(0, w, 2*np.sqrt(np.log(2)/np.pi) / w)(f, fourier=True)
            else:
                kernel = 1.0
            fast_phase = -2.0 * np.pi * (self.end+self.start)/2  * f
            return (interpSpecFuncR(f) + 1j*interpSpecFuncI(f)) * np.exp(1j*fast_phase) * kernel

        cls.freqFunc = types.MethodType(freqFunc, None, cls)
        cls.dt = dt
        return cls
    return decorator
'''
return 2.0 * np.pi * f * (env.start - Env.end - padTime )) 

'''
def test_numerical2(dt, padTime, EnvType, *args, **kw):
    '''
    Test numerical interpolation using the NumericalPulse class.  This
    works by making a copy of the input class dict, removing the
    provided freqFunc and constructing a new class that inherets from
    NumericalPulse instead of Envelope

    Because of the weird way this works, it will fail on classes that
    have been decorated by memoize_envelope.  Disable caching in order
    to use this.
    '''
    class_dict = EnvType.__dict__.copy()
    del class_dict['freqFunc']
    class_name = 'Numerical%s' % (EnvType.__name__)
    base_classes = (NumericalPulse,)
    class_dict['NP_dt'] = dt
    class_dict['NP_padTime'] = padTime
    class_dict['NP_w'] = 0*ns
    new_class = type(class_name, base_classes, class_dict)
    try:
        env1 = new_class(*args, **kw)
    except TypeError:
        raise RuntimeError('Disable caching in envelopes.py to use test_numerical2')
    plot_env(env1, padding=20*ns)

def test_numerical(dt, padTime, EnvType, *args, **kwargs):
    '''
    Tests the numerical pulse mechanism.  Takes an input envelope class and constructor arguments
    and builds a numerical approximation of the fourier transform, then it plots that vs. the 
    time domain data.  This function can be used on envelopes like cosine() that have analytic
    fourier transforms, and calculate a numerical approximation.
    '''
    @numerical_pulse(dt=dt, padTime=padTime)
    class NumericalEnv(EnvType):
        pass
    env1 = NumericalEnv(*args, **kwargs)
    plot_env(env1)

# empty envelope
NOTHING = EnvZero(start=None, end=None)
ZERO = EnvZero(start=0, end=0)

# this envelope doesn't make sense.  It is never used, so I removed it -- ERJ
#_one = lambda x: 1
# ONE = DynamicEnvelope(_one, _one, start=0, end=0) 

@envelope_factory_t0
@memoize_envelope
class gaussian(Envelope):
    @convertUnits(t0='ns', w='ns', amp=None, phase=None, df='GHz')
    def __init__(self, t0, w, amp=1.0, phase=0.0, df=0.0):
        """A gaussian pulse with specified center and full-width at half max."""
        self.sigma = w / np.sqrt(8*np.log(2)) # convert fwhm to std. deviation
        self.t0 = t0
        self.amp = amp
        self.phase = phase
        self.df = df
        Envelope.__init__(self, start=t0-2*w, end=t0+2*w)
    def timeFunc(self, t):
        return self.amp * np.exp(-(t-self.t0)**2/(2*self.sigma**2) - 2j*np.pi*self.df*(t-self.t0) + 1j*self.phase)
        
    def freqFunc(self, f):
        sigmaf = 1 / (2*np.pi*self.sigma) # width in frequency space
        ampf = self.amp * np.sqrt(2*np.pi*self.sigma**2) # amp in frequency space
        return ampf * np.exp(-(f+self.df)**2/(2*sigmaf**2) - 2j*np.pi*f*self.t0 + 1j*self.phase)

@envelope_factory_t0
@memoize_envelope
class cosine(Envelope):
    @convertUnits(t0='ns', w='ns', amp=None, phase=None)
    def __init__(self, t0, w, amp=1.0, phase=0.0):
        """A cosine function centered at t0 with FULL WIDTH w"""
        self.t0 = t0
        self.w = w
        self.amp = amp
        self.phase = phase
        Envelope.__init__(self, t0-w/2.0, t0+w/2.0)
    def timeFunc(self, t):
        return self.amp * 0.5 * (1+np.cos(2*np.pi*(t-self.t0)/self.w)) * (((t-self.t0)+self.w/2.)>0) * ((-(t-self.t0)+self.w/2.)>0) * np.exp(1j*self.phase)
    def freqFunc(self, f):
        wf = self.w*f
        a = 1.-wf
        b = 1.+wf
        return self.amp*np.exp(-2j*np.pi*f*self.t0+1j*self.phase)*self.w/4.*(2.*np.sinc(wf)+np.sinc(a)+np.sinc(b))


@envelope_factory_t0
@memoize_envelope
class cosineLambda(Envelope):
    @convertUnits(t0='ns', w='ns', amp=None, phase=None, lambdas=None)
    def __init__(self, t0, w, amp=1.0, phase=0.0,lambdas=()):
        """Cosine-like function: (1-cos(2*pi*n*(t-w/2)/w)/2, centered at t0 with FULL WIDTH w. Useful for mitigating parasitic rotations on other qubits. RB."""
        self.t0 = t0
        self.w = w
        self.amp = amp
        self.phase = phase
        self.lambdas = (1.-np.sum(lambdas),) + lambdas #initialize first element
        Envelope.__init__(self, t0-w/2.0, t0+w/2.0)
    def timeFunc(self, t):
        envelope=0.
        for idx,lambdaValue in enumerate(self.lambdas):
            n=idx+1.
            envelope += lambdaValue * self.amp * 0.5 * (1-np.cos(2*np.pi*n*(t-self.t0-self.w/2)/self.w)) * (((t-self.t0)+self.w/2.)>0) * ((-(t-self.t0)+self.w/2.)>0) * np.exp(1j*self.phase)
        return envelope
    def freqFunc(self, f):
        envelope=0.
        for idx,lambdaValue in enumerate(self.lambdas):
            n=idx+1.      
            wf = self.w*f
            a = n-wf
            b = n+wf            
            envelope += lambdaValue*self.w/4.*(2.*np.sinc(wf)+ (-1)**(n-1)*(np.sinc(a)+np.sinc(b)))
        return self.amp*np.exp(-2j*np.pi*f*self.t0+1j*self.phase)*envelope
        
        
@memoize_envelope
class CosineHDBase(Envelope):
    '''
    cosine pulse with drag, which is more efficient for caching.  Apply time and phase shifts
    separately to avoid caching issue
    '''
    @convertUnits(w='ns', amp=None, drag=None)
    def __init__(self, w, amp=1.0, drag=0.0):
        cosenv = cosine(0.0, w, amp, 0.0)
        dragenv = 1j * drag * deriv(cosenv)
        self.env = cosenv + dragenv
        Envelope.__init__(self, -w/2.0, w/2.0)
        
    def timeFunc(self, t):
        return self.env.timeFunc(t)
    def freqFunc(self, f):
        return self.env.freqFunc(f)

@convertUnits(t0='ns', w='ns', amp=None, phase=None, drag=None)
def cosineHD(t0, w, amp=1.0, phase=0.0, drag=0.0):
    '''
    This wraps the underlying CosineHD class with zero time and phase shifts,
    then does the shift at the end.  This reduces unnecessary redundant caching.
    '''
    env = CosineHDBase(w, amp, drag)
    if phase:
        env = env*np.exp(1j*phase)
    if t0 or phase:
        env = shift(env, t0)
    return env


@memoize_envelope
class CosineWWFullBase(Envelope):
    '''
    Creates a cosine pulse with up to 3 notches in the frequency spectrum, at freqs RB.
    More efficient for caching.  Apply time and phase shifts
    separately to avoid caching issue.
    '''
    @convertUnits(w='ns', amp=None)
    def __init__(self, w, amp=1.0, freqs=()):
        cosenv = cosine(0.0, w, amp, 0.0)
        if len(freqs)==0:
            alpha=0
            beta=0
            gamma=0        
        elif len(freqs)==1:
            #DRAG
            w0=freqs[0]*(2*np.pi)
            alpha = 1. / w0
            beta=0
            gamma=0        
        elif len(freqs)==2:
            #WAHWAH
            w0=freqs[0]*(2*np.pi)
            w1=freqs[1]*(2*np.pi)
            alpha = (w0+w1) / (w0*w1)
            beta  = -1. / (w0*w1)
            gamma = 0
        elif len(freqs)==3:
            w0=freqs[0]*(2*np.pi)
            w1=freqs[1]*(2*np.pi)
            w2=freqs[2]*(2*np.pi)
            alpha = (w0*w1+w0*w2+w1*w2) / (w0*w1*w2)
            beta  = -(w0+w1+w2) / (w0*w1*w2)
            gamma = -1. / (w0*w1*w2)
        else:
            raise Exception('only up to three notch frequencies')
        Xdot=deriv(cosenv)
        Xdotdot=deriv(Xdot)
        Xdotdotdot=deriv(Xdotdot)
        X = cosenv + beta *  Xdotdot
        Y = alpha * Xdot + gamma*Xdotdotdot
        self.env = X + 1j * Y
        Envelope.__init__(self, -w/2.0, w/2.0)
    def timeFunc(self, t):
        return self.env.timeFunc(t)
    def freqFunc(self, f):
        return self.env.freqFunc(f)

@convertUnits(t0='ns', w='ns', amp=None, phase=None)
def cosineWWFull(t0, w, amp=1.0, phase=0.0, freqs=()):
    '''
    This wraps the underlying CosineWAHWAH class with zero time and phase shifts,
    then does the shift at the end.  This reduces unnecessary redundant caching.
    ''' 
    env = CosineWWFullBase(w, amp, freqs)
    if phase:
        env = env*np.exp(1j*phase)
    if t0 or phase:
        env = shift(env, t0)
    return env    
    
 

@memoize_envelope
class CosineLambdaHDBase(Envelope):
    '''
    cosine pulse with drag, which is more efficient for caching.  Apply time and phase shifts
    separately to avoid caching issue
    '''
    @convertUnits(w='ns', amp=None, drag=None, lambdas=None)
    def __init__(self, w, amp=1.0, drag=0.0,lambdas=()):
        cosenv = cosineLambda(0.0, w, amp, 0.0,lambdas)
        dragenv = 1j * drag * deriv(cosenv)
        self.env = cosenv + dragenv
        Envelope.__init__(self, -w/2.0, w/2.0)        
    def timeFunc(self, t):
        return self.env.timeFunc(t)
    def freqFunc(self, f):
        return self.env.freqFunc(f)

@convertUnits(t0='ns', w='ns', amp=None, phase=None, drag=None, lambdas=None)
def cosineLambdaHD(t0, w, amp=1.0, phase=0.0, drag=0.0,lambdas=()):
    '''
    This wraps the underlying CosineHD class with zero time and phase shifts,
    then does the shift at the end.  This reduces unnecessary redundant caching.
    '''
    env = CosineLambdaHDBase(w, amp, drag, lambdas)
    if phase:
        env = env*np.exp(1j*phase)
    if t0 or phase:
        env = shift(env, t0)
    return env
    
    
@envelope_factory_t0
@memoize_envelope
class triangle(Envelope):
    @convertUnits(t0='ns', tlen='ns', amp=None)
    def __init__(self, t0, tlen, amp, fall=True):
        """A triangular pulse, either rising or falling."""
        self.t0 = t0
        self.tlen = tlen
        self.amp = amp
        if not fall:
            self.t0 = t0+tlen
            self.tlen = -tlen
            self.amp = amp
    
        tmin = min(t0, t0+tlen)
        tmax = max(t0, t0+tlen)
        Envelope.__init__(self, tmin, tmax)

    def timeFunc(self, t):
        return self.amp * (t >= self.start) * (t < self.end) * (1 - (t-self.t0)/self.tlen)
    
    def freqFunc(self, f):
        if self.tlen == 0 or self.amp == 0:
            return 0.0*f
        # this is tricky because the fourier transform has a 1/f term, which blows up for f=0
        # the z array allows us to separate the zero-frequency part from the rest
        z = f == 0
        f = 2j*np.pi*(f + z)
        return self.amp * ((1-z)*np.exp(-f*self.t0)*(1.0/f - (1-np.exp(-f*self.tlen))/(f**2*self.tlen)) + z*self.tlen/2.0)
    
@envelope_factory_t0
@memoize_envelope
class exponential(Envelope):
    @convertUnits(t0='ns', len='ns', tau='ns', amp=None)
    def __init__(self, t0, len, tau, amp):
        """
        An exponential pulse.
        
        INPUT PARAMETERS
        t0 - time: Initial time
        alen - time: Duration
        tau - time: Time constant
        amp - float: Amplitude at initial time
        """
        tmin = min(t0, t0+len)
        tmax = max(t0, t0+len)
        self.tau = tau
        self.amp = amp
        self.len = len
        if tau==0 or len == 0 or amp == 0:
            self.timeFunc = lambda x: 0*x
            self.freqFunc = lambda x: 0*x
        Envelope.__init__(tmin, tmax)
    
    def timeFunc(self, t):
        return self.amp * np.exp((t-self.start)/self.tau) * (t>=self.start) * (t<self.end)
    
    def freqFunc(self, f):
        iw = 2j * np.pi * f
        gamma = 1.0/self.tau - iw
        return self.amp * np.exp(-iw*t0) * (np.exp(gamma*self.len)-1) / gamma

@envelope_factory_t0
@memoize_envelope        
class rect(Envelope):
    @convertUnits(t0='ns', tlen='ns', amp=None, overshoot=None)
    def __init__(self, t0, tlen, amp, overshoot=0.0, overshoot_w=1.0):
        """A rectangular pulse with sharp turn on and turn off.
        
        Note that the overshoot_w parameter, which defines the FWHM of the gaussian overshoot peaks
        is only used when evaluating the envelope in the time domain.  In the fourier domain, as is
        used in the dataking code which uploads sequences to the boards, the overshoots are delta
        functions.
        """
        self.t0 = t0
        self.amp = amp
        self.overshoot = overshoot * np.sign(amp) # overshoot will be zero if amp is zero
        tmin = min(t0, t0+tlen)
        tmax = max(t0, t0+tlen)
        self.tmid = (tmin + tmax) / 2.0
        self.tlen = tlen

        # to add overshoots in time, we create an envelope with two gaussians
        if overshoot:
            o_w = overshoot_w
            o_amp = 2*np.sqrt(np.log(2)/np.pi) / o_w # total area == 1
            self.o_env = gaussian(tmin, o_w, o_amp) + gaussian(tmax, o_w, o_amp)
        else:
            self.o_env = EnvZero(tmin, tmax)

        Envelope.__init__(self, tmin, tmax)
    def timeFunc(self, t):
        return (self.amp * (t >= self.start) * (t < self.end) +
                self.overshoot * self.o_env(t))
    
    # to add overshoots in frequency, use delta funcs (smoothed by filters)
    def freqFunc(self, f):
        return (self.amp * abs(self.tlen) * np.sinc(self.tlen*f) * np.exp(-2j*np.pi*f*self.tmid) +
                self.overshoot * (np.exp(-2j*np.pi*f*self.start) + np.exp(-2j*np.pi*f*self.end)))
    
"""
@envelope_factory_t0
@memoize_envelope
class flattop(Envelope):
    def _asdf__new__(cls, t0, tlen, w, amp=1.0, overshoot=0.0, overshoot_w=1.0, cache=None):
        '''
        __new__ optimizes the case where amp=0 by constructing an EnvZero instance
        instead of a flattop.  This seems to happen unnecessarily often, so this saves memory,
        and maybe a bit of performance
        '''
        if amp==0:
            return EnvZero(t0, t0+tlen)
        elif w==0:
            return rect(t0,tlen)
        else:
            # __init__ will be called!
            return Envelope.__new__(cls)
        
    @convertUnits(t0='ns', tlen='ns', w='ns', amp=None)
    def __init__(self, t0, tlen, w, amp=1.0, overshoot=0.0, overshoot_w=1.0):
        '''A rectangular pulse convolved with a gaussian to have smooth rise and fall.'''
        self.t0 = t0
        self.tlen = tlen
        self.amp = amp
        self.w = w
        self.overshoot = overshoot * np.sign(amp) # overshoot will be zero if amp is zero
        tmin = min(t0, t0+tlen)
        tmax = max(t0, t0+tlen)
        Envelope.__init__(self, tmin-w, tmax+w)
    # to add overshoots in time, we create an envelope with two gaussians
        if overshoot:
            o_w = overshoot_w
            o_amp = 2*np.sqrt(np.log(2)/np.pi) / o_w # total area == 1
            self.o_env = gaussian(tmin, o_w, o_amp) + gaussian(tmax, o_w, o_amp)
        else:
            self.o_env = EnvZero(tmin-w, tmax+w)

    def timeFunc(self, t):
        a = 2*np.sqrt(np.log(2)) / self.w
        tmin = min(self.t0, self.t0+self.tlen)
        tmax = max(self.t0, self.t0+self.tlen)
        return (self.amp * (erf(a*(tmax - t)) - erf(a*(tmin - t)))/2.0 +
                self.overshoot * self.o_env(t))

    # to add overshoots in frequency, use delta funcs (smoothed by filters)
    def freqFunc(self, f):
        #
        #  This function calculates a gaussian convolved with a
        #  top-hat, and then adds overshoot delta functions.  I
        #  don't know why it does not convolve the overshoot with
        #  the gaussian.  -- ERJ
        rect_env = rect(self.t0, self.tlen, 1.0)
        kernel = gaussian(0, self.w, 2*np.sqrt(np.log(2)/np.pi) / self.w) # area = 1
        return (self.amp * rect_env(f, fourier=True) * kernel(f, fourier=True) + # convolve with gaussian kernel
                self.overshoot * (np.exp(-2j*np.pi*f*self.t0) + np.exp(-2j*np.pi*f*(self.t0+self.tlen))))
"""
                
@envelope_factory_t0
@memoize_envelope
class flattop(Envelope):
    def _asdf__new__(cls, t0, tlen, w, amp=1.0):
        '''
        __new__ optimizes the case where amp=0 by constructing an EnvZero instance
        instead of a flattop.  This seems to happen unnecessarily often, so this saves memory,
        and maybe a bit of performance
        '''
        if amp==0:
            return EnvZero(t0, t0+tlen)
        else:
            # __init__ will be called!
            return Envelope.__new__(cls)
        
    @convertUnits(t0='ns', tlen='ns', w='ns', amp=None)
    def __init__(self, t0, tlen, w, amp=1.0):
        self.t0 = t0
        self.tlen = tlen
        self.amp = amp
        self.w = w
        Envelope.__init__(self, t0, t0+tlen)

    def timeFunc(self, t):
        t0 = self.t0
        length = self.tlen
        w = self.w
        amp = self.amp
        if w>0:
            return amp/2.*(1+np.sin(np.pi*(t-t0)/w - np.pi/2.))*(t>t0)*(t<(t0+w)) + \
                   amp/2.*(1+np.sin(np.pi*(t-t0-length)/w-np.pi/2.))*(t>(t0+length-w))*(t<(t0+length)) + \
                   amp*(t>=(t0+w))*(t<=(t0+length-w))
        else:
            return rect(t0,length,amp=amp)(t, fourier=False)
            
    def freqFunc(self, f):
        rect_env = rect(self.t0+self.w/2., self.tlen-self.w, 1.0)
        if self.w>0:            
            kernel = cosine(0, self.w, amp = 2/self.w) # area=1        
            return self.amp * rect_env(f, fourier=True) * kernel(f, fourier=True)# convolve with cosine kernel
        else:
            return self.amp * rect_env(f, fourier=True)
                
@envelope_factory_t0
@memoize_envelope
class adiabaticCZ(NumericalPulse):
    @convertUnits(t0='ns', length='ns', amplitude='GHz', w='ns', 
                  thi='au', thf='au', lam2='au', lam3='au')
    def __init__(self, t0, length, amplitude=1.0, 
                 thf = 0.55*np.pi/2, thi = 0.05, 
                 lam2 = -0.18, lam3 = 0.04, 
                 resolution = 1024, w = 0.0,
                 ftozparams=None, ztofparams=None, zOffsetDC=0.0):
        """
        IMPORTANT NOTE: Convolution (width w) with a Gaussian is done in numericalPulse, but only for the fourier domain.
        Therefore, timeFunc passed in through EnvIn should NOT contain the Gaussian.
        Hence, the gaussian rounding will NOT show up in the timeFunc (plotted on the computer screen), but it will show up on the scope.
        """    
        # start john code
        self.t0 = t0 + w/2.0
        self.length = length - w
        self.amplitude = amplitude
        self.thf = thf
        self.thi = thi
        self.lam2 = lam2
        self.lam3 = lam3
        self.resolution = resolution
        self.w = w
        self.ftozparams = ftozparams
        self.ztofparams = ztofparams
        self.zOffsetDC = zOffsetDC
    
        Envelope.__init__(self, t0, t0+length)

    def timeFunc(self, tpts):
        
        t0 = self.t0
        length = self.length
        amplitude = self.amplitude
        thf = self.thf
        thi = self.thi
        lam2 = self.lam2
        lam3 = self.lam3
        resolution = self.resolution
        w = self.w
        ftozparams = self.ftozparams
        ztofparams = self.ztofparams
        zOffsetDC = self.zOffsetDC

        # start john code
        t=np.linspace(0,1,resolution)
        han2 = (1-lam3)*(1-np.cos(2*np.pi*t))+lam2*(1-np.cos(4*np.pi*t))+lam3*(1-np.cos(6*np.pi*t)) #original from john
        thsl=thi+(thf-thi)*han2/np.max(han2)
        tlu=np.cumsum(np.sin(thsl))*t[1]
        tlu=tlu-tlu[0]
        th=interp1d(tlu,thsl, bounds_error=False, fill_value = 0)    
        t=np.linspace(0, tlu[-1], resolution)
        th=1/np.tan(th(t))
        th -=th[0]
        th *=1/np.min(th)
        # end john code, now have normalized cz pulse of length resolution
        
        t = np.linspace(t0, t0+length, resolution)
        if ftozparams and ztofparams:

        #here, amplitude is assumed to be the relative frequency. The '+' sign is a convention I choose
        #freqOffset = np.array([ f10['GHz'] - b for b in th*amplitude])        
        #cz = interp1d_cubic(t, f1func(freqOffset), bounds_error=False, fill_value = 0.0)
        
        #below is changed to be compatible with zOffsetDC. This way, you won't have to redo spectroscopyzfunc
        
            freqOffset = np.array([ zfuncs.AmpToFrequencyFunc(ztofparams, 0.0) + b for b in th*amplitude])
            cz = interp1d_cubic(t, zfuncs.FrequencyToAmpFunc(ftozparams, freqOffset)-0.0 , bounds_error=False, fill_value = 0.0)        
        else:
            cz = interp1d_cubic(t, th*amplitude, bounds_error=False, fill_value = 0.0)

        return cz(tpts)

        
@envelope_factory_t0
@memoize_envelope         
class adiabaticSWAP(NumericalPulse):
    @convertUnits(t0='ns', length='ns', Deltaf1='GHz', w='ns',holdTime='ns')
    def __init__(self, t0, length, Deltaf1, 
                 thf = np.pi-0.1, thi = 0.1,
                 lam2 = -0.2, lam3 = 0.0, lam4 = 0.0,
                 resolution = 1024, w = 0.0,holdTime=0.0,
                 ftozparams=None,ztofparams=None,zOffsetDC=0.0):
        """
        IMPORTANT NOTE: Convolution (width w) with a Gaussian is done in numericalPulse, but only for the fourier domain.
        Therefore, timeFunc passed in through EnvIn should NOT contain the Gaussian.
        Hence, the gaussian rounding will NOT show up in the timeFunc (plotted on the computer screen), but it will show up on the scope.
        Deltaf1 = f_wheretoswap - f_qubit (<0 if qubit needs to go up in freq)
        Deltaf2 = f_rest  - f_qubit. Deltaf2 is implied with the choice for thf, thi
        RB
        """    
        # start rami code
        self.t0 = t0
        self.length = length
        self.Deltaf1 = Deltaf1
        self.thf = thf
        self.thi = thi
        self.lam2 = lam2
        self.lam3 = lam3
        self.lam4 = lam4
        self.resolution = resolution
        self.w = w
        self.holdTime = holdTime
        self.ftozparams = ftozparams
        self.ztofparams = ztofparams
        self.zOffsetDC = zOffsetDC
    
        Envelope.__init__(self, t0, t0+length+holdTime)

    def timeFunc(self, tpts):
        
        t0 = self.t0
        length = self.length
        Deltaf1 = self.Deltaf1
        thf = self.thf
        thi = self.thi
        lam2 = self.lam2
        lam3 = self.lam3
        lam4 = self.lam4        
        resolution = self.resolution
        w = self.w
        holdTime = self.holdTime
        ftozparams = self.ftozparams
        ztofparams = self.ztofparams
        zOffsetDC = self.zOffsetDC        
        
        # start rami code
        amps=np.array([1.-lam2-lam3-lam4, lam2, lam3,lam4])
        amps*= (thf-thi) / np.sum(amps)
        tau=np.linspace(0,1,resolution)
        dthdtau = amps[0]*(1-np.cos(2*np.pi*tau))+amps[1]*(1-np.cos(4*np.pi*tau))+amps[2]*(1-np.cos(6*np.pi*tau))+amps[3]*(1-np.cos(8*np.pi*tau))     
        thtau=np.cumsum(dthdtau) * tau[1] #np.max(tau)/(resolution-1)
        thtau-=thtau[0]
        thtau+=thi
        ttau=np.cumsum(np.sin(thtau))* tau[1] #np.max(tau)/(resolution-1)
        ttau=ttau-ttau[0]
        th=interp1d(ttau,thtau, bounds_error=False, fill_value = 0)    
        t=np.linspace(0, ttau[-1], resolution)
        hx=-Deltaf1*np.tan(thi)
        hz=hx/np.tan(th(t)) 
        hz-=hz[0]      
        # end rami code, hz will start at the qubit frequency, swap at Deltaf1, and rest at Deltaf2, with Deltaf2=Deltaf1+Hx/tan(thf)
        
        #add a wait Time, holding at the end frequency. The wait time needs to be in units of dt (length/(resolution-1))
        dt=1.0*length/(resolution-1)
        nforholding = np.round( holdTime/dt)
        holdTime = nforholding*dt
        hz=np.hstack( (hz, hz[-1]*np.ones(nforholding) ) )
        #now, t, and hz have length: resolution + nforholding        
        
        t = np.linspace(t0, t0+length+holdTime, resolution+nforholding)
        if ftozparams and ztofparams:

        #here, amplitude is assumed to be the relative frequency. The '+' sign is a convention I choose
        #freqOffset = np.array([ f10['GHz'] - b for b in th*amplitude])        
        #cz = interp1d_cubic(t, f1func(freqOffset), bounds_error=False, fill_value = 0.0)
        
        #below is changed to be compatible with zOffsetDC. This way, you won't have to redo spectroscopyzfunc
        
            freqOffset = np.array([ zfuncs.AmpToFrequencyFunc(ztofparams, 0.0) + b for b in hz])
            cz = interp1d_cubic(t, zfuncs.FrequencyToAmpFunc(ftozparams, freqOffset)-0.0 , bounds_error=False, fill_value = 0.0)        
        else:
            cz = interp1d_cubic(t, hz, bounds_error=False, fill_value = 0.0)
        return cz(tpts)
        
        
        
        
@envelope_factory_t0
@memoize_envelope
class wigglyZ(NumericalPulse):

    @convertUnits(amplitude=None)
    def __init__(self, t0, t, z, amplitude = 1.0):
        self.length = t0 + t[-1]
        self.z = interp1d_cubic(np.array(t), z*np.array(amplitude), bounds_error=False, fill_value = 0.0)
        Envelope.__init__(self, start=t0, end=t0+self.length)
    
    def timeFunc(self,t):
        return self.z(t)


class GRAPE_CZ(NumericalPulse):
    @convertUnits(t0='ns')
    def __init__(self, t0, t, samples):
        """
        Turns a time axis and detuning samples into an Envelope.
        
        IMPORTANT NOTE: Convolution (width w) with a Gaussian is done in numericalPulse, but only for the fourier domain.
        Therefore, timeFunc passed in through EnvIn should NOT contain the Gaussian.
        Hence, the gaussian rounding will NOT show up in the timeFunc (plotted on the computer screen), but it will show up on the scope.
        """
        length = t[-1] - t[0]
        dt = t[1]-t[0]
        
        self.cz = interp1d_cubic(t, samples, bounds_error=False, fill_value = 0.0)
        Envelope.__init__(self, t0, t0+length)
    
    def timeFunc(self, t):
        return self.cz(t)

@convertUnits(t0='ns', tlen='ns')
def wait(t0, tlen):
    tmin=min(t0, t0+tlen)
    tmax=max(t0, t0+tlen)
    return EnvZero(start=tmin, end=tmax)
    
@convertUnits(t0='ns', rise='ns', hold='ns', fall='ns', amp=None)
def trapezoid(t0, rise, hold, fall, amp):
    """Create a trapezoidal pulse, built up from triangles and rectangles."""
    return (triangle(t0, rise, amp, fall=False) +
            rect(t0+rise, hold, amp) +
            triangle(t0+rise+hold, fall, amp))

class mix(Envelope):
    """Apply sideband mixing at difference frequency df."""
    @convertUnits(df='GHz')
    def __init__(self, env, df=0.0, phase=0.0):
        if df is None:
            raise Exception
        self.df = df
        self.phase = phase
        self.env = env
        Envelope.__init__(self, env.start, env.end)
    def timeFunc(self, t):
        return self.env(t) * np.exp(-2j*np.pi*self.df*t - 1.0j*self.phase)
    def freqFunc(self, f):
        return self.env(f + self.df, fourier=True)*np.exp(-1.0j*self.phase)

class deriv(Envelope):
    @convertUnits(dt='ns')
    def __init__(self, env, dt=0.1):
        """Get the time derivative of a given envelope."""
        self.env = env
        self.dt = dt
        Envelope.__init__(self,env.start, env.end)
    def timeFunc(self, t):
        return (self.env(t+self.dt) - self.env(t-self.dt)) / (2*self.dt)
    def freqFunc(self, f):
        return 2j*np.pi*f * self.env(f, fourier=True)

class dragify(Envelope):
    @convertUnits(dt='ns')
    def __init__(self, env, alpha, dt):
        self.env = env
        self.alpha = alpha
        self.dt = dt
        Envelope.__init__(self, env.start, env.end)
    def freqFunc(self, f):
        return (1+2j*np.pi*f*self.alpha) * self.env(f, fourier=True)
    def timeFunc(self, t):
        return (self.env(t) + self.alpha*(self.env(t+self.dt) - self.env(self.t-self.dt))/(2*self.dt))

class shift(Envelope):
    @convertUnits(dt='ns')
    def __init__(self, env, dt):
        self.dt = dt
        self.env = env
        Envelope.__init__(self, env.start+dt if env.start is not None else None, 
                                env.end+dt if env.end is not None else None)
    def timeFunc(self, t):
        return self.env(t-self.dt)
    def freqFunc(self, f):
        return self.env(f, fourier=True) * exp(-2j*np.pi*f*self.dt)

# utility functions

def timeRange(envelopes):
    """Calculate the earliest start and latest end of a list of envelopes.
    
    Returns a tuple (start, end) giving the time range.  Note that one or
    both of start and end may be None if the envelopes do not specify limits.
    """
    starts = [env.start for env in envelopes if env.start is not None]
    start = min(starts) if len(starts) else None
    ends = [env.end for env in envelopes if env.end is not None]
    end = max(ends) if len(ends) else None
    return start, end


def fftFreqs(time=1024):
    """Get a list of frequencies for evaluating fourier envelopes.
    
    The time is rounded up to the nearest power of two, since powers
    of two are best for the fast fourier transform.  Returns a tuple
    of frequencies to be used for complex and for real signals.
    """
    nfft = 2**int(math.ceil(math.log(time, 2)))
    f_complex = np.fft.fftfreq(nfft)
    f_real = f_complex[:nfft/2+1]
    return f_complex, f_real


def ifft(envelope, t0=-200, n=1000):
    f = np.fft.fftfreq(n)
    return np.fft.ifft(envelope(f, fourier=True) * np.exp(2j*np.pi*t0*f))


def fft(envelope, t0=-200, n=1000):
    t = t0 + np.arange(n)
    return np.fft.fft(envelope(t))


def plotFT(envelope, t0=-200, n=1000):
    t = t0 + np.arange(n)
    y = ifft(envelope, t0, n)
    plt.plot(t, np.real(y))
    plt.plot(t, np.imag(y))


def plotTD(envelope, t0=-200, n=1000):
    t = t0 + np.arange(n)
    y = envelope(t)
    plt.plot(t, np.real(y))
    plt.plot(t, np.imag(y))
    
def plot_env(env, padding=100.0, N=1024):
    tstart = env.start - padding
    tend = env.end + padding
    t = np.linspace(tstart, tend, N, endpoint=False)
    x = env(t)
    f = np.fft.fftfreq(N, (tend-tstart)*1.0/N)
    x2 = np.fft.ifft(env(f, fourier=True)*np.exp(2j*np.pi*f*tstart)) * N / (tend-tstart)
    plt.figure()
    plt.plot(t, x, label='Time domain data')
    plt.plot(t, x2.real, label='Re(IFFT) of Frequeny envelope')
    plt.plot(t, x2.imag, label='Im(IFFT) of Frequeny envelope')
    plt.legend()
    plt.figure()
    plt.plot(f, np.abs(env(f, fourier=True)))
    plt.plot(f, np.abs(np.fft.fft(env(t))/(2*np.sqrt(2*np.pi))))
