import numpy as np

from pyle import envelopes as env
from pyle.dataking import envelopehelpers as eh
from pyle.registry import AttrDict
import copy
import pyle.gates as gates

from labrad.units import Unit, Value
import logging
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
ns, GHz, us = (Unit(un) for un in ['ns', 'GHz', 'us'])

def getQubit(qubitType,name='q'):
    q = AttrDict()
    q['__name__'] = name
    q['piAmp'] = 0.6
    q['piLen'] = 20*ns
    q['piFWHM'] = 8*ns
    q['uwavePhase'] = 0.0
    q['readoutLen'] = 200*ns
    q['readoutAmp'] = 0.1
    q['readoutWidth'] = 10*ns
    q['readoutFrequency'] = 4.0*GHz
    q['readout fc'] = 4.1*GHz
    q['f10'] = 6.1*GHz
    q['f21'] = 6.05*GHz
    q['fc'] = 6.0*GHz
    q['measureAmp'] = 0.1
    q['measureLen'] = 200*ns
    return AttrDict2Qubit(q, qubitType)
    
###############
### DEVICES ###
###############

class Agent(AttrDict): #I hate this but I did it to not break code using the . as dict item access. Sorry.
    def __init__(self):
        AttrDict.__init__(self)
        self['_t'] = 0.0*ns
        self['gates'] = []
    
class Qubit(Agent):
    """
    Each Qubit object has a current time, self._t which is the current
    start time of this qubit's sequence.
    """
    def __init__(self):
        Agent.__init__(self)

    def excite(self):
        t = self['_t']
        l = self['piLen']
        self['xy'] += eh.mix(self, eh.piPulse(self, t-(l/2)))
        self['_t'] -= l
        
    def measure(self):
        raise Exception('Not implemented')
        
class PhaseQubit(Qubit):
    pass
    
class Readout(Agent):
    pass
    
class CoherentPumpReadout(Readout):
    def __init__(self):
        Agent.__init__(self)
        self['qubits'] = []
    
    def measure(self):
        pumpSideband = self['pumpFrequency'] - (self['carrierFrequency']*self['pumpHarmonic'])
        self['readoutLen'] = max([q['readoutLen'] for q in self['qubits']])
        pumpDelay = self.get('pumpDelay', 0*ns)
        if pumpDelay['ns'] > 0:
            #print "Using pump delay of %s" % pumpDelay  
            pass
        # There is a 500 nanosecond extra pad on the end of the pump to capture the ring-down.
        # There is certainly a better way to do this.
        self['pp'] = env.mix(env.flattop(pumpDelay['ns'], self['readoutLen']-2*pumpDelay['ns']+1500*ns, w=self['readoutWidth'], amp=self['pumpAmplitude']), pumpSideband)
        #print "pump amplitude %s" % self['pumpAmplitude']
    
class SimpleReadout(Readout):
    def __init__(self):
        Agent.__init__(self)
        self['qubits'] = []
    
    def measure(self, *args, **kwargs):
        pass
        
class Resonator(Qubit):
    def __init__(self):
        Agent.__init__(self)
    
    def measure(self):
        pass

class Transmon(Qubit):
    '''
    Measure can be called multiple times to generate multiple readout windows.
    The delay parameter is the start of the next readout pulse relative to the start
    of the readout dac.
    
    To make general multiple-readout capability, we need to unify the time axis for readout
    and xyz, which can only happen with the future jump table based gate compiler.
    '''
    def __init__(self):
        Qubit.__init__(self)
        self['adcReadoutWindows'] = {}

CONSTRUCTORS = {
                'phaseQubit': PhaseQubit,
                'transmon': Transmon,
                'resonator' : Resonator,
                'simpleReadout': SimpleReadout,
                }

def AttrDict2Qubit(q, qubitType):
    """Convert from a pure registry.AttrDict to a Qubit"""
    raise Exception('Depricated: use registryWrapper2GateCompiler instead')
    qubit = qubitType()
    #Copy all values from q into qubit
    for key,value in q.items():
        qubit[key] = value
    object.__setattr__(qubit, '__name__', q.__name__)
    return qubit
    
def AttrDict2QubitForSim(q, qubitType):
    """Convert from a pure registry.AttrDict to a Qubit"""
    qubit = qubitType()
    #Copy all values from q into qubit
    for key,value in q.items():
        qubit[key] = value
    object.__setattr__(qubit, '__name__', q.__name__)
    return qubit
    
def attrDict2GateCompilerObject(deviceAsAttrDict):
    devType = deviceAsAttrDict['_type']
    dev = CONSTRUCTORS[devType]()
    for key,value in deviceAsAttrDict.items():
        dev[key]=value
    object.__setattr__(dev, '__name__', deviceAsAttrDict.__name__)
    object.__setattr__(dev, '_dir', deviceAsAttrDict._dir) # This shouldn't really be here, but it is part of AttrDict
    return dev
    
def registryWrapper2GateCompiler(sample, qubits, measure):
    '''
    Converts ordinary dictionary-like objects into gate compiler objects.

    This: constructs the derived Qubit/Resonator/Readout objects
    Updates device pointers for each qubit to its readout device and vice versa
    Adds the appropriate ADC readout channels to readout devices for all qubits being read out
    '''
    #Make a copy of qubit list
    devs = dict(qubits.items())

    if not any([q.get('readout', False) for q in devs.values()]):
        raise RuntimeError("No readout devices selected.  q['readout']=True must be set before calling registryWrapper2GateCompiler")

    #Add readout attr dicts to devs
    #for q in qubits:
    for q in qubits.values(): 
        if 'readoutConfig' in q: # resonators are not read out, only qubits
            readoutDev = q['readoutConfig']
            if readoutDev not in devs:
                devs[readoutDev] = sample[readoutDev]

    #Convert to gate compiler classes
    devs = dict([(name, attrDict2GateCompilerObject(dev)) for name,dev in devs.items()])
    readoutDevices = dict([(name, dev) for name,dev in devs.items() if isinstance(dev,Readout)])
    #Map of device name -> device
    devPointers = {}
    for dev in devs.values():
        devPointers[dev.__name__] = dev
    #Give each qubit device a pointer to its readout device
    #and give each readout device a pointer to its qubits
    for dev in devs.values():
        readoutConfig = dev.get('readoutConfig', None)
        if readoutConfig:
            dev['readoutDevice'] = devs[readoutConfig]
            devs[readoutConfig]['qubits'].append(dev)
    #Add readout channels to the readout devices
    for qubit_name in measure:
        qubit = devs[qubit_name]
        readoutDevice = qubit['readoutDevice']
        readoutDevice['channels'].append(('readout%s'%qubit_name, ('ADC', [readoutDevice['readoutADC']])))

    return devs
    
def loadQubits(Sample, measure, write_access=False):
    '''
    New version of loadQubits that combines loadQubits and
    registryWrapper2GateCompiler.  Also, now 'measure' uses the
    string-based names of the qubits, not the index in config.

    '''

    if isinstance(Sample, tuple):
        # This indicates a high level function is passing an already
        # loaded qubit object to a lower level function.  In that case
        # we copy the qubit object and return them.
        #
        # If you load the first time with write_access=False
        # you can't re-load with write_access=True
        new_sample = copy.deepcopy(Sample)
        if write_access:
            if len(new_sample) < 4:
                raise RuntimeError('Asked for write access, but not allowed by previous caller to loadQubits')
            return new_sample[0:4]
        else:
            return new_sample[0:3]
 
    Qubits = dict([(q, Sample[q]) for q in Sample['config']])  # dict of RegistryWrappers (writeable)
    sample = Sample.copy()                                     # AttrDict copy of registry
    qubits = dict([(q, sample[q]) for q in Sample['config']])  # dict of AttrDicts

    if isinstance(measure, (basestring, int)):
        measure = [measure]
        scalar_measure=True
    else:
        scalar_measure=False
        
    if any([isinstance(x, int) for x in measure]):
        raise RuntimeError('measure must be specified as strings, not integers!')

    #measure_idx = [ sample['config'].index(m) for m in measure ]
    for idx,m in enumerate(measure):
        qubits[m]['readout'] = True
        qubits[m]['readoutOrder'] = idx

    devs = registryWrapper2GateCompiler(sample, qubits, measure)
    q = [devs[x] for x in measure]
    Q = [Qubits[x] for x in measure]

    if scalar_measure:
        Q = Q[0]
        q = q[0]
    if write_access:
        return sample, devs, q, Q
    else:
        return sample, devs, q
##################
### ALGORITHMS ###
##################

class Algorithm(object):

    def __init__(self, agents=None):  # I removed the 'gates' option which probably doesn't work
        self.gates = [] # if gates is None else gates

        #  OK, here is the deal: if agents are specified in the constructor
        #  we make a local copy of them to prevent multiple-overwrite errors.
        #  we also set a flag. If that flag is set, adding a gate that references
        #  an unknown agent is an error.  That means that all gates have to be
        #  constructed by something like PiPulse([alg.q]).  If the agents aren't
        #  specified, we build the list of agents dynamically

        if agents:
            self.preset_agents = True
            self.agents = copy.deepcopy(agents)
            tmp_qubits = [agent for agent in self.agents.values() if agent.get('readout', False) ]
            self.qubits = [[]]*len(tmp_qubits) 
            for q in tmp_qubits:
                q_idx = q['readoutOrder']
                self.qubits[q_idx] = q
                setattr(self, 'q%d'%q_idx, q)
            self.q = self.qubits
        else:
            self.preset_agents = False
            self.agents = {}
        self.compiled = False # Flag to avoid double compiling errors
        #for ag in self.agents.values(): # preseed algorithms with empty start gate to make sure all devs get used
        #    self[gates.Wait([ag], 1*ns)]
        
    def __getitem__(self, gate):
        #For each agent involved in this gate, if we haven't seen this agent
        #before, add it to our list of agents.
        if self.compiled:
            raise RuntimeError("Adding a gate to a compiled algorithm")
        for agent in gate.agents:
            if agent not in self.agents.values():
                if self.preset_agents:
                    raise RuntimeError("""Attempted to add gate with an unknown agent.  Please
                                       construct gates using the copy stored in the Algorithm object""")
                else:
                    self.agents[agent.__name__] = agent
        #Add this gate to our list of gates.
        self.gates.append(gate)
        return self
        
    def compile(self,correctXtalkZ=False,config=None):

        """Compile the algorithm, updating all participating agents with relevant sequence data"""
        #Initialize all participating agents
        #correctCrosstalkZ = self.correctCrosstalkZ
            
        def printTimings(self):
            for agent in self.agents.values():
                if isinstance(agent, Transmon):
                    print "agent: ", agent.__name__
                    for idx, z in enumerate(agent['z_s']):
                        print "block %d z envelopes start, end: (%s, %s) " % (idx, z.start, z.end)
                    for idx, xy in enumerate(agent['xy_s']):
                        print "block %d xy envelopes start, end: (%s, %s) " % (idx, xy.start, xy.end)
                    print "readout start, end: (%s, %s)" % (agent['rr'].start, agent['rr'].end)
                    print ""
        
        if self.compiled:
            raise RuntimeError("Algorithm compiled twice.  You probably don't want that")
        self.compiled=True
        for agent in self.agents.values():
            agent['_t'] = 0.0*ns
            #set all envelopes to zero
            agent['xy'] = env.NOTHING
            agent['xy_s'] = []
            agent['z'] = env.NOTHING
            agent['z_s'] = []
            agent['rr'] = env.NOTHING
            agent['xy_phase'] = 0.0 # this is for compensating for z phase shifts. e.g. from a swap
            agent['compiled'] = True # fpgaseqTransmon checks this to avoid running if you forgot
                                    # to compile.
        #Compile the sequence
        for gate in self.gates[::-1]:
            gate.compile(self.agents) # these are the global agents not the gate specific agents     
        for agent in self.agents.values():
            agent['xy_s'] = [agent['xy']] + agent['xy_s']
            #print "agent %s has %d xy blocks" % (agent.__name__, len(agent['xy_s']))
            agent['z_s'] = [agent['z']] + agent['z_s']
            #print "agent %s has %d z blocks" % (agent.__name__, len(agent['z_s']))
        #Now we correct for the fact that we built the sequence backwards.
        #Find the longest sequence and shift all sequences to the right by that
        #amount.
                    
        #printTimings(self)
        tShift = -1*min([ag.get('_t', 0.0*ns)['ns'] for ag in self.agents.values()])*ns
        # tShift must be on 4ns intervals for FPGA start delays
        tShift = eh.fourNsIntervalFloor(tShift)
        #print "gate compiler tShift: ", tShift

        for agent in self.agents.values():
            if isinstance(agent, Transmon):
                agent['xy'] = env.shift(agent['xy'], dt=tShift)
                agent['z']  = env.shift(agent['z'],  dt=tShift)
                agent['rr'] = env.shift(agent['rr'], dt=tShift)# * np.exp(-2j*np.pi*df*tShift) don't need time shift for adc V7, demodulator doesn't free run through sequence. 
                agent['xy_s'] = [env.shift(x, dt=tShift) for x in agent['xy_s']]
                agent['z_s']  = [env.shift(x,  dt=tShift) for x in agent['z_s']]
                #Store the time of the end of coherence manipulations, ie. the time
                #at which the redaout pulse starts. Note that this is shifted to
                #the right by tShift.
                agent['_t'] = tShift
                if 'adcReadoutWindows' in agent:
                    for key in agent['adcReadoutWindows']:
                        window = agent['adcReadoutWindows'][key]
                        shiftedWindow = (window[0]+tShift, window[1]+tShift)
                        agent['adcReadoutWindows'][key] = shiftedWindow
                    # print "GC: %s has adcWindows %s" % (agent.__name__,agent['adcReadoutWindows'])
        # printTimings(self)
        if correctXtalkZ:
            #qubits need to be put in order for the Xtalk matrix

            if config is None:
                #alphabetical order: q0,q1,q2... etc
                names,qubits = zip( *sorted(self.agents.items()))
            else:
                #in order of the config
                #master = ['q3','q9','q8','q5']
                #target = ['q5','q8','q3']
                #indices = [master.index(element) for element in target]
                #i,targetnew = zip( *sorted(zip(indices,target)))
                #out: ['q3','q8','q5']
                
                qubits=[]
                names=[]
                for qname in config:
                    names.append(qname)
                    qubits.append(self.agents[qname])
            #print 'gateCompiler: order for Z Xtalk: ', names
                
            #correct for Z crosstalk
            qub=[]
            for q in qubits:
                if q['_type'] == 'transmon':
                    qub.append(q)
                
            eh.correctCrosstalkZ( qub )

def PrintGateEnvelopes(agents):
    idx=0
    for q in agents:
        print "agent %s has t_: %f" % (q['__name__'], q['_t'])
        for channel in ['xy_s', 'z_s']:
            for idx, e in enumerate(q[channel]):
                print "Envelope %d for agent %s channel %s has span (%f,%f)" % (idx, q['__name__'],channel, e.start, e.end)
##############
### EXAMPLE ##
##############
"""
from pyle.dataking import util
from pyle import gateCompiler as gc
sample, qubits = util.loadQubits(s)
q = qubits[0]
q['readout']=True
q = gc.registryWrapper2GateCompiler(sample, [q])[0]
#q = gc.AttrDict2Qubit(q, gc.Transmon)
alg = gc.Algorithm()
alg[gates.PiPulse([q])][gates.Wait([q], 10*ns)][gates.PiPulse([q])]
alg.compile()

t=np.linspace(-500,500,1001)
s=q['xy'](t)

plt.figure()
plt.plot(t,s)
plt.show()
"""



"""
import time
from pyle.dataking import util
from pyle import gateCompiler as gc

sample, qubits = util.loadQubits(s)
q = qubits[0]
q['readout']=True
q = gc.registryWrapper2GateCompiler(sample, [q])[0]
#q = gc.AttrDict2Qubit(q, gc.Transmon)

alg = gc.Algorithm()
t0a=np.linspace(0,1000,201)*ns
for t0 in t0a:
    alg[gates.PiPulse([q])]
alg.compile()

tic=time.time()
t=np.linspace(-500,500,1001)
s=q['xy'](t)
toc=time.time()
print toc-tic

alg.compile()

tic=time.time()
t=np.linspace(-500,500,1001)
s=q['xy'](t)
toc=time.time()
print toc-tic

plt.figure()
plt.plot(t,s)
plt.show()
"""
