import numpy as np
from pyle import envelopes as env
from pyle.dataking import envelopehelpers as eh
from labrad.units import Unit, Value
import matplotlib.pyplot as plt
import scipy
from pyle.dataking import zfuncs

ns, GHz, us, dBm = (Unit(un) for un in ['ns', 'GHz', 'us', 'dBm'])

#############
### GATES ###
#############

class Gate(object):

    def __init__(self, agents):
        self.agents = agents
        self.setSubgates(agents)
        for agent in agents:
            agent['gates'].append(self)
        self.name = self._name()
        
    def setSubgates(self, agents):
        """Define subgates for this gate
        
        This method may be overloaded
        """
        self.subgates = []
        
    def compile(self, globalAgents):
        """Recursively called method to build sequence
        """
        self.globalAgents = globalAgents
        self.updateAgents()
        for gate in self.subgates[::-1]:
            gate.compile(globalAgents)

    def updateAgents(self):
        """
        Gate specific agent mutations

        This method must be overloaded
        """
        raise Exception
    
    def _name(self):
        """This method should be overloaded"""
        return 'Gate'
        
    def __str__(self):
        return '%s on agents'%self.name +' ['+' '.join([agent for agent in self.agents]) + ']'
    
class PiPulse(Gate):
    def __init__(self, agents, alpha = None, phase = 0.0, freq=None, state = 1):
        Gate.__init__(self, agents)
        self.phase = phase
        self.freq  = freq
        self.state = state
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        l = ag['piLen']
        phase = self.phase
        freq  = self.freq
        state = self.state
        ag['xy'] += eh.mix(ag, eh.piPulse(ag, t-(l/2), phase = phase + ag['xy_phase'], state = state) , freq, state=state)
        ag['_t'] -= l

class PiHalfPulse(Gate):
    def __init__(self, agents, alpha = None, phase = 0.0, freq=None, state = 1, amp=1.0):
        Gate.__init__(self, agents)
        self.phase = phase
        self.freq  = freq
        self.state = state
        self.amp = amp
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        l = ag['piHalfLen']
        phase = self.phase
        freq  = self.freq
        state = self.state
        ag['xy'] += self.amp * eh.mix(ag, eh.piHalfPulse(ag, t-l/2, phase=phase + ag['xy_phase'], state=state) , freq, state=state)
        ag['_t'] -= l
   
class RotPulse(Gate):
    """
    User beware, for rotating to arbitrary angles the amplitude is not exactly linear with rotation theta.
    ONLY USE THIS FUNCTION FOR COARSE (read: low fidelity) GENERAL ROTATIONS
    """
    def __init__(self, agents, angle, phase = 0.0, alpha = None, state=1):
        Gate.__init__(self, agents)
        self.angle = angle
        self.alpha = alpha
        self.phase = phase
        self.state = state
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        l = ag['piLen']
        ag['xy'] += eh.mix(ag, eh.rotPulse(ag, t-(l/2), angle = self.angle, alpha = self.alpha, phase = self.phase + ag['xy_phase'], state=self.state), state = self.state)
        ag['_t'] -= l

class AmpPulse(Gate):
    """
    User beware, for rotating to arbitrary angles the amplitude is not exactly linear with rotation theta.
    ONLY USE THIS FUNCTION FOR COARSE (read: low fidelity) GENERAL ROTATIONS
    """
    def __init__(self, agents, amp, phase = 0.0, alpha = None, state=1):
        Gate.__init__(self, agents)
        self.amp = amp
        self.alpha = alpha
        self.phase = phase
        self.state = state
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        l = ag['piLen']
        ag['xy'] += eh.mix(ag, eh.piPulse(ag, t-(l/2), amp = self.amp, alpha = self.alpha, phase = self.phase + ag['xy_phase'], state=self.state), state = self.state)
        ag['_t'] -= l
        
class RabiDrive(Gate):

    def __init__(self, agents, amp, length, w = 10*ns, state=1):
        Gate.__init__(self, agents)
        self.amp = amp
        self.length = length
        self.w = w
        self.state = state
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        ag['xy'] += eh.mix(ag, env.flattop(t-self.length, self.length, self.w, self.amp), state = self.state)
        ag['_t'] -= self.length
        
class NPiPulses(Gate):
    def __init__(self, agents, N, alpha = None, state=1, phase=0):
        self.N = N
        self.alpha = alpha
        self.state = state
        self.phase = phase
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        self.subgates = (self.N) * [PiPulse(agents, alpha = self.alpha, state=self.state, phase=self.phase)]
        
class NPiHalfPulses(Gate):
    def __init__(self, agents, N, alpha = None, state=1, phase=0):
        self.N = N
        self.alpha = alpha
        self.state = state
        self.phase = phase
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        self.subgates = (self.N) * [PiHalfPulse(agents, alpha = self.alpha, state=self.state, phase=self.phase)]
    
class PiPulseZ(Gate):
    def __init__(self, agents):
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        l = ag['piLenZ']
        ag['z'] += eh.piPulseZ(ag, t-l/2.0)
        ag['_t'] -= l
        
class PiHalfPulseZ(Gate):
    def __init__(self, agents):
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        l = ag['piHalfLenZ']
        ag['z'] += eh.piHalfPulseZ(ag, t-l/2.0)
        ag['_t'] -= l
        
class NPiPulsesZ(Gate):
    def __init__(self, agents, N):
        self.N = N
        Gate.__init__(self, agents)          
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        self.subgates = (self.N) * [PiPulseZ(agents)]
        
class NPiHalfPulsesZ(Gate):
    def __init__(self, agents, N):
        self.N = N      
        Gate.__init__(self, agents) 
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        self.subgates = (self.N) * [PiHalfPulseZ(agents)]   
    
class MoveToState(Gate):
    def __init__(self, agents, initialState, endState, alpha = None):
        self.initialState = initialState
        self.endState = endState
        self.alpha = alpha
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        initialState = self.initialState
        endState = self.endState
        if initialState<endState:
            piPulses = range(initialState+1,endState+1)
        else:
            piPulses = range(initialState, endState, -1)
        self.subgates = [PiPulse(agents, alpha = self.alpha, state=state) for state in piPulses]

class RFHadamard(Gate):
    '''
    RF Hadamard gate.  This takes three separate pulses but avoids
    using the z-line
    '''
    def __init__(self, agents):
        Gate.__init__(self, agents)
    def updateAgents(self):
        pass
    def setSubgates(self, agents):
        self.subgates = [ RotPulse(agents, angle=np.pi/4, phase=np.pi/2),
                          PiPulse(agents),
                          RotPulse(agents, angle=-np.pi/4, phase=np.pi/2) ]

class PiPulseZ_RF(Gate):
    '''
    Z-axis pi pulse made with an X and Y pulse.
    '''
    def __init__(self, agents):
        Gate .__init__(self, agents)
    def updateAgents(self):
        pass
    def setSubgates(self, agents):
        self.subgates = [ PiPulse(agents, phase=0),
                          PiPulse(agents, phase=np.pi/2)]

class RFZPulse(Gate):
    '''
    Arbitrary Z pulse made with two pi-pulses
    '''
    def __init__(self, agents, angle):
        self.angle = angle
        Gate.__init__(self,agents)
    def updateAgents(self):
        pass
    def setSubgates(self, agents):
        self.subgates = [ PiPulse(agents, phase=0),
                          PiPulse(agents, phase=np.pi-self.angle/2) ]
    
class FastRFHadamard(Gate):
    '''
    RF hadamard with only two pulses
    '''
    def __init__(self, agents):
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
    def setSubgates(self, agents):
        self.subgates = [PiHalfPulse(agents, phase=np.pi/2), PiPulse(agents)] 
        #H = i * np.dot(X,np.dot(Y/2,I)) with X and Y/2 as implemented physically, i.e. X=[0 -i ; -i 0] etc
        #but here, we only apply X*Y/2, i.e. we apply exp(-1j*pi/2)*H. However, this is a single qubit phase, 
        #so we need to keep track of that by subtracting it from the reference frame with respect to other qubits.
        
                                  
class RFUnitary(Gate):
    '''
    Implements an arbitrary unitary gate with only microwave drive
    '''
    def __init__(self, agents, U):
        Gate.__init__(self, agents)
        self.U = U
    def updateAgents(self):
        pass
    def setSubgates(self,agents):
        U = self.U
        w, v = np.linalg.eig(U)
        phase_shift = np.angle(w[1]*w[0].conj())
        S = quantum.rho2bloch(v[:,0])
        latitude = np.arccos(S[3])
        longitude = np.atan2(S[2], S[1])
        self.subgates = [RotPulse(agents, angle=latitude, phase=longitude-pi/2),
                         RotPulse(agents, angle=phase_shift, phase=0),
                         RotPulse(agents, angle=-latitude, phase=longitude-pi/2)]

class FastRFUnitary(Gate):
    '''
    Implements an arbitrary unitary gate with only microwave drive
    This uses only two microwave pulses instead of three.  I think there
    is a closed form solution, but it is ugly, so I just do optimization
    '''
    def __init__(self, agents, U):
        Gate.__init__(self, agents)
        self.U = U
    def updateAgents(self):
        pass
    def setSubgates(self,agents):
        def fitfunc(x):
            U1 = quantum.rf_unitary(x[0], x[1])
            U2 = quantum.rf_unitary(x[2], x[3])
            Ux = np.dot(U2, U1)
            tmp  = np.dot(Ux, self.U.conj().T)
            tmp = tmp * tmp[0,0].conjugate() # remove global phase
            err = tmp - eye(2)
            return np.linalg.norm(err)
        xopt, fval, it, fcalls, warn = scipy.optimize.min(fitfunc, np.array([1.23, 0.32, -.23, .4]), full_output=True)
        self.subgates = [RotPulse(agents, angle=x[0], phase=x[1]),
                         RotPulse(agents, angle=x[2], phase=x[3])]
      
class PingPong(Gate):
    def __init__(self, agents, N, theta = 0.0):
        self.N = N
        self.theta = theta
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        N = self.N
        theta = self.theta
        self.subgates = N*[PiHalfPulse(agents, phase=theta,alpha = self.alpha), PiHalfPulse(agents, phase=theta+np.pi,alpha = self.alpha)]
        
class RandomizedBenchmarkingSingleQubitClifford(Gate):
    
    def __init__(self, agents, gateList, alpha = None):
        self.gateList = gateList
        #IMPORTANT: IGN is nothing, I & IW is idle with wait
        #identitLen = identitLen, or the shorted uw gate
        self.gateOps = {'I':lambda q: Wait([q], q.get( 'identityLen' , np.min( [q['piLen'], q['piHalfLen']] )*ns )  ),
                        'IW':lambda q: Wait([q], q['identityWaitLen']),
                        'IWSE':lambda q: EchoWait([q], q['identityWaitLen']),
                        'SE': lambda q: Echo([q], q['identityWaitLen']),
                        'IGN':lambda q: Wait([q], 0*ns),                        
                        'X':lambda q: PiPulse([q], alpha=alpha),
                        'Y':lambda q: PiPulse([q], alpha=alpha, phase=np.pi/2.),
                        'X/2':lambda q: PiHalfPulse([q], alpha=alpha),
                        'Y/2':lambda q: PiHalfPulse([q], alpha=alpha, phase=np.pi/2.),                        
                        '-X':lambda q: PiPulse([q], alpha=alpha, phase = np.pi),
                        '-Y':lambda q: PiPulse([q], alpha=alpha, phase=3*np.pi/2.),
                        '-X/2':lambda q: PiHalfPulse([q], alpha=alpha, phase = np.pi),
                        '-Y/2':lambda q: PiHalfPulse([q], alpha=alpha, phase=3*np.pi/2.),
                        'H': lambda q: FastRFHadamard([q]),
                        'Z': lambda q: Detune([q]),
                        'Zpi': lambda q: PiPulseZ([q]),           
                        'Zpi/2': lambda q: PiHalfPulseZ([q]),
                        'Zpi/4': lambda q: PiQuarterPulseZ([q]),  
                        '2T': lambda q: NPiQuarterPulsesZ([q],2),                          
                        'ZW': lambda q: Detune([q],wiggly=True),                           
                        'ACZShape': lambda q: adiabaticCZShape([q])}
        Gate.__init__(self, agents)          
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        gateList = self.gateList        
        ag = self.agents[0]        
        subgates = []
        for op in gateList:
            for sop in op[0]:
                subgates.append(self.gateOps[sop](ag))            
        self.subgates = subgates  
   
class RandomizedBenchmarkingMultiQubitClifford(Gate):
    """gateList is of the form: [ ('I','X') , ('I','Y/2') , ('CZ','CZ')  ... )  for 2 qubits
    alpha is a list of drag parameters, e.g. [0.5,0.5] for 2 qubits
    Code can do a single CZ on 2 qubits and other gates on other qubits
    First qubit with 'CZ' is the inject qubit, swapping into the bus. The adiabatic Z pulse is applied to the last qubit. RB
    
    CZR: only applies adiabatic waveform on retrieve.
    SWAP: swaps back and forth state from qubit to bus.    
    """
    def __init__(self, agents, gateList, alphaList = None,syncEachClifford=True,**kwargs):
        #prepare drag coefficients, so that alphas['q1'] = 
        alphas={}
        qubits = [agent for agent in agents if agent['_type']=='transmon']
        for idx,ag in enumerate(qubits):
            if alphaList:
                alphas[ag.__name__]=alphaList[idx]
            else:
                alphas[ag.__name__]=None  
        self.numQubits=len(agents)
        self.gateList = gateList
        self.syncEachClifford=syncEachClifford
        #IMPORTANT: IGN is nothing, I, IW is idle with wait
        self.gateOps = {'I':lambda q:  Wait([q], q.get( 'identityLen' , np.min( [q['piLen'], q['piHalfLen']] )*ns )  ),
                        'IW':lambda q: Wait([q], q['identityWaitLen']),
                        'IGN':lambda q: Wait([q], 0*ns),                        
                        'X':lambda q: PiPulse([q], alpha=alphas[q.__name__]),
                        'Y':lambda q: PiPulse([q], alpha=alphas[q.__name__], phase=np.pi/2.),
                        'X/2':lambda q: PiHalfPulse([q], alpha=alphas[q.__name__]),
                        'Y/2':lambda q: PiHalfPulse([q], alpha=alphas[q.__name__], phase=np.pi/2.),                        
                        '-X':lambda q: PiPulse([q], alpha=alphas[q.__name__], phase = np.pi),
                        '-Y':lambda q: PiPulse([q], alpha=alphas[q.__name__], phase=3*np.pi/2.),
                        '-X/2':lambda q: PiHalfPulse([q], alpha=alphas[q.__name__], phase = np.pi),
                        '-Y/2':lambda q: PiHalfPulse([q], alpha=alphas[q.__name__], phase=3*np.pi/2.),
                        'H': lambda q: FastRFHadamard([q]),
                        'Z': lambda q: Detune([q]),
                        'Zpi': lambda q: PiPulseZ([q]),           
                        'Zpi/2': lambda q: PiHalfPulseZ([q]),
                        'Zpi/4': lambda q: PiQuarterPulseZ([q]),                  
                        'CZ':lambda devs: adiabaticCZ(devs, **kwargs),
                        'CZphi': lambda devs: adiabaticCZphi(devs, **kwargs),             
                        'CNOT':lambda devs: adiabaticCNOT(devs, **kwargs),
                        'CZR':lambda devs: adiabaticCZ(devs, **kwargs), # Control Z trajectory with unloaded bus
                        'CZRL':lambda devs: adiabaticCZ(devs, **kwargs), # Control Z trajectory with loaded bus
                        'SWAP': lambda devs: ResonatorFullSwap(devs),
                        'sz': lambda devs: szsz(devs, **kwargs),
                        'trotterStep': lambda devs: trotterStepRB(devs, **kwargs)}
        #'CZ':lambda devs: adiabaticCZwBus(devs, **kwargs),
        Gate.__init__(self, agents)   
        
    def updateAgents(self):
        pass
        
    def gateListToOpList(self,gateListIn):
        '''
        This is per clifford!
        Turns  [ ['A','B','C' ..] , ['D','E','F'...]] into  [ [A,D],[B,E]...] etc
        Padds with IGNs to align CZ, and make it the sequences the same length
        '''        
        gateList=[];gateList.extend(gateListIn)
        if self.numQubits==1:
            #out=gateList[0]
            out=[]
            for el0 in gateList[0]:
                #print el0
                out.append(el0)        
        elif self.numQubits==2:
            numCZs=gateList[0].count('CZ')
            if numCZs!=gateList[1].count('CZ'):
                raise Exception('dissimilar number of two qubit gates')
            if numCZs>0:
                #use the first CZ to align, and pad IGN's around to get the same length
                idx0=gateList[0].index('CZ');idx1=gateList[1].index('CZ')
                if idx0<idx1:
                    gateList[0]=['IGN',]*(idx1-idx0)+gateList[0]
                elif idx0>idx1:
                    gateList[1]=['IGN',]*(idx0-idx1)+gateList[1]
            len0=len(gateList[0]);len1=len(gateList[1])
            if len0<len1:
                gateList[0]+=['IGN',]*(len1-len0)
            elif len0>len1:
                gateList[1]+=['IGN',]*(len0-len1)
            out=[]
            for el0,el1 in zip(gateList[0],gateList[1]):
                out.append([el0,el1])
        return out        
        
    def setSubgates(self, agents):
        qubits = [agent for agent in agents if agent['_type']=='transmon']
        resonators = [agent for agent in agents if agent['_type']=='resonator']
        gateList = self.gateList
        subgates = []
        twoQubitGateList=['CZ','CNOT','sz','CZphi','trotterStep']
        #for op in gateList:
        for clifford in gateList:
            cop = self.gateListToOpList(clifford)
            for op in cop:            
                #if type(op) is not tuple:
                #    op = (op,)
                #op=tuple(op)
                twoQubitGatePresent=any( [ twoQubitGatesElement in op for twoQubitGatesElement in twoQubitGateList] )
                if not twoQubitGatePresent:
                    for idx,ag in enumerate(qubits):
                        #print op,ag.__name__
                        #print op[idx]
                        if op[idx] in ['SWAP','CZR','CZRL']:
                            devs = [ag]
                            for res in resonators:
                                if res['coupled'].__contains__(devs[0].__name__):
                                    devs.append(res)
                            subgates.append(self.gateOps[ op[idx] ](devs))
                        else:
                            subgates.append(self.gateOps[ op[idx] ](ag))
                else:
                    #the code below allows for doing a CZ on 2 qubits and other gates on others
                    CZdevs=[]
                    for idx,sop in enumerate(op):
                        if sop in twoQubitGateList:
                            CZdevs.append(qubits[idx])
                            twoQubitGate=sop #this wont work if there is a CZ and CNOT
                        else:
                            subgates.append(self.gateOps[sop](qubits[idx]))
                    for res in resonators:
                        if res['coupled'].__contains__(CZdevs[0].__name__) and res['coupled'].__contains__(CZdevs[1].__name__):
                            CZdevs.append(res)
                    #supports one resonator to mediate a CZ between 2 qubits
                    #now we run the CZ gate once, and send qubits 0,1 and the resonator to the gate
                    subgates.append(self.gateOps[twoQubitGate](CZdevs))
            if self.syncEachClifford:
                subgates.append( Sync(agents) )
        self.subgates = subgates
 
   
class Tomography(Gate):
    def __init__(self, agents, tomo, alpha = None):
        # tomo is a list with one operation for each qubit
        # e.g. for one qubit: ['I'] OR ['X'] OR ['X/2']...
        # for two qubits: ['I','I] OR ['I','X'] OR ['X','I'] OR ['X','X']...
        #self.tomoOps = {'I':(0,0),'X':(np.pi, 0),
        #               'X/2':(np.pi/2.0,0), 'Y/2':(np.pi/2.0, np.pi/2.0),
        #               '-X/2':(-np.pi/2.0,0), '-Y/2':(-np.pi/2.0, np.pi/2.0)}
        self.tomoOps = {'I':lambda q: Wait([q], q['piLen']),
                        'X':lambda q: PiPulse([q], alpha=alpha),
                        'Y':lambda q: PiPulse([q], alpha=alpha, phase=np.pi/2.),
                        'X/2':lambda q: PiHalfPulse([q], alpha=alpha),
                        'Y/2':lambda q: PiHalfPulse([q], alpha=alpha, phase=np.pi/2.),                        
                        '-X':lambda q: PiPulse([q], alpha=alpha, phase = np.pi),
                        '-Y':lambda q: PiPulse([q], alpha=alpha, phase=3*np.pi/2.),
                        '-X/2':lambda q: PiHalfPulse([q], alpha=alpha, phase = np.pi),
                        '-Y/2':lambda q: PiHalfPulse([q], alpha=alpha, phase=3*np.pi/2.)}
        self.tomo = tomo
        self.alpha = alpha
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        self.subgates = [Sync(agents)]
        for agent,op in zip(agents, self.tomo):
            self.subgates += [self.tomoOps[op](agent)]
        
class Spectroscopy(Gate):
    def __init__(self, agents, df=0.0, z=None, zlonger=0.0*ns):
        self.df = df
        self.z=z
        self.zlonger=zlonger #this allows for starting the z pulse earlier, and ending it later than the spectroscopy pulse    
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        ag = self.agents[0]
        df=self.df
        z=self.z
        zlonger=self.zlonger
        ag['_t'] -= zlonger
        t = ag['_t']
        len=ag['spectroscopyLen']
        #ag['xy'] += eh.spectroscopyPulse(ag, t-len,df)
        ag['xy'] += env.mix(env.flattop(t-len, len , amp=ag['spectroscopyAmp'], w=10*ns), df)
        if z is not None:
            #ag['z']  += env.rect(t-len-zlonger,len+2*zlonger,z) #set a z amplitude
            if zlonger>0.0:
                ag['z']  += env.mix(env.flattop(t-len-zlonger, len+2*zlonger, w=zlonger/4, amp=z), 0) #this way it's a little smooth around the start and end
            else:
                ag['z']  += env.rect(t-len-zlonger,len+2*zlonger,z) #set a z amplitude
        ag['_t'] -= len
        ag['_t'] -= zlonger  

class SpectroscopyWithZonOther(Gate):
    def __init__(self, agents, df=None,z=None,zlonger=0.0*ns):
        self.df = df
        self.z=z
        self.zlonger=zlonger #this allows for starting the z pulse earlier, and ending it later than the spectroscopy pulse    
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        ag = self.agents[0]
        agForZ = self.agents[1]
        df=self.df
        z=self.z
        zlonger=self.zlonger
        ag['_t'] -= zlonger
        agForZ['_t'] -= zlonger        
        t = ag['_t']
        len=ag['spectroscopyLen']
        #ag['xy'] += eh.spectroscopyPulse(ag, t-len,df)
        ag['xy'] += env.mix(env.rect(t-len, len , amp=ag['spectroscopyAmp']), df)
        if z is not None:
            t2 = agForZ['_t']
            #ag['z']  += env.rect(t-len-zlonger,len+2*zlonger,z) #set a z amplitude
            if zlonger>0.0:
                agForZ['z']  += env.mix(env.flattop(t2-len-zlonger, len+2*zlonger, w=zlonger/4, amp=z), 0) #this way it's a little smooth around the start and end
            else:
                agForZ['z']  += env.rect(t2-len-zlonger,len+2*zlonger,z) #set a z amplitude
        ag['_t'] -= (len + zlonger)
        agForZ['_t'] -= (len + zlonger)
        
class Detune(Gate):
    def __init__(self, agents, length=None, w=None, amp=None, phase=None):
        self.length=length
        self.amp=amp
        self.w=w
        self.phase=phase
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        ag = self.agents[0]
        length = self.length
        amp = self.amp
        if self.phase is None:
            phase = ag.get('detunePhase',0.0)            
        elif self.phase is False:
            phase = 0.0
        else:
            phase=self.phase        
        if self.length is None:
            length = ag['detuneLen']
        else:
            length = self.length
            
        if self.amp is None:
            amp = ag['detuneAmp']
        else:
            amp = self.amp            
        if self.w is None:
            w = ag['detuneW']
        else:
            w = self.w            
        detunePadding = ag.get('detunePadding',0.0*ns)
        t = ag['_t']
        if w['ns']>0.0:
            ag['z'] += env.flattop(t-length-detunePadding, length, w, amp)
        else:
            ag['z'] += env.rect(t-length, length, amp)
        ag['_t'] -= length + 2*detunePadding
        ag['xy_phase'] -=  phase
        
class virtualZ(Gate):
    def __init__(self, agents, angle=0.0):
        raise Exception("Duplicate Function to ChangePhase")
        self.angle=angle
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        ag = self.agents[0]
        ag['xy_phase'] -=  self.angle
        
class RabiXTalk(Gate):
    def __init__(self, agents, amp, z=None, zlonger = 20*ns, phase = 0.0, state=1):
        Gate.__init__(self, agents)
        self.amp = amp
        self.phase = phase
        self.state = state
        self.z = z
        self.zlonger = zlonger
        
    def updateAgents(self):
        qmeas = self.agents[0]
        qdrive = self.agents[1]
        t = min(qmeas['_t'], qdrive['_t'])
        piLen = qdrive['piLen']
        totalLen = piLen + 2*self.zlonger
        qdrive['xy'] += eh.mix(qdrive, eh.piPulse(qdrive, t-(piLen/2) - self.zlonger, amp = self.amp, phase = self.phase + qdrive['xy_phase'], state=self.state), state = self.state)
        qmeas['z'] += env.rect(t-totalLen,totalLen,self.z) #set a z amplitude
        qdrive['_t'] -= totalLen
        if qdrive != qmeas:
            qmeas['_t'] -= totalLen
        
class adiabaticCZ(Gate):
    def __init__(self, agents, length=None, amp=None, w=None, z=None, padding=None, detuneOthers=True, detuneTarget=True, **kwargs):
        self.length=length
        self.amp=amp
        self.kwargs=kwargs
        self.w=w
        self.z=z
        self.padding=padding
        self.detuneTarget=detuneTarget
        self.detuneOthers=detuneOthers
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        q0 = self.agents[0] #control
        q1 = self.agents[1] #target
        if q0['_surf']=='measure' and q1['_surf']=='data':
            control=q0
            target=q1
        elif q0['_surf']=='data' and q1['_surf']=='measure':
            control=q1
            target=q0
        else:
            raise Exception('adiabaticCZ: cannot figure out control or target')
        globalAgents = self.globalAgents
        allAgents = [agent for agent in globalAgents.values() if agent['_type']=='transmon' ]
        # this t has to be min of anybody that is involved in the gate, including others in allAgents
        if self.detuneOthers:
            agentsInvolved = []
            for agent in allAgents:
                if agent.__name__ in [control.__name__, target.__name__]:
                    agentsInvolved.append(agent)
                else:
                    involved = allAgent_detuneFreq = agent.get( control.__name__ + target.__name__+'aczDetune' , None)
                    if involved:
                        agentsInvolved.append(agent)
            t = min([agent['_t'] for agent in agentsInvolved])
        else:
            t = min(control['_t'], target['_t'])
        
        #control parameters
        if self.length is None:
            length = control[target.__name__+'aczTime']
        else:
            length = self.length
        if self.amp is None:
            amp = control[target.__name__+'aczAmp']
        else:
            amp = self.amp
        if self.w is None:
            w=control[target.__name__+'aczW']
        else:
            w= self.w
        if self.padding is None:
            self.padding=control['aczPadding']
        if np.alen(self.padding) == 1:
            self.padding = [self.padding, self.padding] # padding before and after the pulse, default symmetric padding if not specified
        thi,thf,lam2,lam3 = control[target.__name__+'aczLambda']
        if control['aczFreqFunc']:
            ftozparams = tuple(control['zFuncAna'])
            ztofparams = tuple(control['zFuncAna'])
        else:
            ftozparams=()
            ztofparams=()
        zOffsetDC = control.get('zOffsetDC',0.0)

        #target parameters
        target_detuneFreq = target.get( control.__name__ +'aczDetune' , None)  
        w_fordetune = target['aczDetuneWidth'] #also used for detuning parasitic qubits
        target_padding = target['aczDetunePadding'] #also used for detuning parasitic qubits     
        if target_detuneFreq and self.detuneTarget:
            target_detuneAmp = zfuncs.FrequencyToAmp(target)(target_detuneFreq)
            
        control['z'] += env.adiabaticCZ(t-length-self.padding[1]-target_padding, length, amplitude=amp,w=w,
                                   thi=thi,thf=thf,lam2=lam2,lam3=lam3, ftozparams=ftozparams,ztofparams=ztofparams,zOffsetDC=zOffsetDC,
                                   **self.kwargs)
                                   
        #apply meet up detune to target qubits:
        if target_detuneFreq and self.detuneTarget:
            target['z'] += env.flattop(t-length - 2*target_padding-self.padding[1], 2*target_padding+length, w_fordetune, target_detuneAmp) 
        #timings:
        control['_t'] = t-length-self.padding[0] - self.padding[1] - 2.*target_padding
        target['_t']  = t-length-self.padding[0] - self.padding[1] - 2.*target_padding  
        #phases:
        control['xy_phase'] -= control[target.__name__+'aczPhase']
        target['xy_phase'] -= target[control.__name__+'aczPhase']
            
        #apply steer away detunes to idle qubits:
        for allAgent in allAgents:
            allAgent_detuneFreq = allAgent.get( control.__name__ + target.__name__+'aczDetune' , None)
            if allAgent_detuneFreq and self.detuneOthers:
                allAgent_detuneAmp = zfuncs.FrequencyToAmp(allAgent)(allAgent_detuneFreq)
                #allAgent_padding = allAgent[ control.__name__ + target.__name__+'aczDetunePadding' ] # 2.0 ns
                #w_fordetune = allAgent[ control.__name__ + target.__name__+'aczDetuneWidth' ] # 1.0 ns
                allAgent_phase = allAgent[ control.__name__ + target.__name__+'aczDetunePhase' ] 
                allAgent['z'] += env.flattop(t-length - 2*target_padding-self.padding[1], 2*target_padding+length, w_fordetune, allAgent_detuneAmp)
                allAgent['_t'] = t-length-self.padding[0] - self.padding[1] - 2.*target_padding
                allAgent['xy_phase'] -= allAgent_phase

class simultaneousAdiabaticCZ(Gate):
    def __init__(self, agents):
        """
        agents = [[pair0Control, pair0Target], [pair1Control, pair1Target]]
        agents = [control0, target0, control1, target1, mutualParasitic]
        First pair runs detuneOthers = True, second pair detuneOthers = False.
        Only works for this configuration:
        Q0 <-> Q1 (Q2) Q3 <-> Q4
        """
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        agents = self.agents
        controlQubits = [agents[0], agents[2]]
        targetQubits = [agents[1], agents[3]]
        parasiticQubit = agents[4]
                        
        # Check that gate lengths are the same for control and target qubits
        lengths = []
        ws = []
        padding = []
        for cq, tq in zip(controlQubits, targetQubits):
            lengths.append(cq[tq.__name__+'aczTime'])
            ws.append(cq[tq.__name__+'aczW'])
            padding.append(cq['aczPadding'])
        if not lengths.count(lengths[0]) == len(lengths):
            raise Exception("Incompatible aczTime")
        if not ws.count(ws[0]) == len(ws):
            raise Exception("Incompatible aczW")
        if not padding.count(padding[0]) == len(padding):
            raise Exception("Incompatible aczPadding")
            
        # add gates
        self.subgates = [Sync(controlQubits + targetQubits + [parasiticQubit])]
        self.subgates += [adiabaticCZ([controlQubits[0], targetQubits[0]], detuneOthers=True)]
        self.subgates += [adiabaticCZ([controlQubits[1], targetQubits[1]], detuneOthers=False)]
        self.subgates += [Sync(controlQubits + targetQubits + [parasiticQubit])]
                
class adiabaticSWAP(Gate):
    def __init__(self, agents, targetName=None, holdTime = 0.0*ns, holdTimeTiming = True, length=None, deltaf1=None, w=None, padding=None, **kwargs):
        self.holdTime=holdTime
        self.holdTimeTiming=holdTimeTiming
        self.length=length
        self.deltaf1=deltaf1
        self.targetName=targetName
        self.kwargs=kwargs
        self.w=w
        self.padding=padding
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        control = self.agents[0] #control
        if self.targetName:
            targetName = self.targetName
            t = control['_t']
        else:
            target = self.agents[1] #target
            targetName = target.__name__
            t = min(control['_t'], target['_t'])
        if self.length is None:
            length = control[targetName+'aSwapTime']
        else:
            length = self.length
        if self.deltaf1 is None:
            deltaf1 = control[targetName+'aSwapAmp']
        else:
            deltaf1 = self.deltaf1
        if self.w is None:
            w=control[targetName+'aSwapW']
        else:
            w= self.w
        if self.padding is None:
            self.padding=control.get('aSwapPadding',0.0)  
        if np.alen(self.padding) == 1:
            self.padding = [self.padding, self.padding] # padding before and after the pulse, default symmetric padding if not specified
        usef1func,thi,thf,lam2,lam3,lam4 = control[targetName+'aSwapLambda']
        if int(usef1func):
            ftozparams = tuple(control.get('zFuncAna',None))
            ztofparams = tuple(control.get('zFuncAna',None))
        else:
            ftozparams=()
            ztofparams=()
        zOffsetDC = control.get('zOffsetDC',0.0)
        if self.holdTimeTiming:
            startTime = t-length-w-self.padding[1] - self.holdTime
            totalLength = length+2*w+self.padding[0]+self.padding[1]+self.holdTime
        else:
            startTime = t-length-w-self.padding[1]
            totalLength = length+2*w+self.padding[0]+self.padding[1]
        control['z'] += env.adiabaticSWAP(startTime, length, Deltaf1=deltaf1, holdTime = self.holdTime, w=w,
                                   thi=thi,thf=thf,lam2=lam2,lam3=lam3,lam4=lam4, 
                                   ftozparams=ftozparams,ztofparams=ztofparams,zOffsetDC=zOffsetDC,
                                   **self.kwargs)
       
        control['_t'] -= totalLength
        
class adiabaticCNOT(Gate):
    """
    CNOT = (Y_-pi/2 on target) - CZ - (Y_+pi/2 on target)
    agents: [controlQubit, targetQubit] note that this is the same order as for adiabaticCZ, but different from adiabaticCZwBus
    RB.
    """
    def __init__(self, agents, length=None, amp=None, syncQubits = True, w=None, **kwargs):
        self.length=length
        self.amp=amp
        self.w=w
        self.kwargs=kwargs
        self.syncQubits = syncQubits
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        controlQubit = self.agents[0] #control/data, also has adiabatic CZ pulse applied on its Z line    
        targetQubit = self.agents[1] #target/measure
        self.subgates = [PiHalfPulse(agents, phase=3*np.pi/2.),
                        Sync([targetQubit, controlQubit]),                        
                        adiabaticCZ([controlQubit, targetQubit], length=self.length, amp = self.amp, w=self.w, **self.kwargs) ,                         
                        Sync([targetQubit, controlQubit]),
                        PiHalfPulse(agents, phase=np.pi/2.)]
        if self.syncQubits:
            self.subgates = [Sync([targetQubit, controlQubit])] + self.subgates + [Sync([targetQubit, controlQubit])]

      
            
class Pulseshape(Gate):
    def __init__(self, agents, offset=None, offsetLen=None, stepHeight=None, stepLen=None, probeTime=None):
        assert probeTime<stepLen
        self.offset = offset
        self.offsetLen = offsetLen
        self.stepHeight = stepHeight
        self.stepLen = stepLen
        self.probeTime = probeTime
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        #Rename variables
        ag = self.agents[0]
        offsetLen = self.offsetLen
        offset = self.offset
        stepLen = self.stepLen
        stepHeight = self.stepHeight
        #Square z pulses
        totalLen = offsetLen + stepLen
        tStart = ag['_t'] - totalLen
        ag['z'] += env.rect(tStart, offsetLen, -1*offset)
        ag['z'] += env.rect(tStart + offsetLen, stepLen, stepHeight-offset)
        #probe pulse
        probeTime = tStart + offsetLen + self.probeTime
        ag['xy'] += eh.mix(ag, eh.rotPulse(ag, probeTime, angle = np.pi ))
        #Update agent timer
        ag['_t'] -= totalLen

class ResonatorSqrtSwap(Gate):
    def __init__(self, agents, length=None, w=None, amp=None):
        self.length=length
        self.amp=amp
        self.w=w
        Gate.__init__(self, agents)

    def updateAgents(self):
        qubit = self.agents[0]
        resonator = self.agents[1]
        length = self.length
        if length is None:
            keyname = qubit.__name__+'sqrtSwapTime'
            length = resonator[keyname]
        amp = self.amp
        if amp is None:
            keyname = qubit.__name__+'sqrtSwapAmp'
            amp = resonator[keyname]
        w = self.w
        if w is None:
            keyname = qubit.__name__+'swapW'
            w = resonator[keyname]
        t = min(qubit['_t'], resonator['_t'])
        qubit['z'] += env.flattop(t-length, length, w, amp)
        qubit['_t'] = t-length
        resonator['_t'] = t-length
        qubit['xy_phase'] -= resonator[qubit.__name__+'sqrtSwapPhase']        
        
class ResonatorSwap(Gate):
    def __init__(self, agents, length=None, w=None, amp=None):
        self.length=length
        self.amp=amp
        self.w=w
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        qubit = self.agents[0]
        resonator = self.agents[1]
        length = self.length
        if length is None:
            keyname = qubit.__name__+'swapTime'
            length = resonator[keyname]
        amp = self.amp
        if amp is None:
            keyname = qubit.__name__+'swapAmp'
            amp = resonator[keyname]
        w = self.w
        if w is None:
            keyname = qubit.__name__+'swapW'
            w = resonator[keyname]
        t = min(qubit['_t'], resonator['_t'])
        qubit['z'] += env.flattop(t-length, length, w, amp)
        qubit['_t'] = t-length
        resonator['_t'] = t-length 
        qubit['xy_phase'] -= resonator[qubit.__name__+'swapPhase']

class ResonatorFullSwap(Gate):
    """There and back again. A qubit's tale.
    Swaps state from qubit to resonator and back."""
    def __init__(self, agents):
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        injectQubit = self.agents[0]
        res = self.agents[1]
        #self.subgates = [ResonatorSwap([injectQubit, res]) , Wait([injectQubit],res[injectQubit.__name__+'swapTime'] ), ResonatorSwap([injectQubit, res])]
        self.subgates = [ResonatorSwap([injectQubit, res]) , Wait([injectQubit],injectQubit['piLen'] ), ResonatorSwap([injectQubit, res])]
        
class testdelayBusSwap(Gate):
    def __init__(self, agents, startTime = 0*ns):
        Gate.__init__(self, agents)
        self.startTime = startTime
        
    def updateAgents(self):
        startTime = self.startTime
    
        qubit0 = self.agents[0]
        qubit1 = self.agents[1]
        resonator = self.agents[2]
        
        length0 = resonator[qubit0.__name__+'swapTime']
        amp0 = resonator[qubit0.__name__+'swapAmp']
        w0 = resonator[qubit0.__name__+'swapW']
        
        length1 = resonator[qubit1.__name__+'swapTime']
        amp1 = resonator[qubit1.__name__+'swapAmp']
        w1 = resonator[qubit1.__name__+'swapW']
        
        # QUBIT 1 recovers photon from resonator, this comes first because we build sequences backwards!
        t = min(qubit0['_t'], qubit1['_t'], resonator['_t'])
        qubit1['z'] += env.flattop(t-length1, length1, w1, amp1)
        qubit0['_t'] = t-length1-startTime
        qubit1['_t'] = t-length1-startTime
        resonator['_t'] = t-length1-startTime
        
        # QUBIT 0 swaps photon into the resonator for recovery by QUBIT 1
        t = min(qubit0['_t'], qubit1['_t'], resonator['_t'])
        qubit0['z'] += env.flattop(t-length0, length0, w0, amp0)
        qubit0['_t'] = t-length0-length1-2*w1
        qubit1['_t'] = t-length0-length1-2*w1
        resonator['_t'] = t-length0-length1

class testdelaySwap(Gate):
    """Direct qubit-qubit swap. RB."""
    def __init__(self, agents, params ,startTime = 0*ns):
        Gate.__init__(self, agents)
        self.startTime = startTime
        self.params=params
    def updateAgents(self):
        startTime = self.startTime
        qubit0 = self.agents[0]
        qubit1 = self.agents[1]        
        [length,w,amp0,amp1]=self.params        
        t = min(qubit0['_t'], qubit1['_t'])
        t0 = t - abs(startTime) - length #we'll center the master Z pulse around this
        qubit0 = self.agents[0]
        qubit1 = self.agents[1]        
        qubit0['z'] += env.flattop(t0-(length)/2., length, w, amp0)
        qubit1['z'] += env.flattop(t0-startTime - (length)/2., length, w, amp1)
        qubit0['_t'] = t0 - abs(startTime) - length
        qubit1['_t'] = t0 - abs(startTime) - length
    
class TestDelayZ(Gate):
    def __init__(self, agents, tshift=0.0,zpa=0.0,zpl=0.0):
        self.tshift=tshift
        self.zpa=zpa
        self.zpl=zpl        
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        """Modify this qubit's time and add the pulse data"""
        ag = self.agents[0]
        tshift=self.tshift
        zpa=self.zpa
        zpl=self.zpl
        t = ag['_t']
        l = ag['piLen']        
        ag['xy'] += eh.mix(ag, eh.piPulse(ag, t-(l/2)))
        ag['z'] += env.rect( t - l/2.0 - zpl/2.0  + tshift, zpl, zpa) #since the pi pulse is centered at t-l/2, we need to center this rectangular window around t-l/2 too
        ag['_t'] -= l
        
class TestDelayRR(Gate):
    def __init__(self, agents, tshift=0.0,rra=-40*dBm):
        self.tshift=tshift
        self.rra=rra        
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        """Modify this qubit's time and add the pulse data"""
        ag = self.agents[0]
        tshift=self.tshift
        rra=self.rra
        rrl=ag['readoutWidth']
        t = ag['_t']
        l = ag['piLen']        
        ag['xy'] += eh.mix(ag, eh.piPulse(ag, t-(l/2)))
        ag['rr'] += eh.readoutRingdownPulse(ag, t - l/2.0 - rrl + tshift, amps = (rra,-100.0*dBm), phases = (0,-np.pi))
        #ag['z'] += env.rect( t - l/2.0 - zpl/2.0  + tshift, zpl, zpa) #since the pi pulse is centered at t-l/2, we need to center this rectangular window around t-l/2 too
        ag['_t'] -= l
    
class ChangePhase(Gate):
    def __init__(self, agents, xyPhase = 0.0 ,**kwargs):
        self.xyPhase = xyPhase
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        ag = self.agents[0]
        ag['xy_phase'] -= self.xyPhase
    
class Sync(Gate):
    def __init__(self, agents):
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        t = min([agent['_t'] for agent in self.agents])
        for agent in self.agents:
            agent['_t'] = t
    
class Wait(Gate):
    def __init__(self, agents, waitTime):
        self.waitTime = waitTime
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        for ag in self.agents:
            waitTime=self.waitTime
            t = ag['_t']
            ag['xy'] += env.wait(t-waitTime, waitTime) # to get the gate windowing
            ag['z'] += env.wait(t-waitTime, waitTime)
            ag['_t'] -= self.waitTime

class EchoWait(Gate):
    def __init__(self, agents, waitTime, fringeFreq = None):
        self.waitTime = waitTime
        self.fringeFreq = fringeFreq
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass

    def setSubgates(self, agents):
        ag = self.agents[0]
        piLen = ag['piLen']
        waitTime=self.waitTime
        dt = (waitTime - 2*piLen)/4.
        if self.fringeFreq:
            phase1 = np.pi*(self.fringeFreq['GHz']*self.waitTime['ns'])/2.0
            phase2 = 3*np.pi*(self.fringeFreq['GHz']*self.waitTime['ns'])/2.0
        else:
            phase1 = 0.0
            phase2 = 0.0
        if dt<0*ns:
            raise Exception("Echo dt < 0, overlapping pi pulses")
        self.subgates = [Wait([ag], dt)] + [PiPulse([ag], phase=phase1)] + [Wait([ag], 2*dt)] + [PiPulse([ag], phase=phase2)] + [Wait([ag], dt)]
        
class Echo(Gate):
    def __init__(self, agents, waitTime):
        self.waitTime = waitTime
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass

    def setSubgates(self, agents):
        ag = self.agents[0]
        piLen = ag['piLen']
        waitTime=self.waitTime
        dt = (waitTime - piLen)/2.
        if dt<0*ns:
            raise Exception("Echo dt < 0")
        self.subgates = [Wait([ag], dt)] + [PiPulse([ag])] + [Wait([ag], dt)]
            
class EmptyWait(Gate):
    def __init__(self, agents, waitTime):
        self.waitTime = waitTime
        Gate.__init__(self, agents)
    def updateAgents(self):
        ag = self.agents[0]
        ag['_t'] -= self.waitTime

class SetTime(Gate):
    def __init__(self, agents, t):
        self.t = t
        Gate.__init__(self, agents)
    def updateAgents(self):
        ag = self.agents[0]
        ag['_t'] = self.t

class DualBlockWaitDetune(Gate):
    '''Detunes for a period of time using the dual-block delay
    '''
    def __init__(self, agents, detune_agent=None, length=0.0, w=0.0, amp=0.0, cache=False):
        # reformat agents dict to list for legacy reasons
        agents = [agents[key] for key in agents.keys()]
        self.length=length
        self.amp=amp
        self.w=w
        self.detune_agent = detune_agent or agents[0]
        self.cache = cache
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        length = self.length
        amp = self.amp
        w = self.w

        # Sync up all agents
        t = min([agent['_t'] for agent in self.agents])
        for agent in self.agents:
            agent['_t'] = t
        # This bit is tricky.  The envelope of the detuning pulse needs to be added
        # to both blocks, but the support of the envelope needs to be truncated at
        # the desired SRAM boundaries.
        #print "DualBlockWaitDetune: length: %s, incoming _t: %s" % (length, t)
        # Add the end of the detune pulse to the current envelope

        if w>0*ns:
            env1 = env.flattop(t-length, length, w, amp)
        else:
            env1 = env.rect(t-length,length,amp)
        env1.start = env1.end - 100 #keep this short
        env1.end += 200 #add 100 ns so the deconv doesn't mess this up
        #print "DualBlockWaitDetune: env1.start: %s, env1.end: %s" % (env1.start, env1.end)
        self.detune_agent['z'] += env1
        #self.detune_agent['_t'] -= 4*w

        # Rebuild the envelope lists (this should be the same for all dual-block gates
        for agent in self.agents:
            #print "enabling dual block for agent: %s" % (agent.__name__)
            agent['dualblock'] = True
            agent['_t'] = t-self.length-2.0*w #all agents need to be shifted with the length of the gate: length+2*w
            agent['xy_s'] = [agent['xy']] + agent['xy_s']
            agent['xy'] = env.EnvZero(agent['_t'],agent['_t'])
            agent['z_s'] = [agent['z']] + agent['z_s']
            agent['z'] = env.EnvZero(agent['_t'], agent['_t'])
        # Add the beginning of the dual block envelope to the newly created 'z' envelope
        if w>0*ns:        
            env2 = env.flattop(t-length, length, w, amp)
        else:
            env2 = env.rect(t-length,length,amp)            
        env2.end = env2.start + 340.0
        #print "DualBlockWaitDetune: env2.start: %s, env2.end: %s" % (env2.start, env2.end)
        self.detune_agent['z'] += env2
        #print "DualBlockWaitDetune: z_s[0].start: %s, z_s[0].end: %s" % (self.detune_agent['z_s'][0].start, self.detune_agent['z_s'][0].end)
        #print "DualBlockWaitDetune: z.start: %s, z.end: %s" % (self.detune_agent['z'].start, self.detune_agent['z'].end)
        
        #self.detune_agent['_t'] -= 4*w

class Measure(Gate):
    def __init__(self, agents, sync='start', name = 'DefaultMeasure'):
        self.sync=sync
        self.name=name
        Gate.__init__(self, agents)
        
    def updateAgents(self):    
        pass
        
    def setSubgates(self, agents):
        self.subgates=[]
        if self.sync is 'start':
            # we have to manually move back the faster readout pulses
            gateLens = []
            self.subgates += [Readout([ag], name = self.name) for ag in agents] 
            for ag in agents:
                readoutPulseLen = ag['readoutPulseLen']
                readoutRingdownLength = ag['readoutRingdownLength']
                totalReadoutTime = readoutPulseLen + readoutRingdownLength
                gateLens.append(totalReadoutTime)
            longestGate = max(gateLens)
            for ag,totalReadoutTime in zip(agents, gateLens):
                self.subgates += [EmptyWait([ag], -totalReadoutTime+longestGate)]
            self.subgates += [Sync(agents)]
        elif self.sync is 'end':
            self.subgates += [Readout([ag],name = self.name) for ag in agents] + [Sync(agents)] # we sync all the qubits at the end
        
class Readout(Gate):
    def __init__(self, agents, name='DefaultReadout', ringdown=None, power=None):
        Gate.__init__(self, agents)
        self.name=str(name)
        self.ringdown=ringdown
        self.power=power
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        readoutPulseLen = ag['readoutPulseLen']
        readoutRingdownLength = ag['readoutRingdownLength']
        totalReadoutTime = readoutPulseLen + readoutRingdownLength
        
        pulseDemodStartTime = t-totalReadoutTime
        pulseDemodStartTime = eh.fourNsIntervalCeil(pulseDemodStartTime)
        
        readoutDemodLen = ag['readoutDemodLen']
        
        adcWindowStart = pulseDemodStartTime
        adcWindowStop = adcWindowStart + readoutDemodLen

        # if self.power is None, eh.readoutPulse takes ag['readoutPulsePower']
        ag['rr'] += env.shift(eh.readoutPulse(ag, 0, self.power), dt=(pulseDemodStartTime)) # zero phase at t=0, shift w/o compensation adc v7
        ag['_t'] = pulseDemodStartTime
        
        if ag['adcReadoutWindows'].has_key(self.name):
            raise Exception("Trying to apply multiple readout windows for one qubit on a single ADC retrigger!")
        ag['adcReadoutWindows'][self.name] = (adcWindowStart, adcWindowStop)
        '''
        What is a adcReadoutWindow wrt to version 7?
        This is essentially a placeholder for the mixer table.
        The mixer table the length of the adcReadoutWindow, and is zero padded
        to make the entire mixer table 1 us. This way we can have an adcReadoutWindow < 1us
        and we can make it wrt the adc retrigger.

        The name value helps us group different measurements into adc retriggers.
        '''

        if self.ringdown is None:
            ringdown = ag.get('ringdown',False)
        else:
            ringdown = self.ringdown
        if ringdown:
            t = ag['_t']
            ag['rr'] += env.shift(eh.readoutRingdownPulse(ag, 0), dt=(pulseDemodStartTime+readoutPulseLen))
  
class ReadoutRingdown(Gate):
    def __init__(self, agents, w=None, amps=None, phases=None, moveBack=None):
        Gate.__init__(self, agents)
        self.w=w
        self.amps=amps
        self.phases=phases
        self.moveBack=moveBack
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        if self.w is None:
            w = ag['readoutWidth']
        else:
            w = self.w
        if self.amps is None:
            amps = tuple(ag['ringdownPulseAmps'])
        else:
            amps = self.amps
        if self.phases is None:
            phases = tuple(ag['ringdownPulsePhases'])
        else:
            phases = self.phases
        if self.moveBack is None:
            moveBack = ag['ringdownPulseMoveBack']
        else:
            moveBack = self.moveBack

        ag['rr'] += eh.readoutRingdownPulse(ag, t-2*w, amps = amps, phases = phases)
        ag['_t'] -= 2*w-moveBack
        
class MeasureState(Gate):
    def __init__(self, agents, state = 2, alpha = None):
        self.state = state
        self.alpha = alpha
        Gate.__init__(self, agents)
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        alpha = self.alpha
        state = self.state
        self.subgates = [MoveToState([ag], state-1, 0, alpha) for ag in agents] + [Measure([ag]) for ag in agents]

class ACStark(Gate):
    def __init__(self, agents, ringUp = 2000*ns, buffer=100*ns, amp=None):
        Gate.__init__(self, agents)
        self.ringUp = ringUp
        self.buffer = buffer
        self.amp = amp
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        piLen = ag['piLen']
        readoutWidth = ag['readoutWidth']
        totalLength = self.ringUp+piLen+self.buffer
        ag['readoutAmp'] = eh.power2amp(ag['readoutPower'])
        carrierFrequency = ag['readoutDevice']['carrierFrequency']
        df = ag['readoutFrequency'] - carrierFrequency
        ag['rr'] += env.mix(env.flattop(t-totalLength, self.ringUp+piLen+self.buffer, w=readoutWidth, amp=self.amp),df)
        ag['xy'] += eh.mix(ag, eh.piPulse(ag, t-(piLen/2)-self.buffer))
        ag['_t'] -= totalLength
        
class ReadoutSpectroscopy(Gate):
    def __init__(self, agents, delay = 0.0*ns,freq=None):
        Gate.__init__(self, agents)
        self.delay = delay
        self.t=t
        self.z=z
        self.freq=freq
        
    def updateAgents(self):
        ag = self.agents[0]
        t = ag['_t']
        readoutPulseLen = ag['readoutPulseLen']
        readoutRingdownLength = ag.get('readoutRingdownLength', 0.0*ns)
        totalReadoutTime = readoutPulseLen + readoutRingdownLength
        
        pulseDemodStartTime = t-totalReadoutTime
        pulseDemodStartTime = eh.fourNsIntervalCeil(pulseDemodStartTime)

        # if self.power is None, eh.readoutPulse takes ag['readoutPulsePower']
        ag['rr'] += env.shift(eh.readoutPulse(ag, 0), dt=(pulseDemodStartTime)) # zero phase at t=0, shift w/o compensation adc v7
        ag['_t'] = pulseDemodStartTime
        
        ag['xy'] += eh.mix(ag, eh.piPulse(ag, pulseDemodStartTime-(ag['piLen']/2)+self.delay), self.freq)
        