import os
import os.path
import itertools

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
import scipy.interpolate
from scipy.optimize import curve_fit
import numpy as np

from labrad.units import Value
from pyle.plotting import dstools
from pyle.fitting import fitting
from pyle import gateCompiler as gc
from pyle import envelopes as env
from pyle.dataking import adjust as adjust

#################################################
### High level functions to get and plot data ###
#################################################
    
def plotSwapFreq(cxn,datasetNumber=None,session=None,dir=None,p=None):
    dv=cxn.data_vault
    if session is None and dir is None:
        raise Exception('You must specify either the session or the directory with the swap spectroscopy.')
    if session is not None and dir is not None:
        raise Exception('You can only specify one of session or dir.')
    if session is not None:
        dir = session._dir
    if datasetNumber is None and session is None:
        raise Exception('You must specify a datasetNumber if providing the directory and not the session.')
    if datasetNumber is None:
        dv.cd(dir)
        datasetNumber=int(dv.dir()[1][-1][:5])
    dataset=dstools.getSwapSpectroscopy(dv,datasetNumber,dir,p)
    dstools.plotSwapSpectro(dataset)
    
def plot1DScans(dataset, legendLocation='upper right', colors=None, title='', **kw):
    """Plot a 1D scan with an arbitrary number of curves
    
    **kw: keyword dictionary to be passed to plotting functions
    """
    if colors is None:
        colors = itertools.cycle(['b','r','g','k'])
    plt.rcParams['font.size']=24
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    indepLabel, indepUnit = dataset.variables[0][0]
    indepData = dataset.data[:,0]
    nIndeps = dataset.data.shape[1]-1
    indepLabel, indepUnit = dataset.indep[0]
    indepData = dataset[:,0]
    nIndeps = dataset.shape[1]-1
    for i in range(nIndeps):
        for i in range(nIndeps):
            depData = dataset.data[:,i+1]
            depLabel,depLegend,depUnit = dataset.variables[1][i]
            depData = dataset[:,i+1]
            depLabel,depLegend,depUnit = dataset.dep[0]
            ax.plot(indepData, depData,colors.next()+'-', label=depLegend, **kw)
    plt.legend(loc=legendLocation)
    depStr = depLabel+' [%s]'%depUnit if depUnit else depLabel
    indepStr = indepLabel+' [%s]'%indepUnit if indepUnit else indepLabel
    plt.xlabel(indepStr)
    plt.ylabel(depStr)
    plt.title(title+': '+str(dataset.path))
    fig.show()
    
def plot2DDataset(data, aspect=None):
    """
    
    extent - tuple: (left, right, bottom, top)
    
    Assumes data was acquired column by column as is done in pyle as of this
    writing (Sept 2013)
    """
    xMin, xMax = np.min(data[:,0]), np.max(data[:,0])
    yMin, yMax = np.min(data[:,1]), np.max(data[:,1])
    #How many rows and columns do we have?
    xVals = set([x for x in data[:,0]])
    yVals = set([y for y in data[:,1]])
    #Reshape data to look like it does in the Grapher
    z = data[:,2]
    z = np.reshape(z, (len(xVals), len(yVals))).T
    z = z[::-1,:]
    im = plt.matshow(z, aspect=aspect)
    plt.colorbar(im, fraction=0.025)
    #return fig, ax
    
def pulses(alg, t = None, timeStep = 0.5, spacing = -2.0, lw = 1, legend=True, spacingByQubit=True, mixerTable=None, triggerTable=None):

    # TODO: Figure out how to rescale xyz so that we can look at them all
    #       at the same time.

    qubits = [q for q in alg.agents.values() if q['_type'] == 'transmon']
    qubits = sorted(qubits, key = lambda name:name.__name__)

    if t is None:
        startTimes = [qubit.get('xy',env.NOTHING).start for qubit in qubits]+[qubit.get('z',env.NOTHING).start for qubit in qubits] + [qubit.get('rr',env.NOTHING).start for qubit in qubits]
        simStart = min([x for x in startTimes if x is not None])
        endTimes = [qubit.get('xy', env.NOTHING).end for qubit in qubits]+[qubit.get('z',env.NOTHING).end for qubit in qubits] + [qubit.get('rr',env.NOTHING).end for qubit in qubits]
        simEnd = max([x for x in endTimes if x is not None])
        t = np.arange(simStart, simEnd+timeStep, timeStep)

    qubitsWithDataIdx = []
    for idx, qubit in enumerate(qubits):
        startTimes = (qubit.get('xy',env.NOTHING).start,qubit.get('z',env.NOTHING).start,qubit.get('rr',env.NOTHING).start)
        if np.alen( ([x for x in startTimes if x is not None]) )>0:
            startTime = min([x for x in startTimes if x is not None])
        else:
            startTime = None
        endTimes = (qubit.get('xy',env.NOTHING).end,qubit.get('z',env.NOTHING).end,qubit.get('rr',env.NOTHING).end)
        if np.alen( ([x for x in endTimes if x is not None]) )>0:
            endTime = max([x for x in endTimes if x is not None])
        else:
            endTime=None
        if endTime is not None and startTime is not None:
            qubitsWithDataIdx.append(idx)
    qubitsWithData = [qubit for idx,qubit in enumerate(qubits) if idx in qubitsWithDataIdx]
        
    if triggerTable:
        masterDemods = []
        timing = 0
        #Trigger table convention:
        #(count, delay, length, rchan)
        for row in triggerTable:
            triggerIdx = 0
            count, delay, length, rchan = row
            while triggerIdx < count:
                masterDemods.append((timing + delay*4,timing+delay*4 + length*4))
                timing = timing+delay*4+length*4
                triggerIdx+=1
                
        demodEnv = sum([env.rect(start, end-start, 0.5) for start,end in masterDemods])
            
    offset = 0
    qubitNum = 0
    labels = []
    
    cols='krgbmc'*100
        
    plt.figure()
    for idx,qubit in enumerate(qubitsWithData):
        labels.append(qubit.__name__ + ' x')
        labels.append(qubit.__name__ + ' y')
        labels.append(qubit.__name__ + ' z')
        labels.append(qubit.__name__ + ' rr')
        if triggerTable and mixerTable:
            labels.append(qubit.__name__ + 'demod')
        col=cols[idx]
        plt.plot(t, qubit.get('xy', env.ZERO)(t)/qubit['piAmp']+offset,col+'-', t, np.imag(qubit.get('xy', env.ZERO)(t))/qubit['piAmp']+offset,col+'-.')   
        if spacingByQubit is False:
            offset+=spacing
        plt.plot(t, qubit.get('z',env.ZERO)(t)+offset,col+'-')
        if spacingByQubit is False:
            offset+=spacing
        plt.plot(t, qubit.get('rr',env.ZERO)(t)+offset,col+'-')
        #plt.plot(t, np.imag(qubit.get('rr',env.ZERO)(t))+offset,col+'-.')
        if triggerTable and mixerTable:
            for demod in masterDemods:
                demodStart, demodStop = demod
                demodt = np.arange(demodStart, demodStop, 2)
                plt.plot(demodt, mixerTable[qubit.__name__][:,0]/256.+offset, col + ':')
        offset+=spacing
    
    if triggerTable:
        labels.append('ADC Demod Triggers')
        plt.plot(t, demodEnv(t)+offset,'-k')
        
    if legend:
        plt.legend(labels, loc = 'upper right',ncol=len(qubits))
    
def levels(Sample,measure,lw=3):

    sample, devs, qubits = gc.loadQubits(Sample, measure, write_access=False)
    '''Plots all the levels for the qubits'''
    offset = 0
    qubitNum = 0
    nqubits=len(qubits)
    labels = []
    
    plt.figure()
    cols='krgbmc'
    fullrange=[-0.5, nqubits-0.5]
    #plot qubit operating levels
    carriers=[]
    for idx,qubit in enumerate(qubits):
        col=cols[np.mod(idx,len(cols))]
        x=[idx-0.5,idx+0.5]
        levels=[ qubit['f10']['GHz'] , qubit['f21']['GHz'] ]
        
        for lev in levels:
            y=[lev,lev]
            lines=plt.plot( x, y,col+'-')
            plt.setp(lines, linewidth=lw) 
            
            lines=plt.plot( fullrange, y,col+':')
            plt.setp(lines, linewidth=lw/2.)             

    #if z dc offset, plot qubit RO levels
        zOffsetDC=qubit.get('zOffsetDC',0.)
        if abs(zOffsetDC)>0.:
            zfunc = gates.AmpToFrequency(qubit)
            nonlin =  qubit['f10']['GHz'] - qubit['f21']['GHz']
            f10=zfunc(-zOffsetDC) 
            levels=[ f10 , f10-nonlin ]
            for lev in levels:
                y=[lev,lev]
                lines=plt.plot( x, y,col+'--')
                plt.setp(lines, linewidth=lw)     
        fc=qubit['fc']['GHz']
        if fc not in carriers:
            carriers.append(fc)
    #plot carriers
    for carrier in carriers:
        x=fullrange
        y=[ carrier, carrier ]
        lines=plt.plot( x, y,'k-.')
        plt.setp(lines, linewidth=lw) 
    plt.xlabel('qubit')
    plt.ylabel('f (GHz)')
    plt.grid('off')
    plt.title('Idling Levels')
    
def aczLevels(Sample,measure,aczPair,lw=3):
    
    sample, devs, qubits = gc.loadQubits(Sample, measure, write_access=False)
    '''Plots all the levels for the qubits during a cz'''
    aczPair = [devs[aczPair[0]], devs[aczPair[1]]]
    offset = 0
    qubitNum = 0
    nqubits=len(qubits)
    labels = []
    
    plt.figure()
    cols='krgbmc'
    fullrange=[-0.5, nqubits-0.5]
    #plot qubit operating levels
    carriers=[]
    for idx,qubit in enumerate(qubits):
        col=cols[np.mod(idx,len(cols))]
        x=[idx-0.5,idx+0.5]
        
        if qubit.__name__==aczPair[0].__name__: # proceed as normal
            levels=[ qubit['f10']['GHz'] , qubit['f21']['GHz'] ]
        elif qubit.__name__==aczPair[1].__name__: # get meet up frequency
            meetUpLevel = qubit.get(aczPair[0].__name__+'aczDetune', None)
            if meetUpLevel:
                nonlin = qubit['f21'] - qubit['f10'] 
                levels = [meetUpLevel['GHz'], (meetUpLevel+nonlin)['GHz']]
            else:
                levels=[ qubit['f10']['GHz'] , qubit['f21']['GHz'] ]
        else: # its not a qubit directly in the CZ
            detuneLevel = qubit.get(aczPair[0].__name__+aczPair[1].__name__+'aczDetune', None)
            if detuneLevel:
                nonlin = qubit['f21'] - qubit['f10'] 
                levels = [detuneLevel['GHz'], (detuneLevel+nonlin)['GHz']]
            else:
                levels=[ qubit['f10']['GHz'] , qubit['f21']['GHz'] ]
        
        for lev in levels:
            y=[lev,lev]
            lines=plt.plot( x, y,col+'-')
            plt.setp(lines, linewidth=lw) 
            
            lines=plt.plot( fullrange, y,col+':')
            plt.setp(lines, linewidth=lw/2.)             
   
        fc=qubit['fc']['GHz']
        if fc not in carriers:
            carriers.append(fc)
    #plot carriers
    for carrier in carriers:
        x=fullrange
        y=[ carrier, carrier ]
        lines=plt.plot( x, y,'k-.')
        plt.setp(lines, linewidth=lw) 
    plt.xlabel('qubit')
    plt.ylabel('f (GHz)')
    plt.grid('off')
    plt.title('Levels for ACZ: '+aczPair[0].__name__+', '+aczPair[1].__name__)
    
def readoutFrequencies(Sample, measure):
    
    sample, devs, qubits = gc.loadQubits(Sample, measure, write_access=False)
    readoutFreqs = [q['readoutPulseFrequency'] for q in qubits]
    carrierFreqs = [devs[q['readoutConfig']]['carrierFrequency'] for q in qubits]
    imageFreqs = [carrier-(freq-carrier) for freq, carrier in zip(readoutFreqs, carrierFreqs)]
    
    plt.figure()
    for rFreq, cFreq, iFreq in zip(readoutFreqs,carrierFreqs,imageFreqs):
        plt.plot([rFreq, rFreq],[0,1], '-b',lw=2)
        plt.plot([cFreq, cFreq],[0,1], '-k',lw=2)
        plt.plot([iFreq, iFreq],[0,1], '-r',lw=2)
    
def matrix(rho, colorMap = matplotlib.cm.RdBu, labels = None, grid=True):

    dim = len(rho)
    amps = abs(rho)
    phases = np.angle(rho)

    fig = plt.figure(figsize = (7,6))
    ax = plt.axes([0,0,1,1])
    ax.grid('off')
    
    if labels:
        ax.set_ylim(-2, 2*dim+1)
        ax.set_xlim(-4, 2*dim)
    else:
        ax.set_ylim(-2, 2*dim)
        ax.set_xlim(-2, 2*dim)
        
    if labels:
        textpos = np.arange(0, 2*dim, 2)
        for tpos, label, revlabel in zip(textpos, labels, labels[::-1]):
            plt.text(tpos, max(textpos)+2, str(label), ha='center', va='center', size=12)
            plt.text(min(textpos)-3, tpos, str(revlabel), ha='center', va='center', size=12)
            
    if grid:
        for tpos in textpos:
            plt.plot([tpos]*2, [min(textpos), max(textpos)], '--k', lw = 1)
            plt.plot([min(textpos), max(textpos)], [tpos]*2, '--k', lw = 1)
    
    pos = np.mgrid[2*(dim-1):-2:-2, 0:2*dim:2]
    
    circlePatches = []
    circleColors = []
    for rownum, row in enumerate(rho):
        for elementnum, element in enumerate(row):
            art = matplotlib.patches.Circle(
                             [pos[1][rownum,elementnum],
                              pos[0][rownum,elementnum]],
                              amps[rownum,elementnum])
            circleColors.append(amps[rownum, elementnum])
            circlePatches.append(art)
    
    # this next crap is for scaling the colors, it's a 
    # dumb way to do it.
    circlePatches.append(matplotlib.patches.Circle([0,0],0))
    circlePatches.append(matplotlib.patches.Circle([-20,-20],1))
    circleColors.append(0)
    circleColors.append(1)
    
    for rownum, row in enumerate(rho):
        for elementnum, element in enumerate(row):
            dx = amps[rownum, elementnum]*np.cos(phases[rownum, elementnum])
            dy = amps[rownum, elementnum]*np.sin(phases[rownum, elementnum])
            x = pos[1][rownum,elementnum]
            y = pos[0][rownum,elementnum]
            art = matplotlib.lines.Line2D([x, x+dx],[y, y+dy], lw = 3, color='k')
            ax.add_line(art)
            
    circleCollection = matplotlib.collections.PatchCollection(circlePatches, 
                                       cmap = colorMap)
    
    circleCollection.set_array(np.array(circleColors))
    ax.add_collection(circleCollection)
    plt.colorbar(circleCollection)
    
def city(rho, d=0.1, cmap=matplotlib.cm.RdBu, vmin=-1, vmax=1):
   """Make a 3d bar plot using a colormap to color each bar.

   d - spacing between adjacent bars
   cmap - the colormap to use.  defaults to RdBu
   vmin - minimum z-value and color value
   vmax - maximum z-value and color value
   """

   norm = matplotlib.colors.Normalize(vmin, vmax)

   fig = plt.figure()
   ax = Axes3D(fig)

   w, h = rho.shape
   for x in range(w):
       for y in range(h):
           ax.bar3d([x+d/2], [y+d/2], [0], 1-d, 1-d, rho[x,y],
           cmap(norm(rho[x,y])))

   ax.set_zlim3d((vmin, vmax))
   
   ax.set_aspect('equal')

   return fig
    
def rho2bloch(rho): # turns a 2x2 rho into a 3d bloch vector

    rho = np.matrix(rho)

    if len(rho)>2:
        rho = rho[:2,:2] # truncate baby!

    sigmaX = np.matrix([[0, 1], [1, 0]], dtype = 'complex')
    sigmaY = np.matrix([[0, -1j], [1j, 0]], dtype = 'complex')
    sigmaZ = np.matrix([[1, 0], [0, -1]], dtype = 'complex')
    
    rx = np.trace( sigmaX * rho )
    ry = np.trace( sigmaY * rho )
    
    rz = np.trace( sigmaZ * rho )
    
    return np.array([rx, ry, rz])
    
def initBlochSphere(labels=True,plotAxes=True,fignum=10, frameColor = (0.7, 0.7, 0.7), frameLw = 2):
    
    u = np.linspace(0, 2 * np.pi, 41)
    v = np.linspace(0, np.pi, 41)

    x = np.outer(np.cos(u), np.sin(v))
    y = np.outer(np.sin(u), np.sin(v))
    z = np.outer(np.ones(np.size(u)), np.cos(v))
    
    plt.close(fignum)
    ax = Axes3D(plt.figure(fignum))
    ax.plot_wireframe(x, y, z, rstride = 10, cstride = 20, color = frameColor, lw = frameLw) #cstride = latitudinal, rstride = longitudinal

    if labels:
        ax.text(0, 0, 1, '|0>')
        ax.text(0, 0, -1, '|1>')
        ax.text(1, 0, 0, '|0> + |1>')
        ax.text(-1, 0, 0, '|0> - |1>')
        ax.text(0, 1, 0, '|0> + i|1>')
        ax.text(0, -1, 0, '|0> - i|1>')
        
    if plotAxes:
        ax.plot([-1,1],[0,0],[0,0], 'black')
        ax.plot([0,0],[-1,1],[0,0], 'black')
        ax.plot([0,0],[0,0],[-1,1], 'black')
        
    ax.set_aspect('equal')
    
    return ax
    
def blochSphere(rhos): # takes a list of rhos and maps them onto the bloch sphere
    
    ax = initBlochSphere(0)
    
    vec = np.array([rho2bloch(rho) for rho in rhos])
    
    ax.plot(vec[:,0],vec[:,1],vec[:,2], 'o', color = 'red')
    
    
def plot2DScatter(data, size=150):
    
    x = (data[:,0],data.indep[0])
    y = (data[:,1],data.indep[1])
    vars = []
    for k in range(data.shape[1]-2):
        vars.append((data[:,k+2],data.dep[k]))
        
    for var in vars:
        plt.figure()
        plt.xlabel(x[1])
        plt.ylabel(y[1])
        plt.scatter(x[0],y[0],s=size, c=var[0], marker='s',lw=0)
        plt.title(var[1])
        plt.colorbar()
        
def nelderMead(dvwData, lastPoint = None):

    deps = dvwData.dep
    parameters = [dep[0] for dep in deps[1:]]
    
    initialValues = dvwData[0,2:]
    
    normalizedValues = []
    
    for k,initialValue in enumerate(list(initialValues)):
    
        unnormalizedValue = dvwData[:lastPoint,k+2]
        normalizedValues.append(unnormalizedValue/initialValue)
        
    normalizedData = np.vstack((dvwData[:lastPoint,0],np.array(normalizedValues))).transpose()
    
    for k in range(len(initialValues)):
        plt.plot(normalizedData[:,0], normalizedData[:,k+1], label = parameters[k])
        plt.xlabel('Function Evaluations')
        plt.ylabel('Percent Change from Initial')
        
    plt.legend()
    return normalizedData
    
def swapSpectroscopyT1(datasetWrapper):

    data = datasetWrapper
    data=np.array(data)
    dataSize = len(matplotlib.mlab.find(data[:,0]==data[0,0]))
    dataLen = len(data)
    print dataLen,dataSize
    numColumns = dataLen/dataSize
    newDataLen = numColumns*dataSize
    data = data[:newDataLen, :] # we take off any unfinished rows
    rawFreqs = data[::dataSize, 0]
    delays = data[:dataSize,1]
    probs = np.reshape(data[:,2], (numColumns, dataSize))

    fitFunc = lambda time, t1, prob, const: prob*np.exp(-time/t1)+const
    p0 = [1e4, 0.9, 0.0]
    t1s = []
    error_bars = []
    freqs = []
    
    for prob,freq in zip(probs,rawFreqs):
    
        popt, pcov = curve_fit(fitFunc, delays, prob, p0=p0)
        
        if type(pcov) is type(np.inf):
            continue
        
        freqs.append(freq)
        t1s.append(popt[0])
        error_bars.append(np.sqrt(pcov[0,0]))
        
    plt.errorbar(freqs, np.array(t1s)/1.e3, yerr=np.array(error_bars)/1.e3, fmt='ob',markerfacecolor='none',markeredgecolor='b',label = 'MQ data')    
    plt.xlabel('Qubit Frequency [GHz]')
    plt.ylabel('Qubit T1 [us]')
    
def dephasingSpectroscopy(datasetWrapper, ymax=20):
    d = datasetWrapper
    freqs = np.unique(d[:,0])
    fitFunc = lambda t, tphi1, tphi2, prob, const: prob*np.exp(-t/tphi1)*np.exp(-t**2/tphi2) + const
    guess = (3000, 3000, 0.9, 0.1)
    tphi1s, tphi2s = [], []
    err1, err2 = [], []
    f1, f2 = [], []
    for f in freqs:
        d2 = d[d[:,0] == f]
        popt, pcov = curve_fit(fitFunc, d2[:,1], d2[:,2], guess)
        try:
            err1.append(np.sqrt(pcov[0,0]))
            tphi1s.append(popt[0])
            f1.append(f)
            err2.append(np.sqrt(pcov[1,1]))
            tphi2s.append(np.sqrt(popt[1]))
            f2.append(f)
        except TypeError:
            print "Bad fit: f=%s" % f
    #plt.errorbar(f1, np.array(tphi1s)/1.e3, yerr=np.array(err1)/1.e3, fmt='ob',markerfacecolor='none', markeredgecolor='b',label = '$T_{\phi1}$')
    plt.errorbar(f2, np.array(tphi2s)/1.e3, yerr=np.array(err2)/1.e3, fmt='or',markerfacecolor='none', markeredgecolor='r',label = '$T_{\phi2}$')
    plt.plot(f1, np.array(tphi1s)/1e3, 'ob', label='$T_{\phi1}$')
    plt.plot(f1, np.array(tphi2s)/1e3, 'or', label='$T_{\phi2}$')
    plt.xlabel('Qubit Frequency [GHz]')
    plt.ylabel('Qubit $T_\phi$ [us]')
    plt.legend(numpoints=1)
    plt.ylim([0,ymax])
    
def isolationMatrix(data, noisy=True):
    
    def getRabiLen(lengths, probs):
        splineCoarse = scipy.interpolate.UnivariateSpline(lengths, probs-0.5, s=0.05)
        if len(splineCoarse.roots())<2:
            return np.nan
        else:
            lengthGuess = (splineCoarse.roots()[0] + splineCoarse.roots()[1])/2.
            spline = scipy.interpolate.UnivariateSpline(lengths, probs, s=0.05)
            length = scipy.optimize.fmin(lambda x: abs(spline.derivatives(x)[1]), lengthGuess)
            return length
    
    qubits = [data.parameters[q] for q in data.parameters['measure']]
    measure = data.parameters['measure']
    experiments = list(itertools.product(range(len(measure)), range(len(measure))))
    driveAmp = data.parameters['driveAmp']
    
    if len(data.indep)==1:
    
        lengths = data[:,0]
        isolationMatrix = np.zeros((len(measure), len(measure)))
        for k, xqubits in enumerate(experiments):
            qmeas,qdrive = xqubits
            length = getRabiLen(lengths, data[:,k+1])
            isolationMatrix[qmeas,qdrive] = length * driveAmp
                
        isolationMatrixNorm = np.zeros((len(measure), len(measure)))
        for k,q in enumerate(qubits):
            currColumn = isolationMatrix[:,k]
            normalizer = isolationMatrix[k,k]
            isolationMatrixNorm[:,k] = (currColumn/normalizer)
        
        isolationMatrixDB = 10*np.log10(isolationMatrixNorm)
        
        print "Isolation Matrix:"
        print isolationMatrixDB
        if noisy:
            print "Isolation Matrix Raw:"
            print isolationMatrix
            
    elif len(data.indep)==2:
        # we're just going to plot the raw coupling vs freq
        lengths = data[:,0]
        lengths_set = np.sort(list(set(lengths)))
        freqs = data[:,1]
        freqs_set = np.sort(list(set(freqs)))
        
        couplingMatrix = np.zeros((len(freqs_set), len(experiments)))
        
        for el,freq in enumerate(freqs_set):
            for k,expt in enumerate(experiments):
                idx = np.argwhere(freqs==freq).flatten()
                probs = np.array(data[idx,k+2])
                length = getRabiLen(lengths_set, probs)
                couplingMatrix[el, k] = length * driveAmp
            
        # convert to dB
        couplingMatrixDB = 10*np.log10(couplingMatrix)
        
        names = ['qubit %s driven by %s' % (measure[expt[0]], measure[expt[1]]) for expt in experiments]
        
        plt.figure()
        for column, name in zip(couplingMatrixDB.transpose(), names):
            plt.plot(freqs_set, column, label=name)
        plt.xlabel('Frequency [GHz]')
        plt.ylabel('Coupling [dB]')
        plt.legend()
        
        return couplingMatrix
        
    else:
        pass

def rabipop(datasetWrapper, verbose=True):
    d = datasetWrapper
    #fitFunc = lambda t, amp, freq, phase, offset: amp*np.cos(freq*t+phase)+offset
    fitFunc = lambda t, amp, offset: amp*np.cos(1.0*t+np.pi)+offset
    popt1, pcov1 = curve_fit(fitFunc, d[:,0], d[:,1], (0.1, 0.8))
    popt2, pcov2 = curve_fit(fitFunc, d[:,0], d[:,2], (0.5, 0.5))
    res = popt1[0] / (popt2[1]+popt2[0])
    if verbose:
        x = np.linspace(d[:,0].min(), d[:,0].max())
        plt.figure()
        plt.plot(d[:,0], d[:,1], 'b.')
        plt.plot(d[:,0], d[:,2], 'r.')
        plt.plot(x, fitFunc(x, *popt1), 'b-')
        plt.plot(x, fitFunc(x, *popt2), 'r-')
        print "Fit 1: %s" % str(popt1)
        print "Fit 2: %s" % str(popt2)
        print "Ae/(Ae+Ag) = %s" % res
    return res
    
def rabipopSpectroscopy(datasetWrapper, showAll=False, plot=True):
    d = datasetWrapper
    freqs = np.unique(d[:,0])
    pops = []
    for f in freqs:
        d2 = d[d[:,0]==f,1:]
        diff = rabipop(d2, verbose=showAll)
        pops.append(diff)
    if plot:
        plt.figure()
        plt.plot(freqs, pops, '-ob')
        plt.xlabel('Qubit Frequency [GHz]')
        plt.ylabel('Rabi Pop amplitude difference')
    return freqs, pops
    
def rabipopTimeConstant(datasetWrapper, showAll=False, plot=True):
    d = datasetWrapper
    times = np.unique(d[:,0])
    pops = []
    for t in times:
        d2 = d[d[:,0]==t,1:]
        diff = rabipop(d2, verbose=showAll)
        pops.append(diff)
    if plot:
        plt.figure()
        plt.plot(times, pops, '-ob')
        plt.xlabel('Thermalization Time [ns]')
        plt.ylabel('Rabi Pop amplitude difference')
    return times, pops
    
def rabipop2effectiveTemp(datasetWrapper, delta):
    
    def population(freq, T):
        # temperature in GHZ, 20.8 GHz/Kelvin
        return np.exp(-freq/(T)) / (1 + np.exp(-freq/(T)) + np.exp(-(2*freq-delta)/(T)) + np.exp(-(3*freq-2*delta)/(T)) )
    
    freqs, pops = rabipopSpectroscopy(datasetWrapper, plot=False)
    T, conv = curve_fit(population, freqs, pops, p0=[1.2])
    
    plt.figure()
    plt.plot(freqs, pops, 'ok',label='Data')
    plt.xlabel('Qubit Frequency [GHz]')
    plt.ylabel('Rabi Pop amplitude difference')
    plt.plot(freqs, population(freqs, T),'-k',label='Boltzmann Statistics')
    plt.plot(freqs, population(freqs, .020*20.8),'--',label='T=20mK')
    plt.plot(freqs, population(freqs, .030*20.8),'--',label='T=30mK')
    plt.plot(freqs, population(freqs, .040*20.8),'--',label='T=40mK')
    plt.plot(freqs, population(freqs, .050*20.8),'--',label='T=50mK')
    plt.plot(freqs, population(freqs, .060*20.8),'--',label='T=60mK')
    plt.plot(freqs, population(freqs, .070*20.8),'--',label='T=70mK')
    plt.plot(freqs, population(freqs, .080*20.8),'--',label='T=80mK')
    plt.plot(freqs, population(freqs, .090*20.8),'--',label='T=90mK')
    plt.plot(freqs, population(freqs, .10*20.8),'--',label='T=100mK')
    plt.legend()
    plt.title('Teff = %0.3f mK' %(T[0]/20.8)) #GHZ/Kelvin
    
def measurePopulation2effectiveTemp(datasetWrapper, delta=0.22):    
    def population(freq, T):
        # temperature in GHZ, 20.8 GHz/Kelvin
        return np.exp(-freq/(T)) / (1 + np.exp(-freq/(T)) + np.exp(-(2*freq-delta)/(T)) + np.exp(-(3*freq-2*delta)/(T)) )        
    freqs = datasetWrapper[:,0]
    pops = 1-datasetWrapper[:,1]    
    T, conv = curve_fit(population, freqs, pops, p0=[1.0])    
    plt.figure()
    plt.plot(freqs, pops, 'ok',label='Data')
    plt.xlabel('Qubit Frequency [GHz]')
    plt.ylabel('Rabi Pop amplitude difference')
    plt.plot(freqs, population(freqs, T),'-k',label='Fit')
    temperatures=np.arange(0.02,0.11,0.01)    
    for temp in temperatures:
        plt.plot(freqs, population(freqs, temp*20.8),'--',label='T=%gmK' % (temp*1000))
    plt.legend()
    plt.title('Teff = %0.1f mK' %(1000*T[0]/20.8)) #GHZ/Kelvin
    
def orbitXYXtalk(listOfDvwData):
    
    #qubits = [data.parameters[q] for q in data.parameters['measure']]
    
    experiments = [data.parameters['measure'] for data in listOfDvwData]
    qNames = list(set([item for sublist in experiments for item in sublist]))
    f10s = [data.parameters[q]['f10'] for q in qNames]
    
    updateDict = {}
    colorDict = {}
    cols = 'krgbmc'*5
    dotTraceIdx = [len(experiments)+k for k,_ in enumerate(qNames)]
    for k,q in enumerate(qNames):
        toUpdate = [idx for idx, expt in enumerate(experiments) if expt[1]==q]
        updateDict[q] = toUpdate + [dotTraceIdx[k]]
        colorDict[q] = cols[k]
        
    # add traces to indicate the dot thing
    
    experiments = [data.parameters['measure'] for data in listOfDvwData]
    traces = [{'x': data[:,0] - data.parameters[data.parameters['measure'][0]]['f10'], 
               'y': data[:,1] - qNames.index(data.parameters['measure'][0]),
               'kw':{'lw':2}} for data in listOfDvwData]
    traces += [{'x': 0, 'y': 0.8-qNames.index(k), 'kw':{'lw':2,'marker':'o','markersize':10,'color':colorDict[k]}} for k in qNames]
               
    params = [{'name': '%s dF10' % k, 'val': 0, 'range':(-1, 1), 'axis': 'x', 'color': colorDict[k], 'update': updateDict[k]} for k in qNames]
    
    result = adjust.adjust(params, traces)
    dfs = [result[q+' dF10'] for q in qNames]
    newF10s = [result[q+' dF10']+f10['GHz'] for q,f10 in zip(qNames, f10s)]
    for q,df in zip(qNames,dfs):
        print '%s DF: %f' % (q, df)
    
def boldify(ax,
            tickEdgeWidth=5, tickSize=8, tickFontSize=40,
            gridLinewidth=2,
            labelSize=30):
    xaxis = ax.xaxis
    yaxis = ax.yaxis
    for axis in [xaxis,yaxis]:
        for line in axis.get_ticklines():
            line.set_markeredgewidth(tickEdgeWidth)
            line.set_markersize(tickSize)
        for label in axis.get_ticklabels():
            label.set_fontsize(tickFontSize)
        axis.label.set_size(labelSize)
    ax.grid(linewidth=gridLinewidth)

def undacify(packed):
    '''
    Undacify's a string representation of a hex SRAM word
    '''
    I = int(packed, 16)
    sign = I & 2**13
    I = I & 0x1FFF
    if sign:
        I = -2**13 + I
    Q = int(packed, 16) >> 14
    sign = Q & 2**13
    Q = Q & 0x1FFF
    if sign:
        Q = -2**13 + Q    
    return float(I)/2**13,float(Q)/2**13
    
DEBUG_FILE = os.path.join(os.path.expanduser('~'), 'packet-dump', 'ghzDacPacket.txt')    
def plotGHzDacDump(fname=DEBUG_FILE, dac=0):
    '''
    Plot data from a GHz FPGA debug
    '''
    with open(fname, 'r') as f:
        lines = f.readlines()
    lines = [l.strip() for l in lines ]
    startidx = lines.index('SRAM')
    lines = lines[startidx+1:]
    stopidx = lines.index('')
    lines = lines[0:stopidx]
    data = []
    for l in lines:
        dat = l.split()[dac]
        data.append(undacify(dat))
    return np.array(data)
        
def makePresentable(ax, datasets=None, xLabel=None, yLabel=None, markerEdgeWidth=10, 
                    markerSize=15, tickLabelFontSize=40, legendFontSize=35, titleSize=40, 
                    labelFontSize=30, gridLineWidth=2, legendLoc='lower left', grid=True):
    xAxis = ax.xaxis
    yAxis = ax.yaxis
    #Set tick line size
    for axis in [xAxis, yAxis]:
        for line in axis.get_ticklines():
            line.set_markeredgewidth(markerEdgeWidth)
            line.set_markersize(markerSize)
        for label in axis.get_ticklabels():
            label.set_fontsize(tickLabelFontSize)
        label = axis.get_label()
        label.set_fontsize(labelFontSize)
    ax.grid(linewidth=gridLineWidth)
    prop = fm.FontProperties(size=legendFontSize)
    if legendLoc:
        ax.legend(loc=legendLoc, numpoints=1, prop=prop)
    if datasets:
        titleStr = '\n'.join([str(dataset.path) for dataset in datasets])
        ax.set_title(titleStr, size=titleSize)
    for label, axis in zip((xLabel, yLabel), (xAxis, yAxis)):
        if label:
            axis.set_label_text(label, size=labelFontSize)
    try:
        plt.tight_layout()
    except AttributeError: # Only works in newer versions of matplotlib
        pass
    
def plot2Dscan(dataset, xbins=100, ybins=100):
    ''' plot a 2D dataset with a color plot '''
    d = dataset.data
    # do manipulations of d here to swap axes, etc.
    assert d.shape[1] >= 3
    xmin, xmax, ymin, ymax = d[:,0].min(), d[:,0].max(), d[:,1].min(), d[:,1].max()
    gridpts = np.array(list(itertools.product(np.linspace(xmin, xmax, xbins), np.linspace(ymin, ymax, ybins))))
    interped = scipy.interpolate.griddata(d[:,0:2], d[:,2], gridpts, method="linear")
    interped = np.vstack(np.split(interped, xbins)).T
    plt.imshow(interped, extent=(xmin, xmax, ymin, ymax), origin='lower', aspect="auto")
    plt.xlabel("%s [%s]" % (dataset.variables[0][0][0], dataset.variables[0][0][1]))
    plt.ylabel("%s [%s]" % (dataset.variables[0][1][0], dataset.variables[0][1][1]))
    
    #return gridpts, interped

    
def plotResFreqs(Sample, measure):        
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
        
    lw=1
    for idx,qubit in enumerate(qubits):
        col='b'
        if qubit['_surf']=='measure':
            col='r'        
        x=[ qubit['readoutResonatorFreq0']]*2
        y=[-0.5,0.5]
        print x
        print y
        lines=plt.plot( x, y,col+'-')
        plt.setp(lines, linewidth=lw)         
        
