#!/usr/bin/python
''' dephasing_funcs_2.py

Run RB simulations for 2 qubit ZZ interaction.
'''

import random
import numpy as np
import numpy_2qubit as np2

def generate_noise(correlated=False, dphi=1e-4):
    if correlated:
        return [np2.rotzz(dphi)]
    else:
        return [np2.rotzz(dphi), np2.rotzz(-dphi)]
        
def _pad_gates(gates, pad_to_length=3):
    ''' pad each tuple of gates to be length X by adding identity gates.
    will truncate any tuple that is longer than X gates (not recommended). '''
    g2 = []
    for g in gates:
        subgate = []
        for i in range(pad_to_length):
            try:
                subgate.append(g[i])
            except IndexError:
                subgate.append(np2.I)
        g2.append(subgate)
    return g2
        
def _pad_check(this_a, this_b, inv_a, inv_b):
    ''' pad either this_a/inv_a or this_b/inv_b (or neither)
        with np2.I so they're the same length. '''
    while len(this_a) < len(this_b):
        this_a = this_a + (np2.I,)
        inv_a = inv_a + (np2.I,)
    while len(this_b) < len(this_a):
        this_b = this_b + (np2.I,)
        inv_b = inv_b + (np2.I,)
    return this_a, this_b, inv_a, inv_b
        
def rb_sequence(gates_a=None, gates_b=None, gates_a_inv=None, gates_b_inv=None,
                noise_gates=None, N=100, noise_after=True, debug=False,
                n_interspersed=1, pad_gates=None, return_all_ops=False):
    ''' run a single two-qubit sequence.
    For the three methods:
    noise_after == True ==> method 1 (apply cliffords, then apply noise)
    noise_after == False and n_interspersed <= 1 ==> method 2 (noise after each pauli gate)
    noise_after == False and n_interspersed > 1  ==> method 3 (noise interspersed with pauli gates) '''
    gates_a = gates_a or np2.cliffords_a
    gates_a_inv = gates_a_inv or np2.cliffords_a_inv
    gates_b = gates_b or np2.cliffords_b
    gates_b_inv = gates_b_inv or np2.cliffords_b_inv
    noise_gates = noise_gates or generate_noise()
    
    state = np2.ground()
    inverter = np2.I
    
    if pad_gates:
        if pad_gates is True:
            pad_gates = 3
        gates_a = _pad_gates(gates_a, pad_to_length=pad_gates)
        gates_b = _pad_gates(gates_b, pad_to_length=pad_gates)
        gates_a_inv = _pad_gates(gates_a_inv, pad_to_length=pad_gates)
        gates_b_inv = _pad_gates(gates_b_inv, pad_to_length=pad_gates)
    
    if return_all_ops:
        ops_list = []
        debug = True
    
    if debug:
        i_list = []
    
    for i in range(N):
        ra = random.randint(0, len(gates_a)-1)
        rb = random.randint(0, len(gates_b)-1)
        if debug:
            i_list.append((ra, rb))
        if noise_after:
            this_a, this_b, inv_a, inv_b = gates_a[ra], gates_b[rb], gates_a_inv[ra], gates_b_inv[rb]
            for ga, gai in zip(this_a, inv_a):
                state = np.dot(ga, state)
                inverter = np.dot(inverter, gai)
                if return_all_ops:
                    ops_list.append(ga)
            for gb, gbi in zip(this_b, inv_b):
                state = np.dot(gb, state)
                inverter = np.dot(inverter, gbi)
                if return_all_ops:
                    ops_list.append(gb)
            n = random.choice(noise_gates)
            state = np.dot(n, state)
            if return_all_ops:
                ops_list.append(n)
        elif n_interspersed <= 1:
            this_a, this_b, inv_a, inv_b = _pad_check(gates_a[ra][:], gates_b[rb][:], gates_a_inv[ra][:], gates_b_inv[rb][:])
            for ga, gb, gai, gbi in zip(this_a, this_b, inv_a, inv_b):
                state = np.dot(ga, state)
                state = np.dot(gb, state)
                inverter = np.dot(inverter, gai)
                inverter = np.dot(inverter, gbi)
                n = random.choice(noise_gates)
                state = np.dot(n, state)
                if return_all_ops:
                    ops_list.append(ga)
                    ops_list.append(gb)
                    ops_list.append(n)
        else:
            this_a, this_b, inv_a, inv_b = _pad_check(gates_a[ra][:], gates_b[rb][:], gates_a_inv[ra][:], gates_b_inv[rb][:])
            for ga, gb, gai, gbi in zip(this_a, this_b, inv_a, inv_b):
                for j in xrange(n_interspersed):
                    state = np.dot(ga, state)
                    state = np.dot(gb, state)
                    n = random.choice(noise_gates)
                    state = np.dot(n, state)
                    if return_all_ops:
                        ops_list.append(ga)
                        ops_list.append(gb)
                        ops_list.append(n)
                inverter = np.dot(inverter, gai)
                inverter = np.dot(inverter, gbi)
                
            
    state = np.dot(inverter, state)
    if return_all_ops:
        return np2.measure(state), i_list, ops_list, inverter
    if debug:
        return np2.measure(state), i_list
    return np2.measure(state)

def debug(nmax=10000):
    i = 0
    m, ilist = rb_sequence(noise_gates=[np2.I], debug=True)
    while i<nmax and np2.equal(m, np2.ground()):
        m, ilist = rb_sequence(noise_gates=[np2.I], debug=True)
        i += 1
    if i==nmax:
        print "SUCCESS"
        return None, None
    else:
        print "FAILURE i=%s" % i
        return m, ilist


def rb_avg(Nlist=range(5,100,5), num_avg=300, no_divide=False, **kwargs):
    data = np.zeros((len(Nlist), 4), dtype=np.float64)
    for i in range(num_avg):
        for j, n in enumerate(Nlist):
            data[j] += rb_sequence(N=n, **kwargs)
    if no_divide:
        return data
    else:
        return data / num_avg
        
def _m_rb_avg(q, *args, **kwargs):
    q.put(rb_avg(*args, **kwargs))

def _queue_get(queue, block=True, timeout=None):
    import errno
    while True:
        try:
            return queue.get(block, timeout)
        except IOError, e:
            if e.errno != errno.EINTR:
                raise
        
def rb_multiprocess(processes=6, Nlist=range(5,100,5), num_avg=400, **kwargs):
    from multiprocessing import Process, Queue
    avg_list = [num_avg/processes] * processes
    avg_list[0] += num_avg%processes
    assert np.sum(avg_list) == num_avg
    procs = []
    queues = []
    data = np.zeros((len(Nlist), 4), dtype=np.float64)
    kwargs['no_divide'] = True
    kwargs['Nlist'] = Nlist
    for i in range(processes):
        q = Queue()
        kw2 = kwargs.copy()
        kw2['num_avg'] = avg_list[i]
        p = Process(target=_m_rb_avg, args=[q,], kwargs=kw2)
        p.start()
        procs.append(p)
        queues.append(q)
    for p in procs:
        p.join()
    for q in queues:
        data += _queue_get(q)
    return data / num_avg
        
