''' numpy_2qubit.py

basically as numpy_states.py, but for 2 qubits.
'''

import numpy as np
import numpy.linalg as LA
import itertools
import numpy_states as nps

dtype = np.complex128
rtol, atol = 1e-7, 1e-12
pi = np.pi

I = np.identity(4, dtype=dtype)
Xa = np.array([ [0,0,1,0],
                [0,0,0,1],
                [1,0,0,0],
                [0,1,0,0]], dtype=dtype)
Xb = np.array([ [0,1,0,0],
                [1,0,0,0],
                [0,0,0,1],
                [0,0,1,0]], dtype=dtype)
Ya = np.array([ [0,0,-1j,0],
                [0,0,0,-1j],
                [ 1j,0,0,0],
                [0, 1j,0,0]], dtype=dtype)
Yb = np.array([ [0,-1j,0,0],
                [ 1j,0,0,0],
                [0,0,0,-1j],
                [0,0, 1j,0]], dtype=dtype)
Za = np.array([ [1,0,0,0],
                [0,1,0,0],
                [0,0,-1,0],
                [0,0,0,-1]], dtype=dtype)
Zb = np.array([ [1,0,0,0],
                [0,-1,0,0],
                [0,0,1,0],
                [0,0,0,-1]], dtype=dtype)
                
def rotxa(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Xa
def rotxb(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Xb
def rotya(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Ya
def rotyb(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Yb
def rotza(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Za
def rotzb(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Zb
    
def rotzz(dphi):
    v = np.exp(-1j/2.0 * dphi/2)
    return np.array( [[1,0,0,0],[0,v,0,0],[0,0,v,0],[0,0,0,1]], dtype=dtype)

def ground():
    return np.array([1,0,0,0], dtype=dtype)
def exciteda():
    return np.array([0,1,0,0], dtype=dtype)
def excitedb():
    return np.array([0,0,1,0], dtype=dtype)
def excitedab():
    return np.array([0,0,0,1], dtype=dtype)
    
def measure(s):
    return (s.conj() * s).real
    
def norm(state):
    return LA.norm(state)
    
def renorm(state):
    return state / norm(state)

def tuplify(x):
    if type(x) is tuple:
        return x
    else:
        return (x,)
        
def reduce(state):
    ''' factor out global phase factor to make the |00> coefficient
    real and non-negative. '''
    #if np.allclose(np.abs(state[0]), 0, rtol=rtol, atol=rtol):
    #    return excited()
    #if np.allclose(np.abs(state[1]), 0, rtol=rtol, atol=rtol):
    #    return ground()
    i = 0
    while np.allclose(np.abs(state[i]), 0, rtol=rtol, atol=atol):
        i += 1
    phase = np.exp(-1j*np.angle(state[i]))
    return phase * state
    #return np.array([np.abs(state[0]),
    #                 np.exp(-1j*np.angle(state[0])) * state[1],
    #                 np.exp(-1j*np.angle(state[0])) * state[2],
    #                 np.exp(-1j*np.angle(state[0])) * state[3]], dtype=dtype)

def equal(a, b):
    return np.allclose(np.dot(reduce(a.conj()), reduce(b)), 1, rtol=rtol, atol=atol)
    
def dotMany(gates, state, inv=False):
    if inv:
        for g in gates:
            state = np.dot(state, g)
    else:
        for g in gates:
            state = np.dot(g, state)
    return state

def generate_cliffords(intersperse=1):
    ''' generate single qubit cliffords for each of the two qubits. '''
    x = float(intersperse)
    clifsa, clifsb = [], []
    for cliffords, rotx, roty, rotz in zip([clifsa, clifsb], [rotxa, rotxb], [rotya, rotyb], [rotza, rotzb]):
        # paulis
        cliffords += [I, rotx(pi/x), roty(pi/x), (rotx(pi/x), roty(pi/x))]
        # 2*pi/3 rotations
        x2s = [rotx(pi/2/x), rotx(-pi/2/x)]
        y2s = [roty(pi/2/x), roty(-pi/2/x)]
        cliffords += list(itertools.product(x2s, y2s)) + list(itertools.product(y2s, x2s))
        # pi/2 rotations
        cliffords += x2s + y2s + [(x2s[1], y2s[0], x2s[0])] + [(x2s[1], y2s[1], x2s[0])]
        # hadamard-like
        cliffords += list(itertools.product([rotx(pi/x)], y2s))
        cliffords += list(itertools.product([roty(pi/x)], x2s))
        cliffords += [(x2s[0], y2s[0], x2s[0]), (x2s[1], y2s[0], x2s[1])]
        cliffords = map(tuplify, cliffords)
    return map(tuplify, clifsa), map(tuplify, clifsb)

def stateIn(s, state_list):
    for s2 in state_list:
        if equal(s, s2):
            return True
    return False

octagon = [np.outer(a,b).reshape(4) for a,b in itertools.product(nps.octagon, nps.octagon)]
cliffords_a, cliffords_b = generate_cliffords()
cliffords_a_inv, cliffords_b_inv = generate_cliffords(-1)
