# Copyright 2012 Daniel Sank
# University of California Santa Barbara

#Numerical analysis tools for transmon qubit

import numpy as np
import scipy.interpolate as interpolate
import labrad.units as units
Value,Unit = units.Value, units.Unit

EJ_OVER_EC_RANGE = (20, 400)

PHI_0 = Value(2.06783e-15,'Wb')
ELECTRON_CHARGE = Value(1.60217657E-19,'C')
PLANCK_CONSTANT = Value(6.626E-34, 'm^2*kg/s')
HBAR = PLANCK_CONSTANT / (2*np.pi)

a = np.diag(np.sqrt(np.arange(10)), 1)
adag = a.T

#The first set of functions numerically computes the transmon energy
#levels and wavefunctions. The energy levels allow computation of
#the anharmonicity. Note that in these functions we work with the
#Hamiltonian in units of the charging energy E_C. As much as possible
#we work with dimensionless quantities to avoid confusion.

def hamiltonian_realSpace(EjOverEc, N=500, alpha=0.0):
    """The discretized real space (phase) Hamiltonian for the transmon
    
    The computation is done by discretizing the Hamiltonian. From basic
    circuit theory you can find the Hamiltonian in the delta (phase)
    basis for the circuit (be careful converting charge to a derivative
    with respect to phase!). We then discretize the phase axis. Writing
    the potential energy term in the discrete form is easy. To get the
    kinetic energy, you have to know how to write a discretized second
    derivative, which is as follows:
    d^2/dx^2 = (1/\Delta x)^2 (f[i+1] - 2f[i] + f[i-1])
    where \Delta x is the grain size of the discretized axis.
    As a matrix this just has +2 on the diagonal and -1 on the first
    off-diagonal.
    
    Note that we deal with the Hamiltonian divided by the charging
    energy to get a unitless quantity.
    """
    delta = np.linspace(-np.pi, np.pi, N) #includes endpoints
    ddelta = delta[1] - delta[0]
    
    derivTerm = -4 * 1.0/(ddelta**2) * ( np.diag([1]*(N-1), -1) + np.diag([-2]*N) + np.diag([1]*(N-1), 1) )
    potentialTerm = -1.0 * EjOverEc * np.diag(np.cos(delta))

    if alpha:
        potentialTerm += np.diag([-alpha*Ej*np.sqrt(1-(np.sin(d)/alpha)**2) for d in delta])
    
    H = derivTerm + potentialTerm
    return delta, H
    
def analyzeH(H):
    """Get ordered eigenvalues and eigenvectors of a Hamiltonian
    
    Note that the eigenvectors are wavefunction almplitudes in the delta
    basis.
    """
    eval,evec = np.linalg.eig(H)
    #Sort eigenvalues and eigenvectors by increasing energy
    inds=np.argsort(eval)
    eval=eval[inds]
    evec=evec[:,inds]
    
    return eval, evec
    
def anharmVsRatio(ratios, N, noisy=False):
    """Calculate transmon parameters as Ej/Ec is varied
    
    Note in particular that anharm is the _relative_ anharmonicity.
    In other words, anharm = (f_10 - f_21) / f_10
    """
    anharms = np.array([])
    E10s = np.array([])
    for i, ratio in enumerate(ratios):
        if noisy:
            print 'Ratio %d' %i
        deltas, H = hamiltonian_realSpace(ratio, N=N)
        eval, evec = analyzeH(H)
        #These energies are in units of E_c
        E10 = eval[1] - eval[0]
        E21 = eval[2] - eval[1]
        #anharm is dimensionless
        anharm = (E21 - E10)/E10
        anharms = np.hstack((anharms, anharm))
        E10s = np.hstack((E10s, E10))
    return np.vstack((ratios, E10s, anharms)).T
    
#One time computation of interpolating functions.
#The idea is to compute E_10 and anharmonicity for a range of Ej/Ec
#ratio. We then make two interpolating functions:
#1. Ej/Ec as a function of relative anharmonicity
#2. E10/Ec as a function of Ej/Ec
#From these, we can turn measurements of frequency and anharmonicity into
#a measurement of Ej and Ec
result = anharmVsRatio(np.linspace(EJ_OVER_EC_RANGE[0], EJ_OVER_EC_RANGE[1], 50), 300)
_anharm2EjOverEc = interpolate.interp1d(result[:,2], result[:,0], kind='cubic')
_EjOverEc2E10OverEc = interpolate.interp1d(result[:,0], result[:,1], kind='cubic')

def anharm2EjOverEc(anharm):
    data = _anharm2EjOverEc(anharm)
    return data
    
def EjOverEc2E10OverEc(EjOverEc):
    data = _EjOverEc2E10OverEc(EjOverEc)
    return data
        
def freqAnharm2EjEc(freq, nonlinearity): 
    anharm = nonlinearity/freq
    """Junction and capacitive characterisitic frequencies as a function of f10 and relative anharmonicity"""
    if anharm>0:
        print 'Anharmonicity specified as positive number... are you sure?'
    EjOverEc = anharm2EjOverEc(anharm)
    E10OverEc = EjOverEc2E10OverEc(EjOverEc)
    Fc = freq / E10OverEc
    Fj = EjOverEc * Fc
    return Fj, Fc
    
    
class Transmon_perturbationTheory(object):
    """Transmon qubit with perturbative formulas for parameters
    
    WARNING:
    This class uses first order perturbation theory to obtain formulas
    relating measured quantities to design parameters. It does NOT use
    the numerical solution provided in this module.
    """


    def __init__(self, f10, anharm):
        if not (isinstance(f10, Value) and isinstance(anharm, Value)):
            raise RuntimeError("Must initialize with Values")
        #Possibly change this to detect positive anharmonicity and negate it
        if anharm>0:
            raise RuntimeError("You probably meant to put in a negative anharmonicity")
        self.f10 = f10
        self.anharm = anharm
    
    #fJ and fC come directly from perturbation theory
    @property
    def fJ(self):
        return (self.f10+self.fC)**2 / (8*self.fC)
    
    @property
    def fC(self):
        return -self.anharm
    
    #Circuit parameters come from Josephson relations and definitions of fJ and fC
    @property
    def C(self):
        return (ELECTRON_CHARGE**2)/(2*PLANCK_CONSTANT*self.fC)
    
    @property
    def L(self):
        return ((PHI_0/(2*np.pi))**2)/(self.fJ*PLANCK_CONSTANT)
    
    @property
    def Z(self):
        """Characteristic impedance"""
        return (self.L/self.C)**0.5
    
    @property
    def Q_zpf(self):
        """Zero point fluctuations of charge"""
        return (HBAR / (2.0*self.Z))**0.5
    
    @property
    def Phi_zpf(self):
        """Zero point fluctuations of flux"""
        return (HBAR*self.Z/2.0)**0.5
    
    def driveCapacitance(self, driveAmplitude, attenuation):
        assert attenuation['dB']>0
        driveAtQubit = driveAmplitude*2*(10**(-attenuation['dB']/20.0))
        Cd = (np.pi/2)*self.C*(2*HBAR/self.Q_zpf)/(driveAtQubit)
        return Cd