import numpy as np
import unittest
import pyle.analysis.readout as readout
import pyle.gateCompiler as gc
 
class testAnalysisReadout(unittest.TestCase):
 
    def setUp(self, channels = 4, stats = 300, nDemod = 10, states=[0,1,2]):
        self.channels = channels
        self.stats = stats
        self.nDemod = nDemod
        self.states = states
        data = np.random.random(size=(channels, stats, nDemod, 2))
        self.randomData = data
        stateData = np.ones((channels, stats, nDemod))
        stateData[...,0] = 0 # make first demod for all channels = 0
        self.stateData = stateData
        centers = {}
        for state in states:
            centers[state] = state
        centersList = [centers]*channels
        self.centersList = centersList
        measure = ['q'+str(k) for k in range(channels)]
        self.measure = measure
        self.qubits = []
        # qubit = {}
        # qubits = [gc.getQubit(self.cxn,gc.Transmon, meas) for meas in measure]
        # for qubit in qubits:
            # qubit['readoutCenters'] = centers.values()
            # qubit['readoutCenterStates'] = centers.keys()
        # self.qubits = qubits
 
    def test_iq_to_z_shape(self):
        """
        checks that (channels, stats, nDemod, IQ) -> 
                    (Channels, stats, nDemod)
        """
        self.assertEqual( readout.IQToZ(self.randomData).shape, 
                         (self.channels, self.stats, self.nDemod) )
                         
    def test_zs_to_states_shape(self):
        """
        checks that (channels, stats, nDemod) -> 
                    (Channels, stats, nDemod)
        """
        zs = readout.IQToZ(self.randomData)
        states = readout.zsToStates(zs, self.centersList)
        self.assertEqual( states.shape, (self.channels, self.stats, self.nDemod))
        
    def test_states_to_probs_shape(self):
        """
        checks that (channels, stats, nDemod) -> 
                    (Channels, stats, nDemod, number of states)
        """
        zs = readout.IQToZ(self.randomData)
        states = readout.zsToStates(zs, self.centersList)
        probs = readout.statesToProbs(states, self.centersList, correlated=False)
        probsCorr = readout.statesToProbs(states, self.centersList, correlated=True)
        self.assertEqual(probs.shape, (self.channels, self.nDemod, len(self.states)))
        self.assertEqual(probsCorr.shape, (self.nDemod, len(self.states)**self.channels))

    def test_IQ_to_probs_shape(self):
        """
        checks that (channels, stats, nDemod, IQ) -> 
                    (Channels, stats, nDemod, number of states)
                    or 
                    (nDemod, number of states ^ channels)
        """
        data = self.randomData
        probs = readout.IQToProbs(data, self.qubits, self.measure, 
                                  centersList = self.centersList, correlated=False)
        probsCorr = readout.IQToProbs(data, self.qubits, self.measure, 
                                  centersList = self.centersList, correlated=True)
        self.assertEqual(probs.shape, (self.channels, self.nDemod, len(self.states)))
        self.assertEqual(probsCorr.shape, (self.nDemod, len(self.states)**self.channels))
        
    def test_average_stats_shape(self):
        """
        checks that (channels, stats, nDemod, IQ) -> 
                    (Channels, nDemod, IQ)
        """
        data = self.randomData
        avgData = readout.averageStats(data)
        self.assertEqual(avgData.shape, (self.channels, self.nDemod, 2))
        
    def test_iq_to_states_shape(self):
        """
        checks that (channels, stats, nDemod, IQ) -> 
                    (Channels, stats  nDemod)
        """
        data = self.randomData
        states = readout.IQToStates(data, self.qubits, self.measure, centersList = self.centersList)
        self.assertEqual(states.shape, (self.channels, self.stats, self.nDemod))
        
    def test_herald_shape(self):
        """
        Check that herald strips the first demod
        """
        stateData = self.stateData
        heraldedStateData,allChannelIndices = readout.heraldStates(stateData)
        self.assertEqual(heraldedStateData.shape, (self.channels, self.stats, self.nDemod-1))
    
    def test_herald_shape_2(self):
        """
        Check that herald strips the right number of measurements
        """
        stateData = np.ones((self.channels, self.stats, self.nDemod))
        stateData[:,:30, 0] = 0 # set first 30 to 0
        heraldedStateData,allChannelIndices = readout.heraldStates(stateData)
        self.assertEqual(heraldedStateData.shape, (self.channels, 30, self.nDemod-1))
        
    def test_herald_shape_3(self):
        """
        Check that heralding takes the system to the MQ ground state,
        and not the ground state for just one of the qubits
        """
        stateData = np.ones((self.channels, self.stats, self.nDemod))
        stateData[0 ,0, 0] = 0 # channel = 0, stat = 0, 
        stateData[1 ,1, 0] = 0 # channel = 1, stat = 1
        stateData[: ,-1, 0] = 0 # channel = 0,1 stat = -1, the real MQ ground state
        heraldedStateData,allChannelIndices = readout.heraldStates(stateData)
        self.assertEqual(heraldedStateData.shape, (self.channels, 1, self.nDemod-1))
        
    def test_correlated_probs(self):
        pass
        
    def test_reshape_for_from_dv(self):
        data = self.randomData
        for_dv = readout.reshapeForDV(data)
        from_dv = readout.reshapeFromDV(for_dv)
        x = data - from_dv
        self.assertFalse(x.any())
          
if __name__ == '__main__':
    unittest.main()