import numpy as np
import unittest
from pyle.workflow import switchSession as pss
from pyle import gateCompiler as gc
from pyle.dataking import zfuncs
from pyle.dataking import singleQubit as sq
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(x) for x in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
        
# Want to be able to pass in sample, measure
# make test functions work on one qubit, have
# them act on all qubits in measure. This way
# we get a whole system view, rather than
# stopping on the first assertion error.
          
class testBase(object):

    def __init__(self, Sample, measure):
        self.Sample = Sample
        self.measure = measure
    
    def setUp(self):
        pass
        
    def tearDown(self):
        pass
        
def run(classToTest, sample, measure=None):
    if measure is None:
        measure = sample['config']
    testClass = classToTest(sample, measure)
    methods = dir(testClass)
    test_methods = [x for x in methods if 'test_' in x]
    for test_method in test_methods:
        for q in measure:
            testClass.setUp()
            testClass.__getattribute__(test_method)(q)
            testClass.tearDown()
        
class registryTest(testBase):

    def setUp(self):
        sample, devs, qubits = gc.loadQubits(self.Sample, self.measure, write_access=False)
        self.devs = devs
        self.qubits = qubits
        qubitDict = {}
        for qubit in qubits:
            qubitDict[qubit.__name__] = qubit
        self.qubitDict = qubitDict
        
    def test_channels(self, qubitName):
        qubit = self.qubitDict[qubitName]
        channels = qubit['channels']
        try: assert( 3 == len(channels) ),"CHANNELS: %s has %d channels" % (qubit.__name__, len(channels))
        except Exception as e: print e
            
    def test_bias_operate_settling_length(self, qubitName):
        qubit = self.qubitDict[qubitName]
        biasOperateSettling = qubit['biasOperateSettling']
        try: assert(biasOperateSettling <= 1000*us),"BIASOPERATESETTLING: %s has settling time %s" % (qubit.__name__, biasOperateSettling)
        except Exception as e: print e
        try: assert(biasOperateSettling >= 100*us),"BIASOPERATESETTLING: %s has settling time %s" % (qubit.__name__, biasOperateSettling)
        except Exception as e: print e 
            
    def test_settling_parameters(self, qubitName):
        qubit = self.qubitDict[qubitName]
        settlingRates = qubit['settlingRates']
        settlingAmplitudes = qubit['settlingAmplitudes']
        try: assert(all([ai!=0*GHz for ai in settlingRates])),"SETTLING RATES: %s has settling rates %s" % (qubit.__name__, settlingRates)
        except Exception as e: print e 
        try: assert(all([ai!=0*GHz for ai in settlingAmplitudes])),"SETTLING AMPS: %s has settling amps %s" % (qubit.__name__, settlingAmplitudes)
        except Exception as e: print e 
    
    def test_carrier_frequency_df(self, qubitName):
        qubit = self.qubitDict[qubitName]
        fc = qubit['fc']
        f10 = qubit['f10']
        df = np.abs(fc['GHz']-f10['GHz'])
        try: assert(0.01 < df < 0.35),"CARRIER DF: %s has carrier df %s" % (qubit.__name__, df)
        except Exception as e: print e 

    def test_timing_lags(self, qubitName):
        qubit = self.qubitDict[qubitName]
        timingLagMeas = qubit['timingLagMeas']
        timingLagUwave = qubit['timingLagUwave']
        timingLagWrtMaster = qubit['timingLagWrtMaster']
        pass
            
    def test_spectroscopy_z_func(self, qubitName):
        qubit = self.qubitDict[qubitName]
        f10 = qubit['f10']
        func = zfuncs.AmpToFrequency(qubit)
        f10_from_zfunc = func(0)
        df = f10 - f10_from_zfunc*GHz
        try: assert(np.abs(df['MHz']) <= 20),"ZFUNC: %s has df %s difference between f10 and zfunc" % (qubit.__name__, df)
        except Exception as e: print e 
            
    def test_nonlinearity(self, qubitName):
        qubit = self.qubitDict[qubitName]
        f10 = qubit['f10']
        f21 = qubit['f21']
        nonlin = f21-f10
        try: assert(-0.15 > nonlin['GHz'] > -0.25),"NONLINEARITY: %s has nonlinearity %s" % (qubit.__name__, nonlin)
        except Exception as e: print e 
            
    def test_overlapping_frequencies(self, qubitName):
        pass
           
    def test_pi_amps(self, qubitName):
        qubit = self.qubitDict[qubitName]
        piAmp = qubit['piAmp']
        piHalfAmp = qubit['piHalfAmp']
        ratio = piHalfAmp / piAmp
        try: assert(0.4 < ratio < 0.6),"PIAMPS: %s has ratio of %s for piHalfAmp to piAmp" % (qubit.__name__, ratio)
        except Exception as e: print e 
                     
    def test_acz_amplitude(self, qubitName):
        qubit = self.qubitDict[qubitName]
        aczKeys = [x for x in qubit.keys() if 'aczAmp' in x]
        for key in aczKeys:
            aczAmp = qubit[key]
            tName = key.strip('aczAmp')
            f21Control = qubit['f21']
            targetQubit = self.qubitDict[tName]
            f10Target = targetQubit.get(qubit.__name__+'aczDetune', targetQubit['f10'])
            # f10Target = self.qubitDict[tName]['f10']
            Delta = f21Control - f10Target
            ratio = -aczAmp / Delta
            # print "f21: %s, f10: %s" % (f21Control, f10Target)
            # print "Delta: %s" % Delta
            # print "ACZ Amp: %s" % aczAmp
            try: assert(0.8 < ratio < 1.2),"ACZ AMPLITUDE: %s -> %s CZ has ratio of %0.2f with expected amplitude" % (qubit.__name__, tName, ratio)
            except Exception as e: print e 
            
    def test_acz_level_crossing(self, qubitName):
        pass
        
    def test_readout_pulse_demod_freq(self, qubitName):
        qubit = self.qubitDict[qubitName]
        demodFreq = qubit['readoutDemodFrequency']
        pulseFreq = qubit['readoutPulseFrequency']
        try: assert(demodFreq == pulseFreq), "READOUT FREQ: %s: Pulse freq: %s, demod freq: %s" % (qubit.__name__, pulseFreq, demodFreq)
        except Exception as e: print e

    def test_readout_pulse_demod_length(self, qubitName):
        qubit = self.qubitDict[qubitName]
        pulseLen = qubit['readoutPulseLen']
        demodLen = qubit['readoutDemodLen']
        try: assert(demodLen >= pulseLen), "READOUT LEN: %s: Pulse len: %s, demod len: %s" % (qubit.__name__, pulseLen, demodLen)
        except Exception as e: print e
        
    def test_z_crosstalk_calibration(self, qubitName):
        """
        Only works for measure passed in the same order as matrix
        """
        qubit = self.qubitDict[qubitName]
        measure = self.measure
        zMat = qubit['calZpaXtalkInv']
                     
class qubitTest(testBase):
    def setUp(self):
        sample, devs, qubits = gc.loadQubits(self.Sample, self.measure, write_access=False)
        self.devs = devs
        self.qubits = qubits
        qubitDict = {}
        for qubit in qubits:
            qubitDict[qubit.__name__] = qubit
        self.qubitDict = qubitDict
        
    def test_readout_fidelity(self, qubitName):
        fids, probs, centers = sq.calculateReadoutCenters(self.Sample, qubitName, stats=900,
                                                          states=[0,1], update=False,
                                                          collect=True, noisy=False)
        avgFidelity = np.mean(fids)
        try: assert(avgFidelity >= 0.9), "READOUT FID: %s: %.4f" % (qubit.__name__, avgFidelity)
        except Exception as e: print e
        
    def test_frequency(self, qubitName):
        data, frequencyError = sq.freqtunerQST(self.Sample, qubitName, update=False, 
                                               noisy=False, stats=600, save=False,
                                               delay = [10*ns, 20*ns, 50*ns, 200*ns], 
                                               collect=True)
        try: assert(np.abs(frequencyError) <= 1.0*MHz), "FREQUENCY: %s: %.2f [MHz]" % (qubit.__name__, frequencyError['MHz'])
        except Exception as e: print e

    def test_delay_z(self, qubitName):
        data, timingError = sq.testdelayZ(self.Sample, qubitName, delay = [-15*ns, -5*ns, 0*ns, 5*ns, 15*ns], 
                      update=False, noisy=False, collect=True, save=False)
                      
        try: assert(np.abs(timingError) <= 2*ns), "TIMING LAG UWAVE: %s: %.2f [ns]" % (qubit.__name__, timingError['ns'])
        except Exception as e: print e
    
# Things to check:
# Reg keys:
# Has channels X
# Bias operate not too long X
# Bias operate settling not too short  X
# settling params set X
# carrier frequency X
# timinglags X
# Spectroscopy z func at zero z is f10 X
# Nonloinearity reasonable X
# Pihalf amp near piamp/2 X
# Acz amp near proper levels X
# Acz doesn't cross X
# Have all acz keys if have 1
# Readout demod near readout f x
# Readout demod Len gequal pulse Len x
# crosstalk calibrated x

# Experiments:
# test delays! X
# Readout zero is small
# Pi pulse has large prob
# Readout fidelity X 
# IQs don't change w repeated readout
# No stim population in zero
# Coarse rb fidelity 
# Acz phases calibrated, single and relative 
# Drag parameter from alpha scan
# Frequency X 
# quick rb scan