''' analysisgroup.py

class: AnalysisGroup

For a group of analysis entries, mostly to plot them together, etc.


'''

import numpy as np, matplotlib.pyplot as plt
from scipy.optimize import curve_fit
# in local folder (FluxNoiseAnalysis)
import meta, ds, ramsey, spectroscopy, echo, rto
from units import *

_t1_popt_guess = (5000, 1.0, 0.04)

class AnalysisGroup(object):
    ''' a container for a list of analysis entries '''
    
    def __init__(self, alist = None, cxn = None, fitEnd=None, fitMin=0):
        ''' first argument: list of analysis entries, e.g. from meta.newEntry '''
        if alist is None:
            alist = []
            
        self.cxn = cxn
        self.alist = []
        for a in alist:
            if type(a) != meta.entryClass:
                self.alist.append(meta.entryClass(cxn, a))
            else:
                self.alist.append(a)
        self.bonusData = []
        self.freqMin = 1e-3
        self.freqMax = 0.4
        self.fitEnd = fitEnd
        self.fitMin = fitMin
    
    def addBonusData(self, ramseys=None, echos=None, t1s=None, spec=None, path=None):
        ''' Add extra ramsey and/or echo data.
        can be dataset numbers/names (in which case you need to supply the path
            argument), or dataset objects (e.g. from ds.getDataset).
        t1 datasets may also be provided to subtract out T1 decay;
            if so, len(t1) must == len(ramseys or echos).
        spectroscopy dataset can be provided to get df/dphi vs frequency.
            Note that only one spec dataset can be provided. If not all ramsey
            datasets apply to one spec dataset, make a new bonus data.
        '''
        assert ramseys or echos, "must provide ramsey or echo datasets"
        if t1s:
            assert (ramseys and len(t1s) == len(ramseys)) or (echos and len(t1s) == len(echos)), "len(t1s) != len(ramseys or echos)"
        thisBonus = {}
        if ramseys:
            thisBonus['ramseyData'] = [self._getDS(x, path) for x in ramseys]
        if echos:
            thisBonus['echoData'] = [self._getDS(x, path) for x in echos]
        if t1s:
            thisBonus['t1Data'] = [self._getDS(x, path) for x in t1s]
        if spec:
            thisBonus['specData'] = self._getDS(spec, path)
        self.bonusData.append(thisBonus)
        
    def runBonusData(self, plot=True, verbose=True, checkSpectroscopy=False, square_df_dphi=False):
        finalOut = []
        for bd in self.bonusData:
            if 't1Data' in bd:
                bd['t1s'] = [self._getT1(x) for x in bd['t1Data']]
            else:
                def nonefunc():
                    while True:
                        yield None
                bd['t1s'] = nonefunc()
            if 'ramseyData' in bd:
                bd['ramseys'] = []
                bd['ramseyFreqs'] = []
                for t1, rd in zip(bd['t1s'], bd['ramseyData']):
                    bd['ramseys'].append(ramsey.ramseyTomo_noLog(rd, T1=t1, plot=False))
                    bd['ramseyFreqs'].append(ds.getParam(rd, 'f10')['GHz'])
            if 'echoData' in bd:
                bd['echos'] = []
                bd['echoFreqs'] = []
                for t1, ed in zip(bd['t1s'], bd['echoData']):
                    bd['echos'].append(echo.analyzeEcho(ed, T1=t1, plot=False))
                    bd['echoFreqs'].append(ds.getParam(ed, 'f10')['GHz'])
            if 'specData' in bd:
                d = bd['specData']
                f_p, A, B = spectroscopy.fitSpectroscopy(d, plot=checkSpectroscopy, quiet=True)
                f_c = ds.getParam(d, 'f10') - ds.getParam(d, 'f21')
                #def dfdphi_closure(f):
                #    return dfdphi_f(f, f_p, f_c['GHz'])
                #bd['dfdphi'] = dfdphi_closure
                bd['dfdphi'] = lambda f: spectroscopy.dfdphi_f(f, f_p, f_c['GHz'])
            else:
                bd['dfdphi'] = lambda x: None
            if verbose:
                print "Dataset\t\tf10 \tdfdphi\t T1 \tTphi1 \tTphi2"
                if 'ramseyData' in bd:
                    for d, f10, t1, tphi in zip(bd['ramseyData'], 
                                bd['ramseyFreqs'], bd['t1s'], bd['ramseys']):
                        print "%s\t%.2f\t%.2f\t%.0f\t%.0f\t%.0f" % (
                            "%s - %s" % (d['path'][-1], ds.getName(d)),
                            f10, bd['dfdphi'](f10), t1, tphi['Tphi1'], tphi['Tphi2'])
                if 'echoData' in bd:
                    for d, f10, t1, tphi in zip(bd['echoData'], 
                                bd['echoFreqs'], bd['t1s'], bd['echos']):
                        print "%s\t%.2f\t%.2f\t%.0f\t%.0f\t%.0f" % (
                            "%s - %s" % (d['path'][-1], ds.getName(d)),
                            f10, bd['dfdphi'](f10), t1, tphi['Tphi1'], tphi['Tphi2'])
            if plot:
                f = plt.figure()
                ax = f.add_subplot(111)
                if 'ramseyData' in bd:
                    if 'specData' in bd:
                        if square_df_dphi:
                            xdata = [bd['dfdphi'](x)**2 for x in bd['ramseyFreqs']]
                            xlabel = "$(df / d\Phi)^2$"
                        else:
                            xdata = [bd['dfdphi'](x) for x in bd['ramseyFreqs']]
                            xlabel = "$df / d\Phi$"
                    else:
                        xdata = bd['ramseyFreqs']
                        xlabel = "Frequency [GHz]"
                    ax.plot(xdata, [x['Tphi1'] for x in bd['ramseys']], 'o',
                            label = "Ramsey T$_{\phi1}$")
                    ax.plot(xdata, [x['Tphi2'] for x in bd['ramseys']], 'o',
                            label = "Ramsey T$_{\phi2}$")
                if 'echoData' in bd:
                    if 'specData' in bd:
                        if square_df_dphi:
                            xdata = [bd['dfdphi'](x)**2 for x in bd['echoFreqs']]
                            xlabel = "($df / d\Phi)^2$"
                        else:
                            xdata = [bd['dfdphi'](x) for x in bd['echoFreqs']]
                            xlabel = "$df / d\Phi$"
                    else:
                        xdata = bd['echoFreqs']
                        xlabel = "Frequency [GHz]"
                    ax.plot(xdata, [x['Tphi1'] for x in bd['echos']], 'o',
                            label = "Echo T$_{\phi1}$")
                    ax.plot(xdata, [x['Tphi2'] for x in bd['echos']], 'o',
                            label = "Echo T$_{\phi2}$")
                if 't1Data' in bd:
                    ax.plot(xdata, bd['t1s'], 'o', label="T$_1$")
                    ymax = np.max(bd['t1s']) * 2
                else:
                    ymax = 20000
                ax.legend(numpoints=1)
                ax.set_xlabel(xlabel, fontsize=19)
                ax.set_ylabel("[ns]", fontsize=19)
                ax.set_ylim(0, ymax)
            retval = {}
            for x in bd:
                if 'Data' not in x:
                    retval[x] = bd[x]
            finalOut.append(retval)
        return finalOut
        
    def ramsey(self, verbose=True, plot=True):
        return self._tphi(meta.processRamsey, verbose, plot)
            
    def spinEcho(self, verbose=True, plot=True):
        return self._tphi(meta.processEcho, verbose, plot, '_echo')
        
    def _tphi(self, func, verbose, plot, suffix=''):
        tphi1s = []
        tphi2s = []
        dfdphis = []
        labels = []
        t1s = []
        for a in self.alist:
            try:
                if 'T1' not in a:
                    meta.processT1(a, plot=False)
                if 'dfdphi' not in a:
                    meta.spectroscopy(a, quiet=True)
                func(a, plot=False)
            except KeyError:
                print "Failed on %s" % a['name']
                continue
            t1s.append(a['T1'])
            tphi1s.append(a['Tphi1'+suffix])
            tphi2s.append(a['Tphi2'+suffix])
            dfdphis.append(a['dfdphi'])
            labels.append(a['name'])
        dfdphi2s = [x**2 for x in dfdphis]
        if plot:
            f = plt.figure()
            ax = f.add_subplot(111)
            ax.plot(dfdphi2s, tphi1s, 'bo', label="$T_{\phi1}$")
            ax.plot(dfdphi2s, tphi2s, 'ro', label="$T_{\phi2}$")
            ax.set_xlabel('$(df / d\phi)^2$ [GHz/$\Phi_0^2$]')
            ax.set_ylabel('$T_\phi$ [ns]')
            ax.set_ylim(0, np.mean(t1s))
            ax.legend(numpoints=1)
        if verbose:
            print "Dataset\t\t\t dfdphi\t T1\t Tphi1\t\tTphi2"
            for i in range(len(t1s)):
                data = (labels[i], dfdphis[i], t1s[i], tphi1s[i], tphi2s[i])
                print "%s\t%.2f\t%.5g\t%.5g\t%.5g\t" % data
        return (labels, dfdphis, t1s, tphi1s, tphi2s)
        
    def fft(self, plotAll=True, plotAverage=True, plotFit=True, plot=True,
            label=None, color=None, ax=None):
        for a in self.alist:
            if 'dfdphi' not in a:
                meta.spectroscopy(a, quiet=True, plot=False)
        results = meta.averageFFTs(self.alist)
        if self.fitEnd:
            results = results[0][:self.fitEnd], results[1][:self.fitEnd]
        if plot:
            if plotAll:
                ax = meta.plotMultipleFFTs(self.alist, skipLabel=True, 
                                            fixColor=color, ax=ax, alpha=0.3)
            elif not ax:
                ax = plt.figure().add_subplot(111)
            if plotAverage:
                ax.loglog(results[0], results[1], 'o', color=color, label=label)
            if plotFit:
                fit = self.fit()
                x = np.logspace(-4, 0)
                ax.loglog(x, rto.aliasFitFunc(x, *fit), '-', color=color)
                ax.loglog(x, np.power(x, fit[0])*10**fit[1], '--', color=color)
        
        return results
        
    def fit(self):
        bins, avg = meta.averageFFTs(self.alist)
        if self.fitEnd:
            bins = bins[:self.fitEnd]
            avg = avg[:self.fitEnd]
        fit = rto.aliasFit(bins, avg, 1.0, freqMin=self.fitMin)
        return fit
        
        
    def _getDS(self, d, path):
        if type(d) == ds.ds_class:
            return d
        else:
            return ds.getDataset(self.cxn.data_vault, d, path)
            
    def _getT1(self, t1d):
        def t1func(t, T1, amp, offset):
            return amp*np.exp(-t/T1) + offset
        popt, pcov = curve_fit(t1func, t1d.data[:,0], t1d.data[:,1], _t1_popt_guess)
        return popt[0]
    
# globals, should put somewhere else instead
basedir = ['', 'Peter', 'analysis']
amlist = [
    basedir + ['transmon-140116am', 'q4'],
    basedir + ['transmon-140117am', 'q4'],
    basedir + ['transmon-140118am', 'q4'],
    basedir + ['transmon-140119am', 'q4'],
    basedir + ['transmon-140120am', 'q3'],
    basedir + ['transmon-140121am', 'q3'],
    basedir + ['transmon-140122am', 'q3'],
    basedir + ['transmon-140123am', 'q1'],
    basedir + ['transmon-140124am', 'q1'],
    basedir + ['transmon-140125am', 'q1'],
    #basedir + ['transmon-140126am', 'q4'], # too close to dfdphi=0
    #basedir + ['transmon-140127am', 'q4'],
]
    
rjlist = [basedir + x for x in [
    ['transmon-130913rj'],
    ['transmon-130915rj'],
    ['transmon-131007rj'],
    ['transmon-131008rj'],
    #['transmon-131111rj'],
    #['transmon-131113rj'],
    #['transmon-131117rj'],
    #['transmon-131202rj'],
    #['transmon-131205rj'],
    #['transmon-131210rj'],    
]]

qslist = [basedir + x for x in [
    ['transmon-130911qs'],
    ['transmon-130912qs'],
    ['transmon-130913qs'],
    ['transmon-130915qs'],
    ['transmon-130920qs'],
    ['transmon-130921qs'],
    ['transmon-130922qs'],
    ['transmon-131012qs'],
    #['transmon-131023qs'],
    #['transmon-131024qs'],
]]

bonusRamseys = [15, 23, 41, 54, 79, 90, 103, 116]
bonusEchos = [16, 24, 42, 55, 80, 91, 104, 117]
bonusT1s = [13, 22, 38, 49, 74, 89, 102, 115]
bonusSpec = 119
bonusPath = ['', 'Peter', 'transmon', '140123am']

def amGroup(cxn):
    g = AnalysisGroup(amlist, cxn, fitEnd=-9, fitMin=1e-3)
    g.addBonusData(bonusRamseys, bonusEchos, bonusT1s, bonusSpec, bonusPath)
    return g
    
def rjGroup(cxn):
    g = AnalysisGroup(rjlist, cxn, fitEnd=-14)
    return g
    
def qsGroup(cxn):
    g = AnalysisGroup(qslist, cxn, fitEnd=-13)
    return g
