'''

FluxNoiseAnalysis/meta.py

Code for handling the storage/recall of flux noise analysis results.

We use the registry for this. Each night of data gets an "entry" (i.e. folder)
in the registry where we keep parameters, results, etc. 

It works as follows: there is a corresponding analysis folder for each sample
folder in the registry, i.e. Peter/transmon/131023qs -> Peter/analysis/131023qs
Important keys:
name -- e.g. '131023qs'
path -- location in data vault
dataset -- RTO dataset number
xxxPath, xxxDataset -- path, number of supporting datasets (spectroscopy, etc)
some keys, such as freqMax, freqMin are parameters to various methods
some, such as alpha, S_phi_1Hz, are results

In the future we should segregate other keys by type, e.g. RTO vs spectroscopy.

Currently I'm also in the process of adding support for multiqubit measurements,
e.g. one RTO dataset that has multiple measure qubits. Each one will get its own
subpath, e.g. 'q1', 'q2' with its own name, path, dataset, xxxPath, xxxDataset,
and so on.

@analysisFunc:
Decorates a function that accepts an analysis registry wrapper and runs, e.g.
RTO processing, spectroscopy fitting, etc. Module global analysisFuncs lists
them all. Decorated functions get required params, optional params, datasetName.
This is used, e.g., when creating a new entry to know what params to prompt for.
Generally path params default to the main path (but can be changed if you want
to use, say, spectroscopy from a different day) and dataset params will default
to a starred dataset matching the dataset name.
(There's not much else that uses this metadata at the moment.)
'''
import csv, types, ast, re

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import is_string_like
from scipy.optimize import curve_fit

from labrad.units import Value
from pyle.plotting import dstools
from pyle.util.registry_wrapper2 import RegistryWrapper

import ds, rto, spectroscopy as spec, ramsey, echo
from units import *



# TODO : change instances of RegistryWrapper to entryClass for later portability
entryClass = RegistryWrapper

def findDatasets(dv, path, *names, **kw):
    ''' find a dataset, recursively starting in _path_,
    containing _names_ in the name '''
    found = []
    dv.cd(path)
    tags = kw.get('tags', None)
    if tags:
        dirs, sets = dv.dir(tags)
    else:
        dirs, sets = dv.dir()
    for d in dirs:
        found.extend(findDatasets(dv, path + [d], *names))
    for s in sets:
        if len([x for x in names if x in s]) == len(names):
            found.append((path, s))
    return found
    
def findStarred(cxn, n, path, measure=None):
    ''' find a starred dataset with name n.
    if measure != none, require measure in ds.measure. '''
    dv = cxn.data_vault
    ctx = dv.context()
    p = dv.packet(context=ctx)
    p.cd(path, context=ctx)
    p.dir('star', context=ctx)
    ans = p.send()
    for dsn in ans.dir[1]:
        if n.lower() in dsn.lower():
            if measure is None:
                return dsn.partition(' ')[0]
            elif type(measure) is int:
                d = ds.getDataset(dv, int(dsn.partition(' ')[0]), path)
                if measure in d['parameters']['measure']:
                    return dsn.partition(' ')[0]
            elif is_string_like(measure):
                d = ds.getDataset(dv, int(dsn.partition(' ')[0]), path)
                if measure in d['parameters']['config'] and any([d['parameters']['config'][x] == measure for x in d['parameters']['measure']]):
                    return dsn.partition(' ')[0]
            else:
                d = ds.getDataset(dv, int(dsn.partition(' ')[0]), path)
                l = [m in d['parameters']['measure'] for m in measure]
                if all(l):
                    return dsn.partition(' ')[0]
    # made it through all datasets, didn't find any
    return 0
    
class _ParamEntry(object):
    def __init__(self, name, type, default=''):
        self.name = name
        self.type = type
        self.default = default
    def processInput(self):
        ''' processes raw string input for this param. called in getValue() '''
        if self.type is list or self.type is dict:
            self.value = ast.literal_eval(self.input)
        else:
            self.value = self.type(self.input)
    def getValue(self):
        ''' query user for value if needed '''
        d = str(self.default)
        self.input = raw_input("%s [%s]: " % (self.name, d))
        if not self.input:
            self.input = d
        self.processInput()
    def changeDefault(self, d):
        self.default = d
        return self

    
def newEntry(s):
    ''' Create a new flux noise entry in the registry, or retrieve existing. '''
    
    def findStarred(n):
        dslist = s._cxn.data_vault.dir('star')[1]
        dslist = [x for x in dslist if n in x]
        return dslist[0].partition(' ')[0] if dslist else '0'
            
    params = [_ParamEntry('name', str).changeDefault('-'.join(s._dir[2:])), 
        _ParamEntry('path', list).changeDefault(s._dir),
        _ParamEntry('dataset', int).changeDefault(findStarred('Ramsey Tomography Oscilloscope')),
        ]
    pdict = dict([(p.name, p) for p in params])
    # get the name first and look in registry
    pdict['name'].getValue()
    user = s._dir[1]
    path = ['', user, 'analysis', pdict['name'].value]
    s._cxn.registry.cd(path, True)
    a = RegistryWrapper(s._cxn, path)
    for p in params:
        if p.name in a.keys():
            p.value = a[p.name]
        elif p.name != 'name':
            p.getValue()
    
    for p in params:
        if p.name not in a or a[p.name] != p.value:
            a[p.name] = p.value
    if a['dataset']:
        d = ds.getDataset(s._cxn.data_vault, a['dataset'], a['path'])
        for m in d['parameters']['measure']:
            print
            print "Now: %s" % d['parameters']['config'][m]
            newQubit(a, d['parameters']['config'][m], interactive=True)
    return a
    
def newQubit(s, q, interactive=True):
    ''' Make a sub-directory for a qubit in a given directory.
    Looks for starred datasets where the measure qubit == q. 
    a -- (RegistryWrapper) original directory, sample or analysis
    q -- qubit name, e.g. 'q1', 'q2', etc.
    interactive -- (bool) query for missing keys?
    '''
    # if sample, create new analysis directory
    if 'analysis' not in s._dir:
        a = newEntry(s)
    # if analysis, get the sample registry
    else:
        a = s
        s = RegistryWrapper(a._cxn, a['path'])
    # make new path
    path = a._dir + [q]
    s._cxn.registry.cd(path, True)
    aq = RegistryWrapper(s._cxn, path)
    if 'name' not in aq or aq['name'] != a['name'] + '.' + q:
        aq['name'] = a['name'] + '.' + q
        aq['qubit'] = q
    for f in analysisFuncs:
        for p in f.requiredParams:
            if 'path' in p.lower():
                pe = _ParamEntry(p, list)
                if p in aq:
                    pe.default = aq[p]
                elif p in a:
                    pe.default = a[p]
                else:
                    pe.default = a['path']
            elif 'dataset' in p.lower():
                pe = _ParamEntry(p, int)
                if p in aq:
                    pe.default = aq[p]
                elif p in a:
                    pe.default = a[p]
                else:
                    pre = p[:p.lower().index('dataset')]
                    if pre:
                        k = pre+'Path'
                    else:
                        k = 'path'
                    if k in aq:
                        pe.default = findStarred(a._cxn, f.datasetName, aq[k], measure=q)
            else:
                pe = _ParamEntry(p, float)
            if interactive and pe.name not in aq:
                pe.getValue()
            else:
                pe.value = pe.default
            # only set it if we're changing it
            if p not in aq or aq[p] != pe.value:
                aq[p] = pe.value
    
def newSoloQubit(a, q):
    ''' For datasets that have interleaved multi-qubit and single-qubit
    measurements. Creates a copy of given qubit.
    '''
    if type(q) is not RegistryWrapper:
        q = a[q]
    qs = q.copy()
    # check to make sure we actually did a solo measurement of this guy
    d = ds.getDataset(a._cxn.data_vault, qs['dataset'], qs['path'])
    for v_tuple in d['variables'][1]:
        if 'single %s' % qs['qubit'] in v_tuple[0]:
            break
    else:
        print "%s is not a solo qubit on this dataset." % qs['name']
        return None
    qs['qubit'] += 's'
    qs['name'] += 's'
    a[qs['qubit']] = qs
    return a[qs['qubit']]
    
    
def loadByDate(a, datestring):
    ''' load a new analysis entry by date string, e.g. 130823 '''
    datestring = str(datestring)
    path = list(a._dir)
    path[-1] = path[-1].rpartition('-')[0] + '-' + datestring
    return RegistryWrapper(a._cxn, path)
    
def loadMultiple(a, getAfter=None, suffix=None, skip=True):
    ''' get multiple analysis entries
    
    getAfter -- string date, e.g., 130901. get only entries after this date
    suffix -- e.g. 'rj', only get with given suffix
    skip -- bool. whether to skip entries with key skip=True
    '''
    root = RegistryWrapper(a._cxn, list(a._dir)[:-1])
    final = []
    for k in root.keys():
        if getAfter and k[9:9+6] < getAfter:
            continue
        if suffix and not k.endswith(suffix):
            continue
        ak = RegistryWrapper(a._cxn, root._dir + [k])
        if skip and 'skip' in ak.keys() and ak['skip']:
            continue
        final.append(ak)
    return final
    
def rescanDatasets(a):
    ''' Attempt to update the dataset numbers for required scans based
    on datasets found (and starred) in the directory. '''
    for func in analysisFuncs:
        if not func.datasetName: continue
        dsKey = [x for x in func.requiredParams if 'dataset' in x.lower()]
        if not dsKey: continue
        dsKey = dsKey[0]
        pathKey = [x for x in func.requiredParams if 'path' in x.lower()][0]
        if dsKey in a and a[dsKey] and type(a[dsKey]) is int: continue
        if pathKey not in a or not a[pathKey]:
            a[pathKey] = a['path']
        if 'qubit' in a.keys():
            measure = a['qubit']
        else:
            measure = None
        dsnum = int(findStarred(a._cxn, func.datasetName, a[pathKey], measure=measure))        
        if not dsnum: continue
        print "FOUND %s for %s" % (func.datasetName, a['name'])
        a[dsKey] = dsnum
        
    
def plotMultipleFFTs(entries, colorBySuffix=True, extraLabel = None,
                    skipLabel=False, fixColor=False, ax=None, alpha=1, **kwargs):
    ffts = [fft(a, plot=False, update=False, quiet=True, **kwargs) for a in entries]
    if ax is None:
        f = plt.figure()
        ax = f.add_subplot(111)
    colors = {}
    count = {}
    clist = ['r', 'b', 'g', 'c', 'm', 'y', 'k']
    mlist = ['o', 'v', 's', 'x', 'D', '^', '+', '<', '>']
    for a,r in zip(entries, ffts):
        label = a['name'][9:]
        marker = 'o'
        color = None
        if extraLabel is not None:
            def replacer(match):
                v = a[match.group(1)]
                return "%.2f" % v if type(v) is float else str(v)
            label += re.sub(r'<(.*?)>', replacer, extraLabel)
        if skipLabel:
            label = None
        if colorBySuffix:
            suf = a._dir[-1][-2:]
            if suf not in colors:
                colors[suf] = clist[len(colors) % len(clist)]
            if suf not in count:
                count[suf] = 0
            else:
                count[suf] += 1
            marker = mlist[count[suf] % len(mlist)]
            color = colors[suf]
        if fixColor:
            color = fixColor
        ax.loglog(r['bins'], r['phi'], marker=marker, label=label, color=color, alpha=alpha)
        ax.legend(numpoints=1)
    if 'plotTheta' in kwargs and kwargs['plotTheta']:
        f2 = plt.figure()
        ax2 = f2.add_subplot(111)
        [ax2.plot(r['theta'], '.', label=a['name']) for r in ffts]
        ax2.legend(numpoints=1)
    return ax
    
def averageFFTs(entries, **kwargs):
    ''' average multiple ffts

        **kwargs -- passed through to FFT function
    '''
    rlist = []
    for a in entries:
        rlist.append(fft(a, quiet=True, plot=False, update=False, **kwargs))
    rlist = [(x['bins'], x['phi']) for x in rlist]
    return rlist[0][0], rto.averageFFTs(rlist)
    
def correlation(a1, a2, **kwargs):
    ''' determine the correlation between two (time trace) signals '''
    d1 = ds.getDataset(a1._cxn.data_vault, a1['dataset'], a1['path'])
    d2 = ds.getDataset(a2._cxn.data_vault, a2['dataset'], a2['path'])
    if 'qubit' in a1:
        q1 = a1['qubit']
    else:
        q1 = None
    if 'qubit' in a2:
        q2 = a2['qubit']
    else:
        q2 = None
    f1 = rto.doFFTstuff(d1, q=q1, timeOnly=True, quiet=True, plot=False)
    f2 = rto.doFFTstuff(d2, q=q2, timeOnly=True, quiet=True, plot=False)
    t1 = d1['parameters']['timeStep']
    t2 = d2['parameters']['timeStep']
    if t1 != t2:
        raise ValueError("Time steps not equal in a1, a2")
    #corr = d1 - d2
    l = min(f1.shape[0], f2.shape[0])
    #print f1.shape, f2.shape
    corr = np.correlate(f1[:l], f2[:l], **kwargs)
    #corr = plt.xcorr(f1[:l], f2[:l], maxlags=None, **kwargs)
    return corr
        
analysisFuncs = []
def analysisFunc(requiredParams=None, optionalParams=None, datasetName=None):
    ''' A decorator factory. Returns a decorator, which is a function that must return a function.
    The purpose of this one is to denote a function that analyzes datasets, and has
    certain required parameters in the registry. '''
    if requiredParams is None:
        requiredParams = []
    if optionalParams is None:
        optionalParams = []
    def decorator(func):
        global analysisFuncs
        analysisFuncs.append(func)
        func.analysisFunc = True
        func.requiredParams = requiredParams
        func.optionalParams = optionalParams
        func.datasetName = datasetName
        return func
    return decorator        

def generateReport(alist, filename = 'analysisReport.csv'):
    ''' generate a csv, doing some minimal coercing of data types '''
    colNames = []
    units = {}
    data = []
    
    def _coerce(v, k):
        if v is None:
            return ''
        if k in units:
            return str(v[units[k]])
        else:
            return str(v)
    
    for a in alist:
        row = [''] * len(colNames)
        for k in a.keys():
            if k in colNames:
                row[colNames.index(k)] = _coerce(a[k], k)
            else:
                colNames.append(k)
                if isinstance(a[k], Value):
                    units[k] = a[k].units
                row.append(_coerce(a[k], k))
        data.append(row)
    for i in range(len(colNames)):
        if colNames[i] in units:
            colNames[i] += ' [%s]' % units[colNames[i]]
    with open(filename, 'wb') as f:
        writer = csv.writer(f)
        writer.writerows([colNames] + data)
    

@analysisFunc(['path','dataset',], 
                optionalParams=['dfdphi', 'dfdDacAmp', 'freqMax', 'freqMin',
                            'iterationStartCutoff', 'iterationEndCutoff'],
                datasetName="Ramsey Tomography Oscilloscope")
def fft(a, update=True, quiet=False, **kwargs):
    d = ds.getDataset(a._cxn.data_vault, a['dataset'], a['path'])
    kw2 = dict(a)
    kw2.update(kwargs)
    kw2.pop('dataset')
    if quiet:
        kw2['plot'] = False
        kw2['quiet'] = True
    if 'qubit' in a:
        kw2['q'] = a['qubit']
    r = rto.doFFTstuff(d, **kw2)
    if update:
        for k in ['alpha', 'S_phi_1Hz', 'S_GHz_1Hz', 'missFraction', 'white']:
            if k in r and r[k] is not None:
                a[k] = r[k]
        try:
            a['S_DacAmp_1Hz'] = a['S_GHz_1Hz'][GHz**2/Hz] / a['dfdDacAmp'] * (DacAmp_u**2/Hz)
        except KeyError:
            if not quiet:
                print "Could not calculate S_DacAmp_1Hz: need spectroscopy fit."
    return r

def processSpectroscopy(a):
    return spectroscopy(a, quiet=False, plot=True)

@analysisFunc(['specPath', 'specDataset',],
                optionalParams=['guessA', 'guessB', 'guessC'],
                datasetName="Spectroscopy Z Func")
def spectroscopy(a, quiet=False, **kwargs):
    d = ds.getDataset(a._cxn.data_vault, a['specDataset'], a['specPath'])
    if 'V_op' not in kwargs:
        kwargs['V_op'] = 0
    dfdphi, dfdDacAmp = spec.fitSpectroscopy(d, returnDacAmp = True, 
                                                quiet=quiet, **kwargs)
    a['dfdphi'] = dfdphi
    a['dfdDacAmp'] = dfdDacAmp

@analysisFunc(['ramseyPath', 'ramseyDataset',], datasetName = 'Ramsey')
def processRamsey(a, plot=True):
    d = ds.getDataset(a._cxn.data_vault, a['ramseyDataset'], a['ramseyPath'])
    if 'T1' in a:
        t1 = a['T1']
    else:
        t1 = None
    r = ramsey.ramseyTomo_noLog(d, T1=t1, plot=plot, title=a['name'])
    for k in ['Tphi1', 'Tphi2']:
        if k in r and r[k]:
            a[k] = r[k]
    
@analysisFunc(['ramseyPath', 'ramseyDataset',], datasetName = 'Ramsey')
def extractSfromRamsey(a, plot=True):
    ''' Extract S from a Ramsey dataset. Looks for alpha, T1 in the registry.
    
    Units: micro-Phi_0 / sqrt(Hz)
    '''
    d = ds.getDataset(a._cxn.data_vault, a['ramseyDataset'], a['ramseyPath'])
    S = ramsey.fitForS(d, a['T1'], alpha=np.abs(a['alpha']),
                        dfdphi=a['dfdphi'], plot=plot, title=a['name'])
    a['S_ramsey'] = np.sqrt(S*10**12) * Unit('uPhi / Hz^1/2')
    return a['S_ramsey']
    
@analysisFunc(['ramseyPath', 'ramseyDataset',], datasetName = 'Ramsey')
def extractAlphaFromRamsey(a, plot=True):
    ''' Extract alpha from a Ramsey dataset, Looks for S_phi_1Hz, T1 in the registry.
    '''
    d = ds.getDataset(a._cxn.data_vault, a['ramseyDataset'], a['ramseyPath'])
    alpha = ramsey.fitForAlpha(d, a['T1'], S=a['S_phi_1Hz'], dfdphi=a['dfdphi'], plot=plot)
    a['alpha_ramsey'] = alpha
    return a['alpha_ramsey']

@analysisFunc(['rabiPath', 'rabiDataset',], datasetName = 'Rabi Oscillation')
def processRabi(a):
    d = ds.getDataset(a._cxn.data_vault, a['rabiDataset'], a['rabiPath'])
    amps = np.unique(d.data[:,1])
    probs = []
    times = []
    for amp in amps:
        probs.append(d.data[d.data[:,1] == amp, 2])
        times.append(d.data[d.data[:,1] == amp, 0])
    ax = plt.figure().add_subplot(111)
    for time, prob, amp in zip(times, probs, amps):
        ax.plot(time, prob, '.-', label="amp=%.3f" % amp)
    ax.set_xlabel("Pulse length [ns]")
    ax.set_ylabel("Probability |1>")
    ax.legend(numpoints=1)
    
@analysisFunc(['fluxPath', 'fluxDataset',], datasetName = 'Spectroscopy flux func')
def processFlux(a):
    d = ds.getDataset(a._cxn.data_vault, a['fluxDataset'], a['rabiPath'])
    spec.fluxSpectroscopy(d, plot=True,)
     
    
@analysisFunc(['T1Path', 'T1Dataset'], datasetName = 'T1')
def processT1(a, plot=True, update=True):
    if 'T1Dataset' not in a and 'T1dataset' in a:
        a['T1Dataset'] = a['T1dataset']
        del a['T1dataset']
    if 'T1Path' not in a and 'T1path' in a:
        a['T1Path'] = a['T1path']
        del a['T1path']
    d = ds.getDataset(a._cxn.data_vault, a['T1Dataset'], a['T1Path'])
    def t1func(t, T1, amp, offset):
        return amp*np.exp(-t/T1) + offset
    popt, pcov = curve_fit(t1func, d.data[:,0], d.data[:,1], (5000, 1.0, 0.04))
    if plot:
        t = np.linspace(0, d.data[:,0].max(), 200)
        ax = plt.figure().add_subplot(111)
        ax.plot(d.data[:,0], d.data[:,1], '.-', label="Data")
        ax.plot(t, t1func(t, *popt), '-', label="Fit, T1 = %.1f us" % (popt[0]/1000,))
        ax.legend(numpoints=1)
    if update:
        a['T1'] = popt[0] * ns

@analysisFunc(['echoPath', 'echoDataset'], datasetName = 'Spin Echo')
def processEcho(a, plot=True, update=True):
    d = ds.getDataset(a._cxn.data_vault, a['echoDataset'], a['echoPath'])
    if 'T1' in a:
        t1 = a['T1']
    else:
        t1 = None
    r = echo.analyzeEcho(d, T1=t1, plot=plot)
    if update:
        for k in ['Tphi1', 'Tphi2']:
            if k in r and r[k]:
                a[k + '_echo'] = r[k]
    
        
