'''

rabi.py -- extract phase noise from Rabi data.

'''

import numpy as np
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt

import pyle.fitting.fitting as fitting

def rabi_analysis(dvwSet, nfftpoints=4096, plot=True, axis_is_time=True):
    guess = (0.9, 0.1, np.pi)
    d = dvwSet
    if d.indep[0][1] != 'ns':
        raise TypeError("First axis (time) must be in units of ns. Or you can fix the code.")
    amps = np.unique(d[:,1])
    noises = np.zeros_like(amps)
    errs = np.zeros_like(amps)
    freqs = np.zeros_like(amps)
    for i, amp in enumerate(amps):
        subd = d[d[:,1] == amp]
        freq = fitting.maxFreq(subd[:,(0,2)], nfftpoints)
        func = makeFunc(freq['GHz'])
        popt, pcov = curve_fit(func, subd[:,0], subd[:,2], guess)
        err = np.sqrt(np.diag(pcov))
        noises[i] = popt[1]
        errs[i] = err[1]
        freqs[i] = freq['MHz']
    if plot:
        ax = plt.figure().add_subplot(111)
        if axis_is_time:
            xdata = 1.0/freqs * 1000.0
            ax.set_xlabel("Noise Timescale (ns)")
        else:
            xdata = freqs
            ax.set_xlabel("Noise Frequency (MHz)")
        
        ax.plot(xdata, noises, 'o')
        ax.set_ylabel("Noise power (rad^2 / Hz)")
        
def print_amps_and_freqs(d, nfftpoints=4096):
    amps = np.unique(d[:,1])
    print "Amp\tf(MHz)\t1/freq(ns)"
    print "--------------------------"
    for i, amp in enumerate(amps):
        subd = d[d[:,1] == amp]
        freq = fitting.maxFreq(subd[:,(0,2)], nfftpoints)
        print "%.3f\t%.1f\t%.1f" % (amp, freq['MHz'], (1.0/freq)['ns'])

def one_rabi(d, freq=None, timescale=None, index=None, noisy=True, plot=True, nfftpoints=4096):
    ''' plot one rabi oscillation + fit at a given freq, timescale, or index
    '''
    if not freq and not timescale and not index:
        raise ValueError("Must have one of freq, timescale, index!")
    amps = np.unique(d[:,1])
    freqs = np.zeros_like(amps)
    # determine the what amp rabi oscillation we're looking for
    if timescale:
        freq = (1.0/timescale)['GHz']
    elif freq:
        freq = freq['GHz']
    if freq:
        for i, amp in enumerate(amps):
            subd = d[d[:,1] == amp]
            f = fitting.maxFreq(subd[:,(0,2)], nfftpoints)
            freqs[i] = f['GHz']
        index = np.abs(freqs-freq).argmin()
    amp = amps[index]
    subd = d[d[:,1] == amp]
    freq = fitting.maxFreq(subd[:,(0,2)], nfftpoints)
    func = makeFunc(freq['GHz'])
    guess = (0.9, 0.1, np.pi)
    popt, pcov = curve_fit(func, subd[:,0], subd[:,2], guess)
    err = np.sqrt(np.diag(pcov))
    
    if noisy:
        print "Rabi amp = %s" % amp
        print "Frequency = %s" % freq
        print "Fit params: %s" % str(popt)
        print "Fit errors: %s" % str(np.sqrt(np.diag(pcov)))
    if plot:
        ax = plt.figure().add_subplot(111)
        ax.plot(subd[:,0], subd[:,2], 'o')
        t = np.linspace(subd[:,0].min(), subd[:,0].max(), 2000)
        ax.plot(t, func(t, *popt), '-')
        ax.plot(t, envelope_func(t, *popt), '--')
        ax.set_xlabel('time [ns]')
        ax.set_ylabel('P1')
        ax.set_title('f = %.0f MHz = %.0f ns' % (freq['MHz'], (1.0/freq)['ns']))
    return popt, np.sqrt(np.diag(pcov))
            
def cosfunc(t, A, s, phi0, fr):
    return A*np.exp(-0.25*s*t)*np.cos(2*np.pi*fr*t + phi0) + 0.5
    
def makeFunc(fr):
    return lambda t, A, s, phi0: cosfunc(t, A, s, phi0, fr)

def envelope_func(t, A, s, *unusedArgs):
    return A*np.exp(-0.25*s*t) + 0.5
    
