'''

FluxNoiseAnalysis/fft.py

Code for the FFT of the noise data.

'''
import sys

import numpy as np
import scipy
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
from matplotlib import is_string_like

import pyle.analysis.signalProcessing as sp

from units import *

def doFFTstuff(dataset, q=None, dfdphi = None, plotTheta=False, plotTomo=False, 
                plotRawFFT=False, aliasFreq=None, plot=True, inMicroPhi0=True,
                holdTime=None, freqMin=None, avgPoints=100, chunks=1,
                smooth=False, iterationEndCutoff=None, iterationStartCutoff=None,
                rawOnly=False, timeOnly=False, quiet=False, **kwargs):
    d = dataset.data
    # cutoff ranges
    if iterationEndCutoff:
        d = d[:iterationEndCutoff]
    if iterationStartCutoff:
        d = d[iterationStartCutoff:]
    # which qubit are we looking at?
    if is_string_like(q):
        if q.endswith('s'):    # special value for solo qubit ("singleMeasure")
            i = -1
        else:
            m = dataset['parameters']['config'].index(q)
            i = dataset['parameters']['measure'].index(m)
    elif type(q) is int:
        i = dataset['parameters']['measure'].index(q)
    else:
        i = 0        
    phase = dataToPhase(d, i)
    # fill in any misses with zeros
    filled = np.zeros(d[:,0].max() + 1)
    filled[d[:,0].astype(int)] = phase
    if timeOnly:
        return filled
    n = filled.shape[0] - phase.shape[0]
    if not quiet:
        print "Num misses: %s / %s -- %.3f %%" % (n, filled.shape[0], float(n) / filled.shape[0] * 100)
    missFraction = float(n) / filled.shape[0]
    if plotTheta:
        f = plt.figure()
        ax = f.add_subplot(111)
        if d.shape[1] > 2:
            ax.plot(d[:,3], phase, '.')
        else:
            ax.plot(phase, '.')
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('$\Delta$ Phase [rad]')
    if plotTomo:
        f = plt.figure()
        ax = f.add_subplot(111)
        plotTomography(dataset, bins=dataset.parameters.stats)
    if aliasFreq is None and 'freqMax' in kwargs:
        aliasFreq = kwargs['freqMax']
    if holdTime is None:
        holdTime = dataset.parameters.holdTime['ns']
    t = len(d[:,0]) * dataset.parameters.timeStep['s'] * sec
    f_nyquist = 1 / dataset.parameters.timeStep['s'] / 4    # div by 4 because the interleaved method doubles the timestep
    fft = sp.dftInterleavedBootstrapped(filled, t, chunks=chunks)
    if rawOnly:
        fft['S'] *= ((1.0/holdTime)**2)
        return fft
    bins, avg = _avgFunc(fft, n_pts=avgPoints)
    if smooth:
        if smooth is True:
            smooth = 5
        avg = movingAverage(avg, window=smooth)
    Sf = avg * ((1.0/holdTime)**2)    # GHz^2 / Hz
    maxFreq = bins[np.isfinite(avg)][-1]
    
    if dfdphi:
        Sphi = Sf*((1/dfdphi)**2)
        units = Phi0_u**2 / Hz
        if inMicroPhi0:
            Sphi *= 10**12
            units = uPhi0_u**2 / Hz
        plotData = Sphi    
    else:
        Sphi = None
        plotData = Sf
        units = GHz**2 / Hz
        
    alpha, off, white = aliasFit(bins, plotData, f_nyquist, freqMin=freqMin)
    fitFunc = lambda x: aliasFitFunc(x, alpha, off, white, f_nyquist)
        
    if not quiet:
        print
        print "alpha = %s" % alpha
        print "Noise power at 1 Hz: %s" % np.sqrt(10**off*units)
        print
    
    if plotRawFFT:
        f = plt.figure()
        ax = f.add_subplot(111)
        ax.loglog(fft.frequencies, fft.S, '.')
        ax.set_xlabel('Frequency [Hz]')
        ax.set_ylabel('S [Phase / Hz]')
                
    if plot:
        f = plt.figure()
        ax = f.add_subplot(111)
        legs = []
        # plot data
        ax.loglog(bins, plotData, 'bo')
        legs.append('FFT')
        # plot 
        ax.loglog(bins, bins**alpha * 10**off, 'r--')
        legs.append('1/f only, $\\alpha = %.2f $' % alpha)
        ax.loglog(bins, fitFunc(bins), 'b--')
        legs.append('1/f + alias + floor, floor = $%.2f$ ' % white)
        if freqMin:
            ax.loglog([freqMin], [fitFunc(freqMin)], 'rs')
            legs.append('fit lower bound')
        if dfdphi and inMicroPhi0:
            ax.set_ylabel(r'$S_\Phi$  [$ ({\mu\Phi_0})^2 $ / Hz]', fontsize=19)
        elif dfdphi:
            ax.set_ylabel(r'$S_\Phi$ [$\Phi_0$ / Hz]', fontsize=19)
        else:
            ax.set_ylabel(r'$S$ [GHz$^2$ / Hz]', fontsize=19)

        ax.legend(legs, numpoints=1)
        ax.set_xlabel('Frequency [Hz]', fontsize=19)
        
    r = {}
    r['bins'] = bins
    r['GHz'] = Sf
    if dfdphi:
        r['phi'] = Sphi
        
    r['alpha'] = alpha
    r['offset'] = off
    r['white'] = white * units
    if dfdphi:
        r['S_phi_1Hz'] = 10**off * units
        r['S_GHz_1Hz'] = r['S_phi_1Hz'][Phi0_u**2/Hz] * dfdphi * (GHz**2/Hz)
    else:
        r['S_phi_1Hz'] = None
        r['S_GHz_1Hz'] = 10**off
    r['fitFunc'] = fitFunc
    r['f_nyquist'] = f_nyquist
    r['theta'] = phase
    r['missFraction'] = missFraction
    if aliasFreq:
        r['maxFreq'] = maxFreq
    else:
        r['maxFreq'] = None
        
    return r
        
def _avgFunc(ret_data, n_pts = 100, f_min_exp = -4, f_max_exp = 0):
    bins = np.logspace(f_min_exp, f_max_exp, n_pts)
    bins_l = bins.copy()
    bins_l[1:] = bins_l[:-1]
    bins_l[0] = 0
    
    avg = np.zeros(n_pts)
    for i, (l, h) in enumerate(zip(bins_l, bins)):
        avg[i] = ret_data.S[np.logical_and(ret_data.frequencies >= l, ret_data.frequencies< h)].mean()
        
    return bins, avg
    
def _getSlope(bins, avg, freqMin=0, freqMax=sys.float_info.max, returnOffset = False):
    def line(x, m, b):
        return x*m + b
    ind = np.logical_and(bins >= freqMin, bins <= freqMax)
    x = np.log10(bins[ind])
    y = np.log10(avg[ind])
    x = x[np.isfinite(y)]
    y = y[np.isfinite(y)]
    popt = (-1, y[-1])
    #plt.plot(x, y, 'o')
    #plt.plot(x, line(x, *popt))
    #print x, y
    popt, pcov = scipy.optimize.curve_fit(line, x, y, popt)
    if returnOffset:
        return popt
    else:
        return popt[0]
        
def _fancyFit(bins, avg, aliasFreq, freqMin=0):
    ''' attempt to account for aliasing from the FFT '''
    slope, off = _getSlope(bins, avg, freqMin=freqMin, freqMax=aliasFreq, returnOffset=True)
    maxFreq = bins[np.isfinite(avg)][-1]
    def func(freq):
        r = freq**slope * 10**off
        #ind = freq > aliasFreq
        #r[ind] = r[ind] + (2*maxFreq-freq[ind])**slope * 10**off
        r += (2*maxFreq-freq)**slope * 10**off
        return r
    return slope, off, func
        
def aliasFit(bins, avg, f_nyquist, S_guess = 1.0, white_guess=0.01, alpha_guess=-1.0,
        fix_alpha=None, fix_S=None, fix_white=None, freqMin = 0):
    ''' fit the full aliased signal, with white noise floor '''
    sigma = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
    i = np.logical_and(bins >= freqMin, np.isfinite(avg))
    x = bins[i]
    y = avg[i]
    sigma = sigma[i]
    def func(freq, alpha, A, white):
        if fix_alpha is not None:
            alpha = fix_alpha
        if fix_S is not None:
            A = fix_S
        if fix_white is not None:
            white = fix_white
        return aliasFitFunc(freq, alpha, A, white, f_nyquist)
    popt = [alpha_guess, S_guess, white_guess]
    popt, pcov = curve_fit(func, x, y, popt, sigma=1.0/sigma)
    return popt
    
def aliasFitFunc(freq, alpha, A, white, f_nyquist=1):
    return np.power(freq, alpha) * 10**A + np.power(2*f_nyquist - freq, alpha) * 10**A + white
    
def noAliasFitFunc(freq, alpha, A, white=None, f_nyquist=None):
    return np.power(freq, alpha) * 10**A
    
def dataToPhase(dataset, i=0, xOnly=False, yOnly=False, unwrap=True):
    ''' takes an RTO dataset and returns phases vs time, in cycles (not radians) '''
    if type(dataset) is np.ndarray:
        d = dataset
    else:
        d = dataset.data
        
    # since we've moved to multiqubit dataking, which columns we use are
    # no longer fixed. i is the index of the qubit we're looking at. the cols
    # are in this order:
    #       iteration, x1, ..., xn, y1, ..., yn, [soloX, soloY,] time
    if i == -1:
        xcol = -3
        ycol = -2
    else:
        xcol = 1 + i
        ycol = d.shape[1] / 2 + i
        
    if xOnly:
        phase = np.arccos(2*d[:,xcol]-1)
    elif yOnly:
        phase = np.arcsin(2*d[:,ycol]-1)
    else:
        phase = np.arctan2(2*d[:,xcol]-1, 2*d[:,ycol]-1)
    if unwrap:
        phase = np.unwrap(phase)
        initialMean = phase[:50].mean()
        phase -= initialMean
        phase += pi
        phase %= 2*pi
        phase -= pi
    phase /= 2*pi
    phase -= phase.mean()
    return phase
    
def checkPhases(s, ds=None, returnPhases=False, returnTime=False, returnDataset=False, plot=True, unwrap=True):
    invalidateCache()
    dv = s._cxn.data_vault
    if ds is None:
        # find last RTO scan
        sets = findDatasets(dv, None, 'Ramsey Tomography')
        ds = int(sets[-1][1].partition(' ')[0])
        print "Using dataset: %d" % ds
    if type(ds) is not AttrDict:
        ds = getDataset(dv, ds, [])
    d = dataToPhase(ds, unwrap=unwrap)
    if plot:
        plt.plot(d*2.0*np.pi)
        plt.xlabel('Iteration')
        plt.ylabel('Phase (rad)')
    if returnPhases:
        if returnTime:
            return (d, ds.parameters.holdTime['ns'])
        else:
            return d
    if returnDataset:
        return ds
        
def movingAverage(d, window=5):
    # remove the nans first
    d2 = d.copy()
    d2 = d2[np.isfinite(d2)]
    d2c = d2.copy()
    d2 = np.convolve(d2, np.ones(window)/float(window), mode='same')
    d2[0:window/2] = d2c[0:window/2]
    d2[-window/2:] = d2c[-window/2:]
    d3 = d.copy()
    d3[np.isfinite(d)] = d2
    return d3

def averageFFTs(datalist):
    ''' average multiple FFTs together.
    
    datalist -- [(frequencies, values), (frequencies, values), ...]
    #note that we interpolate to the x-values (frequencies) of the first dataset
    '''
    datalist = [(np.log10(f), np.log10(v)) for f, v in datalist]
    # check freq bins
    f0, badlist = datalist[0][0], []
    for i, (f, v) in enumerate(datalist):
        if not np.allclose(f, f0):
            badlist.append(i)
    if badlist:
        raise ValueError("The following entries in the data list do not have"
                        + " the same frequency bins: %s" % str(badlist))
    # total up, keeping track of how many non-nan's we add
    final, count = np.zeros_like(f0), np.zeros_like(f0)
    for f, v in datalist:
        data = np.isfinite(v)
        final[data] += v[data]
        count += np.isfinite(v) * 1
    final = final / count
    final = np.power(10, final)
    return final
