import sys, csv
import numpy as np
import scipy
from scipy.optimize import curve_fit
from numpy import floor, log, exp, pi, cos
import matplotlib.pyplot as plt
import ramseyIntegrals
import ast, types
import random

import labrad.units
from labrad.units import Value,Unit
sec, ns, GHz, Hz, MHz, V = (Unit(s) for s in ['s','ns','GHz','Hz','MHz', 'V'])
# define Phi_0 and micro-Phi_0
Phi0_v = labrad.units.hbar*pi/labrad.units.e
Phi0_u = Unit('Phi', 1, Phi0_v)
uPhi0_u = Unit('uPhi', 10**-6, Phi0_u)
# A DAC Amplitude of 1 corresponds to 380 mV
DacAmp_u = Unit('DacAmp', 0.38, 'V')

import signalProcessing as sp
import mathHelpers as mh

from scipy.optimize import leastsq
from pyle.registry import AttrDict
from pyle.plotting import dstools
from pyle.util.registry_wrapper2 import RegistryWrapper
from scipy.optimize import curve_fit

## util
ds_cache = {}
def getDataset(dv, num, path):
    k = '%s - %s' % (num, path)
    if k not in ds_cache:
        ds_cache[k] = dstools.getDeviceDataset(dv, num, path)
    return ds_cache[k]
def invalidateCache():
    global ds_cache
    ds_cache = {}
    

###################################
## TRANSMONS
###################################

## 2013-05-07 The transmon datasets are of a slightly different format.

def findDatasets(dv, path, *names):
    ''' find a dataset, recursively starting in _path_,
    containing _names_ in the name '''
    found = []
    dv.cd(path)
    dirs, sets = dv.dir()
    for d in dirs:
        found.extend(findDatasets(dv, path + [d], *names))
    for s in sets:
        if len([x for x in names if x in s]) == len(names):
            found.append((path, s))
    return found
    
def newEntry(s, **kwargs):
    ''' Create a new flux noise entry in the registry.
    keywords recognized: name, path, dataset, specPath, specDataset '''
    
    def findStarred(n):
        dslist = s._cxn.data_vault.dir('star')[1]
        dslist = [x for x in dslist if n in x]
        return dslist[0].partition(' ')[0] if dslist else '0'
    
    class Param(object):
        def __init__(self, name, type):
            self.name = name
            self.type = type
        def default(self):
            ''' generates default argument for this param '''
            return ''
        def processInput(self):
            ''' processes raw string input for this param. called in getValue() '''
            if self.type is list or self.type is dict:
                self.value = ast.literal_eval(self.input)
            else:
                self.value = self.type(self.input)
        def getValue(self):
            ''' query user for value if needed '''
            if self.name in kwargs:
                return kwargs[self.name]
            d = str(self.default())
            self.input = raw_input("%s [%s]: " % (self.name, d))
            if not self.input:
                self.input = d
            self.processInput()
        def changeDefault(self, f):
            self.default = types.MethodType(f, self)
            return self
            
    params = [Param('name', str).changeDefault(lambda self: '-'.join(s._dir[2:])), 
        Param('path', list).changeDefault(lambda self: s._dir),
        Param('specPath', list).changeDefault(lambda self: s._dir),
        Param('dataset', int).changeDefault(lambda self: findStarred('Ramsey Tomography Oscilloscope')),
        Param('specDataset', int).changeDefault(lambda self: findStarred('Spectroscopy Z func')),
        ]
    pdict = dict([(p.name, p) for p in params])
    # get the name first and look in registry
    pdict['name'].getValue()
    user = s._dir[1]
    path = ['', user, 'analysis', pdict['name'].value]
    s._cxn.registry.cd(path, True)
    a = RegistryWrapper(s._cxn, path)
    for p in params:
        if p.name in a.keys():
            p.value = a[p.name]
        elif p.name != 'name':
            p.getValue()
    
    for p in params:
        a[p.name] = p.value
    return a
    
def loadByDate(a, datestring):
    ''' load a new analysis entry by date string, e.g. 130823 '''
    datestring = str(datestring)
    path = list(a._dir)
    path[-1] = path[-1].rpartition('-')[0] + '-' + datestring
    return RegistryWrapper(a._cxn, path)
    
def loadMultiple(a, getAfter=None, suffix=None):
    ''' get multiple analysis entries
    
    getAfter -- string date, e.g., 130901. get only entries after this date
    suffix -- e.g. 'rj', only get with given suffix
    '''
    root = RegistryWrapper(a._cxn, list(a._dir)[:-1])
    final = []
    for k in root.keys():
        if getAfter and k[9:9+6] < getAfter:
            continue
        if suffix and not k.endswith(suffix):
            continue
        final.append(RegistryWrapper(a._cxn, root._dir + [k]))
    return final
    
    
def plotMultipleFFTs(entries, colorBySuffix=True, **kwargs):
    ffts = [fft(a, plot=False, **kwargs) for a in entries]
    f = plt.figure()
    ax = f.add_subplot(111)
    colors = {}
    count = {}
    clist = plt.rcParams['axes.color_cycle']
    mlist = ['o', 'v', 's', 'x', 'D', '^', '+', '<', '>']
    for a,r in zip(entries, ffts):
        if colorBySuffix:
            suf = a._dir[-1][-2:]
            if suf not in colors:
                colors[suf] = clist[len(colors) % len(clist)]
            if suf not in count:
                count[suf] = 0
            else:
                count[suf] += 1
            ax.loglog(r['bins'], r['phi'], label=a._dir[-1][9:], marker=mlist[count[suf] % len(mlist)], color=colors[suf])
        else:
            ax.loglog(r['bins'], r['phi'], 'o', label=a._dir[-1][9:])
        ax.legend()
    if 'plotTheta' in kwargs and kwargs['plotTheta']:
        f2 = plt.figure()
        ax2 = f2.add_subplot(111)
        [ax2.plot(r['theta'], '.', label=a._dir[-1]) for r in ffts]
        ax2.legend(numpoints=1)
    

def getRTOdatasets(s):
    if 'analysis' in s._dir:
        dv = s._cxn.data_vault
        ctx = dv.context()
        dv.cd(s['path'])
        d = dv.dir()[1]
    else:
        d = s._cxn.data_vault.dir()[1]
    named = [x for x in d if 'RTO Sweep' in x]
    times = [int(x.rpartition(':')[2][:-2]) for x in named]
    inds = [int(x.partition('-')[0]) for x in named]
    # only get contiguous datasets
    inds.reverse()
    keeps = []
    for i in inds:
        if not keeps:
            keeps.append(i)
        elif keeps[-1] - i == 1:
            keeps.append(i)
        else:
            break
    if not keeps:
        raise Exception("Could not find any RTO Sweep datasets")
    print "using these: %s" % keeps
    return keeps
    
def plotRTOtomo(s, datasets=None):
    if not datasets:
        datasets = getRTOdatasets(s)
    for ds in datasets:
        #plt.figure()
        d = getDataset(s._cxn.data_vault, ds, [])
        plotTomography(d)
        plt.title(d.parameters.holdTime)
        
def plotRTOtomo2(s, datasets=None):
    COLORS = ['red', 'green', 'blue', 'cyan', 'magenta', 'goldenrod', 'black', 'Brown', 'Grey', 'OliveDrab', 'DarkCyan', 'DarkMagenta', 'MediumSlateBlue']
    if not datasets:
        datasets = getRTOdatasets(s)
        datasets.reverse()
    plt.figure()
    plt.xlim([-1,1])
    plt.ylim([-1,1])
    i = 0
    for ds in datasets:
        d = getDataset(s._cxn.data_vault, ds, [])
        plt.plot(2*d.data[:,1]-1, 2*d.data[:,2]-1, '.', alpha=0.5, color=COLORS[i%len(COLORS)])
        cx, cy = (2*d.data[:,1]-1).mean(), (2*d.data[:,2]-1).mean()
        plt.text(cx, cy, str(d['parameters']['holdTime']), horizontalalignment='center', verticalalignment='center')
        i += 1
    
def plotRTOsweep(s, datasets=None, plot=True, plotUnwrap=True):
    ''' plot std dev vs hold time from RTO sweep datasets. '''
    if not datasets:
        datasets = getRTOdatasets(s)
    # now plot the datasets
    r = [checkPhases(s, ds=ds, returnPhases=True, returnTime=True, plot=plot) for ds in datasets]
    plt.figure()
    plt.plot([p[1] for p in r], [np.std(p[0]) for p in r], 'o')
    plt.xlabel('Hold Time [ns]')
    plt.ylabel(r'STD of $\Delta \theta$')
    if plotUnwrap:
        [checkUnwrap(s, x, title=p[1]) for x, p in zip(datasets, r)]
        
analysisFuncs = []
def analysisFunc(requiredParams=None, optionalParams=None):
    ''' A decorator factory. Returns a decorator, which is a function that must return a function.
    The purpose of this one is to denote a function that analyzes datasets, and has
    certain required parameters in the registry. '''
    if requiredParams is None:
        requiredParams = []
    if optionalParams is None:
        optionalParams = []
    def decorator(func):
        global analysisFuncs
        analysisFuncs.append(func)
        func.analysisFunc = True
        func.requiredParams = requiredParams
        func.optionalParams = optionalParams
        return func
    return decorator        
        
        
resultList = ['alpha', 'S_phi_1Hz', 'S_GHz_1Hz', 'S_DacAmp_1Hz', 'missFraction', 'dfdphi', 'dfdDacAmp', 'dfdFastBias']
def runComparison(cxn, interactive=True, directory=['', 'Peter', 'analysis'], verbose=True, filename="fluxNoiseResults.csv", ignoreBefore='130901'):
    '''
    
    Run all the data we have, and generate a CSV for comparison. 
    
    interactive (bool) -- query user for missing info (dataset numbers, etc)
    verbose (bool) -- print out params 
    
    '''
    # get list of datasets to do
    reg = cxn.registry
    reg.cd(directory)
    dirs = reg.dir()[0]
    sets = [RegistryWrapper(cxn, directory + [x]) for x in dirs]
    results = [['dataset'] + resultList]
    # run through them
    for a in sets:
        if 'ignore' in a.keys() and a['ignore']:
            continue
        date = a['name'][9:9+6]
        if ignoreBefore and date < ignoreBefore:
            continue
        if verbose:
            print "\n\t\tDataset: %s" % a['name']
        # run all possible analyses
        for af in analysisFuncs:
            if verbose:
                print "\n\tAttempting %s" % af.func_name
            # check params
            hasAll = all([checkParam(a, p, interactive=interactive, verbose=verbose) for p in af.requiredParams])
            hasOptional = all([checkParam(a, p, interactive=False, verbose=verbose) for p in af.optionalParams])
            if hasAll:
                af(a, quiet=True)
        # extract the results
        results.append([a.name] + [getResult(a, r) for r in resultList])
    # save data as csv
    with open(filename, 'w') as f:
        [f.write(','.join(r) + '\r\n') for r in results]
    

def checkParam(a, param, interactive=True, verbose=True):
    ''' check that a given param is present in the registry wrapper. query for it if it doesn't exist. '''
    originalParam = param
    while '.' in param:
        folder, dot, param = param.partition('.')
        a = a[folder]
    if param in a.keys():
        if verbose:
            print "Found %s = %s" % (originalParam, a[param])
        return True
    if interactive:
        ans = raw_input("Need value for %s: " % originalParam)
        if ans:
            a[param] = Value(ans)
            return True
    return False
    
def getResult(a, result, verbose=True):
    ''' extract a result ''' 
    originalResult = result
    while '.' in result:
        folder, dot, result = result.partition('.')
        a = a[folder]
    if result in a.keys():
        if type(a[result]) is Value:
            return str(a[result].value)
        else:
            return str(a[result])
    if verbose:
        print "No result: " + originalResult
    return ''
    
def getResultsColumn(col, filename="fluxNoiseResults.csv", order='grouped', returnNames=False):
    ''' get a single results column, by name. see resultList (module global) for possible names.
    
    order -- "grouped" (by name suffix, e.g. rj or qs) or "date" (by date)
    returnNames (boolean) -- return the dataset names as well '''
    with open(filename, 'r') as f:
        d = list(csv.DictReader(f))
        if order == 'grouped':
            d = sorted(d, key = lambda x: x['dataset'][15:] + x['dataset'][:15])
        elif order == 'date':
            d = sorted(d, key = lambda x: x['dataset'])
        r = [float(x[col]) for x in d]
        if returnNames:
            r = (r, [x['dataset'][9:] for x in d])
        return r
        

@analysisFunc(['specDataset', 'specPath'], optionalParams=['guessA', 'guessB', 'guessC'])
def spectroscopy(a, quiet=False, **kwargs):
    ds = getDataset(a._cxn.data_vault, a['specDataset'], a['specPath'])
    if 'V_op' not in kwargs:
        kwargs['V_op'] = 0
    dfdphi, dfdDacAmp = fitSpectroscopy(ds, returnDacAmp = True, quiet=quiet, **kwargs)
    a['dfdphi'] = dfdphi
    a['dfdDacAmp'] = dfdDacAmp

@analysisFunc(['dataset', 'path',], 
                optionalParams=['dfdphi', 'dfdDacAmp', 'freqMax', 'freqMin', 'iterationStartCutoff', 'iterationEndCutoff'])
def fft(a, update=True, quiet=False, **kwargs):
    ds = getDataset(a._cxn.data_vault, a['dataset'], a['path'])
    kw2 = dict(a)
    kw2.update(kwargs)
    kw2.pop('dataset')
    if quiet:
        kw2['plot'] = False
        kw2['quiet'] = True
    r = doFFTstuff(ds, **kw2)
    if update:
        for k in ['alpha', 'S_phi_1Hz', 'S_GHz_1Hz', 'missFraction']:
            a[k] = r[k]
        try:
            a['S_DacAmp_1Hz'] = a['S_GHz_1Hz'][GHz/Hz**0.5] / a['dfdDacAmp'] * (DacAmp_u/Hz**0.5)
        except KeyError:
            if not quiet:
                print "Could not calculate S_DacAmp_1Hz: need spectroscopy fit."
    return r

#############################
# SPECTROSCOPY
#############################

def getExtrema(data, axisCol, indepCol, valueCol, max=True):
    ''' Gets the value of axisCol where valueCol is max/min, keeping indepCol as independent.
    Useful for reducing a 2D dataset to a 1D one (e.g. flux, Z spectroscopy). '''
    ivals, avals = [], []
    if max:
        f = np.argmax
    else:
        f = np.argmin
    for indepval in np.unique(data[:,indepCol]):
        ivals.append(indepval)
        thisRow = data[data[:,indepCol]==indepval]
        avals.append(thisRow[f(thisRow[:,valueCol]), axisCol])
    return np.array([ivals, avals]).T
    
def fluxSpectroscopy(dataset, plot=True, V_op=None, guessA = 1.0, guessB = 0.0, whichMeasure=0):
    ''' Fits and extracts operating slope from a flux spectroscopy dataset.
    f_10 [GHz] = f_p * Sqrt( Cos(A * V_bias - B) ) - f_c'''
    q = dataset['parameters'][dataset['parameters']['config'][dataset['parameters']['measure'][whichMeasure]]]
    d = getExtrema(dataset.data, 0,1,2, True)
    fc_GHz = (q.f10 - q.f21)['GHz']
    if V_op is None:
        V_op = q.biasOperate['V']
    
    def freqFunc(v, f_p, A, B):
        return f_p * np.sqrt(np.abs(np.cos(A*v - B))) - fc_GHz
    popt = (d[:,1].max(), guessA, guessB)
    popt, pcov = scipy.optimize.curve_fit(freqFunc, d[:,0], d[:,1], popt)
    f_p, A, B = popt
    dfdv = -f_p*np.sin(A*V_op-B)*A/(2*np.sqrt(np.cos(A*V_op-B)))
    #print popt
    if not quiet:
        print "Period: %.2f V" % (np.pi / A)
        print "df/dv: %.2f GHz/V" % dfdv
        print "df/dphi: %.2f GHz/Phi0" % (dfdv * np.pi/A)
    if plot:
        plt.figure()
        plt.plot(d[:,0], d[:,1], 'bo')
        plt.plot(d[:,0], freqFunc(d[:,0], *popt), 'b-', linewidth=2.0)
        plt.plot(np.linspace(0,1.0,1000), freqFunc(np.linspace(0, 1.0,1000), *popt), 'b-', linewidth=2.0)
        plt.plot(V_op, freqFunc(V_op, *popt), 'ro', markersize=10)
        dpts = np.linspace(V_op-0.1, V_op+0.1)
        plt.plot(dpts, dfdv*(dpts-V_op)+freqFunc(V_op, *popt), 'r--', linewidth=2.0)
        plt.xlabel('Flux Bias [V]', fontsize=19)
        plt.ylabel("Frequency [GHz]", fontsize=19)
    
def fitSpectroscopy(dataset, whichMeasure=0, plot=False, V_op = None, asym=False, 
                    zbMin=None, zbMax=None, returnDacAmp=False,
                    guessA = 1.0, guessB = 0.5, guessC = 0.0, quiet=False):
    ''' Fits a spectroscopy dataset to the qubit model.
    the data should have been taken with the Spectroscopy Z func.
    Qubit model:    f(phi) = f_p * Cos( pi * phi - offset )^(1/2) - f_c
    or:             f(phi) = f_p * Cos( A * V_bias - B )^(1/2) - f_c
    ARGUMENTS:
        dataset
        whichMeasure=0
            which qubit in 'measure' we're working with
            i.e. we use config[measure[whichMeasure]]
            if you only measured one (as you should have), leave whichMeasure=0.
        plot=False
            if True, plot the data + fit
        V_op=None
            if not None, calculate df/dPhi0 at V_op and return it
            
    RETURNS:
        if V_op is None:
            (f_p, A, B) -- from model above
        else:
            dfdPhi0 -- evaluated at V_op
            if returnDacAmp:
                (dfdPhi0, dfdDacAmp) -- df/dDacAmp = df/dphi * A / pi
            
    for an asymmetric squid we have
    I_m^2 = I_1^2 + I_2^2 + 2I_1 I_2 Cos(2*pi*phi / Phi0)
        '''
    if type(dataset) is list:
        d = np.concatenate([x.data for x in dataset])
        dataset = dataset[0]
    else:
        d = dataset.data
    if zbMin is not None:
        d = d[d[:,1] >= zbMin]
    if zbMax is not None:
        d = d[d[:,1] <= zbMax]
    q = dataset['parameters'][dataset['parameters']['config'][dataset['parameters']['measure'][whichMeasure]]]
    # TODO: check for dataset name / scan type?
    # get the maximum frequencies
    zbs = []
    freqs = []
    for zbVal in np.unique(d[:,1]):
        inds = d[:,1] == zbVal
        sd = d[ inds ]
        zbs.append(zbVal)
        freqs.append(sd[ sd[:,2] == sd[:,2].max(), 0][0])
        
    fc_GHz = (q.f10 - q.f21)['GHz']
    if asym:
        def freqFunc(zb, amp, A, B, C):
            return amp * np.power((C + np.cos(A*zb + B)**2), 1.0/4.0) - fc_GHz
        popt = [max(freqs), guessA, guessB, guessC]
    else:
        def freqFunc(zb, amp, A, B):
            return amp*np.sqrt(np.abs(np.cos(A*zb + B))) - fc_GHz
        popt = [max(freqs), guessA, guessB]
    try:
        popt, pcov = scipy.optimize.curve_fit(freqFunc, np.array(zbs), np.array(freqs), popt)
        if asym:
            f_p, A, B, C = popt
            if not quiet:
                print "A = %s, B = %s, f_p = %s, C = %s" % (A, B, f_p, C)
        else:
            f_p, A, B = popt
            if not quiet:
                print "A = %s, B = %s, f_p = %s" % (A, B, f_p)
        fit = True
    except RuntimeError:
        print "Unable to fit"
        fit = False
        f_p, A, B, C, D= None, None, None, None, None
    
    retval = (f_p, A, B)
    if V_op is not None:
        if fit:
            if asym:
                dfdphi = -np.pi/2 * f_p * D * np.cos(A*V_op+B) * np.sin(A*V_op+B) / np.power(C+D*np.cos(A*V_op+B)**2, 3.0/4.0)
            else:
                dfdphi = -np.pi/2 * f_p * np.sin(A*V_op + B) / np.sqrt(np.cos(A*V_op + B))
            dfdDacAmp = dfdphi * A / np.pi
        else:
            idx = (np.abs(np.array(zbs) - V_op)).argmin()
            dfdDacAmp = np.polyfit(zbs[idx-2:idx+3], freqs[idx-2:idx+3], 1)[0]
            dfdphi = 0
        if returnDacAmp:
            retval = dfdphi, dfdDacAmp
        else:
            retval = dfdphi
        if not quiet:
            print "df/dphi = %.2f" % dfdphi
            print "df/dDacAmp = %.2f" % dfdDacAmp
        
    if plot:
        f = plt.figure()
        ax = f.add_subplot(111)
        ax.plot(zbs, freqs, 'bo')
        leg = ['Data']
        if fit:
            maxzb = -B / A
            zbps = np.linspace(min(zbs + [maxzb*1.2]), max(zbs + [maxzb*1.2]))
            ax.plot(zbps, freqFunc(np.array(zbps), *popt), 'b-', linewidth=2)
            #pts = np.linspace(-2,2,1000)
            #ax.plot(freqFunc(pts, *popt), pts, 'b--', linewidth=2)
            leg.append('Fit ($\sqrt{\cos}$)')
        if V_op is not None:
            zbps2 = np.linspace(V_op - 0.1, V_op + 0.1)
            if fit:
                ax.plot(V_op, freqFunc(V_op, *popt), 'ro', markersize=10.0)
                ax.plot(zbps2, dfdphi/np.pi*A * (zbps2 - V_op) + freqFunc(V_op, *popt), 'r-', linewidth = 2)
            else:
                ax.plot(V_op, freqs[idx], 'ro', markersize=10.0)
                ax.plot(zbps2, dfdphi * (zbps2 - V_op) + freqs[idx], 'r-', linewidth = 2)
            
            leg += ['Operating Point', 'Calculated Slope']
        ax.set_ylabel("Frequency [GHz]");
        ax.set_xlabel("Dac Amplitude");
        plt.legend(leg, numpoints=1)

    return retval
    
def fitAsym(zbs, freqs, fc_GHz):
    def freqFunc(zb, amp, A, B, C, D):
        return amp * np.power((C + D * np.cos(A*zb + B)), 1.0/4.0) - fc_GHz
    

#############################
# FFT
#############################

def calibration(a, bandwidth = 0.01, **kwargs):
    ac = a['RTO Calibration']
    ds = getDataset(a._cxn.data_vault, ac['dataset'], a['path'])
    #kwargs['plot'] = kwargs.get('plot', False)
    #kwargs['plotRawFFT'] = kwargs.get('plotRawFFT', True)
    kwargs['rawOnly'] = kwargs.get('rawOnly', True)
    r = doFFTstuff(ds, **kwargs)
    f = plt.figure()
    ax = f.add_subplot(111)
    ax.loglog(r['frequencies'], r['S'], '.')
    
    ax.set_xlabel("Frequency [Hz]")
    ax.set_ylabel("Power [GHz^2 / Hz]")
    #if 'powerFloor' not in ac.keys():
    #    pf = input("Power floor, GHz^2/Hz: ")
    #    ac['powerFloor'] = pf * (GHz*GHz/Hz)
    power, bins = integrateRange(r['frequencies'], r['S'], ac['freqApplied']['Hz'], bandwidth, returnBins=True)
    #power -= ac['powerFloor']['GHz^2/Hz'] * bandwidth
    ax.loglog(r['frequencies'][bins], r['S'][bins], '.')
    f.show()
    print "Integrated power: %.2e GHz^2" % power
    print "Cosine power for %.2e V = %.2e V^2" % (ac['voltApplied']['V'], ac['voltApplied']['V']**2/2)
    print "df/dV = %.2e GHz/V" % np.sqrt(power / (ac['voltApplied']['V']**2/2))
    print "df/dV from Z spectroscopy = %.2e GHz/V" % np.sqrt(a['dfdDacAmp']/0.38)**2
    #print "Multiplied: %.2e V^2" % (power / (a['dfdDacAmp']*0.38)**2)
    #print "Expected: %.2e V => %.2e V^2" % (ac['voltApplied'][V], ac['voltApplied'][V]**2)

    return r
    
def integrateRange(x, y, center, bandwidth, returnBins=False):
    bins = np.logical_and(x > center-bandwidth/2.0, x < center+bandwidth/2.0)
    widths = np.gradient(x)
    if returnBins:
        return np.sum(y[bins]*widths[bins]), bins
    return np.sum(y[bins]*widths[bins])
    
def calibrationRaw(a, plot=True, fit=True, **kwargs):
    ac = a['RTO Calibration']
    ds = getDataset(a._cxn.data_vault, ac['dataset'], a['path'])
    phases, holdTime = checkPhases(a, ds=ds, returnPhases=True, returnTime=True, plot=False)
    times = ds.data[:,0] * ds.parameters.timeStep
    freqs = phases / holdTime
    if fit:
        def cos(t, f, amp, phase, off):
            return amp*np.cos(2*pi*f * t + phase) + off
        popt = [ac['freqApplied']['Hz'], (freqs.max() - freqs.min())/2, np.pi/2, freqs.mean()]
        popt, pcov = curve_fit(cos, times, freqs, popt)
        print "Fit: f=%.4e Hz, amp=%.2e GHz" % (popt[0], popt[1])
    if plot:
        plt.figure()
        plt.plot(times, freqs, 'o', label='Raw Calibration Data')
        if fit:
            plt.plot(times, cos(times, *popt), '--', label="Fit: f=%.2e Hz, amp=%.2e GHz" % (popt[0], popt[1]))
        plt.xlabel('Time [s]', fontsize=19)
        plt.ylabel('$\Delta$ Frequency [GHz]', fontsize=19)
        l = 3.0/ac['freqApplied']['Hz']
        startTime = random.random() * (times.max()-l)
        plt.xlim([startTime, startTime+l])
        plt.legend(numpoints=1, loc='best')
    ampFreq = np.sqrt(2) * np.std(freqs)
    ampVolt = ac['voltApplied']['V']
    print "freq amp = %.2e GHz" % ampFreq
    print "volt amp = %.2e V" % ampVolt
    print "df/dV = %.2e GHz / V" % (ampFreq/ampVolt)
    print "compare to df/dV from Z spectroscopy = %.2e GHz/V" % (a['dfdDacAmp']*0.38)
    ac['frequency amplitude'] = ampFreq * GHz
    ac['df/dV'] = ampFreq/ampVolt * (GHz/V)
    
    return times, phases
    
def movingAverage(d, window=5):
    # remove the nans first
    d2 = d.copy()
    d2 = d2[np.isfinite(d2)]
    d2c = d2.copy()
    d2 = np.convolve(d2, np.ones(window)/float(window), mode='same')
    d2[0:window/2] = d2c[0:window/2]
    d2[-window/2:] = d2c[-window/2:]
    d3 = d.copy()
    d3[np.isfinite(d)] = d2
    return d3
    
def plotTomography(dataset, bins=50):
    d = dataset.data.copy()
    d[:,1] = np.around(d[:,1] * 150) / 150
    d[:,2] = np.around(d[:,2] * 150) / 150
    H, x, y = np.histogram2d(2*d[:,2]-1, 2*d[:,1]-1, bins=bins)
    ex = [-1,1,-1,1]
    H[H==0] = np.nan
    plt.imshow(H, extent=ex, interpolation='nearest')
    plt.colorbar()
    plt.show()
    
def dataToPhase(dataset, xOnly=False, yOnly=False, unwrap=True):
    ''' takes an RTO dataset and returns phases vs time, in cycles (not radians) '''
    if type(dataset) is np.ndarray:
        d = dataset
    else:
        d = dataset.data
    if xOnly:
        phase = np.arccos(2*d[:,1]-1)
    elif yOnly:
        phase = np.arcsin(2*d[:,2]-1)
    else:
        phase = np.arctan2(2*d[:,2]-1, 2*d[:,1]-1)
    if unwrap:
        phase = np.unwrap(phase)
        initialMean = phase[:50].mean()
        phase -= initialMean
        phase += pi
        phase %= 2*pi
        phase -= pi
    phase /= 2*pi
    phase -= phase.mean()
    return phase
    
def checkPhases(s, ds=None, returnPhases=False, returnTime=False, returnDataset=False, plot=True, unwrap=True):
    invalidateCache()
    dv = s._cxn.data_vault
    if ds is None:
        # find last RTO scan
        sets = findDatasets(dv, None, 'Ramsey Tomography')
        ds = int(sets[-1][1].partition(' ')[0])
        print "Using dataset: %d" % ds
    if type(ds) is not AttrDict:
        ds = getDataset(dv, ds, [])
    d = dataToPhase(ds, unwrap=unwrap)
    if plot:
        plt.plot(d*2.0*np.pi)
        plt.xlabel('Iteration')
        plt.ylabel('Phase (rad)')
    if returnPhases:
        if returnTime:
            return (d, ds.parameters.holdTime['ns'])
        else:
            return d
    if returnDataset:
        return ds
        
def checkUnwrap(s, ds, title=None):
    plt.figure()
    checkPhases(s, ds, unwrap=True)
    checkPhases(s, ds, unwrap=False)
    plt.legend(['Unwrap=True', 'Unwrap=False'], numpoints=1)
    if title:
        plt.title(title)

def doFFTstuff(dataset, dfdphi = None, plotTheta=False, plotTomo=False, plotRawFFT=False, aliasFreq=None,
                plot=True, inMicroPhi0=True, holdTime=None, freqMin=None, avgPoints=100, chunks=1, smooth=False,
                iterationEndCutoff=None, iterationStartCutoff=None, rawOnly=False, quiet=False, **kwargs):
    d = dataset.data
    # cutoff ranges
    if iterationEndCutoff:
        d = d[:iterationEndCutoff]
    if iterationStartCutoff:
        d = d[iterationStartCutoff:]
    phase = dataToPhase(d)
    # fill in any misses with zeros
    filled = np.zeros(d[:,0].max() + 1)
    filled[d[:,0].astype(int)] = phase
    n = filled.shape[0] - phase.shape[0]
    if not quiet:
        print "Num misses: %s / %s -- %.3f %%" % (n, filled.shape[0], float(n) / filled.shape[0] * 100)
    missFraction = float(n) / filled.shape[0]
    if plotTheta:
        f = plt.figure()
        ax = f.add_subplot(111)
        if d.shape[1] > 2:
            ax.plot(d[:,3], phase, '.')
        else:
            ax.plot(phase, '.')
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('$\Delta$ Phase [rad]')
    if plotTomo:
        f = plt.figure()
        ax = f.add_subplot(111)
        plotTomography(dataset, bins=dataset.parameters.stats)
    if aliasFreq is None and 'freqMax' in kwargs:
        aliasFreq = kwargs['freqMax']
    if holdTime is None:
        holdTime = dataset.parameters.holdTime['ns']
    t = d[:,0].max() * dataset.parameters.timeStep['s'] * sec
    f_nyquist = 1 / dataset.parameters.timeStep['s'] / 4    # div by 4 because the interleaved method doubles the timestep
    fft = sp.dftInterleavedBootstrapped(filled, t, chunks=chunks)
    if rawOnly:
        fft['S'] *= ((1.0/holdTime)**2)
        return fft
    bins, avg = avgFunc(fft, n_pts=avgPoints)
    if smooth:
        if smooth is True:
            smooth = 5
        avg = movingAverage(avg, window=smooth)
    Sf = avg * ((1.0/holdTime)**2)    # GHz^2 / Hz
    maxFreq = bins[np.isfinite(avg)][-1]
    
    if dfdphi:
        Sphi = Sf*((1/dfdphi)**2)
        if inMicroPhi0:
            Sphi *= 10**12
        if aliasFreq:
            alpha, off, fitFunc = fancyFit(bins, Sphi, aliasFreq, freqMin=freqMin)
        else:
            alpha, off = getSlope(bins, Sphi, returnOffset=True, freqMin=freqMin)
            fitFunc = lambda x: x**alpha * 10**off
    else:
        Sphi = None
        if aliasFreq:
            alpha, off, fitFunc = fancyFit(bins, Sf, aliasFreq, freqMin=freqMin)
        else:
            alpha, off = getSlope(bins, Sf, returnOffset=True, freqMin=freqMin)
            fitFunc = lambda x: x**alpha * 10**off
            
    if dfdphi and inMicroPhi0:
        units = uPhi0_u / Hz**0.5
    elif dfdphi:
        units = Phi0_u / Hz**0.5
    else:
        units = GHz / Hz**0.5
    if not quiet:
        print
        print "alpha = %s" % alpha
        print "at 1 Hz: %s" % (np.sqrt(10**off)*units)
        print
    
    if plotRawFFT:
        f = plt.figure()
        ax = f.add_subplot(111)
        ax.loglog(fft.frequencies, fft.S, '.')
        ax.set_xlabel('Frequency [Hz]')
        ax.set_ylabel('S [Phase / Hz]')
                
    if plot:
        f = plt.figure()
        ax = f.add_subplot(111)
        legs = []
        if Sphi is not None:
            ax.loglog(bins, Sphi, 'bo')
            if inMicroPhi0:
                ax.set_ylabel('$S^2$  [$ (\\frac{\\phi}{\\mu\\Phi_0})^2 $ / Hz]')
            else:
                ax.set_ylabel('S^2 [$\phi^2$ / Hz]')
            legs.append('FFT of noise data')
        else:
            ax.loglog(bins, Sf, 'bo')
            ax.set_ylabel('S [GHz^2 / Hz]')
            legs.append('FFT of arctan2(Y,X)')
        if aliasFreq:
            ax.loglog(bins, bins**alpha * 10**off, 'r--')
            legs.append('non-alias-corrected')
            
            limit_pts = [aliasFreq, maxFreq]
            if freqMin:
                limit_pts.append(freqMin)
            ax.loglog(limit_pts, [fitFunc(x) for x in limit_pts], 'rs')
            legs.append('fit cutoff, reflection point')
        ax.loglog(bins, fitFunc(bins), 'b--')
        legs.append('fit, $\\alpha = %.2f $' % alpha)
        ax.legend(legs, numpoints=1)
        ax.set_xlabel('Frequency [Hz]')
        
    r = {}
    r['bins'] = bins
    r['GHz'] = Sf
    if dfdphi:
        r['phi'] = Sphi
        
    r['alpha'] = alpha
    r['offset'] = off
    if dfdphi:
        r['S_phi_1Hz'] = np.sqrt(10**off) * units
        r['S_GHz_1Hz'] = r['S_phi_1Hz'][Phi0_u/Hz**0.5] * dfdphi * (GHz/Hz**0.5)
    else:
        r['S_GHz_1Hz'] = np.sqrt(10**off) * units
    r['fitFunc'] = fitFunc
    r['f_nyquist'] = f_nyquist
    r['theta'] = phase
    r['missFraction'] = missFraction
    if aliasFreq:
        r['maxFreq'] = maxFreq
        
    return r
        
def avgFunc(ret_data, n_pts = 100, f_min_exp = -4, f_max_exp = 0):
    bins = np.logspace(f_min_exp, f_max_exp, n_pts)
    bins_l = bins.copy()
    bins_l[1:] = bins_l[:-1]
    bins_l[0] = 0
    
    avg = np.zeros(n_pts)
    for i, (l, h) in enumerate(zip(bins_l, bins)):
        avg[i] = ret_data.S[np.logical_and(ret_data.frequencies >= l, ret_data.frequencies< h)].mean()
        
    return bins, avg
    
def getSlope(bins, avg, freqMin=0, freqMax=sys.float_info.max, returnOffset = False):
    def line(x, m, b):
        return x*m + b
    ind = np.logical_and(bins >= freqMin, bins <= freqMax)
    x = np.log10(bins[ind])
    y = np.log10(avg[ind])
    x = x[np.isfinite(y)]
    y = y[np.isfinite(y)]
    popt = (-1, y[-1])
    #plt.plot(x, y, 'o')
    #plt.plot(x, line(x, *popt))
    #print x, y
    popt, pcov = scipy.optimize.curve_fit(line, x, y, popt)
    if returnOffset:
        return popt
    else:
        return popt[0]
        
def fancyFit(bins, avg, aliasFreq, freqMin=0):
    ''' attempt to account for aliasing from the FFT '''
    slope, off = getSlope(bins, avg, freqMin=freqMin, freqMax=aliasFreq, returnOffset=True)
    maxFreq = bins[np.isfinite(avg)][-1]
    def func(freq):
        r = freq**slope * 10**off
        #ind = freq > aliasFreq
        #r[ind] = r[ind] + (2*maxFreq-freq[ind])**slope * 10**off
        r += (2*maxFreq-freq)**slope * 10**off
        return r
    return slope, off, func
        
def aliasFit(bins, avg, f_nyquist, S_guess = 1, white_guess=0.01, alpha_guess=-1.0,
        fix_alpha=None, fix_S=None, fix_white=None,
        freqMin = 0):
    ''' fit the full aliased signal, with white noise floor '''
    sigma = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
    i = np.logical_and(bins >= freqMin, np.isfinite(avg))
    bins = bins[i]
    avg = avg[i]
    sigma = sigma[i]
    # our fitting function
    def func(freq, alpha, A, white):
        if fix_alpha is not None:
            alpha = fix_alpha
        if fix_S is not None:
            A = fix_S
        if fix_white is not None:
            white = fix_white
        return np.power(freq, alpha) * A + np.power(2*f_nyquist - freq, alpha) * A + white
    popt, pcov = curve_fit(func, bins, avg, [alpha_guess, S_guess, white_guess], sigma=1/sigma)
    return func, popt

###################################
##Ramsey Envelope
###################################
def ramseyEnvelope(data):
    '''
    Fit the envelope to Ramsey data.
    '''
    env = makeEnvelope(data)
    
        
###################################
##RAMSEY OSCILLOSCOPE
###################################

def fixRamseyOscilloscopeData(dataset, timeStep, useArcTan=False):
    x = dataset.data[:,1]-dataset.data[:,3]
    y = dataset.data[:,2]-dataset.data[:,4]
    if useArcTan:
        phase = np.arctan2(y,x)/(2.0*np.pi)
    else:
        phase = mh.xy2cycles(x,y)
    patches = sp.findPatches(dataset.data[:,5],timeStep)
    if patches != []:
        phase = sp.insertZeros(phase,patches)
    return phase

def ramseyOscilloscope(dataset, weighted=False, chunks=1, smooth=None, fitRange=None, useArcTan=False, plotPhase=False, plotSpectrum=False,
                       magOnly=False, slope=None):
    #Get parameters needed for computation of power spectrum
    holdTime = float(dataset.parameters.holdTime['ns'])
    totalTime = float(dataset.data[-1,5])*sec
    timeStep = float(dataset.parameters.timeStep['s'])
    phase  = fixRamseyOscilloscopeData(dataset, timeStep, useArcTan=useArcTan)
    phase = phase - np.mean(phase)
    #Compute power spectra
    #Sphase = sp.DFT_interleaved(phase,totalTime, smooth=('1f',200.0))
    Sphase = sp.dftInterleavedBootstrapped(phase, totalTime, chunks=chunks, smooth=smooth)
    frequencies = Sphase['frequencies']
    #Convert to frequency and flux units
    Sf = Sphase['S']*((1.0/holdTime)**2) #GHz^2/Hz
    if dataset.parameters.dfdPhi0.isCompatible('GHz'):
        Sphi = Sf*((1/dataset.parameters.dfdPhi0['GHz'])**2)
    elif dataset.parameters.dfdPhi0.isCompatible('GHz/PhiZero'):
        Sphi = Sf*((1/dataset.parameters.dfdPhi0['GHz/PhiZero'])**2)
    #Plot spectrum
    plt.figure()
    plt.loglog(frequencies,Sphi,'.')
    #Fit
    if fitRange is None:
        freqMin = float(raw_input('Minimum frequency to fit [Hz]: '))
        freqMax = float(raw_input('Maximum frequency to fit [Hz]: '))
    else:
        freqMin=fitRange[0]
        freqMax=fitRange[1]
    indexList = (frequencies>freqMin) * (frequencies<freqMax)
    freqsToFit = frequencies[indexList]
    SphiToFit = Sphi[indexList]
    plt.loglog(frequencies[indexList],Sphi[indexList],'r.')
    if weighted and not magOnly:
        weights = 1/np.linspace(1,len(freqsToFit),len(freqsToFit))
        p = mh.weightedLinearFit(np.log(frequencies[indexList]),np.log(Sphi[indexList]),weights)
        p = [p[1][0],p[0][0]]
        slope = p[0]
        A = exp(p[1])
    elif weighted and magOnly:
        weights = 1/np.linspace(1,len(freqsToFit),len(freqsToFit))
        mag = mh.weightedLinearFitFixedSlope(np.log(frequencies[indexList]),np.log(Sphi[indexList]),slope,weights)
        A = exp(mag)
    else:
        p = np.polyfit(log(frequencies[indexList]),np.log(Sphi[indexList]),1)
        slope = p[0]
        A = exp(p[1])
        
    #Show fit line
    plt.loglog(frequencies[indexList],A*frequencies[indexList]**slope,'k.')
    #Package up data to return
    S = AttrDict({'frequencies':frequencies,'Sf':Sf,'Sphi':Sphi,
                  'A':A,'slope':slope,'phase':phase})
    return S

def plotRamseyOscilloscope(dataset):
    amp = str(np.sqrt(dataset.S.A)*1.0E-6)[0:3] #uPhi0/rtHz
    slope = str(dataset.S.slope)[0:4]
    frequencies = dataset.S.frequencies
    description = raw_input('Description of device: ')
    plt.figure()
    plt.loglog(frequencies,dataset.S.Sphi,'.')
    plt.text(0.01,1.0E-7,'Amplitude: '+ amp +'uPhi0/rtHz')
    plt.text(0.01,3.0E-8,'slope: '+slope)
    plt.xlabel('Frequency [Hz]')
    plt.ylabel('Sphi [Phi0^2/rtHz]')
    plt.title('Low Frequency Flux Noise - '+dataset.path[3]+
              dataset.path[4]+ dataset.parameters.config[0] +
              '_'+dataset.path[5]+'_'+str(dataset.path[6])+
              '_'+description)
    plt.loglog(frequencies, dataset.S.A*(frequencies**dataset.S.slope),'r.')
    plt.grid()

#################################
## RAMSEY  ######################
#################################

def plotRamsey(data,T1):
    T1 = float(T1)
    t = data[:,0]
    p = data[:,1]
    pCentered = p - np.mean(p)
    pScaled = pCentered/exp(-t/(2.0*T1))
    plotWithAxes(t,pScaled,'Time [ns]','Probability (Centered)')

def plotScaledRamsey(data,T1):
    T1 = float(T1)
    t = data[:,0]
    p = data[:,1]
    pCentered = p-np.mean(p)
    pScaled = pCentered/exp(-t/(2.0*T1))
    plt.figure()
    plt.subplot(3,1,1)
    plt.plot(t, pCentered,'.')
    plt.xlabel('Time [ns]')
    plt.ylabel('Probability (centered)')
    plt.subplot(3,1,2)
    plt.plot(t, pScaled,'r.')
    plt.xlabel('Time [ns]')
    plt.ylabel('Probability Scaled by T1')
    plt.subplot(3,1,3)
    plt.plot(t,pCentered,'.')
    plt.plot(t,pScaled,'r.')
    plt.xlabel('Time [ns]')
    plt.ylabel('Probability')    

#################################
##RAMSEY TOMOGRAPHY
#################################

def makeEnvelope(dataset):
    xp = dataset.data[:,1]
    xm = dataset.data[:,3]
    yp = dataset.data[:,2]
    ym = dataset.data[:,4]
    env = np.sqrt((xp-xm)**2+(yp-ym)**2)
    return env

def scaledEnvelope(dataset,T1):
    T1 = float(T1)
    t = dataset.data[:,0]
    env = makeEnvelope(dataset)
    envScaled = env/exp(-t/(2.0*T1))
    return envScaled

def ramseyTomo_noLog(dataset, T1=None, timeRange=None, plot=False):
    """
    """
    #CHANGELOG
    # 17 Aug 2011-Daniel Sank
    #  Now uses the plotting functions in pyle.plotting.dstools
    #  Checked that the plots seems to work properly
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    freq = dataset.parameters['fringeFrequency']['GHz']
    
    if T1 is None:
        T1 = float(raw_input('T1 [ns]: '))
    T1=float(T1)
    colors = ['.','r.','g.','k.']
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2)
    envScaled = env/exp(-time/(2*T1))
    #Plot raw data, envelope, and evelope with T1 scaled out
    if plot:
        figRaw=dstools.plotDataset1D(dataset.data,dataset.variables[0],dataset.variables[1],
                                     title='Raw Data')
        figRaw.show()
        figEnv=dstools.plotDataset1D(np.vstack((time,env)).T,[('Time','ns')],
                                    [('Envelope Probability','','')],title='Envelope')
        figEnv.show()
        figEnvScaled=dstools.plotDataset1D(np.vstack((time,envScaled)).T,[('Time','ns')],
                                          [('Scaled Envelope Probability','','')],
                                          title='Scaled Envelope')
        figEnvScaled.show()
    #If time range for fitting has not been specified, ask user
    if timeRange is None:
        minTime = float(raw_input('Minimum time to keep [ns]: '))
        maxTime = float(raw_input('Maximum time to keep [ns]: '))
        timeRange = (minTime,maxTime)
    #Generate array mask for elements to be fitted
    indexList = (time>timeRange[0]) * (time<timeRange[1])
    t = time[indexList]
    e = env[indexList]
    #Create fit function. We fit the unscaled data
    fitFunc = lambda v,x: v[0]*exp(-((x/v[1])**2))*\
              exp(-x/(2*T1))*exp(-x/v[2])+v[3]
    errorFunc = lambda v,x,y: fitFunc(v,x)-y
    vGuess = [0.85, 200.0, 700.0, 0.02]
    result = leastsq(errorFunc, vGuess, args=(t,e), maxfev=10000, full_output=True)
    v = result[0]
    cov=result[1]
    success = result[4]
    #Jim's formula for error bars. Got this from the Resonator fit server.
    rsd = (result[2]['fvec']**2).sum()/(len(t)-len(v))
    bars = np.sqrt(rsd*np.diag(result[1]))
    #Plot the fitted curve
    if plot:
        ax = figEnv.get_axes()[0]
        ax.plot(time, fitFunc(v,time),'g',linewidth=4)
        figEnv.show()
    T2 = v[1]*ns
    dT2 = bars[1]*ns
    Tphi = v[2]*ns
    dTphi = bars[2]*ns
    return AttrDict({'T2':T2, 'dT2':dT2, 'Tphi':Tphi, 'dTphi':dTphi, 'success':success,
                     'fitFunc':fitFunc,'fitVals':v,'envScaled':envScaled})


def ramseyTomoWithLogFactorNew(dataset, T1, minTime=0, maxTime=1e5, TIME=3600.0e9, dfdphi=None):
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    freq = dataset.parameters['fringeFrequency']['GHz']
    
    T1=float(T1)
    colors = ['.','r.','g.','k.']
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2)
    plt.figure()
    for i in range(4):
        plt.plot(time,probabilities[:,i],colors[i])
    plt.figure()
    plt.plot(time,env)
    #Generate array mask for elements you want to keep
    indexList = (time>minTime) * (time<maxTime)
    t = time[indexList]
    e = env[indexList]

    fitFunc = lambda v,x: v[0]*exp(-(x**2)*np.log(0.4*TIME/x)/(v[1]**2))*\
              exp(-x/(2*T1))+v[2]
    errorFunc = lambda v,x,y: fitFunc(v,x)-y
    
    vGuess = [0.85, 130.0, 0.02]
    result = leastsq(errorFunc, vGuess, args=(t,e), maxfev=10000, full_output=True)
    v = result[0]
    cov=result[1]
    success = result[4]
    #Jim's formula for error bars. Got this from the Resonator fit server.
    rsd = (result[2]['fvec']**2).sum()/(len(t)-len(v))
    bars = np.sqrt(rsd*np.diag(result[1]))
    #Plot the fitted curve
    plt.plot(time, fitFunc(v,time),'g')
    T2 = v[1]*ns
    dT2 = bars[1]*ns
    if dfdphi is None:
        dfdphi = (dataset.parameters.dfdPhi0['GHz'])
    S = 2.0*(1.0/(2*pi*T2['ns'])**2)*dfdphi**-2
    
    return AttrDict({'T2':v[1]*ns, 'dT2':dT2, 'success':success, 'rtS':np.sqrt(S)})

def ramseyTomoAlphaNew(dataset, T1, alphas, dfdphi=None, fitRange=[0,40000], makePlots=False):
    if not np.iterable(alphas):
        alphas = np.array([alphas])
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    #freq = dataset.parameters['fringeFrequency']['GHz']
    
    TIME = 3600.0*(10**9) #One hour in nanoseconds
    T1=float(T1)
    colors = ['.','r.','g.','k.']
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2)
    if makePlots:
        plt.figure()
        for i in range(4):
            plt.plot(time,probabilities[:,i],colors[i])
        plt.figure()
        plt.plot(time,env)
    if fitRange is None:
        minTime = float(raw_input('Minimum time to keep [ns]: '))
        maxTime = float(raw_input('Maximum time to keep [ns]: '))
    else:
        minTime = float(fitRange[0])
        maxTime = float(fitRange[1])
    #Generate array mask for elements you want to keep
    indexList = (time>minTime) * (time<maxTime)
    t = time[indexList]
    e = env[indexList]
    noiseAmplitudes = np.array([])
    T2s = np.array([])
    if dfdphi is None:
        dfdphi = dataset.parameters.dfdPhi0['GHz']
    for alpha in alphas:
        integralFunc = ramseyIntegrals.interpForAlpha(alpha,t,1.0/TIME, 1e7) # upper freq = 10 MHz
        fitFunc = lambda x,amp,b,offset: amp*exp(-x/(2*T1)) * exp(-(x**(1.0+alpha))*b*integralFunc(x))+offset
        result,cov = scipy.optimize.curve_fit(fitFunc,t,e,[1.0,1.0/1000000,0.02])
        plt.plot(t, fitFunc(t, *result), '-', linewidth=2.0)
        a,b,c = result
        S = 2.0*(1.0/(2*pi)**2)*b*dfdphi**-2
        noiseAmplitudes = np.hstack((noiseAmplitudes,np.sqrt(S)))
        T2s = np.hstack((T2s, 1.0/np.sqrt(b)))
    return AttrDict({'noiseAmplitudes':noiseAmplitudes, 'alphas':alphas, 'T2s':T2s})
    
    
def ramseyTomoWithLogFactor(dataset, T1):
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    freq = dataset.parameters['fringeFrequency']['GHz']
    
    TIME = 3600.0*(10**9) #One hour in nanoseconds
    T1=float(T1)
    colors = ['.','r.','g.','k.']
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2)
    plt.figure()
    for i in range(4):
        plt.plot(time,probabilities[:,i],colors[i])
    plt.figure()
    plt.plot(time,env)
    minTime = float(raw_input('Minimum time to keep [ns]: '))
    maxTime = float(raw_input('Maximum time to keep [ns]: '))
    #Generate array mask for elements you want to keep
    indexList = (time>minTime) * (time<maxTime)
    t = time[indexList]
    e = env[indexList]

    fitFunc = lambda v,x: v[0]*exp(-(x**2)*np.log(0.4*TIME/x)/(v[1]**2))*\
              exp(-x/(2*T1))+v[2]
    errorFunc = lambda v,x,y: fitFunc(v,x)-y
    
    vGuess = [0.85, 130.0, 0.02]
    result = leastsq(errorFunc, vGuess, args=(t,e), maxfev=10000, full_output=True)
    v = result[0]
    cov=result[1]
    success = result[4]
    #Jim's formula for error bars. Got this from the Resonator fit server.
    rsd = (result[2]['fvec']**2).sum()/(len(t)-len(v))
    bars = np.sqrt(rsd*np.diag(result[1]))
    #Plot the fitted curve
    plt.plot(time, fitFunc(v,time),'g')
    T2 = v[1]*ns
    dT2 = bars[1]*ns
    S = 2.0*(1.0/(2*pi*T2['ns'])**2)*(dataset.parameters.dfdPhi0['GHz'])**-2
    
    return AttrDict({'T2':v[1]*ns, 'dT2':dT2, 'success':success, 'rtS':np.sqrt(S)})

def ramseyTomoAlpha(dataset, T1, alphas, fitRange=None, makePlots=False):
    if not np.iterable(alphas):
        alphas = np.array([alphas])
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    #freq = dataset.parameters['fringeFrequency']['GHz']
    
    TIME = 3600.0*(10**9) #One hour in nanoseconds
    T1=float(T1)
    colors = ['.','r.','g.','k.']
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2)
    if makePlots:
        plt.figure()
        for i in range(4):
            plt.plot(time,probabilities[:,i],colors[i])
        plt.figure()
        plt.plot(time,env)
    if fitRange is None:
        minTime = float(raw_input('Minimum time to keep [ns]: '))
        maxTime = float(raw_input('Maximum time to keep [ns]: '))
    else:
        minTime = float(fitRange[0])
        maxTime = float(fitRange[1])
    #Generate array mask for elements you want to keep
    indexList = (time>minTime) * (time<maxTime)
    t = time[indexList]
    e = env[indexList]
    noiseAmplitudes = np.array([])
    for alpha in alphas:
        integralFunc = ramseyIntegrals.interpForAlpha(alpha,t,1.0/TIME)
        fitFunc = lambda x,amp,b,offset: amp*exp(-x/(2*T1)) * exp(-(x**(1.0+alpha))*b*integralFunc(x))+offset
        result,cov = scipy.optimize.curve_fit(fitFunc,t,e,[1.0,1.0/1000000,0.02])
        a,b,c = result
        S = 2.0*(1.0/(2*pi)**2)*b*(dataset.parameters.dfdPhi0['GHz'])**-2
        noiseAmplitudes = np.hstack((noiseAmplitudes,np.sqrt(S)))
    return AttrDict({'noiseAmplitudes':noiseAmplitudes, 'alphas':alphas})
    

def spinEchoTomo(dataset,T1,timeRange,makePlots=False):
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    T1=float(T1)
    colors = ['.','r.','g.','k.']
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2)
    if makePlots:
        plt.figure()
        for i in range(4):
            plt.plot(time,probabilities[:,i],colors[i])
        plt.figure()
        plt.plot(time,env)
    if timeRange is None:
        minTime = float(raw_input('Minimum time to keep [ns]: '))
        maxTime = float(raw_input('Maximum time to keep [ns]: '))
    else:
        minTime = float(timeRange[0])
        maxTime = float(timeRange[1])
    #Generate array mask for elements you want to keep
    indexList = (time>minTime) * (time<maxTime)
    t = time[indexList]
    e = env[indexList]
    fitFunc = lambda x,amp,rate,offset: (amp*exp(-x/(2*T1)) * exp(-(x/rate)**2))+offset
    result,cov = scipy.optimize.curve_fit(fitFunc,t,e,[1.0,0.0025,0.02])
    a,tEcho,c = result
    try:
        dfdPhi0 = dataset.parameters.dfdPhi0['GHz/PhiZero']
    except:
        print 'Frequency sensitivity was not in GHz/Phi0, trying GHz'
        dfdPhi0 = dataset.parameters.dfdPhi0['GHz']
    #Compute noise amplitude, in units of Phi0. There is no frequency unit because
    #we're assuming the spectral density is S
    S = (4.0)/(((2*np.pi)**2)*(tEcho**2)*(dfdPhi0**2)*1.35)
    noiseAmplitude = np.sqrt(S)
    return AttrDict({'noiseAmplitude':noiseAmplitude,'fitFunc':fitFunc})


def spinEchoTomoAlpha(dataset,inputT1,alphas,timeRange=None,makePlots=False,totalTime=3600*sec,upperFreq=10*MHz):
    results = []
    colors = ['.','r.','g.','k.']
    if not np.iterable(alphas):
        alphas = np.array([alphas])
    if dataset.parameters.dfdPhi0.isCompatible('GHz/PhiZero'):
        sUnit = 'GHz/PhiZero'
    elif dataset.parameters.dfdPhi0.isCompatible('GHz'):
        sUnit = 'GHz'
    if not np.iterable(alphas):
        alphas = np.array([alphas])
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    lowerFreq = (1.0/(totalTime['ns']))
    T1 = inputT1['ns']
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2 )
    #Plot raw data and envelope
    if makePlots:
        plt.figure()
        for i in range(4):
            plt.plot(time,probabilities[:,i],colors[i])
        plt.grid()
        plt.figure()
        plt.plot(time,env/(exp(-time/(2.0*T1))))
        plt.grid()
    if timeRange is None:
        minTime = float(raw_input('Minimum time to keep [ns]: '))
        maxTime = float(raw_input('Maximum time to keep [ns]: '))
    else:
        minTime = float(fitRange[0])
        maxtime = float(fitRange[1])
    indexList = (time>minTime) * (time<maxTime)
    t = time[indexList]
    e = env[indexList]
    noiseAmplitudes = np.array([])
    for alpha in alphas:
        integralFunc = ramseyIntegrals.echoInterpForAlpha(alpha,t,lowerFreq,upperFreq)
        fitFunc = lambda x,amp,b,offset: amp*exp(-t/(2*T1)) * exp(-(x**(1.0+alpha))*b*integralFunc(x))+offset
        result,cov = scipy.optimize.curve_fit(fitFunc,t,e,[1.0,1.7E-8,0.02])
        a,b,c = result
        S = 2.0*(1.0/(2*pi)**2)*b*(dataset.parameters.dfdPhi0[sUnit])**-2
        noiseAmplitudes = np.hstack((noiseAmplitudes,np.sqrt(S)))
        results.append(AttrDict({'noiseAmplitude':noiseAmplitude,'alpha':alpha,
                                 'fitFunc':fitFunc}))
    return results
    
def rabi(dataset, freq, T1, timeRange):
    T1 = float(T1['ns'])
    time = dataset.data[:,0]
    probability = dataset.data[:,1]
    probability = probability-np.mean(probability)
    plt.figure()
    plt.plot(time,probability,'.')
    plt.xlabel('Time [ns]')
    plt.ylabel('Probability')
    #Select range to fit
    if timeRange is None:
        minTime = float(raw_input('Minimum time to keep [ns]: '))
        maxTime = float(raw_input('Maximum time to keep [ns]: '))
    else:
        minTime=timeRange[0]
        maxTime=timeRange[1]
    #Generate array mask for elements you want to keep
    indexList = (time>minTime) * (time<maxTime)
    t = time[indexList]
    p = probability[indexList]
    plt.plot(t,p,'r.')
    #Fit oscillations with decaying exponential
    fitFunc = lambda v,x: v[0]*exp(-(1.0/2.0)*x/v[1])*exp(-(3.0/4.0)*x/T1)*\
              cos(2*pi*x*v[2]+v[3])+v[4]
    errorFunc = lambda v,x,y: fitFunc(v,x)-y
    vGuess = [0.5, 400, freq['GHz'], np.pi, 0.0]
    result = leastsq(errorFunc, vGuess, args=(t,p), maxfev=10000, full_output=True)
    v = result[0]
    cov=result[1]
    if cov is None:
        raise Exception('Singular Covariance matrix was found.')
    success = result[4]
    #Jim's formula for error bars. Got this from the Resonator fit server.
    rsd = (result[2]['fvec']**2).sum()/(len(t)-len(v))
    bars = np.sqrt(rsd*np.diag(result[1]))
    #Plot the fitted curve
    timePlot = np.arange(time[0],time[-1],1.0)
    plt.plot(timePlot, fitFunc(v,timePlot),'g')
    plt.grid()
    Tnu = v[1]*ns
    dTnu = bars[1]*ns
    freq=v[2]
    #Compute Sf and Sphi
    Sf = ((1.0/Tnu['ns'])*((1.0/np.pi)**2)*(1.0E-9))*((1.0*GHz)**2)/(1.0*Hz) #GHz^2/Hz
    Sphi = Sf*(dataset.parameters.dfdPhi0)**-2

    return AttrDict({'Tnu':Tnu, 'dTnu':dTnu, 'freq':freq*GHz, 'Sf':Sf, 'Sphi':Sphi,
                     'fitFunc':fitFunc, 'fitVals':result[0]})
    
def plotWithAxes(x,y,xlabel='',ylabel='',marker='.'):
    plt.figure()
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.plot(x,y,marker)

    

def ramseyTomo_noLog_ExpAndGauss(dataset, T1=None, timeRange=None, plot=False,fitter='ExpAndGauss'):
    """
    Fit ramsey/spin echo envelope, to a white noise and/or a 1/f noise process. RB.
    fitter='Exponential':
    single exponential decay
    fitter='Gaussian':
    single exponential decay AND gaussian decay    
    """
    time = dataset.data[:,0]
    probabilities = dataset.data[:,1:]
    xp = probabilities[:,0]
    yp = probabilities[:,1]
    xm = probabilities[:,2]
    ym = probabilities[:,3]
    env = np.sqrt( (xp-xm)**2 + (yp-ym)**2)
    if T1:
        envScaled = env/np.exp(-time/(2*T1))
    else:
        envScaled = None
    te,dte,t1f,dt1f,fitFunc,p1,success=fitdecay(time,env,fitter=fitter,p0=None)     #We fit the unscaled data
    #Plot raw data, envelope, and evelope with T1 scaled out
    if plot:
        figRaw=dstools.plotDataset1D(dataset.data,dataset.variables[0],dataset.variables[1],
                                     title='Raw Data')
        figRaw.show()
        figEnv=dstools.plotDataset1D(np.vstack((time,env)).T,[('Time','ns')],
                                    [('Envelope Probability','','')],title='Envelope')
        figEnv.show()
        if T1:
            figEnvScaled=dstools.plotDataset1D(np.vstack((time,envScaled)).T,[('Time','ns')],
                                          [('Scaled Envelope Probability','','')],
                                          title='Scaled Envelope')
            figEnvScaled.show()
        ax = figEnv.get_axes()[0]
        ax.plot(time, fitFunc(time,*p1),'g',linewidth=2)
        figEnv.show()
    if T1:
        te = 1./ ( 1./te-1/T1 ) #subtract T1 from exponential decay
        
    return AttrDict({'Te':te, 'dTe':dte, 'T1f':t1f, 'dT1f':dt1f, 'fitFunc':fitFunc,'fitVals':p1,'envScaled':envScaled,'env':env, 'success':success })
    
    
def fitdecay(time,envel,fitter='ExpAndGauss',p0=None):
    """Fit ramsey/spin echo envelope. RB.
    fitter='Exponential':
    single exponential decay
    fitter='Gaussian':
    single exponential decay AND gaussian decay.
    """
    success=True
    if fitter=='Exponential':
        #single exponential decay
        #fitFunc = lambda time, t2, prob: prob*np.exp(-time/t2) #+off  
        fitFunc = lambda time, t1, prob,off: prob*np.exp(-time/t1) +off  
        if p0 is None:
            p0 = [7000, 0.7,0.0]
        try:
            p1, pcov = curve_fit(fitFunc, time, envel, p0=p0)
        except:
            success=False
            p1=np.zeros(len(p0))
            pcov=np.eye(len(p0))*0.0
        if type(pcov) is np.ndarray:
            error = np.sqrt(abs(pcov[0,0])) #sigma = sqrt(var) #sum(abs(prob - fitFunc(delays, *p1))**2)    
        else:
            error = p1[0]         
        te=p1[0]
        dte=error
        t1f=None
        dt1f=None        
    elif fitter=='Gaussian':
        #gaussian decay: 1/f dephasing
        #fitFunc = lambda time, t1f, prob: prob*np.exp( -(time/t1f)**2 ) #+off
        fitFunc = lambda time, t1f, prob,off: prob*np.exp( -(time/t1f)**2 ) +off
        if p0 is None:        
            p0 = [7000, 0.7,0.0]
        try:
            p1, pcov = curve_fit(fitFunc, time, envel, p0=p0)
        except:
            success=False
            p1=np.zeros(len(p0))
            pcov=np.eye(len(p0))*0.0
        if type(pcov) is np.ndarray: 
            error1f = np.sqrt(abs(pcov[0,0])) #sigma = sqrt(var) #sum(abs(prob - fitFunc(delays, *p1))**2)                
        else:
            error1f = p1[0]
        te=None
        dte=None
        t1f=p1[0]
        dt1f=error1f
    elif fitter=='ExpAndGauss':
        #gaussian + exponential decay: 1/f & white noise dephasing
        #fitFunc = lambda time, texp, t1f, prob: prob*np.exp( -(time/t1f)**2 -time/texp ) #+off
        fitFunc = lambda time, texp, t1f, prob, off: prob*np.exp( -(time/t1f)**2 -time/texp ) +off
        if p0 is None:        
            p0 = [5000, 5000, 0.7,0.0]
        try:
            p1, pcov = curve_fit(fitFunc, time, envel, p0=p0)
        except:
            success=False
            p1=np.zeros(len(p0))
            pcov=np.eye(len(p0))*0.0
        if type(pcov) is np.ndarray:
            errore = np.sqrt(abs(pcov[0,0])) #sigma = sqrt(var) #sum(abs(prob - fitFunc(delays, *p1))**2)    
            error1f = np.sqrt(abs(pcov[1,1])) #sigma = sqrt(var) #sum(abs(prob - fitFunc(delays, *p1))**2)                
        else:
            errore = p1[0] 
            error1f = p1[1]
        te=p1[0]
        dte=errore
        t1f=p1[1]
        dt1f=error1f        
    else:
        raise Exception("Fitter not defined, use Gaussian or Exponential")
        success=False
    return te,dte,t1f,dt1f,fitFunc,p1,success
    
    
