import numpy as np
import scipy
import scipy.special as special
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
import matplotlib
from mpl_toolkits.mplot3d import Axes3D

from labrad.units import Unit,Value

import pyle.analysis.iq as iq
import pyle.dataking.adjust as adjust
import pyle.fitting.fitting as fitting
import pyle.dataking.envelopehelpers as eh

from pyle import gateCompiler as gc

from labrad.units import GHz, dBm

import pyle.datavault as dv

#Utility

def fitGaussianOutliers(x):
    """Fit a Gaussian curve to a collection of data while rejecting outliers
    
    x is an array of measured values. This function converts the data to a
    hitogram with outliers removed, and fits a Gaussian curve to the histogram.
    
    The fitted function is
    1/sqrt(2pi sigma^2) * y0 * exp(-(x-x0)^2 / (2 sigma^2))
    
    returns y0, x0, sigma
    """
    x0 = np.median(x)
    sigma = np.std(x)
    mask = np.logical_and(x > (x0-2*sigma), x<(x+2*sigma))
    hist, bin_edges = np.histogram(x[mask], bins=50)
    # find bin centers
    bin_centers = np.convolve(bin_edges, [.5, .5], mode='valid')
    def gaussian_fit(x, y0, x0, sigma):
        return 1/(np.sqrt(2*np.pi*sigma**2)) * \
                    y0*np.exp(-(x-x0)**2/(2 * sigma**2))
    x0_guess = x0
    sigma_guess = sigma
    y0_guess = np.max(hist)
    p_opt, pcov = scipy.optimize.curve_fit(gaussian_fit, bin_centers, hist,
                                           (y0_guess, x0_guess, sigma_guess))
    return p_opt

def fitGaussianOutliersHistogram(counts, binEdges):
    """Fit a Gaussian curve to a histogram while rejecting outliers
    
    h is a tuple of (counts, bin edges)
    
    The fitted function is
    1/sqrt(2pi sigma^2) * y0 * exp(-(x-x0)^2 / (2 sigma^2))
    
    returns y0, x0, sigma
    """
    bin_centers = np.convolve(bin_edges, [.5, .5], mode='valid')
    def gaussian_fit(x, y0, x0, sigma):
        return 1/(np.sqrt(2*np.pi*sigma**2)) * \
                    y0*np.exp(-(x-x0)**2/(2 * sigma**2))
    x0_guess = x0
    sigma_guess = sigma
    y0_guess = np.max(counts)
    p_opt, pcov = scipy.optimize.curve_fit(gaussian_fit, bin_centers, counts,
                                           (y0_guess, x0_guess, sigma_guess))
    return p_opt
    
def determineState(z, qubit):
    centers = np.array(qubit['readoutCenters'])
    diffs = np.abs(np.vstack([z - c for c in centers]))
    states = np.argmin(diffs, 0)
    return states
    
# Data set analyzers
    
def analyze_readout(Sample, measure):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    """Compute interesting readout parameters from a qubit registry"""
    eta = q['f21'] - q['f10']
    delta = (q['f10'] - q['readoutFrequency'])
    chi = 0.5*(q['readoutResonatorFreq1'] - q['readoutResonatorFreq0'])
    GHz_per_dacamp2 = q['calAcStark']
    print "eta: ", eta
    print "delta: ", delta
    print "chi: ", chi    
    g = (chi * delta * (1+delta/eta)).sqrt()
    print "g :", g

    ncrit = 0.25*((delta/g)**2)['']
    print "ncrit: ", ncrit
    n_at_dacamp2eq1 = GHz_per_dacamp2 / (2*chi)
    print "n at dacamp = 1: ", n_at_dacamp2eq1
    dacamp2_ncrit = ncrit/n_at_dacamp2eq1
    print "dacamp^2 @ ncrit = %f (%f dBm)" %(dacamp2_ncrit,
                                        eh.amp2power(np.sqrt(dacamp2_ncrit)))
    
def resonator_ringdown_kappa(dvwData, timeRange=None):
    
    data = dvwData
    data=np.array(data)
    dataSize = len(matplotlib.mlab.find(data[:,0]==data[0,0]))
    dataLen = len(data)
    print dataLen,dataSize
    numColumns = dataLen/dataSize
    newDataLen = numColumns*dataSize
    data = data[:newDataLen, :] # we take off any unfinished rows
    delays = data[::dataSize, 0]
    freqs = data[:dataSize,1]
    probs = np.reshape(data[:,2], (numColumns, dataSize))
    
    maxFreqs = []
    for prob in probs:
        currFreq = fitting.getMaxGauss(freqs, prob)
        maxFreqs.append(currFreq)
    maxFreqs=np.array(maxFreqs)
    
    plt.plot(delays, maxFreqs, '-ob')
    if timeRange is None:
        minTime = float(raw_input('Minimum time to keep [ns]: '))
        maxTime = float(raw_input('Maximum time to keep [ns]: '))
        timeRange = (minTime,maxTime)
    #Generate array mask for points that will be kept
    indexList = (delays>timeRange[0]) * (delays<timeRange[1])
    t = delays[indexList]
    p=maxFreqs[indexList]
    #Fit the data to an exponential
    vGuess = [np.min(maxFreqs), 30.0, maxFreqs[0]]
    fits, cov, fitFunc = fitting.fitCurve('exponential', t-timeRange[0], p, vGuess)
    a, timeConst, dcOffset = fits
    print timeConst
    plt.plot(t, fitFunc(t-timeRange[0], *fits),'-r',lw=3)
    plt.title('kappa = %0.3f ns' % timeConst)
    plt.xlabel('time')
    plt.ylabel('qubit frequency [GHz]')

def chiVsFreq(datasetWrapper):
    
    data = datasetWrapper
    data=np.array(data)
    dataSize = len(matplotlib.mlab.find(data[:,0]==data[0,0]))
    dataLen = len(data)
    print dataLen,dataSize
    numColumns = dataLen/dataSize
    newDataLen = numColumns*dataSize
    data = data[:newDataLen, :] # we take off any unfinished rows
    freqs = data[::dataSize, 0]
    qubitFreqs = data[:dataSize,1]
    I0 = np.reshape(data[:,2], (numColumns, dataSize)).transpose()
    Q0 = np.reshape(data[:,3], (numColumns, dataSize)).transpose()
    I1 = np.reshape(data[:,4], (numColumns, dataSize)).transpose()
    Q1 = np.reshape(data[:,5], (numColumns, dataSize)).transpose()
    
    dip0 = []
    dip1 = []
    freq0 = []
    freq1 = []
    for subI0, subQ0, subI1, subQ1 in zip(I0, Q0, I1, Q1):
        s0 = subI0 + 1.0j*subQ0
        s1 = subI1 + 1.0j*subQ1
        plt.plot(freqs, abs(s0),'-b')
        plt.plot(freqs, abs(s1),'-r')
        dip0.append(abs(s0))
        dip1.append(abs(s1))
        #freq0.append(fitting.getMaxGauss(freqs, abs(s0), fitToMax=False))
        #freq1.append(fitting.getMaxGauss(freqs, abs(s1), fitToMax=False))
        freq0.append(freqs[np.argmin(abs(s0))])
        freq1.append(freqs[np.argmin(abs(s1))])
    dip0 = np.array(dip0)
    dip1 = np.array(dip1)
    chi = (np.array(freq0)-np.array(freq1))/2.
    plt.title('Chi')
    plt.figure()
    plt.plot(qubitFreqs, chi)
    plt.xlabel('Qubit Frequency [GHz]')
    plt.ylabel('Dispersive Shift Chi [GHz]')
    plt.figure()
    plt.plot(qubitFreqs, freq0,'b', label='resonator |0>')
    plt.plot(qubitFreqs, freq1,'r', label='resonator |1>')
    plt.xlabel('Qubit Frequency [GHz]')
    plt.ylabel('Frequency [GHz]')
    plt.legend()
 
def measurementDephasing(datasetMeasDephasing, datasetStarkShift, plotAll=False):
    
    p, ampsSquared, frequencies = fitting.acStark(datasetStarkShift)
    print("Slope of line in GHz/dacAmp^2: %f"%p[0])
    indices = np.argsort(ampsSquared)
    if plotAll:
        plt.figure()
        plt.plot(ampsSquared[indices], frequencies['GHz'][indices], 'b.')
        plt.plot(ampsSquared[indices], np.polyval(p, ampsSquared)[indices],'r-')
        plt.xlabel('DAC amplitude squared')
        plt.ylabel('Frequency [GHz]')
        plt.title("AC Stark shift vs readout power")
    qkey = datasetStarkShift.parameters['measure'][0]
    q = datasetStarkShift.parameters['q1']
    
    chi = 0.5*(q['readoutResonatorFreq1']\
             - q['readoutResonatorFreq0'])
    photons = (p[0]*GHz / (2.0 * chi))[''] * ampsSquared
    
    if plotAll:
        fig = plt.figure()
        plt.plot(photons[indices], frequencies['GHz'][indices], 'b.')
        plt.xlabel('Photons')
        plt.ylabel('Frequency [GHz]')
        plt.title("AC Stark shift vs resonator photons")
        
        plt.figure()
        plt.plot(ampsSquared[indices], photons[indices],'-ob')
        plt.xlabel('DAC amplitude squared')
        plt.ylabel('Photons')
    
    
    data = datasetMeasDephasing
    data=np.array(data)
    dataSize = len(matplotlib.mlab.find(data[:,0]==data[0,0]))
    dataLen = len(data)
    print dataLen,dataSize
    numColumns = dataLen/dataSize
    newDataLen = numColumns*dataSize
    data = data[:newDataLen, :] # we take off any unfinished rows
    delays = data[::dataSize, 0]
    dacAmp2 = data[:dataSize,1]
    envelope = np.reshape(data[:,2], (numColumns, dataSize))

    fitFunc = lambda time, tp, coef, const: coef*np.exp(-time/tp)+const
    p0 = [100.0, 0.5, 0.1]
    tps = []
    error_bars = []
    
    for currEnv in envelope.transpose():
    
        popt, pcov = curve_fit(fitFunc, delays, currEnv, p0=p0)
        
        if type(pcov) is type(np.inf):
            continue
        
        tps.append(popt[0])
        error_bars.append(np.sqrt(pcov[0,0]))
        
    plt.figure()
    ax = plt.subplot(1,1,1)
    plt.errorbar(dacAmp2, tps, error_bars)   
    plt.xlabel('DAC amplitude squared')
    plt.ylabel('Measurement Induced Tphi')
    ax.set_yscale('log')
    
    #convert dacAmp2 to photonNumber
    # ampsSquared[indices], photons[indices]
    p = np.polyfit(ampsSquared[indices], photons[indices],1)
    ampsSquared2photons = p[0]
    
    plt.figure()
    ax = plt.subplot(1,1,1)
    plt.errorbar(dacAmp2*ampsSquared2photons, tps, error_bars)   
    plt.xlabel('Photons')
    plt.ylabel('Measurement Induced Tphi [ns]')
    ax.set_yscale('log')
    
    
    p = np.polyfit(dacAmp2*ampsSquared2photons, 1/np.array(tps),1)
    plt.figure()
    plt.plot(dacAmp2*ampsSquared2photons, 1/np.array(tps),'-ob')   
    
    theoryRange = np.linspace(-1.0, np.max(dacAmp2*ampsSquared2photons))
    
    plt.plot(theoryRange, np.polyval(p, theoryRange),'-k')
    plt.xlabel('Photons')
    plt.ylabel('Measurement Induced Tphi [GHz]')
    plt.title('%.03f GHz/Photon' % p[0])
    
    
    
 
# Fidelity/IQ plane distributions, etc.
    
def selectPreparationAndMeasureTimes(indep, data):
    """Get herald time points and measure start time point
    
    indep - ndarray: probably a time axis
    data - ndarray: 2d array of time traces. Each column is a time trace
    
    Returns a tuple of the state preparation start time, state preparation end
    time, and measure start time
    """
    markers = ['b-','r-','g-', 'k-']
    colors = ['b', 'r', 'g', 'k']
    r = (min(indep), max(indep))
    vals = [0.0, 200, 300, 400]
    names = ['State prep. start', 'State prep. end', 'Measure start',
             'Measure end']
    traces = [{'x': indep, 'y': dat, 'args': (marker,)} \
               for dat, marker in zip(data.T, markers)]
    params = [{'name': n, 'val': v, 'range': r, 'axis': 'x', 'color': c} \
               for n,v,c in zip(names, vals, colors)]
    result = adjust.adjust(params, traces)
    if result is None:
        return None
    return tuple([result[n] for n in names])
    
def IQTrace(data, debug=False,
            statePrepStart_ns=None, statePrepEnd_ns=None,
            measureStart_ns=None):
    """Process time traces with heralded state preparation
    
    Need to note where the heralding is supposed to be for data analyzed by
    this function.
    
    Data columns should be in the form: |Repetition|Time|I0|Q0|I1|Q1|...|In|Qn|
    where n+1 is the number of prepared qubit states.

    Returns a tuple of (time_ns, cumulatives, statePrepS, ax)
    
    time_ns - array: Array of time values
    cumulatives - dict: Keys are heralded qubit states
                  values are lists of integrated (cumulative) IQ worldlines
    statePrepS - array: One column for each prepared (heralded qubit state)
                        Each column gives the transmitted S value found during
                        the heralding stage of the measurement.
    ax - plot axis: Axis on which the 3D cumulative world lines are plotted.
    
    """
    measure = data.parameters['measure'][0]
    qubit = data.parameters[data.parameters['config'][measure]]
    readoutDevice = data.parameters[qubit['readoutConfig']]
    #Demodulation parameters for this qubit
    readoutFrequency = qubit['readoutFrequency']
    carrierFrequency = readoutDevice['carrierFrequency']
    sidebandFrequency = readoutFrequency - carrierFrequency
    #Reshape data such that first axis is repetition number
    numberOfRepetitions = int(max(set(data[:,0])))+1
     #Figure out how many different traces (ie, states) we prepared
    numTraceCols = data.shape[1]-2
    assert numTraceCols%2 == 0
    numTraces = numTraceCols/2
    #reshaped has three indices, repetition, time, dependentVariable
    reshaped = np.reshape(data, (numberOfRepetitions, -1, numTraceCols + 2))
    time_ns = reshaped[0][:,1] * ((Value(1.0, data.indep[1][1]))['ns'])
    
    #Get start/end times of state preparation and measurement from user
    if not all((statePrepStart_ns, statePrepEnd_ns, measureStart_ns)):
        print("Shape of reshaped: %s"%str(reshaped.shape))
        compressed = np.mean(reshaped[:,:,2:], 0)
        print("Shape of compressed: %s"%str(compressed.shape))
        traces = np.vstack((np.abs(compressed[:,2]+1.0j*compressed[:,3]),
                            np.abs(compressed[:,4]+1.0j*compressed[:,5]))).T
        statePrepStart_ns, statePrepEnd_ns, measureStart_ns, _ = \
            selectPreparationAndMeasureTimes(time_ns, traces)
    #Grab state preparation and measure data sections
    #We have to force the shape of the logical indices to have only one
    #dimension because np.argwhere returns 2D arrays. I don't know why.
    indices = np.argwhere(np.logical_and(time_ns<statePrepEnd_ns,
                                         time_ns>statePrepStart_ns))
    indices.shape = (indices.shape[0],)
    statePrepData = reshaped[:, indices, :]
    indices = np.argwhere(time_ns>measureStart_ns)
    indices.shape = (indices.shape[0],)
    measureData = reshaped[:, indices, :]
    #Demodulate state preparation data
    statePrepS = []
    for repetition in statePrepData:
        for trace in range(numTraces):
            rep, time_ns, I, Q = repetition[:,0], \
                                 repetition[:,1], \
                                 repetition[:, 2*(1+trace)], \
                                 repetition[:, 2*(1+trace)+1]
            statePrepS.append(iq.demodulateAndSum(time_ns, I, Q,
                             sidebandFrequency['GHz'], phase_rad=0,
                             window=None))
    statePrepS = np.reshape(np.array(statePrepS), (-1, numTraces))
    """START FUNCTION BLOCK"""
    #Find state discrimination axis from state preparation data
    #Only discriminate using first two colums, ie |0> and |1>
    
    #Delete this block
    #centers = iq.discriminationCenters(*(statePrepS[:,i] \
    #                                   for i in range(numTraces)))
    
    centers = [iq.iqCenter(statePrepS[:,i]) for i in range(numTraces)]
    centers = np.array(centers)
    #For each repetition find state after first measurement
    #0 means |0>, 1 means|1>, etc
    initialStatesByRepetition = []
    def whichState(z):
        return np.argmin(np.abs((z - centers)))
    for elem in statePrepS.flat:
        initialStatesByRepetition.append(whichState(elem))
    initialStatesByRepetition = np.reshape(np.array(initialStatesByRepetition),
                                         (-1, numTraces))
    """END FUNCTION BLOCK"""
    if debug:
        fig = plt.figure(); ax=fig.add_subplot(111)
        ax.plot(statePrepS[:,0].real, statePrepS[:,0].imag, 'b.',
                label="No Pulse", alpha=0.5)
        ax.plot(statePrepS[:,1].real, statePrepS[:,1].imag, 'r.',
                label="Pulse", alpha=0.5)
        ax.plot(statePrepS[:,2].real, statePrepS[:,2].imag, 'g.',
                label="Pulse Pulse", alpha=0.5)
        ax.legend(loc="upper right")
        ax.set_title("IQ points from state preparation")
        fig = plt.figure(); ax = fig.add_subplot(111)
        ax.plot(initialStatesByRepetition[:,0], 'b.', label="Prepared |0>")
        ax.plot(initialStatesByRepetition[:,1], 'r.', label="Prepared |1>")
        ax.plot(initialStatesByRepetition[:,2], 'g.', label="Prepared |2>")
        ax.legend(loc='upper right')
        ax.set_title("Determined states - High means |1>, Low means |0>")
    #Sort traces by whether measured preparation state
    #state -> list of meaured complex time traces
    buckets = dict([(state, []) for state in range(numTraces)])
    for initialStates, repetitionData in zip(initialStatesByRepetition,
                                             measureData):
        for n in range(numTraces):
            z = repetitionData[:, 2*(n+1)] + 1.0j*repetitionData[:, 2*(n+1)+1]
            state = initialStates[n]
            buckets[state].append(z)
    #Get cumulative demodulation sums for all traces
    time_ns = measureData[0, :, 1] - measureData[0, 0, 1]
    cumulatives = {} #state -> list of integrated worldlines
    for state, listOfWorldlines in buckets.items():
        integratedWorldlines = [iq.demodulateAndCumulativeSum(time_ns,
                                    z.real, z.imag,
                                    sidebandFrequency['GHz'],
                                    phase_rad=0,
                                    window=None) for z in listOfWorldlines]
        cumulatives[state] = np.array(integratedWorldlines)
    fig = plt.figure()
    try:
        ax = fig.add_subplot(111, projection='3d')
    except ValueError: #matplotlib prior to 1.0 don't support projection='3d'
        ax = Axes3D(fig)
    ptsToPlot=100
    tracesToPlot=80
    for trace in cumulatives[0][0:tracesToPlot,:]:
        ax.plot(trace.real[0:ptsToPlot], trace.imag[0:ptsToPlot],
                time_ns[0:ptsToPlot], 'b-', alpha=0.5, linewidth=2)
    for trace in cumulatives[1][0:tracesToPlot,:]:
        ax.plot(trace.real[0:ptsToPlot], trace.imag[0:ptsToPlot],
                time_ns[0:ptsToPlot], 'r-', alpha=0.5, linewidth=2)
    for trace in cumulatives[2][0:tracesToPlot,:]:
        ax.plot(trace.real[0:ptsToPlot], trace.imag[0:ptsToPlot],
                time_ns[0:ptsToPlot], 'g-', alpha=0.5, linewidth=2)
    #Plot a dot to mark the center point
    ax.plot([0], [0], [0], 'y.', markersize=10)
    return time_ns, cumulatives, statePrepS, ax
    
def qubitExcitedInState(s, q, N):
    return bool((s >> (N-q-1))%2)
    
def IQTrace_multiQubit(data, debug=False, heraldRanges=None,
                       optimalWindow=True):
    """Process time traces with heralded state preparation
    
    The pulse for data analyzed by this function should be...
    
    Data columns should be in the form: |Repetition|Time|I0|Q0|I1|Q1|...|In|Qn|
    where n+1 is the number of prepared qubit states.

    Returns...
    """
    #Data retreival and preparation
    measure = data.parameters['measure']
    #Use this line for Dan style data sets
    #qubits = [data.parameters[data.parameters['config'][m]] for m in measure]
    #Use this line for Evan style data sets
    qubits = [data.parameters[m] for m in measure]
    nQubits = len(qubits)
    readoutDevice = data.parameters[qubits[0]['readoutConfig']]
    for q in qubits:
        q['sidebandFrequency'] = q['readoutFrequency'] - \
                                 readoutDevice['carrierFrequency']
    numberOfRepetitions = int(max(set(data[:,0])))+1
    #Figure out how many different traces (ie, states) we prepared. Number of
    #columns should be 2 per qubit per state. There are 2**nQubits states so
    #there should be 2**(N+1) columns of IQ data, plus 2 more for repetition
    #number and time axis
    nStates = 2**nQubits
    nColumnsPerQubitPerState = 2 #I and Q
    assert data.shape[1] == nColumnsPerQubitPerState*nQubits*nStates + 2
    numTraceCols = nColumnsPerQubitPerState*nQubits*nStates
    #reshaped has three indices, repetition, time, dependentVariable
    reshaped = np.reshape(data, (numberOfRepetitions, -1, numTraceCols + 2))
    time_ns = reshaped[0][:,1] * ((Value(1.0, data.indep[1][1]))['ns'])
    #SECTION I - Group data by qubit and switch from I/Q to z = I + jQ
    dataQubits = []
    for i, qubit in enumerate(qubits):
        I = reshaped[:, :, 2 + 2*i::2*nQubits]
        Q = reshaped[:, :, 2 + 2*i + 1::2*nQubits]
        z = I + 1.0j*Q
        sbGHz = qubit['sidebandFrequency']['GHz']
        phasor = np.exp(1.0j*2*np.pi*time_ns*sbGHz)
        phasor.shape = (phasor.shape[0], 1) #allow broadcasted multiplication
        z = z * phasor
        dataQubits.append(z)
    del i, qubit, I, Q, z, sbGHz, phasor
    #SECTION II - Get heralding and integration ranges from user
    if heraldRanges is None:
        heraldRanges = []
        for dat in dataQubits:
            compressedIQ = np.mean(dat, 0)
            heraldRanges.append(selectPreparationAndMeasureTimes(time_ns,
                                                            compressedIQ.real))
        del compressedIQ
    #SECTION III - Heralding
    #We determine ground state-ness by whether or not a point is close to the
    #center of the distribution
    groundStateMaskQubits = [] #Each element flattened array of True/False
    for i, (dat, heraldPoints) in enumerate(zip(dataQubits, heraldRanges)):
        statePrepStart_ns, statePrepEnd_ns, _, _ = heraldPoints
        indicesTime = np.argwhere(np.logical_and(time_ns > statePrepStart_ns,
                                                 time_ns < statePrepEnd_ns))
        integrated = np.sum(dat[:, indicesTime, :], 1)
        integrated.shape = (integrated.shape[0], integrated.shape[2])
        integrated = integrated.flatten()
        center = iq.iqCenter(integrated)
        std = np.std(integrated)
        groundStateMask = np.abs(integrated - center) < std
        indicesGroundState = np.argwhere(groundStateMask)
        groundStateMaskQubits.append(groundStateMask)
        if debug:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(integrated.real, integrated.imag, 'b.')
            ax.plot(integrated[indicesGroundState].real,
                    integrated[indicesGroundState].imag, 'r.')
            ax.plot(center.real, center.imag, 'g.', markersize=20)
    del i, dat, heraldPoints
    del statePrepStart_ns, statePrepEnd_ns, indicesTime, integrated
    del center, std, groundStateMask, indicesGroundState
    #SECTION IV - Integrate
    worldLinesQubits = []
    for i, (dat, heraldPoints) in enumerate(zip(dataQubits, heraldRanges)):
        _, _, measureStart_ns, measureEnd_ns = heraldPoints
        indicesTime = np.argwhere(np.logical_and(time_ns > measureStart_ns,
                                                 time_ns < measureEnd_ns))
        indicesTime.shape = (indicesTime.shape[0],)
        #Find optimal integration window
        if optimalWindow:
            gIndices = tuple([s for s in range(nStates) \
                              if not qubitExcitedInState(s, i, nQubits)])
            g = dat[:, indicesTime, :][:, :, gIndices]
            #Note that the second mean operates on the 2nd axis of g because
            #the first mean strips an index off.
            g = np.mean(np.mean(g, axis=0), axis=1)
            eIndices = tuple([s for s in range(nStates) \
                              if qubitExcitedInState(s, i, nQubits)])
            e = dat[:, indicesTime, :][:, :, eIndices]
            e = np.mean(np.mean(e, axis=0), axis=1)
            window = np.abs(g - e) #Optimal window
            window = np.convolve(window, np.ones(10), mode='same')
            if debug:
                fig = plt.figure()
                ax = fig.add_subplot(111)
                ax.plot(window, 'b.')
            window.shape = (1, window.shape[0], 1)
            worldLinesQubits.append(np.cumsum(window * dat[:, indicesTime, :],
                                              axis=1))
            del gIndices, eIndices, g, e
        else:
            worldLinesQubits.append(np.cumsum(dat[:, indicesTime, :], axis=1))
    del i, dat, heraldPoints
    del measureStart_ns, measureEnd_ns, indicesTime
    #SECTION VI - Combine heralding data from all qubits
    #Require that all qubits heralded in ground state.
    #groundStateMaskQubits = np.logical_and(*tuple(groundStateMaskQubits))
    groundStateMaskQubits = reduce(np.logical_and, groundStateMaskQubits)
    groundStateMaskQubits = np.reshape(groundStateMaskQubits, (-1, nStates))
    #SECTION VII - 
    result = []
    for q, worldLines in zip(qubits, worldLinesQubits):
        dat = {}
        for s in range(nStates):
            mask = groundStateMaskQubits[:, s]
            dat[s] = np.vstack((wl for wl in worldLines[mask, :, s]))
        result.append(dat)
    del dat, s, mask
    return result
    
def iqTrace_copyForQubits(data, nQubits):
    """
    
    The iq trace analysis code (eg IQTrace_multiQubit) was set up expecting to
    have a pair of IQ traces for each state _for each qubit_. We don't
    actually store multiple copies of the IQ time trace any more so we need to
    make the copies in local memory.
    """
    dat = data[:,2:]
    exploded = []
    assert (data.shape[1]-2) % 2 == 0
    nStates = (data.shape[1]-2) // 2
    for i in range(nStates):
        exploded.extend([dat[:,2*i:2*(i+1)]]*nQubits)
    dat = np.hstack(exploded)
    return np.hstack((data[:,0:2], dat))
    
def iqTrace_combineDatasets(datasets):
    """Combine multiple iq trace data to be analyzed with IQTrace_multiqubit
    
    Combine a series of iq trace datasets (taken with multiple qubits) into a
    single data set. The only thing we're doing here is incrimenting the
    independent variable "repetition" as we vstack the data blocks together.
    We then pull the metadata from the first data set and stamp it onto the
    combined data set.
    """
    datasets = [d.copy() for d in datasets]
    currentStartingRep = 0
    for dataset in datasets:
        print currentStartingRep
        dataset[:,0] += currentStartingRep
        currentStartingRep = max(dataset[:,0]) + 1
    data = np.vstack(datasets)
    data = iqTrace_copyForQubits(data, len(datasets[0].parameters['measure']))
    print("Done with packing")
    print("Now extracting meta data")
    dataset = dv.LocalDataSet(data, datasets[0].indep,
                                    datasets[0].dep,
                                    datasets[0].parameters,
                                    'iqTraces')
    return dataset
    
def cumulatives2Histograms(cumulativesA, cumulativesB, bins=70):
    """
    cumulativesA - ndarray: Each row is a complex valued cumulatively
                   integrated IQ world line. The many rows correspond to the
                   many repetitions of the experiment.
    cumulativesB - ndarray: Similar to cumulativesA
    bins - int: number of bins for each histogram
    
    Returns two lists of histograms, one for each set of worldlines, eg. one
    for each prepared state.
    """
    assert cumulativesA.shape[1] == cumulativesB.shape[1]
    histsA = []
    histsB = []
    for idx in range(len(cumulativesA[0,:])):
        hA, hB = iq.iq2Histograms(cumulativesA[:,idx], cumulativesB[:,idx],
                                    bins=bins)
        histsA.append(hA)
        histsB.append(hB)
    return histsA, histsB
    
def fidelity(histogramA, histogramB, plot=False, fullOutput=False):
    """Find the separation fidelity from two histograms
    
    histogramA - tuple: First element is an array of counts, second element is
                        an array of bin edges. Note that the computations done
                        here assume the bins have constant width.
    histogramB - tuple: Similar to histogramB but for a different prepared
                        state.
    
    Returns a tuple of (separation fidelity, |0> fidelity, |1> fidelity)
    """
    def model(x, maxAmp, mean, sigma):
        return fitting.gaussian(x, maxAmp, mean, sigma, 0)
    if plot:
        fig = plt.figure(); ax = fig.add_subplot(111)
    results = []
    actualFid = []
    colors = ['b','r']
    for h, color in zip([histogramA, histogramB], colors):
        binCenters = (h[1][:-1] + h[1][1:]) / 2 #Assumes uniform bin width
        binWidth = binCenters[1] - binCenters[0]
        counts = h[0]
        maxBin = binCenters[np.argmax(counts)]
        maxVal = np.max(counts)
        if plot:
            plotIndices = np.argwhere(counts > 0)
            ax.bar(binCenters[plotIndices]-binWidth/2,
                   counts[plotIndices], binWidth,
                   log=True, alpha=0.5, color=color, linewidth=0)
        vGuess = (maxVal, maxBin, np.std(counts*binCenters))
        v, cov, fitFunc = fitting.fitCurve(model, binCenters, h[0], vGuess)
        results.append(v)
        
        #Actual error
        if v[1] > 0:
            indices = np.argwhere(binCenters < 0)
        elif v[1] < 0:
            indices = np.argwhere(binCenters > 0)
        else:
            raise RuntimError("I'm confused")
        indices.shape = (indices.shape[0],)
        errorFraction = float(sum(counts[indices]))/sum(counts)
        actualFid.append(1 - errorFraction)
        
        if plot:
            ax.semilogy(binCenters, model(binCenters, *v), color+'-',
                        linewidth=6)
            ax.set_ylim([0.6,200])
    fidG, fidE = actualFid
    separation = results[0][1] - results[1][1]
    sA, sB = results[0][2], results[1][2]
    sepFid = 0.5*special.erfc(separation/(2*np.sqrt(sA**2 + sB**2)))
    toReturn = sepFid, fidG, fidE
    if fullOutput:
        return toReturn, ax
    else:
        return toReturn
    
####QUARANTINE
    
def realFidelity(histogramA, histogramB, plot=False):
    raise RuntimeError("This doesn't work yet")
    
    def model(x, maxAmp, mean, sigma):
        return fitting.gaussian(x, maxAmp, mean, sigma, 0)
    centers = []
    for h in [histogramA, histogramB]:
        binCenters = (h[1][:-1] + h[1][1:]) / 2 #Assumes uniform bin width
        counts = h[0]
        maxBin = binCenters[np.argmax(counts)]
        maxVal = np.max(counts)
        vGuess = (maxVal, maxBin, np.std(counts*binCenters))
        v, cov, fitFunc = fitting.fitCurve(model, binCenters, h[0], vGuess)
        centers.append(v[1])
    separationPoint = np.mean(centers)
    errors = []
    for center, h in zip(centers, (histogramA, histogramB)):
        #Find fraction of points in the wrong place
        binCenters = (h[1][:-1] + h[1][1:]) / 2
        if center > separationPoint:
            errorIndices = np.argwhere(binCenters < separationPoint)
        else:
            errorIndices = np.argwhere(binCenters > separationPoint)
        errors.append(np.sum(h[0][errorIndices]) / np.sum(h[0]))
    return errors
    
def dan2Evan(data, timeIdx):
    stuff = []
    for state in range(4):
        q0 = np.vstack((data[0][state][:, timeIdx].real, data[0][state][:, timeIdx].imag))
        q1 = np.vstack((data[1][state][:, timeIdx].real, data[1][state][:, timeIdx].imag))
        temp = np.vstack((q0,q1))
        temp = np.vstack((np.zeros(temp.shape[1])+state, temp)).T
        stuff.append(temp)
    return np.vstack(stuff)
    
def analyze_two_qubit(ds):
    qubits = ['q2', 'q4']
    z = 1j*np.zeros((ds.shape[0], len(qubits)))
    discrimination_parameters = [()] * len(qubits)
    for idx, q in enumerate(qubits):
        z[:,idx] = ds[:,2*idx+1] + 1j*ds[:,2*idx+2]
        mask_1 = np.bitwise_and(np.array(ds[:,0], dtype=int), 1<<(len(qubits)-idx-1)) > 0
        mask_0 = np.logical_not(mask_1)
        z0 = z[mask_0, idx]
        z1 = z[mask_1, idx]
        center, axis, _, _ = iq.discriminationAxis(z0, z1)
        discrimination_parameters[idx] = (center, axis)
    plt.figure()
    for state in [0,1,2,3]:#util.uniq(ds[:,0]):
        mask = ds[:,0] == state
        x = np.real((z[mask,0] - discrimination_parameters[0][0]) / discrimination_parameters[0][1])
        y = np.real((z[mask,1] - discrimination_parameters[1][0]) / discrimination_parameters[1][1])
        state_name = "{0:0{width}b}".format(int(state), width=len(qubits))
        plt.plot(x, y, label='State |%s>' % state_name)
        plt.xlabel('Readout separation |%s>' % (qubits[0],))
        plt.ylabel('Readout separation |%s>' % (qubits[1],))
        x_err = np.sum(x * np.sign(np.median(x)) < 0)*1.0 / np.shape(x)[0]
        y_err = np.sum(y * np.sign(np.median(y)) < 0)*1.0 / np.shape(y)[0]
        (_, x_sep, x_std) = fitGaussianOutliers(x)
        (_, y_sep, y_std) = fitGaussianOutliers(y)
        #print "x separation: %s, std: %s" % (x_sep, x_std)
        #print "y separation: %s, std: %s" % (y_sep, y_std)
        
        print "State |%s> error: %s: %f, %s: %f" % (state_name, qubits[0], x_err, qubits[1], y_err)
        x_sep_f =  0.5-0.5*scipy.special.erf(np.abs(-1.0*x_sep/x_std)/np.sqrt(2))
        y_sep_f =  0.5-0.5*scipy.special.erf(np.abs(-1.0*y_sep/y_std)/np.sqrt(2))
        print "State |%s> separation error: %s: %f, %s: %f" % (state_name, qubits[0], x_sep_f, qubits[1], y_sep_f)
    plt.legend()
    return z
    
####QUARANTINE###

def separationFidelityVsTime(histogramsA, histogramsB):
    """Find separation fidelity vs time from a set of IQ histograms
    
    histogramsA - list: IQ histograms for qubit/state A. Each element is a
                  histogram of the IQ points for a single time slice.
    histogramsB - list: similar to histogramsB
    """
    fidelities = []
    for i, (hA, hB) in enumerate(zip(histogramsA, histogramsB)):
        sep, actG, actE = fidelity(hA, hB)
        fidelities.append(sep)
    return np.array(fidelities)
    
def fidelitiesVsTime(histogramsA, histogramsB):
    """Find fidelities vs time from a set of IQ histograms
    
    histogramsA - list: IQ histograms for qubit/state A. Each element is a
                  histogram of the IQ points for a single time slice.
    histogramsB - list: similar to histogramsB
    """
    sep = []
    fidG = []
    fidE = []
    for i, (hA, hB) in enumerate(zip(histogramsA, histogramsB)):
        sepFid, actG, actE = fidelity(hA, hB)
        sep.append(sepFid)
        fidG.append(actG)
        fidE.append(actE)
    return np.array(sep), np.array(fidG), np.array(fidE)
    
def plotTimeDomain(result, highlightAt=None, legendLoc='lower left',
                   whichQubit=None, qubitNames=None, bins=70):
    """Show measurement fidelity as a function of time
    
    result is what comes out of IQTrace_multiQubit
    whichQubit - int: Chooses a single qubit to plot. If left as default(None)
                      all qubits are plotted
    """
    nQubits = len(result)
    if nQubits > 2:
        print("This probably won't work because you have more than 2 qubits")
    if qubitNames is None:
        qubitNames = ['Q%d'%i for i in range(nQubits)]
    states = np.arange(2**nQubits)
    if whichQubit is not None:
        #Re-assign result to have data only from selected qubit
        result = [result[whichQubit]]
    fig = plt.figure()
    ax = fig.add_subplot(111)
    for i, (dat, mrkr, size, qubitName) in \
                    enumerate(zip(result,['.','s'], [40,20], qubitNames)):
        g = np.vstack([dat[s] for s in states if not qubitExcitedInState(s, i, nQubits)])
        e = np.vstack([dat[s] for s in states if qubitExcitedInState(s, i, nQubits)])
        #XXX This is a hack that should not exist, keeping only first 200ns
        e = e[:,0:100]
        g = g[:,0:100]
        
        hg, he = cumulatives2Histograms(g, e, bins=bins)
        sep, actualG, actualE = fidelitiesVsTime(hg, he)
        t = np.arange(len(hg))*2
        print('WARNING - assuming time axis is in units of 2ns!!!')
        #t = np.arange(g.shape[1])*2
        ax.semilogy(t, 1-sep, 'g'+mrkr, markeredgewidth=0, markersize=size,
                    label='%s - '%qubitName + r'$\epsilon_{s}$')
        ax.semilogy(t, 1-actualG, 'b'+mrkr, markeredgewidth=0,
                    markersize=int(2*size/3.0), alpha=0.7,
                    label='%s - '%qubitName + r'$\epsilon_{|0\rangle}$')
        ax.semilogy(t, 1-actualE, 'r'+mrkr, markeredgewidth=0,
                    markersize=int(2*size/3.0), alpha=0.7,
                    label='%s - '%qubitName + r'$\epsilon_{|1\rangle}$')
        if highlightAt is not None:
            idx = np.argwhere(t == highlightAt)
            idx.shape = (idx.shape[0],)
            ax.semilogy(t[idx], 1-sep[idx], 'r'+mrkr, markeredgewidth=0, markersize=size)
    ax.xaxis.set_label_text('Pulse integration time [ns]')
    ax.yaxis.set_label_text('Error')
    ax.legend(loc=legendLoc, numpoints=1, prop={'size':42})
    return ax
    
def plotTimeDomain_multiplot(result, xpts=150):
    """Plot multiqubit readout with one panel per qubit"""
    nQubits = len(result)
    qubitNames = ['Q%d'%(i+1,) for i in range(nQubits)]
    states = np.arange(2**nQubits)
    size=10
    fig = plt.figure()
    for i, (dat, qubitName) in enumerate(zip(result, qubitNames)):
        ax = fig.add_subplot(2,2,i+1)
        g = np.vstack([dat[s] for s in states if not qubitExcitedInState(s, i, nQubits)])
        e = np.vstack([dat[s] for s in states if qubitExcitedInState(s, i, nQubits)])
        hg, he = cumulatives2Histograms(g, e)
        sep, actualG, actualE = fidelitiesVsTime(hg[:xpts], he[:xpts])
        t = np.arange(xpts)*2
        #t = np.arange(g.shape[1])*2
        ax.semilogy(t, 1-sep, 'g', markeredgewidth=0, markersize=size,
                    label='%s - '%qubitName + r'sep')
        ax.semilogy(t, 1-actualG, 'b', markeredgewidth=0,
                    markersize=int(2*size/3.0), alpha=0.7,
                    label='%s - '%qubitName + r'$|0\rangle$')
        ax.semilogy(t, 1-actualE, 'r', markeredgewidth=0,
                    markersize=int(2*size/3.0), alpha=0.7,
                    label='%s - '%qubitName + r'$|1\rangle$')
        ax.xaxis.set_label_text('Pulse integration time [ns]')
        ax.yaxis.set_label_text('Error')
        ax.set_title(qubitName)
        #ax.legend(loc='lower right', numpoints=1, prop={'size':24})
    return fig

def iqToExcited(data):
    """Convert stimulated transition data taken in IQ mode to probabilities"""
    def readoutProbabilities(q, z):
        center = q['readoutCenterZ']['']
        axis = q['readoutAxis']['']
        corrected = (z - center) * axis.conjugate()
        return corrected.real > 0
    m = data.parameters['measure'][0]
    q = data.parameters[data.parameters['config'][m]]
    assert data.shape[1]%2 == 0
    traceCols = (data.shape[1]-2)/2
    depData = data[:,2:]
    depData = np.column_stack((depData[:,2*i]+1.0j*depData[:,(2*i)+1] \
                               for i in range(traceCols)))
    depData = np.column_stack((readoutProbabilities(q, depData[:,i])\
                               for i in range(traceCols)))
    return np.concatenate((data[:,0:2], depData), 1)
    
def stimulatedEmission(data):
    """Stimulated emission/excitation
    
    Data format should be 
    indep = Drive power, Drive time
    dep = Prepared state no pi pulse
          Measured state no pi pulse
          Prepared state w/ pi pulse
          Measured state w/ pi pulse
    """
    result = []
    uniqueXVals = set(data[:,0])
    uniqueYVals = set(data[:,1])
    numUniqueXVals = len(uniqueXVals)
    numUniqueYVals = len(uniqueYVals)
    for x in uniqueXVals:
        for y in uniqueYVals:
            xMask = data[:,0] == x
            yMask = data[:,1] == y
            thisData = data[np.logical_and(xMask, yMask), 4:]
            thisData = thisData[thisData[:,0] == 0, 1]
            fractionUp = np.sum(thisData) / float(len(thisData))
            result.extend([x, y, fractionUp])
    result = np.reshape(result, (-1, 3))
    return result
    
def stimulatedEmission2(data):
    """Stimulated emission/exciation
    
    Data format should be 
    indep = Drive power, Drive time
    dep = Measured state no pi pulse
          Measured state w/ pi pulse
    """
    uniquePowers = set(data[:,0])
    uniqueTimes = set(data[:,1])
    fig = plt.figure()
    ax = fig.add_subplot(111)
    colors = ('b', 'r', 'g')
    for t,c in zip(uniqueTimes, colors):
        print("Processing at time %f"%t)
        indices = np.argwhere(data[:,1] == t)
        indices.shape = (indices.shape[0],)
        dat = data[indices, :]
        indices = np.argsort(dat[:,0])
        indices.shape = (indices.shape[0],)
        dat = dat[indices, :]
        pows = []
        probs0 = []
        probs1 = []
        for pow in set(dat[:,0]):
            indices = np.argwhere(dat[:,0] == pow)
            pows.append(pow)
            probs0.append(np.mean(dat[indices, 2]))
            probs1.append(np.mean(dat[indices, 3]))
        ax.semilogy(pows, probs0, c+'o', label=str(t)+' ns, 0->1', markersize=8,
                    markeredgewidth=0, alpha=0.7)
        ax.semilogy(pows, probs1, c+'s', label=str(t)+' ns, 1->0', markersize=8,
                    markeredgewidth=0, alpha=0.7)
    ax.xaxis.set_label_text('Power [dBm]')
    ax.yaxis.set_label_text('Time [ns]')
    ax.legend(loc='lower right', numpoints=1)
    return ax
        
#############
### TRASH ###
#############

def TRASH_cumulatives2Histograms(time_ns, cumulatives, stateA, stateB,
                                 numHistBins=70):
    """Generate projected IQ histograms from cumulative integrated time traces
    
    time_ns - array: time for each slice
    cumulatives - dict: keys are the state in which the qubit was prepared.
                        values are ndarrays where each row is a complex valued
                        readout worldline. The many rows correspond to many
                        repetitions of the experiment in which the initialized
                        state was the one indicated by the key.
    
    Returns a dict mapping prepared state to a list of histograms. Each index i
    into the list corresponds to data taken at time time_ns[i]. The list
    elements are tuples of ndarray: (counts, bin edges). Note that len(counts)
    is 1 less than len(bin edges).
    """
    raise RuntimeError("Depricated, use cumulatives2Histograms")
    states = [stateA, stateB]
    data = dict([(state, []) for state in states]) #state -> list of histograms
    for idx, t in enumerate(time_ns):
        slices = dict([ \
            (state, cumulatives[state][:,idx]) \
            for state in states])
        #Find the line separating the two clouds...
        center, axis, (_, _), (_, _) = iq.discriminationAxis(slices[stateA],
                                                             slices[stateB])
        #...and use it to project the IQ points into a line.
        slicesProjected = dict([ \
            (state, ((slices[state] - center) * axis.conjugate()).real) \
            for state in states])
        #Make a histogram
        for state in states:
            dat = scipy.histogram(slicesProjected[state], bins=numHistBins)
            data[state].append(dat)
    return data
    
