# pyle.analysis.sr770
# for analyzing data from the SR 770 Spectrum analyzer
# see pyle.dataking.sr770.noiseTest for info.

import matplotlib.pyplot as plt
from pyle.plotting import dstools
import FluxNoiseAnalysis as fna
import numpy as np
from scipy.optimize import curve_fit
from pyle.util.registry_wrapper2 import RegistryWrapper

# DAC amp/Z pulse amp of 1 corresponds to about 400 mV
ZPA_to_V = 0.4
## util
ds_cache = {}
def getDataset(dv, num, path):
    k = '%s - %s' % (num, path)
    if k not in ds_cache:
        ds_cache[k] = dstools.getDeviceDataset(dv, num, path)
    return ds_cache[k]
def invalidateCache():
    ds_cache = {}
    
def plotDatasets(s, sets):
    dv = s._cxn.data_vault
    datasets = [getDataset(dv, x, s._dir) for x in sets]
    f = plt.figure()
    ax = f.add_subplot(111)
    for d in datasets:
        ax.loglog(d.data[:,0], d.data[:,1], 'o')
        
        
def dfdZPA(s, dataset, path=None, **kwargs):
    ''' get d-frequency / d-Z-pulse-amplitude from a given spectroscopy Z pulse dataset.
    units: GHz / DAC amp '''
    if path is None:
        path = s._dir
    d = getDataset(s._cxn.data_vault, dataset, path)
    f_p, A, B = fna.fitSpectroscopy(d, **kwargs)
    slope = - A * f_p / 2 * np.sin(B) / np.sqrt(np.cos(B))
    return slope
    
def dacVoltsToFreq(v, slope):
    ''' convert from volts to frequency. '''
    zpa = v / ZPA_to_V
    freq = zpa * slope
    return freq
    
def makeFrequencySpectrum(s, dataset, specDataset, path=None, specPath=None, **kwargs):
    if path is None:
        path = s._dir
    dv = s._cxn.data_vault
    d = getDataset(dv, dataset, path)
    gain = d['parameters']['preamp']['gain']
    slope = np.abs(dfdZPA(s, specDataset, specPath, **kwargs))
    freq = dacVoltsToFreq(d.data[:,1], slope) / gain
    f = plt.figure()
    ax = f.add_subplot(111)
    ax.loglog(d.data[:,0], freq, 'o')

class DSGroup (object):
    def __init__ (self, dsNums, path, cxn):
        self.dsNums = dsNums
        self.path = path
        self.dv = cxn.data_vault
        self.dv.cd(self.path)
        self.dsNames = self.dv.dir([])[1]
        self.title = self.getDSMiddlePart(self.dsNums[0])
        self.axesNames = None
        self.axesUnits = None
        self.gain = 1
        
    def __str__ (self):
        return self.title
        
    def __repr__ (self):
        # not supposed to do this but whatever
        return self.__str__()
        
    def getDSMiddlePart(self, num):
        ''' get the middle part of the dataset name'''
        dsName = self.dsNames[num-1]
        dsnum, dash, right = dsName.partition('-')
        assert num == int(dsnum.strip())
        return right.rpartition('-')[0].strip()
        
    def isValid (self):
        ''' make sure that all datasets are for the same measurement '''
        if not self.dsNums:
            return False
        for n in self.dsNums[1:]:
            if self.getDSMiddlePart(n) != self.title:
                return False
        return True
        
    def getData (self, applyGain = True, sort = True):
        ''' load all the datasets and concatenate the data '''
        d = np.vstack([getDataset(self.dv, n, self.path)['data'] 
                        for n in self.dsNums]) 
        leadDataset = getDataset(self.dv, self.dsNums[0], self.path)
        if applyGain and 'preamp' in leadDataset['parameters']:
            d[:,1] /= leadDataset['parameters']['preamp']['gain']
            self.gain = leadDataset['parameters']['preamp']['gain']
        self.axesNames = (leadDataset['variables'][0][0][0],
                            leadDataset['variables'][1][0][0])
        self.axesUnits = (leadDataset['variables'][0][0][1],
                            leadDataset['variables'][1][0][2])
        d = d[d[:,0].argsort()]
        return d
        
    def getSmoothedData(self, window=5, applyGain=True):
        d = self.getData(applyGain=applyGain)
        d[:,1] = np.convolve(d[:,1], np.ones(window)/float(window), mode='same')
        return d
        
    def getBinnedData(self, nbins=40, applyGain=True):
        d = self.getData(applyGain=applyGain)
        bins = np.logspace(np.log10(d[:,0].min()), np.log10(d[:,0].max()), nbins)
        bins = np.concatenate([np.array([0]), bins])
        vals = np.zeros(nbins)
        for i in range(nbins):
            vals[i] = d[np.logical_and(d[:,0] > bins[i], d[:,0] <= bins[i+1]), 1].mean()
        return np.vstack([bins[1:], vals]).T
        
    def loglog(self, ax, label='', smooth=0, addGainToLabel=False):
        ''' plot on provided axes '''
        if smooth:
            d = self.getSmoothedData(window=smooth)
        else:
            d = self.getData()
        if label == '':
            label = self.title
        if addGainToLabel:
            label += ' g=%d' % self.gain
        ax.loglog(d[:,0], d[:,1], '.', label=label)
        
    def fitData(self, alpha_guess=0.5, S_guess=None, white_guess=None):
        ''' fit to 1/f + white '''
        d = self.getData()
        # white noise guess is min measured
        if white_guess is None:
            white_guess = d[:,1].min()
        # S_1Hz guess is value at 1Hz
        if S_guess is None:
            x = np.abs(np.log10(d[:,0])).min()
            _1Hz_idx = np.nonzero(x == x.min())[0][0]
            S_guess = d[_1Hz_idx,1]
        popt = [alpha_guess, S_guess, white_guess]
        sigma = np.logspace(np.log10(d[:,0].min()), np.log10(d[:,0].max()), len(d[:,0]))
        popt, pcov = curve_fit(fitfunc, d[:,0], d[:,1], popt, sigma=1.0/sigma)
        return popt
        
    def loglogFit(self, ax, label='', quiet=False):
        ''' run fit and plot it '''
        if label == '':
            label = self.title + ' Fit'
        popt = self.fitData()
        d = self.getData()
        if not quiet:
            print "Fit parameters: slope: %.2f, S_1Hz: %.2g, floor: %.2g" % tuple(popt)
        ax.loglog(d[:,0], fitfunc(d[:,0], *popt), '.', label=label)
    
def fitfunc(freq, alpha, S, white):
    return np.power(freq, alpha) * S + white
        
def buildGroups(s):
    sr560 = [range(129, 144)]
    sr770 = [range(253, 268)]
    cross_50ohm = [range(464,479)]
    diffamps = [range(314,329), range(329, 344),
                # distrust everything else, diffamp may not have been plugged in
                #range(145, 160), range(160, 175),
                #range(175, 190), range(190, 205), # these are 100MHz
                #range(205, 223), range(223, 238), # these have gain = 100
                #range(296, 314)
                ]
    
    dac_dual = [range(99,114), range(114,128),      # these 4 appear good and
                range(344, 359), range(359,374),    # agree w/each other
                #range(52, 59), range(59, 66), range(66,99), # these are no good
                #range(268, 283) # gain=100
                ]
    dac_cross = [range(374,389), range(389,404)]
    diff_cross =[range(404, 419),range(419, 434),range(434, 448),range(448,464)]
    return [DSGroup(x, s._dir, s._cxn) for x in sr770 + sr560 +
            diffamps + dac_dual + dac_cross + diff_cross]
            
def buildNewGroup(s):
    diff_brooks = [range(2,17), range(17,32), range(32,47), range(47, 62)]
    return [DSGroup(x, ['', 'Peter', 'rtNoise', '140122'], s._cxn) for x in diff_brooks]

def getRegWrapper(cxn):
    return RegistryWrapper(cxn, ['', 'Peter', 'rtNoise', '131230'])
