import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K')]
import numpy as np
import time
import pyle.dataking.util as util

def freqScan(sample, PNA = None, save = True,start=None,stop=None):

    cxn = sample._cxn
    if PNA:
        ampMeas = PNA
    else:
        samp, dev = util.loadDeviceType(sample, 'PNA')
        ampMeas = dev[0]
    if start is None:
        start = ampMeas['startFreq']*GHz
    if stop is None:
        stop = ampMeas['stopFreq']*GHz
    num_points = ampMeas['number of points']
    averages = ampMeas['averages']
    power = ampMeas['power']*dBm
    bandwidth = ampMeas['bandwidth']*Hz
    
    # num_points is the number of points per pna sweep
        
    if ampMeas['port'] == 'Port1':
        independents = ['frequency [GHz]']
        dependents = ['Magnitude (S21) [dB]', 'Phase (S21) [Deg]']
    elif ampMeas['port'] == 'Port2':
        independents = ['frequency [GHz]']
        dependents = ['Magnitude (S12) [dB]']
    
    # get the parameters from the data vault
    
    def get_pna_data(start, stop):
        pna.frequency_range(start, stop)
        raw_data = pna.freq_sweep()
        f = raw_data[0].asarray
        s = raw_data[1].asarray.T
        f.shape = (len(f), 1)
        pna_data = np.hstack((f, s))
        return pna_data
    
    def getMagPhase(c): #returns magnitude and phase of pna data
        return 2 * 10 * np.log10(abs(c)), np.angle(c)
    
    pna = cxn.pna_x
    pna.select_device(ampMeas['pna_gpib'])
    pna.power(power)
    pna.bandwidth(bandwidth)
    pna.num_points(int(num_points))
    
    if ampMeas['port'] == 'Port1':
        pna.s_parameters(['S11', 'S21'])
    elif ampMeas['port'] == 'Port2':
        pna.s_parameters(['S12', 'S22'])
        
    # set up data vault parameters
    raw_data = get_pna_data(start, stop)
    freqs = raw_data[:,0]*1e-9 # (downconvert to GHz)
    S_through = getMagPhase(raw_data[:,2]) # S12 or S21, depending on PORT
    S_reflected = getMagPhase(raw_data[:,1]) # S11 or S22 depending on PORT
    data = np.array(np.vstack((freqs.T, S_through[0], S_through[1]))).T.astype(float)
    if save == True:        
        saveData(sample, data)
    else:
        return data
    
 
def freqScanPhase(sample):

    cxn = sample._cxn
    samp, dev = util.loadDeviceType(sample, 'PNA')
    ampMeas = dev[0]
    start = ampMeas['startFreq']*GHz
    stop = ampMeas['stopFreq']*GHz
    num_points = ampMeas['number of points']
    averages = ampMeas['averages']
    power = ampMeas['power']*dBm
    bandwidth = ampMeas['bandwidth']*Hz
    edelay = ampMeas['electricalDelay'].value
    
    # num_points is the number of points per pna sweep
        
    if ampMeas['port'] == 'Port1':
        independents = ['frequency [GHz]']
        dependents = ['Phase (S21) [rad]']
    elif ampMeas['port'] == 'Port2':
        independents = ['frequency [GHz]']
        dependents = ['Phase (S12) [rad]']
    
    # get the parameters from the data vault
    
    def get_pna_phase(start_freq, stop_freq):
        pna.frequency_range(start_freq, stop_freq)
        raw_data = pna.freq_sweep_phase()
        return raw_data
    
    pna = cxn.pna_x
    pna.select_device(ampMeas['pna_gpib'])
    pna.power(power)
    pna.bandwidth(bandwidth)
    pna.num_points(int(num_points))
    pna.electrical_delay(edelay)
    
    if ampMeas['port'] == 'Port1':
        pna.s_parameters(['S11'])
    elif ampMeas['port'] == 'Port2':
        pna.s_parameters(['S12'])
        
    # set up data vault parameters
    dv = cxn.data_vault
    dv.cd('')
    dv.cd(sample._dir, True)
    raw_data = get_pna_phase(start, stop)
    freqs = np.asarray(raw_data[0])*1e-9 # (downconvert to GHz)
    freqs.shape = (len(freqs),1)
    phase21 = np.asarray(raw_data[1][0])
    
    data = np.array(np.hstack((freqs,phase21)))
    
    if save == True:    
        saveData(sample, data)
    else:
        return data
    
def freqScanFast(pna):
    def getMagPhase(c): #returns magnitude and phase of pna data
        return 2 * 10 * np.log10(abs(c)), np.angle(c)
    raw_data = pna.freq_sweep()
    f = raw_data[0].asarray
    s = raw_data[1].asarray.T
    f.shape = (len(f), 1)
    pna_data = np.hstack((f, s))
    freqs = pna_data[:,0]*1e-9 # (downconvert to GHz)
    S_through = getMagPhase(pna_data[:,2]) # S12 or S21, depending on PORT
    data = np.array(np.vstack((freqs.T, S_through[0]))).T.astype(float)
    return data
    
    
def saveData(sample, data):
    cxn = sample._cxn
    ampMeas = sample['PNA']
    dv = cxn.data_vault
    dv.cd(sample._dir, True)
    nameString = 'Frequency Sweep ' + ampMeas['name']
    if ampMeas['port'] == 'Port1':
        independents = ['frequency [GHz]']
        dependents = ['Magnitude (S21) [dB]', 'Phase (S21) [Deg]']
    elif ampMeas['port'] == 'Port2':
        independents = ['frequency [GHz]']
        dependents = ['Magnitude (S12) [dB]']
    nameString = 'Frequency Sweep ' + ampMeas['name']
    dv.new(nameString, independents, dependents)
    p = dv.packet()
    for key in ampMeas.keys():
        p.add_parameter(key, ampMeas.__getitem__(key))
    p.send()
    dv.add(data)
    
def saveData3D(sample, data, dataVault = None):
    if dataVault == None:
        cxn = sample._cxn
        ampMeas = sample['PNA']
        dv = cxn.data_vault
        dv.cd(sample._dir, True)
        nameString = 'Frequency Sweep ' + ampMeas['name']
        if ampMeas['port'] == 'Port1':
            independents = ['frequency [GHz]','Pump Power [dBm]']
            dependents = ['Magnitude (S21) [dB]', 'Phase (S21) [Deg]']
        elif ampMeas['port'] == 'Port2':
            independents = ['frequency [GHz]','Pump Power [dBm]']
            dependents = ['Magnitude (S12) [dB]']
        nameString = 'Frequency Sweep ' + ampMeas['name']
        dv.new(nameString, independents, dependents)
        p = dv.packet()
        for key in ampMeas.keys():
            p.add_parameter(key, ampMeas.__getitem__(key))
        p.send()
        dv.add(data)
        return dv
    else:
        dataVault.add(data)
    
    