import numpy as np
import itertools

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

import scipy.signal as spsignal

from pyle import datavault
from pyle.dataking import util

from pyle.dataking.DevDaniel.Qubit import readout

from labrad.units import Unit,Value

MHz,GHz = (Unit(u) for u in ['MHz','GHz'])

def plotFt(t, z):
    plt.figure()
    freq = np.fft.fftfreq(len(t), t[1]-t[0])
    plt.semilogy(freq, np.abs(np.fft.fft(z)))
    
def mix(t, z, f):
    """t and f are assumed to be in complementary units"""
    phasor = np.exp(2*np.pi*1.0j*t*f)
    return phasor*z
    
def phasorFunc(freq):
    def func(t):
        return np.exp(2*np.pi*1.0j*freq*t)
    return func
    
def interquartileRange(x):
    dat = np.sort(x)
    idx0 = len(dat)//4
    idx1 = len(dat)*3//4
    return np.mean(dat[idx0:idx1])
    
### DATASET INSPECTION/PARSING ###
    
def signalIdlerSidebandFrequencies(dataset, qubitName, pumpFrequency=None):
    """
    As of May2013, the readout configuration data is in its own registry
    directory. This includes data like the pump frequency, which is
    important for determining the detection sideband frequencies of the
    signal and idler. For datasets with this style these values are
    extracted from the datavault parameter set.
    
    For datasets taken before the pump frequency and readout carrier were
    stored in their own registry directory, we look for the carrier in the
    qubit's own directory and expect the user to supply the pump frequency.
    """
    if dataset.parameters[qubitName].has_key('readoutConfig'):
        readoutConfigName = dataset.parameters[qubitName]['readoutConfig']
        readoutConfig = dataset.parameters[readoutConfigName]
        
        signalFrequency = dataset.parameters[qubitName]['readoutFrequency']
        carrierFrequency = readoutConfig['carrierFrequency']
        pumpFrequency = readoutConfig['pumpFrequency']

    else:
        signalFrequency = dataset.parameters[qubitName]['readoutFrequency']
        carrierFrequency = dataset.parameters[qubitName]['readout fc']
        assert pumpFrequency is not None
    idlerFrequency = pumpFrequency - signalFrequency
    signalSb = signalFrequency - carrierFrequency
    idlerSb = idlerFrequency - carrierFrequency
    return signalSb, idlerSb
    
def parse_iq_multiple(dataset, time_start=0, time_stop=np.inf):
    '''
    Parses data from IQ vs time traces with multiple repetitions.
    The input should be a dataset wrapper, the output is a 3-D array
    indexed as x[repetition, time_step, col].  
    '''
    nreps = dataset['Repetition'][-1]+1
    t = dataset['Time']
    data_by_rep = []
    time_mask = np.logical_and(dataset['Time'] >= time_start, dataset['Time'] < time_stop)
    for x in range(nreps):
        mask = np.logical_and(dataset['Repetition']==x, time_mask)
        idx = np.flatnonzero(mask)
        mask = slice(idx[0], idx[-1]+1)
        rep_data = dataset[mask, 1:]
        data_by_rep.append(rep_data)
    return np.array(data_by_rep)

### DEMODULATION ###

def datasetTimeChannels2SignalIdlerIq(dataset):
    """Convert a set of IQ time traces to lollipops
    
    RETURNS:
    np.ndarray
    Number of columns is twice the number of IQ time trace channels.
    For each channel you get two columns, one for signal, one for idler.
    Each row corresponds to a repetition of the measurement
    """
    signalSb, idlerSb = signalIdlerSidebandFrequencies(dataset)
    signalPhasorFunc = phasorFunc(signalSb['GHz'])
    idlerPhasorFunc = phasorFunc(idlerSb['GHz'])
    data = parse_iq_multiple(dataset)
    #Make sure there is an even number of columns
    #One colum for repetition, one for time, and then pairs of I and Q
    assert (data.shape[2]-1)%2 == 0
    numChannels = (data.shape[2]-1)//2
    results = []
    for repetition in data:
        t = repetition[:,0]
        dat = repetition[:,1:]
        for ch in range(numChannels):
            I = dat[:,ch*2]
            Q = dat[:,ch*2+1]
            z = I+1.0j*Q
            results.append(np.sum(signalPhasorFunc(t)*z))
            results.append(np.sum(idlerPhasorFunc(t)*z))
    results = np.array(results)
    results = np.reshape(results,(-1,2*numChannels))
    return results
    
def datasetTimeChannels2SignalIq_cumulative(dataset):
    """Convert a set of IQ time traces to cumulatively integrated phasors
    
    RETURNS:
    np.ndarray
    Three indices. First index is repetition
    Second index is channel
    Final index is time
    """
    signalSb, idlerSb = signalIdlerSidebandFrequencies(dataset)
    signalPhasorFunc = phasorFunc(signalSb['GHz'])
    idlerPhasorFunc = phasorFunc(idlerSb['GHz'])
    data = parse_iq_multiple(dataset)
    t = data[0,:,0]
    numTPoints = len(t)
    signalPhasor = signalPhasorFunc(t)
    idlerPhasor = idlerPhasorFunc(t)
    #Make sure there is an even number of columns
    #One colum for repetition, one for time, and then pairs of I and Q
    assert (data.shape[2]-1)%2 == 0
    numChannels = (data.shape[2]-1)//2
    results = []
    for repetition in data:
        dat = repetition[:,1:]
        for ch in range(numChannels):
            I = dat[:,ch*2]
            Q = dat[:,ch*2+1]
            z = I+1.0j*Q
            results.append(np.cumsum(signalPhasor*z))
    results = np.array(results)
    results = np.reshape(results,(-1, numChannels, numTPoints))
    return results

### IQ ANALYSIS ###
    
### SIGNAL/NOISE ###
    
def iqHistogram(SNoPulse, SPulse, center, axis, bins=30):
    """Makes a historam of readout shots"""
    noPulse = ((SNoPulse-center)/axis.conjugate()).real
    pulse = ((SPulse-center)/axis.conjugate()).real
    noPulseMed = np.median(noPulse)
    pulseMed = np.median(pulse)
    totalCounts = len(pulse)    
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    countsNoPulse,binEdgesNoPulse,_ = ax.hist(noPulse, bins, alpha=0.5, color='b', linewidth=0)
    countsPulse,binEdgesPulse,_ = ax.hist(pulse, bins, alpha=0.5, color='r', linewidth=0)
    
    countsNoPulse = np.array(countsNoPulse)
    countsPulse = np.array(countsPulse)
    binEdgesNoPulse = np.array(binEdgesNoPulse)
    binEdgesPulse = np.array(binEdgesPulse)
    
    #Compute errors
    if noPulseMed<0:
        noPulseErrorMask = binEdgesNoPulse>0
        pulseErrorMask = binEdgesPulse<0
    else:
        noPulseErrorMask = binEdgesNoPulse<0
        pulseErrorMask = binEdgesPulse>0
    noPulseErrorMask = np.array(noPulseErrorMask)[1:] #Throw away an element to make mask right length
    pulseErrorMask = np.array(pulseErrorMask)[1:]     #This is slightly bad
    
    numNoPulseErrors = np.sum(countsNoPulse[noPulseErrorMask])
    numPulseErrors = np.sum(countsPulse[pulseErrorMask])
    errorRateNoPulse = float(numNoPulseErrors)/totalCounts
    errorRatePulse = float(numPulseErrors)/totalCounts
    print("Total counts: %d"%totalCounts)
    print("No pulse")
    print("Num errors: %d"%numNoPulseErrors)
    print("Error rate: %f"%errorRateNoPulse)
    print("Pulse")
    print("Num errors: %d"%numPulseErrors)
    print("Error rate: %f"%errorRatePulse)
    

    
def iqSpread_rejectOutliers(projections0, projections1):
    separation = np.median(projections0) - np.median(projections1)
    spread0 = interquartileRange(projections0)
    spread1 = interquartileRange(projections1)
    return separation, spread0, spread1
    
def iqSpread(projections0, projections1):
    separation = np.mean(projections0) - np.mean(projections1)
    spread0 = np.std(projections0)
    spread1 = np.std(projections1)
    return separation, spread0, spread1
    
def datasetSignalNoiseVsTime(dataset):
    signalNoiseRatios=[]
    #Get data as REP x CHANNEL x TIME
    data = datasetTimeChannels2SignalIq_cumulative(dataset)
    #for each time slice, analyze the signal/noise ratio
    for t in range(len(data[0,0,:])):
        z0 = data[:,0,t]
        z1 = data[:,1,t]
        projections = projectLollipops(z0, z1)
        sep, spread0, spread1 = iqSpread(projections[0], projections[1])
        signalNoiseRatios.append((sep/np.sqrt(spread0**2 + spread1**2))**2)
    return np.array(signalNoiseRatios)
    
### UNSORTED ###
    
def iqVsTime2SignalIdlerIq(t, z, fSignal, fCarrier, fPump, idlerAngle=None, plot=False):
    """
    fSignal: frequency of signal you put into the system
    fCarrier: frequency of RF carrier, ie. the A/D board's reference
    fPump: effective paramp pump frequency. If you 2\omega pump, this is
           \omega
    
    All frequencies are assumed to be in units compatible with the units
    of t
    """
    #Mix so that the pump frequency is at DC
    z = mix(t, z, fPump-fCarrier)
    if plot:
        plotFt(t, z)
    #We are now in the frame of the pump. The signal and idler lie symmetrically about zero
    fSideband = fSignal - fPump
    signalPhasorFunc = phasorFunc(fSideband)
    idlerPhasorFunc = phasorFunc(-fSideband)
    zSignal = z*signalPhasorFunc(t)
    if idlerAngle:
        zIdler = z*idlerPhasorFunc(t)
        zIdler = zIdler*np.exp(1.0j*idlerAngle)
        return np.sum(zSignal+zIdler)
    else:
        return np.sum(zSignal)
    
def timeTraces2SignalIdlerIqs(dataset, idlerAngle=None):
    pumpHarmonic = dataset.parameters['readout0.pumpHarmonic']
    fPump = dataset.parameters['readout0.pumpFrequency']/pumpHarmonic
    fSignal = dataset.parameters['q4.readoutFrequency']
    fCarrier = dataset.parameters['readout0.carrierFrequency']
    data = parse_iq_multiple(dataset)
    demoddedNoPulse = []
    demoddedPulse = []
    for dat in data:
        t, INoPulse, QNoPulse, IPulse, QPulse = dat.T
        zDemodNoPulse = iqVsTime2SignalIdlerIq(t, INoPulse+1.0j*QNoPulse, fSignal['GHz'],
                                               fCarrier['GHz'], fPump['GHz'],
                                               idlerAngle)
        zDemodPulse = iqVsTime2SignalIdlerIq(t, IPulse+1.0j*QPulse, fSignal['GHz'],
                                             fCarrier['GHz'], fPump['GHz'],
                                             idlerAngle)
        demoddedNoPulse.append(zDemodNoPulse)
        demoddedPulse.append(zDemodPulse)
    return np.array(demoddedNoPulse), np.array(demoddedPulse)
    
def thingy(dataset, angles, axNoPulse, axPulse=None):
    for angle in angles:
        noPulse, pulse = timeTraces2SignalIdlerIqs(dataset, angle)
        #Show pulse and no pulse in same color
        #dat = np.hstack((pulse,noPulse))
        axNoPulse.plot(noPulse.real/2,noPulse.imag/2)
        #axNoPulse.plot(noPulse.real,noPulse.imag)
        #axNoPulse.text(np.mean(noPulse.real), np.mean(noPulse.imag), str(angle/(2*np.pi)))
        if axPulse:
            axPulse.plot(pulse.real, pulse.imag)
        #ax.text(np.mean(result.real), np.mean(result.imag), str(angle/(2*np.pi)))
    
def projectPhaseCoherent(dataset, idlerAngles):
    """Projected phase sensitive IQs. One set for each idlerAngle"""
    results = []
    for angle in idlerAngles:
        zs, _ = timeTraces2SignalIdlerIqs(dataset, angle)
        rotationAngle = np.angle(np.mean(zs))
        zs = zs*np.exp(-1.0j*rotationAngle)
        results.append(zs.real)
    return results
    
def projectLollipops(zs0, zs1):
    raise Exception('Depricated: use singleQubit.readoutAxis instead')
    """project two clusters of IQ points onto their bisection axis"""
    zs = [zs0, zs1]
    angles = np.array([np.mean(np.angle(z)) for z in zs])
    rotator = np.exp(-1.0j*np.sum(angles)/2)
    #Rotate both clusters of points so that they are symmetric about the real axis
    zs = [z*rotator for z in zs]
    projections = [z.imag for z in zs]
    return projections
    
def signalNoiseRatiosIncoherent(datasets):
    #Demodulate each dataset
    zs = [timeTraces2SignalIdlerIqs(dataset)[0] for dataset in datasets]
    separations = []
    noises = []
    for i in range(len(datasets)-1):
        z0 = zs[i]
        z1 = zs[i+1]
        projections = projectLollipops(z0,z1)
        separations.append(np.mean(projections[0]) - np.mean(projections[1]))
        noises.append(np.sqrt(np.var(projections[0]) + np.var(projections[1])))
    return np.vstack((separations,noises)).T
    
### PLOTS ###
    
def plotLollipop(Sample, measure, readoutIqNum, lollipopNum):
    """Show (corrected) averaged IQ circle with single shot Fresnel lollipops superimposed"""
    s, qubits, = util.loadQubits(Sample)
    q = qubits[measure]
    dvw = datavault.DataVaultWrapper(Sample)
    data = dvw[readoutIqNum]
    vAveraged = data[:,1]+1.0j*data[:,2]
    sAveraged = readout.correctEverything(q, data[:,0], vAveraged)
    
    data = dvw[lollipopNum]
    vNoPulse = data[:,1]+1.0j*data[:,2]
    sNoPulse = readout.correctEverything(q, data[:,0], vNoPulse)
    vPulse = data[:,3]+1.0j*data[:,4]
    sPulse = readout.correctEverything(q, data[:,0], vPulse)
    
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot(sAveraged.real, sAveraged.imag, 'k.', markersize=20)
    ax.plot(sNoPulse.real, sNoPulse.imag, 'b.', markersize=20, label='$|0\\rangle$')
    ax.plot(sPulse.real, sPulse.imag, 'r.', markersize=20, label='$|1\\rangle$')
    ax.legend(loc='lower right', numpoints=1, markerscale=2, prop={'size':30})
    
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot(vAveraged.real, vAveraged.imag, 'k.', markersize=20)
    ax.plot(vNoPulse.real, vNoPulse.imag, 'b.', markersize=20, label='$|0\\rangle$')
    ax.plot(vPulse.real, vPulse.imag, 'r.', markersize=20, label='$|1\\rangle$')
    ax.legend(loc='lower right', numpoints=1, markerscale=2, prop={'size':30})
    
    xaxis = ax.xaxis
    yaxis = ax.yaxis
    for line in xaxis.get_ticklines():
        line.set_markeredgewidth(5)
        line.set_markersize(8)
    for line in yaxis.get_ticklines():
        line.set_markeredgewidth(5)
        line.set_markersize(8)
    for label in xaxis.get_ticklabels():
        label.set_fontsize(40)
    for label in yaxis.get_ticklabels():
        label.set_fontsize(40)
    
def plotIqTracePhase(data, start, stop):
    """Plot phase of a |0> and |1> IQ trace
    Note that data is not corrected here so you're looking at a phase
    computed from raw IQ values
    """
    t, INoPulse, QNoPulse, IPulse, QPulse = data.T
    zNoPulse = INoPulse + 1.0j*QNoPulse
    zPulse = IPulse + 1.0j*QPulse
    timeMask = np.logical_and(t<stop,t>start)
    t = t[timeMask]
    zNoPulse = zNoPulse[timeMask]
    zPulse = zPulse[timeMask]
    phaseNoPulse = np.unwrap(np.angle(zNoPulse))
    phasePulse = np.unwrap(np.angle(zPulse))
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot(t, zNoPulse, 'b-')
    ax.plot(t, zPulse, 'r-')
    
def plotIqTraces3D(data, start, stop, signalSb, filterBw=None):
    """Plot many IQ traces vs time in 3D"""
    raise Exception('Use better parsing code, etc')
    timeMask = np.logical_and(data[:,1]<stop,data[:,1]>start)
    data = data[timeMask,:]
    repetitions=set()
    
    fig = plt.figure()
    ax = Axes3D(fig)
    ax.set_title("Warp Drive")
    for repetition in data[:,0]:
        repetitions.add(repetition)
    
    for repetition in repetitions:
        if repetition>20:
            continue
        mask = data[:,0]==repetition
        dataAtIteration = data[mask,1:]
        t = dataAtIteration[:,0]
        zNoPulse = dataAtIteration[:,1]+1.0j*dataAtIteration[:,2]
        zPulse = dataAtIteration[:,3]+1.0j*dataAtIteration[:,4]
        if filterBw:
            phasor = np.exp(2*np.pi*1.0j*t*signalSb['GHz'])
            zNoPulse = zNoPulse * phasor
            zPulse = zPulse * phasor
            lpfilter = spsignal.firwin(512, (filterBw/(250*MHz))[''])
            zNoPulse = np.convolve(zNoPulse, lpfilter, 'same')
            zPulse = np.convolve(zPulse, lpfilter, 'same')
        ax.plot(zNoPulse.real, zNoPulse.imag, t, 'b-')
        ax.plot(zPulse.real, zNoPulse.imag, t, 'r-')
    
def plotIqTracesPhase(data, start, stop, presentable=False):
    """Plot the phase of many IQ traces"""
    timeMask = np.logical_and(data[:,1]<stop,data[:,1]>start)
    data = data[timeMask,:]
    repetitions=set()
    
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    
    for repetition in data[:,0]:
        repetitions.add(repetition)
    
    for repetition in repetitions:
        mask = data[:,0]==repetition
        dataAtIteration = data[mask,1:]
        t_us = dataAtIteration[:,0]/1000.0
        phaseNoPulse = np.unwrap(np.angle(dataAtIteration[:,1]+1.0j*dataAtIteration[:,2]))
        phasePulse = np.unwrap(np.angle(dataAtIteration[:,3]+1.0j*dataAtIteration[:,4]))
        labels = ['$|0\\rangle$','$|1\\rangle$'] if repetition==0 else ['','']
        ax.plot(t_us, phaseNoPulse, 'b-', linewidth=3, label=labels[0])
        ax.plot(t_us, phasePulse, 'r-', linewidth=3, label=labels[1])
    ax.legend(loc='lower left', numpoints=1, markerscale=2, prop={'size':30})
    
    if presentable:
        xaxis = ax.xaxis
        yaxis = ax.yaxis
        for line in xaxis.get_ticklines():
            line.set_markeredgewidth(5)
            line.set_markersize(8)
        for line in yaxis.get_ticklines():
            line.set_markeredgewidth(5)
            line.set_markersize(8)
        for label in xaxis.get_ticklabels():
            label.set_fontsize(40)
        for label in yaxis.get_ticklabels():
            label.set_fontsize(40)
    
### TWO QUBIT STUFF ###

def lollipops2Qubit(data, tMin, tMax):
    """This function is specifically set up to plot [Daniel,Transmon,Transmon6,130221],41"""
    repetitions = set()
    states = set()
    for repetition in data[:,0]:
        repetitions.add(repetition)
    print('Repetitions: %s'%repetitions)
    for state in data[:,2]:
        states.add(state)
    figs = []
    for state in states:
        f = plt.figure(figsize=(20,8))
        ax = f.add_subplot(1,2,1)
        ax.set_autoscaley_on(False)
        ax.set_autoscalex_on(False)
        ax.set_xlim([-120000,20000])
        ax.set_ylim([-50000,90000])
        ax = f.add_subplot(1,2,2)
        ax.set_autoscaley_on(False)
        ax.set_autoscalex_on(False)
        ax.set_xlim([-50000,90000])
        ax.set_ylim([-100000,40000])
        figs.append(f)
    print figs
    for i,(state,color) in enumerate(zip(states,['b','r','g','k'])):
        stateMask = data[:,2]==state
        dataState = data[stateMask,:]
        dataState = dataState[:,[0,1,3,4,5,6]]
        for repetition in repetitions:
            dataRep = dataState[dataState[:,0]==repetition,1:]
            timeMask = np.logical_and(dataRep[:,0]>tMin, dataRep[:,0]<tMax)
            dataTime = dataRep[timeMask, 1:]
            z0 = np.sum(dataTime[:,0] + 1.0j*dataTime[:,1])
            z1 = np.sum(dataTime[:,2] + 1.0j*dataTime[:,3])
            for j,fig in enumerate(figs[i:]):
                if j==0:
                    alpha = 1
                else:
                    alpha = 0.2
                ax0, ax1 = fig.axes
                ax0.plot(z0.real, z0.imag, color+'.', markersize=20, alpha=alpha)
                ax1.plot(z1.real, z1.imag, color+'.', markersize=20, alpha=alpha)
    for fig in figs:
        for ax in fig.axes:
            for xyaxes in [ax.xaxis, ax.yaxis]:
                n = len(xyaxes.get_ticklabels())
              #  xyaxes.set_ticklabels(['']*n)
    plt.show()
    return figs
            
### WARNING, THIS IS A TOTAL HACK ###

def plotTwoQubitTimeTrace(data, start, stop, filterBw=None):
    """IQ signals must be derotated before using this function"""
    figQ0 = plt.figure()
    figQ1 = plt.figure()
    ax0 = Axes3D(figQ0)
    ax1 = Axes3D(figQ1)
    repetitions = set()
    states=set()
    for rep in data[:,0]:
        repetitions.add(rep)
    for state in data[:,2]:
        states.add(state)
    
    for repetition in repetitions:
        mask = data[:,0]==repetition
        dataRep = data[mask,1:]
        for state,color in zip(states,['b-','r-','g-','k-']):
#            if state not in [0.0, 10.0]:
#                continue
            mask = dataRep[:,1]==state
            dataState = np.vstack((dataRep[mask,0],dataRep[mask,2:].T)).T
            mask = np.logical_and(dataState[:,0]>start,dataState[:,0]<stop)
            dataTime = dataState[mask,:]
            t = dataTime[:,0]
            z0 = dataTime[:,1] + 1.0j*dataTime[:,2]
            z1 = dataTime[:,3] + 1.0j*dataTime[:,4]
            if filterBw:
                lpfilter = spsignal.firwin(512, (filterBw/(250*MHz))[''])
                z0 = np.convolve(z0, lpfilter, 'same')
                z1 = np.convolve(z1, lpfilter, 'same')
            plt.figure(figQ0.number)
            ax0.plot(z0.real, z0.imag, t, color)
            plt.figure(figQ1.number)
            ax1.plot(z1.real, z1.imag, t, color)
