import numpy as np

import readout
from readout import nonCompactPhase_rad, getLastDataset
import pyle.dataking.singleQubit as sq
#import scans
#from scans import readoutIq

import labrad
from labrad.units import Value,Unit
MHz,GHz,ns = (Unit(s) for s in ['MHz','GHz','ns'])

from pyle.plotting import dstools
from pyle.dataking import util
from pyle.util.structures import ValueArray

import matplotlib.pyplot as plt

import pyle.dataking.adjust as adjust


def backgroundBoth(sample, measure, name='Cable calibration',
                   df=None, freq=None, power = None, readoutLen = None,
                   stats=600, numPts=300,
                   cutOutResonator=True,
                   plot=True):
    """Measure background transmission magnitude and phase
    
    DOCSTRING OUT OF DATE. WE USED TO KEEP THE IF FREQUENCY FIXED,
    BUT NOT ANY MORE SINCE YOU CAN'T DO THAT WITH MULTIPLE QUBITS
    
    Run this scan near the readout resonator, but do not include the dip in
    your frequency axis.
    Paramp should be biased far away.
    
    REGISTRY KEYS UPDATED
    qubit['calReadoutRfParams']: ((p[0],p[1]), sbFreq)
        p is the polynomial fit to the cable delay in radians vs GHz, as
        the RF frequency is varied, keeping the IF frequency fixed.
    qubit['calReadoutCableDelay']: Value(ns)
        A convenience key giving the time of flight through the RF
        part of the circuit, equal to p[0]/2pi * 1ns
    qubit['calReadoutMagnitudeParams']: (pMag, power, qubit['adcFilterWindow'][1]-qubit['adcFilterWindow'][0])
        pMag is a polynomial fit to the transmitted I^2 + Q^2.
    """
    sampleReg, qubits, Qubits = util.loadQubits(sample, write_access=True)
    qubit = qubits[measure]
    Qubit = Qubits[measure]
    if power is None:
        power = qubit['readoutPower']
    if df is None:
        df = 3.0*MHz
    if freq is None:
        freqStart_GHz = (qubit['readoutFrequency']-df)['GHz']
        freqEnd_GHz = (qubit['readoutFrequency']+df)['GHz']
        freq = ValueArray(np.linspace(freqStart_GHz,freqEnd_GHz, numPts),'GHz')
    #Measure IQ data
    data = sq.readoutIq(sample, measure, name=name,
                        freq=freq, power=power, readoutLen = readoutLen,
                        stats=stats, dataFormat='iq',
                        save=True, collect=True,
                        plot=False, regParams=(sampleReg, qubits))
    dataset = getLastDataset(sample)
    
    phaseParams, delayTime, magParams = getCableParameters(dataset, cutOutResonator, plot)
    
    sbFreq = 10*GHz # We need to set something to keep the type tag right,
                    # but you shouldn't use this.
    #Store to registry
    Qubit['calReadoutRfParams'] = (phaseParams, sbFreq)
    Qubit['calReadoutCableDelay'] = delayTime
    Qubit['calReadoutMagnitudeParams'] = (magParams, power, qubit['adcFilterWindow'][1]-qubit['adcFilterWindow'][0])
    
def adjustMagnitude(freqs_GHz, mags):
    traces = [{'x': freqs_GHz, 'y': mags, 'args': ('b.',)}]
    params = [{'name': 'Left side start', 'val': freqs_GHz[0], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'b'},
              {'name': 'Left side end', 'val': freqs_GHz[0], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'r'},
              {'name': 'Right side start', 'val': freqs_GHz[-1], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'g'},
              {'name': 'Right side end', 'val': freqs_GHz[-1], 'range':(freqs_GHz[0], freqs_GHz[-1]), 'axis':'x', 'color': 'k'}]
    result = adjust.adjust(params, traces)
    return result
    
def getCableParameters(dataset, cutOutResonator=True, plot=False, unwrapPhase=True):
    """
    dataset columns 1 and 2 must be real and imaginary parts of S21 in linear units
    """
    #Fetch data from datavault and parse
    freqUnit = dataset.variables[0][0][1]
    frequencies = ValueArray(dataset.data[:,0],freqUnit)
    I, Q = dataset.data[:,1],dataset.data[:,2]
    mags = np.sqrt(I**2 + Q**2)
    if unwrapPhase:
        phases_rad = nonCompactPhase_rad(I,Q)
    else:
        phases_rad = np.angle(I+1j*Q)
    figMag = plt.figure()
    plt.plot(frequencies['GHz'],mags,'.')
    figPhase = plt.figure()
    plt.plot(frequencies['GHz'],phases_rad,'.')
    #Remove data at resonator
    if cutOutResonator:
        results = adjustMagnitude(frequencies['GHz'], mags)
        calStart = Value(results['Left side start'], 'GHz')
        calStop = Value(results['Right side end'], 'GHz')
        dipStart = Value(results['Left side end'])
        dipStop = Value(results['Right side start'])
        maskMags = np.logical_or(
                                 np.logical_and(frequencies['GHz']>calStart['GHz'],frequencies['GHz']<dipStart['GHz']),
                                 np.logical_and(frequencies['GHz']>dipStop['GHz'],frequencies['GHz']<calStop['GHz'])
                                )
        maskPhase1 = frequencies['GHz']<dipStart
        maskPhase2 = frequencies['GHz']>dipStop
    else:
        maskMags = np.array([True]*len(frequencies))
        maskPhase1 = np.array([True]*len(frequencies))
        maskPhase2 = np.array([True]*len(frequencies))
    #Plot mag data being fit
    if plot:
        plt.figure(figMag.number)
        plt.plot(frequencies['GHz'][maskMags], mags[maskMags],'r.')
        plt.figure(figPhase.number)
    #Fit magnitude vs frequency in GHz
    pMag = np.polyfit(frequencies['GHz'][maskMags], mags[maskMags], 1)
    #Fit phase vs frequency in GHz
    #Plot frequency data being fit
    if plot:
        plt.figure(figPhase.number)
        plt.plot(frequencies['GHz'][maskPhase1],phases_rad[maskPhase1],'r.')
        plt.plot(frequencies['GHz'][maskPhase2],phases_rad[maskPhase2],'r.')
    minFreq = frequencies[0]
    shiftedFrequencies = frequencies - minFreq
    pPhase1 = np.polyfit(shiftedFrequencies['GHz'][maskPhase1], phases_rad[maskPhase1], 1)
    pPhase2 = np.polyfit(shiftedFrequencies['GHz'][maskPhase2], phases_rad[maskPhase2], 1)
    phaseDiff = pPhase1[1]-pPhase2[1] #Phase difference before and after resonance
    phaseDiff2Pi = round(phaseDiff/(2*np.pi))*(2*np.pi)
    phasesToFit_rad = np.hstack((phases_rad[maskPhase1],phases_rad[maskPhase2]+phaseDiff2Pi))
    freqsToFit = np.hstack((frequencies['GHz'][maskPhase1],frequencies['GHz'][maskPhase2]))
    pPhase = np.polyfit(freqsToFit, phasesToFit_rad, 1)
    delayTime = Value(pPhase[0]/(2*np.pi), 'ns')
    if plot:
        plotFreqs = np.linspace(frequencies['GHz'][0],frequencies['GHz'][-1],200)
        plt.figure(figMag.number)
        plt.plot(plotFreqs,np.polyval(pMag, plotFreqs),'r-')
        plt.xlabel('Frequency [GHz]')
        plt.ylabel('Transmission magnitude')
        plt.figure(figPhase.number)
        plt.plot(plotFreqs, np.polyval(pPhase, plotFreqs), 'r-')
        plt.xlabel('Frequency [GHz]')
        plt.ylabel('Transmitted phase [rad]')
    return (tuple(pPhase), delayTime, tuple(pMag))
    
def parampResponse(sample, measure,
                   freq=None, df=5*MHz,
                   stats=150,
                   plot=False):
    sampleReg, qubits, Qubits = util.loadQubits(sample, write_access=True)
    qubit = qubits[measure]
    Qubit = Qubits[measure]
    if freq is None:
        freqStart_GHz = (qubit['readoutFrequency']-df)['GHz']
        freqEnd_GHz = (qubit['readoutFrequency']+df)['GHz']
        freq = ValueArray(np.linspace(freqStart_GHz,freqEnd_GHz,80),'GHz')
    sq.readoutIq(sample, freq=freq, measure=measure, stats=1200, dataFormat='iq',
                 save=True, plot=False, name='Paramp calibration', regParams=(sampleReg,qubits))
    with labrad.connect() as cxn:
        dv = cxn.data_vault
        dataset = dstools.getDeviceDataset(dv, -1, session=sample._dir)
    gainParams, phaseParams = readout.getParampParameters_polynomial(dataset)
    #Save parameters to registry
    Qubit['calReadoutParampGainFunc'] = gainParams
    Qubit['calReadoutParampPhaseFunc'] = phaseParams
    if plot:
        v = dataset['data'][:,1]+1.0j*dataset['data'][:,2]
        S = readout.correctEverything(Qubit, dataset['data'][:,0], v)
        plt.figure()
        plt.plot(S.real,S.imag)
        plt.title('Corrected readout circle')
        plt.xlabel('I')
        plt.ylabel('Q')