#!/usr/bin/python
from __future__ import division

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from pyle import datavault
from pyle.dataking import util
import scipy
import scipy.signal

def phasor(freq, t):
    return np.exp(2j*np.pi*freq*t)

def simple_dispersion_estimate(data, fraction_to_cover=0.5):
    '''
    Calculate the dispersion of a sample set in the presence of outliers. 
    This is useful for estimating the separation of IQ points for qubit
    readout when there can be state decay, hot qubits, or imperfect pi-pulses.
    
    The way this works is to find the smallest interval or disk that contains 50%
    of the points.
    '''
    cover_pts = int(len(data)* fraction_to_cover)
    sorted_points = np.sort(data)
    best_range = sorted_points[-1]-sorted_points[0]
    best_idx = 0
    for idx in range(len(sorted_points)-cover_pts):
        size = sorted_points[cover_pts+idx] - sorted_points[idx]
        if size < best_range:
            best_range=size
            best_idx = idx
    middle_point = np.mean(sorted_points[best_idx:best_idx+cover_pts])
    return best_range

def interquartile_range(data):
    '''
    returns the interquartile range 
    '''
    sorted_data = np.sort(data)
    idx1 = len(sorted_data)//4
    idx2 = len(sorted_data)*3//4
    iqr = sorted_data[idx2] - sorted_data[idx1]
    return iqr

def interquartile_mean(data):
    sorted_data = np.sort(data)
    idx1 = len(sorted_data)//4
    idx2 = len(sorted_data)*3//4
    iqm = mean(sorted_data[idx1:idx2])
    return iqm
    
def analyze_separation(z_nopulse, z_pulse):
    axis = np.mean(z_pulse) - np.mean(z_nopulse)
    axis = axis / abs(axis)
    
    x_nopulse = np.real(z_nopulse * axis.conj())
    x_pulse = np.real(z_pulse * axis.conj())
    separation = np.median(x_nopulse) - np.median(x_pulse)
    nopulse_spread = interquartile_range(x_nopulse)
    pulse_spread = interquartile_range(x_pulse)
    #separation = np.mean(x_nopulse) - np.mean(x_pulse)
    #nopulse_spread = np.std(x_nopulse)
    #pulse_spread = np.std(x_pulse)
    return separation, nopulse_spread, pulse_spread
    
def cheater_stats(data):
    '''
    This does a cheating statistical analysis.  Calculates the mean and standard deviation,
    rejects points more than 3 sigma from the mean, then recalculates:
    
    (correted_mean, corrected_std, outliers) = cheater_stats(data)
    '''
    mean = np.mean(data)
    std = np.std(data)
    outliers = np.abs(data-mean) > (3*std)
    corrected_std = np.std(data[np.logical_not(outliers)])
    corrected_mean = np.mean(data[np.logical_not(outliers)])
    return corrected_mean, corrected_std, outliers
    
def parse_iq_multiple(data, start=0.0, stop=np.inf):
    '''
    Parses data from IQ vs time traces with multiple repetitions.
    The input should be a dataset wrapper, the output is a 3-D array
    that should be indexed by x[repetition, time_step, col].  
    '''
    nreps = int(data['Repetition'][-1]+1)
    t = data['Time']
    data_by_rep = []
    time_mask = np.logical_and(data['Time'] >= start, data['Time'] < stop)
    for x in range(nreps):
        mask = np.logical_and(data['Repetition']==x, time_mask)
        idx = np.flatnonzero(mask)
        mask = slice(idx[0], idx[-1]+1)
        rep_data = data[mask, 1:]
        data_by_rep.append(rep_data)
    return np.array(data_by_rep)

def incoherent_demod(ds, channel=0, start=0.0, stop=np.inf, qubit_num=0, sideband_freq=None):
    '''
    Demodulate and sum a signal at a single frequency and sum.  By default,
    the demodulation frequency will be computed from the dataset parameters.
    '''
    config = ds.parameters['config']
    qubitName = config[ds.parameters['measure'][qubit_num]]
    readoutName = ds.parameters[qubitName]['readoutConfig']
    data = parse_iq_multiple(ds, start, stop)
    t = data[0,:,0]
    if sideband_freq is None:
        sideband_freq = ds.parameters[readoutName]['carrierFrequency'] - ds.parameters[qubitName]['readoutFrequency']
    demod_phasor = phasor(-sideband_freq['GHz'], t)
    z_demod = []
    for rep in range(len(data)):
        z = data[rep,:,1+2*channel] + 1j*data[rep,:,2+2*channel]
        z_demod.append(np.sum(z*demod_phasor))
    z_demod = np.array(z_demod)
    return z_demod

def mix_to_dc(ds, channel, qubit_num=0, sideband_freq=None):
    config = ds.parameters['config']
    qubitName = config[ds.parameters['measure'][qubit_num]]
    readoutName = ds.parameters[qubitName]['readoutConfig']
    data = parse_iq_multiple(ds)#, start, stop)
    t = data[0,:,0]
    if sideband_freq is None:
        sideband_freq = ds.parameters[readoutName]['carrierFrequency'] - ds.parameters[qubitName]['readoutFrequency']
    demod_phasor = phasor(-sideband_freq['GHz'], t)
    z_demod = []
    for rep in range(len(data)):
        z = data[rep,:,1+2*channel] + 1j*data[rep,:,2+2*channel]
        z_demod.append(z*demod_phasor)
    z_demod = np.array(z_demod)
    return z_demod
    
def coherent_demod(ds, readout_phase, channel=0, start=0.0, stop=np.inf):
    data = parse_iq_multiple(ds, start, stop)
    t = data[0,:,0]
    sideband_freq = ds.parameters['readout0.carrierFrequency'] - ds.parameters['readout0.pumpFrequency']/2
    demod_phasor = phasor(-sideband_freq['GHz'], t)
    readout_sideband_freq = ds.parameters['readout0.pumpFrequency']/2 - ds.parameters['q4.readoutFrequency']
    readout_phasor = phasor(-readout_sideband_freq['GHz'], t)
    rv = []
    for x in range(len(data)):
        z = data[x,:,1+2*channel] + 1j*data[x,:,2+2*channel]
        z_demod = 2*np.real(z * demod_phasor * np.exp(1j*readout_phase))
        rv.append(np.sum(z_demod * readout_phasor))
    return np.array(rv)

def idler_demod(ds, channel=0, start=0.0, stop=np.inf):
    '''
    Demodulate and sum the idler frequency of the given data set.
    '''
    sideband_freq = (ds.parameters['readout0.carrierFrequency'] - 
        (ds.parameters['readout0.pumpFrequency'] - 
            ds.parameters['q4.readoutFrequency']))
    return incoherent_demod(ds, channel, start, stop, sideband_freq=sideband_freq)
    
def incoherent_separation_plot(ds):
    '''
    Plot pulse/no-pulse separation clouds
    '''
    data = parse_iq_multiple(ds)
    t = data[0,:,0]
    sideband_freq = ds.parameters['readout0.carrierFrequency'] - ds.parameters['q4.readoutFrequency']
    demod_phasor = phasor(-sideband_freq['GHz'], t)
    z_pulse = []
    z_nopulse = []
    for rep in range(len(data)):
        z_nopulse.append(np.sum((data[rep,:,1] + 1j*data[rep,:,2])*demod_phasor))
        z_pulse.append(np.sum((data[rep,:,3] + 1j*data[rep,:,4])*demod_phasor))
    z_pulse = np.array(z_pulse)
    z_nopulse = np.array(z_nopulse)
    plt.plot(z_nopulse.real, z_nopulse.imag, label="nopulse")
    plt.plot(z_pulse.real, z_pulse.imag, label="pulse")
    plt.legend(numpoints=1)
    return z_pulse, z_nopulse

def signal_idler_separation_plot(ds):
    '''
    Plot pulse/no-pulse separation clouds for both signal and idler/
    '''
    signal_nopulse = incoherent_demod(ds, channel=0)
    signal_pulse = incoherent_demod(ds, channel=1)
    idler_nopulse = idler_demod(ds, channel=0)
    idler_pulse = idler_demod(ds, channel=1)
    plt.figure()
    plt.plot(signal_nopulse.real, signal_nopulse.imag, label="Signal No-pulse")
    plt.plot(signal_pulse.real, signal_pulse.imag, label="Signal w/ Pulse")
    plt.plot(idler_nopulse.real, idler_nopulse.imag, label="Idler No-pulse")
    plt.plot(idler_pulse.real, idler_pulse.imag, label="Idler w/ Pulse")
    
    plt.legend(numpoints=1)
    
def phase_sensitive_readout_plot(ds, start=0, stop=np.inf):
    plt.figure()
    phases = np.arange(16)*np.pi/16
    signal = []
    noise = []
    for phi in phases:
        x = coherent_demod(ds, phi, start=start, stop=stop)
        corrected_mean, corrected_std, outliers = cheater_stats(x)
        print "Rejecting %d outliers from %d points" % (np.sum(outliers), len(outliers))
        noise.append(corrected_std)
        signal.append(corrected_mean)
        plt.plot(x.real, x.imag)
    x = incoherent_demod(ds, start=start, stop=stop)
    plt.plot(x.real, x.imag)
    incoherent_signal, incoherent_noise, outliers = cheater_stats(x)
    plt.figure()
    plt.plot(phases, np.abs(np.array(signal))**2 / np.array(noise)**2)
    incoherent_snr = np.abs(incoherent_signal)**2 / np.abs(incoherent_noise)**2
    print "plotting horizontal line at %f" % incoherent_snr
    plt.axhline(y=incoherent_snr, linewidth=2, linestyle='-', color='r')

def SNR_incoherent_simple(ds1, channel1=0, ds2=None, channel2=None):
    '''
    Calculates a robust signal-to-noise ratio for distinguishing two singals that may have 
    outliers.  ds1 and ds2 are the datasets (default ds2=ds1) and col1 and col2 are the
    channel number in that dataset (i.e., 0 for "no-pulse" and 1 for "pulse")
    '''
    if ds2 is None:
        ds2 = ds1
        if channel2 is None:
            channel2 = channel1
    elif channel2 is None:
        channel2 = 1
    
    z_nopulse = incoherent_demod(ds1, channel1)
    z_pulse = incoherent_demod(ds2, channel2)
    separation, range_nopulse, range_pulse = analyze_separation(z_nopulse, z_pulse)
    SNR = abs(separation) / np.sqrt(range_nopulse**2+range_pulse**2)
    print "SNR: ", SNR
    return SNR

def separation_vs_time(ds):
    '''
    Input is from IqVsTime2D
    '''
    x = parse_iq_multiple(ds)
    config = ds.parameters['config']
    measure = ds.parameters['measure'][0]
    q = ds.parameters[config[measure]]
    readout = ds.parameters[q['readoutConfig']]
    sb_freq = q['readoutFrequency'] - readout['carrierFrequency']
    print "sideband frequency: %s" % sb_freq
    plt.figure()
    separation = 0
    for idx, rep in enumerate(x):
        t = rep[:,0]
        phasor = np.exp(1j*2*np.pi*t*sb_freq['GHz'])
        z_no = np.convolve(phasor*(rep[:,1]+1j*rep[:,2]), np.kaiser(5,3), mode='same')
        z_pulse = np.convolve(phasor*(rep[:,3]+1j*rep[:,4]), np.kaiser(5,3), mode='same')
        separation = separation + (z_no-z_pulse)
        plt.plot(t, np.abs(z_no-z_pulse).real, label=("Separation rep %d" % idx))
        
    plt.legend()
    plt.figure()
    plt.plot(t, np.abs(separation).real/np.max(np.abs(separation).real))

def blah(ds):
    '''
    Input is from IqVsTime2D
    ''' 
    x = parse_iq_multiple(ds)
    config = ds.parameters['config']
    measure = ds.parameters['measure'][0]
    q = ds.parameters[config[measure]]
    readout = ds.parameters[q['readoutConfig']]
    sb_freq = q['readoutFrequency'] - readout['carrierFrequency']
    print "sideband frequency: %s" % sb_freq
    plt.figure()
    for idx, rep in enumerate(x):
        t = rep[:,0]
        phasor = np.exp(1j*2*np.pi*t*sb_freq['GHz'])
        z_no = np.convolve(phasor*(rep[:,1]+1j*rep[:,2]), np.kaiser(50,5), mode='same')
        plt.plot(t, z_no.real, label=("real part rep %d" % idx))
    plt.legend()
    
    plt.figure()
    for idx, rep in enumerate(x):
        t = rep[:,0]
        phasor = np.exp(1j*2*np.pi*t*sb_freq['GHz'])
        z_no = np.convolve(phasor*(rep[:,1]+1j*rep[:,2]), np.kaiser(50,5), mode='same')
        plt.plot(t, z_no.imag, label=("imaginary part rep %d" % idx))
    plt.legend()

def optimal_window(ds, Sample=None, update=False, threshold=.2):
    '''
    Return an optimal window function.  ds should be a dataset
    collected with IqVsTime2D.  threshold is the fraction of full
    scale to cut off the window.
    '''
    x = parse_iq_multiple(ds)
    config = ds.parameters['config']
    measure = ds.parameters['measure'][0]
    q = ds.parameters[config[measure]]
    readout = ds.parameters[q['readoutConfig']]
    sb_freq = q['readoutFrequency'] - readout['carrierFrequency']
    window = 0
    for idx, rep in enumerate(x):
        t = rep[:,0]
        phasor = np.exp(1j*2*np.pi*t*sb_freq['GHz'])
        z_no = np.convolve(phasor*(rep[:,1]+1j*rep[:,2]), np.kaiser(5,3), mode='same')
        z_pulse = np.convolve(phasor*(rep[:,3]+1j*rep[:,4]), np.kaiser(5,3), mode='same')
        window = window + np.abs(z_no-z_pulse).real
    window = window[0::2] + window[1::2] # ADC window is in 4 nanosecond chunks
    t = (t[0::2] + t[1::2])/2
    window = window / np.max(window)
    window = window * (window > .2)
    plt.figure()
    plt.plot(t, window)
    f = np.fft.fftfreq(len(t), .004)
    plt.figure()
    plt.plot(f, 20*np.log10(np.abs(np.fft.fft(window))))
    nz = np.nonzero(window)[0]
    start, stop = nz[0], nz[-1]+1
    if update and Sample:
        Q = Sample[config[measure]]
        Q['optimalWindow'] = window[start:stop]
    return window[start:stop], t[start]
    
def SNR_coherent_simple(ds1, channel1=0, ds2=None, channel2 = None, phases=[0]):
    if ds2 is None:
        ds2 = ds1
        if channel2 is None:
            channel2 = channel1
    elif channel2 is None:
        channel2 = 1
    z1_signal = incoherent_demod(ds1, channel1)
    z1_idler = idler_demod(ds1, channel1)
    z2_signal = incoherent_demod(ds2, channel2)
    z2_idler = idler_demod(ds2, channel2)
    SNR = np.zeros_like(phases)
    for idx, phi in enumerate(phases):
        s1 = z1_signal + z1_idler*np.exp(1j*phi)
        s2 = z2_signal + z2_idler*np.exp(1j*phi)
        separation, range1, range2 = analyze_separation(s1, s2)
        SNR[idx] = abs(separation) / np.sqrt(range1**2+range2**2)
    return SNR
    
def SNR_vs_time_phase_insensitive(ds, start, stop, dt):
    '''
    Plots the IQ values integrated from zero to T as a function of T.
    '''
    stop_times = np.arange(start+dt, stop, dt)
    pulse_data = []
    nopulse_data = []
    for st in stop_times:
        z_nopulse = incoherent_demod(ds, 0, start, st)
        z_pulse = incoherent_demod(ds, 1, start, st)
        nopulse_data.append(np.column_stack(((st-start)*np.ones_like(z_nopulse), z_nopulse)))
        pulse_data.append(np.column_stack(((st-start)*np.ones_like(z_pulse), z_pulse)))
    pulse_data = np.vstack((pulse_data))
    nopulse_data = np.vstack((nopulse_data))
    plt.figure()
    t = nopulse_data[:,0]
    plt.plot(t, nopulse_data[:,1].real, label='nopulse real')
    plt.plot(t, pulse_data[:,1].real, label='pulse real')
    plt.legend(numpoints=1)

    plt.figure()
    plt.plot(t, nopulse_data[:,1].imag, label='nopulse imag')
    plt.plot(t, pulse_data[:,1].imag, label='pulse imag')
    plt.legend(numpoints=1)
    
def find_separation(data, start, stop):
    nreps = int(max(data[:,0])+1)
    t = data[:,1]
    t_mask = np.logical_and(t>=start, t<=stop)
    data = data[t_mask,:]
    # Go through all the reps and isolate the given time slice, use that to find the
    # center and axis so we can bin data into |1> and |0> for calculating the SNR
    std_s = []
    pulse_data = []
    nopulse_data = []
    for x in range(nreps):
        mask = data[:,0] == x
        rep_data = data[mask,:]
        t = rep_data[:,1]
        Z_nopulse = rep_data[:,2] + 1j*rep_data[:,3]
        Z_pulse = rep_data[:,4] + 1j*rep_data[:,5]
        std_s.append(np.std(Z_pulse))
        std_s.append(np.std(Z_nopulse))
        nopulse_data.append(np.mean(Z_nopulse))
        pulse_data.append(np.mean(Z_pulse))
    pulse_data = np.array(pulse_data)
    nopulse_data = np.array(nopulse_data)
    center = np.mean(np.hstack((pulse_data, nopulse_data)))
    axis = np.mean(pulse_data) - np.mean(nopulse_data)
    separation = abs(axis)
    orientation = axis/ separation
    return center, separation, orientation

    
def SNR_vs_time(ds, start, stop, label=""):
    dt = np.logspace(1, np.log10(stop-start), 15)
    
    err_data = []
    data = []
    for x in dt:
        print "x: ", x
        data.append(SNR(ds,start,stop,x))
        #err_data.append(error_rate)
    signal_level, one_noise, zero_noise, error_rate = zip(*data)
    signal = np.array(signal_level)
    noise = 0.5*(np.array(one_noise) + np.array(zero_noise) )
    plt.plot(dt, signal/noise)
    #plt.loglog(dt, err_data, label=label)
    return dt, err_data

def SNR(data, start, stop, dt=10):

    center, separation, orientation = find_separation(data, start, stop)
    nreps = int(max(data[:,0])+1)
    time_windows = np.arange(start, stop-dt, dt)
    zero_data = []
    one_data = []

    for rep in range(nreps):
        mask = data[:,0] == rep
        rep_data = data[mask,:]
        ZNoPulse = rep_data[:,2]+1j*rep_data[:,3]
        ZPulse = rep_data[:,4]+1j*rep_data[:,5]
        filt = scipy.signal.firwin(512, 3.0/250.0)

        ZNoPulse_filt = np.convolve(ZNoPulse, filt, mode="same")
        ZPulse_filt = np.convolve(ZPulse, filt, mode="same")
        ZNoPulse_state = np.real((ZNoPulse_filt-center)/orientation)>0
        ZPulse_state = np.real((ZPulse_filt-center)/orientation)>0

        #plt.figure()
        #t = rep_data[:,1]
        #plt.plot(t,np.angle(ZNoPulse_filt))
        #plt.plot(t,np.angle(ZPulse_filt))
        #plt.plot(t,ZNoPulse_state)
        #plt.plot(t,ZPulse_state)

        for t_start in time_windows:
            t = rep_data[:,1]
            start_idx = np.flatnonzero(t>t_start)[0]
            stop_idx = np.flatnonzero(t<t_start+dt)[-1]
            mask = slice(start_idx, stop_idx)
            state = np.mean(ZPulse_state[mask])>0
            measurement = (np.mean(ZPulse[mask])-center)/orientation
            if state:
                one_data.append(measurement)
            else:
                zero_data.append(measurement)

            state = np.mean(ZNoPulse_state[mask])>0
            measurement = (np.mean(ZNoPulse[mask])-center)/orientation
            if state:
                one_data.append(measurement)
            else:
                zero_data.append(measurement)
                
    zero_data = np.array(zero_data)
    one_data = np.array(one_data)
    zero_noise = np.std(zero_data)
    one_noise = np.std(one_data)
    signal_level = abs(np.mean(one_data) - np.mean(zero_data))/2
    error_rate = (1-scipy.special.erf(signal_level / max(zero_noise, one_noise)))/2
    return signal_level, one_noise, zero_noise, error_rate

def plotIqTracePhase(data, start, stop):
    """Plot phase of a |0> and |1> IQ trace
    Note that data is not corrected here so you're looking at a phase
    computed from raw IQ values
    """
    t, INoPulse, QNoPulse, IPulse, QPulse = data.T
    zNoPulse = INoPulse + 1.0j*QNoPulse
    zPulse = IPulse + 1.0j*QPulse
    timeMask = np.logical_and(t<stop,t>start)
    t = t[timeMask]
    zNoPulse = zNoPulse[timeMask]
    zPulse = zPulse[timeMask]
    phaseNoPulse = np.unwrap(np.angle(zNoPulse))
    phasePulse = np.unwrap(np.angle(zPulse))
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1)
    ax.plot(t, zNoPulse, 'b-')
    ax.plot(t, zPulse, 'r-')

def readout_histogram(ds, start=1000, stop=2000):
    '''
    Select a window on each repetition and plot a histogram
    showing the separation
    '''
    plt.figure()
    nreps = int(max(ds[:,0])+1)
    t = ds[:,1]
    t_mask = np.logical_and(t>=start, t<=stop)
    ds = ds[t_mask,:]
    pulse_data = []
    nopulse_data = []
    std_s = []
    for x in range(nreps):
        mask = ds[:,0] == x
        rep_data = ds[mask,:]
        t = rep_data[:,1]
        Z_nopulse = rep_data[:,2] + 1j*rep_data[:,3]
        Z_pulse = rep_data[:,4] + 1j*rep_data[:,5]
        std_s.append(np.std(Z_pulse))
        std_s.append(np.std(Z_nopulse))
        nopulse_data.append(np.mean(Z_nopulse))
        pulse_data.append(np.mean(Z_pulse))
    pulse_data = np.array(pulse_data)
    nopulse_data = np.array(nopulse_data)
    center = np.mean(np.hstack((pulse_data, nopulse_data)))
    axis = np.mean(pulse_data) - np.mean(nopulse_data)
    separation = abs(axis)
    total_std = np.sqrt(np.sum(np.array(std_s)**2) / len(std_s))
    print "separation: %f, std: %f" % (separation, total_std)
    axis = axis / abs(axis)
    pulse_readout_value = np.real((pulse_data - center)/axis)
    nopulse_readout_value = np.real((nopulse_data - center)/axis)
    plt.hist([nopulse_readout_value, pulse_readout_value], 40, label=["No Pi Pulse", "Pi Pulse"], histtype='barstacked')
    #plt.hist(pulse_readout_value, 30, label="Pi Pulse")
    plt.legend()
    plt.title("Single shot readout histogram: %d ns averaging" % (stop-start))
    plt.xlabel("Software Demodulator average (ADC counts)")
    
def plotPowerT1(datasets, t0, t1):
    plt.figure()
    plt.title("Population remaining after 4 microseconds")
    plt.xlabel("Readout power")
    plt.ylabel("Population")
    
    for data  in datasets:
        t = data[:,2]
        t0 = np.min(t[t>t0])
        t1 = np.min(t[t>t1])
        data0=data[t==t0,:]
        data1=data[t==t1,:]
        z0_nopulse = data0[:,3]+1j*data0[:,4]
        z0_pulse = data0[:,5]+1j*data0[:,6]
        z1_nopulse = data1[:,3]+1j*data1[:,4]
        z1_pulse = data1[:,5]+1j*data1[:,6]
        power_dBm = data0[:,0]
        #normalized_separation = abs((z_pulse-z_nopulse)) / np.sqrt(10**(power_dBm/10))
        normalized_separation = abs((z1_nopulse - z1_pulse)/(z0_nopulse - z0_pulse))

        qubit = data.parameters['config'][data.parameters['measure'][0]]
        bias = data.parameters['%s.biasOperate'%qubit]
        legend_str = ('Bias %s' % bias)
        plt.plot(power_dBm, normalized_separation, label=legend_str)
    plt.legend(numpoints=1)

def plot_IQ_ft(data, avg_reps, start, stop):
    t = data.column('Time')[:,0]
    mask = np.logical_and((t> start) , (t<stop))
    data = data[mask,:]
    reps = data.column('Repetition')[:,0]
    if avg_reps > np.max(reps)+1:
        raise ValueError('Not enough reps in data set: %d' % (np.max(reps)+1))

    psd_pulse = 1
    psd_nopulse = 1
    for idx in range(avg_reps):
        mask = reps==idx
        t = data.column('Time')[mask,0]
        z_nopulse = data[mask,2] + 1j*data[mask,3]
        z_pulse = data[mask,4] + 1j*data[mask,5]
        f_pulse = np.fft.fft(z_pulse)
        f_nopulse = np.fft.fft(z_nopulse)
        psd_pulse = psd_pulse + np.abs(f_pulse)**2
        psd_nopulse = psd_nopulse + np.abs(f_nopulse)**2
    freqs = np.fft.fftfreq(len(t), t[1]-t[0])
    plt.figure()
    plt.plot(freqs*1000, 10*np.log10(psd_pulse), label="w/ $\pi$-Pulse")
    #plt.plot(freqs*1000, 10*np.log10(psd_nopulse), label="w/o $\pi$-pulse")
    plt.xlabel('Detuning from Signal (MHz)')
    plt.ylabel('Spectral Density (dB - A.U.)')
    plt.legend()
    make_presentable(plt.gca())
    
def make_presentable(ax):
    xaxis = ax.xaxis
    yaxis = ax.yaxis
    
    for line in ax.get_lines():
        line.set_markersize(12)
        
    xaxis.label.set_fontsize(20)
    yaxis.label.set_fontsize(20)
    ax.title.set_fontsize(20)
    ax.legend(prop={'size': 16}, numpoints=1)
    
    for line in xaxis.get_ticklines():
        line.set_markeredgewidth(5)
        line.set_markersize(5)
    for line in yaxis.get_ticklines():
        line.set_markeredgewidth(5)
        line.set_markersize(5)
    for label in xaxis.get_ticklabels():
        label.set_fontsize(20)
    for label in yaxis.get_ticklabels():
        label.set_fontsize(20)
    

def plot_multi_qubit_time_traces(data, start, stop):
    '''
    Colums of data are: [repetition, time, state, I1, Q1, I2, Q2]
    '''
    
    # Plan:
    # find centroid and axis in I/Q plane for Qubit1 and Qubit2
    # Collapse complex I/Q to real value
    # Iterate over states, 1 plot per state
    # Plot Q1 + i*Q2 vs time for each state
    
    t_mask = (data[:,0] < start) or (data[:,0] > stop)
    data = data[t_mask,:]
    del t_mask
    
    Q1_mean = np.mean(data[:,4:5])
    Q2_mean = np.mean(data[:,6:6])
    Q1_axis = []
    Q2_axis = []

def new_snr_vs_t(ds, label=None):
    z_nopulse = mix_to_dc(ds, 0)
    z_pulse = mix_to_dc(ds, 1)

    nopulse_sum = np.cumsum(z_nopulse, 1)
    pulse_sum = np.cumsum(z_pulse, 1)

    signal = np.median(pulse_sum.real, 0) - np.median(nopulse_sum.real, 0)
    signal = signal + 1j*np.median(pulse_sum.imag, 0) - np.median(nopulse_sum.imag, 0)
    noise = 0.5 * (np.std(pulse_sum.real, 0) + np.std(nopulse_sum.real, 0))
    x = np.hanning(11)
    snr = np.convolve(x, np.abs(signal)/noise) / np.sum(x)
    t = np.arange(len(snr)) * 2.0
    plt.plot(t[5:-10], snr[5:-10], label=label)
    return pulse_sum, nopulse_sum
