#!/usr/bin/python

from __future__ import division
import inspect
import sys
import numbers

import matplotlib.pyplot as plt
import pyle.dataking.util as util
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.quantum as quantum
import pyle.dataking.envelopehelpers as eh

import numpy as np
import pyle.util.sweeptools as st
import pyle.dataking.sweeps as sweeps
import pyle.dataking.singleQubit as sq

from pyle.pipeline import returnValue, FutureList
from pyle.gates import PiPulse, Spectroscopy, Wait, Measure, MeasureWait
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.util.structures import ValueArray
import pyle.envelopes as env

import pyle.dataking.DevDaniel.Qubit.scans as scans

from labrad.units import Unit,Value
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]

def plotCZentanglement():
    theta_pts = np.array([0, 0.05, 0.1, 0.15, 0.2])
    phi_pts = np.linspace(0.5, 20, 200)
    for theta in theta_pts:
        phase = []
        entropy = []
        B = []
        for phi in phi_pts:
            psi = np.cos(theta) * (quantum.ket('001') + np.exp(1j*phi)*quantum.ket('101')) + np.sin(theta)*(quantum.ket('010') + quantum.ket('110'))
            psi = psi/np.sqrt(2)
            rho = quantum.psi2rho(psi)
            rhoq1 = quantum.partial_trace(rho, 1, [2,4])
            B.append(quantum.rho2bloch(rhoq1))
            phase.append(np.angle(rhoq1[1,1]))
            entropy.append(quantum.entropy(rhoq1))
        B = np.array(B)
        plt.plot(phi_pts, np.sqrt(B[:,1]**2 + B[:,2]**2), label="theta = %f" % theta)
        #plt.plot(phi_pts, np.array(entropy), label="theta = %f" % theta)
    plt.legend()
def T1StartDelay(Sample, measure, delay = st.r[50:100e3:1e3, ns], name="T1", stats=1200, 
       save=True, collect=False, state=1):
    
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    axes = [(delay, 'Time')]
    deps = [("Probabiliy |%s>" % state,'','')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)

    def func(server, t):
        alg = gc.Algorithm()
        alg[gates.MoveToState([q], 0, state)]
        if t<1e3*ns:
            alg[gates.Wait([q], t)]
        else:
            start_delay = (t['ns'] //4)*4*ns
            wait_delay = t - start_delay
            alg[gates.Wait([q], wait_delay)]
            q['readoutDevice']['readoutStartDelay'] = start_delay
        alg[gates.Measure(devs)]
        alg.compile()
        print t
        data = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = np.mean(sq.readoutProbabilities(q,np.array(Is)+1.0j*np.array(Qs)))
        
        returnValue([prob])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    if collect:
        return data

def SNR_vs_power(Sample, measure, name='SNR vs Power',
                powers=st.r[-80:-30:2, dBm], stats=900):
    
    snr = []
    for p in powers:
        err0, err1, x = sq.calculateReadoutAxis(Sample, measure, power=p, stats=stats, plot=False, update=False)
        snr.append(x)
    return np.array(zip(powers, snr))
    
def uniq(it):
    '''
    A generator that compresses repeated values
    '''
    first = True
    for x in it:
        if first or last_val != x:
            first = False
            last_val = x
            yield x
def outlier_std(x):
    std = np.std(x)
    med = np.median(x)
    mask = np.logical_and(x< (med+2*std), x> (med-2*std))
    return np.std(x[mask])
def fit_gaussian_outliers(x):
    x0 = np.median(x)
    sigma = np.std(x)
    mask = np.logical_and(x > (x0-2*sigma), x<(x+2*sigma))
    hist, bin_edges = np.histogram(x[mask], bins=50)
    bin_centers = np.convolve(bin_edges, [.5, .5], mode='valid') # find bin centers
    def gaussian_fit(x, y0, x0, sigma):
        return 1/(np.sqrt(2*np.pi*sigma**2)) * y0*np.exp(-(x-x0)**2/(2 * sigma**2))
    x0_guess = x0
    sigma_guess = sigma
    y0_guess = np.max(hist)
    p_opt, pcov = scipy.optimize.curve_fit(gaussian_fit, bin_centers, hist, (y0_guess, x0_guess, sigma_guess))
    return p_opt
    
def analyze_SNR(ds, plot=True):
    '''
    dataset from lollipop(s, m, power=st.r[], dataformat='iqRaw')
    '''
    indep = ds[:,0]
    result = []
    for p in uniq(indep):
        mask = indep == p
        z0 = ds['I |0> pulse'][mask] + 1j * ds['Q |0> pulse'][mask]
        z1 = ds['I |1> pulse'][mask] + 1j * ds['Q |1> pulse'][mask]
        sep = np.mean(z0) - np.mean(z1)
        centroid = (np.mean(z0) + np.mean(z1))/2
        x0 = np.real((z0 - centroid) * np.exp(-1j*np.angle(sep)))
        x1 = np.real((z1 - centroid) * np.exp(-1j*np.angle(sep)))
        sep = np.mean(x0 - x1)
        std0 = fit_gaussian_outliers(x0)[2]
        std1 = fit_gaussian_outliers(x1)[2]
        snr = sep**2 / (std0**2 + std1**2)
        result.append([p, snr, sep**2, std0**2 + std1**2])
    result = np.array(result)
    mask = result[:,0] < -50
    if plot:
        plt.figure()
        plt.plot(result[:,0], 10*np.log10(result[:,1]))
        p = np.polyfit(result[mask,0], 10*np.log10(result[mask,1]), 1)
        plt.plot(result[:,0], np.polyval(p, result[:,0]), '-')
        plt.xlabel("%s (%s)" % (ds.indep[0][0], ds.indep[0][1]))
        plt.ylabel("log10 SNR")
        plt.figure()
        plt.plot(result[:,0], np.exp(-.25 * result[:,1]))
        plt.xlabel("%s (%s)" % (ds.indep[0][0], ds.indep[0][1]))
        plt.ylabel("maximum visibility")   
        plt.figure()
        plt.semilogy(result[:,0], result[:,2], label='Separation^2')
        plt.semilogy(result[:,0], result[:,3], label='variance')
        plt.legend()
    return result

def find_threshold(x, y, threshold=0.5):
    '''
    Returns the x axis value where y=threshold using an interpolator
    '''
    f = scipy.interpolate.interp1d(x, y, bounds_error=False)
    def optim_func(z):
        return (f(z) - threshold)**2
    guess_idx = np.argmin(abs(y - threshold))
    rv = scipy.optimize.fmin(optim_func, x[guess_idx])
    return rv
    
def analyze_qe(dephasing_data, snr_data, threshold=0.5):
    dephasing = analyze_dephasing(dephasing_data, False)
    snr = analyze_SNR(snr_data, False)
    dephasing[:,1] = dephasing[:,1]/np.max(dephasing[:,1])
    max_visibility = np.exp(-.25*snr[:,1])
    snr_thresh = find_threshold(snr[:,0], max_visibility, threshold)
    dephasing_thresh = find_threshold(dephasing[:,0], dephasing[:,1], threshold)
    qe_est = dephasing_thresh-snr_thresh

    print "Quantum efficiency estimate: %s" % qe_est
    plt.figure()
    plt.semilogy(dephasing[:,0], dephasing[:,1], '*', markersize=8, label='Measured Ramsey Visibility')
    plt.semilogy(snr[:,0], np.exp(-.25*snr[:,1]), 'o', markersize=6, label='Ideal Measurement Dephasing')
    plt.semilogy(snr[:,0]+qe_est, np.exp(-.25*snr[:,1]), '+', markersize=8, label='Measurement Dephasing w/ QE=%.1f dB' % qe_est)
    plt.legend(loc='lower left')
    plt.xlabel('%s (%s)' % (dephasing_data.indep[0][0], dephasing_data.indep[0][1]))
    plt.ylabel('Ramsey Visibility (relative)')
    
def readout_dephasing(Sample, measure, name='Readout Dephasing', powers=st.r[-60:-30:2, dBm], 
                        excitationLen=None, excitationFreq=None, df=None, decayLen = 400*ns, 
                        phases=np.linspace(0, 4*np.pi, 40, endpoint=False), stats=900, save=False,
                        mode = 'both'):
    '''
    measure the qubit dephasing due to readout drive.
    '''
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)

    if excitationLen is None:
        excitationLen = q['readoutLen']
    if excitationFreq is None:
        excitationFreq = q['readoutFrequency']
    if df is not None:
        excitationFreq = st.r[excitationFreq['GHz']-df['GHz']:excitationFreq['GHz']+df['GHz']:df['GHz']/20, GHz]
    
    axes = [(powers, 'Power'), (excitationFreq, 'Excitation Frequency'), (phases, 'Ramsey Phase')]
    deps = [('Probability', 'P', '')]
    kw = {'stats': stats, 'decayLen': decayLen, 'excitationLen': excitationLen, 'mode': mode}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw={'stats':stats})
    if mode not in ('first', 'second', 'both'):
        raise RuntimeError('Invalid Mode')
        
    def func(server, power, ex_freq, phase):
        alg = gc.Algorithm(agents=devs)
        if mode in ('first', 'both'):
            alg[gates.PiHalfPulse([alg.q])]
        else:
            alg[gates.PiHalfPulse([alg.q], amp=0)]
#        alg[gates.MeasurePulse([q], power=power, len=readoutLen]
        alg[gates.Wait([alg.q], excitationLen+decayLen)]
        if mode in ('second', 'both'):
            alg[gates.PiHalfPulse([alg.q], phase=phase)]
        else:
            alg[gates.PiHalfPulse([alg.q], phase=phase, amp=0)]
            
        alg[gates.Measure([alg.q])]
        alg.compile()
        amp = eh.power2amp(power)
        sb_freq = ex_freq-q['readoutDevice']['carrierFrequency'] #sideband freq for readout
        alg.q['rr'] += env.mix(env.flattop(-excitationLen-q['piHalfLen']-decayLen, excitationLen, 20, amp=amp), sb_freq)
        data = yield runQubits(server, alg.agents, stats=stats, dataFormat='iqRaw')        
        Is, Qs = data[0]
        prob = sq.readoutProbabilities(q, Is.asarray+1.0j*Qs.asarray)
        returnValue([np.mean(prob)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
def analyze_dephasing(ds, plot=True):
    indep = ds[:,0]
    phase = ds['Ramsey Phase']
    P1 = ds['P']
    
    def sin_fit(x, A0, V, phi0):
        y = A0*(1+V*np.sin(x+phi0))
        return y
    idx = 0
    rv = []
    for idx, p in enumerate(uniq(indep)):
        mask = indep==p
        x = phase[mask]
        y = P1[mask]
        p_opt, pcov = scipy.optimize.curve_fit(sin_fit, x, y, (np.mean(y), 0.5, np.pi/4))
        '''if (idx%5)==0:
            plt.figure()
            plt.plot(x, y)
            plt.plot(x, sin_fit(x, *p_opt), '-')
        '''
        rv.append((p, np.abs(p_opt[1])))
    data = np.array(rv)
    if plot:
        plt.figure()
        plt.plot(data[:,0], data[:,1]/max(data[:,1]))
        plt.xlabel('%s (%s)' % (ds.indep[0][0], ds.indep[0][1]))
        plt.ylabel('Relative Ramsey Visbility')
    return data

def ramsey_LF(Sample, measure, name='Ramsey LF Noise', delay=500*ns,
    phase = np.pi/2, interval=1000*us, extra_delay=12*us, stats=30000, averages=1, 
    save=False, collect=True, noisy=False):

    """Low frequency phase noise estimator based on Ramsey interference.
    
    PARAMETERS
    sample: object defining qubits to measure, loaded from registry
    measure - scalar: number of qubit to measure. Only one qubit allowed.
    delay - scalar: Ramsey delay.  This should be large, but less than T2*
    phase - scalar: The ramsey phase. pi/2 should give the best sensitivity.
    interval - scalar: The measurement interval.  This sets the frequency 
                resolution for the FFT.  delay/interval determines the noise floor
    extra_delay - scalar: estimate of the extra per-shot delay in the FPGAs used
                    to calibrate the time axis properly.
    stats - scalar: number of times a point will be measured per iteration over
            the time axis.  1/(stats*interval) = measurement time
    name - string: Name of dataset.
    save - bool: Whether or not to save data to the datavault
    collect - bool: Whether or not to return data to the local scope.
    reps - scalar: Number of sequence blocks to run  These can be averaged to improve FFT
                        resolution
    """
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    for d in devs:
        d['biasOperateSettling'] = interval    
    if averages==1:
        rep_vec = 1
    else:
        rep_vec = np.arange(averages)
    V_est = 0.8 # ramsey visibility used for calibration
    axes = [(rep_vec, 'Repetition')]
    deps = [('Probability', 'P1', '')]
    kw = {'stats': stats, 'delay':delay, 'interval': interval, 'phase':phase}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw,explicitIndependents=[('Time','ns')])
    def func(server, rep):
        reqs = []

        alg = gc.Algorithm(agents=devs)
        alg[gates.PiHalfPulse([alg.q], phase = 0.0)]
        alg[gates.Wait([alg.q], delay)]
        alg[gates.PiHalfPulse([alg.q], phase = phase)]
        alg[gates.Measure([alg.q])]
        alg.compile()
        result = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = result[0]
        prob = sq.readoutProbabilities(alg.q, Is.asarray+1.0j*Qs.asarray)

        ts = np.arange(len(prob))*(interval['ns']+extra_delay['ns'])
        returnValue(np.column_stack((ts, prob)))
    #Run the scan and save data
    data = sweeps.grid(func, axes, dataset = dataset, save=save, collect=collect, noisy=noisy, pipesize=1) # Don't know if this is needed
    #
    # In the linear regime, f(t) - f0 = (P(t)-0.5) / pi*V*t
    # where V is the ramsey visibility (nominally 1, typically ~0.8)
    # and t is the Ramsey time.  The readout state is a binary 0/1, but
    # we can imagine that it is a measurement of the frequency with some large
    # noise term.  Because projection is memoryless, the noise spectrum is white
    # and has the standard noise power of quantization noise 1 LSB/12 over the
    # entire bandwidth, for a propability spectral density of 
    # sqrt(2*dt/(12)) in per-root-Hz units.
def analyze_ramsey_LF(data):

    if averages>1:
        Zf = 0
        reps = data['Repetition']
        for rep in range(averages):
            mask = reps == rep
            t = data['Time'][mask]
            p = data['P1'][mask]
            f = 2*interval['s'] * (p - np.mean(p)) / (1.0 * V_est * np.pi * delay['s'])
            x = np.abs(np.fft.fft(f)**2) / len(p)
            Zf = Zf + x
        Zf = Zf / averages
    else:
        t = data['Time']
        p = data['P1']
        f = 2 * interval['s'] * (p - np.mean(p)) / ( 1.0 * V_est * np.pi * delay['s'])
        Zf = np.abs(np.fft.fft(f))**2/len(p)
    psd = Zf
    
    freqs = np.fft.fftfreq(len(psd), interval['s'])
    noise_floor = 0.25 * (2*interval['s']/(1.0 *  V_est * np.pi * delay['s']))**2
    print "Noise floor: %f Hz^2 / Hz" % noise_floor
    mask = freqs>0
    plt.figure()
    plt.loglog(freqs[mask], psd[mask], label='f10(f) PSD')
    plt.loglog(freqs[mask], noise_floor*np.ones_like(freqs[mask]), '-', label='Quantization Noise Floor') 
    plt.ylabel('f10 (Hz)^2 / Hz')
    plt.xlabel('sb freq(Hz)')
    if collect:
        return data

def calculateReadoutError(SNoPulse, SPulse):
    x1 = np.mean(SNoPulse)
    std1 = np.std(SNoPulse)
    mask = abs(SNoPulse-x1) < 3*std1
    x1 = np.mean(SNoPulse[mask])
    
    x2 = np.mean(SPulse)
    std2 = np.std(SPulse)
    mask = abs(SPulse-x2) < 3*std2
    x2 = np.mean(SPulse[mask])
    center = (x1+x2)/2
    axis = (x2-x1)/abs(x2-x1)
    err0 = np.mean(np.real((SNoPulse - center) * axis.conjugate())>0)
    err1  = np.mean(np.real((SPulse - center) * axis.conjugate())<=0)
    return err0, err1

def optimizeReadout(Sample, measure, name='Readout Optimizer',
            readout_power=st.r[-50:-30:1,dBm],
            readout_len=st.r[200:2000:200, ns],
            stats=900,
            save=True, collect=False, noisy=True):
    """sweep over readout parameters to find optial readout power and length"""
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]

    axes = [(readout_power,'Readout Power'), (readout_len, 'Readout Time')]
    deps = [('Probability','|0> fidlity',''), ('Probabilty','|1> fidelity','')]
    kw = {'stats': stats}
        
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, power, length):
        q['readoutPower'] = power
        q['readoutLen'] = length
        q['adcFitlerWindow'] = [250*ns, length+250*ns]
        reqs = []
        #Measure with no pi pulse
        alg = gc.Algorithm()
        alg[gates.Measure(devs)]
        alg.compile()
        reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        #Measure with pi pulse
        alg = gc.Algorithm()
        alg[gates.PiPulse([q])]
        alg[gates.Measure(devs)]
        alg.compile()
        reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        data = yield FutureList(reqs)
        noPulse, pulse = data
        InoPulse, QnoPulse = noPulse[0]
        IPulse, QPulse = pulse[0]
        SNoPulse = np.array(InoPulse) + 1j*np.array(QnoPulse)
        SPulse = np.array(IPulse) + 1j*np.array(QPulse)
        err0, err1 = calculateReadoutError(SNoPulse, SPulse)
        print "power %s readoutlen: %s |0> fidelity: %s |1> fidelity: %s" % (power, length, 1-err0, 1-err1)
        returnValue([1-err0,1-err1])
     
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    return data

def readout_adc_window(Sample, measure, name="ADC window optimization",
        stats=900, start_delay = st.r[50:500:10, ns], stop_delay=0*ns,
        save=True):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    devs = gc.registryWrapper2GateCompiler(sample, qubits, measure)
    q = devs[measure]
    
    axes = [(start_delay, "Start Delay"), (stop_delay, "Stop Delay")]
    deps = [("Amplitude", "I No Pulse", "Clicks"), ("Amplitude", "Q No Pulse", "Clicks"),
            ("Amplitude", "I Pulse", "Clicks"), ("Amplitude", "Q Pulse", "Clicks")]
    kw = {'stats':stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)

    def func(server, tStart, tDelay):
        alg1 = gc.Algorithm(agents=devs, measure=measure)
        readoutLen = alg1.q['readoutLen']
        alg1.q['adcFilterWindow'] = (tStart, tStart+readoutLen+tDelay)
        alg1[gates.Measure([alg1.q])]
        alg1.compile()
        result = yield runQubits(server, alg1.agents, stats, dataFormat='iq')
        INoPulse, QNoPulse = result[0]
        
        alg2 = gc.Algorithm(agents=devs, measure=measure)
        readoutLen = alg1.q['readoutLen']
        alg2.q['adcFilterWindow'] = (tStart, tStart+readoutLen+tDelay)
        alg2[gates.PiPulse([alg2.q])]
        alg2[gates.Measure([alg2.q])]
        alg2.compile()
        result = yield runQubits(server, alg2.agents, stats, dataFormat='iq')
        IPulse, QPulse = result[0]
        returnValue([INoPulse, QNoPulse, IPulse, QPulse])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    t = data[:,0]
    zNoPulse = data[:,1] + 1j*data[:,2]
    zPulse = data[:,3] + 1j*data[:,4]
    plt.figure()
    plt.plot(t, abs(zNoPulse-zPulse))
def measurement_crosstalk(Sample, measure,
            name='Measurement Crosstalk',
            stats=900, repetitions=20, save=True, collect=False, noisy=True):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    for tmp in [qubits[x] for x in measure]:
        tmp['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = [devs[m] for m in measure]

    axes = [(np.arange(repetitions), 'Repetition')]
    deps = []
    for qubit in q:
        axes.append(((0,1), '%s state'%qubit.__name__))
        deps.extend([('Amp', '%s I'%qubit.__name__, ''), ('Amp', '%s Q'%qubit.__name__, '')])

    kw = {'stats': stats}
    print axes
    print deps
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)   
    def func(server, rep, *qubit_states):
        alg = gc.Algorithm()
        for (idx,qubit) in enumerate(q):
            if qubit_states[idx]:
                alg[gates.PiPulse([qubit])]
        alg[gates.Measure(devs)]
        alg.compile()
        data = yield runQubits(server, devs, stats, dataFormat='iq')
        result = []
        for l in data:
            result.extend(l) 
        returnValue(result)
    d= sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    return d
    
def paramp_gain_saturation(sample, measure, power=None):
    '''
    Measure paramp gain saturation
    '''
    if power is None:
        power = ValueArray(np.arange(-50,-25,1), dBm)
    data=sq.lollipop(sample, measure, stats=600, name="Paramp Gain Saturation", 
        dataFormat='iq', save=False, power=power, plot=False)
    return data
   

def FluxNoiseLF(Sample, measure, reps, stats, interval):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    axes = [(delay, 'Time'), (np.arange(reps), 'Repetition'), (tomoops, "Tomo Op")]
    deps = [("Outcome",'','')]
    kw = {'stats': stats}
    q['readout'] = True
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)

def demod_out_vs_time(Sample, measure, name="Demodulator Output vs. Readout Time", 
    readoutTime=st.r[80:5000:40, ns], save=True, stats=1500):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    q['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    q['readout'] = True
    q['adc mode'] = 'demodulate'
    readoutTimeRef = q['readoutLen']
    axes = [(readoutTime, 'Readout Time')]
    deps = [("INoPulse",'',''), ("QNoPulse", "", ""),("IPulse",'',''), ("QPulse", "", "")]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, t):
        q['readoutLen'] = t
        q['adcFilterWindow'] = (200.0*ns, 240.0*ns+t)
        alg = gc.Algorithm()
        alg[gates.Wait([q], 30*ns)]
        alg[gates.Measure(devs)]
        alg.compile()
        dataNoPulse = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        Is,Qs = dataNoPulse[0]
        ZNoPulse = np.mean(np.array(Is) +1j*np.array(Qs))*readoutTimeRef/t
        q['readoutLen'] = t
        q['adcFilterWindow'] = (200.0*ns, 240.0*ns+t)
        alg = gc.Algorithm()
        alg[gates.PiPulse([q])]
        alg[gates.Measure(devs)]
        alg.compile()
        dataPulse = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        Is,Qs = dataPulse[0]
        ZPulse = np.mean(np.array(Is)+1j*np.array(Qs))*readoutTimeRef/t
        returnValue([ZNoPulse.real, ZNoPulse.imag, ZPulse.real, ZPulse.imag])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
def readout_fidelity_vs_time(Sample, measure, name='Readout fidelity vs. Time', 
    readoutTime=st.r[100:2000:50, ns], save=True, stats=1500):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    q['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    q['readout'] = True
    q['adc mode'] = 'demodulate'
    readoutTimeRef = q['readoutLen']
    axes = [(readoutTime, 'Readout Time')]
    deps = [("P 0->1",'',''), ("Probability 1->1", "", "")]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, t):
        q['readoutLen'] = t
        q['adcFilterWindow'] = (200.0*ns, 250.0*ns+t)      
        algNoPulse = gc.Algorithm()
        algNoPulse[gates.Measure(devs)]
        algNoPulse.compile()
        dataNoPulse = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        Is,Qs = dataNoPulse[0]
        p01 = np.mean(sq.readoutProbabilities(q,(np.array(Is)+1.0j*np.array(Qs))*readoutTimeRef/t))
        print "p00: ", p01
        q['readoutLen'] = t
        q['adcFilterWindow'] = (200.0*ns, 250.0*ns+t)        
        algPulse = gc.Algorithm()
        algPulse[gates.PiPulse([q])]
        algPulse[gates.Measure(devs)]
        algPulse.compile()
        dataPulse = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        Is,Qs = dataPulse[0]
        p10 = np.mean(sq.readoutProbabilities(q,(np.array(Is)+1.0j*np.array(Qs))*readoutTimeRef/t))
        print "p11: ", p10
        returnValue([p01, p10])
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
def dualQubitTest(Sample, measure, stats=900):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    qubits = [ devs[m] for m in measure ]
    
    for q in qubits:
        q['readout'] = True
        q['adc mode'] = 'demodulate'
    alg = gc.Algorithm()
    for q in qubits:
        alg[gates.PiPulse([q])]
    alg[Measure(devs)]
    alg.compile()
    rv = runQubits(Sample._cxn.qubit_sequencer, devs, stats=stats, dataFormat='iqRaw', debug=True)
    return rv.wait()
    
def dualBlockTest(Sample, measure, delay=500*ns, dual_block=True):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    q['readout'] = True
    q['adc mode'] = 'demodulate'    
    alg = gc.Algorithm()
    alg[gates.PiPulse([q])]
    if dual_block:
        alg[gates.DualBlockWaitDetune(devs, q, len=delay, w=15*ns, amp=0.5)]
    else:
        alg[gates.Detune([q],len=delay,w=15*ns,amp=0.5)]     
    alg[Measure(devs)]
    alg.compile()
    rv = runQubits(Sample._cxn.qubit_sequencer, devs, stats=100, dataFormat='iqRaw', debug=False)
    return np.array(rv.wait())
    
def analyze_caller(start_depth=1, max_frames=float("inf")):
    frames = inspect.stack()
    nframes = min(max_frames, len(frames)-start_depth)
    parameters={}
    for k in range(nframes):
        frame = frames[k+start_depth]
        parameters['stack.frame%d.file' % k] = frame[1]
        parameters['stack.frame%d.func' % k] = frame[3]
        parameters['stack.frame%d.line' % k] = frame[2]
        args, varargs, keywords, locals = inspect.getargvalues(frame[0])
        parameters['stack.frame%d.args' % k] = inspect.formatargvalues(args, varargs, keywords, locals)
        for a in args:
            parameters['stack.frame%d.arg.%s' % (k, a)] = repr(locals[a])
        if frame[1] == '<ipython console>':
            break
    return parameters

def iqVsTime2Q(Sample, measure, name = 'IQ vs time multi-qubit',
                stats=1, averages=1, repetitions=20, tmax_ns=5000, save=True, noisy=False):
    """IQ vs time traces"""
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = [qubits[m] for m in measure]
    for qubit in q:
        qubit['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = [devs[m] for m in measure]
    axes = [(np.arange(repetitions), 'Repetition')]
    for qubit in q:
        axes.append(([0,1], '|%s> state' % qubit.__name__))
    explicit_indeps = [('Time', 'ns')]
    deps = [('Amplitude', 'I', ''), ('Amplitude', 'Q', '')]
    kw = {'stats':stats, 'averages': averages, 'repetitions': repetitions}
    dataset = sweeps.prepDataset(sample, name, axes, deps, 
                explicitIndependents = explicit_indeps, 
                measure=measure, kw=kw)
    def func(server, rep, *states):
        alg = gc.Algorithm()
        for (qubit, s) in zip(q, states):
            qubit['adc mode'] = 'average'
            if s:
                alg[gates.PiPulse([qubit])]
        for qubit in q:
            alg[gates.Measure([qubit])]
        alg.compile()
        npts = int(tmax_ns//2)
        t = 2.0*np.arange(npts)
        reqs = []
        for x in range(averages):
            reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        data = yield FutureList(reqs)
        I=0.0
        Q=0.0
        for x in data: 
            I = I + np.array(x[0][0])[0:npts]
            Q = Q + np.array(x[0][1])[0:npts]
        I = 1.0*I / averages
        Q = 1.0*Q / averages
        returnValue(np.column_stack((t,I,Q)))
    data = sweeps.grid(func, axes, dataset = save and dataset, noisy=noisy)
    
def iqVsTime2DPhaseCoherent(Sample, measure, name='IQ vs time Phase Coherent',
                            freq=None, power=None, averages=1, stats=1, repetitions=20,
                            filterBw=None, downmix=False, downsampleFactor=1, timeMax_ns=5000):
    """IQ vs time traces"""
    sample, qubits = util.loadQubits(Sample, write_access=False)
    q = qubits[measure]
    q['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    filterPts = 512
    if freq is not None:
        q['readoutFrequency'] = freq
    if power is not None:
        q['readoutPower'] = power
    
    indeps = [('Repetition', ''), ('Time','ns')]
    deps=[('Amplitude','I no pulse',''),
          ('Amplitude','Q no pulse',''),
          ('Amplitude','I with pulse',''),
          ('Amplitude','Q with pulse','')]
    kw = {'stats':stats, 'averages':averages, 'repeitions': repetitions, 'filterBw':filterBw, 'downsampleFactor':downsampleFactor,'filterPts':filterPts}
    dataset = sweeps.prepDataset(sampleReg, name, axes=[], explicitIndependents = indeps, dependents=deps, measure=measure, kw=kw, cxn=Sample._cxn)  
    #Loop over repetitions
    with dataset:
        for repetition in range(repetitions):
            INoPulse,QNoPulse = readoutTimeTracePhaseCoherent(Sample, measure,
                                                              freq=freq, power=power,
                                                              piPulse=False,
                                                              averages=averages, stats=stats,
                                                              collect=True,
                                                              plot=False, regParams=(sampleReg,qubits))
            IPulse,QPulse = readoutTimeTracePhaseCoherent(Sample, measure,
                                                          freq=freq, power=power,
                                                          piPulse=True,
                                                          averages=averages, stats=stats,
                                                          collect=True,
                                                          plot=False, regParams=(sampleReg,qubits))
            t = np.linspace(0, (8192*2)-2, 8192)
            repetitionVector = np.zeros(8192)+repetition
            downsampler = slice(0,None,downsampleFactor)
            tMask = t<timeMax_ns
            dataset.add(np.vstack((
                                   repetitionVector[tMask][downsampler],
                                   t[tMask][downsampler],
                                   INoPulse[tMask][downsampler],
                                   QNoPulse[tMask][downsampler],
                                   IPulse[tMask][downsampler],
                                   QPulse[tMask][downsampler])).T)

def readoutPreEmph(Sample, measure, name = 'Readout Pre-Emphasis',
                  preEmphFactor=3.0,
                  preEmphLen=st.r[0:200:25,ns],
                  averages=300, stats=30, collect=False, save=True, noisy=True):
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    if not isinstance(q, gc.TransmonPreEmphasis):
        raise RuntimeError('Device type %s does not support pre-emphasis' % type(q))
    Q['preEmphFactor'] = preEmphFactor
    axes = [(preEmphLen, 'Pre-Emphasis length')]
    explicitIndependents= [("Time", "ns")]
    deps = [ ('Amplitude', 'I No Pulse', ''), 
             ('Amplitude', 'Q No Pulse', ''), 
             ('Amplitude', 'I Pulse', ''), 
             ('Amplitude', 'Q Pulse', '')]
    kw = {'stats': stats}        
                
    dataset = sweeps.prepDataset(sample, name, axes, deps, explicitIndependents=explicitIndependents, measure=measure, kw=kw)
    
    def func(server, length):
        Q['preEmphLen'] = length
        INoPulse,QNoPulse = yield scans.readoutTimeTracePhaseCoherent(Sample, measure, averages=averages, stats=stats, piPulse=False, collect=True, plot=False)
        IPulse,QPulse = yield scans.readoutTimeTracePhaseCoherent(Sample, measure, averages=averages, stats=stats, piPulse=True, collect=True, plot=False)
        t = np.arange(8192)*2.0
        mask = t<5000
        
        rv = np.column_stack((t[mask], INoPulse[mask], QNoPulse[mask], IPulse[mask], QPulse[mask]))
        print rv.shape
        returnValue(rv)
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False) 
    
    if collect:
        return data

def alphaScan2State(Sample, measure, name='Alpha Scan 2 state',
    phases=np.linspace(0, np.pi*2, 20), 
    alphas=np.linspace(0.3,0.7,20), 
    num_pulses=10, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(alphas, 'Alpha'), (phases, 'Phase')]
    deps = [('Probability', 'P1', ''), ('Probability', 'P2', '')]
    
    def func(server, alpha, phase):
        alg = gc.Algorithm(agents=devs)
        for idx in range(num_pulses):
            alg[gates.PiPulse([alg.q], phase=idx*phase)]
        alg[gates.Measure([alg.q])]
        data  = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        prob = readoutProbabilities(q, Is.asarray+1.0j*Qs.asarray)
        
def iqVsPhase(s, measure, basename='IQ vs time',phases=None):
    if phases is None:
        phases = np.pi*np.arange(8)/8.0
    for phi in phases:
        s['q4']['readoutphase'] = phi
        iqVsTimePhaseCoherent(s, measure, name='IQ vs time phase=%f' % phi,
                                  averages=1, stats=1,
                                  timeMax_ns=17000,
                                  save=True, collect=False, noisy=True)

from pyle.signalprocessing import interp1d_cubic as rami_interp
from scipy.interpolate import interp1d
import scipy.signal
import time

def interp_erj(x, y, bounds_error=False,fill_value=None):
    '''
    Emulate a general purpose interpolator using the scipy.signal cspline code (which requires that the
    data points be equally spaced
    '''
    if x[0] > x[-1]: # don't want to deal with reversed arrays -- only matters for bounds checking
        x = x[::-1]
        y = y[::-1]
    spacing = np.diff(x)
    if any(np.abs(spacing-spacing[0]) > 5*np.max(np.abs(x))*sys.float_info.epsilon):
        raise RuntimeError('X points not equally spaced')
    spline_data = scipy.signal.cspline1d(y)
    def interp_func(x_new):
        low_mask = x_new < x[0]
        high_mask = x_new > x[-1]
        if bounds_error and (any(low_mask) or any(high_mask)):
            raise RuntimeError('interpolation out-of-bounds')

        dx = 1.0*(x[-1]-x[0])/(len(x)-1)
        tmp = scipy.signal.cspline1d_eval(spline_data, x_new, dx, x[0])
        tmp[low_mask] = fill_value if fill_value is not None else y[0]
        tmp[high_mask] = fill_value if fill_value is not None else y[-1]
        return tmp
    return interp_func

def time_interp(N, f=5, reps=10):
    t = np.linspace(0, 1, N)
    x = np.sin(t*2*np.pi/f)
    x = x * np.exp(8* (t-0.5)**2)
    t1 = time.time()
    for idx in range(reps):
        fn_rami = rami_interp(t, x)
    t2 = time.time()
    t_create_rami = t2-t1
    t1 = time.time()
    for idx in range(reps):
        fn_erj = interp_erj(t, x)
    t2 = time.time()
    t_create_erj = t2-t1
    t_new = np.linspace(0.2, 0.5, N/2.0)
    t1 = time.time()
    for idx in range(reps):
        y = fn_rami(t_new)
    t2 = time.time()
    dt_rami = t2 - t1
    t1 = time.time()
    for idx in range(reps):
        y = fn_erj(t_new)
    t2 = time.time()
    dt_erj = t2- t1
    print "create_time_rami: %f, create_time_erj: %f" % (t_create_rami, t_create_erj)
    print "dt_rami: %f, dt_erj: %f" % (dt_rami, dt_erj)

def test_interp_residual(N, f=5):
    t = np.linspace(1, 0, N)
    def func(x):
        tmp = np.sin(x*2*np.pi*f)
        tmp = tmp * np.exp(-4* (x-0.5)**2)
        #tmp = scipy.special.erf((x-0.5)*4)
        return tmp
    y = func(t)
    fn_rami = rami_interp(t, y)
    fn_erj = interp_erj(t, y)
    t_new = np.linspace(0, 1.0, .7*N)
    y_new = func(t_new)
    y_rami = fn_rami(t_new)
    y_erj = fn_erj(t_new)
    plt.figure()
    plt.plot(t_new, y_new)
    plt.grid()
    plt.figure()
    plt.plot(t_new, y_rami-y_new, label='rami')
    plt.plot(t_new, y_erj-y_new, label='erj')
    plt.grid()
    plt.legend()
    plt.show()
