import os

from labrad.units import Unit
ns, us, V, mV, GHz = [Unit(s) for s in ('ns', 'us', 'V', 'mV', 'GHz')]

import pyle.envelopes as env
from pyle.dataking import envelopehelpers as eh
import pyle.dataking.DevJim.CatchRelease.gateCompiler as gc
import numpy as np

# default channel identifiers
FLUX = lambda q: (q.__name__, 'flux')
SQUID = lambda q: (q.__name__, 'squid')

# bias command names
FAST = 'dac1'
SLOW = 'dac1slow'

# default sram padding
PREPAD = 1000
POSTPAD = 50


# debug dumps 
DEBUG_PATH = os.path.join(os.path.expanduser('~'), '.packet-dump')


def filterBytes(filterLen = 16384*ns, filterWindowStart= 0*ns, filterWindowEnd=16384*ns):
    if filterWindowStart['ns'] < 0 or filterWindowStart['ns'] > filterLen['ns']:
        filterWindowStart = 0*ns
    if filterWindowEnd['ns'] < 0 or filterWindowEnd['ns'] > filterLen['ns']:
        filterWindowEnd = filterLen
    filter_len = round(filterLen['ns']/4)
    filt = np.zeros(filter_len,dtype='<u1')
    indexStart = filterWindowStart['ns']/4
    indexEnd = filterWindowEnd['ns']/4 +1
    filt[indexStart:indexEnd] = 128
    return filt.tostring()
    

def checkChannel(qubit, name):
    """Finds out if a ADC is used by qubit"""
    channelNames = dict(qubit['channels']).keys()
    if name in channelNames:
        return True
    else:
        return False

def sortDevices(devices):
    sortedDevices = {}
    for type in gc.CONSTRUCTORS.keys():
        sortedDevices[type] = []
    for device in devices:
        for type in gc.CONSTRUCTORS.values():
            if isinstance(device, type):
                sortedDevices[type].append(device)
                break
        else:
            raise Exception('Unrecognized device type')
    return sortedDevices

def sortqubits_ADC(devices):
    """Get a dictionary where ADC names key a list of devices that use that ADC"""
    groups = {}
    for d in [d for d in devices if d.get('readout',False)]:
        if isinstance(d, gc.Resonator):
            adcBoard = dict(d['channels']).get('readoutADC', None)
            if adcBoard:
                ADCname = adcBoard[1][0]
                if ADCname in groups.keys():
                    groups[ADCname].append(d)
                else:
                    groups[ADCname] = [d]
        if isinstance(d, gc.Qubit):
            r = d.get('readoutDevice',None)
            if r and isinstance(r,gc.Readout):
                adcBoard = dict(r['channels']).get('readoutADC', None)
                if adcBoard:
                    ADCname = adcBoard[1][0]
                    if ADCname in groups.keys():
                        groups[ADCname].append(d)
                    else:
                        groups[ADCname] = [d]
    return groups
        
def sortqubits_readoutDAC(qubits):
    groups = {}
    for q in qubits:
        r = q['readoutDevice']
        channels = dict(r['channels'])
        readoutDACname = channels['uwave'][1][0]
        if readoutDACname in groups.keys():
            groups[readoutDACname].append(q)
        else:
            groups[readoutDACname] = [q]
    return groups                       
    
def runQubits(server, devices, stats, probs=None, dataFormat=None,
              debug=False):
    """A generic sequence for running multiple devices with the Qubit Sequencer.
    qubits
    These parameters control the sequence and how it should be run:
    
    server - labrad server to which we send the sequence (Qubit Sequencer).
    devices - list of dict-like objects (typically loaded from the registry)
        containing the configuration parameters of all devices involved in
        the sequence.
    stats - the number of repetitions of the experiment to be run.
    
    
    These parameters control how the data should be processed before it is
    returned to the caller (note that this processing is done by the Qubit
    Sequencer server for speed; see the server documentation for more info): 
    
    dataFormat - string description of how data should be returned
    probs - a list of which probabilities should be returned.  For example,
        when only measuring a single qubit this should be [1] since we only
        care about the |1>-state probability.
    
    DEPRICATED DATA FORMAT FLAGS
    raw - a flag indicating whether to return raw switching data (converted
        to microseconds) or processed probabilities (the default). Replaced
        by 'raw_microseconds' for SQUID readout and 'iq' for resonator
        readout.
    separate - a flag indicating whether to return separate probabilities
        for each qubit (N possibilities), or combined probabilities of the
        multi-qubit states (2**N posibilities; the default). Replaced by
        'probs_separate'.
    switches - Replaced by 'raw_switches'
    """
    if dataFormat is None:
        dataFormat = 'probs'
    
    p = server.packet()
    makeSequence(p, devices)
    p.build_sequence()
    p.run(long(stats))
    
    # get the data in the desired format
    if dataFormat == 'raw':
        p.get_data_raw(key='data')
    elif dataFormat == 'raw_microseconds':
        p.get_data_raw_microseconds(key='data')
    elif dataFormat == 'iq':
        p.get_data_raw(key='data')
    elif dataFormat == 'phases':
        p.get_data_raw_phases(key='data')
    elif dataFormat == 'raw_switches':
        p.get_data_raw_switches(key='data')
    elif dataFormat == 'probs_separate':
        p.get_data_probs_separate(probs, key='data')
    elif dataFormat == 'probs':
        p.get_data_probs(probs, key='data')
    else:
        raise Exception('dataFormat %s not recognized' %str(dataFormat))
    
    #Send the packet with wait=False, therefore returning a Future
    return sendPacket(p, debug)


def runInterlaced(server, qubits, stats, probs=None, raw=False, separate=False, debug=False):
    # TODO: handle generic interlacing (not just defferent SRAM blocks)
    raise Exception('not implemented yet')


def runInterlacedSRAM(server, qubits, stats, probs=None, raw=False, separate=False, debug=False):
    """A generic sequence for running multiple qubits with the Qubit Sequencer.
    
    These parameters control the sequence and how it should be run:
    
    server - labrad server to which we send the sequence (Qubit Sequencer).
    qubits - list of dict-like objects (typically loaded from the registry)
        containing the configuration parameters of all qubits involved in
        the sequence.
    stats - the number of repetitions of the experiment to be run.
    
    
    These parameters control how the data should be processed before it is
    returned to the caller (note that this processing is done by the Qubit
    Sequencer server for speed; see the server documentation for more info): 
    
    probs - a list of which probabilities should be returned.  For example,
        when only measuring a single qubit this should be [1] since we only
        care about the |1>-state probability.
    raw - a flag indicating whether to return raw switching data (converted
        to microseconds) or processed probabilities (the default).
    separate - a flag indicating whether to return separate probabilities
        for each qubit (N possibilities), or combined probabilities of the
        multi-qubit states (2**N posibilities; the default).
    """
    raise Exception('TODO: Revised Interlaced for separate devices.')
    numSRAMs = set([len(q['xy']) for q in qubits if 'xy' in q] + [len(q['z']) for q in qubits if 'z' in q])
    if len(numSRAMs) > 1:
        raise Exception('all qubits must have same number of xy a z sequences')
    numSRAM = numSRAMs.pop()
    
    p = server.packet()
    makeSequenceInterlaced(p, qubits, numSRAM)
    p.build_sequence()

    p.run(long(stats))
    
    # get the data in the desired format with deinterlacing enabled
    if raw: p.get_data_raw_microseconds(numSRAM, key='data')
    elif separate: p.get_data_probs_separate(probs, numSRAM, key='data')
    else: p.get_data_probs(probs, numSRAM, key='data')
    
    return sendPacket(p, debug)


def runDualblock(server, qubits, stats, probs=None, raw=False, separate=False, debug=False):
    # TODO: handle dual-block SRAM (need another argument)
    raise Exception('not implemented yet')


def makeSequence(p, devices): 
    """Make a memory/sram sequence to be passed to the Qubit Sequencer server.
    
    Sequences are made in several steps
    1. First we add anything to the SRAM sequences that we need but wasn't added by the user.
    For example, the phase qubits with resonator readout need to have a readout pulse in the SRAM
    of the DAC controlling the readout resonator. We add that pulse here.
    2. Find the total time spanned by all SRAM sequences and save this information for later
    3. Initialize the qubit sequencer, telling it all of the channels we need to use.
    4. Configure external resources. For example, we must declare the microwave source frequency to
    the qubit sequencer so that it can check that shared sources have the same frequency, etc.
    5. Add memory commands to the DAC boards
    6. Add SRAM to the dac boards
    """
    sortedDevices = sortDevices(devices)
    phaseQubits = sortedDevices['phaseQubit']
    tunableCouplers = sortedDevices['tunableCoupler']
    resonators = sortedDevices['resonator']
    groups_ADC = sortqubits_ADC(phaseQubits+resonators)
    #1. Add SRAM where needed
    #1a. Add readout pulses to phaseQubits with resonator readout
    """
    for q in phaseQubits:
        if ((readoutType(q) is 'resonator') and (q.get('readout', False))):
            q['adc demod frequency'] = q['readout frequency']- q['readout fc']
            q['rr'] = eh.readoutPulse(q,0)
    """
    #2. Find total time of SRAM sequences
    #Check total xyz (SRAM) envelope length, taking all devices into account. Devices with no 'xy', 'z',
    #or 'rr' defined have their envelope default to env.NOTHING 
    envelopes = []
    for dev in devices:
        envelopes.extend([dev.get('xy',env.NOTHING),dev.get('z',env.NOTHING),dev.get('rr',env.NOTHING)])
    t = checkTiming(envelopes)
    #3. construct the packet for the Qubit Sequencer
    p.initialize([(d.__name__, d['channels']) for d in devices])
    addConfig(p, devices)
    p.new_mem()
    #checked to here
    memqubits = [q for q in phaseQubits+tunableCouplers if 'flux' in dict(q['channels'])]
    addMem(p, memqubits, ['block0'])
    addSram(p, devices, [t], ['block0'])   
    for key in groups_ADC.keys():
        addADC(p, groups_ADC[key])


def makeSequenceInterlaced(p, qubits, numSRAM):
    """Make a memory/sram sequence using the Qubit Sequencer server.
    
    This version allows for interlacing with multiple SRAM blocks.
    """
    
    p.initialize([(q.__name__, q['channels']) for q in qubits])
    addConfig(p, qubits)
    p.new_mem() # call this once, even if we use multiple blocks
    memqubits = [q for q in qubits if 'flux' in dict(q['channels'])]
    
    for i in range(numSRAM):
        # calculate time range from given control envelopes
        xys = [q['xy'][i] for q in qubits if 'xy' in q]
        zs = [q['z'][i] for q in qubits if 'z' in q]
        t = checkTiming(xys + zs)
        
        block = 'block%d' % i
        addMem(p, memqubits, [block])
        addSram(p, qubits, [t], [block], idx=i)


def sendPacket(p, debug):
    """Finalize a packet for the Qubit Sequencer, send it, and add callback to get data.
    
    Also sets up debugging if desired to log the packet being sent to the
    Qubit Sequencer, as well as the packet sent from the Qubit Sequencer to the GHz DACs.
    
    
    """
    if debug:
        fname = os.path.join(DEBUG_PATH, 'qubitServerPacket.txt')
        with open(fname, 'w') as f:
            print >>f, p
        p.dump_sequence_packet(key='pkt')
    
    # send the request    
    req = p.send(wait=False)
    
    if debug:
        def dump(result):
            from pyle.dataking.qubitsequencer import prettyDump
            fname = os.path.join(DEBUG_PATH, 'ghzDacPacket.txt')
            with open(fname, 'w') as f:
                print >>f, prettyDump(result['pkt'])
            return result
        req.addCallback(dump)
    
    # add a callback to unpack the data when it comes in
    req.addCallback(lambda result: result['data'])
    return req


def checkTiming(envelopes):
    """Calculate the timing interval to encompass a set of envelopes."""
    start, end = env.timeRange(envelopes)
    if start is not None and end is None:
        raise Exception('sequence has start but no end')
    elif start is None and end is not None:
        raise Exception('sequence has end but no start')
    elif start is None and end is None:
        t = 0, 40 # default time range
    else:
        t = start, end
    return t


def addConfig(p, devices, autotrigger='S3'):
    """Add config information to a Qubit Sequencer packet.
    
    Config information includes:
        - which qubits to read out (timing order)
        - microwave source settings (freq, power)
        - preamp settings (offset, polarity, etc.)
        - settling rates for analog channels
        - autotrigger
    """   
    p.new_config()
    timing_order = []
    for d in devices:
        if d.get('readout',False):
            if isinstance(d, gc.Qubit):
                if isinstance(d.get('readoutDevice',None), gc.Readout):
                    adcString = 'readout ADC::'+str(d['adc channel']) if (d.get('adc mode', None)=='demodulate') else 'readout ADC'
                    timing_order.append((d.__name__, adcString))
                elif checkChannel(d, 'squid'):
                    timing_order.append((d.__name__, 'timing'))
                else:
                    raise Exception('Qubit readout type not determined.')
            elif isinstance(d, gc.Resonator):
                adcString = 'readout ADC::'+str(d['adc channel']) if (d.get('adc mode', None)=='demodulate') else 'readout ADC'
                timing_order.append((d.__name__, adcString))
            else:
                raise Exception('Can only read out qubits and resonators.')
    p.config_timing_order(timing_order)
    for d in devices:
        channels = dict(d['channels'])
        if 'uwave' in channels:
            if isinstance(d,gc.Readout):
                p.config_microwaves((d.__name__, 'uwave'), d['carrierFrequency'], d['carrierPower'])
            else:
                p.config_microwaves((d.__name__, 'uwave'), d['fc'], d['uwavePower'])
        if 'squid' in channels:
            if 'squidPreampConfig' in d:
                p.config_preamp(d.__name__, *d['squidPreampConfig'])
            if 'squidSwitchIntervals' in d:
                p.config_switch_intervals(d.__name__, d['squidSwitchIntervals'])
        if 'settlingRates' in d:
            p.config_settling((d.__name__,'meas'), d['settlingRates'], d['settlingAmplitudes'])
    if autotrigger is not None:
        p.config_autotrigger(autotrigger)


def addSram(p, devices, ts, blocks, idx=None):
    """Add SRAM data to a Qubit Sequencer packet.
    
    INPUTS
    
    p: packet for the (qubit sequencer) server to which we add the
    commands to add SRAM data.
    
    devices: list of device (qubit) objects that have sequences
    
    ts: list of time ranges. Each range is in the form (tStart, tEnd)
    in nanoseconds.
    
    blocks: SRAM blocks to use for this SRAM
    
    idx: I have no idea.
    
    We input the data for a set of sram blocks.  Currently, at most two
    blocks are supported, which will make use of the split sram feature.
    The first block will be prepadded and the last block postpadded to
    prevent aliasing, or if only one block is given, then that block will
    be padded on both ends.
    
    Note that the sequences are also shifted relative to the time intervals
    to compensate for various delays, as controlled by the 'timingLagUwave'
    and 'timingLagWrtMaster' parameters.  Because of this shift, when using
    dual-block SRAM, you should make sure that both time intervals are
    given with at least enough padding already included to handle any shifts
    due to timing delays.
    """
    sortedDevices = sortDevices(devices)
    phaseQubits = sortedDevices['phaseQubit']
    resonators = sortedDevices['resonator']
    tunableCouplers = sortedDevices['tunableCoupler']
    groups_ADC = sortqubits_ADC(phaseQubits)
    
    for i, (t, block) in enumerate(zip(ts, blocks)):
        #Determine total time span of sequence, including pre and post padding,
        #then compute frequency samples.
        prepad = PREPAD if (i == 0) else 0 # only prepad the first block
        postpad = POSTPAD if (i == len(ts)-1) else 0 # only postpad the last block
        time = prepad + t[1]-t[0] + postpad
        fxy, fz = env.fftFreqs(time)
        p.new_sram_block(block, len(fxy))
        for key in groups_ADC.keys():
            qubits = groups_ADC[key]
            rr = env.NOTHING            
            if len(sortqubits_readoutDAC(qubits)) is not 1:
                raise Exception('DAC setup error: All qubits connected to the same ADC should be connected to the same DAC.')
            for q in groups_ADC[key]:
                r = q['readoutDevice']
                channels = dict(r['channels'])                                
                if 'uwave' in channels:
                    ofs = r['timingLagWrtMaster']
                    t0  = -ofs - prepad
                    rr = rr + q.get('rr', env.NOTHING)  
                    p.sram_iq_data_fourier((r.__name__, 'uwave'), rr(fxy, fourier = True), float(t0))
                    start_delay = round(q['readout DAC start delay']['ns']/4)*us     # this 4 corresponds to 4ns per clock cycle
                    p.set_start_delay((r.__name__, 'uwave'), start_delay)
        for d in phaseQubits+resonators+tunableCouplers:
            channels = dict(d['channels'])
            if 'uwave' in channels:
                ofs = d['timingLagWrtMaster']+d.get('timingLagUwave',0)
                t0 = -ofs - prepad
                # XXX should throw error if t[0] < t0
                xy = d.get('xy', env.NOTHING)
                if idx is not None and xy is not env.NOTHING:
                    xy = xy[idx]
                p.sram_iq_data_fourier((d.__name__, 'uwave'), xy(fxy, fourier=True), float(t0))
            if 'meas' in channels:
                ofs = d['timingLagWrtMaster']
                t0 = -ofs - prepad
                # XXX should throw error if t[0] < t0
                z = d.get('z', env.NOTHING)
                if idx is not None and z is not env.NOTHING:
                    z = z[i]
                # add constant offset in z, if specified
                # FIXME: this might break if PRE- or POSTPAD is 0.  Need to check how such sequences are deconvolved
                if 'zOffsetDC' in d:
                    z += env.rect(t0 + prepad/2, time - prepad/2 - postpad/2, d['zOffsetDC'])
                p.sram_analog_data_fourier((d.__name__,'meas'), z(fz, fourier=True), float(t0))
            if 'trigger' in channels:
                # TODO: support trigger channels
                pass


def addMem(p, devices, blocks):
    """Add memory commands to a Qubit Sequencer packet.
    
    The sequence consists of resetting all qubits then setting
    them to their operating bias.  Next, the SRAM is called,
    then the qubits are read out.  Finally, all DC lines
    are set to zero.
    """
    sortedDevices = sortDevices(devices)
    qubits = sortedDevices['phaseQubit']
    operate_settle = max(d['biasOperateSettling'][us] for d in devices) * us
    
    p.mem_delay(4.3*us)
    #Run through qubit reset procedure
    resetQubits(p, qubits)
    #Go to qubit (and SQUID) operating point
    p.mem_bias([(FLUX(d), FAST, d['biasOperate'][V]*V) for d in devices] +
               [(SQUID(d), FAST, d['squidBias'][V]*V) for d in devices if checkChannel(d, 'squid')], operate_settle)
    #Call SRAM
    p.mem_call_sram(*blocks)
    #Readout
    readoutQubits(p, qubits)
    #Finally, set everything to zero
    p.mem_bias([(FLUX(d), FAST, 0*V) for d in devices] +
               [(SQUID(d), FAST, 0*V) for d in devices if checkChannel(d, 'squid')])


def addADC(p, qubits):
    """ Add ADC configration
    
    """
    for q in qubits:
        if q['adc mode'] == 'demodulate':
            p.adc_set_mode(q.__name__, (q['adc mode'], q['adc channel']))
        else:
            p.adc_set_mode(q.__name__, q['adc mode'])
        if q['adc mode'] == 'demodulate':
            adc_phase = q['adc demod phase']+q['adc adjusted phase']
            p.adc_demod_phase(q.__name__, (q['adc demod frequency'], adc_phase))
            p.adc_set_trig_magnitude(q.__name__, q['adc sinAmp'], q['adc cosAmp'])
            p.config_critical_phase(q.__name__, q['readoutCriticalPhase'])
            p.reverse_critical_phase_comparison(q.__name__, q['readoutCriticalPhaseGreater'])
            p.set_iq_offsets(q.__name__, q['readoutIqOffset'])
            p.adc_set_filter_function(q.__name__, filterBytes(q['adcFilterLen'],q.get('adcFilterWindowStart', 0*ns),q.get('adcFilterWindowEnd', 16384*ns)), q['adcFilterStretchLen'], q['adcFilterStretchAt'])
        start_delay = round(q['adc_start_delay']['ns']/4)*us    # this 4 corresponds to 4ns per clock cycle, and there is a conversion error between ns and us
        p.set_start_delay((q.__name__, 'readout ADC'), start_delay)
            

def resetQubits(p, qubits):
    """Add commands to reset a list of qubits.
    
    The resets are performed simultaneously on all qubits, with
    the reset settling time chosen to be the maximum reset settling
    time for any of the qubits.
    """
    reset_settle = max(q['biasResetSettling'][us] for q in qubits) * us
    for q in qubits:
        if not isinstance(q['biasReset'], list):
            q['biasReset'] = [q['biasReset']]
    reset_steps = max(len(q['biasReset']) for q in qubits)
    for q in qubits:
        shortfall = reset_steps - len(q['biasReset']) #shortfall>=0
        if shortfall:
            q['biasReset'].extend([q['biasOperate']] * shortfall)
    for i in xrange(reset_steps):
        p.mem_bias([(FLUX(q), FAST, q['biasReset'][i][V]*V) for q in qubits], reset_settle)


def readoutQubits(p, qubits):
    """Add commands to readout a list of qubits with SQUID ramps.
    
    The SQUID ramps are performed in groups, based on the value of
    squid_readout_delay for each qubit.
    """
    measure_settle = max(q['biasReadoutSettling'][us] for q in qubits) * us
    squidQubits = [q for q in qubits if not isinstance(q.get('readoutDevice',None), gc.Readout)]
    p.mem_bias([(FLUX(q), FAST, q['biasReadout'][V]*V) for q in qubits] + 
               [(SQUID(q), FAST, 0*V) for q in squidQubits], measure_settle)
    
    if squidQubits:
        # build groups based on readout_delay
        readout = [q for q in squidQubits if q.get('readout', False)]
        no_readout = [q for q in squidQubits if not q.get('readout', False)]
        readout_groups = {}
        for q in readout:
            delay = q['squidReadoutDelay'][us]
            if delay not in readout_groups:
                readout_groups[delay] = []
            readout_groups[delay].append(q)
            
        # readout each qubit group in sequence
        # along with the first group, we run the timers for all qubits
        # that are not actually being read out, just to make sure the timers run
        time = 0
        first = True
        for delay in sorted(readout_groups.keys()):
            group = readout_groups[delay]
            if delay > time:
                p.mem_delay((delay - time)*us)
            ramp_len = max(q['squidRampLength'][us] for q in group)
            time += ramp_len + 10
            if first:
                timers = group + no_readout
                first = False
            else:
                timers = group
            p.mem_start_timer([q.__name__ for q in timers])
            p.mem_bias([(SQUID(q), FAST, q['squidRampBegin'][V]*V) for q in group])
            p.mem_bias([(SQUID(q), SLOW, q['squidRampEnd'][V]*V) for q in group], ramp_len*us)
            p.mem_stop_timer([q.__name__ for q in timers])
            p.mem_bias([(SQUID(q), FAST, q['squidReset'][V]*V) for q in group])
            p.mem_bias([(SQUID(q), FAST, 0*V) for q in group])