'''

experimentRunner.py

This is to be something similar to the Qubit Sequencer, but for simple (i.e. non-qubit) experiments.
Basically, you choose what instruments to use (from LabRAD servers), what measurements to take,
what parameters to scan, and it does it.

'''

import time
import numpy as np

# how to do samples/qubits/devices?
# first you switch session on your cxn.
# then you load your sample[s?] from the registry
# give the same cxn to the experiment to save data to the dv.

DEFAULT_PARAMETERS = {
    'takeDataAlways' :          False,
    'autoAdjust' :              False,
    'dataDelay' :               0.5,
    'pointsPerStep' :           5,
    'name':                     '',
}

class Experiment(object):
    '''
    Experiment class.
    
    Members include:
    cxn -- labrad connection. used for the data vault.
    parameters -- a dict-like object that determines how this experiment goes.
    measureFuncs -- a list of no-arg callables to be called for each data point. must have .label, .legend, .units attributes
    parameterFuncs -- a list of no-arg callables to be called once before the experiment starts.
    scans -- hierarchy of scan objects.
    '''
    def __init__(self, cxn, parameters = None, measureFuncs=None, parameterFuncs=None):
        if not measureFuncs:
            self.measureFuncs = []
        else:
            self.measureFuncs = measureFuncs
        if not parameterFuncs:
            self.parameterFuncs = []
        else:
            self.parameterFuncs = parameterFuncs
        self.parameters = DEFAULT_PARAMETERS
        if parameters:
            self.parameters.update(parameters)
        self.cxn = cxn
        self.scans = []
        
    def addScan(self, scan):
        ''' add a scan or sweep to the bottom of the scan list.'''
        if not isinstance(scan, Scan):
            raise TypeError("%s is not a Scan object" % scan)
        if self.scans and self.scans[-1].isSweep():
            raise ValueError("Cannot append a new scan after a sweep.")
        self.scans.append(scan)
        
    def run(self):
        ''' run the experiment. '''
        # first, set the parameter funcs
        for f in self.parameterFuncs:
            f()
        self.startDataset()
        # the main loop: go through all of our scans.
        self._doScan(self.scans)
        
    def getDatasetName(self):
        ''' returns the name of the dataset for this experiment '''
        if hasattr(self, 'dvName'):
            return self.dvName
        return self.parameters['name']
        
    def startDataset(self):
        ''' create the dataset '''
        # assume we've already cd'd to the correct place
        indeps = [(s.setFunc.label, s.setFunc.unit) for s in self.scans]
        deps = [(f.label, f.legend, f.unit) for f in self.measureFuncs] + [('time', 'timestamp', 's')]
        self.dvPath, self.dvName = self.cxn.data_vault.new(self.getDatasetName(), indeps, deps)
        for k,v in self.parameters.items():
            self.cxn.data_vault.add_parameter(k,v)
        
    def _doScan(self, scanList):
        ''' do the first scan in scanList, recursively doing lower ones. '''
        thisScan = scanList[0]
        scanList = scanList[1:]
        thisScan.reset()
        nextValue = thisScan.goToNextValue()
        while nextValue is not None:
            # two possibilities: we have further subscans, or this is the bottom
            if scanList:
                # we have subscans: wait for the next value, then do subordinate scans
                while not thisScan.atNextValue():
                    time.sleep(0.5)
                self._doScan(scanList)
                nextValue = thisScan.goToNextValue()
            else:
                # this is bottom: we are responsible for data taking
                while not thisScan.atNextValue():
                    # do we take data while moving?
                    if thisScan.isSweep() or self.parameters['takeDataAlways']:
                        self._takePoint()
                    time.sleep(self.parameters['dataDelay'])
                # do we take data when we get there?
                if not thisScan.isSweep():
                    time.sleep(self.parameters['dataDelay'])
                    for i in range(self.parameters['pointsPerStep']):
                        self._takePoint()
                        time.sleep(self.parameters['dataDelay'])
                # go to next value
                nextValue = thisScan.goToNextValue()
    def _takePoint(self):
        ''' take and save a single data point. '''
        data = [s.getFunc() for s in self.scans] + [f() for f in self.measureFuncs] + [time.time()]
        self.cxn.data_vault.add(data)
        
class Scan(object):
    ''' Represents a single scan or sweep of a parameter.
    setFunc is the callable to set the parameter, taking one value--a float. must have .label and .unit attributes.
    atFunc is a one-parameter (float) function that returns True if we're at the given value (and ready to go--settled, persistent, etc), else false
    start and stop are the values to run the sweep/scan from and to.
    If step=None, it is "sweep", meaning a continuous change of the parameter (not stepped).
    (Take data during the sweep, as opposed to at each point in a scan.)
    Only one parameter can be swept in an experiment. (Others must be scanned.)
    If switchDirection==True, then each repeat of this sweep/scan goes in the reverse order.
    (Scans are repeated when nested; e.g. scan parameter A, and at each step, sweep parameter B.
    Switching the direction of parameter B could save time if the sweep is long.)
    '''
    def __init__(self, setFunc, getFunc, atFunc, start, stop, step=None, switchDirection=True):
        self.setFunc = setFunc
        self.getFunc = getFunc
        self.atFunc = atFunc
        self.start = start
        self.stop = stop
        self.step = step
        self.switchDirection = switchDirection
        if self.step:
            self.values = list(np.arange(start, stop, step)) + [stop]
        else:
            self.values = [start, stop]
        self.originalValues = self.values
        self.nextValue = None
        
    def atNextValue(self):
        ''' Have we finished going to the next value in our sweep or scan? '''
        return (self.nextValue is None) or (self.atFunc(self.nextValue))
        
    def goToNextValue(self):
        ''' Start going to the next value in our sweep/scan. Returns that value, or None if there are no more values to go to. '''
        if not self.values:
            self.nextValue = None
        else:
            self.nextValue = self.values[0]
            self.values = self.values[1:]
            self.setFunc(self.nextValue)
        return self.nextValue
        
    def reset(self):
        ''' reset the scan/sweep for another run. '''
        if self.values == self.originalValues:
            return
        self.values = self.originalValues
        if self.switchDirection:
            self.values.reverse()   # note that this reverses both values and originalValues
            
    def isSweep(self):
        return self.step is None
        
        
# now we have some tester stuff
class ERTester(object):
    def __init__(self):
        self.paramA = 1
        self.paramB = 1
    
    def getValue1(self):
        return self.paramA * self.paramB
    getValue1.label, getValue1.legend, getValue1.unit = 'val', 'v1', 'A'
    def getValue2(self):
        return self.paramA**2 / self.paramB
    getValue2.label, getValue2.legend, getValue2.unit = 'val', 'v2', 'A'
    
    def setParamA(self, a):
        self.paramA = a
    setParamA.label, setParamA.unit = 'pA', 'V'
    def setParamB(self, b):
        self.paramB = b
    setParamB.label, setParamB.unit = 'pB', 'V'
    def getParamA(self):
        return self.paramA
    def getParamB(self):
        return self.paramB
        
    def cmpPA(self, v):
        return self.paramA == v
    def cmpPB(self, v):
        return self.paramB == v
        
def runERTest(cxn):
    ert = ERTester()
    ex = Experiment(cxn, measureFuncs=[ert.getValue1, ert.getValue2], parameters={'dataDelay':0.01})
    scanA = Scan(ert.setParamA, ert.getParamA, ert.cmpPA, -1.0, 3.0, 0.1)
    scanB = Scan(ert.setParamB, ert.getParamB, ert.cmpPB, 21.0, 3.0, -1.0)
    ex.addScan(scanA)
    ex.addScan(scanB)
    ex.run()