#!/usr/bin/python
"""
def template(Sample, measure, name=<default name>,
             <scan ranges with defaults>,
             <sequence options>,
             <other data related options like stats>,
             save=False, collect=False, noisy=True,
             <scripting options like plots>):
    pass
"""

#Standard packages
import numpy as np
import matplotlib.pyplot as plt
import scipy.optimize
import scipy.signal
#LabRAD
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
import pyle.datavault as datavault
import pyle.envelopes as env
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
import pyle.util.structures as structures
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.dataking.adjust as adjust
import pyle.analysis.iq as iq
#named functions
ValueArray = structures.ValueArray
from pyle.datavault import DataVaultWrapper
from pyle.dataking.fpgaseqTransmon import runQubits, runQubitsLocalDemod
from pyle.util import sweeptools as st
from pyle.pipeline import returnValue, FutureList
from pyle.dataking import sweeps
from pyle.dataking import util
import pyle.dataking.singleQubit as sq
import pyle.dataking.readout as readout
from pyle.registry import AttrDict
import logging
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
    
def iqCorrelation(Sample, measure, name='IQ correlation (ground state)',
                  delay=None, repetitions=1,
                  freq=None, power=None,
                  stats=30, dataFormat='iq',
                  save=True, collect=False, noisy=False):
    log.warn("WARNING - This routine MUST be run with pipesize=1")
    log.warn("          due to bad interactions between runQubitsLocalDemod")
    log.warn("          and the pipeline. Complain to Dan or Evan.")
    x = util.loadQubits(Sample, write_access=False)
    sample,qubits = x
    q = qubits[measure]
    q['readout'] = True  
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    q = devs[measure]
    
    if delay is None:
        delay = st.r[250:2000:100,ns]
    if freq is None:
        freq = q['readoutFrequency']
    if power is None:
        power = q['readoutPower']
    
    sweep = st.repeatedScan(delay, repetitions)
    
    axes = [(sweep(), 'delay')]
    deps = [('Detection amplitude', 'I first', 'clicks'), ('Detection amplitude', 'Q first', 'clicks'),
            ('Detection amplitude', 'I second', 'clicks'), ('Detection amplitude', 'Q second', 'clicks')]
    kw = {'stats':stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)

    def func(server, d):
        q['adcReadoutWindows'] = [(200*ns,450*ns),(200*ns+d,450*ns+d)]
        #Measure with no pi pulse
        alg = gc.Algorithm()
        alg[gates.Measure(devs, delay=0*ns)]
        alg[gates.Measure(devs, delay=d)]
        alg.compile()
        result = yield runQubitsLocalDemod(server, devs, stats, dataFormat=dataFormat)
        #Run the experiment
        rv = np.vstack(result[0]).T
        returnValue(rv)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=noisy, pipesize=1)        

    if collect:
        return data
        
def fidelityVsReadoutTPostSelect(Sample, measure, name='fidelity Vs Readout T PS',
                  readoutT=None, delay=None, repetitions=1,
                  stats=30, dataFormat='iqRaw',
                  save=True, collect=False, noisy=False):
    log.info("WARNING - This routine MUST be run with pipesize=1")
    log.info("          due to bad interactions between runQubitsLocalDemod")
    log.info("          and the pipeline. Complain to Dan or Evan.")
    sample, devs, q = gc.loadQubits(Sample, measure, write_access=False)

    if readoutT is None:
        readoutT = st.r[30:200:10,ns]
    if delay is None:
        delay = 350*ns
        
    sweep = st.repeatedScan(readoutT, repetitions)
    axes = [(sweep(), 'readoutT'), ([0,1], 'state')]        
    deps = [('Detection amplitude', 'I first', 'clicks'), ('Detection amplitude', 'Q first', 'clicks'),
            ('Detection amplitude', 'I second', 'clicks'), ('Detection amplitude', 'Q second', 'clicks')]
    kw = {'stats':stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, t, state):
        #Measure without a pi pulse
        alg = gc.Algorithm(devs)
        alg.q['adcReadoutWindows'] = [(250*ns,250*ns+t),(250*ns+delay+t,250*ns+delay+2*t)]
        alg.q['readoutLen'] = t
        if state:
            alg[gates.PiPulse([alg.q])]
        alg[gates.Measure(alg.agents, delay=0*ns)]
        alg[gates.Measure(alg.agents, delay=(delay+t))]
        alg.compile()
        result = yield runQubitsLocalDemod(server, alg.agents, stats, dataFormat=dataFormat)
        rv = np.vstack(result[0]).T
        returnValue(rv)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=noisy, pipesize=1)        

    if collect:
        return data

def stimulated_emission(Sample, measure, name="Stimulated Emission",
        readout_power = st.r[-60:-30:3, dBm], power_reps=1, stats=150, 
        excitationLen=None, pre_emph=False, post_select=False, readoutDelay=None, zAmps = None,
        save=True, collect=False, noisy=False):
        
    sample, devs, q = gc.loadQubits(Sample, measure, write_access=False)

    if excitationLen is None:
        excitationLen = q['readoutLen']
    if readoutDelay is None:
        readoutDelay = 200*ns
    if zAmps is None:
        zAmps = 0.0
        
    axes = [(st.repeatedScan(readout_power, power_reps)(), 'Readout Power'), 
            (readoutDelay, 'Delay'), 
            (excitationLen, 'Excitation Length'),
            (zAmps, 'Z')]
    deps = [('Probability', 'No Pulse', 'probability'), ('Probability', 'w/ Pulse', 'probability')]
    kw = {'stats':stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, pow, delay, pulse_len, z):
        df = q['readoutFrequency']-q['readoutDevice']['carrierFrequency']
        excitationAmp = eh.power2amp(pow)

        alg1 = gc.Algorithm(agents=devs)
        alg1[gates.Wait([alg1.q], 20*ns)]
        alg1[gates.Detune([alg1.q], length=pulse_len, amp=z)]
        alg1[gates.Wait([alg1.q], delay)]
        alg1[gates.Measure([alg1.q])] 
        alg1.compile()
        alg1.q['rr'] += env.mix(env.flattop(-pulse_len['ns']-delay['ns'], pulse_len['ns'], q['readoutWidth'], amp=excitationAmp), df)
        if pre_emph:
            alg1.q['rr'] += env.mix(env.flattop(-pulse_len['ns']-delay['ns'], alg1.q['preEmphLen'], alg1.q['readoutWidth'], amp=excitationAmp*(alg1.q['preEmphFactor']-1)), df)
        if post_select:
            readout.add_herald(alg1, 500*ns)
            deferNoPulse = runQubits(server, alg1.agents, stats, dataFormat='iqRaw', localDemod=True)
            deferNoPulse.addCallback(readout.do_post_select, alg1)
        else:
            deferNoPulse = runQubits(server, alg1.agents, stats, dataFormat='iqRaw')

        alg2 = gc.Algorithm(agents=devs)
        alg2[gates.PiPulse([alg2.q])]
        alg2[gates.Wait([alg2.q], 20*ns)]
        alg2[gates.Detune([alg2.q], length=pulse_len, amp=z)]
        alg2[gates.Wait([alg2.q], delay)]
        alg2[gates.Measure([alg2.q])] 
        alg2.compile()
        alg2.q['rr'] += env.mix(env.flattop(-pulse_len['ns']-delay['ns'], pulse_len['ns'], alg2.q['readoutWidth'], amp=excitationAmp), df)
        if pre_emph:
            alg2.q['rr'] += env.mix(env.flattop(-pulse_len['ns']-delay['ns'], alg2.q['preEmphLen'], alg2.q['readoutWidth'], amp=excitationAmp*(alg2.q['preEmphFactor']-1)), df)
        if post_select:
            readout.add_herald(alg2, 500*ns)
            deferPulse = runQubits(server, alg2.agents, stats, dataFormat='iqRaw', localDemod=True)
            deferPulse.addCallback(readout.do_post_select, alg2)
        else:
            deferPulse = runQubits(server, alg2.agents, stats, dataFormat='iqRaw')

        resultNoPulse, resultPulse = yield FutureList([deferNoPulse, deferPulse])
        Is,Qs = resultNoPulse[0]
        probNoPulse = sq.readoutProbabilities(q, np.array(Is)+1.0j*np.array(Qs))

        Is,Qs = resultPulse[0]
        probPulse = sq.readoutProbabilities(q, np.array(Is)+1.0j*np.array(Qs))

        returnValue([np.mean(probNoPulse), np.mean(probPulse)])
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=noisy, pipesize=10)        

    if collect:
        return data
    
def analyze_fidelity_vs_T(ds, plot=True):
    mask0 = ds['state'] == 0
    mask1 = ds['state'] == 1
    readoutT = set(ds['readoutT'])
    sorted_t = np.sort(list(readoutT))
    
    iq1 = ds['I first'] + 1j*ds['Q first']
    iq2 = ds['I second'] + 1j*ds['Q second']
    result = []
    success = 0
    total = 0
    for t in sorted_t:
        mask_t = ds['readoutT'] == t
        iq_nopulse = iq1[mask0*mask_t][:6500]
        iq_pulse = iq1[mask1*mask_t][:6500]
        center, axis, _, _ = iq.discriminationAxis(iq_nopulse, iq_pulse)
        param = (center, axis)
        log.info("discrimination parameters: (%s, %s)", *param)
        mask_nopulse = iq.discriminateState(iq_nopulse, param) == 0
        mask_pulse = iq.discriminateState(iq_pulse, param) == 1
        log.info("post selection : %s, %s / %s", sum(mask_nopulse), sum(mask_pulse), len(mask_nopulse))
        success += (sum(mask_nopulse) + sum(mask_pulse))
        total += len(iq_nopulse) + len(iq_pulse)
        err0 = np.sum(iq.discriminateState(iq2[mask0*mask_t],param)[mask_nopulse])/ (1.0*sum(mask_nopulse))
        err1 = 1.0 - np.sum(iq.discriminateState(iq2[mask1*mask_t],param)[mask_pulse])/ (1.0*sum(mask_pulse))
        result.append([t, err0, err1])
    log.info("total post-selection success: %s / %s = %s", success, total, 1.0*success/total)
    log.info("Hot qubit population: %s ", 1.0 - 1.0*success/total)
    result = np.array(result)
    if plot:
        plt.figure()
        plt.plot(result[:,0], result[:,1], label='zero state')
        plt.plot(result[:,0], result[:,2], label='one state')
        t = result[:,0]
        pdecay = 1-np.exp(-(t+200)/10000)
        plt.plot(result[:,0], pdecay, 'r-', label="T1 = 10 us")
        plt.xlabel('Measurement Time (ns)')
        plt.ylabel('Detection Error Probabilty')
        plt.legend()
    return result

def histogram_post_selected(ds, t=150.0, bins=50):
    mask0 = ds['state'] == 0
    mask1 = ds['state'] == 1
    readoutT = set(ds['readoutT'])
    sorted_t = np.sort(list(readoutT))
    
    iq1 = ds['I first'] + 1j*ds['Q first']
    iq2 = ds['I second'] + 1j*ds['Q second']

    mask_t = ds['readoutT'] == t
    iq_nopulse = iq1[mask0*mask_t][:6500]
    iq_pulse = iq1[mask1*mask_t][:6500]
    center, axis, _, _ = iq.discriminationAxis(iq_nopulse, iq_pulse)
    param = (center, axis)
    log.info("discrimination parameters: (%s, %s)", *param)
    mask_nopulse = iq.discriminateState(iq_nopulse, param) == 0
    mask_pulse = iq.discriminateState(iq_pulse, param) == 1
    log.info("post selection : %s, %s / %s", sum(mask_nopulse), sum(mask_pulse), len(mask_nopulse))

    x_no_pulse = ((iq2[mask0*mask_t] - center ) * axis.conjugate()).real
    x_pulse = ((iq2[mask1*mask_t] - center ) * axis.conjugate()).real
    fig = plt.figure()
    ax = fig.add_subplot(111)
    hist_count, hist_bins, _ = plt.hist([x_no_pulse,x_pulse], bins=bins, label=['No Pulse', 'Pi Pulse'])
    plt.legend()
    iq.analyze_readout_hist(hist_count, hist_bins)
    return ax
    
def resonator_Q(Sample, measure, name='Resonator Q ring-down', 
                power=None, averages=1, stats=1,
                timeMax_ns=17000, update=True):
    """
    This excites the readout resonator and measures the IQ vs time, which can then be used
    to fit an exponential decay and calculate the resonator Q.

    The excitation pulse is at the resonant frequencies of the |0> and |1> state resonators,
    not the readout frequency.
    """

    GCobjects = gc.loadQubits(Sample, measure, write_access=True)
    sample, devs, q, Q = GCobjects
    if power is not None:
        q['readoutPower'] = power
    
    indeps = [('Time','ns')]
    deps=[('Amplitude','I no pulse',''),
          ('Amplitude','Q no pulse',''),
          ('Amplitude','I with pulse',''),
          ('Amplitude','Q with pulse','')]
    kw = {'stats':stats, 'averages':averages}
    dataset = sweeps.prepDataset(sample, name, axes=[], explicitIndependents = indeps, dependents=deps, measure=measure, kw=kw, cxn=Sample._cxn)  
    
    fNoPulse = q['readoutResonatorFreq0']
    fPulse = q['readoutResonatorFreq1']
    INoPulse,QNoPulse = sq.readoutTimeTrace(GCobjects, None,
                                         freq=fNoPulse, power=power,
                                         piPulse=False,
                                         averages=averages, stats=stats,
                                         collect=True,
                                         plot=False)
    IPulse,QPulse = sq.readoutTimeTrace(GCobjects, None,
                                     freq=fPulse, power=power,
                                     piPulse=True,
                                     averages=averages, stats=stats,
                                     collect=True,
                                     plot=False)
    t = np.linspace(0,(8192*2),8192, endpoint=False)
    tMask = t<timeMax_ns
    data = np.vstack((
                      t[tMask],
                      INoPulse[tMask],
                      QNoPulse[tMask],
                      IPulse[tMask],
                      QPulse[tMask])).T
    with dataset:
        dataset.add(data)
    ds = datavault.LocalDataSet(data, dataset.independents, dataset.dependents, dataset.params, dataset.name)
    if update:
        analyze_resonator_q(ds, timeMax_ns)
    return ds
   
def exp_fit(t, amp, tau, offset):
    return amp * np.exp(-1.0*t/tau) + offset

def analyze_resonator_q(ds, tmax_ns = None):
    t = ds['Time']
    if tmax_ns is not None:
        mask = t < tmax_ns
        ds = ds[mask, :]
        t = ds['Time']
    zNoPulse = ds['I no pulse'] + 1j*ds['Q no pulse']
    zPulse = ds['I with pulse'] + 1j*ds['Q with pulse']

    qubitName = ds.parameters['measure'][0]
    q = ds.parameters[qubitName]
    readout = ds.parameters[q['readoutConfig']]

    freq0 = q['readoutResonatorFreq0'] - readout['carrierFrequency']
    freq1 = q['readoutResonatorFreq1'] - readout['carrierFrequency']
    z0 = zNoPulse * np.exp(-1j*(2*np.pi*t*freq0))
    z1 = zPulse * np.exp(-1j*(2*np.pi*t*freq1))
    z0 = np.convolve(z0, scipy.signal.gaussian(21, 5), 'same') # Gaussian filters have the minimum impact on rise-time and
    z1 = np.convolve(z1, scipy.signal.gaussian(21, 5), 'same') # excellent attenuation of far-off-resonant signals (idler, DC)
    
    idx =np.argmax(np.abs(z0))
    phi0 = np.angle(z0[idx])
    z0 = z0 * np.exp(-1j*phi0)
    
    idx =np.argmax(np.abs(z1))
    phi1 = np.angle(z1[idx])
    z1 = z1 * np.exp(-1j*phi1)
    
    x0 = z0.real
    x1 = z1.real
    plt.figure()
    plt.plot(t, z0.real)
    plt.plot(t, z0.imag)
    plt.plot(t, np.abs(z0))
    plt.title('Zero state signal')
    plt.figure()
    plt.plot(t, z1.real)
    plt.plot(t, z1.imag)
    plt.plot(t, np.abs(z1))
    plt.title('One state signal')
    traces = [{'x': t, 'y': x0, 'args': ('b.',)},
              {'x': t, 'y': x1,   'args': ('r.',)}]
    params = [{'name': 'start', 'val': min(t), 'range': (min(t), max(t)), 'axis': 'x', 'color': 'b'},
              {'name': 'stop' , 'val': max(t), 'range': (min(t), max(t)), 'axis': 'x', 'color': 'r'},]

    result = adjust.adjust(params, traces)
    #result = None
    if result:
        start_idx = np.nonzero(t > result['start'])[0][0]
        log.info("start idx: %d", start_idx)
        stop_idx = np.nonzero(t < result['stop'])[0][-1]
        log.info("stop idx: %d", stop_idx)
        tau_est = (result['stop'] - result['start'])/3.0
        t_fit = t[start_idx:stop_idx] - result['start']

        fit0, _ = scipy.optimize.curve_fit(exp_fit, t_fit, x0[start_idx:stop_idx], (x0[start_idx], tau_est, x0[stop_idx]))
        fit1, _ = scipy.optimize.curve_fit(exp_fit, t_fit, x1[start_idx:stop_idx], (x1[start_idx], tau_est, x1[stop_idx]))
        log.warn('fit0: %s', fit0)
        log.warn('fit1: %s', fit1)
        plt.figure()
        plt.plot(t, x0, 'b.')
        plt.plot(t_fit+result['start'], exp_fit(t_fit, *fit0), 'b-')
        plt.plot(t, x1, 'r.')
        plt.plot(t_fit+result['start'], exp_fit(t_fit, *fit1), 'r-')
        print "Time constant 0 state: %f ns / Q= %f" % (fit0[1], np.pi*q['readoutResonatorFreq0']['GHz'] * fit0[1])
        print "Time constant 1 state: %f ns / Q= %f" % (fit1[1], np.pi*q['readoutResonatorFreq1']['GHz'] * fit1[1])

def two_qubit_readout(Sample, measure, name='Two qubit readout', stats=900, plot=True,
            save=True):
    sample, devs, q = gc.loadQubits(Sample, measure, write_access=False)

    axes = [(range(2**len(measure)), 'State')]
    deps = [[('Amplitude', 'I |%s>' %(m,),'clicks'), ('Amplitude', "Q |%s>" % (m,), 'clicks')] for m in measure]
    deps = sum(deps, []) # concatenate lists.
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, state):
        alg = gc.Algorithm(agents=devs)
        for idx, qubit in enumerate(reversed(alg.q)):
            if (state>>idx) & 1:
                alg[gates.PiPulse([qubit])]
        alg[gates.Measure(alg.q)]
        alg.compile()
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        result = []
        for idx, qubit in enumerate(alg.q):
            Is, Qs = data[idx]
            result.extend([Is, Qs])
        result = 1.0*np.column_stack(result)
        returnValue(result)
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    if plot:
        plot_two_qubit_readout(data)
        
def plot_two_qubit_readout(data):
    qconfig = data.parameters['config']
    q = []
    for qname in qconfig:
        qubit = AttrDict(data.parameters[qname])
        qubit.__name__ = qname
        q.append(qubit)
    for idx, qubit in enumerate(q):
        plt.figure()
        plt.xlabel('I (clicks)')
        plt.ylabel('Q (clicks)')
        plt.title('Readout for |%s>' % (qubit.__name__,))
        for state in range(2**len(q)):
            mask = data['State']==state
            I = data['I |%s>' % (qubit.__name__,)][mask]
            Q = data['Q |%s>' % (qubit.__name__,)][mask]
            Z = I+1j*Q
            plt.plot(I, Q, label="state %d" % (state,))
        z_center = qubit['readoutCenterZ']
        z_axis = qubit['readoutAxis']
        xrange = np.linspace(min(I), max(I), 100)
        pts = z_center + z_axis*1j*np.linspace(-np.max(np.abs(Z)), np.max(np.abs(Z)))
        plt.plot(pts.real, pts.imag, 'k-', label="Axis")
        plt.legend()
    plt.show()
                    
def two_qubit_readout_heralding(Sample, measure, name='Two qubit readout Post-select', stats=900, plot=True,
            save=True):
    sample, devs, q = gc.loadQubits(Sample, measure, write_access=False)

    axes = [(range(2**len(measure)), 'State')]
    deps = [[('Amplitude', 'I |%s>' %(m,),'clicks'), ('Amplitude', "Q |%s>" % (m,), 'clicks')] for m in measure]
    deps = sum(deps, []) # concatenate lists.
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, state):
        alg = gc.Algorithm(agents=devs)
        for idx, qubit in enumerate(reversed(alg.q)):
            if state & 2**idx:
                alg[gates.PiPulse([qubit])]
        alg[gates.Measure(alg.q)]
        alg.compile()
        readout.add_herald(alg, 1000*ns)
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', localDemod=True)
        data = readout.do_post_select(data, alg)
        result = []
        for idx, qubit in enumerate(alg.q):
            Is, Qs = data[idx]
            result.extend([Is, Qs])
        result = 1.0*np.column_stack(result)
        returnValue(result)
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    if plot:
        plot_two_qubit_readout(data)
