import logging
from multiprocessing import Process

import labrad
from labrad.units import Unit,Value
from pyle import gateCompiler as gc
mK = Unit('mK')

confDict = {'Peter': 'U:/Peter/labrad/dev/pyle/pyle/dataking/confPeter.py',
            'Julian': 'U:/Julian/dataking/init_julian.py'}

def loadDeviceType(sample ,deviceType, write_access=False):
    Devices=[]
    devices=[]
    deviceNames = sample['config']
    #First get writeable Devices
    for deviceName in deviceNames:
        if sample[deviceName]['_type'] == deviceType:
            Devices.append(sample[deviceName])
    #Now make the unwritable devices
    sample = sample.copy()
    for deviceName in deviceNames:
        if sample[deviceName]['_type'] == deviceType:
            devices.append(sample[deviceName])
    if write_access:
        return sample, devices, Devices
    else:
        return sample, devices

def loadQubits(sample, write_access=False):
    """Get local copies of the sample configuration stored in the registry.
    
    Returns the local sample config, and also extracts the individual
    qubit configurations, as specified by the sample['config'] list.  If
    write_access is True, also returns the qubit registry wrappers themselves,
    so that updates can be saved back into the registry.
    """
    Qubits = [sample[q] for q in sample['config']]  #RegistryWrappers
    sample = sample.copy()                          #AttrDict
    qubits = [sample[q] for q in sample['config']]  #AttrDicts
    
    # only return original qubit objects if requested
    if write_access:
        return sample, qubits, Qubits
    else:
        return sample, qubits
    
def datasetToReg(sample,dataset,qubitsToLoad):
    """ Load Registry values from a dataset. Target qubit must be in config"""
    sample, qubits, Qubits = loadQubits(sample, write_access=True)
    
    QubitDict = {}
    for Qubit in Qubits:
        QubitName = Qubit._dir[-1]
        QubitDict[QubitName]=Qubit
    
    for qubit in qubitsToLoad:
        for key in dataset['parameters'][qubit].keys():
            if key != 'readout':
                QubitDict[qubit][key] = dataset['parameters'][qubit][key]
                
            
def dvwsetToReg(sample, dvwData, qubitsToLoad):
    '''
    syntax: dvwsetToReg(s,dvw[...],['q0','q1'])
    '''
    sample, qubits, Qubits = loadQubits(sample, write_access=True)
    
    QubitDict = {}
    for Qubit in Qubits:
        QubitName = Qubit._dir[-1]
        QubitDict[QubitName]=Qubit
    
    for meas, Qubit in zip(qubitsToLoad, Qubits):
        for key in dvwData.parameters[meas].keys():
            if key != 'readout':
                QubitDict[meas][key] = dvwData.parameters[meas][key]
                
def channels2backup(Sample, measure):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for q,Q in zip(qubits,Qubits):
        origChannels = q['channels']
        Q['channelsBak'] = origChannels
        Q['channels'] = [origChannels[0]]
    
def backup2channels(Sample, measure):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for q,Q in zip(qubits,Qubits):  
        realChannels = q['channelsBak']
        Q['channels'] = realChannels

def loadDevices(sample, write_access=False, configKey='config'):
    """Get local copies of the sample configuration stored in the registry.

    Returns the local sample config, and also extracts the individual
    device configurations, as specified by the sample['config'] list.  If
    write_access is True, also returns the qubit registry wrappers themselves,
    so that updates can be saved back into the registry.
    """
    devices={}
    Devices={}
    if configKey is None:
        configKey = 'config'
    #The order of these lines is important, as we want devices to be assigned
    #after we make a copy of sample, whereas Devices is not a copy.
    for d in sample[configKey]:
        Devices[d]=sample[d]
    sample = sample.copy()
    for d in sample[configKey]:
        devices[d]=sample[d]

    if write_access:
        return sample, devices, Devices
    else:
        return sample, devices

def dcZero():
    for id in [2, 3, 5, 6]:
        board = 'DR Lab FastBias %d' % id
        for chan in ['A', 'B', 'C', 'D']:
            dcVoltage(board, chan, 0)


def dcVoltage(board, chan, voltage):
    with labrad.connect() as cxn:
        channels = [('b', ('FastBias', [board, chan]))]
        p = cxn.qubit_sequencer.packet()
        p.initialize([('dev', channels)])
        p.mem_start_timer()
        p.mem_bias([('b', 'dac1', voltage)])
        p.mem_stop_timer()
        p.build_sequence()
        p.run(30)
        p.send()


def getMixTemp(cxn, sample, measure=0):
    sample, qubits, Qubits = loadQubits(sample, write_access=True)
    q, Q = qubits[measure], Qubits[measure]
    device_name = q['lakeshoreName']
    lr = cxn.lakeshore_ruox
    lr.select_device(device_name)
    temp = lr.temperatures()[0][0][mK]
    Q['temperature'] = temp
    return temp

def _multi_wrapper(conf, funcName, measure, **kwargs):
    with open(conf) as f:
        exec(f)
    exec('import %s' % func.rpartition('.')[0])
    exec('func = %s' % funcName)
    func(s, measure, **kwargs)

def repeatOverSettings(sample, measure, dvw, func, multiprocess=True, dsList = None, **kwargs):
    ''' repeat a given scan multiple times with different registry settings,
    optionally with each in a new process (to work around memory leaks).
    
    The registry settings are loaded from the starred datasets found in dvw, or a list of indices
    passed in as the argument dsList.
    func is the function to be repeated. must take sample, measure, and any kwargs as arguments.
    sample, measure and kwargs are passed through to the function.
    
    Note that you must have an entry in confDict for this to work, as your conf
    will be run in each new python process. The key is read from the second entry
    of your sample (i.e. your registry root). '''
    logging.basicConfig(filename='qubitLog.txt', filemode='w', level=logging.DEBUG)
    conf = confDict[sample._dir[1]]
    if dsList is None:
        dsList = dvw.keys(['star'])
    logging.info("Starting repeatOverSettings, doing %s datasets" % len(dsList))
    if callable(func):
        oldfunc = func
        func = func.__module__ + '.' + func.__name__
        logging.info("Converted function '%s' into string '%s'" % (str(oldfunc), func))
    for star in dsList:
        #try:
            dvwsetToReg(sample, dvw[star], [measure])
            logging.info("now starting: f10 = %s" % sample[measure]['f10'])
            if multiprocess:
                logging.info("spawning new python process...")
                p = Process(target=_multi_wrapper, args=(conf, func, measure), kwargs=kwargs)
                p.start()
                p.join()
            else:
                exec('import %s' % func.rpartition('.')[0])
                exec('f2 = %s' % func)
                f2(sample, measure, **kwargs)
        #except Exception as e:
        #    print "Exception: %s" % str(e)
        #    logging.error(str(e))
    
    
