from __future__ import division
import numpy as np

# some helper functions for dealing with matrices and computing fidelity

def tensor(matrices):
    """Compute the tensor product of a list (or array) of matrices"""
    return reduce(np.kron, matrices)

def dots(matrices):
    """Compute the dot product of a list (or array) of matrices"""
    return reduce(np.dot, matrices)

def dot3(A, B, C):
    """Compute the dot product of three matrices"""
    return np.dot(np.dot(A, B), C)

def normalize(X):
    """Normalize the representation of X.  This means:
    For state vectors, X' * X = 1 and the first non-zero
    element of X is positive real.  For density matricies,
    X is hermetian, unit trace, and non-negative"""
    if X.ndim == 1 :
        nz = X[np.nonzero(X)][0]
        X *= nz.conj()
        amp = np.abs(np.dot(X.conj().T, X))
        X /= np.sqrt(amp)
    else:
        (d, U) = np.linalg.eigh(X)
        d = d.real
        d[np.nonzero(d<0)] = 0
        X = np.dot(np.dot(U, np.diag(d)), U.conj().T)
        X = (X.conj().T + X)
        X = X / np.trace(X)
    return X

def commutator(A, B):
    """Compute the commutator of two matrices"""
    return np.dot(A, B) - np.dot(B, A)
    

def dagger(vector):
    """Compute the hermitian conjugate of a vector or matrix"""
    return vector.conjugate().transpose()


def lindblad(rho, L, Ld=None):
    """Compute the contribution of one Lindblad term to the master equation""" 
    if Ld is None:
        Ld = L.conj().T
    return dot3(L, rho, Ld) - 0.5*dot3(Ld, L, rho) - 0.5*dot3(rho, Ld, L)


def ket2rho(ket):
    """Convert a state (ket) to a density matrix (rho)."""
    return np.outer(ket, ket.conj())


def trace_distance(rho, sigma):
    """Compute the trace distance between matrices rho and sigma
    See Nielsen and Chuang, p. 403
    """
    A = rho - sigma
    abs = sqrtm(np.dot(A.conj().T, A))
    return np.real(np.trace(abs)) / 2.0


def fidelity(rho, sigma):
    """Compute the fidelity between matrices rho and sigma
    See Nielsen and Chuang, p. 409
    """
    rhosqrt = sp.linalg.sqrtm(rho)
    return  np.trace(sp.linalg.sqrtm(dots((rhosqrt, sigma, rhosqrt))).real)

def overlap(rho, sigma):
    """Trace of the product of two matrices."""
    # XXX this is only correct if at least one of rhos or sigma is a pure state
    return np.trace(np.dot(rho, sigma))


def concurrence(rho):
    """Concurrence of a two-qubit density matrix.
    see http://qwiki.stanford.edu/wiki/Entanglement_of_Formation
    """
    yy = np.array([[ 0, 0, 0,-1],
                   [ 0, 0, 1, 0],
                   [ 0, 1, 0, 0],
                   [-1, 0, 0, 0]], dtype=complex)
    m = dots([rho, yy, rho.conj(), yy])
    eigs = [np.abs(e) for e in np.linalg.eig(m)[0]]
    e = [np.sqrt(x) for x in sorted(eigs, reverse=True)]
    return max(0, e[0] - e[1] - e[2] - e[3])


def eof(rho):
    """Entanglement of formation of a two-qubit density matrix.
    see http://qwiki.stanford.edu/wiki/Entanglement_of_Formation
    """
    def h(x):
        if x <= 0 or x >= 1:
            return 0
        return -x*np.log2(x) - (1-x)*np.log2(1-x)
    C = concurrence(rho)
    arg = max(0, np.sqrt(1-C**2))
    return h((1 + arg)/2.0)

def random_unitary(N):
    """Random unitary matrix of dimension N according to the Haar metric"""
    H = np.diag(np.complex128(np.random.randn(N)))
    for j in range(N):
        for k in range(j+1,N):
            H[j,k] = (np.random.randn() + 1j*np.random.randn())/np.sqrt(2)
            H[k,j] = H[j,k].conj()
    (d,U) = np.linalg.eigh(H)
    for j in range(N):
        U[:,j] = U[:,j] * np.exp(-1j*2*np.pi*np.random.rand())
    return U

def random_density(N, rank=None):
    """Generates a random density matrix.  The distribution does not
    necessarily mean anything.  N is the dimension and rank is the
    number of non-zero eigenvalues.  Alternately, N can be a vector
    of eigenvalues which will be transformed by a Haar-distributed
    random unitary."""
    
    N = np.atleast_1d(N)
    if N.size == 1:
        if rank is None: rank = N[0]
        d = np.zeros(N, dtype=np.complex128)
        d[0:rank] = np.random.rand(rank)
    else:
        d = N
    U = random_unitary(d.shape[0])
    tmp = np.dot(U, np.diag(d))
    tmp = np.dot(tmp, U.conj().T)
    return normalize(tmp)

def random_state(N):
    """Generates a random pure state."""
    U = random_unitary(N)
    psi = U[:,1]
    return psi


