from __future__ import division

import numpy as np
import matplotlib.pyplot as pyplot
import mpl_toolkits.mplot3d
import pyquantum

s_ = {
    "H": np.array([1, 0j],  dtype=np.complex),
    "V": np.array([0, 1],   dtype=np.complex),
    "D": np.array([1, 1],   dtype=np.complex)/np.sqrt(2),
    "A": np.array([1, -1],  dtype=np.complex)/np.sqrt(2),
    "R": np.array([1, 1j],  dtype=np.complex)/np.sqrt(2),
    "L": np.array([1, -1j], dtype=np.complex)/np.sqrt(2)}

def sphere(N=20):
    """mplot_3d coordinate meshes for a unit sphere."""
    theta = np.linspace(0, np.pi, N)
    phi = np.linspace(0, 2*np.pi, 2*N)
    X = np.outer(np.sin(theta), np.cos(phi))
    Y = np.outer(np.sin(theta), np.sin(phi))
    Z = np.outer(np.cos(theta), np.ones_like(phi))
    return X, Y, Z
    
def rotation_matrix(axis,theta):
    '''3D rotation matrix for the given axis and angle'''
    axis = axis/np.sqrt(np.dot(axis,axis))
    a = np.cos(theta/2)
    b,c,d = -axis*np.sin(theta/2)
    return np.array([[a*a+b*b-c*c-d*d, 2*(b*c-a*d), 2*(b*d+a*c)],
                     [2*(b*c+a*d), a*a+c*c-b*b-d*d, 2*(c*d-a*b)],
                     [2*(b*d-a*c), 2*(c*d+a*b), a*a+d*d-b*b-c*c]])

def bloch_color_map(X,Y,Z):
    '''Color map for poincare sphere plots.  This is a weird but
    convenient formula.  It maps the 6 cardinal points onto primary
    and secondary colors.  6 of the 8 octants are mapped to half-faces
    of the color cube, while the +++ and --- octants map to planes
    through the color cube obtained by cutting off the white and black
    corners.  This gives high saturation colors almost everywhere, is
    a 1-1 map, and specifcally avoids white and black'''
    X = np.sign(X) * X**2;
    Y = np.sign(Y) * Y**2;
    Z = np.sign(Z) * Z**2;
    
    red = ((X-Y-Z)+1)/2;
    green = ((-X+Y-Z)+1)/2;
    blue = ((-X-Y+Z)+1)/2;
    red = np.expand_dims(red, axis=red.ndim)
    green = np.expand_dims(green, axis=green.ndim)
    blue = np.expand_dims(blue, axis=blue.ndim)
    red[red<0] = 0
    green[green<0] = 0
    blue[blue<0] = 0
    return np.concatenate((red, green, blue), axis=red.ndim-1)

def rho_color(rho):
    S = rho2stokes_1q(rho)
    return  bloch_color_map(*S)

def plot_ball(ax, r, radius=1, color=(0,0,0)):
    '''Plots a small sphere of the given color'''
    X, Y, Z = ( foo*radius+r for foo in sphere(6) )
    ax.plot_surface(X,Y,Z, facecolor=color)

def rho2stokes_1q(rho):
    if rho.ndim == 1:
        rho = np.outer(rho, rho.conj())
    rho = rho/np.trace(rho)
    S = np.zeros(3)
    S[0] = rho[1,0].real * 2
    S[1] = rho[1,0].imag * 2
    S[2] = rho[0,0].real - rho[1,1].real
    return S

def stokes2rho_1q(S):
    rho = np.zeros((2,2), dtype=np.complex)
    rho[0,0] = (S[2] + 1)/2
    rho[1,1] = (1 - S[2])/2
    rho[1,0] = (S[0] + 1j*S[1])/2
    rho[0,1] = rho[1,0].conj()
    return rho

def rho2psi(rho):
    '''Converts a density matrix to a pure state.  No sanity checking
    is done, this simply returns the eigenvector corresponding to the
    largest eigenvalue.'''
    [v,d] = np.linalg.eigh(rho)
    psi = d[:,np.argmax(v)]
    return psi

def plot_ring(ax, center, normal, radius=1, color=None):
    '''Plots a circle on the given 3D axis.'''
    normal = np.atleast_1d(normal)
    normal = normal / np.sqrt(np.dot(normal, normal))
    center = np.atleast_1d(center)
    theta = np.arange(30) * 2 * np.pi / 29
    X = np.cos(theta)*radius
    Y = np.sin(theta)*radius
    Z = np.zeros_like(X)
    pts = np.vstack((X,Y,Z)).T
    axis = np.cross(np.array([0,0,1]), normal)
    theta = -np.arcsin(np.linalg.norm(axis))
    if np.abs(theta) > .001:
        op = rotation_matrix(axis, theta)
    else:
        op = np.eye(3)

    pts = np.dot(pts, op.T) + center
    ax.plot(pts[:,0], pts[:,1], pts[:,2], color=color or 'k', linewidth=2)

def plot_3rings(ax):
    '''Plots a set of 3 orthogonal rings on the surface of the unit sphere'''
    plot_ring(ax, [0,0,0], [0,0,1], radius = 1.02, color='k')
    plot_ring(ax, [0,0,0], [0,1,0], radius = 1.02, color='k')
    plot_ring(ax, [0,0,0], [1,0,0], radius = 1.02, color='k')

def plot_state_1q(rho, label=r'$|\psi\rangle$'):
    '''Plots a single qubit on the Poincare/Bloch sphere'''
    fig = pyplot.figure()
    ax = fig.add_subplot(111, projection='3d')
    X, Y, Z = sphere(16)
    C = bloch_color_map(X,Y,Z)
    ax.plot_surface(X, Y, Z, rstride = 1, cstride = 1, facecolors = C, alpha = 0.2, linewidth = 0) #cmap = xxx
    plot_3rings(ax)
    ax.axis('tight')
    # poles
    coords = array([[0, 0, 1], [0, 0, -1]]).transpose()  # easier to read this way
    ax.scatter(*coords, c = 'r', marker = 'o')
    ax.text(0, 0,  1.1, '$|0\\rangle$')
    ax.text(0, 0, -1.2, '$|1\\rangle$')
    if rho is not None:
        S = rho2stokes_1q(rho)
        ax.scatter(*S, c = 'k', marker = 'o')
        ax.text(S[0], S[1]+.1, S[2]+.1, label)
    fig.show()

def plot_cardinal(ax):
    for x in [ 'H', 'V', 'D', 'R', 'L' ]:
        S = rho2stokes_1q(s_[x])
        ax.scatter(*(S*1.01), c = bloch_color_map(*S), marker='o')

def plot_transform_sphere(ax, rho):
    rho = pyquantum.psi2rho(rho)
    X,Y,Z = sphere(16)
    Xout = np.zeros_like(X)
    Yout = np.zeros_like(X)
    Zout = np.zeros_like(X)
    alpha_out = np.zeros_like(X)
    C = bloch_color_map(X,Y,Z)

    for idx in np.ndindex(X.shape):
        S = [X[idx], Y[idx], Z[idx]]
        rho1 = stokes2rho_1q(S)
        psi = rho2psi(rho1)
        rho2, P = pyquantum.partial_project(rho, psi, 1, [2, 2])
        Xout[idx], Yout[idx], Zout[idx] = rho2stokes_1q(rho2)
        alpha_out[idx] = P.real if P.real > 0.4 else 0

    C = np.dstack((C, alpha_out))
    ax.plot_surface(Xout, Yout, Zout, rstride = 1, cstride = 1, facecolors = C, linewidth = 2)
    #ax.scatter(Xout.flat, Yout.flat, Zout.flat)

def plot_transform_cardinal(ax, rho):
    for x in [ 'H', 'V', 'D', 'R', 'L' ]:
        psi1 = s_[x]
        rho2 = pyquantum.partial_project(rho, psi1, 1, [2, 2])[0]
        C = rho_color(psi1)
        S = rho2stokes_1q(rho2)
        ax.scatter(*S, c = C, marker='o')

    
def plot_state_2q(rho):
    '''Plots a 2-qubit state on a pair of Poincare/Bloch spheres'''
    fig = pyplot.figure(figsize=pyplot.figaspect(0.5))
    ax1 = fig.add_subplot(121, projection='3d')
    ax2 = fig.add_subplot(122, projection='3d')

    plot_3rings(ax1)
    plot_3rings(ax2)

    plot_cardinal(ax1)
    plot_cardinal(ax2)

    plot_transform_sphere(ax1, rho)
    plot_transform_sphere(ax2, pyquantum.permute_qudits(rho)[0])
    plot_transform_cardinal(ax1, rho)
    plot_transform_cardinal(ax2, pyquantum.permute_qudits(rho)[0])
    
    fig.show()
    
