#!/usr/bin/env python
from __future__ import division
import numpy as np
import scipy
import itertools
import re

phiplus = np.array([1.,0.,0.,1.], dtype=np.complex128)/np.sqrt(2)
phiminus = np.array([1.0, 0., 0., -1.], dtype=np.complex128)/np.sqrt(2);
psiplus = np.array([0., 1., 1., 0.], dtype=np.complex128)/ np.sqrt(2);
psiminus = np.array([0., 1., -1., 0.], dtype=np.complex128)/np.sqrt(2);


def make_hermetian(rho):
    '''What it says'''
    return (rho + rho.conj().T)/2

def normalize_trace(rho):
    '''What it says'''
    return rho/rho.trace()
    
def normalize_psi(psi):
    '''What it says.'''
    psi = np.atleast_1d(psi)
    tmp = np.dot(psi, psi.conj());
    return psi / np.sqrt(tmp);

def dotN(*l):
    '''Computes the matrix inner product of a series of vectors'''
    return reduce(np.dot, l, 1)

def legal_rho(rho):
    '''Takes an input rho and forces it to be a legal density matrix.
    That is, it must be hermetian, positive semi-definite, and unit
    trace.'''
    D,V = np.linalg.eigh(rho)
    D[D<0] = 0
    tmp = np.dot(np.dot(V, np.diag(D)), V.conj().T)
    return normalize_trace(tmp)

def basis_state(x, dim=2):
    '''
    Returns a single element of a state vector.  Each component is
    either a digit (0,1,2,...) representing the computational basis,
    or one or two letters representing a special state (H,V,D,A,R,L,
    X,Y,Z,Xm,Ym,Zm.  B1..4 are the 4 bell states
    '''
    if x.isdigit():
        rv = np.zeros((dim))
        rv[int(x)] = 1
    else:
        d = {'H': [1,0],
             'V': [0,1],
             'D': [1,1],
             'A': [1,-1],
             'R': [1,1j],
             'L': [1,-1j],
             'X': [1,1],
             'Xp': [1,1],
             'Xm': [1,-1],
             'Y': [1,1j],
             'Yp': [1,1j],
             'Ym': [1,-1j],
             'Z': [1,0],
             'Zp': [1,0],
             'Zm': [0,1],
             'B1': phiminus,
             'B2': phiplus,
             'B3': psiminus,
             'B4': psiplus}
        rv = np.array(d[x], dtype=np.complex)
        rv = rv / np.linalg.norm(rv)
    return rv
             
def ket(sstr, dim=2, norm=1):
    '''
    Constructs a state from a string description.  It can be of the
    form 101 + .5*010 - 001 ...
    
    By default, everything is a qubit.  if 'dim' is specified, it is
    the dimension of the hilbert space (2 for qubit, 3 for qutrit...).
    This can't be higher than 9, since every digit is a single
    subsystem.  The special values : H,V,D,A,R,L,Z,X,Y,Xm,Ym,Zm are
    treated as qubits regardless of 'dim'.  B1..4 are the 4 bell
    states and treated as two qubits.  Invalid input will generate
    wierd error messages or in some cases ("2H + H2" with dim=3)
    give non-sensical results.
    '''
    sstr = re.sub(r'\s', '', sstr)
    term_re = r"([-+])?([0-9.]+\*)?(\w+)"
    rv =0
    for m in re.findall(term_re, sstr):
        coeff = float("0" + m[1]) * (-1 if m[0] == '-' else 1)
        it = re.findall(r'B[1-4]|\w(?:m|p)?', m[2])
        x = [ basis_state(x, dim) for x in it ]
        term = reduce(np.kron, x)
        rv = rv + term
    if norm:
        rv = rv / np.linalg.norm(rv)
    return rv

def density(*state_strings, **kwargs):
    '''
    Taks a list of state description strings (see state method) and
    returns a density matrix corresponding to their incoherent mixture.
    Takes keyword arguments dim and norm with the same meaning as 'state'
    '''
    dim=kwargs.pop(dim, 2)
    norm=kwargs.pop(norm,True)
    if len(kwargs):
        raise RuntimeError('Unknown keyword arguments')
    rv =sum(map(lambda x: outer(x, x.conj()), map(ket, state_strings, dim=dim, norm=0)))
    if norm:
        rv = rv / np.trace(rv)
    return rv

def psi2rho(psi):
    '''
    Converts input state vector to density matrix represenation.  If psi
    is already a density matrix, leave it alone, so this function can be
    used to coerce input arguments to the desired type.
    '''
    if psi.ndim == 1:
        return legal_rho(np.outer(psi, psi.conj()))
    else: # Already a density matrix
        return legal_rho(psi)

def get_D(rho, D):
    '''
    Helper function.  D specified how the state is partitioned into 
    subsystems. If D is not specifeid, assume it is all qubits
    otherwise raise an error.
    '''
    dim = rho.shape[0];
    if D is None:
        nqubits = int(np.log2(dim))
        if dim != 2**nqubits:
            raise RuntimeError("Qubits specified, but rho is not power of 2")
        D = 2*np.ones(nqubits, dtype=np.integer)
    else:
        D = np.atleast_1d(D)
    nparts = D.shape[0]
    return D

def permute_qudits(rho, perm=None, D=None):
    '''
    Permutes the order of qudits in the specified density matrix,
    divided into parts according to the vector D specifiying the
    dimension of each part.  If D is not specified, length(rho) must
    be a power of two and rho is treated as log2(length(rho)) qubits.
    perm is a vector specifying the permutation.  If not specified or
    empty, a cyclic permutation is performed (the last part is made
    first).  If perm is a positive scalar, the n'th part will be
    placed first, and all other parts will appear in order.  That is
    useful for simplifying operations like partial trace / partial
    projection.  Finally, if perm is a negative scalar, the reverse
    happens: the first element is inserted into the n'th place.
    '''

    # Handle arguments.
    dim = rho.shape[0]
    D = get_D(rho, D)
    nparts = D.shape[0]
    if perm is None:
        perm = np.concatenate(([nparts-1], np.arange(nparts-1)))
    perm = np.atleast_1d(perm)
    if perm.size == 1:
        if perm[0] > 0:
            perm = np.concatenate((perm, np.arange(perm), np.arange(perm+1, nparts)))
        elif perm[0] < 0:
            perm = np.concatenate((np.arange(1,-perm+1), [0], np.arange(-perm+1, nparts)))
        else: # perm = 0: do nothing
            # perm = np.arange(dim)
            return rho, D

    newD = D[perm]
    
    # Now we have the "small" permutation vector in the that applies to qubits
    # we can generate a "big" permutation vector that applies to the rows/cols of rho

    big_perm = np.zeros(dim, dtype=np.integer)
    it = itertools.product(*map(np.arange, newD))
    for old_idx, v in enumerate(it):
        tmp = np.array(v)[perm]
        new_idx = np.sum(tmp * np.prod(D) // np.cumprod(D))
        big_perm[old_idx] = new_idx

    rho_out = rho[:,big_perm][big_perm,:]
    return rho_out, newD

def matrix2tensor4(rho, n, D=None):
    '''
    Changes the normal flattened 2D density matrix representation and
    pulls out a single subsystem into its own tensor dimensions.  The
    result is a rank-4 tensor where the first two coordinates index
    the states of the desired subsystem, while the sencond two coordinates
    index everything else.
    '''
    sz = rho.shape[0]
    D = get_D(rho, D)
    idx = np.arange(sz)
    coord = np.unravel_index(idx, D)

    index_lists = [ [] for j in range(D[n]) ]
    for j in idx:
        index_lists[coord[n][j]].append(j)
    
    rho_out = np.zeros((D[n], D[n], sz//D[n], sz//D[n]), dtype=np.complex128)
    for j in range(D[n]):
        for k in range(D[n]):
            rho_out[j,k,:,:] = rho[index_lists[j],:][:,index_lists[k]]
    return rho_out

def schmidt_decomposition(psi, D=(2,2)):
    tmp = np.reshape(psi, D)
    U, s, Vh = scipy.linalg.svd(tmp)
    
def schmidt_rank(psi, threshold=.001, D=(2,2)):
    U,s,Vh = schmidt_decomposition(psi, D)
    np.sum(s > threshold)
    
def rf_unitary(angle, phase):
    '''
    Gives the unitary operator for an ideal microwave gate.
    Phase gives the rotation axis on the plane of the bloch sphere (RF drive phase)
    angle is the rotation angle of the gate (pulse area)
    '''
    alpha = np.array([1, np.exp(1j*phase)])/np.sqrt(2)
    alpha_p = np.array([-np.exp(-1j*phase), 1])/np.sqrt(2)
    U = np.outer(alpha.conj(), alpha) + np.exp(1j*angle)*np.outer(alpha_p.conj(),alpha_p)
    return U

def tensor42matrix(rho, n, D):
    '''
    Reverses the matrix2tensor4 operation.  Flattens a rank 4 tensor to a 2d matrix.
    'n' gives the position to insert the isolated subsystem, 
    '''
    raise NotImplemented()
    pass

def partial_trace(rho, n, D=None):
    '''
    Traces over one part of a multi-partite quantum system and gives
    the reduced density matrix of the remainder.
    '''
    return np.trace(matrix2tensor4(rho, n, D))

def partial_project(rho, psi, n, D=None):
    '''
    Calculates the result of a partial projection of one part of a
    multi-partite probability.  This returns the conditional state of
    the remainder of the system, along with the probability of success
    of the original projection.
    '''
    x = matrix2tensor4(rho, n, D)
    tmp = np.tensordot(psi.conj(), x, [0,1])
    out = np.tensordot(tmp, psi, [0,0])
    P = np.trace(out)
    out /= P
    return out, P

def partial_transpose(rho, n, D):
    x = matrix2tensor4(rho, n, D)
    np.transpose(x, [1,0,2,3])
    return tensor42matrix(rho, n, D)

def concurrence(rho):
    """Usage: C=concurrence(rho)
    
    Returns the two-qubit concurrence of the density matrix rho or a
    two qubit state."""

    rho = psi2rho(rho)
    Z = np.array([[ 0, 0, 0,-1],
                  [ 0, 0, 1, 0],
                  [ 0, 1, 0, 0],
                  [-1, 0, 0, 0]])
    R = dotN(rho, Z, rho.T, Z) 
    v = np.linalg.eigvals(R)
    tmp = np.sqrt(v).real
    C = 2 * max(tmp) - sum(tmp)
    C = C if C>0 else 0
    return C

def tangle(rho):
    '''For two qubits, tangle is just the square of the concurrence.'''
    return concurrence(rho)**2

def negativity(rho, n=1, D=None):
    '''
    Negativity measure of entanglement. 
    '''
    rho = psi2rho(rho)
    x = partial_transpose(rho, n, D)
    v = np.linalg.eigvalsh(x)
    return max(-2*v.min(),0)

def gell_mann(j, N):
    a = j//N
    b = j - a*N
    if a<b:
        rv = np.zeros((N,N), dtype="complex")
        rv[a,b] = 1
        rv[b,a] = 1
    elif a>b:
        rv = np.zeros((N,N), dtype="complex")
        rv[a,b] = 1j
        rv[b,a] = -1j
    elif a>0:
        rv = np.diag((np.arange(N)<=a) * np.sqrt(N/(a**2+a)))
        rv[a,a] = -a*np.sqrt(N/(a**2 + a))
    else:
        rv = np.eye(N)/N
    return rv
        
def sigma_all(N=1):
    '''
    Return the pauli basis for N qubits
    '''
    s0 = np.eye(2, dtype=np.complex)
    s1 = np.array([[0,1],[1,0]], dtype=np.complex)
    s2 = np.array([[0,-1j],[1j,0]], dtype=np.complex)
    s3 = np.array([[1,0],[0,-1]], dtype=np.complex)
    it =  itertools.product(*([ [s0, s1, s2, s3] ] * N))
    for x in it:
        yield reduce(np.kron, x)

def sigma(j, N=1):
    s0 = np.eye(2, dtype=np.complex)
    s1 = np.array([[0,1],[1,0]], dtype=np.complex)
    s2 = np.array([[0,-1j],[1j,0]], dtype=np.complex)
    s3 = np.array([[1,0],[0,-1]], dtype=np.complex)
    s = [s0, s1, s2, s3]
    dims = [4]*N
    idx = np.unravel_index(j, dims)
    return reduce(np.kron, (s[x] for x in idx))

def bloch2rho(S):
    '''
    Convert a generalized bloch vector S to a denisty matrix.  Currently only works for
    N-qubits
    '''
    N = int(np.log2(len(S))/2)
    x = sigma_all(N)
    rho = reduce(np.add, (a * b for (a,b) in zip(S, x)))
    return rho / rho.shape[0]

def rho2bloch(rho):
    rho = psi2rho(rho)
    N = int(np.log2(rho.shape[0]))
    S = np.array([ np.trace(np.dot(rho,x)).real for x in sigma_all(N) ])
    return S

def qubitAngles2Matrix(theta,phi):
    return np.array([[cos(theta/2.0)**2,exp(-I*phi)*cos(theta/2.0)*sin(theta/2.0)],
                     [exp(I*phi)*cos(theta/2.0)*sin(theta/2.0), sin(theta/2.0)**2]])
    
def qubitAngles2Vector(theta,phi):
    return np.array([sin(theta)*cos(phi), sin(theta)*sin(phi), cos(theta)])
    
def qubitMatrix2Vector(rho):
    return np.array([np.trace(np.dot(rho, sigma)) for sigma in SIGMAS])
    
def qubitMatrix2Angles(rho):
    vec = np.real(qubitMatrix2Vector(rho))
    angles = qubitVector2Angles(vec)
    return angles
    
def qubitVector2Matrix(blochVector):
    return 0.5*(np.eye(2)+sum([v*sig for v,sig in zip(blochVector,SIGMAS)]))

def qubitVector2Angles(blochVector):
    """Convert a bloch vector into spherical angles theta and phi"""
    x,y,z = np.real(blochVector)
    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arccos(z/r)
    phi = np.arctan2(y,x)
    return theta,phi
    
def angleBetweenVectors(v, u):
    """The angle between two vectors v and u"""
    return np.arccos(np.dot(u,v)/np.sqrt(np.dot(v,v)*np.dot(u,u)))
        

    
def entanglement_of_formation(rho):
    '''For 2 qubit, entanglement of formation can be computed from
    concurrence.'''
    C = concurrence(rho)
    if C <= 0: # Should do an expansion here for small C...
        return 0
    elif C >= 1:
        return 1
    else:
        x = (1+np.sqrt(1-C**2))/2
        Ef = -x*np.log2(x) - (1-x)*np.log2(1-x)
        return Ef

def trace_dist(rho1, rho2):
    '''The trace distance between two denstiy matricies'''

    rho1 = psi2rho(rho1)
    rho2 = psi2rho(rho2)
    l = np.linalg.eigvalsh(rho1 - rho2)
    return np.sum(np.abs(l))*0.5

def fully_entangled_fraction(rho):
    '''This is the fully entangled fraction as defined by Bennett et
    al. in PRA 54, p.3824.  The expression below is closed form, but
    it is equivalent to max(<e|rho|e>) over all maximally entangled
    states |e>.'''
    rho = psi2rho(rho)
    O = np.column_stack((phiplus, 1j*phiminus, 1j*psiplus, psiminus))
    M = dotN(O.conj().T, rho, O)
    l = np.linalg.eigvalsh(M.real)
    return l.max()

def entropy(rho):
    '''von Neumann / Shannon entropy function'''
    if rho.ndim==1: # pure states have no entropy
        return 0
    v = np.linalg.eigvalsh(rho).real
    return -np.sum(v[v>0] * np.log2(v[v>0]))

def fidelity(rho1, rho2):
    '''
    The fidelity of two quantum states.
    '''
    if rho1.ndim == 1:
        return np.dot(rho1.conj(), np.dot(psi2rho(rho2), rho1))
    elif rho2.ndim == 1:
        return np.dot(rho1.conj(), np.dot(psi2rho(rho2), rho1))
    else:
        tmp = scipy.linalg.sqrtm(rho1)
        ev = np.linalg.eigvalsh(dotN(tmp, rho2, tmp))
        return np.sum(np.sqrt(ev.real))

def linear_entropy(rho):
    '''Linear entropy function.  I think this is only used by people
    uncomforatble with logarithms, but it may have some practical use'''
    if rho.ndim == 1:
        return 0
    d = rho.shape[0]
    return d * (1.0 - np.trace(np.dot(rho,rho).real)) / (d-1.0)

    
def apply_chi(chi, rho):
    rho = psi2rho(rho)
    raise RuntimeError()
