
import warnings

class AttrDict(dict):
    """A dict whose entries can also be accessed as attributes.
    
    The copy method returns a copy, including recursive copies of any
    sub directories as AttrDict's.  Mutable values of other types,
    such as lists, are not copied.  A better approach would be to use
    the copy module's deepcopy functionality which would copy
    containers.
    
    The where method returns a copy with updates passed as a dictionary
    or keyword parameters.  This also allows updates to be made to
    keys in subdirectories by passing dotted names (or, when using
    keyword parameters, by using '__' in place of '.').
    """

    # ERJ has disabled the __*attr__ methods here.  Aliasing the
    # dictionary contents to instance attributes can cause confusing
    # bugs and weird behavior, as well as make debugging and
    # introspection harder.  If you really like this feature or need
    # it for existing code, please talk to me before changing it back.

    def disabled__getattr__(self, name):
        try:
            attr = dict.__getitem__(self, name)
            warnings.warn('attribute access for AttrDict is deprecated', DeprecationWarning, stacklevel=2)
            return attr
        except KeyError:
#            return dict.__getattr__(self, name) # This code is redundant -- python always checks for 
#        except:                                 # normal attributes before calling __getattr__ -- ERJ
            raise AttributeError(name)

    def disabled__setattr__(self, name, value):
        warnings.warn('attribute access for AttrDict is deprecated', DeprecationWarning, stacklevel=2)
        dict.__setitem__(self, name, value)
    
    def disabled__delattr__(self, name):
        try:
            dict.__delitem__(self, name)
        except KeyError:
            raise AttributeError(name)
    
    def set_readonly(self):
        '''
        Converts an attrdict to be readonly.  This is useful after
        making a copy when you don't want someone to accidentally
        mutate the original and think it affects the copy.
        '''
        self._readonly = True

    def __setitem__(self, key, value):
        if getattr(self, '_readonly', False):
            raise RuntimeError("Attempting to mutate read-only AttrDict")
        dict.__setitem__(self, key, value)

    def copy(self, readonly=None):
        if readonly is None:
            readonly = getitem(self, '_readonly', False)
        d = AttrDict(self)
        object.__setattr__(d, '_dir', self._dir)
        object.__setattr__(d, '__name__', self.__name__)
        for k, v in d.iteritems():
            if isinstance(v, AttrDict):
                d[k] = v.copy(readonly=readonly)
        if readonly:
            d.set_readonly()
        return d
    
    def where(self, dict={}, **kwds):
        d = self.copy()
        def add_all(updates):
            for k, v in updates.iteritems():
                k = k.replace('__', '.')
                if '.' in k:
                    path = k.split('.')
                    path, k = path[:-1], path[-1]
                    subdir = d
                    for dir in path:
                        subdir = subdir[dir]
                    subdir[k] = v
                else: # just set the key
                    d[k] = v
        add_all(dict)
        add_all(kwds)
        return d


class RegistryWrapper(object):
    """Accesses the labrad registry with a dict-like interface.
    
    Keys or directories in the registry can be accesed using either
    dict-like notation d["key"] or attribute access d.key.  Obviously
    the former is required if you want to compute the string name of
    an element to access and then pass the key as a variable.
    
    For every packet that gets sent to the registry, we first cd
    into the directory wrapped by this wrapper, then at the end
    cd back into the root directory.  That way this directory
    (and any subdirectories that we visit which get wrapped) can
    always be deleted because there is not a stray context hanging
    out in any particular directory.
    
    The copy method returns a local copy of this registry directory
    in the form of an AttrDict. The copy includes recursive copies
    of all subdirectories. 
    """
    def __init__(self, cxn, dir='', ctx=None):
        if isinstance(dir, str):
            dir = [dir]
        else:
            dir = list(dir)
        if '' in dir[1:]:
            raise Exception('Empty string is invalid subdirectory name')
        if dir[0] != '':
            dir = [''] + dir
        if ctx is None:
            ctx = cxn.context()
        srv = cxn.registry
            
        object.__setattr__(self, '_dir', dir)
        object.__setattr__(self, '_cxn', cxn)
        object.__setattr__(self, '_srv', srv)
        object.__setattr__(self, '_ctx', ctx)
        
        # make sure the directory gets created
        self._send(self._packet())
        
    def _subdir(self, name):
        """Create a wrapper for a subdirectory, reusing our connection."""
        if name == '':
            raise Exception('Empty string is invalid subdirectory name')
        return RegistryWrapper(self._cxn, self._dir + [name], self._ctx)
        
    def _packet(self):
        """Create a packet with the correct context and directory."""
        return self._srv.packet(context=self._ctx).cd(self._dir, True)
    
    def _send(self, pkt):
        """Change back into the root directory after each request."""
        return pkt.cd(['']).send()
    
    def _get_list(self):
        """Get current directory listing (dirs and keys)."""
        return self._send(self._packet().dir()).dir
        
    ## dict interface
    def __getitem__(self, name):
        dirs, keys = self._get_list()
        if name in dirs:
            return self._subdir(name)
        elif name in keys:
            return self._send(self._packet().get(name)).get
        else:
            raise KeyError(name)

    def __setitem__(self, name, value):
        if self.__contains__(name):
            print 'changing',name,'from',self[name],'to',value
        else:
            #print 'creating key',name,value
            pass
        if isinstance(value, (dict, RegistryWrapper)):
            subdir = self._subdir(name)
            for element in value:
                subdir[element] = value[element]
        else:
            self._send(self._packet().set(name, value))

    def __delitem__(self, name):
        dirs, keys = self._get_list()
        if name in dirs:
            subdir = self._subdir(name)
            for k in subdir:
                del subdir[k]
            self._send(self._packet().rmdir(name))
        elif name in keys:
            self._send(self._packet()['del'](name))
        else:
            raise KeyError(name)

    ## attribute interface
    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self.__setitem__(name, value)

    def __delattr__(self, name):
        try:
            return self.__delitem__(name)
        except KeyError:
            raise AttributeError(name)

    def copy(self):
        """Make a local copy, recursively copying subdirs as well.
        Returns an AttrDict, NOT another RegistryWrapper"""
        dirs, keys = self._get_list()
        d = AttrDict()
        object.__setattr__(d, '_dir', self._dir)
        object.__setattr__(d, '__name__', self._dir[-1])
        for name in dirs:
            d[name] = self._subdir(name).copy()
        if len(keys):
            p = self._packet()
            for name in keys:
                p.get(name, key=name)
            ans = self._send(p)
            for name in keys:
                d[name] = ans[name]
        return d

    def __contains__(self, name):
        dirs, keys = self._get_list()
        return name in keys or name in dirs
    
    def keys(self):
        dirs, keys = self._get_list()
        return list(dirs) + list(keys)
    
    def __iter__(self):
        return iter(self.keys())
    
    def __repr__(self):
        return '<RegistryWrapper: %r>' % (self._dir,)

