import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy.interpolate import interp1d

def hamiltonian(Ec, Ej, points=500,extraTerms={},frust=0.0,plot=False):

    # extra terms: seriesJunction: relativeLargerArea
    #              seriesInductor: Ej for the inductor

    delta=np.linspace(-np.pi,np.pi,points)
    ddelta=delta[1]-delta[0]
    
    derivTerm=-4*Ec/(ddelta**2)*(np.diag([1]*(points-1),-1)+np.diag([-2]*points)+np.diag([1]*(points-1),1))
    
    U=-Ej*np.cos(np.pi*frust)*np.cos(delta)
    potentialTerm=np.diag(U)
    
    if extraTerms.has_key('seriesJunction'):
    
        Ejp=extraTerms['seriesJunction']
        
        deltaA=np.arcsin(Ej/Ejp*np.sin(delta))+delta
        deltaB=interp1d(deltaA, delta)
        deltaA=delta
        
        U=-Ej*np.cos(deltaB(delta))-Ejp*np.cos(deltaA-deltaB(deltaA))
        
        potentialTerm=np.diag(U)
        
    elif extraTerms.has_key('seriesInductor'):
    
        Ejp=extraTerms['seriesInductor']
    
        delta=np.linspace(-np.pi, np.pi, 500)

        deltaA=Ej/Ejp*np.sin(delta)+delta
        deltaB=interp1d(deltaA, delta)
        deltaA=delta
        
        U=-Ej*np.cos(np.pi*frust)*np.cos(deltaB(deltaA))+Ejp*(deltaA-deltaB(deltaA))**2
    
        potentialTerm=np.diag(U)
        
    elif extraTerms.has_key('harmonic'):
        
        coef=extraTerms['harmonic']
        
        U=-Ej*np.cos(np.pi*frust)*np.cos(delta)+coef*(delta)**2
    
        potentialTerm=np.diag(U)
    
    H=derivTerm+potentialTerm
    
    if plot:
        plt.figure(0)
        plt.plot(delta, U-np.mean(U))
        plt.xlabel('Phase (delta)')
        plt.ylabel('Potential (V)')
    
    return delta, H
    
def analyze(H,sparse=False):
    
    if sparse:
        eval,evec=scipy.sparse.linalg.eigs(H,which='SM',k=4)
    else:
        eval,evec=np.linalg.eig(H)
    
    inds=np.argsort(eval)   
    eval=eval[inds]
    evec=evec[:,inds]
    
    f10=eval[1]-eval[0]
    f21=eval[2]-eval[1]
    
    nonlin=f10-f21
    
    return f10,nonlin
    
def plotEvecs(delta, H, levels=3):

    eval,evec=np.linalg.eig(H)
    
    inds=np.argsort(eval)
    eval=eval[inds]
    evec=-evec[:,inds] # vectors coming out negative... dunno why
    
    spacing=3*max(evec[:,0])
    
    for ind in range(levels):
        plt.plot(delta, evec[:,ind]+spacing*ind)
    
def rToEj(r):
    
    phi0=2.05e-15
    Delta=160e-6*1.6e-19
    q=1.6e-19
    h=2*np.pi*1.05e-34
    
    Ej=phi0*Delta/(4*q*r)/h
    
    return Ej
    
def cToEc(c):

    q=1.6e-19
    h=np.pi*2*1.05e-34
    Ec=q**2/(2*c)/h
    
    return Ec
    
def mvToF(Mutual, Voltage):

    Current=Voltage/(1.5e3) # cold resistance 1.5kOhm
    Flux=Mutual*Current
    phi0=2.05e-15
    frust=Flux/phi0
    
    return frust
    
def rcToFreqs(r,c,sparse=False,**kwargs):

    Ec=cToEc(c)
    Ej=rToEj(r)
    
    delta,H=hamiltonian(Ec,Ej,**kwargs)
    
    f10,nonlin=analyze(H,sparse=sparse)
    
    return f10/1.0e9,nonlin/1.0e9
    
def EjEcToFreqs(Ej,Ec,sparse=False,**kwargs):
    
    delta,H=hamiltonian(Ec,Ej,**kwargs)
    
    f10,nonlin=analyze(H,sparse=sparse)
    
    return f10/1.0e9,nonlin/1.0e9
    
def nonlin(Ecs=[1]*100, Ejs=np.logspace(0,2,100),points=500,alpha=0):

    relAnh=[]
    anh=[]
    
    for Ec, Ej in zip(Ecs,Ejs):
        delta, H = hamiltonian(Ec,Ej,points=points,alpha=alpha)
        
        f10,nonlin=analyzeH(H)
        
        relAnh.append(nonlin/f10)
        anh.append(nonlin)
    return anh,relAnh
    
