''' analysis.py

analysis for RB sim stuff.
'''

import numpy as np

def fit_slope(Nlist, data, dphi=1e-4, guess=1.0):
    def line(x, slope, off):
        return (x+off)*slope*dphi**2 / 6.0
    from scipy.optimize import curve_fit
    fit, pcov = curve_fit(line, np.array(Nlist), data, [guess, 0.1])
    return fit[0], fit[1]


def analyze(Nlist, results_list, names, plot=True, dphi=1e-4):
    slopes = [fit_slope(np.array(Nlist), x, dphi=dphi)[0] for x in results_list]
    
    if plot:
        import matplotlib.pyplot as plt
        plt.ion()
        ax = plt.figure().add_subplot(111)
        i, l = 0, len(plt.rcParams['axes.color_cycle'])
        for slope, name, results in zip(slopes, names, results_list):
            color = plt.rcParams['axes.color_cycle'][i%l]            
            ax.plot(Nlist, results, 'o', label=name, color=color)
            x = np.linspace(0, np.max(Nlist))
            ax.plot(x, slope*x*dphi**2/6.0, '-', label="slope=%.4f" % slope, color=color)
            i += 1
        ax.legend(numpoints=1, loc='upper left')
        plt.draw()
    return slopes

def analyze_2qubit(data, Nlist=range(5,100,5), plot=True, dphi=1e-4, title=None, extra_label=''):
    qa = np.array([x[1] + x[3] for x in data])
    qb = np.array([x[2] + x[3] for x in data])
    if len(Nlist) != len(qa):
        raise ValueError("Nlist not same length as data.")
    slopea, offa = fit_slope(Nlist, qa, dphi=dphi)
    slopeb, offb = fit_slope(Nlist, qb, dphi=dphi)
    qa /= (dphi**2)/6.0
    qb /= (dphi**2)/6.0
    if plot:
        import matplotlib.pyplot as plt
        plt.ion()
        if plot is True:
            ax = plt.figure().add_subplot(111)
        else:
            ax = plot
        ax.plot(Nlist, qa, 'o', label=extra_label + "q$_a$=$a_1^2+a_3^2$, slope=%.4f, off=%.2f" % (slopea, offa))
        ax.plot(Nlist, qb, 'o', label=extra_label + "q$_b$=$a_2^2+a_3^2$, slope=%.4f, off=%.2f" % (slopeb, offb))
        ax.set_xlabel("N gates", fontsize=19)
        ax.set_ylabel(r"Error$\times 6 / \phi_g^2$", fontsize=19)
        if title:
            ax.set_title(title, fontsize=19)
        ax.legend(numpoints=1, loc='upper left')
        plt.draw()
    return slopea, slopeb
