import random, itertools
import numpy as np
import numpy_states as nps


pi = np.pi
gates = []
angles = []
compiled_gates = []
inverted_gates = []
noise_gate = None
compiled_noise = []
compiled_noise_correlated = None

n_interspersed = None
compiled_gates_interspersed = []
compiled_noise_interspersed = []
compiled_noise_correlated_interspersed = None

cliffords = []
cliffords_interspersed = []
cliffords_inverted = []

def tuplify(x):
    if type(x) is tuple or type(x) is list:
        return x
    else:
        return (x,)

def _pad_gates(gates, pad_to_length=3):
    ''' pad each tuple of gates to be length X by adding identity gates.
    will truncate any tuple that is longer than X gates (not recommended). '''
    g2 = []
    for g in gates:
        subgate = []
        for i in range(pad_to_length):
            try:
                subgate.append(g[i])
            except IndexError:
                subgate.append(nps.I)
        g2.append(tuple(subgate))
    return g2

def generate_cliffords(intersperse=1):
    x = float(intersperse)
    cliffords = []
    # paulis
    cliffords += [nps.I, nps.rotx(pi/x), nps.roty(pi/x), (nps.rotx(pi/x), nps.roty(pi/x))]
    # 2*pi/3 rotations
    x2s = [nps.rotx(pi/2/x), nps.rotx(-pi/2/x)]
    y2s = [nps.roty(pi/2/x), nps.roty(-pi/2/x)]
    cliffords += list(itertools.product(x2s, y2s)) + list(itertools.product(y2s, x2s))
    # pi/2 rotations
    cliffords += x2s + y2s + [(x2s[1], y2s[0], x2s[0])] + [(x2s[1], y2s[1], x2s[0])]
    # hadamard-like
    cliffords += list(itertools.product([nps.rotx(pi/x)], y2s))
    cliffords += list(itertools.product([nps.roty(pi/x)], x2s))
    cliffords += [(x2s[0], y2s[0], x2s[0]), (x2s[1], y2s[0], x2s[1])]
    cliffords = map(tuplify, cliffords)
    return cliffords
    

def compile_gates(_gate_list=[nps.rotx, nps.roty],
                  _angle_list=[pi/2, -pi/2, pi],
                  _noise_gate=nps.rotz,
                  interspersed_n = 10,
                  dphi=0.0001):
    global gates, angles, compiled_gates, inverted_gates, noise_gate
    global compiled_noise, compiled_noise_correlated
    # basic gates
    gates = list(_gate_list)
    angles = list(_angle_list)
    noise_gate = _noise_gate
    compiled_gates = [g(a) for g in gates for a in angles]
    inverted_gates = [g(-a) for g in gates for a in angles]
    compiled_noise = [noise_gate(dphi), noise_gate(-dphi)]
    compiled_noise_correlated = compiled_noise[0]
    # interspersed gates
    global compiled_gates_interspersed, compiled_noise_interspersed
    global compiled_noise_correlated_interspersed, n_interspersed
    n_interspersed = interspersed_n
    compiled_gates_interspersed = [g(a/n_interspersed) for g in gates for a in angles]
    compiled_noise_interspersed = [noise_gate(dphi/np.sqrt(n_interspersed)), noise_gate(-dphi/np.sqrt(n_interspersed))]
    compiled_noise_correlated_interspersed = noise_gate(dphi/n_interspersed)
    # cliffords
    global cliffords, cliffords_interspersed, cliffords_inverted
    cliffords = generate_cliffords()
    cliffords_interspersed = generate_cliffords(intersperse=n_interspersed)
    cliffords_inverted = generate_cliffords(intersperse=-1)
    #cliffords_inverted = [x[::-1] for x in cliffords_inverted]

def uncorrelated(N=100):
    ''' Uncorrelated noise: each step gets z-rotation (+/- dphi)
    returns probability of 1 state '''
    state = nps.ground()
    inverter = nps.I
    for i in range(N):
        r = random.randint(0, len(compiled_gates)-1)
        state = np.dot(compiled_gates[r], state)
        state = np.dot(random.choice(compiled_noise), state)
        inverter = np.dot(inverter, inverted_gates[r])
    state = np.dot(inverter, state)
    return nps.measure(state)

def correlated(N=100):
    ''' correlated noise: each step gets z-rotation (dphi)
    returns probability of 1 state '''
    state = nps.ground()
    inverter = nps.I
    for i in range(N):
        r = random.randint(0, len(compiled_gates)-1)
        state = np.dot(compiled_gates[r], state)
        state = np.dot(compiled_noise_correlated, state)
        inverter = np.dot(inverter, inverted_gates[r])
    state = np.dot(inverter, state)
    return nps.measure(state)
    
def uncorrelated_interspersed(N=100):
    ''' uncorrelated noise, interspersed with the rotation '''
    state = nps.ground()
    inverter = nps.I
    for i in range(N):
        r = random.randint(0, len(compiled_gates_interspersed)-1)
        for j in range(n_interspersed):
            state = np.dot(compiled_gates_interspersed[r], state)
            state = np.dot(random.choice(compiled_noise_interspersed), state)
        inverter = np.dot(inverter, inverted_gates[r])
    state = np.dot(inverter, state)
    return nps.measure(state)

def correlated_interspersed(N=100):
    ''' uncorrelated noise, interspersed with the rotation '''
    state = nps.ground()
    inverter = nps.I
    for i in range(N):
        r = random.randint(0, len(compiled_gates_interspersed)-1)
        for j in range(n_interspersed):
            state = np.dot(compiled_gates_interspersed[r], state)
            state = np.dot(compiled_noise_correlated_interspersed, state)
        inverter = np.dot(inverter, inverted_gates[r])
    state = np.dot(inverter, state)
    return nps.measure(state)
    
def correlated_cliffords(N=100):
    state = nps.ground()
    inverter = nps.I
    for i in range(N):
        r = random.randint(0, len(cliffords)-1)
        for g, inv in zip(cliffords[r], cliffords_inverted[r]):
            state = np.dot(g, state)
            state = np.dot(compiled_noise_correlated, state)
            inverter = np.dot(inverter, inv)
    state = np.dot(inverter, state)
    return nps.measure(state)
    
def uncorrelated_cliffords(N=100):
    state = nps.ground()
    inverter = nps.I
    for i in range(N):
        r = random.randint(0, len(cliffords)-1)
        for g, inv in zip(cliffords[r], cliffords_inverted[r]):
            state = np.dot(g, state)
            state = np.dot(random.choice(compiled_noise), state)
            inverter = np.dot(inverter, inv)
    state = np.dot(inverter, state)
    return nps.measure(state)
    
def generate_noise(correlated=False, phi_g=1e-4):
    if correlated:
        return [nps.rotz(phi_g)]
    else:
        return [nps.rotz(phi_g), nps.rotz(-phi_g)]

def generic(gates_tup, gates_inv_tup, noise, N=100, start_state=nps.ground(),
            final_inv=nps.I, n_interspersed=1, noise_after=False):
    ''' Run a single RB instance, rotate back to zero, return the measurement.
    parameters:
        gates_tup: list of tuples of gates. each tuple forms a single logical
            gate, implemented by the physical gates in the tuple. each physical
            gate should be a rotation from numpy_states
        gates_inv_tup: list of tuples of gates, each tuple being the inverse
            of the gate at the same position in the gates_tup list. (for
            simplicity usually just the same rotations as gates_tup but with the
            opposite (negative) angle.)
        noise: list of noise gates. each time noise is applied it is pulled
            randomly from this list. usually [rotz(phi_g)] for correlated noise
            and [rotz(phi_g), rotz(-phi_g)] for uncorrelated.
        N: length of RB sequance, e.g. number of times to pick a (logical) gate 
            at random from gates_tup and apply it.
        start_state: in case you have a funny list of gates (e.g. z-rots only)
            and want to start from something other than the ground state.
        final_inv: gate that rotates the starting state to ground.
        n_interspersed: number of times to apply each physical gate. used for
            interspersing noise. e.g. if n_interspersed=10 then X/2 gate should
            be rotz(pi/2/10), and the noise should be scaled appropriately.
        noise_after:
            if true, only apply noise after all physical gates in the tuple.
    '''
    state = start_state
    inverter = final_inv
    for i in range(N):
        r = random.randint(0, len(gates_tup)-1)
        for g, inv in zip(gates_tup[r], gates_inv_tup[r]):
            for j in range(n_interspersed):
                state = np.dot(g, state)
                if not noise_after:
                    state = np.dot(random.choice(noise), state)
            inverter = np.dot(inverter, inv)
        if noise_after:
            state = np.dot(random.choice(noise), state)
    state = np.dot(inverter, state)
    return nps.measure(state)
    
def generic_avg(gates_tup, gates_inv_tup, noise, num_avg=400,
                Nlist=range(5,100,5), no_divide=False, **kwargs):
    ''' Run generic() num_avg times for each N in Nlist, return the results.
    parameters:
        gates_tup, gates_inv_tup, noise: as generic()
        **kwargs: as generic(), except for N
        no_divide: if true, each element of return value is sum of num_avg
            measurements. if false (default), divide by num_avg to give average
            measurement.
    returns:
        ndarray of length Nlist, 
    '''
    data = np.zeros_like(Nlist, dtype=np.float64)
    for i in range(num_avg):
        for j, n in enumerate(Nlist):
            data[j] += generic(gates_tup, gates_inv_tup, noise, N=n, **kwargs)
    if no_divide:
        return data
    else:
        return data / num_avg
    
def _m_generic_avg(q, *args, **kwargs):
    q.put(generic_avg(*args, **kwargs))

def _queue_get(queue, block=True, timeout=None):
    import errno
    while True:
        try:
            return queue.get(block, timeout)
        except IOError, e:
            if e.errno != errno.EINTR:
                raise
    
def multiprocess_generic(gates, gates_inv, noise, processes=6, num_avg=400, Nlist=range(5,100,5), **kwargs):
    ''' Run and average an RB sequence.
    params:
        gates, gates_inv, noise, **kwargs: see generic()
        num_avg, Nlist: see generic_avg()
        processes: number of processes to use.
    
    '''
    from multiprocessing import Process, Queue
    avg_list = [num_avg/processes] * processes
    avg_list[0] += num_avg%processes
    assert np.sum(avg_list) == num_avg
    procs = []
    queues = []
    data = np.zeros_like(Nlist, dtype=np.float64)
    gates_tup = map(tuplify, gates)
    gates_inv_tup = map(tuplify, gates_inv)
    kwargs['no_divide'] = True
    kwargs['Nlist'] = Nlist
    for i in range(processes):
        q = Queue()
        kw2 = kwargs.copy()
        kw2['num_avg'] = avg_list[i]
        p = Process(target=_m_generic_avg, args=[q, gates_tup, gates_inv_tup, noise], kwargs=kw2)
        p.start()
        procs.append(p)
        queues.append(q)
    for p in procs:
        p.join()
    for q in queues:
        data += _queue_get(q)
    return data / num_avg
    
def test_visitation(gates, N=100, check_inside_tuples=False):
    ''' test how many times we end up at each point on the octagon. '''
    gates = map(tuplify, gates)
    results = np.zeros(len(nps.octagon), dtype=float)
    def _count_state(state):
        for i, s in enumerate(nps.octagon):
            if nps.equal(s, state):
                results[i] += 1
                break
        else:
            raise ValueError("%s not on the octagon!" % state)
            
    state = nps.ground()
    for n in range(N):
        if not check_inside_tuples:
            state = nps.dotMany(random.choice(gates), state)
            _count_state(state)
        else:
            for g in random.choice(gates):
                state = np.dot(g, state)
                _count_state(state)

    results /= np.sum(results)
    print "fraction \t state"
    for i, s in enumerate(nps.octagon):
        print "%.5f \t %s" % (results[i], s)
    
compile_gates()
