#!/usr/bin/python

''' manual_test.py

Tests outside the python unittest framework.
'''

import numpy as np
import numpy_2qubit as np2, dephasing_funcs_2 as df2

pi = np.pi

ca, cb = np2.generate_cliffords()
ca_inv, cb_inv = np2.generate_cliffords(intersperse=-1)
ca10, cb10 = np2.generate_cliffords(intersperse=10)

round_level = 2

def _format_v(v):
    if np.isclose(v, 0):
        return '0'
    if np.isclose(v, 1):
        return '1'
    if np.isclose(v, -1):
        return '-1'
    return '{:.{width}f}'.format(v, width=round_level).rstrip('0')

def format_complex(v):
    ipart = _format_v(v.imag)
    rpart = _format_v(v.real)
    if ipart == '0':
        s = rpart
    elif rpart == '0':
        s = ipart + 'j'
    else:
        if ipart[0] == '-':
            s = rpart + ipart + 'j'
        else:
            s = rpart + '+' + ipart + 'j'
    return s

def pg(g):
    ''' pretty print a gate '''
    g = g.round(round_level)
    strs = []
    width = 0
    for row in g:
        strs.append([])
        for v in row:
            s = format_complex(v)
            strs[-1].append(s)
            width = max(width, len(s))
    for row in strs:
        for s in row:
            print '{:^{width}s}'.format(s, width=width) + ' ',
        print
        
def p(s):
    ''' pretty print a state '''
    labels = ['|00>', '|01>', '|10>', '|11>']
    st = []
    for l, v in zip(labels, s):
        st.append(format_complex(v) + ' ' + l)
    print ' + '.join(st)

def examine_gates(method=1, N=1, n_interspersed=10, correlated=False, noise=1e-4):
    noise = df2.generate_noise(correlated=correlated, dphi=noise)
    print "**** Examining gates ****"
    if method == 1:
        meas, i_list, ops_list = df2.rb_sequence(N=N, return_all_ops=True,
            noise_gates=noise)
    elif method == 2:
        meas, i_list, ops_list = df2.rb_sequence(N=N, return_all_ops=True,
            noise_after=False, noise_gates=noise)
    elif method == 3:
        ca, cb = np2.generate_cliffords(intersperse=n_interspersed)
        meas, i_list, ops_list = df2.rb_sequence(gates_a=ca, gates_b=cb, N=N,
            return_all_ops=True, noise_after=False, n_interspersed=n_interspersed,
            noise_gates=noise)
    print "Method %s (N=%s): " % (method, N)
    print "i_list = " + str(i_list)
    state = np2.ground()
    print "state: ",
    p(state)
    for g in ops_list:
        print '------------------------------------'
        print "GATE:"
        pg(g.copy())
        state = np.dot(g, state)
        print "STATE: ",
        p(state)
        
if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--method', help='which method to use (1,2,3)', type=int, default=3)
    parser.add_argument('-n', '--numGates', help='how many cliffords', type=int, default=1)
    parser.add_argument('-i', '--nInterspersed', help='for method 3, how many interspersings', type=int, default=10)
    parser.add_argument('-c', '--correlated', help='use correlated noise', action='store_true')
    parser.add_argument('-z', '--noise', help='noise magnitude', type=float, default=1e-4)
    args = parser.parse_args()
    examine_gates(method=args.method, N=args.numGates, noise=args.noise,
        n_interspersed=args.nInterspersed, correlated=args.correlated)
