#!/usr/bin/python
''' rb_dephasing.py

Basic estimates for dephasing during randomized benchmarking, based on
Korotkov and Veitia: "Error per gate in UCSB randomized benchmarking" (email)

'''

import random
import numpy as np
import dephasing_funcs_np as dfn, numpy_states as nps

pi = np.pi


def perr_vs_n(func, Nlist=range(1, 100, 10)):
    perr = np.zeros_like(np.array(Nlist), np.float64)    
    for i, N in enumerate(Nlist):
        perr[i] = func(N)
    return perr
        
def avg_perr(Nlist = range(10, 100, 10), num_avg = 20, dphi=0.01, plot=True,
             correlated=False, interspersed=False, cliffords=False):
    if isinstance(interspersed, bool) or not interspersed:
        inter_n = 10
    else:
        inter_n = int(interspersed)
    if inter_n == 1:
        interspersed = False
    # prep
    dfn.compile_gates(dphi=dphi, interspersed_n=inter_n)
    # which func do we use?
    if correlated:
        if cliffords:
            func = dfn.correlated_cliffords
        elif interspersed:
            func = dfn.correlated_interspersed
        else:
            func = dfn.correlated
    else:
        if cliffords:
            func = dfn.uncorrelated_cliffords
        elif interspersed:
            func = dfn.uncorrelated_interspersed
        else:
            func = dfn.uncorrelated
    # do the simulation and average
    perr = np.zeros_like(np.array(Nlist), np.float64)
    for i in range(num_avg):
        perr += perr_vs_n(func, Nlist)
    perr /= num_avg
    
    if plot:
        import matplotlib.pyplot as plt
        ax = plt.figure().add_subplot(111)
        ax.plot(Nlist, perr, 'o')
        plt.show()
        
    return perr

def compare_correlated_uncorrelated(Nlist, num_avg, dphi=0.01, plot=True,
                            interspersed=False, verbose=True, cliffords=False):
    ''' compares correlated and uncorrelated noise sequences.
    returns (uncorrelatedSlope, correlatedSlope) '''
    # these could be done in parallel, i suppose
    if verbose:
        print "Starting interspersed=%s" % interspersed
    uncorrelated = avg_perr(Nlist, num_avg, dphi, plot=False, correlated=False, interspersed=interspersed, cliffords=cliffords)
    slopeUncorr = fit_slope(Nlist, uncorrelated, dphi=dphi, guess=1.0/6.0)
    correlated = avg_perr(Nlist, num_avg, dphi, plot=False, correlated=True, interspersed=interspersed, cliffords=cliffords)
    slopeCorr = fit_slope(Nlist, correlated, dphi=dphi, guess=1.0/12.0)
    if verbose:
        print "  finished interspersed=%s" % interspersed
    if plot:
        x = np.linspace(0, np.max(Nlist))
        import matplotlib.pyplot as plt
        ax = plt.figure().add_subplot(111)
        ax.plot(Nlist, uncorrelated, 'bo', label="Uncorrelated")
        ax.plot(x, x*slopeUncorr*dphi**2, 'b-', label="Slope = $%.4f \, \Delta \phi^2$" % slopeUncorr)
        ax.plot(Nlist, correlated, 'ro', label="Correlated")
        ax.plot(x, x*slopeCorr*dphi**2, 'r-', label="Slope = $%.4f \, \Delta \phi^2$" % slopeCorr)         
        ax.set_xlabel("N gates", fontsize=19)
        ax.set_ylabel("Probability |1>", fontsize=19)
        ax.legend(numpoints=1, loc='upper left')
        ax.set_title("num_avg=%d, dphi=%f, interspersed=%s" % (num_avg, dphi, interspersed))
        plt.show()
    return slopeUncorr, slopeCorr
        
        
if __name__ == '__main__':
    import time, argparse
    t = time.time()
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--num_avg', help="number of averages for N",
            type=int, default=300)
    parser.add_argument('-P', '--processes', help="number of processes to use",
            type=int, default=4)
    parser.add_argument('-N', '--n_gates', help="max number of gates to apply",
            type=int, default=250)
    parser.add_argument('-s', '--n_step', 
            help="step size for N_gates, i.e. N = range(N_step. N_max, N_step)", 
            type=int, default=5)
    parser.add_argument('-g', '--phi_g', help="phi_g (should be << 1)",
            nargs='*', type=float, default=[0.0001])
    parser.add_argument('-p', '--plot', help="plot when done",
            action="store_true")
    parser.add_argument('-i', '--interspersings', help="N interspersing(s)",
            nargs='*', type=int, default=[1])
    parser.add_argument('-c', '--cliffords', help="use clifford gate set",
            action="store_true")
    parser.add_argument('-o', '--out', help="output file",
            type=argparse.FileType('w'))
    args = parser.parse_args()
    Nlist = np.array(range(args.n_step, args.n_gates, args.n_step), dtype=int)
    if len(args.interspersings) > 1:
        def f(i):
            return compare_correlated_uncorrelated(Nlist, args.num_avg,
                interspersed=i, dphi=args.phi_g[0], plot=False)
        if args.processes > 1:
            from multiprocessing import Pool
            pool = Pool(processes=args.processes)
            res = pool.map(f, args.interspersings)
        else:
            res = map(f, args.interspersings)
        uncorrelated_slopes = [x[0] for x in res]
        correlated_slopes = [x[1] for x in res]
        print res
        print "Total time (s): %s" % (time.time() - t)
        if args.plot:
            import matplotlib.pyplot as plt
            ax = plt.figure().add_subplot(111)
            ax.plot(args.interspersings, correlated_slopes, 'ro', label="Correlated")
            ax.plot(args.interspersings, uncorrelated_slopes, 'bo', label="Uncorrelated")
            ax.set_xlabel("N Interspersing", fontsize=19)
            ax.set_ylabel("Slope", fontsize=19)
            ax.legend(numpoints=1, loc='best')
            plt.show()
    elif len(args.phi_g) > 1:
        def f(i):
            return compare_correlated_uncorrelated(Nlist, args.num_avg,
                interspersed=args.interspersings[0], dphi=i, plot=False,
                cliffords=args.cliffords)
        if args.processes > 1:
            from multiprocessing import Pool
            pool = Pool(processes=args.processes)
            res = pool.map(f, args.phi_g)
        else:
            res = map(f, args.phi_g)
        uncorrelated_slopes = [x[0] for x in res]
        correlated_slopes = [x[1] for x in res]
        print res
        print "Total time (s): %s" % (time.time() - t)
        if args.plot:
            import matplotlib.pyplot as plt
            ax = plt.figure().add_subplot(111)
            ax.loglog(args.phi_g, correlated_slopes, 'ro', label="Correlated")
            ax.loglog(args.phi_g, uncorrelated_slopes, 'bo', label="Uncorrelated")
            ax.set_xlabel(r"$\varphi_{g}$", fontsize=19)
            ax.set_ylabel("Slope", fontsize=19)
            ax.legend(numpoints=1, loc='best')
            plt.grid(which="both")
            plt.show()
            
    else:
        res = compare_correlated_uncorrelated(Nlist, args.num_avg, 
                            interspersed=args.interspersings[0], dphi=args.phi_g[0],
                            plot=args.plot, cliffords=args.cliffords)
        print res

