#!/usr/bin/python

# tests for the dephasing simulations

import unittest, random
import numpy as np
import numpy_states as nps
import dephasing_funcs_np as dfn

pi = np.pi
ALMOST_EQUAL_PLACES = 14    # number of decimal places for almost equal

class TestRBSimulation(unittest.TestCase):
    
    def assertEqualStates(self, a, b, msg=None):
        if not nps.equal(a, b):
            msg = msg or "%s does not equal %s" % (a, b)
            raise self.failureException(msg)
            
    def assertNotEqualStates(self, a, b, msg=None):
        if nps.equal(a, b):
            msg = msg or "%s == %s" % (a, b)
            raise self.failureException(msg)
        
    def assertStateIn(self, state, stateList, msg=None):
        for s in stateList:
            if nps.equal(state, s):
                break
        else:
            msg = msg or "%s not in %s" % (state, stateList)
            raise self.failureException(msg)
            
    
    def setUp(self):
        self.addTypeEqualityFunc(type(nps.ground()), self.assertEqualStates)
        self.pis = [nps.rotx(pi), nps.roty(pi), nps.rotz(pi)]
        self.pi2s = [nps.rotx(pi/2), nps.roty(pi/2), nps.rotz(pi/2)]
        self.places = ALMOST_EQUAL_PLACES

    def test_measurement(self):
        self.assertEqual(nps.measure(nps.excited()), 1.0)
        self.assertEqual(nps.measure(nps.ground()), 0.0)
        self.assertAlmostEqual(nps.measure(np.dot(self.pi2s[0], nps.ground())), 0.5, places=self.places)
        self.assertAlmostEqual(nps.measure(np.dot(self.pi2s[1], nps.ground())), 0.5, places=self.places)
        self.assertTrue(nps.equal(nps.ground(), nps.ground()))
        self.assertFalse(nps.equal(nps.ground(), np.dot(self.pi2s[0], nps.ground())))
        self.assertTrue(nps.opposite(nps.ground(), nps.excited()))
        self.assertFalse(nps.opposite(nps.ground(), np.dot(self.pi2s[0], nps.ground())))

    # each rotation puts qubit where expected
    def test_rotations(self):
        self.assertEqualStates(np.dot(self.pis[0], nps.ground()), nps.excited())
        self.assertEqualStates(np.dot(self.pis[1], nps.excited()), nps.ground())
        for gate, dest in zip(self.pi2s[:2], nps.equator):
            self.assertEqual(np.dot(gate, nps.ground()), dest)
        self.assertEqual(np.dot(self.pis[2], nps.equator[1]), nps.equator[2])
        self.assertEqual(np.dot(self.pi2s[2], nps.equator[1]), nps.equator[3])
        
    def test_octagon(self):
        self.assertEqual(len(nps.equator), 4)
        self.assertEqual(len(nps.octagon), 6)
        for s in nps.equator:
            orig = np.copy(s)
            s = np.dot(self.pi2s[2], s)
            self.assertStateIn(s, nps.equator)
            self.assertNotEqualStates(s, orig)
            
        

    # qubit always end up on poles or equators
    def test_rotation_chain(self):
        for i in range(10):
            state = nps.ground()
            for j in range(10):
                state = np.dot(random.choice(self.pis+self.pi2s), state)
            self.assertAlmostEqual(nps.norm(state), 1.0, places=self.places)
            self.assertStateIn(state, nps.octagon)
            
    # now test the dephasing functions
    def test_dephasing_gates(self):
        dfn.compile_gates(dphi=0)
        for g in dfn.compiled_gates + dfn.inverted_gates:
            s = np.dot(g, random.choice(nps.octagon))
            self.assertStateIn(s, nps.octagon)
        for gg in dfn.cliffords + dfn.cliffords_inverted:
            s = random.choice(nps.octagon)
            for g in gg:
                s = np.dot(g, s)
            self.assertStateIn(s, nps.octagon)
    
    def test_inversions(self):
        dfn.compile_gates(dphi=0)
        for g, gi in zip(dfn.compiled_gates, dfn.inverted_gates):
            s = random.choice(nps.octagon)
            orig = np.copy(s)
            s = np.dot(g, s)
            self.assertEqual(np.dot(gi, s), orig)
        for gg, gi in zip(dfn.cliffords, dfn.cliffords_inverted):
            s = random.choice(nps.octagon)
            orig = np.copy(s)
            for g in gg:
                s = np.dot(g, s)
            for g in gi[::-1]:
                s = np.dot(g, s)
            self.assertEqual(s, orig)

    def test_interspersed_gates(self):
        nint = 10
        dfn.compile_gates(dphi=0, interspersed_n=nint)
        for g, gint in zip(dfn.compiled_gates, dfn.compiled_gates_interspersed):
            s = random.choice(nps.octagon)
            s2 = np.copy(s)
            for i in range(nint):
                s = np.dot(gint, s)
            self.assertEqual(s, np.dot(g, s2))
        for gg, ggint in zip(dfn.cliffords, dfn.cliffords_interspersed):
            s = random.choice(nps.octagon)
            s2 = np.copy(s)
            for g, gint in zip(gg, ggint):
                for i in range(nint):
                    s = np.dot(gint, s)
                s2 = np.dot(g, s2)
            self.assertEqual(s, s2)
            
    def test_dephasing_noiseless(self):
        dfn.compile_gates(dphi=0)
        self.assertAlmostEqual(dfn.uncorrelated(), 0, places=self.places)
        self.assertAlmostEqual(dfn.correlated(), 0, places=self.places)
        self.assertAlmostEqual(dfn.correlated_interspersed(), 0, places=self.places)
        self.assertAlmostEqual(dfn.uncorrelated_interspersed(), 0, places=self.places)
        self.assertAlmostEqual(dfn.correlated_cliffords(), 0, places=self.places)
        self.assertAlmostEqual(dfn.uncorrelated_cliffords(), 0, places=self.places)
        


if __name__ == '__main__':
    unittest.main()
